/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#include "attr.h"

/* -- Begin Profiling Symbol Block for routine MPI_Win_set_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Win_set_attr = PMPI_Win_set_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Win_set_attr  MPI_Win_set_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Win_set_attr as PMPI_Win_set_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_set_attr
#define MPI_Win_set_attr PMPI_Win_set_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_set_attr

/*@
MPI_Win_set_attr - Stores attribute value associated with a key

Input Parameters:
+ win - MPI window object to which attribute will be attached (handle) 
. keyval - key value, as returned by  'MPI_Win_create_keyval' (integer) 
- attribute_val - attribute value 

Notes:

The type of the attribute value depends on whether C or Fortran is being used.
In C, an attribute value is a pointer ('void *'); in Fortran, it is an 
address-sized integer.

If an attribute is already present, the delete function (specified when the
corresponding keyval was created) will be called.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_KEYVAL
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_set_attr(MPI_Win win, int win_keyval, void *attribute_val)
{
	static const wchar_t FCNAME[] = L"MPI_Win_set_attr";
	int mpi_errno = MPI_SUCCESS;

	MPID_Win *win_ptr = NULL;
	MPID_Keyval *keyval_ptr = NULL;
	MPID_Attribute *p, **old_p;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_SET_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	/* The thread lock prevents a valid attr delete on the same window
	but in a different thread from causing problems */
	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_SET_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			MPIR_ERRTEST_KEYVAL(win_keyval, MPID_WIN, L"window", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(win_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	MPID_Keyval_get_ptr( win_keyval, keyval_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */
			/* Validate keyval */
			MPID_Keyval_valid_ptr( keyval_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_set_attr %W %d %p", win, win_keyval, attribute_val); }

	/* Look for attribute.  They are ordered by keyval handle.  This uses 
	a simple linear list algorithm because few applications use more than a 
	handful of attributes */

	old_p = &win_ptr->attributes;
	p = win_ptr->attributes;
	while (p)
	{
		if (p->keyval->handle == keyval_ptr->handle)
		{
			/* If found, call the delete function before replacing the 
			attribute */
			mpi_errno = MPIR_Call_attr_delete( win, p );
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno)
			{
				/* FIXME : communicator of window? */
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			p->value = attribute_val;
			/* Does not change the reference count on the keyval */
			break;
		}
		else
		{
			if (p->keyval->handle > keyval_ptr->handle)
			{
				MPID_Attribute *new_p = (MPID_Attribute *)MPIU_Handle_obj_alloc( &MPID_Attr_mem );
				MPIU_ERR_CHKANDJUMP1(!new_p,mpi_errno,MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPID_Attribute" );
				new_p->keyval		 = keyval_ptr;
				new_p->pre_sentinal		 = 0;
				new_p->value		 = attribute_val;
				new_p->post_sentinal = 0;
				new_p->next				 = p->next;
				MPIU_Object_add_ref( keyval_ptr );
				p->next				 = new_p;
				break;
			}
		}
		old_p = &p->next;
		p = p->next;
	}
	if (!p)
	{
		MPID_Attribute *new_p = (MPID_Attribute *)MPIU_Handle_obj_alloc( &MPID_Attr_mem );
		MPIU_ERR_CHKANDJUMP1(!new_p,mpi_errno,MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPID_Attribute" );
		/* Did not find in list.  Add at end */
		new_p->keyval			 = keyval_ptr;
		new_p->pre_sentinal  = 0;
		new_p->value			 = attribute_val;
		new_p->post_sentinal = 0;
		new_p->next			 = 0;
		MPIU_Object_add_ref( keyval_ptr );
		*old_p					 = new_p;
	}

	/* Here is where we could add a hook for the device to detect attribute
	value changes, using something like
	MPID_Dev_win_attr_hook( win_ptr, keyval, attribute_val );
	*/

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_SET_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr"); 
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_set_attr", 
			L"**mpi_win_set_attr %W %d %p", win, win_keyval, attribute_val);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Win_get_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Win_get_attr = PMPI_Win_get_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Win_get_attr  MPI_Win_get_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Win_get_attr as PMPI_Win_get_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_get_attr
#define MPI_Win_get_attr PMPI_Win_get_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_get_attr

/*@
MPI_Win_get_attr - Get attribute cached on an MPI window object

Input Parameters:
+ win - window to which the attribute is attached (handle) 
- win_keyval - key value (integer) 

Output Parameters:
+ attribute_val - attribute value, unless flag is false 
- flag - false if no attribute is associated with the key (logical) 

Notes:
The following attributes are predefined for all MPI Window objects\:

+ MPI_WIN_BASE - window base address. 
. MPI_WIN_SIZE - window size, in bytes. 
- MPI_WIN_DISP_UNIT - displacement unit associated with the window. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_KEYVAL
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_attr(MPI_Win win, int win_keyval, void *attribute_val, int *flag)
{
	static const wchar_t FCNAME[] = L"MPI_Win_get_attr";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_GET_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_GET_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			MPIR_ERRTEST_KEYVAL(win_keyval, MPID_WIN, L"window", mpi_errno);
#		   ifdef NEEDS_POINTER_ALIGNMENT_ADJUST
			/* A common user error is to pass the address of a 4-byte
			int when the address of a pointer (or an address-sized int)
			should have been used.  We can test for this specific
			case.  Note that this code assumes sizeof(MPI_Aint) is 
			a power of 2. */
			if ((MPI_Aint)attribute_val & (sizeof(MPI_Aint)-1))
			{
				MPIU_ERR_SET(mpi_errno,MPI_ERR_ARG,L"**attrnotptr");
			}
#		   endif
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(attribute_val, L"attribute_val", mpi_errno);
			MPIR_ERRTEST_ARGNULL(flag, L"flag", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */	

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_get_attr %W %d %p %p", win, win_keyval, attribute_val, flag); }

	/* Check for builtin attribute */
	/* This code is ok for correct programs, but it would be better
	to copy the values from the per-process block and pass the user
	a pointer to a copy */
	/* Note that if we are called from Fortran, we must return the values,
	not the addresses, of these attributes */
	if (HANDLE_GET_KIND(win_keyval) == HANDLE_KIND_BUILTIN)
	{
		int attr_idx = win_keyval & 0x0000000f;
		void **attr_val_p = (void **)attribute_val;
#ifdef HAVE_FORTRAN_BINDING
		/* Note that this routine only has a Fortran 90 binding,
		so the attribute value is an address-sized int */
		MPI_Aint  *attr_int = (MPI_Aint *)attribute_val;
#endif
		*flag = 1;

		/* 
		* The C versions of the attributes return the address of a 
		* *COPY* of the value (to prevent the user from changing it)
		* and the Fortran versions provide the actual value (as an Fint)
		*/
		switch (attr_idx)
		{
		case 1: /* WIN_BASE */
			*attr_val_p = win_ptr->base;
			break;
		case 3: /* SIZE */
			win_ptr->copySize = win_ptr->size;
			*attr_val_p = &win_ptr->copySize;
			break;
		case 5: /* DISP_UNIT */
			win_ptr->copyDispUnit = win_ptr->disp_unit;
			*attr_val_p = &win_ptr->copyDispUnit;
			break;
#ifdef HAVE_FORTRAN_BINDING
		case 2: /* Fortran BASE */
			/* The Fortran routine that matches this routine should
			provide an address-sized integer, not an MPI_Fint */
			*attr_int = MPIU_PtrToAint(win_ptr->base);
			break;
		case 4: /* Fortran SIZE */
			/* We do not need to copy because we return the value,
			not a pointer to the value */
			*attr_int = win_ptr->size;
			break;
		case 6: /* Fortran DISP_UNIT */
			/* We do not need to copy because we return the value,
			not a pointer to the value */
			*attr_int = win_ptr->disp_unit;
			break;
#endif
		}
	}
	else
	{
		MPID_Attribute *p = win_ptr->attributes;

		*flag = 0;
		while (p)
		{
			if (p->keyval->handle == win_keyval)
			{
				*flag = 1;
				(*(void **)attribute_val) = p->value;
				break;
			}
			p = p->next;
		}
	}

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_GET_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_get_attr", 
			L"**mpi_win_get_attr %W %d %p %p", win, win_keyval, attribute_val, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Win_free_keyval */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Win_free_keyval = PMPI_Win_free_keyval
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Win_free_keyval  MPI_Win_free_keyval
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Win_free_keyval as PMPI_Win_free_keyval
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_free_keyval
#define MPI_Win_free_keyval PMPI_Win_free_keyval

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_free_keyval

/*@
MPI_Win_free_keyval - Frees an attribute key for MPI RMA windows

Input Parameter:
. win_keyval - key value (integer) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER
.N MPI_ERR_KEYVAL
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_free_keyval(int *win_keyval)
{
	static const wchar_t FCNAME[] = L"MPI_Win_free_keyval";
	int mpi_errno = MPI_SUCCESS;
	MPID_Keyval *keyval_ptr = NULL;
	int		  in_use;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_FREE_KEYVAL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_FREE_KEYVAL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(*win_keyval, L"win_keyval", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			MPIR_ERRTEST_KEYVAL(*win_keyval, MPID_WIN, L"window", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(*win_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Keyval_get_ptr( *win_keyval, keyval_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Keyval_valid_ptr( keyval_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_free_keyval %p", win_keyval); }

	MPIU_Object_release_ref( keyval_ptr, &in_use);
	if (!in_use)
	{
		MPIU_Handle_obj_free( &MPID_Keyval_mem, keyval_ptr );
	}
	*win_keyval = MPI_KEYVAL_INVALID;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_FREE_KEYVAL);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_free_keyval", 
			L"**mpi_win_free_keyval %p", win_keyval);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Win_delete_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Win_delete_attr = PMPI_Win_delete_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Win_delete_attr  MPI_Win_delete_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Win_delete_attr as PMPI_Win_delete_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_delete_attr
#define MPI_Win_delete_attr PMPI_Win_delete_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_delete_attr

/*@
MPI_Win_delete_attr - Deletes an attribute value associated with a key on 
a datatype

Input Parameters:
+ win - window from which the attribute is deleted (handle) 
- win_keyval - key value (integer) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_KEYVAL
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_delete_attr(MPI_Win win, int win_keyval)
{
	static const wchar_t FCNAME[] = L"MPI_Win_delete_attr";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_Attribute *p, **old_p;
	MPID_Keyval *keyval_ptr=0;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_DELETE_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	/* The thread lock prevents a valid attr delete on the same window
	but in a different thread from causing problems */
	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_DELETE_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			MPIR_ERRTEST_KEYVAL(win_keyval, MPID_WIN, L"window", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(win_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	MPID_Keyval_get_ptr( win_keyval, keyval_ptr );

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */
			/* Validate keyval_ptr */
			MPID_Keyval_valid_ptr( keyval_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_delete_attr %W %d", win, win_keyval); }

	/* Look for attribute.  They are ordered by keyval handle */

	old_p = &win_ptr->attributes;
	p	 = win_ptr->attributes;
	while (p)
	{
		if (p->keyval->handle == keyval_ptr->handle)
		{
			break;
		}
		old_p = &p->next;
		p = p->next;
	}

	/* We can't unlock yet, because we must not free the attribute until
	we know whether the delete function has returned with a 0 status
	code */

	if (p)
	{
		/* Run the delete function, if any, and then free the attribute 
		storage */
		mpi_errno = MPIR_Call_attr_delete( win, p );

		/* --BEGIN ERROR HANDLING-- */
		if (!mpi_errno)
		{
			int in_use;
			/* We found the attribute.  Remove it from the list */
			*old_p = p->next;
			/* Decrement the use of the keyval */
			MPIU_Object_release_ref( p->keyval, &in_use);
			if (!in_use)
			{
				MPIU_Handle_obj_free( &MPID_Keyval_mem, p->keyval );
			}
			MPID_Attr_free(p);
		}
		/* --END ERROR HANDLING-- */
	}

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_DELETE_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_delete_attr", 
			L"**mpi_win_delete_attr %W %d", win, win_keyval);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Win_create_keyval */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Win_create_keyval = PMPI_Win_create_keyval
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Win_create_keyval  MPI_Win_create_keyval
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Win_create_keyval as PMPI_Win_create_keyval
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_create_keyval
#define MPI_Win_create_keyval PMPI_Win_create_keyval

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_create_keyval

/*@
MPI_Win_create_keyval - Create an attribute keyval for MPI window objects

Input Parameters:
+ win_copy_attr_fn - copy callback function for win_keyval (function) 
. win_delete_attr_fn - delete callback function for win_keyval (function) 
- extra_state - extra state for callback functions 

Output Parameter:
. win_keyval - key value for future access (integer) 

Notes:
Default copy and delete functions are available.  These are
+ MPI_WIN_NULL_COPY_FN   - empty copy function
. MPI_WIN_NULL_DELETE_FN - empty delete function
- MPI_WIN_DUP_FN		 - simple dup function

.N AttrErrReturn

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_create_keyval(MPI_Win_copy_attr_function *win_copy_attr_fn, 
												MPI_Win_delete_attr_function *win_delete_attr_fn, 
												int *win_keyval, void *extra_state)
{
	static const wchar_t FCNAME[] = L"MPI_Win_create_keyval";
	int mpi_errno = MPI_SUCCESS;
	MPID_Keyval *keyval_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_CREATE_KEYVAL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_CREATE_KEYVAL);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(win_keyval,L"win_keyval",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_create_keyval %p %p %p %p", win_copy_attr_fn, win_delete_attr_fn, win_keyval, extra_state); }

	keyval_ptr = (MPID_Keyval *)MPIU_Handle_obj_alloc( &MPID_Keyval_mem );
	MPIU_ERR_CHKANDJUMP1(!keyval_ptr,mpi_errno,MPI_ERR_OTHER,L"**nomem", L"**nomem %s", L"MPID_Keyval" );
	/* Initialize the attribute dup function */
	if (!MPIR_Process.attr_dup)
	{
		MPIR_Process.attr_dup  = MPIR_Attr_dup_list;
		MPIR_Process.attr_free = MPIR_Attr_delete_list;
	}

	/* The handle encodes the keyval kind.  Modify it to have the correct field */
	keyval_ptr->handle = (keyval_ptr->handle & ~(0x03c00000)) | (MPID_WIN << 22);
	*win_keyval = keyval_ptr->handle;
	MPIU_Object_set_ref(keyval_ptr,1);
	keyval_ptr->language = MPID_LANG_C;
	keyval_ptr->kind = MPID_WIN;
	keyval_ptr->extra_state = extra_state;
	keyval_ptr->copyfn.C_CopyFunction = win_copy_attr_fn;
	keyval_ptr->delfn.C_DeleteFunction = win_delete_attr_fn;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_CREATE_KEYVAL);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_create_keyval",
			L"**mpi_win_create_keyval %p %p %p %p", win_copy_attr_fn, win_delete_attr_fn, win_keyval, extra_state);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Type_set_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Type_set_attr = PMPI_Type_set_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Type_set_attr  MPI_Type_set_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Type_set_attr as PMPI_Type_set_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_set_attr
#define MPI_Type_set_attr PMPI_Type_set_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_set_attr

/*@
MPI_Type_set_attr - Stores attribute value associated with a key

Input Parameters:
+ type - MPI Datatype to which attribute will be attached (handle) 
. keyval - key value, as returned by  'MPI_Type_create_keyval' (integer) 
- attribute_val - attribute value 

Notes:

The type of the attribute value depends on whether C or Fortran is being used.
In C, an attribute value is a pointer ('void *'); in Fortran, it is an 
address-sized integer.

If an attribute is already present, the delete function (specified when the
corresponding keyval was created) will be called.
.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_KEYVAL
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_set_attr(MPI_Datatype type, int type_keyval, void *attribute_val)
{
	static const wchar_t FCNAME[] = L"MPI_Type_set_attr";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *type_ptr = NULL;
	MPID_Keyval *keyval_ptr = NULL;
	MPID_Attribute *p, **old_p;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_SET_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	/* The thread lock prevents a valid attr delete on the same datatype
	but in a different thread from causing problems */
	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_SET_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(type, L"datatype", mpi_errno);
			MPIR_ERRTEST_KEYVAL(type_keyval, MPID_DATATYPE, L"datatype", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(type_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( type, type_ptr );
	MPID_Keyval_get_ptr( type_keyval, keyval_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate type_ptr */
			MPID_Datatype_valid_ptr( type_ptr, mpi_errno );
			/* If type_ptr is not valid, it will be reset to null */
			/* Validate keyval_ptr */
			MPID_Keyval_valid_ptr( keyval_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_set_attr %D %d %p", type, type_keyval, attribute_val); }

	/* Look for attribute.  They are ordered by keyval handle.  This uses 
	a simple linear list algorithm because few applications use more than a 
	handful of attributes */

	old_p = &type_ptr->attributes;
	p = type_ptr->attributes;
	while (p)
	{
		if (p->keyval->handle == keyval_ptr->handle)
		{
			/* If found, call the delete function before replacing the attribute */
			mpi_errno = MPIR_Call_attr_delete( type, p );
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno)
			{
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			p->value = attribute_val;
			break;
		}
		else
		{
			if (p->keyval->handle > keyval_ptr->handle)
			{
				MPID_Attribute *new_p = (MPID_Attribute *)MPIU_Handle_obj_alloc( &MPID_Attr_mem );
				MPIU_ERR_CHKANDJUMP1(!new_p,mpi_errno,MPI_ERR_OTHER, L"**nomem",L"**nomem %s", L"MPID_Attribute" );
				new_p->keyval		 = keyval_ptr;
				new_p->pre_sentinal		 = 0;
				new_p->value		 = attribute_val;
				new_p->post_sentinal = 0;
				new_p->next				 = p->next;
				MPIU_Object_add_ref( keyval_ptr );
				p->next				 = new_p;
				break;
			}
		}
		old_p = &p->next;
		p = p->next;
	}
	if (!p)
	{
		MPID_Attribute *new_p = (MPID_Attribute *)MPIU_Handle_obj_alloc( &MPID_Attr_mem );
		MPIU_ERR_CHKANDJUMP1(!new_p,mpi_errno,MPI_ERR_OTHER, L"**nomem",L"**nomem %s", L"MPID_Attribute" );
		/* Did not find in list.  Add at end */
		new_p->keyval			 = keyval_ptr;
		new_p->pre_sentinal  = 0;
		new_p->value			 = attribute_val;
		new_p->post_sentinal = 0;
		new_p->next			 = 0;
		MPIU_Object_add_ref( keyval_ptr );
		*old_p					 = new_p;
	}

	/* Here is where we could add a hook for the device to detect attribute
	value changes, using something like
	MPID_Dev_type_attr_hook( type_ptr, keyval, attribute_val );
	*/

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_SET_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_set_attr", 
			L"**mpi_type_set_attr %D %d %p", type, type_keyval, attribute_val);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Type_get_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Type_get_attr = PMPI_Type_get_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Type_get_attr  MPI_Type_get_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Type_get_attr as PMPI_Type_get_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_get_attr
#define MPI_Type_get_attr PMPI_Type_get_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_get_attr

/*@
MPI_Type_get_attr - Retrieves attribute value by key

Input Parameters:
+ type - datatype to which the attribute is attached (handle) 
- type_keyval - key value (integer) 

Output Parameters:
+ attribute_val - attribute value, unless flag = false 
- flag - false if no attribute is associated with the key (logical) 

Notes:
Attributes must be extracted from the same language as they were inserted  
in with 'MPI_Type_set_attr'.  The notes for C and Fortran below explain 
why. 

Notes for C:
Even though the 'attr_value' arguement is declared as 'void *', it is
really the address of a void pointer.  See the rationale in the 
standard for more details. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_KEYVAL
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_attr(MPI_Datatype type, int type_keyval, void *attribute_val, int *flag)
{
	static const wchar_t FCNAME[] = L"MPI_Type_get_attr";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *type_ptr = NULL;
	MPID_Attribute *p;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_GET_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_GET_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(type, L"datatype", mpi_errno);
			MPIR_ERRTEST_KEYVAL(type_keyval, MPID_DATATYPE, L"datatype", mpi_errno);
#		   ifdef NEEDS_POINTER_ALIGNMENT_ADJUST
			/* A common user error is to pass the address of a 4-byte
			int when the address of a pointer (or an address-sized int)
			should have been used.  We can test for this specific
			case.  Note that this code assumes sizeof(MPI_Aint) is 
			a power of 2. */
			if ((MPI_Aint)attribute_val & (sizeof(MPI_Aint)-1))
			{
				MPIU_ERR_SET(mpi_errno,MPI_ERR_ARG,L"**attrnotptr");
			}
#		   endif
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( type, type_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype pointer */
			MPID_Datatype_valid_ptr( type_ptr, mpi_errno );
			/* If type_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_get_attr %D %d %p %p", type, type_keyval, attribute_val, flag); }

	*flag = 0;
	p = type_ptr->attributes;
	while (p)
	{
		if (p->keyval->handle == type_keyval)
		{
			*flag = 1;
			(*(void **)attribute_val) = p->value;
			break;
		}
		p = p->next;
	}

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_GET_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_get_attr",
			L"**mpi_type_get_attr %D %d %p %p", type, type_keyval, attribute_val, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Type_free_keyval */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Type_free_keyval = PMPI_Type_free_keyval
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Type_free_keyval  MPI_Type_free_keyval
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Type_free_keyval as PMPI_Type_free_keyval
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_free_keyval
#define MPI_Type_free_keyval PMPI_Type_free_keyval

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_free_keyval

/*@
MPI_Type_free_keyval - Frees an attribute key for datatypes

Input Parameter:
. keyval - Frees the integer key value (integer) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
.N MPI_ERR_KEYVAL
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_free_keyval(int *type_keyval)
{
	static const wchar_t FCNAME[] = L"MPI_Type_free_keyval";
	MPID_Keyval *keyval_ptr = NULL;
	int		  in_use;
	int		  mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_FREE_KEYVAL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_FREE_KEYVAL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(type_keyval, L"type_keyval", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			MPIR_ERRTEST_KEYVAL(*type_keyval, MPID_DATATYPE, L"datatype", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(*type_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Keyval_get_ptr( *type_keyval, keyval_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Keyval_valid_ptr( keyval_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_free_keyval %p", type_keyval); }

	MPIU_Object_release_ref( keyval_ptr, &in_use);
	if (!in_use)
	{
		MPIU_Handle_obj_free( &MPID_Keyval_mem, keyval_ptr );
	}
	*type_keyval = MPI_KEYVAL_INVALID;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_FREE_KEYVAL);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_free_keyval", 
			L"**mpi_type_free_keyval %p", type_keyval);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Type_delete_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Type_delete_attr = PMPI_Type_delete_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Type_delete_attr  MPI_Type_delete_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Type_delete_attr as PMPI_Type_delete_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_delete_attr
#define MPI_Type_delete_attr PMPI_Type_delete_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_delete_attr

/*@
MPI_Type_delete_attr - Deletes an attribute value associated with a key on 
a datatype

Input Parameters:
+  type - MPI datatype to which attribute is attached (handle)
-  type_keyval - The key value of the deleted attribute (integer) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
.N MPI_ERR_KEYVAL
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_delete_attr(MPI_Datatype type, int type_keyval)
{
	static const wchar_t FCNAME[] = L"MPI_Type_delete_attr";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *type_ptr = NULL;
	MPID_Attribute *p, **old_p;
	MPID_Keyval *keyval_ptr = 0;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_DELETE_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	/* The thread lock prevents a valid attr delete on the same datatype
	but in a different thread from causing problems */
	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_DELETE_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(type, L"datatype", mpi_errno);
			MPIR_ERRTEST_KEYVAL(type_keyval, MPID_DATATYPE, L"datatype", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(type_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Validate parameters and objects (post conversion) */
	MPID_Datatype_get_ptr( type, type_ptr );
	MPID_Keyval_get_ptr( type_keyval, keyval_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate type_ptr */
			MPID_Datatype_valid_ptr( type_ptr, mpi_errno );
			/* If type_ptr is not valid, it will be reset to null */
			/* Validate keyval_ptr */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_delete_attr %D %d", type, type_keyval); }

	/* Look for attribute.  They are ordered by keyval handle */

	old_p = &type_ptr->attributes;
	p	 = type_ptr->attributes;
	while (p)
	{
		if (p->keyval->handle == keyval_ptr->handle)
		{
			break;
		}
		old_p = &p->next;
		p	 = p->next;
	}

	/* We can't unlock yet, because we must not free the attribute until
	we know whether the delete function has returned with a 0 status
	code */

	if (p)
	{
		/* Run the delete function, if any, and then free the attribute 
		storage */
		mpi_errno = MPIR_Call_attr_delete( type, p );

		/* --BEGIN ERROR HANDLING-- */
		if (!mpi_errno)
		{
			int in_use;
			/* We found the attribute.  Remove it from the list */
			*old_p = p->next;
			/* Decrement the use of the keyval */
			MPIU_Object_release_ref( p->keyval, &in_use);
			if (!in_use)
			{
				MPIU_Handle_obj_free( &MPID_Keyval_mem, p->keyval );
			}
			MPID_Attr_free(p);
		}
		/* --END ERROR HANDLING-- */
	}

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_DELETE_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER,
			L"**mpi_type_delete_attr",
			L"**mpi_type_delete_attr %D %d", type, type_keyval);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Type_create_keyval */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Type_create_keyval = PMPI_Type_create_keyval
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Type_create_keyval  MPI_Type_create_keyval
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Type_create_keyval as PMPI_Type_create_keyval
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_create_keyval
#define MPI_Type_create_keyval PMPI_Type_create_keyval

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_create_keyval

/*@
MPI_Type_create_keyval - Create an attribute keyval for MPI datatypes

Input Parameters:
+ type_copy_attr_fn - copy callback function for type_keyval (function) 
. type_delete_attr_fn - delete callback function for type_keyval (function) 
- extra_state - extra state for callback functions 

Output Parameter:
. type_keyval - key value for future access (integer) 

Notes:

Default copy and delete functions are available.  These are
+ MPI_TYPE_NULL_COPY_FN   - empty copy function
. MPI_TYPE_NULL_DELETE_FN - empty delete function
- MPI_TYPE_DUP_FN		 - simple dup function

.N AttrErrReturn

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_keyval(MPI_Type_copy_attr_function *type_copy_attr_fn, 
												 MPI_Type_delete_attr_function *type_delete_attr_fn,
												 int *type_keyval, void *extra_state)
{
	static const wchar_t FCNAME[] = L"MPI_Type_create_keyval";
	int mpi_errno = MPI_SUCCESS;
	MPID_Keyval *keyval_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_KEYVAL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CREATE_KEYVAL);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(type_keyval,L"type_keyval",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_create_keyval %p %p %p %p", type_copy_attr_fn, type_delete_attr_fn, type_keyval, extra_state); }

	keyval_ptr = (MPID_Keyval *)MPIU_Handle_obj_alloc( &MPID_Keyval_mem );
	MPIU_ERR_CHKANDJUMP(!keyval_ptr,mpi_errno,MPI_ERR_OTHER,L"**nomem");

	/* Initialize the attribute dup function */
	if (!MPIR_Process.attr_dup)
	{
		MPIR_Process.attr_dup  = MPIR_Attr_dup_list;
		MPIR_Process.attr_free = MPIR_Attr_delete_list;
	}

	/* The handle encodes the keyval kind.  Modify it to have the correct field */
	keyval_ptr->handle = (keyval_ptr->handle & ~(0x03c00000)) | (MPID_DATATYPE << 22);
	*type_keyval = keyval_ptr->handle;
	MPIU_Object_set_ref(keyval_ptr,1);
	keyval_ptr->language = MPID_LANG_C;
	keyval_ptr->kind = MPID_DATATYPE;
	keyval_ptr->extra_state = extra_state;
	keyval_ptr->copyfn.C_CopyFunction = type_copy_attr_fn;
	keyval_ptr->delfn.C_DeleteFunction = type_delete_attr_fn;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CREATE_KEYVAL);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_create_keyval",
			L"**mpi_type_create_keyval %p %p %p %p", type_copy_attr_fn, type_delete_attr_fn, type_keyval, extra_state);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Keyval_free */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Keyval_free = PMPI_Keyval_free
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Keyval_free  MPI_Keyval_free
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Keyval_free as PMPI_Keyval_free
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Keyval_free
#define MPI_Keyval_free PMPI_Keyval_free

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Keyval_free

/*@

MPI_Keyval_free - Frees an attribute key for communicators

Input Parameter:
. keyval - Frees the integer key value (integer) 

Note:
Key values are global (they can be used with any and all communicators)

.N Deprecated
The replacement for this routine is 'MPI_Comm_free_keyval'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_PERM_KEY

.seealso: MPI_Keyval_create, MPI_Comm_free_keyval
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Keyval_free(int *keyval)
{
	static const wchar_t FCNAME[] = L"MPI_Keyval_free";
	int mpi_errno = MPI_SUCCESS;
	MPIU_THREADPRIV_DECL;									   \
		MPID_MPI_STATE_DECL(MPID_STATE_MPI_KEYVAL_FREE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_KEYVAL_FREE);
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(keyval, L"keyval", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_keyval_free %p", keyval); }

	MPIU_THREADPRIV_GET;
	NMPI_Comm_free_keyval(mpi_errno, ( keyval ));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

fn_exit: 
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_KEYVAL_FREE);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_keyval_free", L"**mpi_keyval_free %p", keyval);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Keyval_create */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Keyval_create = PMPI_Keyval_create
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Keyval_create  MPI_Keyval_create
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Keyval_create as PMPI_Keyval_create
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Keyval_create
#define MPI_Keyval_create PMPI_Keyval_create

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Keyval_create

/*@

MPI_Keyval_create - Greates a new attribute key

Input Parameters:
+ copy_fn - Copy callback function for 'keyval' 
. delete_fn - Delete callback function for 'keyval' 
- extra_state - Extra state for callback functions 

Output Parameter:
. keyval - key value for future access (integer) 

Notes:
Key values are global (available for any and all communicators).

There are subtle differences between C and Fortran that require that the
copy_fn be written in the same language that 'MPI_Keyval_create'
is called from.
This should not be a problem for most users; only programers using both 
Fortran and C in the same program need to be sure that they follow this rule.

.N ThreadSafe

.N Deprecated
The replacement for this routine is 'MPI_Comm_create_keyval'.

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_ARG

.seealso  MPI_Keyval_free, MPI_Comm_create_keyval
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Keyval_create(MPI_Copy_function *copy_fn, 
											MPI_Delete_function *delete_fn, 
											int *keyval, void *extra_state)
{
	static const wchar_t FCNAME[] = L"MPI_Keyval_create";
	int mpi_errno = MPI_SUCCESS;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_KEYVAL_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_KEYVAL_CREATE);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(keyval, L"keyval", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_keyval_create %p %p %p %p", copy_fn, delete_fn, keyval, extra_state); }

	MPIU_THREADPRIV_GET;
	NMPI_Comm_create_keyval(mpi_errno, ( copy_fn, delete_fn, keyval, extra_state ));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_KEYVAL_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_keyval_create",
			L"**mpi_keyval_create %p %p %p %p", copy_fn, delete_fn, keyval, extra_state);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/*D

MPI_DUP_FN - A function to simple-mindedly copy attributes  

D*/
int MPIR_Dup_fn ( MPI_Comm comm, int keyval, void *extra_state, void *attr_in, void *attr_out, int *flag )
{
	/* No error checking at present */

	MPIU_UNREFERENCED_ARG(comm);
	MPIU_UNREFERENCED_ARG(keyval);
	MPIU_UNREFERENCED_ARG(extra_state);

	/* Set attr_out, the flag and return success */
	(*(void **)attr_out) = attr_in;
	(*flag) = 1;
	return (MPI_SUCCESS);
}

/* -- Begin Profiling Symbol Block for routine MPI_Comm_set_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Comm_set_attr = PMPI_Comm_set_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Comm_set_attr  MPI_Comm_set_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Comm_set_attr as PMPI_Comm_set_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_set_attr
#define MPI_Comm_set_attr PMPI_Comm_set_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_set_attr

/*@
MPI_Comm_set_attr - Stores attribute value associated with a key

Input Parameters:
+ comm - communicator to which attribute will be attached (handle) 
. keyval - key value, as returned by  'MPI_Comm_create_keyval' (integer) 
- attribute_val - attribute value 

Notes:
Values of the permanent attributes 'MPI_TAG_UB', 'MPI_HOST', 'MPI_IO', 
'MPI_WTIME_IS_GLOBAL', 'MPI_UNIVERSE_SIZE', 'MPI_LASTUSEDCODE', and 
'MPI_APPNUM' may not be changed. 

The type of the attribute value depends on whether C, C++, or Fortran
is being used. 
In C and C++, an attribute value is a pointer ('void *'); in Fortran, it is an 
address-sized integer.

If an attribute is already present, the delete function (specified when the
corresponding keyval was created) will be called.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_KEYVAL
.N MPI_ERR_PERM_KEY

.seealso MPI_Comm_create_keyval, MPI_Comm_delete_attr
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_set_attr(MPI_Comm comm, int comm_keyval, void *attribute_val)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_set_attr";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_Keyval *keyval_ptr = NULL;
	MPID_Attribute *p, **old_p;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SET_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SET_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			MPIR_ERRTEST_KEYVAL(comm_keyval, MPID_COMM, L"communicator", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(comm_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Keyval_get_ptr( comm_keyval, keyval_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			/* Validate keyval_ptr */
			MPID_Keyval_valid_ptr( keyval_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_set_attr %C %d %p", comm, comm_keyval, attribute_val); }

	/* Look for attribute.  They are ordered by keyval handle.  This uses 
	a simple linear list algorithm because few applications use more than a 
	handful of attributes */

	old_p = &comm_ptr->attributes;
	p = comm_ptr->attributes;
	while (p)
	{
		if (p->keyval->handle == keyval_ptr->handle)
		{
			/* If found, call the delete function before replacing the attribute */
			mpi_errno = MPIR_Call_attr_delete( comm, p );
			if (mpi_errno)
			{
				goto fn_fail;
			}
			p->value = attribute_val;
			/* Does not change the reference count on the keyval */
			break;
		}
		else
		{
			if (p->keyval->handle > keyval_ptr->handle)
			{
				MPID_Attribute *new_p = (MPID_Attribute *)MPIU_Handle_obj_alloc( &MPID_Attr_mem );
				MPIU_ERR_CHKANDJUMP(!new_p,mpi_errno,MPI_ERR_OTHER,L"**nomem");
				new_p->keyval		 = keyval_ptr;
				new_p->pre_sentinal		 = 0;
				new_p->value		 = attribute_val;
				new_p->post_sentinal = 0;
				new_p->next				 = p->next;
				MPIU_Object_add_ref( keyval_ptr );
				p->next				 = new_p;
				break;
			}
		}
		old_p = &p->next;
		p = p->next;
	}
	if (!p)
	{
		MPID_Attribute *new_p = (MPID_Attribute *)MPIU_Handle_obj_alloc( &MPID_Attr_mem );
		MPIU_ERR_CHKANDJUMP(!new_p,mpi_errno,MPI_ERR_OTHER,L"**nomem");
		/* Did not find in list.  Add at end */
		new_p->keyval			 = keyval_ptr;
		new_p->pre_sentinal  = 0;
		new_p->value			 = attribute_val;
		new_p->post_sentinal = 0;
		new_p->next			 = 0;
		MPIU_Object_add_ref( keyval_ptr );
		*old_p					 = new_p;
	}

	/* Here is where we could add a hook for the device to detect attribute
	value changes, using something like
	MPID_Dev_comm_attr_hook( comm_ptr, keyval, attribute_val );
	*/

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SET_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_set_attr",
			L"**mpi_comm_set_attr %C %d %p", comm, comm_keyval, attribute_val);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Comm_get_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Comm_get_attr = PMPI_Comm_get_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Comm_get_attr  MPI_Comm_get_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Comm_get_attr as PMPI_Comm_get_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_get_attr
#define MPI_Comm_get_attr PMPI_Comm_get_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_get_attr

/*@
MPI_Comm_get_attr - Retrieves attribute value by key

Input Parameters:
+ comm - communicator to which attribute is attached (handle) 
- keyval - key value (integer) 

Output Parameters:
+ attr_value - attribute value, unless 'flag' = false 
- flag -  true if an attribute value was extracted;  false if no attribute is
associated with the key 

Notes:
Attributes must be extracted from the same language as they were inserted  
in with 'MPI_Comm_set_attr'.  The notes for C and Fortran below explain 
why. 

Notes for C:
Even though the 'attr_value' arguement is declared as 'void *', it is
really the address of a void pointer.  See the rationale in the 
standard for more details. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_KEYVAL
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_attr(MPI_Comm comm, int comm_keyval, void *attribute_val, int *flag)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_get_attr";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	static PreDefined_attrs attr_copy;	/* Used to provide a copy of the predefined attributes */
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_GET_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_GET_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			MPIR_ERRTEST_KEYVAL(comm_keyval, MPID_COMM, L"communicator", mpi_errno);
#		   ifdef NEEDS_POINTER_ALIGNMENT_ADJUST
			/* A common user error is to pass the address of a 4-byte
			int when the address of a pointer (or an address-sized int)
			should have been used.  We can test for this specific
			case.  Note that this code assumes sizeof(MPI_Aint) is 
			a power of 2. */
			if ((MPI_Aint)attribute_val & (sizeof(MPI_Aint)-1))
			{
				MPIU_ERR_SET(mpi_errno,MPI_ERR_ARG,L"**attrnotptr");
			}
#		   endif
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(attribute_val, L"attr_val", mpi_errno);
			MPIR_ERRTEST_ARGNULL(flag, L"flag", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_get_attr %C %d %p %p", comm, comm_keyval, attribute_val, flag); }

	/* Check for builtin attribute */
	/* This code is ok for correct programs, but it would be better
	to copy the values from the per-process block and pass the user
	a pointer to a copy */
	/* Note that if we are called from Fortran, we must return the values,
	not the addresses, of these attributes */
	if (HANDLE_GET_KIND(comm_keyval) == HANDLE_KIND_BUILTIN)
	{
		int attr_idx = comm_keyval & 0x0000000f;
		void **attr_val_p = (void **)attribute_val;
#ifdef HAVE_FORTRAN_BINDING
		/* This is an address-sized int instead of a Fortran (MPI_Fint)
		integer because, even for the Fortran keyvals, the C interface is 
		used which stores the result in a pointer (hence we need a
		pointer-sized int).  Thus we use MPI_Aint instead of MPI_Fint.
		On some 64-bit plaforms, such as Solaris-SPARC, using an MPI_Fint
		will cause the value to placed into the high, rather than low,
		end of the output value. */
		MPI_Aint  *attr_int = (MPI_Aint *)attribute_val;
#endif
		*flag = 1;

		/* FIXME : We could initialize some of these here; only tag_ub is 
		used in the error checking. */
		/* 
		* The C versions of the attributes return the address of a 
		* *COPY* of the value (to prevent the user from changing it)
		* and the Fortran versions provide the actual value (as an Fint)
		*/
		attr_copy = MPIR_Process.attrs;
		switch (attr_idx) {
		case 1: /* TAG_UB */
			*attr_val_p = &attr_copy.tag_ub;
			break;
		case 3: /* HOST */
			*attr_val_p = &attr_copy.host;
			break;
		case 5: /* IO */
			*attr_val_p = &attr_copy.io;
			break;
		case 7: /* WTIME */
			*attr_val_p = &attr_copy.wtime_is_global;
			break;
		case 9: /* UNIVERSE_SIZE */
			/* This is a special case.  If universe is not set, then we
			attempt to get it from the device.  If the device is doesn't
			supply a value, then we set the flag accordingly */
			if (attr_copy.universe >= 0)
			{ 
				*attr_val_p = &attr_copy.universe;
			}
			else if (attr_copy.universe == MPIR_UNIVERSE_SIZE_NOT_AVAILABLE)
			{
				*flag = 0;
			}
			else
			{
				mpi_errno = MPID_Get_universe_size(&attr_copy.universe);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					attr_copy.universe = MPIR_UNIVERSE_SIZE_NOT_AVAILABLE;
					goto fn_fail;
				}
				/* --END ERROR HANDLING-- */

				if (attr_copy.universe >= 0)
				{
					*attr_val_p = &attr_copy.universe;
				}
				else
				{
					attr_copy.universe = MPIR_UNIVERSE_SIZE_NOT_AVAILABLE;
					*flag = 0;
				}
			}
			break;
		case 11: /* LASTUSEDCODE */
			*attr_val_p = &attr_copy.lastusedcode;
			break;
		case 13: /* APPNUM */
			/* This is another special case.  If appnum is negative,
			we take that as indicating no value of APPNUM, and set
			the flag accordingly */
			if (attr_copy.appnum < 0)
			{
				*flag = 0;
			}
			else
			{
				*attr_val_p = &attr_copy.appnum;
			}
			break;
#ifdef HAVE_FORTRAN_BINDING
		case 2: /* Fortran TAG_UB */
			*attr_int = attr_copy.tag_ub;
			break;
		case 4: /* Fortran HOST */
			*attr_int = attr_copy.host;
			break;
		case 6: /* Fortran IO */
			*attr_int = attr_copy.io;
			break;
		case 8: /* Fortran WTIME */
			*attr_int = attr_copy.wtime_is_global;
			break;
		case 10: /* UNIVERSE_SIZE */
			/* This is a special case.  If universe is not set, then we
			attempt to get it from the device.  If the device is doesn't
			supply a value, then we set the flag accordingly */
			if (attr_copy.universe >= 0)
			{ 
				*attr_int = attr_copy.universe;
			}
			else if (attr_copy.universe == MPIR_UNIVERSE_SIZE_NOT_AVAILABLE)
			{
				*flag = 0;
			}
			else
			{
				mpi_errno = MPID_Get_universe_size(&attr_copy.universe);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					attr_copy.universe = MPIR_UNIVERSE_SIZE_NOT_AVAILABLE;
					goto fn_fail;
				}
				/* --END ERROR HANDLING-- */

				if (attr_copy.universe >= 0)
				{
					*attr_int = attr_copy.universe;
				}
				else
				{
					attr_copy.universe = MPIR_UNIVERSE_SIZE_NOT_AVAILABLE;
					*flag = 0;
				}
			}
			break;
		case 12: /* LASTUSEDCODE */
			*attr_int = attr_copy.lastusedcode;
			break;
		case 14: /* APPNUM */
			/* This is another special case.  If appnum is negative,
			we take that as indicating no value of APPNUM, and set
			the flag accordingly */
			if (attr_copy.appnum < 0)
			{
				*flag = 0;
			}
			else {
				*attr_int = attr_copy.appnum;
			}
			break;
#endif
		}
	}
	else
	{
		MPID_Attribute *p = comm_ptr->attributes;

		*flag = 0;
		while (p)
		{
			if (p->keyval->handle == comm_keyval)
			{
				*flag = 1;
				(*(void **)attribute_val) = p->value;
				break;
			}
			p = p->next;
		}
	}
	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_GET_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_get_attr",
			L"**mpi_comm_get_attr %C %d %p %p", comm, comm_keyval, attribute_val, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Comm_free_keyval */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Comm_free_keyval = PMPI_Comm_free_keyval
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Comm_free_keyval  MPI_Comm_free_keyval
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Comm_free_keyval as PMPI_Comm_free_keyval
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_free_keyval
#define MPI_Comm_free_keyval PMPI_Comm_free_keyval

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_free_keyval

/*@
MPI_Comm_free_keyval - Frees an attribute key for communicators

Input Parameter:
. comm_keyval - Frees the integer key value (integer) 

Notes:
Key values are global (they can be used with any and all communicators)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_PERM_KEY
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_free_keyval(int *comm_keyval)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_free_keyval";
	MPID_Keyval *keyval_ptr = NULL;
	int		  in_use;
	int		  mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_FREE_KEYVAL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_FREE_KEYVAL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(comm_keyval, L"comm_keyval", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			MPIR_ERRTEST_KEYVAL(*comm_keyval, MPID_COMM, L"communicator", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(*comm_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Keyval_get_ptr( *comm_keyval, keyval_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Keyval_valid_ptr( keyval_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_free_keyval %p", comm_keyval); }

	MPIU_Object_release_ref( keyval_ptr, &in_use);
	if (!in_use)
	{
		MPIU_Handle_obj_free( &MPID_Keyval_mem, keyval_ptr );
	}
	*comm_keyval = MPI_KEYVAL_INVALID;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_FREE_KEYVAL);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_free_keyval",
			L"**mpi_comm_free_keyval %p", comm_keyval);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Comm_delete_attr */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Comm_delete_attr = PMPI_Comm_delete_attr
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Comm_delete_attr  MPI_Comm_delete_attr
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Comm_delete_attr as PMPI_Comm_delete_attr
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_delete_attr
#define MPI_Comm_delete_attr PMPI_Comm_delete_attr

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_delete_attr

/*@
MPI_Comm_delete_attr - Deletes an attribute value associated with a key on 
a  communicator

Input Parameters:
+ comm - communicator to which attribute is attached (handle) 
- comm_keyval - The key value of the deleted attribute (integer) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_PERM_KEY

.seealso MPI_Comm_set_attr, MPI_Comm_create_keyval
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_delete_attr(MPI_Comm comm, int comm_keyval)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_delete_attr";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_Attribute *p, **old_p;
	MPID_Keyval *keyval_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_DELETE_ATTR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_DELETE_ATTR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			MPIR_ERRTEST_KEYVAL(comm_keyval, MPID_COMM, L"communicator", mpi_errno);
			MPIR_ERRTEST_KEYVAL_PERM(comm_keyval, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Keyval_get_ptr( comm_keyval, keyval_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			/* Validate keyval_ptr */
			MPID_Keyval_valid_ptr( keyval_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_delete_attr %C %d", comm, comm_keyval); }

	/* Look for attribute.  They are ordered by keyval handle */

	old_p = &comm_ptr->attributes;
	p	 = comm_ptr->attributes;
	while (p)
	{
		if (p->keyval->handle == keyval_ptr->handle)
		{
			break;
		}
		old_p = &p->next;
		p	 = p->next;
	}

	/* We can't unlock yet, because we must not free the attribute until
	we know whether the delete function has returned with a 0 status
	code */

	if (p)
	{
		/* Run the delete function, if any, and then free the attribute 
		storage */
		mpi_errno = MPIR_Call_attr_delete( comm, p );

		if (!mpi_errno)
		{
			int in_use;
			/* We found the attribute.  Remove it from the list */
			*old_p = p->next;
			/* Decrement the use of the keyval */
			MPIU_Object_release_ref( p->keyval, &in_use);
			if (!in_use)
			{
				MPIU_Handle_obj_free( &MPID_Keyval_mem, p->keyval );
			}
			MPID_Attr_free(p);
		}
	}

	/* ... end of body of routine ... */

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_DELETE_ATTR);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_delete_attr",
			L"**mpi_comm_delete_attr %C %d", comm, comm_keyval);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Comm_create_keyval */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Comm_create_keyval = PMPI_Comm_create_keyval
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Comm_create_keyval  MPI_Comm_create_keyval
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Comm_create_keyval as PMPI_Comm_create_keyval
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_create_keyval
#define MPI_Comm_create_keyval PMPI_Comm_create_keyval

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_create_keyval

/*@
MPI_Comm_create_keyval - Create a new attribute key 

Input Parameters:
+ copy_fn - Copy callback function for 'keyval' 
. delete_fn - Delete callback function for 'keyval' 
- extra_state - Extra state for callback functions 

Output Parameter:
. comm_keyval - key value for future access (integer) 

Notes:
Key values are global (available for any and all communicators).

Default copy and delete functions are available.  These are
+ MPI_COMM_NULL_COPY_FN   - empty copy function
. MPI_COMM_NULL_DELETE_FN - empty delete function
- MPI_COMM_DUP_FN		 - simple dup function

There are subtle differences between C and Fortran that require that the
copy_fn be written in the same language from which 'MPI_Comm_create_keyval'
is called.
This should not be a problem for most users; only programers using both 
Fortran and C in the same program need to be sure that they follow this rule.

.N AttrErrReturn

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS

.seealso MPI_Comm_free_keyval
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_create_keyval(MPI_Comm_copy_attr_function *comm_copy_attr_fn, 
												 MPI_Comm_delete_attr_function *comm_delete_attr_fn, 
												 int *comm_keyval, void *extra_state)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_create_keyval";
	int mpi_errno = MPI_SUCCESS;
	MPID_Keyval *keyval_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_CREATE_KEYVAL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_CREATE_KEYVAL);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(comm_keyval, L"comm_keyval", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_create_keyval %p %p %p %p", comm_copy_attr_fn, comm_delete_attr_fn, comm_keyval, extra_state); }

	keyval_ptr = (MPID_Keyval *)MPIU_Handle_obj_alloc( &MPID_Keyval_mem );
	MPIU_ERR_CHKANDJUMP(!keyval_ptr,mpi_errno,MPI_ERR_OTHER,L"**nomem");

	/* Initialize the attribute dup function */
	if (!MPIR_Process.attr_dup)
	{
		MPIR_Process.attr_dup  = MPIR_Attr_dup_list;
		MPIR_Process.attr_free = MPIR_Attr_delete_list;
	}

	/* The handle encodes the keyval kind.  Modify it to have the correct field */
	keyval_ptr->handle = (keyval_ptr->handle & ~(0x03c00000)) | (MPID_COMM << 22);
	*comm_keyval = keyval_ptr->handle;
	MPIU_Object_set_ref(keyval_ptr,1);
	keyval_ptr->language = MPID_LANG_C;
	keyval_ptr->kind = MPID_COMM;
	keyval_ptr->extra_state = extra_state;
	keyval_ptr->copyfn.C_CopyFunction = comm_copy_attr_fn;
	keyval_ptr->delfn.C_DeleteFunction = comm_delete_attr_fn;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_CREATE_KEYVAL);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_create_keyval",
			L"**mpi_comm_create_keyval %p %p %p %p", comm_copy_attr_fn, comm_delete_attr_fn, comm_keyval, extra_state);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Attr_put */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Attr_put = PMPI_Attr_put
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Attr_put  MPI_Attr_put
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Attr_put as PMPI_Attr_put
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Attr_put
#define MPI_Attr_put PMPI_Attr_put

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Attr_put

/*@

MPI_Attr_put - Stores attribute value associated with a key

Input Parameters:
+ comm - communicator to which attribute will be attached (handle) 
. keyval - key value, as returned by  'MPI_KEYVAL_CREATE' (integer) 
- attribute_val - attribute value 

Notes:
Values of the permanent attributes 'MPI_TAG_UB', 'MPI_HOST', 'MPI_IO', 
'MPI_WTIME_IS_GLOBAL', 'MPI_UNIVERSE_SIZE', 'MPI_LASTUSEDCODE', and
'MPI_APPNUM'  may not be changed. 

The type of the attribute value depends on whether C, C++, or Fortran
is being used. 
In C and C++, an attribute value is a pointer ('void *'); in Fortran,
it is a single  
integer (`not` a pointer, since Fortran has no pointers and there are systems 
for which a pointer does not fit in an integer (e.g., any > 32 bit address 
system that uses 64 bits for Fortran 'DOUBLE PRECISION').

If an attribute is already present, the delete function (specified when the
corresponding keyval was created) will be called.

.N ThreadSafe

.N Deprecated
The replacement for this routine is 'MPI_Comm_set_attr'.

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_KEYVAL
.N MPI_ERR_PERM_KEY

.seealso MPI_Attr_get, MPI_Keyval_create, MPI_Attr_delete, MPI_Comm_set_attr
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Attr_put(MPI_Comm comm, int keyval, void *attr_value)
{
	static const wchar_t FCNAME[] = L"MPI_Attr_put";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ATTR_PUT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ATTR_PUT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_attr_put %C %d %p", comm, keyval, attr_value); }

	MPIU_THREADPRIV_GET;
	NMPI_Comm_set_attr(mpi_errno, ( comm, keyval, attr_value ));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ATTR_PUT);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_attr_put",
			L"**mpi_attr_put %C %d %p", comm, keyval, attr_value);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Attr_get */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Attr_get = PMPI_Attr_get
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Attr_get  MPI_Attr_get
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Attr_get as PMPI_Attr_get
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Attr_get
#define MPI_Attr_get PMPI_Attr_get

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Attr_get

/*@

MPI_Attr_get - Retrieves attribute value by key

Input Parameters:
+ comm - communicator to which attribute is attached (handle) 
- keyval - key value (integer) 

Output Parameters:
+ attr_value - attribute value, unless 'flag' = false 
- flag -  true if an attribute value was extracted;  false if no attribute is
associated with the key 

Notes:
Attributes must be extracted from the same language as they were inserted  
in with 'MPI_ATTR_PUT'.  The notes for C and Fortran below explain why.

Notes for C:
Even though the 'attr_value' arguement is declared as 'void *', it is
really the address of a void pointer (i.e., a 'void **').  Using
a 'void *', however, is more in keeping with C idiom and allows the
pointer to be passed without additional casts.


.N ThreadSafe

.N Deprecated
The replacement for this routine is 'MPI_Comm_get_attr'.

.N Fortran

The 'attr_value' in Fortran is a pointer to a Fortran integer, not
a pointer to a 'void *'.  

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_KEYVAL

.seealso MPI_Attr_put, MPI_Keyval_create, MPI_Attr_delete, MPI_Comm_get_attr
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Attr_get(MPI_Comm comm, int keyval, void *attr_value, int *flag)
{
	static const wchar_t FCNAME[] = L"MPI_Attr_get";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ATTR_GET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ATTR_GET);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
#		   ifdef NEEDS_POINTER_ALIGNMENT_ADJUST
			/* A common user error is to pass the address of a 4-byte
			int when the address of a pointer (or an address-sized int)
			should have been used.  We can test for this specific
			case.  Note that this code assumes sizeof(MPI_Aint) is 
			a power of 2. */
			if ((MPI_Aint)attr_value & (sizeof(MPI_Aint)-1))
			{
				MPIU_ERR_SET(mpi_errno,MPI_ERR_ARG,L"**attrnotptr");
			}
#		   endif
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(attr_value, L"attr_value", mpi_errno);
			MPIR_ERRTEST_ARGNULL(flag, L"flag", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_attr_get %C %d %p %p", comm, keyval, attr_value, flag); }

	MPIU_THREADPRIV_GET;
	NMPI_Comm_get_attr(mpi_errno, ( comm, keyval, attr_value, flag ));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ATTR_GET);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_attr_get",
			L"**mpi_attr_get %C %d %p %p", comm, keyval, attr_value, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Attr_delete */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Attr_delete = PMPI_Attr_delete
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Attr_delete  MPI_Attr_delete
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Attr_delete as PMPI_Attr_delete
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Attr_delete
#define MPI_Attr_delete PMPI_Attr_delete

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Attr_delete

/*@

MPI_Attr_delete - Deletes an attribute value associated with a key on a 
communicator

Input Parameters:
+ comm - communicator to which attribute is attached (handle) 
- keyval - The key value of the deleted attribute (integer) 

.N ThreadSafe

.N Deprecated
The replacement for this routine is 'MPI_Comm_delete_attr'.

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_PERM_KEY
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Attr_delete(MPI_Comm comm, int keyval)
{
	static const wchar_t FCNAME[] = L"MPI_Attr_delete";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ATTR_DELETE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("attr");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ATTR_DELETE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_attr_delete %C %d", comm, keyval); }

	MPIU_THREADPRIV_GET;
	NMPI_Comm_delete_attr(mpi_errno, ( comm, keyval ));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ATTR_DELETE);
	MPIU_THREAD_SINGLE_CS_EXIT("attr");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_attr_delete",
			L"**mpi_attr_delete %C %d", comm, keyval);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
