/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#ifdef HAVE_FORTRAN_BINDING
#include "mpi_fortlogical.h"
#endif

/* MINLOC and MAXLOC structures */
typedef struct MPIR_2int_loctype
{
	int  value;
	int  loc;
} MPIR_2int_loctype;

typedef struct MPIR_floatint_loctype
{
	float  value;
	int	loc;
} MPIR_floatint_loctype;

typedef struct MPIR_longint_loctype
{
	long  value;
	int	loc;
} MPIR_longint_loctype;

typedef struct MPIR_shortint_loctype
{
	short  value;
	int	loc;
} MPIR_shortint_loctype;

typedef struct MPIR_doubleint_loctype
{
	double  value;
	int	 loc;
} MPIR_doubleint_loctype;

#if defined(HAVE_LONG_DOUBLE)
typedef struct MPIR_longdoubleint_loctype
{
	long double   value;
	int		   loc;
} MPIR_longdoubleint_loctype;
#endif

/* -- Begin Profiling Symbol Block for routine MPI_Op_create */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Op_create = PMPI_Op_create
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Op_create  MPI_Op_create
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Op_create as PMPI_Op_create
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Op_create
#define MPI_Op_create PMPI_Op_create

#ifndef MPID_OP_PREALLOC 
#define MPID_OP_PREALLOC 16
#endif

/* Preallocated op objects */
MPID_Op MPID_Op_builtin[MPID_OP_N_BUILTIN] = { {0} };
MPID_Op MPID_Op_direct[MPID_OP_PREALLOC] = { {0} };
MPIU_Object_alloc_t MPID_Op_mem = { 0, 0, 0, 0, MPID_OP, 
sizeof(MPID_Op), 
MPID_Op_direct,
MPID_OP_PREALLOC, };

#ifdef HAVE_CXX_BINDING
#undef FUNCNAME
#define FUNCNAME MPIR_Op_set_cxx
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Op_set_cxx( MPI_Op op, void (*opcall)(void) )
{
	MPID_Op *op_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_OP_SET_CXX);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_OP_SET_CXX);

	MPID_Op_get_ptr( op, op_ptr );
	op_ptr->language				= MPID_LANG_CXX;
	MPIR_Process.cxx_call_op_fn		= (void (*)(void *, void *, int, 
		MPI_Datatype, MPI_User_function *))opcall;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_OP_SET_CXX);
}
#endif
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Op_create

/*@
MPI_Op_create - Creates a user-defined combination function handle

Input Parameters:
+ function - user defined function (function) 
- commute -  true if commutative;  false otherwise. (logical)

Output Parameter:
. op - operation (handle) 

Notes on the user function:
The calling list for the user function type is
.vb
typedef void (MPI_User_function) ( void * a, 
void * b, int * len, MPI_Datatype * ); 
.ve
where the operation is 'b[i] = a[i] op b[i]', for 'i=0,...,len-1'.  A pointer
to the datatype given to the MPI collective computation routine (i.e., 
'MPI_Reduce', 'MPI_Allreduce', 'MPI_Scan', or 'MPI_Reduce_scatter') is also
passed to the user-specified routine.

.N ThreadSafe

.N Fortran

.N collops

.N Errors
.N MPI_SUCCESS

.seealso: MPI_Op_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Op_create(MPI_User_function *function, int commute, MPI_Op *op)
{
#undef FCNAME
	static const wchar_t FCNAME[] = L"MPI_Op_create";
	MPID_Op *op_ptr;
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_OP_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("coll");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_OP_CREATE);

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_op_create %p %d %p", function, commute, op); }

	op_ptr = (MPID_Op *)MPIU_Handle_obj_alloc( &MPID_Op_mem );
	/* --BEGIN ERROR HANDLING-- */
	if (!op_ptr)
	{
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPI_Op" );
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	*op					 = op_ptr->handle;
	op_ptr->language = MPID_LANG_C;
	op_ptr->kind	 = commute ? MPID_OP_USER : MPID_OP_USER_NONCOMMUTE;
	op_ptr->function.c_function = (void (*)(const void *, void *, const int *, const MPI_Datatype *))function;
	MPIU_Object_set_ref(op_ptr,1);

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_OP_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT("coll");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_op_create",
			L"**mpi_op_create %p %d %p", function, commute, op);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* -- Begin Profiling Symbol Block for routine MPI_Op_free */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Op_free = PMPI_Op_free
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Op_free  MPI_Op_free
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Op_free as PMPI_Op_free
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Op_free
#define MPI_Op_free PMPI_Op_free

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Op_free

/*@
MPI_Op_free - Frees a user-defined combination function handle

Input Parameter:
. op - operation (handle) 

Notes:
'op' is set to 'MPI_OP_NULL' on exit.

.N NULL

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_PERM_OP

.seealso: MPI_Op_create
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Op_free(MPI_Op *op)
{
	static const wchar_t FCNAME[] = L"MPI_Op_free";
	MPID_Op *op_ptr = NULL;
	int	 in_use;
	int	 mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_OP_FREE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("coll");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_OP_FREE);

	MPID_Op_get_ptr( *op, op_ptr );
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Op_valid_ptr( op_ptr, mpi_errno );
			if (!mpi_errno)
			{
				if (op_ptr->kind < MPID_OP_USER_NONCOMMUTE)
				{
					mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**permop", 0 );
				}
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_op_free %p", op); }

	MPIU_Object_release_ref( op_ptr, &in_use);
	if (!in_use)
	{
		MPIU_Handle_obj_free( &MPID_Op_mem, op_ptr );
	}
	*op = MPI_OP_NULL;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_OP_FREE);
	MPIU_THREAD_SINGLE_CS_EXIT("coll");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_op_free", L"**mpi_op_free %p", op);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran integer,
* and byte data items (4.9.2 Predefined reduce operations)
*/
#ifndef MPIR_LBAND
#define MPIR_LBAND(a,b) ((a)&(b))
#endif
#undef FUNCNAME
#define FUNCNAME MPIR_BAND
void MPIR_BAND(void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_BAND";
	int i, len = *Len;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_BAND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_BAND);

	switch (*type)
	{
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL:
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
#endif
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; 
			int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
	case MPI_BYTE:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBAND(a[i],b[i]);
			break;
		}
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_BAND" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_BAND);
}


#undef FUNCNAME
#define FUNCNAME MPIR_BAND_check_dtype
int MPIR_BAND_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_BAND_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_BAND_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_BAND_CHECK_DTYPE);
	switch (type)
	{
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL:
	case MPI_INTEGER: 
#endif
	case MPI_INT: 
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_BYTE: 
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_BAND" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_BAND_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran integer,
* and byte data items (4.9.2 Predefined reduce operations)
*/
#ifndef MPIR_LBOR
#define MPIR_LBOR(a,b) ((a)|(b))
#endif
#undef FUNCNAME
#define FUNCNAME MPIR_BOR
void MPIR_BOR ( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_BOR";
	int i, len = *Len;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_BOR);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_BOR);

	switch (*type)
	{
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL:
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
#endif
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
	case MPI_BYTE:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBOR(a[i],b[i]);
			break;
		}
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_BOR" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_BOR);
}


#undef FUNCNAME
#define FUNCNAME MPIR_BOR_check_dtype
int MPIR_BOR_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_BOR_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_BOR_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_BOR_CHECK_DTYPE);

	switch (type)
	{
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL: case MPI_INTEGER: 
#endif
	case MPI_INT: 
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_BYTE: 
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_BOR" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_BOR_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran integer,
* and byte data items (4.9.2 Predefined reduce operations)
*/
#ifndef MPIR_LBXOR
#define MPIR_LBXOR(a,b) ((a)^(b))
#endif
#undef FUNCNAME
#define FUNCNAME MPIR_BXOR
void MPIR_BXOR ( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_BXOR";
	int i, len = *Len;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_BXOR);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_BXOR);

	switch (*type)
	{
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL:
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
#endif
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; 
			int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
	case MPI_BYTE:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LBXOR(a[i],b[i]);
			break;
		}
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_BXOR" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_BXOR);
}

#undef FUNCNAME
#define FUNCNAME MPIR_BXOR_check_dtype
int MPIR_BXOR_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_BXOR_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_BXOR_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_BXOR_CHECK_DTYPE);

	switch (type)
	{
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL: case MPI_INTEGER: 
#endif
	case MPI_INT: 
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_BYTE: 
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_BXOR" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_BXOR_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran logical
* data items (4.9.2 Predefined reduce operations)
*/
#ifndef MPIR_LLAND
#define MPIR_LLAND(a,b) ((a)&&(b))
#endif
#undef FUNCNAME
#define FUNCNAME MPIR_LAND
#undef FCNAME
void MPIR_LAND ( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_LAND";
	int i, len = *Len;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_LAND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_LAND);

	switch (*type)
	{
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; 
			int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
		/* FIXME: This assume C float = Fortran real */
	case MPI_REAL: 
#endif
		{
			float * restrict a = (float *)inoutvec; 
			float * restrict b = (float *)invec;
			for ( i=0; i<len; i++ )
				a[i] = (float)MPIR_LLAND(a[i],b[i]);
			break;
		}
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
		/* FIXME: This assume C double = Fortran double precision */
	case MPI_DOUBLE_PRECISION: 
#endif
		{
			double * restrict a = (double *)inoutvec; 
			double * restrict b = (double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
		{
			MPIR_REAL4_CTYPE * restrict a = (MPIR_REAL4_CTYPE *)inoutvec; 
			MPIR_REAL4_CTYPE * restrict b = (MPIR_REAL4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
		{
			MPIR_REAL8_CTYPE * restrict a = (MPIR_REAL8_CTYPE *)inoutvec; 
			MPIR_REAL8_CTYPE * restrict b = (MPIR_REAL8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
		{
			MPIR_REAL16_CTYPE * restrict a = (MPIR_REAL16_CTYPE *)inoutvec; 
			MPIR_REAL16_CTYPE * restrict b = (MPIR_REAL16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE:
		{
			long double * restrict a = (long double *)inoutvec; 
			long double * restrict b = (long double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for (i=0; i<len; i++) 
				a[i] = MPIR_TO_FLOG(MPIR_LLAND(MPIR_FROM_FLOG(a[i]),
				MPIR_FROM_FLOG(b[i])));
			break;
		}
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLAND(a[i],b[i]);
			break;
		}
#endif
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_LAND" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_LAND);
}

#undef FUNCNAME
#define FUNCNAME MPIR_LAND_check_dtype
#undef FCNAME
int MPIR_LAND_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_LAND_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_LAND_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_LAND_CHECK_DTYPE);

	switch (type)
	{
	case MPI_INT: 
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
#endif
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE: 
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL: 
	case MPI_INTEGER: 
#endif
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_LAND" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_LAND_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran logical
* data items (4.9.2 Predefined reduce operations)
*/
#ifndef MPIR_LLOR
#define MPIR_LLOR(a,b) ((a)||(b))
#endif
#undef FUNCNAME
#define FUNCNAME MPIR_LOR
#undef FCNAME
void MPIR_LOR ( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_LOR";
	int i, len = *Len;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_LOR);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_LOR);

	switch (*type)
	{
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
		{
			float * restrict a = (float *)inoutvec; 
			float * restrict b = (float *)invec;
			for ( i=0; i<len; i++ )
				a[i] = (float)MPIR_LLOR(a[i],b[i]);
			break;
		}
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
		/* FIXME : This assumes Fortran double precision == C double */
	case MPI_DOUBLE_PRECISION: 
#endif
		{
			double * restrict a = (double *)inoutvec; 
			double * restrict b = (double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE:
		{
			long double * restrict a = (long double *)inoutvec; 
			long double * restrict b = (long double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for (i=0; i<len; i++) 
				a[i] = MPIR_TO_FLOG(MPIR_LLOR(MPIR_FROM_FLOG(a[i]),
				MPIR_FROM_FLOG(b[i])));
			break;
		}
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for (i=0; i<len; i++) 
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
		{
			MPIR_REAL4_CTYPE * restrict a = (MPIR_REAL4_CTYPE *)inoutvec; 
			MPIR_REAL4_CTYPE * restrict b = (MPIR_REAL4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
		{
			MPIR_REAL8_CTYPE * restrict a = (MPIR_REAL8_CTYPE *)inoutvec; 
			MPIR_REAL8_CTYPE * restrict b = (MPIR_REAL8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
		{
			MPIR_REAL16_CTYPE * restrict a = (MPIR_REAL16_CTYPE *)inoutvec; 
			MPIR_REAL16_CTYPE * restrict b = (MPIR_REAL16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLOR(a[i],b[i]);
			break;
		}
#endif
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_LOR" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_LOR);
}

#undef FUNCNAME
#define FUNCNAME MPIR_LOR_check_dtype
#undef FCNAME
int MPIR_LOR_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_LOR_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_LOR_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_LOR_CHECK_DTYPE);

	switch (type)
	{
	case MPI_INT: 
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
		/* FIXME : This assumes Fortran double precision == C double */
	case MPI_DOUBLE_PRECISION: 
#endif
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE: 
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL: 
	case MPI_INTEGER: 
#endif
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_LOR" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_LOR_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran logical
* data items (4.9.2 Predefined reduce operations)
*/
#ifndef MPIR_LLXOR
#define MPIR_LLXOR(a,b) (((a)&&(!b))||((!a)&&(b)))
#endif
#undef FUNCNAME
#define FUNCNAME MPIR_LXOR
#undef FCNAME
void MPIR_LXOR ( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_LXOR";
	int i, len = *Len;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_LXOR);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_LXOR);

	switch (*type)
	{
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; 
			int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
		/* FIXME: This assumes C float = Fortran real */
#endif
		{
			float * restrict a = (float *)inoutvec; 
			float * restrict b = (float *)invec;
			for ( i=0; i<len; i++ )
				a[i] = (float)MPIR_LLXOR(a[i],b[i]);
			break;
		}
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
		/* FIXME: This assumes C double = Fortran double precision */
#endif
		{
			double * restrict a = (double *)inoutvec; double * restrict b = (double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE:
		{
			long double * restrict a = (long double *)inoutvec; 
			long double * restrict b = (long double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for (i=0; i<len; i++) 
				a[i] = MPIR_TO_FLOG(MPIR_LLXOR(MPIR_FROM_FLOG(a[i]),
				MPIR_FROM_FLOG(b[i])));
			break;
		}
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for (i=0; i<len; i++) 
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
		{
			MPIR_REAL4_CTYPE * restrict a = (MPIR_REAL4_CTYPE *)inoutvec; 
			MPIR_REAL4_CTYPE * restrict b = (MPIR_REAL4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
		{
			MPIR_REAL8_CTYPE * restrict a = (MPIR_REAL8_CTYPE *)inoutvec; 
			MPIR_REAL8_CTYPE * restrict b = (MPIR_REAL8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
		{
			MPIR_REAL16_CTYPE * restrict a = (MPIR_REAL16_CTYPE *)inoutvec; 
			MPIR_REAL16_CTYPE * restrict b = (MPIR_REAL16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LLXOR(a[i],b[i]);
			break;
		}
#endif
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_LXOR" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_LXOR);
}

#undef FUNCNAME
#define FUNCNAME MPIR_LXOR_check_dtype
#undef FCNAME
int MPIR_LXOR_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_LXOR_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_LXOR_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_LXOR_CHECK_DTYPE);

	switch (type)
	{
	case MPI_INT: 
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
#endif
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE: 
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_LOGICAL: 
	case MPI_INTEGER: 
#endif
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_LXOR" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_LXOR_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran integer,
* and floating point data items (4.9.2 Predefined reduce operations)
*/
#undef FUNCNAME
#define FUNCNAME MPIR_MAXF
void MPIR_MAXF( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_MAXF";
	int i, len = *Len;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_MAXF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_MAXF);

	switch (*type)
	{
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; 
			int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#ifdef HAVE_FORTRAN_BINDING
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
		{
			MPIR_REAL4_CTYPE * restrict a = (MPIR_REAL4_CTYPE *)inoutvec; 
			MPIR_REAL4_CTYPE * restrict b = (MPIR_REAL4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
		{
			MPIR_REAL8_CTYPE * restrict a = (MPIR_REAL8_CTYPE *)inoutvec; 
			MPIR_REAL8_CTYPE * restrict b = (MPIR_REAL8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
		{
			MPIR_REAL16_CTYPE * restrict a = (MPIR_REAL16_CTYPE *)inoutvec; 
			MPIR_REAL16_CTYPE * restrict b = (MPIR_REAL16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED:
		{
			unsigned int * restrict a = (unsigned int *)inoutvec; 
			unsigned int * restrict b = (unsigned int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
		/* FIXME: This assumes C float = Fortran real */
#endif
		{
			float * restrict a = (float *)inoutvec; 
			float * restrict b = (float *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
		/* FIXME: This assumes C double = Fortran double precision */
#endif
		{
			double * restrict a = (double *)inoutvec; 
			double * restrict b = (double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE:
		{
			long double * restrict a = (long double *)inoutvec; 
			long double * restrict b = (long double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MAX(a[i],b[i]);
			break;
		}
#endif
		/* --BEGIN ERROR HANDLING-- */
	default: {
		MPICH_PerThread_t *p;
		MPIR_GetPerThread(&p);
		p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_MAX" );
		break;
			 }
			 /* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_MAXF);
}

#undef FUNCNAME
#define FUNCNAME MPIR_MAXF_check_dtype
int MPIR_MAXF_check_dtype( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_MAXF_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_MAXF_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_MAXF_CHECK_DTYPE);

	switch (type)
	{
	case MPI_INT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_INTEGER: 
#endif
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
#endif
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE: 
#endif
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_MAX" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_MAXF_CHECK_DTYPE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIR_MAXLOC
void MPIR_MAXLOC( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_MAXLOC";
	int i, len = *Len, flen;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_MAXLOC);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_MAXLOC);

	flen = len * 2; /* used for Fortran types */

	switch (*type)
	{
		/* first the C types */
	case MPI_2INT:
		{
			MPIR_2int_loctype *a = (MPIR_2int_loctype *)inoutvec;
			MPIR_2int_loctype *b = (MPIR_2int_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value < b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
	case MPI_FLOAT_INT:
		{
			MPIR_floatint_loctype *a = (MPIR_floatint_loctype *)inoutvec;
			MPIR_floatint_loctype *b = (MPIR_floatint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value < b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
	case MPI_LONG_INT:
		{
			MPIR_longint_loctype *a = (MPIR_longint_loctype *)inoutvec;
			MPIR_longint_loctype *b = (MPIR_longint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value < b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
	case MPI_SHORT_INT:
		{
			MPIR_shortint_loctype *as = (MPIR_shortint_loctype *)inoutvec;
			MPIR_shortint_loctype *bs = (MPIR_shortint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (as[i].value == bs[i].value)
					as[i].loc = MPIR_MIN(as[i].loc,bs[i].loc);
				else if (as[i].value < bs[i].value)
				{
					as[i].value = bs[i].value;
					as[i].loc   = bs[i].loc;
				}
			}
			break;
		}
	case MPI_DOUBLE_INT:
		{
			MPIR_doubleint_loctype *a = (MPIR_doubleint_loctype *)inoutvec;
			MPIR_doubleint_loctype *b = (MPIR_doubleint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value < b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}

#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE_INT:
		{
			MPIR_longdoubleint_loctype *a = (MPIR_longdoubleint_loctype *)inoutvec;
			MPIR_longdoubleint_loctype *b = (MPIR_longdoubleint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value < b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
#endif

		/* now the Fortran types */
#ifdef HAVE_FORTRAN_BINDING
#ifndef HAVE_NO_FORTRAN_MPI_TYPES_IN_C
	case MPI_2INTEGER:
		{
			int *a = (int *)inoutvec; int *b = (int *)invec;
			for ( i=0; i<flen; i+=2 )
			{
				if (a[i] == b[i])
					a[i+1] = MPIR_MIN(a[i+1],b[i+1]);
				else if (a[i] < b[i])
				{
					a[i]   = b[i];
					a[i+1] = b[i+1];
				}
			}
			break;
		}
	case MPI_2REAL:
		{
			float *a = (float *)inoutvec; float *b = (float *)invec;
			for ( i=0; i<flen; i+=2 )
			{
				if (a[i] == b[i])
					a[i+1] = MPIR_MIN(a[i+1],b[i+1]);
				else if (a[i] < b[i])
				{
					a[i]   = b[i];
					a[i+1] = b[i+1];
				}
			}
			break;
		}
	case MPI_2DOUBLE_PRECISION:
		{
			double *a = (double *)inoutvec; double *b = (double *)invec;
			for ( i=0; i<flen; i+=2 )
			{
				if (a[i] == b[i])
					a[i+1] = MPIR_MIN(a[i+1],b[i+1]);
				else if (a[i] < b[i])
				{
					a[i]   = b[i];
					a[i+1] = b[i+1];
				}
			}
			break;
		}
#endif
#endif
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_MAXLOC" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_MAXLOC);
}

#undef FUNCNAME
#define FUNCNAME MPIR_MAXLOC_check_dtype
int MPIR_MAXLOC_check_dtype( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_MAXLOC_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_MAXLOC_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_MAXLOC_CHECK_DTYPE);
	switch (type)
	{
		/* first the C types */
	case MPI_2INT: 
	case MPI_FLOAT_INT: 
	case MPI_LONG_INT: 
	case MPI_SHORT_INT: 
	case MPI_DOUBLE_INT: 
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE_INT: 
#endif
		/* now the Fortran types */
#ifdef HAVE_FORTRAN_BINDING
#ifndef HAVE_NO_FORTRAN_MPI_TYPES_IN_C
	case MPI_2INTEGER: 
	case MPI_2REAL: 
	case MPI_2DOUBLE_PRECISION: 
#endif
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_MAXLOC" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_MAXLOC_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran integer,
* and floating point data items (4.9.2 Predefined reduce operations)
*/
#undef FUNCNAME
#define FUNCNAME MPIR_MINF
void MPIR_MINF ( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_MINF";
	int i, len = *Len;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_MINF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_MINF);

	switch (*type)
	{
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; 
			int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#ifdef HAVE_FORTRAN_BINDING
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
		{
			MPIR_REAL4_CTYPE * restrict a = (MPIR_REAL4_CTYPE *)inoutvec; 
			MPIR_REAL4_CTYPE * restrict b = (MPIR_REAL4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
		{
			MPIR_REAL8_CTYPE * restrict a = (MPIR_REAL8_CTYPE *)inoutvec; 
			MPIR_REAL8_CTYPE * restrict b = (MPIR_REAL8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
		{
			MPIR_REAL16_CTYPE * restrict a = (MPIR_REAL16_CTYPE *)inoutvec; 
			MPIR_REAL16_CTYPE * restrict b = (MPIR_REAL16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
		/* FIXME: This assumes C float = Fortran real */
#endif
		{
			float * restrict a = (float *)inoutvec; 
			float * restrict b = (float *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
		/* FIXME: This assumes C double = Fortran double precision */
#endif
		{
			double * restrict a = (double *)inoutvec; 
			double * restrict b = (double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE:
		{
			long double * restrict a = (long double *)inoutvec; 
			long double * restrict b = (long double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_MIN(a[i],b[i]);
			break;
		}
#endif
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_MIN" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_MINF);
}

#undef FUNCNAME
#define FUNCNAME MPIR_MINF_check_dtype
int MPIR_MINF_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_MINF_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_MINF_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_MINF_CHECK_DTYPE);

	switch (type)
	{
	case MPI_INT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_INTEGER: 
#endif
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
#endif
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE: 
#endif
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_MIN" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_MINF_CHECK_DTYPE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIR_MINLOC
void MPIR_MINLOC( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_MINLOC";
	int i, len = *Len, flen;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_MINLOC);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_MINLOC);

	flen = len * 2; /* used for Fortran types */

	switch (*type)
	{
		/* first the C types */
	case MPI_2INT:
		{
			MPIR_2int_loctype *a = (MPIR_2int_loctype *)inoutvec;
			MPIR_2int_loctype *b = (MPIR_2int_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value > b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
	case MPI_FLOAT_INT:
		{
			MPIR_floatint_loctype *a = (MPIR_floatint_loctype *)inoutvec;
			MPIR_floatint_loctype *b = (MPIR_floatint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value > b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
	case MPI_LONG_INT:
		{
			MPIR_longint_loctype *a = (MPIR_longint_loctype *)inoutvec;
			MPIR_longint_loctype *b = (MPIR_longint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value > b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
	case MPI_SHORT_INT:
		{
			MPIR_shortint_loctype *a = (MPIR_shortint_loctype *)inoutvec;
			MPIR_shortint_loctype *b = (MPIR_shortint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value > b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
	case MPI_DOUBLE_INT:
		{
			MPIR_doubleint_loctype *a = (MPIR_doubleint_loctype *)inoutvec;
			MPIR_doubleint_loctype *b = (MPIR_doubleint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value > b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE_INT:
		{
			MPIR_longdoubleint_loctype *a = (MPIR_longdoubleint_loctype *)inoutvec;
			MPIR_longdoubleint_loctype *b = (MPIR_longdoubleint_loctype *)invec;
			for (i=0; i<len; i++)
			{
				if (a[i].value == b[i].value)
					a[i].loc = MPIR_MIN(a[i].loc,b[i].loc);
				else if (a[i].value > b[i].value)
				{
					a[i].value = b[i].value;
					a[i].loc   = b[i].loc;
				}
			}
			break;
		}
#endif

		/* now the Fortran types */
#ifdef HAVE_FORTRAN_BINDING
#ifndef HAVE_NO_FORTRAN_MPI_TYPES_IN_C
	case MPI_2INTEGER:
		{
			int *a = (int *)inoutvec; int *b = (int *)invec;
			for ( i=0; i<flen; i+=2 )
			{
				if (a[i] == b[i])
					a[i+1] = MPIR_MIN(a[i+1],b[i+1]);
				else if (a[i] > b[i])
				{
					a[i]   = b[i];
					a[i+1] = b[i+1];
				}
			}
			break;
		}
	case MPI_2REAL:
		{
			float *a = (float *)inoutvec; float *b = (float *)invec;
			for ( i=0; i<flen; i+=2 )
			{
				if (a[i] == b[i])
					a[i+1] = MPIR_MIN(a[i+1],b[i+1]);
				else if (a[i] > b[i])
				{
					a[i]   = b[i];
					a[i+1] = b[i+1];
				}
			}
			break;
		}
	case MPI_2DOUBLE_PRECISION:
		{
			double *a = (double *)inoutvec; double *b = (double *)invec;
			for ( i=0; i<flen; i+=2 )
			{
				if (a[i] == b[i])
					a[i+1] = MPIR_MIN(a[i+1],b[i+1]);
				else if (a[i] > b[i])
				{
					a[i]   = b[i];
					a[i+1] = b[i+1];
				}
			}
			break;
		}
#endif
#endif
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_MINLOC" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_MINLOC);
}

#undef FUNCNAME
#define FUNCNAME MPIR_MINLOC_check_dtype
int MPIR_MINLOC_check_dtype( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_MINLOC_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_MINLOC_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_MINLOC_CHECK_DTYPE);

	switch (type)
	{
		/* first the C types */
	case MPI_2INT: 
	case MPI_FLOAT_INT: 
	case MPI_LONG_INT: 
	case MPI_SHORT_INT: 
	case MPI_DOUBLE_INT: 
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE_INT: 
#endif
		/* now the Fortran types */
#ifdef HAVE_FORTRAN_BINDING
#ifndef HAVE_NO_FORTRAN_MPI_TYPES_IN_C
	case MPI_2INTEGER: 
	case MPI_2REAL: 
	case MPI_2DOUBLE_PRECISION: 
#endif
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_MINLOC" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_MINLOC_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran integer,
* floating point, and complex data items (4.9.2 Predefined reduce operations)
*/
#define MPIR_LPROD(a,b) ((a)*(b))

#undef FUNCNAME
#define FUNCNAME MPIR_PROD
void MPIR_PROD ( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_PROD";
	int i, len = *Len;
#ifdef HAVE_FORTRAN_BINDING
	typedef struct
	{
		float re;
		float im; 
	} s_complex;

	typedef struct
	{
		double re;
		double im; 
	} d_complex;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIR_PROD);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_PROD);

	switch (*type)
	{
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; 
			int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#ifdef HAVE_FORTRAN_BINDING
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
		{
			MPIR_REAL4_CTYPE * restrict a = (MPIR_REAL4_CTYPE *)inoutvec; 
			MPIR_REAL4_CTYPE * restrict b = (MPIR_REAL4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
		{
			MPIR_REAL8_CTYPE * restrict a = (MPIR_REAL8_CTYPE *)inoutvec; 
			MPIR_REAL8_CTYPE * restrict b = (MPIR_REAL8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
		{
			MPIR_REAL16_CTYPE * restrict a = (MPIR_REAL16_CTYPE *)inoutvec; 
			MPIR_REAL16_CTYPE * restrict b = (MPIR_REAL16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
		{
			float * restrict a = (float *)inoutvec; 
			float * restrict b = (float *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
#endif
		{
			double * restrict a = (double *)inoutvec; 
			double * restrict b = (double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE:
		{
			long double * restrict a = (long double *)inoutvec; 
			long double * restrict b = (long double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LPROD(a[i],b[i]);
			break;
		}
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_COMPLEX8:
	case MPI_COMPLEX:
		{
			s_complex * restrict a = (s_complex *)inoutvec; 
			s_complex * restrict b = (s_complex *)invec;
			for ( i=0; i<len; i++ )
			{
				s_complex c;
				c.re = a[i].re; c.im = a[i].im;
				a[i].re = c.re*b[i].re - c.im*b[i].im;
				a[i].im = c.im*b[i].re + c.re*b[i].im;
			}
			break;
		}
	case MPI_COMPLEX16:
	case MPI_DOUBLE_COMPLEX:
		{
			d_complex * restrict a = (d_complex *)inoutvec; 
			d_complex * restrict b = (d_complex *)invec;
			for ( i=0; i<len; i++ )
			{
				d_complex c;
				c.re = a[i].re; c.im = a[i].im;
				a[i].re = c.re*b[i].re - c.im*b[i].im;
				a[i].im = c.im*b[i].re + c.re*b[i].im;
			}
			break;
		}
#endif
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_PROD" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_PROD);
}

#undef FUNCNAME
#define FUNCNAME MPIR_PROD_check_dtype
#undef FCNAME
int MPIR_PROD_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_PROD_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_PROD_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_PROD_CHECK_DTYPE);
	switch (type)
	{
	case MPI_INT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_INTEGER: 
#endif
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
#endif
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE: 
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_COMPLEX: 
	case MPI_DOUBLE_COMPLEX: 
#endif
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
	case MPI_COMPLEX8:
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
	case MPI_COMPLEX16:
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
	case MPI_COMPLEX32:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default: 
		mpi_errno =  MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_PROD" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_PROD_CHECK_DTYPE);
	return mpi_errno;
}

/* 
* In MPI-1, this operation is valid only for  C integer, Fortran integer,
* floating point, and complex data items (4.9.2 Predefined reduce operations)
*/
#define MPIR_LSUM(a,b) ((a)+(b))

#undef FUNCNAME
#define FUNCNAME MPIR_SUM
void MPIR_SUM ( void *invec, void *inoutvec, int *Len, MPI_Datatype *type )
{
	static const wchar_t FCNAME[] = L"MPIR_SUM";
	int i, len = *Len;
#ifdef HAVE_FORTRAN_BINDING
	typedef struct
	{
		float re;
		float im; 
	} s_complex;

	typedef struct
	{
		double re;
		double im; 
	} d_complex;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIR_SUM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_SUM);

	switch (*type)
	{
	case MPI_INT:
		{
			int * restrict a = (int *)inoutvec; 
			int * restrict b = (int *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#ifdef HAVE_FORTRAN_BINDING
	case MPI_INTEGER:
		{
			MPI_Fint * restrict a = (MPI_Fint *)inoutvec; 
			MPI_Fint * restrict b = (MPI_Fint *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
		{
			MPIR_INTEGER1_CTYPE * restrict a = (MPIR_INTEGER1_CTYPE *)inoutvec; 
			MPIR_INTEGER1_CTYPE * restrict b = (MPIR_INTEGER1_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
		{
			MPIR_INTEGER2_CTYPE * restrict a = (MPIR_INTEGER2_CTYPE *)inoutvec; 
			MPIR_INTEGER2_CTYPE * restrict b = (MPIR_INTEGER2_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
		{
			MPIR_INTEGER4_CTYPE * restrict a = (MPIR_INTEGER4_CTYPE *)inoutvec; 
			MPIR_INTEGER4_CTYPE * restrict b = (MPIR_INTEGER4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
		{
			MPIR_INTEGER8_CTYPE * restrict a = (MPIR_INTEGER8_CTYPE *)inoutvec; 
			MPIR_INTEGER8_CTYPE * restrict b = (MPIR_INTEGER8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
		{
			MPIR_INTEGER16_CTYPE * restrict a = (MPIR_INTEGER16_CTYPE *)inoutvec; 
			MPIR_INTEGER16_CTYPE * restrict b = (MPIR_INTEGER16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
		{
			MPIR_REAL4_CTYPE * restrict a = (MPIR_REAL4_CTYPE *)inoutvec; 
			MPIR_REAL4_CTYPE * restrict b = (MPIR_REAL4_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
		{
			MPIR_REAL8_CTYPE * restrict a = (MPIR_REAL8_CTYPE *)inoutvec; 
			MPIR_REAL8_CTYPE * restrict b = (MPIR_REAL8_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
		{
			MPIR_REAL16_CTYPE * restrict a = (MPIR_REAL16_CTYPE *)inoutvec; 
			MPIR_REAL16_CTYPE * restrict b = (MPIR_REAL16_CTYPE *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
	case MPI_UNSIGNED:
		{
			unsigned * restrict a = (unsigned *)inoutvec; 
			unsigned * restrict b = (unsigned *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
	case MPI_LONG:
		{
			long * restrict a = (long *)inoutvec; 
			long * restrict b = (long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG:
		{
			/* case MPI_LONG_LONG_INT: defined to be the same as long_long */
			long long * restrict a = (long long *)inoutvec; 
			long long * restrict b = (long long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif

	case MPI_UNSIGNED_LONG:
		{
			unsigned long * restrict a = (unsigned long *)inoutvec; 
			unsigned long * restrict b = (unsigned long *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
	case MPI_SHORT:
		{
			short * restrict a = (short *)inoutvec; 
			short * restrict b = (short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_SHORT:
		{
			unsigned short * restrict a = (unsigned short *)inoutvec; 
			unsigned short * restrict b = (unsigned short *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
		{
			char * restrict a = (char *)inoutvec; 
			char * restrict b = (char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
	case MPI_UNSIGNED_CHAR:
		{
			unsigned char * restrict a = (unsigned char *)inoutvec; 
			unsigned char * restrict b = (unsigned char *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
		{
			float * restrict a = (float *)inoutvec; 
			float * restrict b = (float *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
#endif
		{
			double * restrict a = (double *)inoutvec; 
			double * restrict b = (double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE:
		{
			long double * restrict a = (long double *)inoutvec; 
			long double * restrict b = (long double *)invec;
			for ( i=0; i<len; i++ )
				a[i] = MPIR_LSUM(a[i],b[i]);
			break;
		}
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_COMPLEX8:
	case MPI_COMPLEX:
		{
			s_complex * restrict a = (s_complex *)inoutvec; 
			s_complex * restrict b = (s_complex *)invec;
			for ( i=0; i<len; i++ )
			{
				a[i].re = MPIR_LSUM(a[i].re ,b[i].re);
				a[i].im = MPIR_LSUM(a[i].im ,b[i].im);
			}
			break;
		}
	case MPI_COMPLEX16:
	case MPI_DOUBLE_COMPLEX:
		{
			d_complex * restrict a = (d_complex *)inoutvec; 
			d_complex * restrict b = (d_complex *)invec;
			for ( i=0; i<len; i++ )
			{
				a[i].re = MPIR_LSUM(a[i].re ,b[i].re);
				a[i].im = MPIR_LSUM(a[i].im ,b[i].im);
			}
			break;
		}
		/* FIXME: Need complex8, 16, and 32 */
#endif
		/* --BEGIN ERROR HANDLING-- */
	default:
		{
			MPICH_PerThread_t *p;
			MPIR_GetPerThread(&p);
			p->op_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_SUM" );
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_SUM);
}

#undef FUNCNAME
#define FUNCNAME MPIR_SUM_check_dtype
#undef FCNAME
int MPIR_SUM_check_dtype ( MPI_Datatype type )
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPIR_SUM_check_dtype";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_SUM_CHECK_DTYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_SUM_CHECK_DTYPE);

	switch (type)
	{
	case MPI_INT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_INTEGER:
#endif
	case MPI_UNSIGNED: 
	case MPI_LONG: 
#if defined(HAVE_LONG_LONG_INT)
	case MPI_LONG_LONG: 
#endif
	case MPI_UNSIGNED_LONG: 
	case MPI_SHORT: 
	case MPI_UNSIGNED_SHORT: 
	case MPI_CHAR: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_CHARACTER: 
#endif
	case MPI_UNSIGNED_CHAR: 
	case MPI_FLOAT: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_REAL: 
#endif
	case MPI_DOUBLE: 
#ifdef HAVE_FORTRAN_BINDING
	case MPI_DOUBLE_PRECISION: 
#endif
#if defined(HAVE_LONG_DOUBLE)
	case MPI_LONG_DOUBLE: 
#endif
#ifdef HAVE_FORTRAN_BINDING
	case MPI_COMPLEX: 
	case MPI_DOUBLE_COMPLEX: 
#endif
		/* The length type can be provided without Fortran, so we do so */
#ifdef MPIR_INTEGER1_CTYPE
	case MPI_INTEGER1:
#endif
#ifdef MPIR_INTEGER2_CTYPE
	case MPI_INTEGER2:
#endif
#ifdef MPIR_INTEGER4_CTYPE
	case MPI_INTEGER4:
#endif
#ifdef MPIR_INTEGER8_CTYPE
	case MPI_INTEGER8:
#endif
#ifdef MPIR_INTEGER16_CTYPE
	case MPI_INTEGER16:
#endif
#ifdef MPIR_REAL4_CTYPE
	case MPI_REAL4:
	case MPI_COMPLEX8:
#endif
#ifdef MPIR_REAL8_CTYPE
	case MPI_REAL8:
	case MPI_COMPLEX16:
#endif
#ifdef MPIR_REAL16_CTYPE
	case MPI_REAL16:
	case MPI_COMPLEX32:
#endif
		break;
		/* --BEGIN ERROR HANDLING-- */
	default:
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opundefined",L"**opundefined %s", L"MPI_SUM" );
		/* --END ERROR HANDLING-- */
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_SUM_CHECK_DTYPE);
	return mpi_errno;
}
