/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#include "mpicomm.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Intercomm_merge
#define MPI_Intercomm_merge PMPI_Intercomm_merge

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Intercomm_merge

/*@
MPI_Intercomm_merge - Creates an intracommuncator from an intercommunicator

Input Parameters:
+ comm - Intercommunicator (handle)
- high - Used to order the groups within comm (logical)
  when creating the new communicator.  This is a boolean value; the group
  that sets high true has its processes ordered `after` the group that sets 
  this value to false.  If all processes in the intercommunicator provide
  the same value, the choice of which group is ordered first is arbitrary.

Output Parameter:
. comm_out - Created intracommunicator (handle)

Notes:
 While all processes may provide the same value for the 'high' parameter,
 this requires the MPI implementation to determine which group of 
 processes should be ranked first. 

.N ThreadSafe

.N Fortran

Algorithm:
.Es
.i Allocate contexts 
.i Local and remote group leaders swap high values
.i Determine the high value.
.i Merge the two groups and make the intra-communicator
.Ee

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Intercomm_create, MPI_Comm_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Intercomm_merge(MPI_Comm intercomm, int high, MPI_Comm *newintracomm)
{
	static const wchar_t FCNAME[] = L"MPI_Intercomm_merge";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_Comm *newcomm_ptr;
	int  local_high, remote_high, i, j, new_size;
	MPIR_Context_id_t new_context_id;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INTERCOMM_MERGE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("comm");  
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INTERCOMM_MERGE);

	MPIU_THREADPRIV_GET;
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(intercomm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( intercomm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (comm_ptr && comm_ptr->comm_kind != MPID_INTERCOMM)
			{
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_COMM, L"**commnotinter", 0 );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_intercomm_merge %C %d %p", intercomm, high, newintracomm); }
	
	/* Make sure that we have a local intercommunicator */
	if (!comm_ptr->local_comm)
	{
		/* Manufacture the local communicator */
		mpi_errno = MPIR_Setup_intercomm_localcomm( comm_ptr );
		MPIU_ERR_CHK_POP(mpi_errno);
	}

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			int acthigh;
			/* Check for consistent valus of high in each local group.
			 The Intel test suite checks for this; it is also an easy
			 error to make */
			acthigh = high ? 1 : 0;   /* Clamp high into 1 or 0 */
			NMPI_Allreduce(mpi_errno, ( MPI_IN_PLACE, &acthigh, 1, MPI_INT, MPI_SUM, comm_ptr->local_comm->handle ));
			MPIU_ERR_CHK_POP(mpi_errno);
			/* acthigh must either == 0 or the size of the local comm */
			if (acthigh != 0 && acthigh != comm_ptr->local_size)
			{
				mpi_errno = MPIR_Err_create_code( mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**notsame", L"**notsame %s %s", L"high", L"MPI_Intercomm_merge" );
				goto fn_fail;
			}
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Find the "high" value of the other group of processes.  This
	   will be used to determine which group is ordered first in
	   the generated communicator.  high is logical */
	local_high = high;
	if (comm_ptr->rank == 0)
	{
		/* This routine allows use to use the collective communication
		   context rather than the point-to-point context. */
		mpi_errno = MPIC_Sendrecv( &local_high, 1, MPI_INT, 0, 0, 
					   &remote_high, 1, MPI_INT, 0, 0, intercomm, 
					   MPI_STATUS_IGNORE );
		MPIU_ERR_CHK_POP(mpi_errno);

		/* If local_high and remote_high are the same, then order is arbitrary.
		   we use the gpids of the rank 0 member of the local and remote
		   groups to choose an order in this case. */
		if (local_high == remote_high)
		{
			int ingpid[2], outgpid[2];
			
			mpi_errno = MPID_GPID_Get( comm_ptr, 0, ingpid );
			MPIU_ERR_CHK_POP(mpi_errno);
			mpi_errno = MPIC_Sendrecv( ingpid, 2, MPI_INT, 0, 1,
						   outgpid, 2, MPI_INT, 0, 1, intercomm, 
						   MPI_STATUS_IGNORE );
			MPIU_ERR_CHK_POP(mpi_errno);

			/* Note that the gpids cannot be the same because we are starting from a valid intercomm */
			if (ingpid[0] < outgpid[0] ||
				(ingpid[0] == outgpid[0] && ingpid[1] < outgpid[1]))
			{
				local_high = 0;
			}
			else
			{
				local_high = 1;
			}
		}
	}

	/* 
	   All processes in the local group now need to get the 
	   value of local_high, which may have changed if both groups
	   of processes had the same value for high
	*/
	NMPI_Bcast(mpi_errno, ( &local_high, 1, MPI_INT, 0, comm_ptr->local_comm->handle ));
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = MPIR_Comm_create( &newcomm_ptr );
	MPIU_ERR_CHK_POP(mpi_errno);

	new_size = comm_ptr->local_size + comm_ptr->remote_size;
    if (local_high)
	{
		newcomm_ptr->context_id	= comm_ptr->recvcontext_id + 2; /* See below */
    }
    else
	{
		newcomm_ptr->context_id	= comm_ptr->context_id + 2; /* See below */
    }
    newcomm_ptr->recvcontext_id	= newcomm_ptr->context_id;
	newcomm_ptr->remote_size  = newcomm_ptr->local_size   = new_size;
	newcomm_ptr->rank		 = -1;
	newcomm_ptr->comm_kind	= MPID_INTRACOMM;

	/* Now we know which group comes first.  Build the new vcr from the existing vcrs */
	mpi_errno = MPID_VCRT_Create( new_size, &newcomm_ptr->vcrt );
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
	MPIU_ERR_CHK_POP(mpi_errno);
	if (local_high)
	{
		/* remote group first */
		j = 0;
		for (i=0; i<comm_ptr->remote_size; i++)
		{
			MPID_VCR_Dup( comm_ptr->vcr[i], &newcomm_ptr->vcr[j++] );
		}
		for (i=0; i<comm_ptr->local_size; i++)
		{
			if (i == comm_ptr->rank) newcomm_ptr->rank = j;
			MPID_VCR_Dup( comm_ptr->local_vcr[i], &newcomm_ptr->vcr[j++] );
		}
	}
	else
	{
		/* local group first */
		j = 0;
		for (i=0; i<comm_ptr->local_size; i++)
		{
			if (i == comm_ptr->rank) newcomm_ptr->rank = j;
			MPID_VCR_Dup( comm_ptr->local_vcr[i], &newcomm_ptr->vcr[j++] );
		}
		for (i=0; i<comm_ptr->remote_size; i++)
		{
			MPID_VCR_Dup( comm_ptr->vcr[i], &newcomm_ptr->vcr[j++] );
		}
	}

	/* We've setup a temporary context id, based on the context id
	   used by the intercomm.  This allows us to perform the allreduce
	   operations within the context id algorithm, since we already
	   have a valid (almost - see comm_create_hook) communicator.
	*/
	/*wprintf( L"About to get context id \n" ); fflush( stdout );*/
	new_context_id = 0;
	mpi_errno = MPIR_Get_contextid( newcomm_ptr, &new_context_id );
	MPIU_ERR_CHK_POP(mpi_errno);
	MPIU_ERR_CHKANDJUMP(new_context_id == 0, mpi_errno, MPI_ERR_OTHER, L"**toomanycomm" );

	/*wprintf( L"Resetting contextid\n" ); fflush( stdout );*/
	newcomm_ptr->context_id = (int16_t)new_context_id;
	newcomm_ptr->recvcontext_id = (int16_t)new_context_id;

	/* Notify the device of this new communicator */
	MPID_Dev_comm_create_hook( newcomm_ptr );
    mpi_errno = MPIR_Comm_commit(newcomm_ptr);
    MPIU_ERR_CHK_POP(mpi_errno);

	*newintracomm = newcomm_ptr->handle;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INTERCOMM_MERGE);
	MPIU_THREAD_SINGLE_CS_EXIT("comm");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_intercomm_merge",
			L"**mpi_intercomm_merge %C %d %p", intercomm, high, newintracomm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}


/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
PMPI_LOCAL int MPIR_CheckDisjointLpids( int [], int, int [], int );
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Intercomm_create
#define MPI_Intercomm_create PMPI_Intercomm_create

/* 128 allows us to handle up to 4k processes */
#ifdef HAVE_ERROR_CHECKING
#define MAX_LPID32_ARRAY 128
#undef FUNCNAME
#define FUNCNAME MPIR_CheckDisjointLpids
PMPI_LOCAL int MPIR_CheckDisjointLpids( int lpids1[], int n1, int lpids2[], int n2 )
{
	static const wchar_t FCNAME[] = L"MPIR_CheckDisjointLpids";
    int i, maxi, idx, bit, maxlpid = -1;
    int mpi_errno = MPI_SUCCESS;
    int32_t lpidmaskPrealloc[MAX_LPID32_ARRAY];
    int32_t *lpidmask;
    MPIU_CHKLMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_MPIR_CHECKDISJOINTLPIDS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_CHECKDISJOINTLPIDS);

	/* Find the max lpid */
    for (i=0; i<n1; i++)
	{
		if (lpids1[i] > maxlpid) maxlpid = lpids1[i];
    }
    for (i=0; i<n2; i++)
	{
		if (lpids2[i] > maxlpid) maxlpid = lpids2[i];
    }

    /* Compute the max index and zero the pids array */
    maxi = (maxlpid + 31) / 32;

    if (maxi >= MAX_LPID32_ARRAY)
	{
		MPIU_CHKLMEM_MALLOC(lpidmask, int32_t*, maxi*sizeof(int32_t), mpi_errno, L"lpidmask");
    }
    else
	{
		lpidmask = lpidmaskPrealloc;
    }
    
    for (i=0; i<maxi; i++) lpidmask[i] = 0;

    /* Set the bits for the first array */
    for (i=0; i<n1; i++)
	{
		idx = lpids1[i] / 32;
		bit = lpids1[i] % 32;
		lpidmask[idx] = lpidmask[idx] | (1 << bit);
    }    

    /* Look for any duplicates in the second array */
    for (i=0; i<n2; i++)
	{
		idx = lpids2[i] / 32;
		bit = lpids2[i] % 32;
		if (lpidmask[idx] & (1 << bit))
		{
			MPIU_ERR_SET1(mpi_errno,MPI_ERR_COMM, L"**dupprocesses", L"**dupprocesses %d", lpids2[i] );
			goto fn_fail;
		}
		/* Add a check on duplicates *within* group 2 */
		lpidmask[idx] = lpidmask[idx] | (1 << bit);
    }

    /* Also fall through for normal return */
 fn_fail:
    MPIU_CHKLMEM_FREEALL();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_CHECKDISJOINTLPIDS);
    return mpi_errno;
}
#endif /* HAVE_ERROR_CHECKING */

#ifndef HAVE_GPID_ROUTINES
/* FIXME: A temporary version for lpids within my comm world */
#undef FUNCNAME
#define FUNCNAME MPID_GPID_GetAllInComm
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
PMPI_LOCAL int MPID_GPID_GetAllInComm( MPID_Comm *comm_ptr, int local_size, int local_gpids[], int *singlePG )
{
	int i;
	int *gpid = local_gpids;
	MPIDI_STATE_DECL(MPID_STATE_MPID_GPID_GETALLINCOMM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_GPID_GETALLINCOMM);
	
	for (i=0; i<comm_ptr->local_size; i++)
	{
		*gpid++ = 0;
		(void)MPID_VCR_Get_lpid( comm_ptr->vcr[i], gpid );
		gpid++;
	}
	*singlePG = 1;
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_GPID_GETALLINCOMM);
	return 0;
}

/* FIXME: A temp for lpids within my comm world */
#undef FUNCNAME
#define FUNCNAME MPID_GPID_ToLpidArray
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
PMPI_LOCAL int MPID_GPID_ToLpidArray( int size, int gpid[], int lpid[] )
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPID_GPID_TOLPIDARRAY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_GPID_TOLPIDARRAY);

	for (i=0; i<size; i++)
	{
		lpid[i] = *++gpid;
		gpid++;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_GPID_TOLPIDARRAY);
	return 0;
}

/* FIXME: for MPI1, all process ids are relative to MPI_COMM_WORLD.
   For MPI2, we'll need to do something more complex */
#undef FUNCNAME
#define FUNCNAME MPID_VCR_CommFromLpids
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
PMPI_LOCAL int MPID_VCR_CommFromLpids( MPID_Comm *newcomm_ptr, int size, const int lpids[] )
{
	MPID_Comm *commworld_ptr;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPID_VCR_COMMFROMLPIDS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_VCR_COMMFROMLPIDS);

	commworld_ptr = MPIR_Process.comm_world;
	/* Setup the communicator's vc table: remote group */
	MPID_VCRT_Create( size, &newcomm_ptr->vcrt );
	MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
	for (i=0; i<size; i++)
	{
		/* For rank i in the new communicator, find the corresponding
		   rank in the comm world (FIXME FOR MPI2) */
		/* wprintf( L"[%d] Remote rank %d has lpid %d\n", 
		   MPIR_Process.comm_world->rank, i, lpids[i] ); */
		if (lpids[i] < commworld_ptr->remote_size)
		{
			MPID_VCR_Dup( commworld_ptr->vcr[lpids[i]], &newcomm_ptr->vcr[i] );
		}
		else
		{
			/* We must find the corresponding vcr for a given lpid */
			/* FIXME: Error */
			MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCR_COMMFROMLPIDS);
			return 1;
			/* MPID_VCR_Dup( ???, &newcomm_ptr->vcr[i] ); */
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCR_COMMFROMLPIDS);
	return 0;
}

#endif /* HAVE_GPID_ROUTINES */

#undef FUNCNAME
#define FUNCNAME MPID_LPID_GetAllInComm
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
PMPI_LOCAL int MPID_LPID_GetAllInComm( MPID_Comm *comm_ptr, int local_size, int local_lpids[] )
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPID_LPID_GETALLINCOMM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_LPID_GETALLINCOMM);

	MPIU_UNREFERENCED_ARG(local_size);

	for (i=0; i<comm_ptr->local_size; i++)
	{
		(void)MPID_VCR_Get_lpid( comm_ptr->vcr[i], &local_lpids[i] );
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_LPID_GETALLINCOMM);
	return 0;
}


#endif /* MPICH_MPI_FROM_PMPI */


#undef FUNCNAME
#define FUNCNAME MPI_Intercomm_create

/*@

MPI_Intercomm_create - Creates an intercommuncator from two intracommunicators

Input Parameters:
+ local_comm - Local (intra)communicator
. local_leader - Rank in local_comm of leader (often 0)
. peer_comm - Communicator used to communicate between a 
			  designated process in the other communicator.  
			  Significant only at the process in 'local_comm' with
			  rank 'local_leader'.
. remote_leader - Rank in peer_comm of remote leader (often 0)
- tag - Message tag to use in constructing intercommunicator; if multiple
  'MPI_Intercomm_creates' are being made, they should use different tags (more
  precisely, ensure that the local and remote leaders are using different
  tags for each 'MPI_intercomm_create').

Output Parameter:
. comm_out - Created intercommunicator

Notes:
   'peer_comm' is significant only for the process designated the 
   'local_leader' in the 'local_comm'.

  The MPI 1.1 Standard contains two mutually exclusive comments on the
  input intracommunicators.  One says that their repective groups must be
  disjoint; the other that the leaders can be the same process.  After
  some discussion by the MPI Forum, it has been decided that the groups must
  be disjoint.  Note that the `reason` given for this in the standard is
  `not` the reason for this choice; rather, the `other` operations on 
  intercommunicators (like 'MPI_Intercomm_merge') do not make sense if the
  groups are not disjoint.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_TAG
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_RANK

.seealso: MPI_Intercomm_merge, MPI_Comm_free, MPI_Comm_remote_group, 
		  MPI_Comm_remote_size

@*/
MPI_DLL_SPEC int MPI_CALL MPI_Intercomm_create(MPI_Comm local_comm, int local_leader, 
						 MPI_Comm peer_comm, int remote_leader, int tag, 
						 MPI_Comm *newintercomm)
{
#undef FCNAME
	static const wchar_t FCNAME[] = L"MPI_Intercomm_create";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_Comm *peer_comm_ptr = NULL;
	MPIR_Context_id_t final_context_id, recvcontext_id;
	int remote_size=0, *remote_lpids=0, *remote_gpids=0, singlePG;
	int local_size, *local_gpids=0, *local_lpids=0;
	int comm_info[3];
	int is_low_group = 0;
	int i;
	MPID_Comm *newcomm_ptr;
	MPIU_CHKLMEM_DECL(4);
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INTERCOMM_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("comm");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INTERCOMM_CREATE);

	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(local_comm, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( local_comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			if (comm_ptr)
			{
				/*  Only check if comm_ptr valid */
				MPIR_ERRTEST_COMM_INTRA(comm_ptr, mpi_errno );
				if ((local_leader < 0 || 
					 local_leader >= comm_ptr->local_size))
				{
					MPIU_ERR_SET2(mpi_errno,MPI_ERR_RANK, L"**ranklocal", L"**ranklocal %d %d", local_leader, comm_ptr->local_size );
				}
				if (comm_ptr->rank == local_leader)
				{
					MPIR_ERRTEST_COMM(peer_comm, mpi_errno);
				}
			}
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_intercomm_create %C %d %C %d %d %p", local_comm, local_leader, peer_comm, remote_leader, tag, newintercomm); }
	
	/*
	 * Error checking for this routine requires care.  Because this
	 * routine is collective over two different sets of processes,
	 * it is relatively easy for the user to try to create an 
	 * intercommunicator from two overlapping groups of processes.
	 * This is made more likely by inconsistencies in the MPI-1
	 * specification (clarified in MPI-2) that seemed to allow
	 * the groups to overlap.  Because of that, we first check that the
	 * groups are in fact disjoint before performing any collective 
	 * operations.  
	 */

	if (comm_ptr->rank == local_leader)
	{

		MPID_Comm_get_ptr( peer_comm, peer_comm_ptr );
#	   ifdef HAVE_ERROR_CHECKING
		{
			MPID_BEGIN_ERROR_CHECKS;
			{
				MPID_Comm_valid_ptr( peer_comm_ptr, mpi_errno );
				/* Note: In MPI 1.0, peer_comm was restricted to 
				   intracommunicators.  In 1.1, it may be any communicator */

				/* In checking the rank of the remote leader, 
				   allow the peer_comm to be in intercommunicator
				   by checking against the remote size */
				if (!mpi_errno && peer_comm_ptr && 
					(remote_leader < 0 || 
					 remote_leader >= peer_comm_ptr->remote_size))
				{
					MPIU_ERR_SET2(mpi_errno,MPI_ERR_RANK, L"**rankremote", L"**rankremote %d %d", local_leader, comm_ptr->local_size );
				}
				/* Check that the local leader and the remote leader are 
				   different processes.  This test requires looking at
				   the lpid for the two ranks in their respective 
				   communicators */
				/* FIXME: Why is this test commented out ? */
/*				if (local_leader == remote_leader) {
					mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_RANK,
													  L"**ranksdistinct", 0 );
													  }*/
				if (mpi_errno) goto fn_fail;
			}
			MPID_END_ERROR_CHECKS;
		}
#	   endif /* HAVE_ERROR_CHECKING */
		
		/* First, exchange the group information.  If we were certain
		   that the groups were disjoint, we could exchange possible 
		   context ids at the same time, saving one communication.
		   But experience has shown that that is a risky assumption.
		*/
		/* Exchange information with my peer.  Use sendrecv */
		local_size = comm_ptr->local_size;

		MPIU_DBG_MSG_FMT(COMM,VERBOSE, (MPIU_DBG_FDEST,L"rank %d sendrecv to rank %d", peer_comm_ptr->rank, remote_leader));
		NMPI_Sendrecv(mpi_errno, ( &local_size,  1, MPI_INT, 
								   remote_leader, tag,
								   &remote_size, 1, MPI_INT, 
								   remote_leader, tag, 
								   peer_comm, MPI_STATUS_IGNORE ));

		MPIU_DBG_MSG_FMT(COMM,VERBOSE, (MPIU_DBG_FDEST, L"local size = %d, remote size = %d", local_size, remote_size ));
		/* With this information, we can now send and receive the global process ids from the peer. */
		MPIU_CHKLMEM_MALLOC(remote_gpids,int*,2*remote_size*sizeof(int), mpi_errno,L"remote_gpids");
		MPIU_CHKLMEM_MALLOC(remote_lpids,int*,remote_size*sizeof(int), mpi_errno,L"remote_lpids");
		MPIU_CHKLMEM_MALLOC(local_gpids,int*,2*local_size*sizeof(int), mpi_errno,L"local_gpids");
		MPIU_CHKLMEM_MALLOC(local_lpids,int*,local_size*sizeof(int), mpi_errno,L"local_lpids");

		mpi_errno = MPID_GPID_GetAllInComm( comm_ptr, local_size, local_gpids, &singlePG );
		MPIU_ERR_CHK_POP(mpi_errno);

		/* Exchange the lpid arrays */
		NMPI_Sendrecv(mpi_errno, ( local_gpids, 2*local_size, MPI_INT, 
					   remote_leader, tag,
					   remote_gpids, 2*remote_size, MPI_INT, 
					   remote_leader, tag, peer_comm, MPI_STATUS_IGNORE ));

		/* Convert the remote gpids to the lpids */
		mpi_errno = MPID_GPID_ToLpidArray( remote_size, remote_gpids, remote_lpids );
		MPIU_ERR_CHK_POP(mpi_errno);

		/* Get our own lpids */
		mpi_errno = MPID_LPID_GetAllInComm( comm_ptr, local_size, local_lpids );
		MPIU_ERR_CHK_POP(mpi_errno);
		
#	   ifdef HAVE_ERROR_CHECKING
		{
			MPID_BEGIN_ERROR_CHECKS;
			{
				/* Now that we have both the local and remote processes,
				   check for any overlap */
				mpi_errno = MPIR_CheckDisjointLpids( local_lpids, local_size, remote_lpids, remote_size );
				MPIU_ERR_CHK_POP(mpi_errno);
			}
			MPID_END_ERROR_CHECKS;
		}
#	   endif /* HAVE_ERROR_CHECKING */
		
		/* Make an arbitrary decision about which group of processs is
		   the low group.  The LEADERS do this by comparing the
		   local process ids of the 0th member of the two groups */
		is_low_group = local_lpids[0] < remote_lpids[0];

		/* At this point, we're done with the local lpids; they'll
		   be freed with the other local memory on exit */

	} /* End of the first phase of the leader communication */

	/* 
	 * Create the contexts.  Each group will have a context for sending 
	 * to the other group. All processes must be involved.  Because 
	 * we know that the local and remote groups are disjoint, this 
	 * step will complete 
	 */
	MPIU_DBG_MSG_FMT(COMM,VERBOSE, (MPIU_DBG_FDEST,L"About to get contextid (commsize=%d) on %d", comm_ptr->local_size, comm_ptr->rank ));
	mpi_errno = MPIR_Get_contextid( comm_ptr, &recvcontext_id );
	MPIU_ERR_CHK_POP(mpi_errno);
	if (recvcontext_id == 0)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**toomanycomm");
	}
	MPIU_DBG_MSG(COMM,VERBOSE,L"Got contextid");

	/* Leaders can now swap context ids and then broadcast the value to the local group of processes */
	if (comm_ptr->rank == local_leader)
	{
		MPIR_Context_id_t remote_context_id;

		NMPI_Sendrecv(mpi_errno, ( &recvcontext_id, 1, MPIR_CONTEXT_ID_T_DATATYPE, remote_leader, tag,
		       &remote_context_id, 1, MPIR_CONTEXT_ID_T_DATATYPE, remote_leader, tag, 
		       peer_comm, MPI_STATUS_IGNORE ));

		final_context_id = remote_context_id;

		/* Now, send all of our local processes the remote_lpids, 
		   along with the final context id */
		comm_info[0] = remote_size;
		comm_info[1] = final_context_id;
		comm_info[2] = is_low_group;
		MPIU_DBG_MSG(COMM,VERBOSE,L"About to bcast on local_comm");
		NMPI_Bcast(mpi_errno, ( comm_info, 3, MPI_INT, local_leader, local_comm ));
		NMPI_Bcast(mpi_errno, ( remote_gpids, 2*remote_size, MPI_INT, local_leader, local_comm ));
		MPIU_DBG_MSG_D(COMM,VERBOSE,L"end of bcast on local_comm of size %d", comm_ptr->local_size );
	}
	else
	{
		/* we're the other processes */
		MPIU_DBG_MSG(COMM,VERBOSE,L"About to receive bcast on local_comm");
		NMPI_Bcast(mpi_errno, ( comm_info, 3, MPI_INT, local_leader, local_comm ));
		remote_size = comm_info[0];
		MPIU_CHKLMEM_MALLOC(remote_gpids,int*,2*remote_size*sizeof(int), mpi_errno,L"remote_gpids");
		MPIU_CHKLMEM_MALLOC(remote_lpids,int*,remote_size*sizeof(int), mpi_errno,L"remote_lpids");
		NMPI_Bcast(mpi_errno, ( remote_gpids, 2*remote_size, MPI_INT, local_leader, local_comm ));

		/* Extract the context and group sign informatin */
		final_context_id = (MPIR_Context_id_t)comm_info[1];
		is_low_group	 = comm_info[2];
	}

	/* Finish up by giving the device the opportunity to update 
	   any other infomration among these processes.  Note that the
	   new intercomm has not been set up; in fact, we haven't yet
	   attempted to set up the connection tables.
	   
	   In the case of the ch3 device, this calls MPID_PG_ForwardPGInfo
	   to ensure that all processes have the information about all
	   process groups.  This must be done before the call 
	   to MPID_GPID_ToLpidArray, as that call needs to know about 
	   all of the process groups.
	*/
#ifdef MPID_ICCREATE_REMOTECOMM_HOOK
	MPID_ICCREATE_REMOTECOMM_HOOK( peer_comm_ptr, comm_ptr, remote_size, remote_gpids, local_leader );
								   
#endif
#if 0
	MPID_PG_ForwardPGInfo( peer_comm_ptr, comm_ptr, remote_size, remote_gpids, local_leader );
#endif

	/* Finally, if we are not the local leader, we need to 
	   convert the remote gpids to local pids.  This must be done
	   after we allow the device to handle any steps that it needs to 
	   take to ensure that all processes contain the necessary process
	   group information */
	if (comm_ptr->rank != local_leader)
	{
		mpi_errno = MPID_GPID_ToLpidArray( remote_size, remote_gpids, remote_lpids );
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* At last, we now have the information that we need to build the intercommunicator */

	/* All processes in the local_comm now build the communicator */

	mpi_errno = MPIR_Comm_create( &newcomm_ptr );
	if (mpi_errno) goto fn_fail;

	newcomm_ptr->context_id   = (int16_t)final_context_id;
	newcomm_ptr->recvcontext_id = (int16_t)recvcontext_id;
	newcomm_ptr->remote_size  = remote_size;
	newcomm_ptr->local_size   = comm_ptr->local_size;
	newcomm_ptr->rank		 = comm_ptr->rank;
	newcomm_ptr->comm_kind	= MPID_INTERCOMM;
	newcomm_ptr->local_comm   = 0;
	newcomm_ptr->is_low_group = is_low_group;

	mpi_errno = MPID_VCR_CommFromLpids( newcomm_ptr, remote_size, remote_lpids );
	MPIU_ERR_CHK_POP(mpi_errno);

	/* Setup the communicator's vc table: local group.  This is just a duplicate of the local_comm's group */
	MPID_VCRT_Create( comm_ptr->local_size, &newcomm_ptr->local_vcrt );
	MPID_VCRT_Get_ptr( newcomm_ptr->local_vcrt, &newcomm_ptr->local_vcr );
	for (i=0; i<comm_ptr->local_size; i++)
	{
		MPID_VCR_Dup( comm_ptr->vcr[i], &newcomm_ptr->local_vcr[i] );
	}

	/* Inherit the error handler (if any) */
	newcomm_ptr->errhandler = comm_ptr->errhandler;
	if (comm_ptr->errhandler)
	{
		MPIR_Errhandler_add_ref( comm_ptr->errhandler );
	}
		
	/* Notify the device of this new communicator */
	MPID_Dev_comm_create_hook( newcomm_ptr );
	mpi_errno = MPIR_Comm_commit(newcomm_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	*newintercomm = newcomm_ptr->handle;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INTERCOMM_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT("comm");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_intercomm_create",
			L"**mpi_intercomm_create %C %d %C %d %d %p", local_comm, local_leader, peer_comm, remote_leader, tag, newintercomm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_test_inter
#define MPI_Comm_test_inter PMPI_Comm_test_inter

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_test_inter

/*@

MPI_Comm_test_inter - Tests to see if a comm is an inter-communicator

Input Parameter:
. comm - communicator to test (handle) 

Output Parameter:
. flag - true if this is an inter-communicator(logical) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_test_inter(MPI_Comm comm, int *flag)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_test_inter";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_TEST_INTER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_TEST_INTER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(flag,L"flag",mpi_errno);

			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_test_inter %C %p", comm, flag); }
	
	*flag = (comm_ptr->comm_kind == MPID_INTERCOMM);
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_TEST_INTER);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_test_inter",
			L"**mpi_comm_test_inter %C %p", comm, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_split
#define MPI_Comm_split PMPI_Comm_split

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_split


typedef struct splittype
{
	int color, key;
} splittype;

PMPI_LOCAL void MPIU_Sort_inttable( splittype *, int );
#ifndef MPICH_MPI_FROM_PMPI
/* Sort the entries in keytable into increasing order by key.  A stable
sort should be used incase the key values are not unique. */
PMPI_LOCAL void MPIU_Sort_inttable( splittype *keytable, int size )
{
	splittype tmp;
	int i, j;

	/* FIXME Bubble sort */
	for (i=0; i<size; i++)
	{
		for (j=i+1; j<size; j++)
		{
			if (keytable[i].key > keytable[j].key)
			{
				tmp	    = keytable[i];
				keytable[i] = keytable[j];
				keytable[j] = tmp;
			}
		}
	}
}
#endif

/*@

MPI_Comm_split - Creates new communicators based on colors and keys

Input Parameters:
+ comm - communicator (handle) 
. color - control of subset assignment (nonnegative integer).  Processes 
with the same color are in the same new communicator 
- key - control of rank assigment (integer)

Output Parameter:
. newcomm - new communicator (handle) 

Notes:
The 'color' must be non-negative or 'MPI_UNDEFINED'.

.N ThreadSafe

.N Fortran

Algorithm:
.vb
1. Use MPI_Allgather to get the color and key from each process
2. Count the number of processes with the same color; create a 
communicator with that many processes.  If this process has
'MPI_UNDEFINED' as the color, create a process with a single member.
3. Use key to order the ranks
4. Set the VCRs using the ordered key values
.ve

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Comm_free
@*/
int MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm *newcomm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_split";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL, *newcomm_ptr, *local_comm_ptr;
	MPI_Comm  local_comm;
	splittype *table, *remotetable=0, *keytable, *remotekeytable=0;
	int       rank, size, remote_size, i, new_size, new_remote_size, 
		first_entry = 0, first_remote_entry = 0,
		*last_ptr;
	MPIR_Context_id_t   new_context_id, remote_context_id = 0;
	MPIU_THREADPRIV_DECL;
	MPIU_CHKLMEM_DECL(4);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SPLIT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SPLIT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}

#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */

	rank        = comm_ptr->rank;
	size        = comm_ptr->local_size;
	remote_size = comm_ptr->remote_size;

	/* Step 1: Find out what color and keys all of the processes have */
	MPIU_CHKLMEM_MALLOC(table,splittype*,size*sizeof(splittype),mpi_errno,L"table");
	table[rank].color = color;
	table[rank].key   = key;

	MPIU_THREADPRIV_GET;

	/* Get the communicator to use in collectives on the local group of processes */
	if (comm_ptr->comm_kind == MPID_INTERCOMM)
	{
		if (!comm_ptr->local_comm)
		{
			MPIR_Setup_intercomm_localcomm( comm_ptr );
		}
		local_comm_ptr = comm_ptr->local_comm;
		local_comm     = local_comm_ptr->handle;
	}
	else
	{
		local_comm_ptr = comm_ptr;
		local_comm     = comm;
	}
	/* Gather information on the local group of processes */
	NMPI_Allgather( mpi_errno, (MPI_IN_PLACE, 2, MPI_INT, table, 2, MPI_INT, local_comm ) );

	/* Step 2: How many processes have our same color? */
	new_size = 0;
	if (color != MPI_UNDEFINED)
	{
		/* Also replace the color value with the index of the *next* value
		in this set.  The integer first_entry is the index of the 
		first element */
		last_ptr = &first_entry;
		for (i=0; i<size; i++)
		{
			/* Replace color with the index in table of the next item
			of the same color.  We use this to efficiently populate 
			the keyval table */
			if (table[i].color == color)
			{
				new_size++;
				*last_ptr = i;
				last_ptr  = &table[i].color;
			}
		}
	}
	/* We don't need to set the last value to -1 because we loop through
	the list for only the known size of the group */

	/* If we're an intercomm, we need to do the same thing for the remote
	table, as we need to know the size of the remote group of the
	same color before deciding to create the communicator */
	if (comm_ptr->comm_kind == MPID_INTERCOMM)
	{
		splittype mypair;
		/* For the remote group, the situation is more complicated.
		We need to find the size of our "partner" group in the
		remote comm.  The easiest way (in terms of code) is for
		every process to essentially repeat the operation for the
		local group - perform an (intercommunicator) all gather
		of the color and rank information for the remote group.
		*/
		MPIU_CHKLMEM_MALLOC(remotetable,splittype*,remote_size*sizeof(splittype),mpi_errno,L"remotetable");
		/* This is an intercommunicator allgather */

		/* We must use a local splittype because we've already modified the
		entries in table to indicate the location of the next rank of the
		same color */
		mypair.color = color;
		mypair.key   = key;
		NMPI_Allgather(mpi_errno, ( &mypair, 2, MPI_INT, remotetable, 2, MPI_INT, comm ) );

		/* Each process can now match its color with the entries in the table */
		new_remote_size = 0;
		last_ptr = &first_remote_entry;
		for (i=0; i<remote_size; i++)
		{
			/* Replace color with the index in table of the next item
			of the same color.  We use this to efficiently populate 
			the keyval table */
			if (remotetable[i].color == color)
			{
				new_remote_size++;
				*last_ptr = i;
				last_ptr  = &remotetable[i].color;
			}
		}
		/* Note that it might find that there a now processes in the remote
		group with the same color.  In that case, COMM_SPLIT will
		return a null communicator */
	}
	else
	{
		/* Set the size of the remote group to the size of our group.
		This simplifies the test below for intercomms with an empty remote
		group (must create comm_null) */
		new_remote_size = new_size;
	}

	/* Step 3: Create the communicator */
	/* Collectively create a new context id.  The same context id will
	be used by each (disjoint) collections of processes.  The
	processes whose color is MPI_UNDEFINED will return the 
	context id to the pool */
	/* In the multi-threaded case, MPIR_Get_contextid assumes that the
	calling routine already holds the single criticial section */
	mpi_errno = MPIR_Get_contextid( local_comm_ptr, &new_context_id );
	MPIU_ERR_CHK_POP(mpi_errno);
	MPIU_Assert(new_context_id != 0);

	/* In the intercomm case, we need to exchange the context ids */
	if (comm_ptr->comm_kind == MPID_INTERCOMM)
	{
		if (comm_ptr->rank == 0)
		{
			mpi_errno = MPIC_Sendrecv( &new_context_id, 1, MPIR_CONTEXT_ID_T_DATATYPE, 0, 0,
				&remote_context_id, 1, MPIR_CONTEXT_ID_T_DATATYPE, 
				0, 0, comm, MPI_STATUS_IGNORE );
			MPIU_ERR_CHK_POP(mpi_errno);
			NMPI_Bcast(mpi_errno, ( &remote_context_id, 1, MPIR_CONTEXT_ID_T_DATATYPE, 0, local_comm ) );
		}
		else
		{
			/* Broadcast to the other members of the local group */
			NMPI_Bcast(mpi_errno, ( &remote_context_id, 1, MPIR_CONTEXT_ID_T_DATATYPE, 0, local_comm ) );
		}
	}

	/* Now, create the new communicator structure if necessary */
	if (color != MPI_UNDEFINED && new_remote_size > 0)
	{
		mpi_errno = MPIR_Comm_create( &newcomm_ptr );
		if (mpi_errno) goto fn_fail;

		newcomm_ptr->context_id	    = new_context_id;
		newcomm_ptr->local_size	    = new_size;
		newcomm_ptr->comm_kind	    = comm_ptr->comm_kind;
		/* Other fields depend on whether this is an intercomm or intracomm */

		/* Step 4: Order the processes by their key values.  Sort the
		list that is stored in table.  To simplify the sort, we 
		extract the table into a smaller array and sort that.
		Also, store in the "color" entry the rank in the input communicator
		of the entry. */
		MPIU_CHKLMEM_MALLOC(keytable,splittype*,new_size*sizeof(splittype),mpi_errno,L"keytable");
		for (i=0; i<new_size; i++)
		{
			keytable[i].key   = table[first_entry].key;
			keytable[i].color = first_entry;
			first_entry	      = table[first_entry].color;
		}

		/* sort key table.  The "color" entry is the rank of the corresponding
		process in the input communicator */
		MPIU_Sort_inttable( keytable, new_size );

		if (comm_ptr->comm_kind == MPID_INTERCOMM)
		{
			MPIU_CHKLMEM_MALLOC(remotekeytable,splittype*,new_remote_size*sizeof(splittype),mpi_errno,L"remote keytable");
			for (i=0; i<new_remote_size; i++)
			{
				remotekeytable[i].key   = remotetable[first_remote_entry].key;
				remotekeytable[i].color = first_remote_entry;
				first_remote_entry	= remotetable[first_remote_entry].color;
			}

			/* sort key table.  The "color" entry is the rank of the
			corresponding process in the input communicator */
			MPIU_Sort_inttable( remotekeytable, new_remote_size );

			MPID_VCRT_Create( new_size, &newcomm_ptr->local_vcrt );
			MPID_VCRT_Get_ptr( newcomm_ptr->local_vcrt, 
				&newcomm_ptr->local_vcr );
			for (i=0; i<new_size; i++)
			{
				MPID_VCR_Dup( comm_ptr->local_vcr[keytable[i].color], &newcomm_ptr->local_vcr[i] );
				if (keytable[i].color == comm_ptr->rank)
				{
					newcomm_ptr->rank = i;
				}
			}

			/* For the remote group, the situation is more complicated.
			We need to find the size of our "partner" group in the
			remote comm.  The easiest way (in terms of code) is for
			every process to essentially repeat the operation for the
			local group - perform an (intercommunicator) all gather
			of the color and rank information for the remote group.
			*/
			/* We apply the same sorting algorithm to the entries that we've
			found to get the correct order of the entries.

			Note that if new_remote_size is 0 (no matching processes with
			the same color in the remote group), then MPI_COMM_SPLIT
			is required to return MPI_COMM_NULL instead of an intercomm 
			with an empty remote group. */

			MPID_VCRT_Create( new_remote_size, &newcomm_ptr->vcrt );
			MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
			for (i=0; i<new_remote_size; i++)
			{
				MPID_VCR_Dup( comm_ptr->vcr[remotekeytable[i].color], &newcomm_ptr->vcr[i] );
			}

			newcomm_ptr->recvcontext_id = remote_context_id;
			newcomm_ptr->remote_size    = new_remote_size;
			newcomm_ptr->local_comm     = 0;
			newcomm_ptr->is_low_group   = comm_ptr->is_low_group;

		}
		else
		{
			/* INTRA Communicator */
			newcomm_ptr->recvcontext_id = new_context_id;
			newcomm_ptr->remote_size    = new_size;
			MPID_VCRT_Create( new_size, &newcomm_ptr->vcrt );
			MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
			for (i=0; i<new_size; i++)
			{
				MPID_VCR_Dup( comm_ptr->vcr[keytable[i].color], &newcomm_ptr->vcr[i] );
				if (keytable[i].color == comm_ptr->rank)
				{
					newcomm_ptr->rank = i;
				}
			}
		}

		/* Inherit the error handler (if any) */
		newcomm_ptr->errhandler = comm_ptr->errhandler;
		if (comm_ptr->errhandler)
		{
			MPIR_Errhandler_add_ref( comm_ptr->errhandler );
		}

		/* Notify the device of this new communicator */
		MPID_Dev_comm_create_hook( newcomm_ptr );
		mpi_errno = MPIR_Comm_commit(newcomm_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);

		*newcomm = newcomm_ptr->handle;
	}
	else
	{
		/* color was MPI_UNDEFINED.  Free the context id */
		*newcomm = MPI_COMM_NULL;
		MPIR_Free_contextid( new_context_id );
	}

	/* ... end of body of routine ... */

fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SPLIT);
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_split",
			L"**mpi_comm_split %C %d %d %p", comm, color, key, newcomm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_size
#define MPI_Comm_size PMPI_Comm_size
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_size

/*@

MPI_Comm_size - Determines the size of the group associated with a communicator

Input Parameter:
. comm - communicator (handle) 

Output Parameter:
. size - number of processes in the group of 'comm'  (integer) 

Notes:

.N NULL

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_size( MPI_Comm comm, int *size ) 
{
	static const wchar_t FCNAME[] = L"MPI_Comm_size";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = 0;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SIZE);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(size,L"size",mpi_errno);
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_size %C %p", comm, size); }
	
	*size = comm_ptr->local_size;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SIZE);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_size",
			L"**mpi_comm_size %C %p", comm, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_set_nameA
#undef MPI_Comm_set_nameW
#define MPI_Comm_set_nameA PMPI_Comm_set_nameA
#define MPI_Comm_set_nameW PMPI_Comm_set_nameW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_set_name

/*@
   MPI_Comm_set_name - Sets the print name for a communicator

   Input Parameters:
+  MPI_Comm comm - communicator to name (handle)
-  char *comm_name - Name for communicator

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_set_nameA(MPI_Comm comm, char *comm_name)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_set_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t name[MPI_MAX_OBJECT_NAME];
	wchar_t *name_ptr = L"";
	size_t len;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SET_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SET_NAMEA);

	/* convert the ASCII string to a wide character string and call the wide character version */

	if (comm_name != NULL)
	{
		error = mbstowcs_s(&len, name, MPI_MAX_OBJECT_NAME, comm_name, _TRUNCATE);
		if (error != 0)
		{
			if (len >= MPI_MAX_OBJECT_NAME)
			{
				/* truncation error */
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"comm_name", (int)len);
				goto fn_fail;
			}
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
		if (len >= MPI_MAX_OBJECT_NAME)
		{
			/* truncation error */
			MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"comm_name", (int)len);
			goto fn_fail;
		}
		if (len > 0 && len < MPI_MAX_OBJECT_NAME)
			name_ptr = name;
	}
	else
	{
		name_ptr = NULL;
	}

	mpi_errno = MPI_Comm_set_nameW(comm, name_ptr);

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SET_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Comm_set_nameW(MPI_Comm comm, wchar_t *comm_name)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_set_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SET_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SET_NAME);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			MPIR_ERRTEST_ARGNULL( comm_name, L"comm_name", mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_set_name %C %s", comm, comm_name); }

	MPIU_Wcsncpy( comm_ptr->name, comm_name, MPI_MAX_OBJECT_NAME );

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SET_NAME);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_set_name",
			L"**mpi_comm_set_name %C %s", comm, comm_name);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_remote_size
#define MPI_Comm_remote_size PMPI_Comm_remote_size

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_remote_size

/*@

MPI_Comm_remote_size - Determines the size of the remote group 
					   associated with an inter-communictor

Input Parameter:
. comm - communicator (handle) 

Output Parameter:
. size - number of processes in the remote group of 'comm'  (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_remote_size(MPI_Comm comm, int *size)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_remote_size";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_REMOTE_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_REMOTE_SIZE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (comm_ptr && comm_ptr->comm_kind != MPID_INTERCOMM)
			{
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_COMM, L"**commnotinter", 0 );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_remote_size %C %p", comm, size); }
	
	*size = comm_ptr->remote_size;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_REMOTE_SIZE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_remote_size",
			L"**mpi_comm_remote_size %C %p", comm, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_remote_group
#define MPI_Comm_remote_group PMPI_Comm_remote_group

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_remote_group

/*@

MPI_Comm_remote_group - Accesses the remote group associated with 
						the given inter-communicator

Input Parameter:
. comm - Communicator (must be an intercommunicator) (handle)

Output Parameter:
. group - remote group of communicator (handle)

Notes:
The user is responsible for freeing the group when it is no longer needed.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM

.seealso MPI_Group_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_remote_group(MPI_Comm comm, MPI_Group *group)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_remote_group";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	int i, lpid, n;
	MPID_Group *group_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_REMOTE_GROUP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("comm");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_REMOTE_GROUP);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (comm_ptr && comm_ptr->comm_kind != MPID_INTERCOMM)
			{
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_COMM, L"**commnotinter", 0 );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_remote_group %C %p", comm, group); }
	
	/* Create a group and populate it with the local process ids */
	if (!comm_ptr->remote_group)
	{
		n = comm_ptr->remote_size;
		mpi_errno = MPIR_Group_create( n, &group_ptr );
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno)
		{
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
		
		for (i=0; i<n; i++)
		{
			group_ptr->lrank_to_lpid[i].lrank = i;
			(void) MPID_VCR_Get_lpid( comm_ptr->vcr[i], &lpid );
			group_ptr->lrank_to_lpid[i].lpid  = lpid;
		}
		group_ptr->size				 = n;
		group_ptr->rank				 = MPI_UNDEFINED;
		group_ptr->idx_of_first_lpid = -1;
		comm_ptr->remote_group   = group_ptr;
	}
	*group = comm_ptr->remote_group->handle;
	MPIR_Group_add_ref( comm_ptr->remote_group );
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_REMOTE_GROUP);
	MPIU_THREAD_SINGLE_CS_EXIT("comm");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_remote_group",
			L"**mpi_comm_remote_group %C %p", comm, group);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_rank
#define MPI_Comm_rank PMPI_Comm_rank
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_rank

/*@

MPI_Comm_rank - Determines the rank of the calling process in the communicator

Input Argument:
. comm - communicator (handle) 

Output Argument:
. rank - rank of the calling process in the group of 'comm'  (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_rank( MPI_Comm comm, int *rank ) 
{
	static const wchar_t FCNAME[] = L"MPI_Comm_rank";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = 0;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_RANK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_RANK);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(rank,L"rank",mpi_errno);
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not value, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_rank %C %p", comm, rank); }
	
	*rank = comm_ptr->rank;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_RANK);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_rank",
			L"**mpi_comm_rank %C %p", comm, rank);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_group
#define MPI_Comm_group PMPI_Comm_group

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_group

/*@

MPI_Comm_group - Accesses the group associated with given communicator

Input Parameter:
. comm - Communicator (handle)

Output Parameter:
. group - Group in communicator (handle)

Notes:
.N COMMNULL

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_group(MPI_Comm comm, MPI_Group *group)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_group";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	int i, lpid, n;
	MPID_Group *group_ptr;
	MPID_VCR   *local_vcr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_GROUP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("comm");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_GROUP);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
	
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_group %C %p", comm, group); }
	
	/* Create a group if necessary and populate it with the local process ids */
	if (!comm_ptr->local_group)
	{
		n = comm_ptr->local_size;
		mpi_errno = MPIR_Group_create( n, &group_ptr );
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno)
		{
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */

		/* Make sure that we get the correct group */
		if (comm_ptr->comm_kind == MPID_INTERCOMM)
		{
			local_vcr = comm_ptr->local_vcr;
		}
		else
			local_vcr = comm_ptr->vcr;
		
		for (i=0; i<n; i++)
		{
			(void) MPID_VCR_Get_lpid( local_vcr[i], &lpid );
			group_ptr->lrank_to_lpid[i].lrank = i;
			group_ptr->lrank_to_lpid[i].lpid  = lpid;
		}
		group_ptr->size				 = n;
		group_ptr->rank				 = comm_ptr->rank;
		group_ptr->idx_of_first_lpid = -1;
		
		comm_ptr->local_group = group_ptr;
	}
	/* FIXME : Add a sanity check that the size of the group is the same as
	   the size of the communicator.  This helps catch corrupted 
	   communicators */

	*group = comm_ptr->local_group->handle;
	MPIU_Object_add_ref( comm_ptr->local_group );

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_GROUP);
	MPIU_THREAD_SINGLE_CS_EXIT("comm");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_group",
			L"**mpi_comm_group %C %p", comm, group);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_get_nameA
#undef MPI_Comm_get_nameW
#define MPI_Comm_get_nameA PMPI_Comm_get_nameA
#define MPI_Comm_get_nameW PMPI_Comm_get_nameW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_get_name

/*@
  MPI_Comm_get_name - Return the print name from the communicator

  Input Parameter:
. comm - Communicator to get name of (handle)

  Output Parameters:
+ comm_name - On output, contains the name of the communicator.  It must
  be an array of size at least 'MPI_MAX_OBJECT_NAME'.
- resultlen - Number of characters in name

 Notes:

.N COMMNULL

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_nameA(MPI_Comm comm, char *comm_name, int *resultlen)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_get_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t name[MPI_MAX_OBJECT_NAME] = L"";
	int len;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_GET_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_GET_NAMEA);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL( comm_name, L"comm_name", mpi_errno );
			MPIR_ERRTEST_ARGNULL( resultlen, L"resultlen", mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	mpi_errno = MPI_Comm_get_nameW(comm, name, &len);
	if (mpi_errno == MPI_SUCCESS)
	{
		if (len == 0)
		{
			*comm_name = '\0';
			*resultlen = 0;
			MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_GET_NAMEA);
			return mpi_errno;
		}

		error = wcstombs_s(&num_converted, comm_name, MPI_MAX_OBJECT_NAME, name, _TRUNCATE);
		if (error != 0)
		{
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
		if (num_converted > 0)
		{
			*resultlen = (int)num_converted - 1;
		}
		else
		{
			/* conversion error */
			mpi_errno = MPI_ERR_INTERN;
		}
	}

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_GET_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_nameW(MPI_Comm comm, wchar_t *comm_name, int *resultlen)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_get_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_GET_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_GET_NAME);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Validate parameters and objects (post conversion) */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );

			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL( comm_name, L"comm_name", mpi_errno );
			MPIR_ERRTEST_ARGNULL( resultlen, L"resultlen", mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_get_name %C %p %p", comm, comm_name, resultlen); }

	/* The user must allocate a large enough section of memory */
	/*
	MPIU_Wcsncpy( comm_name, comm_ptr->name, MPI_MAX_OBJECT_NAME );
	*resultlen = (int)wcslen( comm_name );
	*/
	MPIU_Wcsncpy( comm_name, comm_ptr->name, MPI_MAX_OBJECT_NAME );
	*resultlen = (int)wcslen( comm_name );

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_GET_NAME);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_get_name",
			L"**mpi_comm_get_name %C %p %p", comm, comm_name, resultlen);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_free
#define MPI_Comm_free PMPI_Comm_free

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_free

/*@
MPI_Comm_free - Marks the communicator object for deallocation

Input Parameter:
. comm - Communicator to be destroyed (handle) 

Notes:
This routine `frees` a communicator.  Because the communicator may still
be in use by other MPI routines, the actual communicator storage will not
be freed until all references to this communicator are removed.  For most
users, the effect of this routine is the same as if it was in fact freed
at this time of this call.  

Null Handles:
The MPI 1.1 specification, in the section on opaque objects, explicitly
disallows freeing a null communicator.  The text from the standard is:
.vb
 A null handle argument is an erroneous IN argument in MPI calls, unless an
 exception is explicitly stated in the text that defines the function. Such
 exception is allowed for handles to request objects in Wait and Test calls
 (sections Communication Completion and Multiple Completions ). Otherwise, a
 null handle can only be passed to a function that allocates a new object and
 returns a reference to it in the handle.
.ve

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_free(MPI_Comm *comm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_free";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_FREE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("comm");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_FREE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(*comm, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Get handles to MPI objects. */
	MPID_Comm_get_ptr( *comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			
			/* Cannot free the predefined communicators */
			if (HANDLE_GET_KIND(*comm) == HANDLE_KIND_BUILTIN)
			{
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_COMM, L"**commperm", L"**commperm %s", comm_ptr->name );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_free %p", comm); }
	
	mpi_errno = MPIR_Comm_release(comm_ptr, 0);
	MPIU_ERR_CHK_POP(mpi_errno);
	
	*comm = MPI_COMM_NULL;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_FREE);
	MPIU_THREAD_SINGLE_CS_EXIT("comm");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_free", L"**mpi_comm_free %p", comm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_dup
#define MPI_Comm_dup PMPI_Comm_dup

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_dup

/*@

MPI_Comm_dup - Duplicates an existing communicator with all its cached
			   information

Input Parameter:
. comm - Communicator to be duplicated (handle) 

Output Parameter:
. newcomm - A new communicator over the same group as 'comm' but with a new
  context. See notes.  (handle) 

Notes:
  This routine is used to create a new communicator that has a new
  communication context but contains the same group of processes as
  the input communicator.  Since all MPI communication is performed
  within a communicator (specifies as the group of processes `plus`
  the context), this routine provides an effective way to create a
  private communicator for use by a software module or library.  In
  particular, no library routine should use 'MPI_COMM_WORLD' as the
  communicator; instead, a duplicate of a user-specified communicator
  should always be used.  For more information, see Using MPI, 2nd
  edition. 

  Because this routine essentially produces a copy of a communicator,
  it also copies any attributes that have been defined on the input
  communicator, using the attribute copy function specified by the
  'copy_function' argument to 'MPI_Keyval_create'.  This is
  particularly useful for (a) attributes that describe some property
  of the group associated with the communicator, such as its
  interconnection topology and (b) communicators that are given back
  to the user; the attibutes in this case can track subsequent
  'MPI_Comm_dup' operations on this communicator.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM

.seealso: MPI_Comm_free, MPI_Keyval_create, MPI_Attr_put, MPI_Attr_delete,
 MPI_Comm_create_keyval, MPI_Comm_set_attr, MPI_Comm_delete_attr
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_dup(MPI_Comm comm, MPI_Comm *newcomm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_dup";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL, *newcomm_ptr;
	MPID_Attribute *new_attributes = 0;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_DUP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("comm");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_DUP);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(newcomm, L"newcomm", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_dup %C %p", comm, newcomm); }

	/* Copy attributes, executing the attribute copy functions */
	/* This accesses the attribute dup function through the perprocess
	   structure to prevent comm_dup from forcing the linking of the
	   attribute functions.  The actual function is (by default)
	   MPIR_Attr_dup_list 
	*/
	if (MPIR_Process.attr_dup)
	{
		mpi_errno = MPIR_Process.attr_dup( comm_ptr->handle, comm_ptr->attributes, &new_attributes );
		if (mpi_errno)
		{
			/* Note: The error code returned here should reflect the error code
			   determined by the user routine called during the
			   attribute duplication step.  Adding additional text to the 
			   message associated with the code is allowable; changing the
			   code is not */
			*newcomm = MPI_COMM_NULL;
			goto fn_fail;
		}
	}

	
	/* Generate a new context value and a new communicator structure */ 
	/* We must use the local size, because this is compared to the 
	   rank of the process in the communicator.  For intercomms, 
	   this must be the local size */
	mpi_errno = MPIR_Comm_copy( comm_ptr, comm_ptr->local_size, &newcomm_ptr );
	MPIU_ERR_CHK_POP(mpi_errno);

	newcomm_ptr->attributes = new_attributes;
	*newcomm = newcomm_ptr->handle;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_DUP);
	MPIU_THREAD_SINGLE_CS_EXIT("comm");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_dup",
			L"**mpi_comm_dup %C %p", comm, newcomm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_create
#define MPI_Comm_create PMPI_Comm_create

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_create

/*@

MPI_Comm_create - Creates a new communicator

Input Parameters:
+ comm - communicator (handle) 
- group - group, which is a subset of the group of 'comm'  (handle) 

Output Parameter:
. comm_out - new communicator (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_GROUP

.seealso: MPI_Comm_free
@*/
int MPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *newcomm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_create";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	int i, j, n, *mapping = 0, *remote_mapping = 0, remote_size = -1;
	MPIR_Context_id_t new_context_id;
	MPID_Comm *newcomm_ptr;
	MPID_Group *group_ptr;
	MPIU_CHKLMEM_DECL(3);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("comm");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_CREATE);

	/* Validate parameters, and convert MPI object handles to object pointers */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;

		MPID_Comm_get_ptr( comm, comm_ptr );

		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */

			MPIR_ERRTEST_GROUP(group, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;

		MPID_Group_get_ptr( group, group_ptr );

		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Check the group ptr */
			MPID_Group_valid_ptr( group_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   else
	{
		MPID_Comm_get_ptr( comm, comm_ptr );
		MPID_Group_get_ptr( group, group_ptr );
	}
#   endif

	/* ... body of routine ...  */

	/* Create a new communicator from the specified group members */

	/* If there is a context id cache in oldcomm, use it here.  Otherwise,
	use the appropriate algorithm to get a new context id. 
	Creating the context id is collective over the *input* communicator,
	so it must be created before we decide if this process is a 
	member of the group */
	/* In the multi-threaded case, MPIR_Get_contextid assumes that the
	calling routine already holds the single criticial section */
	if (comm_ptr->comm_kind == MPID_INTERCOMM)
	{
		if (!comm_ptr->local_comm)
		{
			MPIR_Setup_intercomm_localcomm( comm_ptr );
		}
		mpi_errno = MPIR_Get_contextid( comm_ptr->local_comm, &new_context_id );
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		mpi_errno = MPIR_Get_contextid( comm_ptr, &new_context_id );
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	MPIU_Assert(new_context_id != 0);

	/* Make sure that the processes for this group are contained within
	the input communicator.  Also identify the mapping from the ranks of 
	the old communicator to the new communicator.
	We do this by matching the lpids of the members of the group
	with the lpids of the members of the input communicator.
	It is an error if the group contains a reference to an lpid that 
	does not exist in the communicator.

	An important special case is groups (and communicators) that
	are subsets of MPI_COMM_WORLD.  This this case, the lpids are
	exactly the same as the ranks in comm world.  Currently, we
	don't take this into account, but if the code to handle the general 
	case is too messy, we'll add this in.
	*/
	if (group_ptr->rank != MPI_UNDEFINED)
	{
		MPID_VCR *vcr, *mapping_vcr=0, *remote_mapping_vcr=0;
		int      vcr_size;
		int      subsetOfWorld = 0;

		/* Default choice of mapping vcr */
		mapping_vcr        = comm_ptr->local_vcr;
		remote_mapping_vcr = comm_ptr->vcr;
		n = group_ptr->size;
		MPIU_CHKLMEM_MALLOC(mapping,int*,n*sizeof(int),mpi_errno,L"mapping");
		if (comm_ptr->comm_kind == MPID_INTERCOMM)
		{
			vcr      = comm_ptr->local_vcr;
			vcr_size = comm_ptr->local_size;
		}
		else
		{
			vcr      = comm_ptr->vcr;
			vcr_size = comm_ptr->remote_size;
		}

		/* Optimize for groups contained within MPI_COMM_WORLD. */
		if (comm_ptr->comm_kind == MPID_INTRACOMM)
		{
			int wsize;
			subsetOfWorld = 1;
			wsize         = MPIR_Process.comm_world->local_size;
			MPIR_Group_setup_lpid_list( group_ptr );
			for (i=0; i<n; i++) {
				int g_lpid = group_ptr->lrank_to_lpid[i].lpid;

				/* This Mapping is relative to comm world */
				MPIU_DBG_MSG_FMT(COMM,VERBOSE,(MPIU_DBG_FDEST, L"comm-create - mapping into world[%d] = %d\n", i, g_lpid ));
				if (g_lpid < wsize)
				{
					mapping[i] = g_lpid;
				}
				else
				{
					subsetOfWorld = 0; 
					break;
				}
			}
		}
		MPIU_DBG_MSG_D(COMM,VERBOSE,L"Are we using the world subset %d\n", subsetOfWorld );
		if (subsetOfWorld)
		{
			/* Override the vcr to be used with the mapping array. */
			mapping_vcr = MPIR_Process.comm_world->vcr;
#           ifdef HAVE_ERROR_CHECKING
			{
				MPID_BEGIN_ERROR_CHECKS;
				{
					int idx;
					mpi_errno = MPIR_GroupCheckVCRSubset( group_ptr, vcr_size, vcr, &idx );
					if (mpi_errno) goto fn_fail;
				}
				MPID_END_ERROR_CHECKS;
			}
#           endif
		}
		else
		{
			for (i=0; i<n; i++)
			{
				/* Mapping[i] is the rank in the communicator of the process 
				that is the ith element of the group */
				/* We use the appropriate vcr, depending on whether this is
				an intercomm (use the local vcr) or an intracomm 
				(remote vcr) 
				Note that this is really the local mapping for intercomm
				and remote mapping for the intracomm */
				/* FIXME : BUBBLE SORT */
				mapping[i] = -1;
				for (j=0; j<vcr_size; j++)
				{
					int comm_lpid;
					MPID_VCR_Get_lpid( vcr[j], &comm_lpid );
					if (comm_lpid == group_ptr->lrank_to_lpid[i].lpid)
					{
						mapping[i] = j;
						break;
					}
				}
				MPIU_ERR_CHKANDJUMP1(mapping[i] == -1,mpi_errno,MPI_ERR_GROUP,L"**groupnotincomm", L"**groupnotincomm %d", i );
			}
		}

		MPIU_DBG_MSG(COMM,VERBOSE,L"End of group in comm check...");

		if (comm_ptr->comm_kind == MPID_INTRACOMM)
		{
			/* If this is an intra comm, we've really determined the remote mapping */
			MPIU_DBG_MSG(COMM,VERBOSE,L"Swapping remote and local mapping" );
			remote_mapping     = mapping;
			remote_mapping_vcr = mapping_vcr;
			remote_size        = n;
			mapping            = 0;
			mapping_vcr        = 0;
		}

		/* Get the new communicator structure and context id */

		mpi_errno = MPIR_Comm_create( &newcomm_ptr );
		MPIU_ERR_CHK_POP(mpi_errno);

		newcomm_ptr->context_id	    = new_context_id;
		newcomm_ptr->rank	    = group_ptr->rank;
		newcomm_ptr->comm_kind	    = comm_ptr->comm_kind;
		/* Since the group has been provided, let the new communicator know about the group */
		newcomm_ptr->local_comm	    = 0;
		newcomm_ptr->local_group    = group_ptr;
		MPIU_Object_add_ref( group_ptr );

		if (comm_ptr->comm_kind == MPID_INTRACOMM)
		{
			newcomm_ptr->remote_group   = group_ptr;
			MPIU_Object_add_ref( group_ptr );
			newcomm_ptr->recvcontext_id = new_context_id;
			newcomm_ptr->remote_size    = newcomm_ptr->local_size = n;
		}
		else
		{
			/* comm_ptr->comm_kind == MPID_INTERCOMM */ 
			int rinfo[2];
			newcomm_ptr->local_size   = group_ptr->size;
			newcomm_ptr->remote_group = 0;
			/* There is an additional step.  We must communicate the information
			on the local context id and the group members, given by the ranks
			so that the remote process can construct the appropriate VCRT 
			First we exchange group sizes and context ids.  Then the 
			ranks in the remote group, from which the remote VCRT can 
			be constructed.  We can't use NMPI_Sendrecv since we need to 
			use the "collective" context in the original intercommunicator */
			if (comm_ptr->rank == 0)
			{
				int info[2];
				info[0] = new_context_id;
				info[1] = group_ptr->size;

				mpi_errno = MPIC_Sendrecv(info, 2, MPI_INT, 
					0, 0, 
					rinfo, 2, MPI_INT,
					0, 0, comm, MPI_STATUS_IGNORE );
				MPIU_ERR_CHK_POP(mpi_errno);
				newcomm_ptr->recvcontext_id = (MPIR_Context_id_t)rinfo[0];
				remote_size                 = rinfo[1];

				MPIU_CHKLMEM_MALLOC(remote_mapping,int*, remote_size*sizeof(int), mpi_errno, L"remote_mapping");

				/* Populate and exchange the ranks */
				mpi_errno = MPIC_Sendrecv( mapping, group_ptr->size, MPI_INT, 
					0, 0, 
					remote_mapping, remote_size, MPI_INT,
					0, 0, comm, MPI_STATUS_IGNORE );
				MPIU_ERR_CHK_POP(mpi_errno);

				/* Broadcast to the other members of the local group */
				NMPI_Bcast(mpi_errno, ( rinfo, 2, MPI_INT, 0, comm_ptr->local_comm->handle ) );
				NMPI_Bcast(mpi_errno, ( remote_mapping, remote_size, MPI_INT, 0, comm_ptr->local_comm->handle ) );
			}
			else
			{
				/* The other processes */
				/* Broadcast to the other members of the local group */
				NMPI_Bcast(mpi_errno, ( rinfo, 2, MPI_INT, 0, comm_ptr->local_comm->handle ) );
				newcomm_ptr->recvcontext_id = (MPIR_Context_id_t)rinfo[0];
				remote_size                 = rinfo[1];
				MPIU_CHKLMEM_MALLOC(remote_mapping,int*, remote_size*sizeof(int), mpi_errno, L"remote_mapping");
				NMPI_Bcast(mpi_errno, ( remote_mapping, remote_size, MPI_INT, 0, comm_ptr->local_comm->handle ) );
			}
			newcomm_ptr->remote_size    = remote_size;
			/* Now, everyone has the remote_mapping, and can apply that to 
			the vcr table.  That step is done below, as it is the 
			same step as needed for the intracommunicator */

			/* Setup the communicator's local vc table from the local mapping. */
			n = group_ptr->size;
			MPID_VCRT_Create( n, &newcomm_ptr->local_vcrt );
			MPID_VCRT_Get_ptr( newcomm_ptr->local_vcrt, &newcomm_ptr->local_vcr );
			for (i=0; i<n; i++)
			{
				/* For rank i in the new communicator, find the corresponding rank in the input communicator */
				MPID_VCR_Dup( mapping_vcr[mapping[i]], &newcomm_ptr->local_vcr[i] );
			}
		}

		/* Setup the communicator's vc table.  This is for the remote group,
		which is the same as the local group for intracommunicators */
		n = remote_size;
		MPID_VCRT_Create( n, &newcomm_ptr->vcrt );
		MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
		for (i=0; i<n; i++)
		{
			/* For rank i in the new communicator, find the corresponding rank in the input communicator */
			MPIU_DBG_MSG_FMT(COMM,VERBOSE,(MPIU_DBG_FDEST, L"dupping %d into %d using %p\n", remote_mapping[i], i, remote_mapping_vcr ));
			MPID_VCR_Dup( remote_mapping_vcr[remote_mapping[i]], &newcomm_ptr->vcr[i] );
		}

		/* Notify the device of this new communicator */
		MPID_Dev_comm_create_hook( newcomm_ptr );
		mpi_errno = MPIR_Comm_commit(newcomm_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);

		*newcomm = newcomm_ptr->handle;
	}
	else
	{
		/* This process is not in the group */
		MPIR_Free_contextid( new_context_id );
		*newcomm = MPI_COMM_NULL;

		/* Dummy to complete collective ops in the intercomm case */
		if (comm_ptr->comm_kind == MPID_INTERCOMM)
		{
			int rinfo[2], ic_remote_size, *ic_remote_mapping = 0;
			NMPI_Bcast(mpi_errno, ( rinfo, 2, MPI_INT, 0, comm_ptr->local_comm->handle ) );
			ic_remote_size = rinfo[1];
			MPIU_CHKLMEM_MALLOC(ic_remote_mapping,int*, ic_remote_size*sizeof(int), mpi_errno, L"ic_remote_mapping");
			NMPI_Bcast(mpi_errno, ( ic_remote_mapping, ic_remote_size, MPI_INT, 0, comm_ptr->local_comm->handle ) );
		}
	}
    
    /* ... end of body of routine ... */

    /* mpi_errno = MPID_Comm_create(); */
    if (mpi_errno != MPI_SUCCESS) goto fn_fail;

  fn_exit:
    MPIU_CHKLMEM_FREEALL();
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT(L"comm");
    return mpi_errno;

  fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
	    mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_create",
	    L"**mpi_comm_create %C %G %p", comm, group, newcomm);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_compare
#define MPI_Comm_compare PMPI_Comm_compare

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_compare

/*@

MPI_Comm_compare - Compares two communicators

Input Parameters:
+ comm1 - comm1 (handle) 
- comm2 - comm2 (handle) 

Output Parameter:
. result - integer which is 'MPI_IDENT' if the contexts and groups are the
same, 'MPI_CONGRUENT' if different contexts but identical groups, 'MPI_SIMILAR'
if different contexts but similar groups, and 'MPI_UNEQUAL' otherwise

Using 'MPI_COMM_NULL' with 'MPI_Comm_compare':

It is an error to use 'MPI_COMM_NULL' as one of the arguments to
'MPI_Comm_compare'.  The relevant sections of the MPI standard are 

$(2.4.1 Opaque Objects)
A null handle argument is an erroneous 'IN' argument in MPI calls, unless an
exception is explicitly stated in the text that defines the function.

$(5.4.1. Communicator Accessors)
where there is no text in 'MPI_COMM_COMPARE' allowing a null handle.

.N ThreadSafe
(To perform the communicator comparisions, this routine may need to
allocate some memory.  Memory allocation is not interrupt-safe, and hence
this routine is only thread-safe.)

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_compare(MPI_Comm comm1, MPI_Comm comm2, int *result)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_compare";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr1 = NULL;
	MPID_Comm *comm_ptr2 = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_COMPARE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_COMPARE);

	MPIU_THREADPRIV_GET;
	
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm1, mpi_errno);
			MPIR_ERRTEST_COMM(comm2, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Get handles to MPI objects. */
	MPID_Comm_get_ptr( comm1, comm_ptr1 );
	MPID_Comm_get_ptr( comm2, comm_ptr2 );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr1, mpi_errno );
			MPID_Comm_valid_ptr( comm_ptr2, mpi_errno );
			MPIR_ERRTEST_ARGNULL( result, L"result", mpi_errno );
			/* If comm_ptr1 or 2 is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_compare %C %C %p", comm1, comm2, result); }

	if (comm_ptr1->comm_kind != comm_ptr2->comm_kind)
	{
		*result = MPI_UNEQUAL;
	}
	else if (comm1 == comm2)
	{
		*result = MPI_IDENT;
	}
	else if (comm_ptr1->comm_kind == MPID_INTRACOMM)
	{
		MPI_Group group1, group2;

		NMPI_Comm_group(mpi_errno, ( comm1, &group1 ));
		NMPI_Comm_group(mpi_errno, ( comm2, &group2 ));
		NMPI_Group_compare(mpi_errno, ( group1, group2, result ));
		/* If the groups are the same but the contexts are different, then
		   the communicators are congruent */
		if (*result == MPI_IDENT) 
			*result = MPI_CONGRUENT;
		NMPI_Group_free(mpi_errno, ( &group1 ));
		NMPI_Group_free(mpi_errno, ( &group2 ));
	}
	else
	{ 
		/* INTER_COMM */
		int	   lresult, rresult;
		MPI_Group group1, group2;
		MPI_Group rgroup1, rgroup2;
		
		/* Get the groups and see what their relationship is */
		NMPI_Comm_group(mpi_errno, (comm1, &group1));
		NMPI_Comm_group(mpi_errno, (comm2, &group2));
		NMPI_Group_compare(mpi_errno, ( group1, group2, &lresult ));

		NMPI_Comm_remote_group(mpi_errno, (comm1, &rgroup1));
		NMPI_Comm_remote_group(mpi_errno, (comm2, &rgroup2));
		NMPI_Group_compare(mpi_errno, ( rgroup1, rgroup2, &rresult ));

		/* Choose the result that is "least" strong. This works 
		   due to the ordering of result types in mpi.h */
		(*result) = (rresult > lresult) ? rresult : lresult;

		/* They can't be identical since they're not the same
		   handle, they are congruent instead */
		if ((*result) == MPI_IDENT)
		  (*result) = MPI_CONGRUENT;

		/* Free the groups */
		NMPI_Group_free(mpi_errno, (&group1));
		NMPI_Group_free(mpi_errno, (&group2));
		NMPI_Group_free(mpi_errno, (&rgroup1));
		NMPI_Group_free(mpi_errno, (&rgroup2));
	}
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_COMPARE);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_HANDLING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_compare",
			L"**mpi_comm_compare %C %C %p", comm1, comm2, result);
	}
#   endif
	/* Use whichever communicator is non-null if possible */
	mpi_errno = MPIR_Err_return_comm( comm_ptr1 ? comm_ptr1 : comm_ptr2, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
