/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#include "errcodes.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_set_errhandler
#define MPI_Win_set_errhandler PMPI_Win_set_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_set_errhandler

/*@
   MPI_Win_set_errhandler - Set window error handler

   Input Parameters:
+ win - window (handle) 
- errhandler - new error handler for window (handle) 
.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_set_errhandler(MPI_Win win, MPI_Errhandler errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Win_set_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	int  in_use;
	MPID_Errhandler *errhan_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_SET_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan"); 
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_SET_ERRHANDLER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			MPIR_ERRTEST_ERRHANDLER(errhandler, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	MPID_Errhandler_get_ptr( errhandler, errhan_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not value, it will be reset to null */

			if (HANDLE_GET_KIND(errhandler) != HANDLE_KIND_BUILTIN)
			{
				MPID_Errhandler_valid_ptr( errhan_ptr,mpi_errno );
				/* Also check for a valid errhandler kind */
				if (!mpi_errno)
				{
					if (errhan_ptr->kind != MPID_WIN)
					{
						mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**errhandnotwin", NULL );
					}
				}
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_set_errhandler %W %E", win, errhandler); }
	
	if (win_ptr->errhandler != NULL)
	{
		if (HANDLE_GET_KIND(errhandler) != HANDLE_KIND_BUILTIN)
		{
			MPIU_Object_release_ref(win_ptr->errhandler,&in_use);
			if (!in_use)
			{
				MPID_Errhandler_free( win_ptr->errhandler );
			}
		}
	}
	
	MPIU_Object_add_ref(errhan_ptr);
	win_ptr->errhandler = errhan_ptr;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_SET_ERRHANDLER);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_set_errhandler",
			L"**mpi_win_set_errhandler %W %E", win, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_win(win_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_get_errhandler
#define MPI_Win_get_errhandler PMPI_Win_get_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_get_errhandler

/*@
   MPI_Win_get_errhandler - Get the error handler for the MPI RMA window

   Input Parameter:
. win - window (handle) 

   Output Parameter:
. errhandler - error handler currently associated with window (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_errhandler(MPI_Win win, MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Win_get_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_GET_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_GET_ERRHANDLER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(errhandler,L"errhandler",mpi_errno);
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_get_errhandler %W %p", win, errhandler); }
	
	if (win_ptr->errhandler)
	{
		*errhandler = win_ptr->errhandler->handle;
		MPIU_Object_add_ref(win_ptr->errhandler);
	}
	else
	{
		/* Use the default */
		*errhandler = MPI_ERRORS_ARE_FATAL;
	}
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_GET_ERRHANDLER);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_get_errhandler",
			L"**mpi_win_get_errhandler %W %p", win, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_win(win_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_create_errhandler
#define MPI_Win_create_errhandler PMPI_Win_create_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_create_errhandler

/*@
   MPI_Win_create_errhandler - Create an error handler for use with MPI window
   objects

   Input Parameter:
. function - user defined error handling procedure (function) 

   Output Parameter:
. errhandler - MPI error handler (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_create_errhandler(MPI_Win_errhandler_fn *function, 
							  MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Win_create_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Errhandler *errhan_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_CREATE_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_CREATE_ERRHANDLER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(function, L"function", mpi_errno);
			MPIR_ERRTEST_ARGNULL(errhandler, L"errhandler", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_create_errhandler %p %p", function, errhandler); }
	
	errhan_ptr = (MPID_Errhandler *)MPIU_Handle_obj_alloc( &MPID_Errhandler_mem );
	MPIU_ERR_CHKANDJUMP1(!errhan_ptr,mpi_errno,MPI_ERR_OTHER,L"**nomem", L"**nomem %s", L"MPI_Errhandler");
	*errhandler				 = errhan_ptr->handle;
	errhan_ptr->language = MPID_LANG_C;
	errhan_ptr->kind		 = MPID_WIN;
	MPIU_Object_set_ref(errhan_ptr,1);
	errhan_ptr->errfn.C_Win_Handler_function = function;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_CREATE_ERRHANDLER);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_create_errhandler", 
			L"**mpi_win_create_errhandler %p %p", function, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_call_errhandler
#define MPI_Win_call_errhandler PMPI_Win_call_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_call_errhandler

/*@
   MPI_Win_call_errhandler - Call the error handler installed on a 
   window object

   Input Parameters:
+ win - window with error handler (handle) 
- errorcode - error code (integer) 

 Note:
 Assuming the input parameters are valid, when the error handler is set to
 MPI_ERRORS_RETURN, this routine will always return MPI_SUCCESS.
 
.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_call_errhandler(MPI_Win win, int errorcode)
{
	static const wchar_t FCNAME[] = L"MPI_Win_call_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_CALL_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_CALL_ERRHANDLER);

	MPIU_THREADPRIV_GET;
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_call_errhandler %W %d", win, errorcode); }
	
	if (!win_ptr->errhandler || win_ptr->errhandler->handle == MPI_ERRORS_ARE_FATAL)
	{
		mpi_errno = MPIR_Err_return_win( win_ptr, L"MPI_Win_call_errhandler", errorcode );
		goto fn_exit;
	}

	if (win_ptr->errhandler->handle == MPI_ERRORS_RETURN)
	{
		/* MPI_ERRORS_RETURN should always return MPI_SUCCESS */
		goto fn_exit;
	}

	/* The user error handler may make calls to MPI routines, so the nesting
	 * counter must be incremented before the handler is called */
	MPIR_Nest_incr();
	
	switch (win_ptr->errhandler->language)
	{
	case MPID_LANG_C:
#ifdef HAVE_CXX_BINDING
	case MPID_LANG_CXX:
#endif
		(*win_ptr->errhandler->errfn.C_Win_Handler_function)( &win_ptr->handle, &errorcode );
		break;
#ifdef HAVE_FORTRAN_BINDING
	case MPID_LANG_FORTRAN90:
	case MPID_LANG_FORTRAN:
		(*win_ptr->errhandler->errfn.F77_Handler_function)( (MPI_Fint *)&win_ptr->handle, &errorcode );
		break;
#endif
	}
	
	MPIR_Nest_decr();
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_CALL_ERRHANDLER);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_call_errhandler", 
			L"**mpi_win_call_errhandler %W %d", win, errorcode);
	}
#   endif
	mpi_errno = MPIR_Err_return_win(win_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_File_set_errhandler
#define MPI_File_set_errhandler PMPI_File_set_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_set_errhandler

/*@
   MPI_File_set_errhandler - Set the error handler for an MPI file

   Input Parameters:
+ file - MPI file (handle) 
- errhandler - new error handler for file (handle) 

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_set_errhandler(MPI_File file, MPI_Errhandler errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_File_set_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	int in_use;
	MPID_Errhandler *errhan_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SET_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SET_ERRHANDLER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE_NULL_OK(file, mpi_errno);
			MPIR_ERRTEST_ERRHANDLER(errhandler, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	MPID_File_get_ptr( file, file_ptr );
	MPID_Errhandler_get_ptr( errhandler, errhan_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_File_valid_ptr( file_ptr, mpi_errno );

			if (HANDLE_GET_KIND(errhandler) != HANDLE_KIND_BUILTIN)
			{
				MPID_Errhandler_valid_ptr( errhan_ptr,mpi_errno );
				/* Also check for a valid errhandler kind */
				if (!mpi_errno)
				{
					if (errhan_ptr->kind != MPID_FILE)
					{
						mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**errhandnotfile", NULL );
					}
				}
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_set_errhandler %F %E", file, errhandler); }

	if (file_ptr->errhandler != NULL)
	{
		if (HANDLE_GET_KIND(errhandler) != HANDLE_KIND_BUILTIN)
		{
			MPIU_Object_release_ref(file_ptr->errhandler,&in_use);
			if (!in_use)
			{
				MPID_Errhandler_free( file_ptr->errhandler );
			}
		}
	}

	MPIU_Object_add_ref(errhan_ptr);
	file_ptr->errhandler = errhan_ptr;

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SET_ERRHANDLER);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_file_set_errhandler",
			L"**mpi_file_set_errhandler %F %E", file, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_file( file_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#ifndef MPICH_MPI_FROM_PMPI
/* Export this routine only once (if we need to compile this file twice
   to get the PMPI and MPI versions without weak symbols */
#undef FUNCNAME
#define FUNCNAME MPIR_Get_file_error_routine
void MPIR_Get_file_error_routine( MPI_Errhandler e, void (**c)(MPI_File *, int *, ...), int *kind )
{
	static const wchar_t FCNAME[] = L"MPIR_Get_file_error_routine";
	MPID_Errhandler *e_ptr = 0;
	int mpi_errno = MPI_SUCCESS;

	/* Convert the MPI_Errhandler into an MPID_Errhandler */

	if (!e)
	{
		*c = 0;
		*kind = 1; /* Use errors return as the default */
	}
	else
	{
		MPIR_ERRTEST_ERRHANDLER(e,mpi_errno);
		if (mpi_errno != MPI_SUCCESS)
		{
			/* FIXME: We need an error return */
			*c = 0;
			*kind = 1;
			return;
		}
		MPID_Errhandler_get_ptr(e,e_ptr);
		if (!e_ptr)
		{
			/* FIXME: We need an error return */
			*c = 0;
			*kind = 1;
			return;
		}
		if (e_ptr->handle == MPI_ERRORS_RETURN)
		{
			*c = 0;
			*kind = 1;
		}
		else if (e_ptr->handle == MPI_ERRORS_ARE_FATAL)
		{
			*c = 0;
			*kind = 0;
		}
		else
		{
			*c = e_ptr->errfn.C_File_Handler_function;
			*kind = 2;
			/* If the language is C++, we need to use a special call
			   interface.  This is MPIR_File_call_cxx_errhandler.  
			   See file_call_errhandler.c */
#ifdef HAVE_CXX_BINDING
			if (e_ptr->language == MPID_LANG_CXX) *kind = 3;
#endif
		}
	}
 fn_fail:
	return;
}
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_File_get_errhandler
#define MPI_File_get_errhandler PMPI_File_get_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_errhandler

/*@
   MPI_File_get_errhandler - Get the error handler attached to a file

   Input Parameter:
. file - MPI file (handle) 

   Output Parameter:
. errhandler - handler currently associated with file (handle) 

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_errhandler(MPI_File file, MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_File_get_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_ERRHANDLER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE_NULL_OK(file, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(errhandler,L"errhandler",mpi_errno);
			MPID_File_valid_ptr(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_errhandler %F %p", file, errhandler); }

	if (file_ptr->errhandler)
	{
		*errhandler = file_ptr->errhandler->handle;
		MPIU_Object_add_ref(file_ptr->errhandler);
	}
	else
	{
		*errhandler = MPI_ERRORS_RETURN;
	}

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_ERRHANDLER);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_errhandler",
			L"**mpi_file_get_errhandler %F %p", file, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_file( file_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_File_create_errhandler
#define MPI_File_create_errhandler PMPI_File_create_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_create_errhandler

/*@
   MPI_File_create_errhandler - Create a file error handler

   Input Parameter:
. function - user defined error handling procedure (function) 

   Output Parameter:
. errhandler - MPI error handler (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_create_errhandler(MPI_File_errhandler_fn *function, MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_File_create_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Errhandler *errhan_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_CREATE_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_CREATE_ERRHANDLER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(function, L"function", mpi_errno);
			MPIR_ERRTEST_ARGNULL(errhandler, L"errhandler", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_create_errhandler %p %p", function, errhandler); }
	
	errhan_ptr = (MPID_Errhandler *)MPIU_Handle_obj_alloc( &MPID_Errhandler_mem );
	MPIU_ERR_CHKANDJUMP1(!errhan_ptr,mpi_errno,MPI_ERR_OTHER,L"**nomem", L"**nomem %s", L"MPI_Errhandler");
	*errhandler				 = errhan_ptr->handle;
	errhan_ptr->language = MPID_LANG_C;
	errhan_ptr->kind		 = MPID_FILE;
	MPIU_Object_set_ref(errhan_ptr,1);
	errhan_ptr->errfn.C_File_Handler_function = function;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_CREATE_ERRHANDLER);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_create_errhandler",
			L"**mpi_file_create_errhandler %p %p", function, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_File_call_errhandler
#define MPI_File_call_errhandler PMPI_File_call_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_call_errhandler
/*@
   MPI_File_call_errhandler - Call the error handler installed on a 
   file

   Input Parameters:
+ file - MPI file with error handler (handle) 
- errorcode - error code (integer) 

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_FILE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_call_errhandler(MPI_File file, int errorcode)
{
	static const wchar_t FCNAME[] = L"MPI_File_call_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_CALL_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_CALL_ERRHANDLER);

	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE_NULL_OK(file, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Convert MPI object handles to object pointers */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* If file_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_call_errhandler %F %d", file, errorcode); }
	
	if (!file_ptr->errhandler || file_ptr->errhandler->handle == MPI_ERRORS_ARE_FATAL)
	{
		mpi_errno = MPIR_Err_return_file( file_ptr, FCNAME, errorcode );
		goto fn_exit;
	}

	if (file_ptr->errhandler->handle == MPI_ERRORS_RETURN)
	{
		/* MPI_ERRORS_RETURN should always return MPI_SUCCESS */
		goto fn_exit;
	}

	/* The user error handler may make calls to MPI routines, so the nesting
	 * counter must be incremented before the handler is called */
	MPIR_Nest_incr();

	switch (file_ptr->errhandler->language)
	{
	case MPID_LANG_C:
#ifdef HAVE_CXX_BINDING
	case MPID_LANG_CXX:
#endif
		(*file_ptr->errhandler->errfn.C_File_Handler_function)( &file_ptr->handle, &errorcode );
		break;
#ifdef HAVE_FORTRAN_BINDING
	case MPID_LANG_FORTRAN90:
	case MPID_LANG_FORTRAN:
		(*file_ptr->errhandler->errfn.F77_Handler_function)( (MPI_Fint *)&file_ptr->handle, &errorcode );
		break;
#endif
	}
	
	MPIR_Nest_decr();

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_CALL_ERRHANDLER);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_file_call_errhandler",
			L"**mpi_file_call_errhandler %F %d", file, errorcode);
	}
#   endif
	mpi_errno = MPIR_Err_return_file( file_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#ifndef MPICH_MPI_FROM_PMPI
/* This is a glue routine that can be used by ROMIO 
   (see mpi-io/glue/mpich2/mpio_err.c) to properly invoke the C++ 
   error handler */
#undef FUNCNAME
#define FUNCNAME MPIR_File_call_cxx_errhandler
int MPIR_File_call_cxx_errhandler( MPI_File *fh, int *errorcode, void (*c_errhandler)(MPI_File *, int *, ... ) )
{
	/* ROMIO will contain a reference to this routine, so if there is 
	   no C++ support, it will never be called but it must be availavle. */
#ifdef HAVE_CXX_BINDING
	(*MPIR_Process.cxx_call_errfn)( 1, fh, errorcode, (void(*)(void))c_errhandler );
	/* The C++ code throws an exception if the error handler 
	   returns something other than MPI_SUCCESS. There is no "return"
	   of an error code. This code mirrors that in errutil.c */
	*errorcode = MPI_SUCCESS;
#endif
	return *errorcode;
}
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Error_stringA
#undef MPI_Error_stringW
#define MPI_Error_stringA PMPI_Error_stringA
#define MPI_Error_stringW PMPI_Error_stringW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Error_string

/*@
   MPI_Error_string - Return a string for a given error code

Input Parameters:
. errorcode - Error code returned by an MPI routine or an MPI error class

Output Parameter:
+ string - Text that corresponds to the errorcode 
- resultlen - Length of string 

Notes:  Error codes are the values return by MPI routines (in C) or in the
'ierr' argument (in Fortran).  These can be converted into error classes 
with the routine 'MPI_Error_class'.  

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Error_stringA(int errorcode, char *string, int *resultlen)
{
	static const wchar_t FCNAME[] = L"MPI_Error_stringA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t wstring[MPI_MAX_ERROR_STRING];
	wchar_t *wstring_ptr = NULL;
	int len;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ERROR_STRINGA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ERROR_STRINGA);

	if (string != NULL)
		wstring_ptr = wstring;

	mpi_errno = MPI_Error_stringW(errorcode, wstring_ptr, &len);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (string != NULL)
	{
		string[0] = '\0';
		error = wcstombs_s(&num_converted, string, MPI_MAX_ERROR_STRING, wstring, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= MPI_MAX_ERROR_STRING)
			{
				/* truncation error */
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"string", (int)num_converted);
				goto fn_fail;
			}
			MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation");
			goto fn_fail;
		}
		if (resultlen != NULL)
			*resultlen = (int)num_converted - 1;
	}

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ERROR_STRINGA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Error_stringW(int errorcode, wchar_t *string, int *resultlen)
{
	static const wchar_t FCNAME[] = L"MPI_Error_string";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ERROR_STRING);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ERROR_STRING);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(string,L"string",mpi_errno);
			MPIR_ERRTEST_ARGNULL(resultlen,L"resultlen",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_error_string %d %s %p", errorcode, string, resultlen); }

	MPIR_Err_disable_error_stack_chop(); /* Don't wrap error strings returned to the user */
	MPIR_Err_get_string( errorcode, string, MPI_MAX_ERROR_STRING, NULL );
	*resultlen = (int)wcslen( string );
	MPIR_Err_restore_error_stack_chop();

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ERROR_STRING);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_error_string",
			L"**mpi_error_string %d %s %p", errorcode, string, resultlen);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Error_class
#define MPI_Error_class PMPI_Error_class

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Error_class

/*@
   MPI_Error_class - Converts an error code into an error class

Input Parameter:
. errorcode - Error code returned by an MPI routine 

Output Parameter:
. errorclass - Error class associated with 'errorcode'

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Error_class(int errorcode, int *errorclass)
{
	static const wchar_t FCNAME[] = L"MPI_Error_class";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ERROR_CLASS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ERROR_CLASS);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(errorclass,L"errorclass",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_error_class %d %p", errorcode, errorclass); }
	
	/* We include the dynamic bit because this is needed to fully describe the dynamic error classes */
	*errorclass = errorcode & (ERROR_CLASS_MASK | ERROR_DYN_MASK);
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ERROR_CLASS);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_error_class",
			L"**mpi_error_class %d %p", errorcode, errorclass);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Errhandler_get
#define MPI_Errhandler_get PMPI_Errhandler_get

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Errhandler_get

/*@
  MPI_Errhandler_get - Gets the error handler for a communicator

Input Parameter:
. comm - communicator to get the error handler from (handle) 

Output Parameter:
. errhandler - MPI error handler currently associated with communicator
(handle) 

.N ThreadSafe

.N Fortran

Note on Implementation:

The MPI Standard was unclear on whether this routine required the user to call 
'MPI_Errhandler_free' once for each call made to this routine in order to 
free the error handler.  After some debate, the MPI Forum added an explicit
statement that users are required to call 'MPI_Errhandler_free' when the
return value from this routine is no longer needed.  This behavior is similar
to the other MPI routines for getting objects; for example, 'MPI_Comm_group' 
requires that the user call 'MPI_Group_free' when the group returned
by 'MPI_Comm_group' is no longer needed.

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Errhandler_get(MPI_Comm comm, MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Errhandler_get";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ERRHANDLER_GET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ERRHANDLER_GET);

	MPIU_THREADPRIV_GET;
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr; if comm_ptr is not value, it will be reset to null */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			MPIR_ERRTEST_ARGNULL(errhandler, L"errhandler", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_errhandler_get %C %p", comm, errhandler); }

	NMPI_Comm_get_errhandler(mpi_errno, ( comm, errhandler ));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ERRHANDLER_GET);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_errhandler_get",
			L"**mpi_errhandler_get %C %p", comm, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Errhandler_free
#define MPI_Errhandler_free PMPI_Errhandler_free

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Errhandler_free

/*@
  MPI_Errhandler_free - Frees an MPI-style errorhandler

Input Parameter:
. errhandler - MPI error handler (handle).  Set to 'MPI_ERRHANDLER_NULL' on 
exit.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Errhandler_free(MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Errhandler_free";
	int mpi_errno = MPI_SUCCESS;
	MPID_Errhandler *errhan_ptr = NULL;
	int in_use;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ERRHANDLER_FREE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ERRHANDLER_FREE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(*errhandler, L"errhandler", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			MPIR_ERRTEST_ERRHANDLER(*errhandler, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Errhandler_get_ptr( *errhandler, errhan_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Errhandler_valid_ptr( errhan_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_errhandler_free %p", errhandler); }
	
	MPIU_Object_release_ref( errhan_ptr,&in_use);
	if (!in_use)
	{
		MPIU_Handle_obj_free( &MPID_Errhandler_mem, errhan_ptr );
	}
	*errhandler = MPI_ERRHANDLER_NULL;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ERRHANDLER_FREE);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_errhandler_free",
			L"**mpi_errhandler_free %p", errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Errhandler_create
#define MPI_Errhandler_create PMPI_Errhandler_create

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Errhandler_create

/*@
  MPI_Errhandler_create - Creates an MPI-style errorhandler

Input Parameter:
. function - user defined error handling procedure 

Output Parameter:
. errhandler - MPI error handler (handle) 

Notes:
The MPI Standard states that an implementation may make the output value 
(errhandler) simply the address of the function.  However, the action of 
'MPI_Errhandler_free' makes this impossible, since it is required to set the
value of the argument to 'MPI_ERRHANDLER_NULL'.  In addition, the actual
error handler must remain until all communicators that use it are 
freed.

.N ThreadSafe

.N Deprecated
The replacement routine for this function is 'MPI_Comm_create_errhandler'.

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Comm_create_errhandler, MPI_Errhandler_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Errhandler_create(MPI_Handler_function *function, 
						  MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Errhandler_create";
	int mpi_errno = MPI_SUCCESS;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ERRHANDLER_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ERRHANDLER_CREATE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(function, L"function", mpi_errno);
			MPIR_ERRTEST_ARGNULL(errhandler, L"errhandler", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_errhandler_create %p %p", function, errhandler); }

	MPIU_THREADPRIV_GET;

	NMPI_Comm_create_errhandler(mpi_errno, ( function, errhandler ));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ERRHANDLER_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_errhandler_create",
			L"**mpi_errhandler_create %p %p", function, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Errhandler_set
#define MPI_Errhandler_set PMPI_Errhandler_set

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Errhandler_set

/*@
  MPI_Errhandler_set - Sets the error handler for a communicator

Input Parameters:
+ comm - communicator to set the error handler for (handle) 
- errhandler - new MPI error handler for communicator (handle) 

.N ThreadSafe

.N Deprecated
The replacement for this routine is 'MPI_Comm_set_errhandler'.

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG

.seealso: MPI_Comm_set_errhandler, MPI_Errhandler_create, MPI_Comm_create_errhandler
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Errhandler_set(MPI_Comm comm, MPI_Errhandler errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Errhandler_set";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ERRHANDLER_SET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ERRHANDLER_SET);

	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Errhandler * errhan_ptr = NULL;
			
			/* Validate comm_ptr; if comm_ptr is not value, it will be reset to null */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			MPIR_ERRTEST_ERRHANDLER(errhandler, mpi_errno);
			if (mpi_errno) goto fn_fail;

			MPID_Errhandler_get_ptr( errhandler, errhan_ptr );
			MPID_Errhandler_valid_ptr( errhan_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_errhandler_set %C %E", comm, errhandler); }

	NMPI_Comm_set_errhandler(mpi_errno, ( comm, errhandler ));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ERRHANDLER_SET);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_errhandler_set",
			L"**mpi_errhandler_set %C %E", comm, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_set_errhandler
#define MPI_Comm_set_errhandler PMPI_Comm_set_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_set_errhandler

/*@
   MPI_Comm_set_errhandler - Set the error handler for a communicator

   Input Parameters:
+ comm - communicator (handle) 
- errhandler - new error handler for communicator (handle) 

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_OTHER

.seealso MPI_Comm_get_errhandler, MPI_Comm_call_errhandler
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_set_errhandler(MPI_Comm comm, MPI_Errhandler errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_set_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	int in_use;
	MPID_Errhandler *errhan_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SET_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SET_ERRHANDLER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Errhandler_get_ptr( errhandler, errhan_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr; if comm_ptr is not valid, it will be reset to null */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );

			if (HANDLE_GET_KIND(errhandler) != HANDLE_KIND_BUILTIN)
			{
				MPID_Errhandler_valid_ptr( errhan_ptr, mpi_errno );
			}
			/* FIXME: Add a check that this is a comm errhandler.
			   This test should be a macro so that it can be used
			   in all calls, along with similar calls for win and file */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_set_errhandler %C %E", comm, errhandler); }
	
	/* We don't bother with the case where the errhandler is NULL; 
	   in this case, the error handler was the original, MPI_ERRORS_ARE_FATAL,
	   which is builtin and can never be freed. */
	if (comm_ptr->errhandler != NULL)
	{
		if (HANDLE_GET_KIND(errhandler) != HANDLE_KIND_BUILTIN)
		{
			MPIU_Object_release_ref(comm_ptr->errhandler,&in_use);
			if (!in_use)
			{
				MPID_Errhandler_free( comm_ptr->errhandler );
			}
		}
	}
	
	MPIR_Errhandler_add_ref(errhan_ptr);
	comm_ptr->errhandler = errhan_ptr;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SET_ERRHANDLER);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_set_errhandler",
			L"**mpi_comm_set_errhandler %C %E", comm, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_get_errhandler
#define MPI_Comm_get_errhandler PMPI_Comm_get_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_get_errhandler

/*@
   MPI_Comm_get_errhandler - Get the error handler attached to a communicator

   Input Parameter:
. comm - communicator (handle) 

   Output Parameter:
. errhandler - handler currently associated with communicator (handle) 

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_errhandler(MPI_Comm comm, MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_get_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_GET_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_GET_ERRHANDLER);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr; if comm_ptr is not valid, it will be reset to null  */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			MPIR_ERRTEST_ARGNULL(errhandler,L"errhandler",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	
	/* Check for default error handler */
	if (!comm_ptr->errhandler)
	{
		*errhandler = MPI_ERRORS_ARE_FATAL;
	}
	else
	{
		*errhandler = comm_ptr->errhandler->handle;
		MPIR_Errhandler_add_ref(comm_ptr->errhandler);
	}
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_GET_ERRHANDLER);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_get_errhandler",
			L"**mpi_comm_get_errhandler %C %p", comm, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_create_errhandler
#define MPI_Comm_create_errhandler PMPI_Comm_create_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_create_errhandler

/*@
   MPI_Comm_create_errhandler - Create a communicator error handler

   Input Parameter:
. function - user defined error handling procedure (function) 

   Output Parameter:
. errhandler - MPI error handler (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_create_errhandler(MPI_Comm_errhandler_fn *function, MPI_Errhandler *errhandler)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_create_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Errhandler *errhan_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_CREATE_ERRHANDLER);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_CREATE_ERRHANDLER);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(function, L"function", mpi_errno);
			MPIR_ERRTEST_ARGNULL(errhandler, L"errhandler", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_create_errhandler %p %p", function, errhandler); }
	
	errhan_ptr = (MPID_Errhandler *)MPIU_Handle_obj_alloc( &MPID_Errhandler_mem );
	MPIU_ERR_CHKANDJUMP(!errhan_ptr,mpi_errno,MPI_ERR_OTHER,L"**nomem");

	*errhandler				 = errhan_ptr->handle;
	errhan_ptr->language = MPID_LANG_C;
	errhan_ptr->kind		 = MPID_COMM;
	MPIU_Object_set_ref(errhan_ptr,1);
	errhan_ptr->errfn.C_Comm_Handler_function = function;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_CREATE_ERRHANDLER);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_create_errhandler",
			L"**mpi_comm_create_errhandler %p %p", function, errhandler);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_call_errhandler
#define MPI_Comm_call_errhandler PMPI_Comm_call_errhandler

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_call_errhandler

/*@
   MPI_Comm_call_errhandler - Call the error handler installed on a 
   communicator

 Input Parameters:
+ comm - communicator with error handler (handle) 
- errorcode - error code (integer) 

 Note:
 Assuming the input parameters are valid, when the error handler is set to
 MPI_ERRORS_RETURN, this routine will always return MPI_SUCCESS.
 
.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_call_errhandler(MPI_Comm comm, int errorcode)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_call_errhandler";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_CALL_ERRHANDLER);
	
	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_CALL_ERRHANDLER);

	MPIU_THREADPRIV_GET;
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr; if comm_ptr is not value, it will be reset to null */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;

			if (comm_ptr->errhandler)
			{
				MPIR_ERRTEST_ERRHANDLER(comm_ptr->errhandler->handle,mpi_errno);
				if (mpi_errno) goto fn_fail;
			}
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_call_errhandler %C %d", comm, errorcode); }
	
	/* Check for predefined error handlers */
	if (!comm_ptr->errhandler || 
		comm_ptr->errhandler->handle == MPI_ERRORS_ARE_FATAL)
	{
		mpi_errno = MPIR_Err_return_comm( comm_ptr, L"MPI_Comm_call_errhandler", errorcode );
		goto fn_exit;
	}

	if (comm_ptr->errhandler->handle == MPI_ERRORS_RETURN)
	{
		/* MPI_ERRORS_RETURN should always return MPI_SUCCESS */
		goto fn_exit;
	}

    /* Check for the special case of errors-throw-exception.  In this case
       return the error code; the C++ wrapper will cause an exception to be thrown.
       */
#if defined(HAVE_CXX_BINDING) && defined(MPIR_ERRORS_THROW_EXCEPTIONS)
    if (comm_ptr->errhandler->handle == MPIR_ERRORS_THROW_EXCEPTIONS)
	{
		mpi_errno = errorcode;
		goto fn_exit;
    }
#endif

	/* The user error handler may make calls to MPI routines, so the nesting
	 * counter must be incremented before the handler is called */
	MPIR_Nest_incr();
	
	/* Process any user-defined error handling function */
	switch (comm_ptr->errhandler->language)
	{
	case MPID_LANG_C:
#ifdef HAVE_CXX_BINDING
	case MPID_LANG_CXX:
#endif
		(*comm_ptr->errhandler->errfn.C_Comm_Handler_function)( &comm_ptr->handle, &errorcode );
		break;
#ifdef HAVE_FORTRAN_BINDING
	case MPID_LANG_FORTRAN90:
	case MPID_LANG_FORTRAN:
		(*comm_ptr->errhandler->errfn.F77_Handler_function)( (MPI_Fint *)&comm_ptr->handle, &errorcode );
		break;
#endif
	}

	MPIR_Nest_decr();
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_CALL_ERRHANDLER);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_call_errhandler",
			L"**mpi_comm_call_errhandler %C %d", comm, errorcode);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Add_error_class
#define MPI_Add_error_class PMPI_Add_error_class

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Add_error_class

/*@
   MPI_Add_error_class - Add an MPI error class to the known classes

   Output Parameter:
.  errorclass - New error class

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Add_error_class(int *errorclass)
{
	static const wchar_t FCNAME[] = L"MPI_Add_error_class";
	int mpi_errno = MPI_SUCCESS;
	int new_class;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ADD_ERROR_CLASS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ADD_ERROR_CLASS);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(errorclass, L"errorclass", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_add_error_class %p", errorclass); }
	
	new_class = MPIR_Err_add_class( );
	MPIU_ERR_CHKANDJUMP(new_class<0,mpi_errno,MPI_ERR_OTHER,L"**noerrclasses");

	*errorclass = ERROR_DYN_MASK | new_class;
	MPIR_Setmax( &MPIR_Process.attrs.lastusedcode, *errorclass );
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ADD_ERROR_CLASS);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_add_error_class",
			L"**mpi_add_error_class %p", errorclass);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Add_error_stringA
#undef MPI_Add_error_stringW
#define MPI_Add_error_stringA PMPI_Add_error_stringA
#define MPI_Add_error_stringW PMPI_Add_error_stringW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Add_error_string

/*@
   MPI_Add_error_string - Associates an error string with an MPI error code or 
   class

   Input Parameters:
+ errorcode - error code or class (integer) 
- string - text corresponding to errorcode (string) 

   Notes:
The string must be no more than 'MPI_MAX_ERROR_STRING' characters long. 
The length of the string is as defined in the calling language. 
The length of the string does not include the null terminator in C or C++.  
Note that the string is 'const' even though the MPI standard does not 
specify it that way.

According to the MPI-2 standard, it is erroneous to call 'MPI_Add_error_string'
for an error code or class with a value less than or equal 
to 'MPI_ERR_LASTCODE'.  Thus, you cannot replace the predefined error messages
with this routine.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Add_error_stringA(int errorcode, char *string)
{
	static const wchar_t FCNAME[] = L"MPI_Add_error_stringA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wstring = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ADD_ERROR_STRINGA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ADD_ERROR_STRINGA);

	if (string != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(string, &wstring);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = MPI_Add_error_stringW(errorcode, wstring);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wstring != NULL)
	{
		MPIU_Free(wstring);
	}
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ADD_ERROR_STRINGA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Add_error_stringW(int errorcode, wchar_t *string)
{
	static const wchar_t FCNAME[] = L"MPI_Add_error_string";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ADD_ERROR_STRING);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ADD_ERROR_STRING);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(string,L"string",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_add_error_string %d %s", errorcode, string); }
	
	mpi_errno = MPIR_Err_set_msg( errorcode, (const wchar_t *)string );
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ADD_ERROR_STRING);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_add_error_string",
			L"**mpi_add_error_string %d %s", errorcode, string);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Add_error_code
#define MPI_Add_error_code PMPI_Add_error_code

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Add_error_code

/*@
   MPI_Add_error_code - Add and MPI error code to an MPI error class

   Input Parameter:
.  errorclass - Error class to add an error code.

   Output Parameter:
.  errorcode - New error code for this error class.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Add_error_code(int errorclass, int *errorcode)
{
	static const wchar_t FCNAME[] = L"MPI_Add_error_code";
	int mpi_errno = MPI_SUCCESS;
	int new_code;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ADD_ERROR_CODE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("errhan");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ADD_ERROR_CODE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* FIXME: verify that errorclass is a dynamic class */
			MPIR_ERRTEST_ARGNULL(errorcode, L"errorcode", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_add_error_code %d %p", errorclass, errorcode); }
	
	new_code = MPIR_Err_add_code( errorclass );
	MPIU_ERR_CHKANDJUMP(new_code<0,mpi_errno,MPI_ERR_OTHER,L"**noerrcodes");

	*errorcode = new_code;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ADD_ERROR_CODE);
	MPIU_THREAD_SINGLE_CS_EXIT("errhan");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_add_error_code",
			L"**mpi_add_error_code %d %p", errorclass, errorcode);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
