/*   Copyright (C) 2009 Deino Software. */

/* stdarg is required to handle the variable argument lists for MPIR_Err_create_code */
#include <stdarg.h>
/* Define USE_ERR_CODE_VALIST to get the prototype for the valist version
   of MPIR_Err_create_code in mpiimpl.h */
#define USE_ERR_CODE_VALIST

#include "mpiimpl.h"
#include "errcodes.h"

/* defmsg is generated automatically from the source files and contains
   all of the error messages, both the generic and specific.  Depending 
   on the value of MPICH_ERROR_MSG_LEVEL, different amounts of message
   information will be included from the file defmsg.h */
#include "defmsg.h" 

/* Define length of the the maximum error message line (or string with 
   newlines?) */
#define MAX_ERRMSG_STRING 4096

/* stdio is needed for vsprintf and _vsnwprintf_s */
#include <stdio.h>

/*
 * Structure of this file
 *
 * This file contains several groups of routines user for error handling
 * and reporting.  
 *
 * The first group provides memory for the MPID_Errhandler objects 
 * and the routines to free and manipulate them
 *
 * MPIR_Err_return_xxx - For each of the MPI types on which an 
 * error handler can be defined, there is an MPIR_Err_return_xxx routine
 * that determines what error handler function to call and whether to 
 * abort the program.  The comm and win versions are here; ROMIO
 * provides its own routines for invoking the error handlers for Files.
 *
 * The next group of code handles the error messages.  There are three
 * options, controlled by the value of MPICH_ERROR_MSG_LEVEL. 
 *
 * MPICH_ERROR_MSG_NONE - No text messages at all
 * MPICH_ERROR_MSG_CLASS - Only messages for the MPI error classes
 * MPICH_ERROR_MSG_GENERIC - Only predefiend messages for the MPI error codes
 * MPICH_ERROR_MSG_ALL - Instance specific error messages (and error message
 *					   stack)
 *
 * In only the latter (MPICH_ERROR_MSG_ALL) case are instance-specific
 * messages maintained (including the error message "stack" that you may
 * see mentioned in various places.  In the other cases, an error code 
 * identifies a fixed message string (unless MPICH_ERROR_MSG_NONE,
 * when there are no strings) from the "generic" strings defined in defmsg.h
 *
 * A major subgroup in this section is the code to handle the instance-specific
 * messages (MPICH_ERROR_MSG_ALL only).  
 * 
 */

/* A few prototypes.  These routines are called from the MPIR_Err_return 
   routines.  checkValidErrcode depends on the MPICH_ERROR_MSG_LEVEL */
static int convertErrcodeToIndexes( int errcode, int *ring_idx, int *ring_id,
									int *generic_idx );
static int checkValidErrcode( int error_class, const wchar_t fcname[], int *errcode );
#if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL
static int ErrGetInstanceString( int errorcode, wchar_t *msg, int num_remaining, 
								 MPIR_Err_get_class_string_func_t fn );
static void MPIR_Err_stack_init( void );
static int checkForUserErrcode( int errcode );
#else
/* We only need special handling for user error codes when we support the
   error message stack */
#define checkForUserErrcode(_a) _a
#endif
static int checkErrcodeIsValid( int errcode );
static const wchar_t *ErrcodeInvalidReasonStr( int reason );
static void handleFatalError( MPID_Comm *comm_ptr, const wchar_t fcname[], int errcode );
static const wchar_t *get_class_msg( int error_class );

static int MPIR_Err_print_stack_flag = TRUE;
static int MPIR_Err_abort_on_error = FALSE;
static int MPIR_Err_chop_error_stack = TRUE;
static int MPIR_Err_chop_width = 79;

/* ------------------------------------------------------------------------- */
/* Provide the MPID_Errhandler space and the routines to free and set them
   from C++  */
/* ------------------------------------------------------------------------- */
/*
 * Error handlers.  These are handled just like the other opaque objects
 * in MPICH
 */

#ifndef MPID_ERRHANDLER_PREALLOC 
#define MPID_ERRHANDLER_PREALLOC 8
#endif

/* Preallocated errorhandler objects */
MPID_Errhandler MPID_Errhandler_builtin[2] = 
		  { { MPI_ERRORS_ARE_FATAL, 0},
			{ MPI_ERRORS_RETURN, 0} }; 
MPID_Errhandler MPID_Errhandler_direct[MPID_ERRHANDLER_PREALLOC] = { {0} };
MPIU_Object_alloc_t MPID_Errhandler_mem = { 0, 0, 0, 0, MPID_ERRHANDLER, 
											sizeof(MPID_Errhandler), 
											MPID_Errhandler_direct,
											MPID_ERRHANDLER_PREALLOC, };

#undef FUNCNAME
#define FUNCNAME MPID_Errhandler_free
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPID_Errhandler_free(MPID_Errhandler *errhan_ptr)
{
	MPIDI_STATE_DECL(MPID_STATE_MPID_ERRHANDLER_FREE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_ERRHANDLER_FREE);
	MPIU_Handle_obj_free(&MPID_Errhandler_mem, errhan_ptr);
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_ERRHANDLER_FREE);
}

#ifdef HAVE_CXX_BINDING
#undef FUNCNAME
#define FUNCNAME MPIR_Errhandler_set_cxx
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Errhandler_set_cxx( MPI_Errhandler errhand, void (*errcall)(void) )
{
	MPID_Errhandler *errhand_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERRHANDLER_SET_CXX);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERRHANDLER_SET_CXX);
	
	MPID_Errhandler_get_ptr( errhand, errhand_ptr );
	errhand_ptr->language				= MPID_LANG_CXX;
	MPIR_Process.cxx_call_errfn		= (void (*)( int, int *, int *, 
											void (*)(void) ))errcall;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERRHANDLER_SET_CXX);
}
#endif

/* ------------------------------------------------------------------------- */
/* 
   Nesting level for routines.
   Note that since these use per-thread data, no locks or atomic update
   routines are required.

   In a single-threaded environment, These are replaced with
   MPIR_Thread.nest_count ++, --.  These are defined in the mpiimpl.h file.
 */
/* ------------------------------------------------------------------------- */

#undef FUNCNAME
#define FUNCNAME MPIR_Nest_incr_export
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/* These routines export the nest increment and decrement for use in ROMIO */
void MPIR_Nest_incr_export( void )
{
	MPICH_PerThread_t *p;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_NEST_INCR_EXPORT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_NEST_INCR_EXPORT);
	MPIR_GetPerThread(&p);
	p->nest_count++;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_NEST_INCR_EXPORT);
}

#undef FUNCNAME
#define FUNCNAME MPIR_Nest_decr_export
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Nest_decr_export( void )
{
	MPICH_PerThread_t *p;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_NEST_DECR_EXPORT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_NEST_DECR_EXPORT);
	MPIR_GetPerThread(&p);
	p->nest_count--;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_NEST_DECR_EXPORT);
}

/* ------------------------------------------------------------------------- */
/* These routines are called on error exit from most top-level MPI routines
   to invoke the appropriate error handler.  Also included is the routine
   to call if MPI has not been initialized (MPIR_Err_preinit) and to 
   determine if an error code represents a fatal error (MPIR_Err_is_fatal). */
/* ------------------------------------------------------------------------- */
/* Special error handler to call if we are not yet initialized */
void MPIR_Err_preinit( void )
{
	MPIU_Error_printf(L"Error encountered before initializing MPICH\n");
	ExitProcess(1);
}

/* Return true if the error code indicates a fatal error */
#undef FUNCNAME
#define FUNCNAME MPIR_Err_is_fatal
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Err_is_fatal(int errcode)
{
	int result;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_IS_FATAL);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_IS_FATAL);
	result = (errcode & ERROR_FATAL_MASK) ? MPIU_TRUE : MPIU_FALSE;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_IS_FATAL);
	return result;
}

/*
 * This is the routine that is invoked by most MPI routines to 
 * report an error 
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Err_return_comm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Err_return_comm( MPID_Comm  *comm_ptr, const wchar_t fcname[], 
						  int errcode )
{
	int rc;
	const int error_class = ERROR_GET_CLASS(errcode);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_RETURN_COMM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_RETURN_COMM);

	MPIU_THREADPRIV_GET; 

	rc = checkValidErrcode( error_class, fcname, &errcode );
	
	/* First, check the nesting level */
	if (MPIR_Nest_value())
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_COMM);
		return errcode;
	}
	
	if (!comm_ptr || comm_ptr->errhandler == NULL) {
		/* Try to replace with the default handler, which is the one on 
		   MPI_COMM_WORLD.  This gives us correct behavior for the
		   case where the error handler on MPI_COMM_WORLD has been changed. */
		if (MPIR_Process.comm_world)
		{
			comm_ptr = MPIR_Process.comm_world;
		}
	}

	if (MPIR_Err_is_fatal(errcode) ||
		comm_ptr == NULL || comm_ptr->errhandler == NULL || 
		comm_ptr->errhandler->handle == MPI_ERRORS_ARE_FATAL)
	{
		/* Calls MPID_Abort */
		handleFatalError( comm_ptr, fcname, errcode );
	}

	/* Check for the special case of a user-provided error code */
	errcode = checkForUserErrcode( errcode );
	
	if (comm_ptr->errhandler->handle == MPI_ERRORS_RETURN)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_COMM);
		return errcode;
	}
	else
	{
		/* The user error handler may make calls to MPI routines, so the
		 * nesting counter must be incremented before the handler is called */
		MPIR_Nest_incr();
	
		/* We pass a final 0 (for a null pointer) to these routines
		   because MPICH-1 expected that */
		switch (comm_ptr->errhandler->language)
		{
		case MPID_LANG_C:
			(*comm_ptr->errhandler->errfn.C_Comm_Handler_function)( 
				&comm_ptr->handle, &errcode, 0 );
			break;
#ifdef HAVE_CXX_BINDING
		case MPID_LANG_CXX:
			(*MPIR_Process.cxx_call_errfn)( 0, &comm_ptr->handle, &errcode, 
					(void (*)(void))*comm_ptr->errhandler->errfn.C_Comm_Handler_function );
			/* The C++ code throws an exception if the error handler 
			 returns something other than MPI_SUCCESS. There is no "return"
			 of an error code. */
			errcode = MPI_SUCCESS;
			break;
#endif
#ifdef HAVE_FORTRAN_BINDING
		case MPID_LANG_FORTRAN90:
		case MPID_LANG_FORTRAN:
			(*comm_ptr->errhandler->errfn.F77_Handler_function)( 
				(MPI_Fint *)&comm_ptr->handle, &errcode, 0 );
			break;
#endif
		}

		MPIR_Nest_decr();
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_COMM);
	return errcode;
}

/* 
 * MPI routines that detect errors on window objects use this to report errors
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Err_return_win
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Err_return_win( MPID_Win  *win_ptr, const wchar_t fcname[], int errcode )
{
	int rc;
	const int error_class = ERROR_GET_CLASS(errcode);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_RETURN_WIN);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_RETURN_WIN);

	MPIU_THREADPRIV_GET;

	if (win_ptr == NULL || win_ptr->errhandler == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_WIN);
		return MPIR_Err_return_comm(NULL, fcname, errcode);
	}

	rc = checkValidErrcode( error_class, fcname, &errcode );

	/* First, check the nesting level */
	if (MPIR_Nest_value())
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_WIN);
		return errcode;
	}

	if (MPIR_Err_is_fatal(errcode) ||
		win_ptr == NULL || win_ptr->errhandler == NULL || 
		win_ptr->errhandler->handle == MPI_ERRORS_ARE_FATAL)
	{
		/* Calls MPID_Abort */
		handleFatalError( NULL, fcname, errcode );
	}

	/* Check for the special case of a user-provided error code */
	errcode = checkForUserErrcode( errcode );
	
	if (win_ptr->errhandler->handle == MPI_ERRORS_RETURN)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_WIN);
		return errcode;
	}
	else
	{
		/* Now, invoke the error handler for the window */

		/* The user error handler may make calls to MPI routines, so the
		 * nesting counter must be incremented before the handler is called */
		MPIR_Nest_incr();
	
		/* We pass a final 0 (for a null pointer) to these routines
		   because MPICH-1 expected that */
		switch (win_ptr->errhandler->language)
		{
			case MPID_LANG_C:
				(*win_ptr->errhandler->errfn.C_Win_Handler_function)( 
					&win_ptr->handle, &errcode, 0 );
				break;
#ifdef HAVE_CXX_BINDING
			case MPID_LANG_CXX:
			(*MPIR_Process.cxx_call_errfn)( 2, &win_ptr->handle, &errcode, 
					(void (*)(void))*win_ptr->errhandler->errfn.C_Win_Handler_function );
			/* The C++ code throws an exception if the error handler 
			 returns something other than MPI_SUCCESS. There is no "return"
			 of an error code. */
			errcode = MPI_SUCCESS;
			break;
#endif
#ifdef HAVE_FORTRAN_BINDING
			case MPID_LANG_FORTRAN90:
			case MPID_LANG_FORTRAN:
				(*win_ptr->errhandler->errfn.F77_Handler_function)( 
					(MPI_Fint *)&win_ptr->handle, &errcode, 0 );
				break;
#endif
		}

		MPIR_Nest_decr();
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_WIN);
	return errcode;
}

/* 
 * MPI routines that detect errors on files use this to report errors 
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Err_return_file
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Err_return_file( MPID_File  *file_ptr, const wchar_t fcname[], int errcode )
{
	int rc;
	const int error_class = ERROR_GET_CLASS(errcode);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_RETURN_FILE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_RETURN_FILE);

	MPIU_THREADPRIV_GET;

	if (file_ptr == NULL || file_ptr->errhandler == NULL)
	{
		file_ptr = MPIR_Process.file_null;
	}

	rc = checkValidErrcode( error_class, fcname, &errcode );
	
	/* First, check the nesting level */
	if (MPIR_Nest_value())
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_FILE);
		return errcode;
	}

	if (MPIR_Err_is_fatal(errcode) || 
		file_ptr->errhandler->handle == MPI_ERRORS_ARE_FATAL)
	{
		/* Calls MPID_Abort */
		handleFatalError( MPIR_Process.comm_world, fcname, errcode );
	}

	/* Check for the special case of a user-provided error code */
	errcode = checkForUserErrcode( errcode );
	
	if (file_ptr->errhandler->handle == MPI_ERRORS_RETURN)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_FILE);
		return errcode;
	}
	else
	{
		/* Invoke the provided error handler */

		/* The user error handler may make calls to MPI routines, so the
		 * nesting counter must be incremented before the handler is called */
		MPIR_Nest_incr();
	
		switch (file_ptr->errhandler->language)
		{
			case MPID_LANG_C:
				(*file_ptr->errhandler->errfn.C_File_Handler_function)( 
				(MPI_File *)&file_ptr->handle, &errcode );
				break;
#ifdef HAVE_CXX_BINDING
			case MPID_LANG_CXX:
				(*MPIR_Process.cxx_call_errfn)( 1, &file_ptr->handle, &errcode, 
					(void (*)(void))*file_ptr->errhandler->errfn.C_File_Handler_function );
			/* The C++ code throws an exception if the error handler 
			 returns something other than MPI_SUCCESS. There is no "return"
			 of an error code. */
			errcode = MPI_SUCCESS;
				break;
#endif
#ifdef HAVE_FORTRAN_BINDING
			case MPID_LANG_FORTRAN90:
			case MPID_LANG_FORTRAN:
				(*file_ptr->errhandler->errfn.F77_Handler_function)( 
					(MPI_Fint *)&file_ptr->handle, &errcode );
				break;
#endif
		}

		MPIR_Nest_decr();
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_RETURN_FILE);
	return errcode;
}


/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

void MPIR_Err_init( void )
{
#   if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL
	MPIR_Err_stack_init();
#   endif
}


/* ------------------------------------------------------------------------- */
/* The following block of code manages the instance-specific error messages  */
/* ------------------------------------------------------------------------- */

/* Check for a valid error code.  If the code is not valid, attempt to
   print out something sensible; reset the error code to have class 
   ERR_UNKNOWN */
/* FIXME: Now that error codes are chained, this does not produce a valid 
   error code since there is no valid ring index corresponding to this code */
static int checkValidErrcode( int error_class, const wchar_t fcname[], 
							  int *errcode_p )
{
	int errcode = *errcode_p;
	int rc = 0;

	if (error_class > MPICH_ERR_LAST_CLASS)
	{
		if (errcode & ~ERROR_CLASS_MASK)
		{
			MPIU_Error_printf(L"INTERNAL ERROR: Invalid error class (%d) encountered while returning from\n"
							  L"%s.  Please file a bug report.  The error stack follows:\n", error_class, fcname);
			/* Note that we don't try to print the error stack; if the 
			   error code is invalid, it can't be used to find
			   the error stack.  We could consider dumping the 
			   contents of the error ring instead (without trying
			   to interpret them) */
		}
		else
		{
			MPIU_Error_printf(L"INTERNAL ERROR: Invalid error class (%d) encountered while returning from\n"
							  L"%s.  Please file a bug report.  No error stack is available.\n", error_class, fcname);
		}
		/* FIXME: We probably want to set this to MPI_ERR_UNKNOWN
		   and discard the rest of the bits */
		errcode = (errcode & ~ERROR_CLASS_MASK) | MPI_ERR_UNKNOWN;
		rc = 1;
	}
	*errcode_p = errcode;
	return rc;
}

/* Check that an encoded error code is valid. Return 0 if valid, positive, 
   non-zero if invalid.  Value indicates reason; see 
   ErrcodeInvalidReasonStr() */
#if MPICH_ERROR_MSG_LEVEL > MPICH_ERROR_MSG_GENERIC
#else
/* This is the shortened version when there are no error messages */
static int checkErrcodeIsValid( int errcode )
{
	if (errcode < 0 || errcode >= MPICH_ERR_LAST_CLASS) {
		return 3;
	}
	return 0;
}

static const wchar_t *ErrcodeInvalidReasonStr( int reason )
{
	const wchar_t *str = NULL;
	switch (reason)
	{
	case 3:
		str = L"Message class is out of range";
		break;
	default:
		str = L"Unknown reason for invalid errcode";
		break;
	}
	return str;
}
#endif

/* This routine is called when there is a fatal error */
static void handleFatalError( MPID_Comm *comm_ptr, 
							  const wchar_t fcname[], int errcode )
{
	wchar_t error_msg[ MAX_ERRMSG_STRING ];
	int len;

	/* FIXME: Not internationalized */
	if (MPIR_ERR_GET_CLASS(errcode) == MPI_ERR_OTHER)
		MPIU_Snprintf(error_msg, MAX_ERRMSG_STRING, _TRUNCATE, L"Fatal error in %s:\n", fcname);
	else
		MPIU_Snprintf(error_msg, MAX_ERRMSG_STRING, _TRUNCATE, L"Fatal error in %s: ", fcname);
	len = (int)wcslen(error_msg);
	MPIR_Err_get_string(errcode, &error_msg[len], MAX_ERRMSG_STRING-len, NULL);
	/* FIXME: 13? */
	MPID_Abort(comm_ptr, MPI_SUCCESS, 13, error_msg);
}

#if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_GENERIC
/* 
 * Given a message string abbreviation (e.g., one that starts "**"), return 
 * the corresponding index.  For the generic (non
 * parameterized messages), use idx = FindGenericMsgIndex( "**msg" );
 */
#undef FUNCNAME
#define FUNCNAME FindGenericMsgIndex
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int FindGenericMsgIndex( const wchar_t *msg )
{
	int i, c;
	MPIDI_STATE_DECL(MPID_STATE_FINDGENERICMSGINDEX);

	MPIDI_FUNC_ENTER(MPID_STATE_FINDGENERICMSGINDEX);
	for (i=0; i<generic_msgs_len; i++) {
		/* Check the sentinals to insure that the values are ok first */
		if (generic_err_msgs[i].sentinal1 != 0xacebad03 ||
			generic_err_msgs[i].sentinal2 != 0xcb0bfa11) {
			/* Something bad has happened! Don't risk trying the
			   short_name pointer; it may have been corrupted */
			break;
		}
		c = wcscmp( generic_err_msgs[i].short_name, msg );
		if (c == 0)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_FINDGENERICMSGINDEX);
			return i;
		}
		if (c > 0)
		{
			/* don't return here if the string partially matches */
			if (wcsncmp(generic_err_msgs[i].short_name, msg, wcslen(msg)) != 0)
			{
				MPIDI_FUNC_EXIT(MPID_STATE_FINDGENERICMSGINDEX);
				return -1;
			}
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_FINDGENERICMSGINDEX);
	return -1;
}

/* 
   Here is an alternate search routine based on bisection.
   int i_low, i_mid, i_high, c;
   i_low = 0; i_high = generic_msg_len - 1;
   while (i_high - i_low >= 0) {
	   i_mid = (i_high + i_low) / 2;
	   c = wcscmp( generic_err_msgs[i].short_name, msg );
	   if (c == 0) return i_mid;
	   if (c < 0) { i_low = i_mid + 1; }
	   else	   { i_high = i_mid - 1; }
   }
   return -1;
*/
#endif

#if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL
/* 
 * Given a message string abbreviation (e.g., one that starts "**"), return 
 * the corresponding index.  For the specific
 * (parameterized messages), use idx = FindSpecificMsgIndex( "**msg" );
 */
#undef FUNCNAME
#define FUNCNAME FindSpecificMsgIndex
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int FindSpecificMsgIndex( const wchar_t *msg )
{
	int i, c;
	MPIDI_STATE_DECL(MPID_STATE_FINDSPECIFICMSGINDEX);

	MPIDI_FUNC_ENTER(MPID_STATE_FINDSPECIFICMSGINDEX);
	for (i=0; i<specific_msgs_len; i++)
	{
		/* Check the sentinals to insure that the values are ok first */
		if (specific_err_msgs[i].sentinal1 != 0xacebad03 ||
			specific_err_msgs[i].sentinal2 != 0xcb0bfa11)
		{
			/* Something bad has happened! Don't risk trying the
			   short_name pointer; it may have been corrupted */
			break;
		}
		c = wcscmp( specific_err_msgs[i].short_name, msg );
		if (c == 0)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_FINDSPECIFICMSGINDEX);
			return i;
		}
		if (c > 0)
		{
			/* don't return here if the string partially matches */
			if (wcsncmp(specific_err_msgs[i].short_name, msg, wcslen(msg)) != 0)
			{
				MPIDI_FUNC_EXIT(MPID_STATE_FINDSPECIFICMSGINDEX);
				return -1;
			}
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_FINDSPECIFICMSGINDEX);
	return -1;
}
/* See FindGenericMsgIndex comments for a more efficient search routine that
   could be used here as well. */
#endif

/* ------------------------------------------------------------------------ */
/* The following routines create an MPI error code, handling optional,	  */
/* instance-specific error message information.  There are two key routines:*/
/*	MPIR_Err_create_code - Create the error code; this is the routine used*/
/*						   by most routines							   */
/*	MPIR_Err_create_code_valist - Create the error code; accept a valist  */
/*						   instead of a variable argument list (this is   */
/*						   used to allow this routine to be used from	 */
/*						   within another varargs routine)				*/
/* ------------------------------------------------------------------------ */

/* Err_create_code is just a shell that accesses the va_list and then
   calls the real routine.  */
#undef FUNCNAME
#define FUNCNAME MPIR_Err_create_code
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Err_create_code( int lastcode, int fatal, const wchar_t fcname[], 
						  int line, int error_class, const wchar_t generic_msg[],
						  const wchar_t specific_msg[], ... )
{
	int rc;
	va_list Argp;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_CREATE_CODE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_CREATE_CODE);
	va_start(Argp, specific_msg);
	rc = MPIR_Err_create_code_valist( lastcode, fatal, fcname, line,
									  error_class, generic_msg, specific_msg,
									  Argp );
	va_end(Argp);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_CREATE_CODE);
	return rc;
}

#if MPICH_ERROR_MSG_LEVEL < MPICH_ERROR_MSG_ALL
/* In this case, the routine ignores all but (possibly) the generic message */
int MPIR_Err_create_code_valist( int lastcode, int fatal, const wchar_t fcname[], 
								 int line, int error_class, 
								 const wchar_t generic_msg[],
								 const wchar_t specific_msg[], va_list Argp )
{
#if MPICH_ERROR_MSG_LEVEL == MPICH_ERROR_MSG_GENERIC
	int generic_idx;
	int errcode;
	generic_idx = FindGenericMsgIndex(generic_msg);
	if (generic_idx >= 0) {
		errcode = (generic_idx << ERROR_GENERIC_SHIFT) | error_class;
		if (fatal)
			errcode |= ERROR_FATAL_MASK;
	}
	return errcode;
#else
	return error_class;
#endif	
}
#endif /* msg_level < msg_all */

/*
 * Accessor routines for the predefined messages.  These can be
 * used by the other routines (such as MPI_Error_string) to
 * access the messages in this file, or the messages that may be
 * available through any message catalog facility 
 */
#undef FUNCNAME
#define FUNCNAME get_class_msg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t *get_class_msg( int error_class )
{
	MPIDI_STATE_DECL(MPID_STATE_GET_CLASS_MSG);

	MPIDI_FUNC_ENTER(MPID_STATE_GET_CLASS_MSG);
#if MPICH_ERROR_MSG_LEVEL > MPICH_ERROR_MSG_NONE
	if (error_class >= 0 && error_class < MPIR_MAX_ERROR_CLASS_INDEX)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_GET_CLASS_MSG);
		return generic_err_msgs[class_to_index[error_class]].long_name;
	}
	else
	{
		MPIDI_FUNC_EXIT(MPID_STATE_GET_CLASS_MSG);
		return L"Unknown error class";
	}
#else
	MPIDI_FUNC_EXIT(MPID_STATE_GET_CLASS_MSG);
	return L"Error message texts are not available";
#endif
}

#undef FUNCNAME
#define FUNCNAME MPIR_Err_get_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Err_get_string( int errorcode, wchar_t * msg, int length, 
						  MPIR_Err_get_class_string_func_t fn )
{
	int error_class;
	int len, num_remaining = length;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_GET_STRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_GET_STRING);
	
	if (num_remaining == 0)
		num_remaining = MPI_MAX_ERROR_STRING;

	/* Convert the code to a string.  The cases are:
	   simple class.  Find the corresponding string.
	   <not done>
	   if (user code) { go to code that extracts user error messages }
	   else {
		   is specific message code set and available?  if so, use it
		   else use generic code (lookup index in table of messages)
	   }
	 */
	if (errorcode & ERROR_DYN_MASK)
	{
		/* This is a dynamically created error code (e.g., with 
		   MPI_Err_add_class) */
		/* If a dynamic error code was created, the function to convert
		   them into strings has been set.  Check to see that it was; this 
		   is a safeguard against a bogus error code */
		if (!MPIR_Process.errcode_to_string)
		{
			/* FIXME: not internationalized */
			/* --BEGIN ERROR HANDLING-- */
			if (MPIU_Wcsncpy(msg, L"Undefined dynamic error code", num_remaining))
			{
				msg[num_remaining - 1] = L'\0';
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			if (MPIU_Wcsncpy(msg, MPIR_Process.errcode_to_string( errorcode ), num_remaining))
			{
				msg[num_remaining - 1] = L'\0';
			}
		}
	}
	else if ( (errorcode & ERROR_CLASS_MASK) == errorcode)
	{
		error_class = MPIR_ERR_GET_CLASS(errorcode);

		if (fn != NULL && error_class > MPICH_ERR_LAST_CLASS /*&& error_class < MPICH_ERR_MAX_EXT_CLASS*/)
		{
			fn(errorcode, msg, length);
		}
		else
		{
			if (MPIU_Wcsncpy(msg, get_class_msg( errorcode ), num_remaining))
			{
				msg[num_remaining - 1] = L'\0';
			}
		}
	}
	else
	{
		/* print the class message first */
		error_class = MPIR_ERR_GET_CLASS(errorcode);

		if (fn != NULL && error_class > MPICH_ERR_LAST_CLASS /*&& error_class < MPICH_ERR_MAX_EXT_CLASS*/)
		{
			fn(errorcode, msg, num_remaining);
		}
		else
		{
			MPIU_Wcsncpy(msg, get_class_msg(ERROR_GET_CLASS(errorcode)), num_remaining);
		}
		msg[num_remaining - 1] = L'\0';
		len = (int)wcslen(msg);
		if ((len > 0) && !((len == 1) && (msg[0] == L' '))) /* Avoid empty strings or the single space error string */
		{
			msg += len;
			num_remaining -= len;
			if (num_remaining > 1)
			{
				msg[0] = L'\n';
				msg[1] = L'\0';
				msg++;
				num_remaining--;
			}
		}


		/* then print the stack or the last specific error message */

#	   if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL
		if (ErrGetInstanceString( errorcode, msg, num_remaining, fn )) 
			goto fn_exit;
#	   elif MPICH_ERROR_MSG_LEVEL > MPICH_ERROR_MSG_NONE
		{
			int generic_idx;

			generic_idx = ((errorcode & ERROR_GENERIC_MASK) >> ERROR_GENERIC_SHIFT) - 1;

			if (generic_idx >= 0)
			{
				MPIU_Snprintf(msg, num_remaining, _TRUNCATE, L", %s", generic_err_msgs[generic_idx].long_name);
				msg[num_remaining - 1] = L'\0';
				goto fn_exit;
			}
		}
#	   endif

#if 0
		if (MPIR_Err_print_stack_flag)
		{
			/*MPIU_Wcsncpy(msg, L", error stack:\n", num_remaining);*/
			MPIU_Wcsncpy(msg, L"\n", num_remaining);
			msg[num_remaining - 1] = L'\0';
			len = (int)wcslen(msg);
			msg += len;
			num_remaining -= len;
			MPIR_Err_print_stack_string_ext(errorcode, msg, num_remaining, fn);
			msg[num_remaining - 1] = L'\0';
		}
		else
		{
#		   if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL
			{
				error_ring_mutex_lock();
				{
					while (errorcode != MPI_SUCCESS)
					{
						int ring_idx;
						int ring_id;
						int generic_idx;

						if (convertErrcodeToIndexes( errorcode, &ring_idx, 
													 &ring_id,
													 &generic_idx ) != 0) {
							MPIU_Error_printf( 
				  L"Invalid error code (%d) (error ring index %d invalid)\n", 
				  errorcode, ring_idx );
							break;
						}
						
						if (generic_idx < 0)
						{
							break;
						}

						if (ErrorRing[ring_idx].id == ring_id)
						{
							/* just keep clobbering old values until the 
							   end of the stack is reached */
							MPIU_Snprintf(msg, num_remaining, _TRUNCATE, L", %s", ErrorRing[ring_idx].msg);
							msg[num_remaining - 1] = L'\0';
							errorcode = ErrorRing[ring_idx].prev_error;
						}
						else
						{
							break;
						}
					}
				}
				error_ring_mutex_unlock();

				if (errorcode == MPI_SUCCESS)
				{
					goto fn_exit;
				}
			}
#		   endif

#		   if MPICH_ERROR_MSG_LEVEL > MPICH_ERROR_MSG_NONE
			{
				int generic_idx;

				generic_idx = ((errorcode & ERROR_GENERIC_MASK) >> ERROR_GENERIC_SHIFT) - 1;

				if (generic_idx >= 0)
				{
					MPIU_Snprintf(msg, num_remaining, _TRUNCATE, L", %s", generic_err_msgs[generic_idx].long_name);
					msg[num_remaining - 1] = L'\0';
					goto fn_exit;
				}
			}
#		   endif
		}
#endif
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_GET_STRING);
	return;
}

/* ------------------------------------------------------------------------- */
/* Routines to convert instance-specific messages into a string			  */
/* This is the only case that supports instance-specific messages			*/
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* This block of code is used to convert various MPI values into descriptive */
/* strings.  The routines are												*/
/*	 GetAssertString   - handle MPI_MODE_xxx (RMA asserts)				 */
/*	 GetFileModeString - handle MPI_MODE_xxx (File open flags)			 */
/*	 GetDTypeString	- handle MPI_Datatypes							  */
/*	 GetMPIOpString	- handle MPI_Op									 */
/* These routines are used in vsnprintf_mpi								  */
/* FIXME: These functions are not thread safe								*/
/* ------------------------------------------------------------------------- */
#define ASSERT_STR_MAXLEN 256

#undef FUNCNAME
#define FUNCNAME GetAssertString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t * GetAssertString(int d)
{
	static wchar_t str[ASSERT_STR_MAXLEN] = L"";
	wchar_t *cur;
	size_t len = ASSERT_STR_MAXLEN;
	size_t n;
	MPIDI_STATE_DECL(MPID_STATE_GETASSERTSTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_GETASSERTSTRING);

	if (d == 0)
	{
		MPIU_Wcsncpy(str, L"assert=0", ASSERT_STR_MAXLEN);
		MPIDI_FUNC_EXIT(MPID_STATE_GETASSERTSTRING);
		return str;
	}
	cur = str;
	if (d & MPI_MODE_NOSTORE)
	{
		MPIU_Wcsncpy(cur, L"MPI_MODE_NOSTORE", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		d ^= MPI_MODE_NOSTORE;
	}
	if (d & MPI_MODE_NOCHECK)
	{
		if (len < ASSERT_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_NOCHECK", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_NOCHECK", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		d ^= MPI_MODE_NOCHECK;
	}
	if (d & MPI_MODE_NOPUT)
	{
		if (len < ASSERT_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_NOPUT", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_NOPUT", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		d ^= MPI_MODE_NOPUT;
	}
	if (d & MPI_MODE_NOPRECEDE)
	{
		if (len < ASSERT_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_NOPRECEDE", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_NOPRECEDE", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		d ^= MPI_MODE_NOPRECEDE;
	}
	if (d & MPI_MODE_NOSUCCEED)
	{
		if (len < ASSERT_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_NOSUCCEED", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_NOSUCCEED", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		d ^= MPI_MODE_NOSUCCEED;
	}
	if (d)
	{
		if (len < ASSERT_STR_MAXLEN)
			MPIU_Snprintf(cur, len, _TRUNCATE, L" | 0x%x", d);
		else
			MPIU_Snprintf(cur, len, _TRUNCATE, L"assert=0x%x", d);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_GETASSERTSTRING);
	return str;
}

#define AMODE_STR_MAXLEN 256

#undef FUNCNAME
#define FUNCNAME GetFileModeString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t * GetFileModeString(int m)
{
	static wchar_t str[AMODE_STR_MAXLEN] = L"";
	wchar_t *cur;
	size_t len = AMODE_STR_MAXLEN;
	size_t n;
	MPIDI_STATE_DECL(MPID_STATE_GETFILEMODESTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_GETFILEMODESTRING);

	if (m == 0)
	{
		MPIU_Wcsncpy(str, L"amode=0", AMODE_STR_MAXLEN);
		MPIDI_FUNC_EXIT(MPID_STATE_GETFILEMODESTRING);
		return str;
	}
	cur = str;
	if (m & MPI_MODE_RDONLY)
	{
		MPIU_Wcsncpy(cur, L"MPI_MODE_RDONLY", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_RDONLY;
	}
	if (m & MPI_MODE_RDWR)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_RDWR", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_RDWR", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_RDWR;
	}
	if (m & MPI_MODE_WRONLY)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_WRONLY", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_WRONLY", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_WRONLY;
	}
	if (m & MPI_MODE_CREATE)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_CREATE", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_CREATE", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_CREATE;
	}
	if (m & MPI_MODE_EXCL)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_EXCL", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_EXCL", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_EXCL;
	}
	if (m & MPI_MODE_DELETE_ON_CLOSE)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_DELETE_ON_CLOSE", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_DELETE_ON_CLOSE", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_DELETE_ON_CLOSE;
	}
	if (m & MPI_MODE_UNIQUE_OPEN)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_UNIQUE_OPEN", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_UNIQUE_OPEN", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_UNIQUE_OPEN;
	}
	if (m & MPI_MODE_APPEND)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_APPEND", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_APPEND", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_APPEND;
	}
	if (m & MPI_MODE_SEQUENTIAL)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Wcsncpy(cur, L" | MPI_MODE_SEQUENTIAL", len);
		else
			MPIU_Wcsncpy(cur, L"MPI_MODE_SEQUENTIAL", len);
		n = wcslen(cur);
		cur += n;
		len -= n;
		m ^= MPI_MODE_SEQUENTIAL;
	}
	if (m)
	{
		if (len < AMODE_STR_MAXLEN)
			MPIU_Snprintf(cur, len, _TRUNCATE, L" | 0x%x", m);
		else
			MPIU_Snprintf(cur, len, _TRUNCATE, L"amode=0x%x", m);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_GETFILEMODESTRING);
	return str;
}

#undef FUNCNAME
#define FUNCNAME GetDTypeString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t * GetDTypeString(MPI_Datatype d)
{
	static wchar_t default_str[64];
	int num_integers, num_addresses, num_datatypes, combiner = 0;
	wchar_t *str;
	MPIDI_STATE_DECL(MPID_STATE_GETDTYPESTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_GETDTYPESTRING);

	if (d == MPI_DATATYPE_NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
		return L"MPI_DATATYPE_NULL";
	}

	if (d == 0)
	{
		MPIU_Wcsncpy(default_str, L"dtype=0x0", 64);
		MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
		return default_str;
	}

	MPID_Type_get_envelope(d, &num_integers, &num_addresses, &num_datatypes, 
						   &combiner);
	if (combiner == MPI_COMBINER_NAMED)
	{
		str = MPIDU_Datatype_builtin_to_string(d);
		if (str == NULL)
		{
			MPIU_Snprintf(default_str, 64, _TRUNCATE, L"dtype=0x%08x", d);
			MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
			return default_str;
		}
		MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
		return str;
	}
	
	/* default is not thread safe */
	str = MPIDU_Datatype_combiner_to_string(combiner);
	if (str == NULL)
	{
		MPIU_Snprintf(default_str, 64, _TRUNCATE, L"dtype=USER<0x%08x>", d);
		MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
		return default_str;
	}
	MPIU_Snprintf(default_str, 64, _TRUNCATE, L"dtype=USER<%s>", str);
	MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
	return default_str;
}

#undef FUNCNAME
#define FUNCNAME GetMPIOpString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t * GetMPIOpString(MPI_Op o)
{
	const wchar_t *str;
	static wchar_t default_str[64];
	MPIDI_STATE_DECL(MPID_STATE_GETMPIOPSTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_GETMPIOPSTRING);

	switch (o)
	{
	case MPI_OP_NULL:
		str = L"MPI_OP_NULL";
		break;
	case MPI_MAX:
		str = L"MPI_MAX";
		break;
	case MPI_MIN:
		str = L"MPI_MIN";
		break;
	case MPI_SUM:
		str = L"MPI_SUM";
		break;
	case MPI_PROD:
		str = L"MPI_PROD";
		break;
	case MPI_LAND:
		str = L"MPI_LAND";
		break;
	case MPI_BAND:
		str = L"MPI_BAND";
		break;
	case MPI_LOR:
		str = L"MPI_LOR";
		break;
	case MPI_BOR:
		str = L"MPI_BOR";
		break;
	case MPI_LXOR:
		str = L"MPI_LXOR";
		break;
	case MPI_BXOR:
		str = L"MPI_BXOR";
		break;
	case MPI_MINLOC:
		str = L"MPI_MINLOC";
		break;
	case MPI_MAXLOC:
		str = L"MPI_MAXLOC";
		break;
	case MPI_REPLACE:
		str = L"MPI_REPLACE";
		break;
	default:
		/* default is not thread safe */
		MPIU_Snprintf(default_str, 64, _TRUNCATE, L"op=0x%x", o);
		str = default_str;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_GETMPIOPSTRING);
	return str;
}

/* ------------------------------------------------------------------------ */
/* This routine takes an instance-specific string with format specifiers	*/
/* This routine makes use of the above routines, along with some inlined	*/
/* code, to process the format specifiers for the MPI objects			   */
/* The current set of format specifiers is undocumented except for their
   use in this routine.  In addition, these choices do not permit the 
   use of GNU extensions to check the validity of these arguments.  
   At some point, a documented set that can exploit those GNU extensions
   will replace these. */
/* ------------------------------------------------------------------------ */
#undef FUNCNAME
#define FUNCNAME vsnprintf_mpi
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int vsnprintf_mpi(wchar_t *str, size_t maxlen, const wchar_t *fmt_orig, va_list list)
{
	wchar_t *begin, *end, *fmt = NULL;
	size_t len;
	MPI_Comm C;
	MPI_Info I;
	MPI_Datatype D;
	MPI_File F;
	MPI_Win W;
	MPI_Group G;
	MPI_Op O;
	MPI_Request R;
	MPI_Errhandler E;
	MPI_Offset o;
	wchar_t *s;
	int t, i, d, mpi_errno=MPI_SUCCESS;
	void *p;
	MPIDI_STATE_DECL(MPID_STATE_VSNPRINTF_MPI);

	MPIDI_FUNC_ENTER(MPID_STATE_VSNPRINTF_MPI);

	fmt = MPIU_Strdup(fmt_orig);
	if (fmt == NULL)
	{
		if (maxlen > 0 && str != NULL)
			*str = L'\0';
		MPIDI_FUNC_EXIT(MPID_STATE_VSNPRINTF_MPI);
		return 0;
	}

	begin = fmt;
	end = wcschr(fmt, L'%');
	while (end)
	{
		len = maxlen;
		if (len > (size_t)(end - begin)) {
			len = (size_t)(end - begin);
		}
		if (len)
		{
			memcpy(str, begin, len*sizeof(wchar_t));
			str += len;
			maxlen -= len;
		}
		end++;
		begin = end+1;
		switch ((int)(*end))
		{
		case (int)L's':
			s = va_arg(list, wchar_t *);
			if (s != NULL)
			{
				MPIU_Wcsncpy(str, s, maxlen);
			}
			break;
		case (int)L'd':
			d = va_arg(list, int);
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%d", d);
			break;
		case (int)L'x':
			d = va_arg(list, int);
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%x", d);
			break;
		case (int)L'o':
			o = va_arg(list, MPI_Offset);
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%d", o);
			break;
		case (int)L'i':
			i = va_arg(list, int);
			switch (i)
			{
			case MPI_ANY_SOURCE:
				MPIU_Wcsncpy(str, L"MPI_ANY_SOURCE", maxlen);
				break;
			case MPI_PROC_NULL:
				MPIU_Wcsncpy(str, L"MPI_PROC_NULL", maxlen);
				break;
			case MPI_ROOT:
				MPIU_Wcsncpy(str, L"MPI_ROOT", maxlen);
				break;
			case MPI_UNDEFINED_RANK:
				MPIU_Wcsncpy(str, L"MPI_UNDEFINED_RANK", maxlen);
				break;
			default:
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%d", i);
				break;
			}
			break;
		case (int)L't':
			t = va_arg(list, int);
			switch (t)
			{
			case MPI_ANY_TAG:
				MPIU_Wcsncpy(str, L"MPI_ANY_TAG", maxlen);
				break;
			case MPI_UNDEFINED:
				MPIU_Wcsncpy(str, L"MPI_UNDEFINED", maxlen);
				break;
			default:
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%d", t);
				break;
			}
			break;
		case (int)L'p':
			p = va_arg(list, void *);
			/* FIXME: A check for MPI_IN_PLACE should only be used 
			   where that is valid */
			if (p == MPI_IN_PLACE)
			{
				MPIU_Wcsncpy(str, L"MPI_IN_PLACE", maxlen);
			}
			else
			{
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"0x%p", p);
			}
			break;
		case (int)L'C':
			C = va_arg(list, MPI_Comm);
			switch (C)
			{
			case MPI_COMM_WORLD:
				MPIU_Wcsncpy(str, L"MPI_COMM_WORLD", maxlen);
				break;
			case MPI_COMM_SELF:
				MPIU_Wcsncpy(str, L"MPI_COMM_SELF", maxlen);
				break;
			case MPI_COMM_NULL:
				MPIU_Wcsncpy(str, L"MPI_COMM_NULL", maxlen);
				break;
			default:
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"comm=0x%x", C);
				break;
			}
			break;
		case (int)L'I':
			I = va_arg(list, MPI_Info);
			if (I == MPI_INFO_NULL)
			{
				MPIU_Wcsncpy(str, L"MPI_INFO_NULL", maxlen);
			}
			else
			{
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"info=0x%x", I);
			}
			break;
		case (int)L'D':
			D = va_arg(list, MPI_Datatype);
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s", GetDTypeString(D));
			break;
			/* Include support for %F only if MPI-IO is enabled */
		case (int)L'F':
			F = va_arg(list, MPI_File);
			if (F == MPI_FILE_NULL)
			{
				MPIU_Wcsncpy(str, L"MPI_FILE_NULL", maxlen);
			}
			else
			{
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"file=0x%x", F);
			}
			break;
		case (int)L'M':
			d = va_arg(list, int);
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s", GetFileModeString(d));
			break;
		case (int)L'W':
			W = va_arg(list, MPI_Win);
			if (W == MPI_WIN_NULL)
			{
				MPIU_Wcsncpy(str, L"MPI_WIN_NULL", maxlen);
			}
			else
			{
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"win=0x%x", W);
			}
			break;
		case (int)L'A':
			d = va_arg(list, int);
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s", GetAssertString(d));
			break;
		case (int)L'G':
			G = va_arg(list, MPI_Group);
			if (G == MPI_GROUP_NULL)
			{
				MPIU_Wcsncpy(str, L"MPI_GROUP_NULL", maxlen);
			}
			else
			{
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"group=0x%x", G);
			}
			break;
		case (int)L'O':
			O = va_arg(list, MPI_Op);
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s", GetMPIOpString(O));
			break;
		case (int)L'R':
			R = va_arg(list, MPI_Request);
			if (R == MPI_REQUEST_NULL)
			{
				MPIU_Wcsncpy(str, L"MPI_REQUEST_NULL", maxlen);
			}
			else
			{
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"req=0x%x", R);
			}
			break;
		case (int)L'E':
			E = va_arg(list, MPI_Errhandler);
			switch (E)
			{
			case MPI_ERRHANDLER_NULL:
				MPIU_Wcsncpy(str, L"MPI_ERRHANDLER_NULL", maxlen);
				break;
			case MPI_ERRORS_RETURN:
				MPIU_Wcsncpy(str, L"MPI_ERRORS_RETURN", maxlen);
				break;
			case MPI_ERRORS_ARE_FATAL:
				MPIU_Wcsncpy(str, L"MPI_ERRORS_ARE_FATAL", maxlen);
				break;
			default:
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"errh=0x%x", E);
				break;
			}
			break;
		default:
			/* Error: unhandled output type */
			if (fmt != NULL)
				MPIU_Free(fmt);
			MPIDI_FUNC_EXIT(MPID_STATE_VSNPRINTF_MPI);
			return 0;
			/*
			if (maxlen > 0 && str != NULL)
				*str = L'\0';
			break;
			*/
		}
		len = wcslen(str);
		maxlen -= len;
		str += len;
		end = wcschr(begin, L'%');
	}
	if (*begin != L'\0')
	{
		MPIU_Wcsncpy(str, begin, maxlen);
	}
	/* Free the dup'ed format string */
	MPIU_Free( fmt );

	MPIDI_FUNC_EXIT(MPID_STATE_VSNPRINTF_MPI);
	return mpi_errno;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/*
 * Instance-specific error messages are stored in a ring.  The elements of this
 * ring are MPIR_Err_msg_t structures, which contain the following fields:
 *	id - this is used to check that the entry is valid; it is computed from
 *		 the error code and location in the ring.  The routine
 *		   ErrcodeToId( errcode, &id ) is used to extract the id from an 
 *		 error code and
 *		   ErrcodeCreateID( class, generic, msg, &id, &seq ) is used to 
 *		 create the id from an error class, generic index, and message 
 *		 string.  The "seq" field is inserted into the error code as a 
 *		 check.
 *
 *	prev_error - The full MPI error code of the previous error attached 
 *		 to this list of errors, or MPI_SUCCESSS (which has value 0).
 *		 This is the last error code, not the index in the ring of the last 
 *		 error code.  That's the right choice, because we want to ensure 
 *		 that the value is valid if the ring overflows.  In addition,
 *		 we allow this to be an error CLASS (one of the predefined MPI
 *		 error classes).  This is particularly important for 
 *		 MPI_ERR_IN_STATUS, which may be returned as a valid error code.
 *		 (classes are valid error codes).
 *		 
 *	use_user_error_code and user_error_code - Used to handle a few cases 
 *		 in MPI where a user-provided routine returns an error code; 
 *		 this allows us to provide information about the chain of 
 *		 routines that were involved, while returning the users prefered
 *		 error value to the users environment.
 *
 *	location - A string that indicates what function and line number
 *		 where the error code was set.
 *
 *	msg - A message about the error.  This may be instance-specific (e.g.,
 *		 it may have been created at the time the error was detected with
 *		 information about the parameters that caused the error).
 *
 * Note that both location and msg are defined as length MAX_xxx+1.  This 
 * isn't really necessary (at least for msg), since the MPI standard 
 * requires that MAX_MPI_ERROR_STRING include the space for the trailing null,
 * but using the extra byte makes the code a little simpler.
 *
 * The "id" value is used to keep a sort of "checkvalue" to ensure that the
 * error code that points at this message is in fact for this particular 
 * message.  This is used to handle the unlikely but possible situation where 
 * so many error messages are generated that the ring is overlapped.
 *
 * The message arrays are preallocated to ensure that there is space for these
 * messages when an error occurs.  One variation would be to allow these
 * to be dynamically allocated, but it is probably better to either preallocate
 * these or turn off all error message generation (which will eliminate these
 * arrays).
 *
 * One possible alternative is to use the message ring *only* for instance
 * messages and use the predefined messages in-place for the generic
 * messages.  The approach used here provides uniform handling of all 
 * error messages.
 *
 * Note on user error codes
 *
 * The "user error codes" is used to handle an ambiguity in the MPI-1 
 * standard about the return value from the attribute callbacks.  The 
 * standard does not specify what values, other than 'MPI_SUCCESS', are
 * valid.  Because the Intel MPI-1 test suite expected 'MPI_Comm_dup' to
 * return the same non-zero value returned by the attribute callback routine,
 * this is the behavior that many (if not all) MPI implementations provide.
 * As a result, the return from those routines is 
 * 
*/
#if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL
#define MAX_ERROR_RING ERROR_SPECIFIC_INDEX_SIZE
#define MAX_LOCATION_LEN 63

/* See the description above for the fields in this structure */
typedef struct MPIR_Err_msg
{
	int  id;
	int  prev_error;
	int  use_user_error_code;
	int  user_error_code;

	wchar_t location[MAX_LOCATION_LEN+1];
	wchar_t msg[MPI_MAX_ERROR_STRING+1];
}
MPIR_Err_msg_t;

static MPIR_Err_msg_t ErrorRing[MAX_ERROR_RING];
static volatile unsigned int error_ring_loc	 = 0;
static volatile unsigned int max_error_ring_loc = 0;

#if defined(MPID_REQUIRES_THREAD_SAFETY)
/* if the device requires internal MPICH routines to be thread safe, the
   MPIU_THREAD_CHECK macros are not appropriate */
static MPID_Thread_mutex_t error_ring_mutex;
#define error_ring_mutex_create(_mpi_errno_p_)						\
	MPID_Thread_mutex_create(&error_ring_mutex, _mpi_errno_p_)
#define error_ring_mutex_destroy(_mpi_errno_p)								\
	MPID_Thread_mutex_destroy(&error_ring_mutex, _mpi_errno_p_)
#define error_ring_mutex_lock()						\
	MPID_Thread_mutex_lock(&error_ring_mutex)
#define error_ring_mutex_unlock()				\
	MPID_Thread_mutex_unlock(&error_ring_mutex)
#elif defined(MPICH_IS_THREADED)
static MPID_Thread_mutex_t error_ring_mutex;
#define error_ring_mutex_create(_mpi_errno_p) MPID_Thread_mutex_create(&error_ring_mutex,_mpi_errno_p)
#define error_ring_mutex_destroy(_mpi_errno_p) MPID_Thread_mutex_destroy(&error_ring_mutex,_mpi_errno_p)
#define error_ring_mutex_lock() \
	MPIU_THREAD_CHECK_BEGIN \
	 MPID_Thread_mutex_lock(&error_ring_mutex) \
	MPIU_THREAD_CHECK_END
#define error_ring_mutex_unlock() \
	MPIU_THREAD_CHECK_BEGIN \
	 MPID_Thread_mutex_unlock(&error_ring_mutex) \
	MPIU_THREAD_CHECK_END
#else
#define error_ring_mutex_create(_a)
#define error_ring_mutex_destroy(_a)
#define error_ring_mutex_lock()
#define error_ring_mutex_unlock()
#endif

/* Create the ring id from information about the message */
static void ErrcodeCreateID( int error_class, int generic_idx, 
							 const wchar_t *msg, int *id, int *seq )
{
	int i;
	int ring_seq = 0, ring_id;

	/* Create a simple hash function of the message to serve as the 
	   sequence number */
	ring_seq = 0;
	for (i=0; msg[i]; i++)
		ring_seq += (unsigned int) msg[i];

	ring_seq %= ERROR_SPECIFIC_SEQ_SIZE;

	ring_id = (error_class & ERROR_CLASS_MASK) |
		((generic_idx + 1) << ERROR_GENERIC_SHIFT) |
		(ring_seq << ERROR_SPECIFIC_SEQ_SHIFT);

	*id  = ring_id;
	*seq = ring_seq;
}

/* Convert an error code into ring_idx, ring_id, and generic_idx.
   Return non-zero if there is a problem with the decode values
   (e.g., out of range for the ring index) */
#undef FUNCNAME
#define FUNCNAME convertErrcodeToIndexes
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int convertErrcodeToIndexes( int errcode, int *ring_idx, int *ring_id,
									int *generic_idx )
{
	MPIDI_STATE_DECL(MPID_STATE_CONVERTERRCODETOINDEXES);

	MPIDI_FUNC_ENTER(MPID_STATE_CONVERTERRCODETOINDEXES);
	*ring_idx = (errcode & ERROR_SPECIFIC_INDEX_MASK) >> 
		ERROR_SPECIFIC_INDEX_SHIFT;
	*ring_id = errcode & (ERROR_CLASS_MASK | 
						  ERROR_GENERIC_MASK | ERROR_SPECIFIC_SEQ_MASK);
	*generic_idx = ((errcode & ERROR_GENERIC_MASK) >> ERROR_GENERIC_SHIFT) - 1;
	
	/* Test on both the max_error_ring_loc and MAX_ERROR_RING to guard 
	   against memory overwrites */
	if (*ring_idx < 0 || *ring_idx >= MAX_ERROR_RING || 
		(unsigned int)(*ring_idx) > max_error_ring_loc)
	{
		return 1;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_CONVERTERRCODETOINDEXES);
	return 0;
}

/* Check that an encoded error code is valid. Return 0 if valid, positive, 
   non-zero if invalid.  Value indicates reason; see 
   ErrcodeInvalidReasonStr() */
static int checkErrcodeIsValid( int errcode )
{
	int ring_id, generic_idx, ring_idx;

	/* If the errcode is a class, then it is valid */
	if (errcode <= MPIR_MAX_ERROR_CLASS_INDEX && errcode >= 0)
	{
		return 0;
	}

	convertErrcodeToIndexes( errcode, &ring_idx, &ring_id, &generic_idx );
	if (ring_idx < 0 || ring_idx >= MAX_ERROR_RING ||
		(unsigned int)ring_idx > max_error_ring_loc)
	{
		return 1;
	}
	if (ErrorRing[ring_idx].id != ring_id)
	{
		return 2;
	}
	/* It looks like the code uses a generic idx of -1 to indicate no
	   generic message */
	if (generic_idx < -1 || generic_idx > generic_msgs_len)
	{
		return 3;
	}
	return 0;
}

static const wchar_t *ErrcodeInvalidReasonStr( int reason )
{
	const wchar_t *str = 0;
	switch (reason) {
	case 1:
		str = L"Ring Index out of range";
		break;
	case 2:
		str = L"Ring ids do not match";
		break;
	case 3:
		str = L"Generic message index out of range";
		break;
	default:
		str = L"Unknown reason for invalid errcode";
		break;
	}
	return str;
}

/* Check to see if the error code is a user-specified error code
   (e.g., from the attribute delete function) and if so, set the error code
   to the value provide by the user */
static int checkForUserErrcode( int errcode )
{
	error_ring_mutex_lock();
	{
		if (errcode != MPI_SUCCESS)
		{
			int ring_idx;
			int ring_id;
			int generic_idx;

			if (convertErrcodeToIndexes( errcode, &ring_idx, &ring_id,
				&generic_idx ) != 0) {
					MPIU_Error_printf( 
						L"Invalid error code (%d) (error ring index %d invalid)\n", 
						errcode, ring_idx );
				}
			else {
				/* Can we get a more specific error message */
				if (generic_idx >= 0 && 
					ErrorRing[ring_idx].id == ring_id && 
					ErrorRing[ring_idx].use_user_error_code)
				{
					errcode = ErrorRing[ring_idx].user_error_code;
				}
			}
		}
	}
	error_ring_mutex_unlock();
	return errcode;
}


/*
 * This is the real routine for generating an error code.  It takes
 * a va_list so that it can be called by any routine that accepts a 
 * variable number of arguments.
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Err_create_code_valist
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Err_create_code_valist( int lastcode, int fatal, const wchar_t fcname[], 
								 int line, int error_class, 
								 const wchar_t generic_msg[],
								 const wchar_t specific_msg[], va_list Argp )
{
	int err_code;
	int generic_idx;
	int use_user_error_code = 0;
	int user_error_code = -1;
	wchar_t user_ring_msg[MPI_MAX_ERROR_STRING+1] = L"";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_CREATE_CODE_VALIST);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_CREATE_CODE_VALIST);

	/* Create the code from the class and the message ring index */

	/* FIXME: We want to make abort invoke any cleanup routines first */
	if (MPIR_Err_abort_on_error)
	{
		/*wprintf(L"aborting from %s, line %d\n", fcname, line);fflush(stdout);*/
		abort();
	}

	/* Check that lastcode is valid */
	if (lastcode != MPI_SUCCESS) {
		int reason;
		reason = checkErrcodeIsValid(lastcode);
		if (reason) {
			MPIU_Error_printf( L"Internal Error: invalid error code %x (%s) in %s:%d\n", 
							   lastcode, ErrcodeInvalidReasonStr( reason ), 
							   fcname, line );
			lastcode = MPI_SUCCESS;
		}
	}

	/* FIXME: ERR_OTHER is overloaded; this may mean "OTHER" or it may
	   mean "No additional error, just routine stack info" */
	if (error_class == MPI_ERR_OTHER)
	{
		if (MPIR_ERR_GET_CLASS(lastcode) > MPI_SUCCESS && 
			MPIR_ERR_GET_CLASS(lastcode) <= MPICH_ERR_LAST_CLASS)
		{
			/* If the last class is more specific (and is valid), then pass it 
			   through */
			error_class = MPIR_ERR_GET_CLASS(lastcode);
		}
		else
		{
			error_class = MPI_ERR_OTHER;
		}
	}

	/* Handle special case of MPI_ERR_IN_STATUS.  According to the standard,
	   the code must be equal to the class. See section 3.7.5.  
	   Information on the particular error is in the MPI_ERROR field 
	   of the status. */
	if (error_class == MPI_ERR_IN_STATUS)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_CREATE_CODE_VALIST);
		return MPI_ERR_IN_STATUS;
	}

	err_code = error_class;

	/* Handle the generic message.  This selects a subclass, based on a text 
	   string */
#   if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_GENERIC
	{
		generic_idx = FindGenericMsgIndex(generic_msg);
		if (generic_idx >= 0)
		{
			if (wcscmp( generic_err_msgs[generic_idx].short_name, L"**user" ) == 0)
			{
				use_user_error_code = 1;
				/* This is a special case.  The format is
				   "**user", "**userxxx %d", intval
				   (generic, specific, parameter).  In this
				   case we must ... save the user value because
				   we store it explicitly in the ring.  
				   We do this here because we cannot both access the 
				   user error code and pass the argp to vsnprintf_mpi . */
				if (specific_msg) {
					const wchar_t *specific_fmt; 
					int specific_idx;
					user_error_code = va_arg(Argp,int);
#if MPICH_ERROR_MSG_LEVEL > MPICH_ERROR_MSG_GENERIC
					specific_idx = FindSpecificMsgIndex(specific_msg);
					if (specific_idx >= 0) {
						specific_fmt = specific_err_msgs[specific_idx].long_name;
					}
					else {
						specific_fmt = specific_msg;
					}
#else
						specific_fmt = specific_msg;
#endif
					MPIU_Snprintf( user_ring_msg, sizeof(user_ring_msg) / sizeof(wchar_t), _TRUNCATE,
								   specific_fmt, user_error_code );
				}
				else {
					user_ring_msg[0] = 0;
				}
			}
			err_code |= (generic_idx + 1) << ERROR_GENERIC_SHIFT;
		}
		else
		{
			/* TODO: lookup index for class error message */
			err_code &= ~ERROR_GENERIC_MASK;
			
#		   ifdef MPICH_DBG_OUTPUT
			{
				if (generic_msg[0] == L'*' && generic_msg[1] == L'*')
				{
					/* FIXME : Internal error.  Generate some debugging information; Fix for the general release */
					fwprintf( stderr, L"Could not find %s in list of messages\n", generic_msg );
					MPID_Abort(MPIR_Process.comm_world, 0, 1, L"Inconsistent internal error tables.");
				}
			}
#		   endif
		}
	}
#   endif

	/* Handle the instance-specific part of the error message */
#   if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL
	{
		int specific_idx;
		const wchar_t * specific_fmt = 0;
		int  ring_idx, ring_seq=0;
		wchar_t * ring_msg;
		
		error_ring_mutex_lock();
		{
			/* Get the next entry in the ring; keep track of what part of the 
			   ring is in use (max_error_ring_loc) */
			ring_idx = error_ring_loc++;
			if (error_ring_loc >= MAX_ERROR_RING) 
				error_ring_loc %= MAX_ERROR_RING;
			if (error_ring_loc > max_error_ring_loc)
				max_error_ring_loc = error_ring_loc;
		
			ring_msg = ErrorRing[ring_idx].msg;

			if (specific_msg != NULL)
			{
				specific_idx = FindSpecificMsgIndex(specific_msg);
				if (specific_idx >= 0)
				{
					specific_fmt = specific_err_msgs[specific_idx].long_name;
				}
				else
				{
					specific_fmt = specific_msg;
				}
				/* See the code above for handling user errors */
				if (!use_user_error_code)
				{
					vsnprintf_mpi( ring_msg, MPI_MAX_ERROR_STRING, 
								   specific_fmt, Argp );
				}
				else
				{
					MPIU_Wcsncpy(ring_msg, user_ring_msg, MPI_MAX_ERROR_STRING);
					/*ring_msg = user_ring_msg;*/
				}
			}
			else if (generic_idx >= 0)
			{
				MPIU_Wcsncpy( ring_msg, generic_err_msgs[generic_idx].long_name, MPI_MAX_ERROR_STRING );
			}
			else
			{
				MPIU_Wcsncpy( ring_msg, generic_msg, MPI_MAX_ERROR_STRING );
			}

			ring_msg[MPI_MAX_ERROR_STRING] = L'\0';
			/*wprintf(L"ring msg: %s\n", ring_msg);fflush(stdout);*/
		
			/* Get the ring sequence number and set the ring id */
			ErrcodeCreateID( error_class, generic_idx, ring_msg, 
							 &ErrorRing[ring_idx].id, &ring_seq );
			/* Set the previous code. */
			ErrorRing[ring_idx].prev_error = lastcode;

			/* */
			if (use_user_error_code)
			{
				ErrorRing[ring_idx].use_user_error_code = 1;
				ErrorRing[ring_idx].user_error_code	 = user_error_code;
			}
			else if (lastcode != MPI_SUCCESS)
			{
				int last_ring_idx;
				int last_ring_id;
				int last_generic_idx;

				if (convertErrcodeToIndexes( lastcode, &last_ring_idx, 
											 &last_ring_id,
											 &last_generic_idx ) != 0) {
					MPIU_Error_printf( 
				  L"Invalid error code (%d) (error ring index %d invalid)\n", 
				  lastcode, last_ring_idx );
				}
				else {
					if (last_generic_idx >= 0 && 
						ErrorRing[last_ring_idx].id == last_ring_id) {
						if (ErrorRing[last_ring_idx].use_user_error_code) {
							ErrorRing[ring_idx].use_user_error_code = 1;
							ErrorRing[ring_idx].user_error_code = 
								ErrorRing[last_ring_idx].user_error_code;
						}
					}
				}
			}

			if (fcname != NULL)
			{
				MPIU_Snprintf(ErrorRing[ring_idx].location, MAX_LOCATION_LEN+1, _TRUNCATE, L"%s(%d)", fcname, line);
				ErrorRing[ring_idx].location[MAX_LOCATION_LEN] = L'\0';
				/*wprintf(L"location: %s\n", ErrorRing[ring_idx].location);fflush(stdout);*/
			}
			else
			{
				ErrorRing[ring_idx].location[0] = L'\0';
			}
		}
		error_ring_mutex_unlock();

		err_code |= ring_idx << ERROR_SPECIFIC_INDEX_SHIFT;
		err_code |= ring_seq << ERROR_SPECIFIC_SEQ_SHIFT;

	}
#   endif

	if (fatal || MPIR_Err_is_fatal(lastcode))
	{
		err_code |= ERROR_FATAL_MASK;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_CREATE_CODE_VALIST);
	return err_code;
}

/* Append an error code, error2, to the end of a list of messages in the error
   ring whose head endcoded in error1_code.  An error code pointing at the
   combination is returned.  If the list of messages does not terminate cleanly
   (i.e. ring wrap has occurred), then the append is not performed. and error1
   is returned (although it may include the class of error2 if the class of
   error1 was MPI_ERR_OTHER). */
int MPIR_Err_combine_codes(const int error1, const int error2)
{
	int error1_code = error1;
	int error2_code = error2;
	int error2_class;
	
	if (error2_code == MPI_SUCCESS) return error1_code;
	if (error2_code & ERROR_DYN_MASK) return error2_code;
	if (error1_code == MPI_SUCCESS) return error2_code;
			
	error2_class = MPIR_ERR_GET_CLASS(error2_code);
	if (MPIR_ERR_GET_CLASS(error2_class) < MPI_SUCCESS ||
		MPIR_ERR_GET_CLASS(error2_class) > MPICH_ERR_LAST_CLASS)
	{
		error2_class = MPI_ERR_OTHER;
	}

#   if MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL
	{
		int error_code;

		error_code = error1_code;
		
		error_ring_mutex_lock();
		{
			for (;;)
			{
				int error_class;
				int ring_idx;
				int ring_id;
				int generic_idx;
				
				if (convertErrcodeToIndexes(error_code, &ring_idx, &ring_id,
					&generic_idx) != 0 || generic_idx < 0 ||
					ErrorRing[ring_idx].id != ring_id)
				{
					break;
				}

				error_code = ErrorRing[ring_idx].prev_error;

				if (error_code == MPI_SUCCESS)
				{
					ErrorRing[ring_idx].prev_error = error2;
					break;
				}
				
				error_class = MPIR_ERR_GET_CLASS(error_code);
				
				if (error_class == MPI_ERR_OTHER)
				{
					ErrorRing[ring_idx].prev_error &= ~(ERROR_CLASS_MASK);
					ErrorRing[ring_idx].prev_error |= error2_class;
				}
			}
		}
		error_ring_mutex_unlock();
	}
#   endif

	if (MPIR_ERR_GET_CLASS(error1_code) == MPI_ERR_OTHER)
	{
		error1_code = (error1_code & ~(ERROR_CLASS_MASK)) | error2_class;
	}

	return error1_code;
}

/* ------------------------------------------------------------------------- */
/* Manage the error reporting stack										  */
/* ------------------------------------------------------------------------- */

static int old_stack_chop_val;

#undef FUNCNAME
#define FUNCNAME MPIR_Err_disable_error_stack_chop
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Err_disable_error_stack_chop()
{
	old_stack_chop_val = MPIR_Err_chop_error_stack;
	MPIR_Err_chop_error_stack = 0;
}

#undef FUNCNAME
#define FUNCNAME MPIR_Err_restore_error_stack_chop
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Err_restore_error_stack_chop()
{
	MPIR_Err_chop_error_stack = old_stack_chop_val;
}

#undef FUNCNAME
#define FUNCNAME MPIR_Err_stack_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Err_stack_init( void )
{
	int n;
	int rc;
#if defined(MPID_REQUIRES_THREAD_SAFETY) || defined(MPICH_IS_THREADED) /* These protect whether error_ring_mutex_create is defined */
	int mpi_errno = MPI_SUCCESS;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_INIT);

		error_ring_mutex_create(&mpi_errno);

		rc = MPIU_GetEnvBool( L"DeinoMPI_ABORT_ON_ERROR", &MPIR_Err_abort_on_error );
		rc = MPIU_GetEnvBool( L"DeinoMPI_PRINT_ERROR_STACK", &MPIR_Err_print_stack_flag );

		rc = MPIU_GetEnvInt( L"DeinoMPI_CHOP_ERROR_STACK", &n );
		if (rc == 1)
		{
			/* If windows, set the default width to the window size */
			HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
			if (hConsole != INVALID_HANDLE_VALUE)
			{
				CONSOLE_SCREEN_BUFFER_INFO info;
				if (GetConsoleScreenBufferInfo(hConsole, &info))
				{
					MPIR_Err_chop_width = info.dwMaximumWindowSize.X - 1;
				}
			}
			if (n > 0)
			{
				MPIR_Err_chop_error_stack = TRUE;
				MPIR_Err_chop_width = n;
			}
			else if (n == 0)
			{
				/* Use the default width */
				MPIR_Err_chop_error_stack = TRUE;
			}
		}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_INIT);
}

/* Given an error code, print the stack of messages corresponding to this
   error code. */
#undef FUNCNAME
#define FUNCNAME MPIR_Err_print_stack
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Err_print_stack(FILE * fp, int errcode)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_PRINT_STACK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_PRINT_STACK);
		error_ring_mutex_lock();
		{
			while (errcode != MPI_SUCCESS)
			{
				int ring_idx;
				int ring_id;
				int generic_idx;

				if (convertErrcodeToIndexes( errcode, &ring_idx, &ring_id,
											 &generic_idx ) != 0) {
					MPIU_Error_printf( 
				  L"Invalid error code (%d) (error ring index %d invalid)\n", 
				  errcode, ring_idx );
					break;
				}

				if (generic_idx < 0)
				{
					break;
				}

				if (ErrorRing[ring_idx].id == ring_id)
				{
					fwprintf(fp, L"%s: %s\n", ErrorRing[ring_idx].location, 
							ErrorRing[ring_idx].msg);
					errcode = ErrorRing[ring_idx].prev_error;
				}
				else
				{
					break;
				}
			}
		}
		error_ring_mutex_unlock();

		/* FIXME: This is wrong.  The only way that you can get here without
		   errcode beign MPI_SUCCESS is if there is an error in the 
		   processing of the error codes.  Dropping through into the next
		   level of code (particularly when that code doesn't check for 
		   valid error codes!) is erroneous */
		if (errcode == MPI_SUCCESS)
		{
			goto fn_exit;
		}

	{
		int generic_idx;
					
		generic_idx = ((errcode & ERROR_GENERIC_MASK) >> ERROR_GENERIC_SHIFT) - 1;
		
		if (generic_idx >= 0)
		{
			fwprintf(fp, L"(unknown)(): %s\n", generic_err_msgs[generic_idx].long_name);
			goto fn_exit;
		}
	}
	
	{
		int error_class;

		error_class = ERROR_GET_CLASS(errcode);
		
		if (error_class <= MPICH_ERR_LAST_CLASS)
		{
			fwprintf(fp, L"(unknown)(): %s\n", get_class_msg(ERROR_GET_CLASS(errcode)));
		}
		else
		{
			/* FIXME: Not internationalized */
			fwprintf(fp, L"Error code contains an invalid class (%d)\n", error_class);
		}
	}
	
  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_PRINT_STACK);
	return;
}

#undef FUNCNAME
#define FUNCNAME MPIR_Err_print_stack_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Err_print_stack_string(int errcode, wchar_t *str, int maxlen)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_PRINT_STACK_STRING);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_PRINT_STACK_STRING);
	MPIR_Err_print_stack_string_ext( errcode, str, maxlen, NULL );
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_PRINT_STACK_STRING);
}

#undef FUNCNAME
#define FUNCNAME MPIR_Err_print_stack_string_ext
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIR_Err_print_stack_string_ext(int errcode, wchar_t *str, int maxlen, 
									 MPIR_Err_get_class_string_func_t fn)
{
	int len;
	wchar_t *str_orig = str;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ERR_PRINT_STACK_STRING_EXT);

	/*printf("MPIR_Err_chop_width = %d\n", MPIR_Err_chop_width);fflush(stdout);*/
	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ERR_PRINT_STACK_STRING_EXT);
		error_ring_mutex_lock();
		{
			/* Find the longest fcname in the stack */
			int max_location_len = 0;
			int tmp_errcode = errcode;
			while (tmp_errcode != MPI_SUCCESS)
			{
				int ring_idx;
				int ring_id;
				int generic_idx;

				if (convertErrcodeToIndexes( tmp_errcode, &ring_idx, &ring_id,
											 &generic_idx ) != 0) {
					MPIU_Error_printf( 
				  L"Invalid error code (%d) (error ring index %d invalid)\n", 
				  errcode, ring_idx );
					break;
				}

				if (generic_idx < 0)
				{
					break;
				}
					
				if (ErrorRing[ring_idx].id == ring_id) {
					len = (int)wcslen(ErrorRing[ring_idx].location);
					max_location_len = MPIR_MAX(max_location_len, len);
					tmp_errcode = ErrorRing[ring_idx].prev_error;
				}
				else
				{
					break;
				}
			}
			max_location_len += 2; /* add space for the L": " */
			/*wprintf(L"max_location_len = %d\n", max_location_len);fflush(stdout);*/
			/* print the error stack */
			while (errcode != MPI_SUCCESS)
			{
				int ring_idx;
				int ring_id;
				int generic_idx;
				int i;
				wchar_t *cur_pos, *last_pos, *iter_pos;

				if (convertErrcodeToIndexes( errcode, &ring_idx, &ring_id,
											 &generic_idx ) != 0) {
					MPIU_Error_printf( 
				  L"Invalid error code (%d) (error ring index %d invalid)\n", 
				  errcode, ring_idx );
					break;
				}

				if (generic_idx < 0)
				{
					break;
				}
					
				if (ErrorRing[ring_idx].id == ring_id)
				{
					MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s", ErrorRing[ring_idx].location);
					len = (int)wcslen(str);
					maxlen -= len;
					str += len;
					for (i=0; i<max_location_len - (int)wcslen(ErrorRing[ring_idx].location) - 2; i++)
					{
						if (MPIU_Snprintf(str, maxlen, _TRUNCATE, L"."))
						{
							maxlen--;
							str++;
						}
					}
					if (MPIU_Snprintf(str, maxlen, _TRUNCATE, L":"))
					{
						maxlen--;
						str++;
					}
					if (MPIU_Snprintf(str, maxlen, _TRUNCATE, L" "))
					{
						maxlen--;
						str++;
					}

					if (MPIR_Err_chop_error_stack)
					{
						cur_pos = ErrorRing[ring_idx].msg;
						len = (int)wcslen(cur_pos);
						if (len == 0)
						{
							if (MPIU_Snprintf(str, maxlen, _TRUNCATE, L"\n"))
							{
								maxlen--;
								str++;
							}
						}
						while (len)
						{
							if (len >= MPIR_Err_chop_width - max_location_len)
							{
								if (len > maxlen)
									break;

								last_pos = cur_pos;
								MPIU_Snprintf(str, MPIR_Err_chop_width - 1 - max_location_len, _TRUNCATE, L"%s", cur_pos);
								str[MPIR_Err_chop_width - 2 - max_location_len] = L'\n';
								cur_pos += MPIR_Err_chop_width - 2 - max_location_len;
								/* back up to the last space character */
								iter_pos = cur_pos;
								while (*iter_pos != L' ' && iter_pos > last_pos)
									iter_pos--;
								if (iter_pos > last_pos)
								{
									cur_pos = iter_pos + 1;
									str += cur_pos - last_pos - 1;
									*str = L'\n';
									str++;
									maxlen -= (int)(cur_pos - last_pos + 1);
								}
								else
								{
									str += MPIR_Err_chop_width - max_location_len;
									maxlen -= MPIR_Err_chop_width - max_location_len;
								}
								if (maxlen < max_location_len)
									break;
								for (i=0; i<max_location_len; i++)
								{
									MPIU_Snprintf(str, maxlen, _TRUNCATE, L" ");
									maxlen--;
									str++;
								}
								len = (int)wcslen(cur_pos);
							}
							else
							{
								MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s\n", cur_pos);
								len = (int)wcslen(str);
								maxlen -= len;
								str += len;
								len = 0;
							}
						}
					}
					else
					{
						MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s\n", ErrorRing[ring_idx].msg);
						len = (int)wcslen(str);
						maxlen -= len;
						str += len;
					}
					/*
					MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s: %s\n", ErrorRing[ring_idx].location, ErrorRing[ring_idx].msg);
					len = (int)wcslen(str);
					maxlen -= len;
					str += len;
					*/
					errcode = ErrorRing[ring_idx].prev_error;
				}
				else
				{
					break;
				}
			}
		}
		error_ring_mutex_unlock();

		if (errcode == MPI_SUCCESS)
		{
			goto fn_exit;
		}

	{
		int generic_idx;
					
		generic_idx = ((errcode & ERROR_GENERIC_MASK) >> ERROR_GENERIC_SHIFT) - 1;
		
		if (generic_idx >= 0)
		{
			const wchar_t *p;
			/* FIXME: (Here and elsewhere)  Make sure any string is
			   non-null before you use it */
			p = generic_err_msgs[generic_idx].long_name;
			if (!p) { p = L"<NULL>"; }
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"(unknown)(): %s\n", p );
			len = (int)wcslen(str);
			maxlen -= len;
			str += len;
			goto fn_exit;
		}
	}
	
	{
		int error_class;

		error_class = ERROR_GET_CLASS(errcode);
		
		if (error_class <= MPICH_ERR_LAST_CLASS)
		{
			MPIU_Snprintf(str, maxlen, _TRUNCATE, L"(unknown)(): %s\n", get_class_msg(ERROR_GET_CLASS(errcode)));
			len = (int)wcslen(str);
			maxlen -= len;
			str += len;
		}
		else
		{
			if (fn != NULL)
			{
				fn(errcode, str, maxlen);
			}
			else
			{
				/* FIXME: Not internationalized */
				MPIU_Snprintf(str, maxlen, _TRUNCATE, L"Error code contains an invalid class (%d)\n", error_class);
			}
			len = (int)wcslen(str);
			maxlen -= len;
			str += len;
		}
	}
	
  fn_exit:
	if (str_orig != str)
	{
		str--;
		*str = L'\0';
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ERR_PRINT_STACK_STRING_EXT);
	return;
}

static int ErrGetInstanceString( int errorcode, wchar_t *msg, int num_remaining, 
							 MPIR_Err_get_class_string_func_t fn )
{
	//int len;

	MPIU_UNREFERENCED_ARG(fn);

	if (MPIR_Err_print_stack_flag) {
		/*
		MPIU_Wcsncpy(msg, L", error stack:\n", num_remaining);
		msg[num_remaining - 1] = L'\0';
		len = (int)wcslen(msg);
		msg += len;
		num_remaining -= len;
		*/
		/* note: this took the "fn" arg, but that appears to be unused
		 and is undocumented.  */
		MPIR_Err_print_stack_string(errorcode, msg, num_remaining);
		msg[num_remaining - 1] = L'\0';
	}
	else {
		error_ring_mutex_lock();
		{
			while (errorcode != MPI_SUCCESS) {
				int ring_idx;
				int ring_id;
				int generic_idx;
				
				if (convertErrcodeToIndexes( errorcode, &ring_idx, 
											 &ring_id,
											 &generic_idx ) != 0) {
					MPIU_Error_printf( 
			  L"Invalid error code (%d) (error ring index %d invalid)\n", 
			  errorcode, ring_idx );
					break;
				}
				
				if (generic_idx < 0) {
					break;
				}
				
				if (ErrorRing[ring_idx].id == ring_id) {
					/* just keep clobbering old values until the 
					   end of the stack is reached */
					MPIU_Snprintf(msg, num_remaining, _TRUNCATE, L", %s", ErrorRing[ring_idx].msg);
					msg[num_remaining - 1] = L'\0';
					errorcode = ErrorRing[ring_idx].prev_error;
				}
				else {
					break;
				}
			}
		}
		error_ring_mutex_unlock();
	}
	/* FIXME: How do we determine that we failed to unwind the stack? */
	if (errorcode != MPI_SUCCESS) return 1;

	return 0;
}
#endif /* (MPICH_ERROR_MSG_LEVEL >= MPICH_ERROR_MSG_ALL) */

/***********************************************************
 *
 * MPIU_Call_history section
 * These functions save the last N MPI function calls in a ring
 *
 */
#define MPIR_MAX_FN_CALL_LENGTH 256
#define MPIR_FN_CALL_HISTORY_SIZE 32

typedef struct MPIR_MPI_Call_history_element
{
	wchar_t fn_call[MPIR_MAX_FN_CALL_LENGTH];
} MPIR_MPI_Call_history_element;

static MPIR_MPI_Call_history_element call_history[MPIR_FN_CALL_HISTORY_SIZE] = { 0 };
static int call_index = 0;
static int history_size = MPIR_FN_CALL_HISTORY_SIZE;
static int history_initialized = 0;

#undef FUNCNAME
#define FUNCNAME MPIU_Save
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Init_call_history(int * argc, wchar_t ***argv)
{
	/*
	int i;
	wchar_t argc_str[20], argv_str[20];
	*/
	const wchar_t *env;
	
	if (history_initialized)
		return MPI_SUCCESS;
	history_initialized = 1;

	env = MPIU_GetStaticEnv(L"DeinoMPI_USE_CALL_HISTORY");
	if (env != NULL)
	{
		if (wcscmp(env, L"1") == 0 || wcscmp(env, L"on") == 0 || wcscmp(env, L"yes") == 0)
		{
			MPIR_Process.save_call_history = TRUE;
		}
		if (wcscmp(env, L"0") == 0 || wcscmp(env, L"off") == 0 || wcscmp(env, L"no") == 0)
		{
			MPIR_Process.save_call_history = FALSE;
		}
		if (env[0] == L'\0')
		{
			/* default */
			MPIR_Process.save_call_history = TRUE;
		}
	}
	else
	{
		/* default */
		MPIR_Process.save_call_history = FALSE;
	}
	env = MPIU_GetStaticEnv(L"DeinoMPI_CALL_HISTORY_SIZE");
	if (env != NULL)
	{
		int size = _wtoi(env);
		if (size > 0 && size <= MPIR_FN_CALL_HISTORY_SIZE)
			history_size = size;
	}
	/*
	for (i=0; i<MPIR_FN_CALL_HISTORY_SIZE; i++)
	{
		call_history[i].fn_call[0] = L'\0';
	}
	if (argc == NULL)
		MPIU_Wcsncpy(argc_str, L"NULL", 20);
	else
		MPIU_Snprintf(argc_str, 20, _TRUNCATE, L"0x%p", argc);
	if (argv == NULL)
		MPIU_Wcsncpy(argv_str, L"NULL", 20);
	else
		MPIU_Snprintf(argv_str, 20, _TRUNCATE, L"0x%p", argv);
	MPIU_Snprintf(call_history[0].fn_call, MPIR_MAX_FN_CALL_LENGTH, _TRUNCATE, L"MPI_Init(%s, %s)", argc_str, argv_str);
	call_index = 1;
	*/
	MPIU_UNREFERENCED_ARG(argc);
	MPIU_UNREFERENCED_ARG(argv);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Save
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Save( const wchar_t format_str[], ... )
{
	int specific_idx;
	const wchar_t *specific_fmt;
	va_list Argp;
	MPIU_THREADPRIV_DECL;

	MPIU_THREADPRIV_GET;
	if (MPIR_Nest_value())
		return MPI_SUCCESS;

	va_start(Argp, format_str);
	specific_idx = FindSpecificMsgIndex(format_str);
	if (specific_idx >= 0)
	{
		wchar_t *str;
		specific_fmt = MPIU_Strdup(specific_err_msgs[specific_idx].long_name);
		str = wcswcs(specific_fmt, L"failed");
		if (str != NULL)
			*str = L'\0';
	}
	else
	{
		specific_fmt = format_str;
	}
	vsnprintf_mpi(call_history[call_index].fn_call, MPIR_MAX_FN_CALL_LENGTH, specific_fmt, Argp);
	va_end(Argp);
	if (specific_fmt != format_str)
		MPIU_Free(specific_fmt);
	call_index++;
	if (call_index >= history_size)
		call_index = 0;
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Print_call_history
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Print_call_history()
{
	int i, index;
	if (!MPIR_Process.save_call_history)
	{
		return MPI_SUCCESS;
	}
	for (i=0; i<history_size; i++)
	{
		index = (call_index + i) % history_size;
		if (call_history[index].fn_call[0] != L'\0')
		{
			wprintf(L"[%d] %s\n", MPIR_Process.comm_world->rank, call_history[index].fn_call);
		}
	}
	fflush(stdout);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Call_history_to_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Call_history_to_string(wchar_t *str, int max_length)
{
	size_t len;
	int i, index;
	if (!MPIR_Process.save_call_history)
	{
		return MPI_SUCCESS;
	}
	for (i=0; i<history_size; i++)
	{
		index = (call_index + i) % history_size;
		if (call_history[index].fn_call[0] != L'\0')
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"[%d] %s\n", MPIR_Process.comm_world->rank, call_history[index].fn_call);
			len = wcslen(str);
			str = str + len;
			max_length = max_length - (int)len;
		}
	}
	return MPI_SUCCESS;
}

