/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#include "group.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_union
#define MPI_Group_union PMPI_Group_union

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_union

/*@

MPI_Group_union - Produces a group by combining two groups

Input Parameters:
+ group1 - first group (handle) 
- group2 - second group (handle) 

Output Parameter:
. newgroup - union group (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Group_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_union(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup)
{
	static const wchar_t FCNAME[] = L"MPI_Group_union";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr1 = NULL;
	MPID_Group *group_ptr2 = NULL;
	MPID_Group *new_group_ptr;
	int g1_idx, g2_idx, nnew, i, k, size1, size2, mylpid;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_UNION);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("group"); 
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_UNION);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group1, mpi_errno);
			MPIR_ERRTEST_GROUP(group2, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group1, group_ptr1 );
	MPID_Group_get_ptr( group2, group_ptr2 );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr1, mpi_errno );
			MPID_Group_valid_ptr( group_ptr2, mpi_errno );
			/* If group_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_union %G %G %p", group1, group2, newgroup); }
	/* Determine the size of the new group.  The new group consists of all
	   members of group1 plus the members of group2 that are not in group1.
	*/
	g1_idx = group_ptr1->idx_of_first_lpid;
	g2_idx = group_ptr2->idx_of_first_lpid;

	/* If the lpid list hasn't been created, do it now */
	if (g1_idx < 0)
	{
		MPIR_Group_setup_lpid_list( group_ptr1 ); 
		g1_idx = group_ptr1->idx_of_first_lpid;
	}
	if (g2_idx < 0)
	{
		MPIR_Group_setup_lpid_list( group_ptr2 ); 
		g2_idx = group_ptr2->idx_of_first_lpid;
	}
	nnew = group_ptr1->size;

	/* Clear the flag bits on the second group.  The flag is set if
	   a member of the second group belongs to the union */
	size2 = group_ptr2->size;
	for (i=0; i<size2; i++)
	{
		group_ptr2->lrank_to_lpid[i].flag = 0;
	}
	/* Loop through the lists that are ordered by lpid (local process
	   id) to detect which processes in group 2 are not in group 1
	*/
	while (g1_idx >= 0 && g2_idx >= 0)
	{
		int l1_pid, l2_pid;
		l1_pid = group_ptr1->lrank_to_lpid[g1_idx].lpid;
		l2_pid = group_ptr2->lrank_to_lpid[g2_idx].lpid;
		if (l1_pid > l2_pid)
		{
			nnew++;
			group_ptr2->lrank_to_lpid[g2_idx].flag = 1;
			g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
		}
		else if (l1_pid == l2_pid)
		{
			g1_idx = group_ptr1->lrank_to_lpid[g1_idx].next_lpid;
			g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
		}
		else
		{
			/* l1 < l2 */
			g1_idx = group_ptr1->lrank_to_lpid[g1_idx].next_lpid;
		}
	}
	/* If we hit the end of group1, add the remaining members of group 2 */
	while (g2_idx >= 0)
	{
		nnew++;
		group_ptr2->lrank_to_lpid[g2_idx].flag = 1;
		g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
	}
	
	/* Allocate a new group and lrank_to_lpid array */
	mpi_errno = MPIR_Group_create( nnew, &new_group_ptr );
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno)
	{
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */
	/* If this process is in group1, then we can set the rank now. 
	   If we are not in this group, this assignment will set the
	   current rank to MPI_UNDEFINED */
	new_group_ptr->rank = group_ptr1->rank;
	
	/* Add group1 */
	size1 = group_ptr1->size;
	for (i=0; i<size1; i++)
	{
		new_group_ptr->lrank_to_lpid[i].lrank = i;
		new_group_ptr->lrank_to_lpid[i].lpid = group_ptr1->lrank_to_lpid[i].lpid;
	}
	
	/* Add members of group2 that are not in group 1 */
	
	if (group_ptr1->rank == MPI_UNDEFINED &&
		group_ptr2->rank >= 0)
	{
		mylpid = group_ptr2->lrank_to_lpid[group_ptr2->rank].lpid;
	}
	else
	{
		mylpid = -2;
	}
	k = size1;
	for (i=0; i<size2; i++)
	{
		if (group_ptr2->lrank_to_lpid[i].flag)
		{
			new_group_ptr->lrank_to_lpid[k].lrank = k;
			new_group_ptr->lrank_to_lpid[k].lpid = 
				group_ptr2->lrank_to_lpid[i].lpid;
			if (new_group_ptr->rank == MPI_UNDEFINED &&
				group_ptr2->lrank_to_lpid[i].lpid == mylpid)
				new_group_ptr->rank = k;
			k++;
		}
	}
	*newgroup = new_group_ptr->handle;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_UNION);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_union",
			L"**mpi_group_union %G %G %p", group1, group2, newgroup);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_translate_ranks
#define MPI_Group_translate_ranks PMPI_Group_translate_ranks

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_translate_ranks

/*@
 MPI_Group_translate_ranks - Translates the ranks of processes in one group to 
							 those in another group

  Input Parameters:
+ group1 - group1 (handle) 
. n - number of ranks in  'ranks1' and 'ranks2'  arrays (integer) 
. ranks1 - array of zero or more valid ranks in 'group1' 
- group2 - group2 (handle) 

  Output Parameter:
. ranks2 - array of corresponding ranks in group2,  'MPI_UNDEFINED'  when no 
correspondence exists. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_translate_ranks(MPI_Group group1, int n, int *ranks1, MPI_Group group2, int *ranks2)
{
	static const wchar_t FCNAME[] = L"MPI_Group_translate_ranks";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr1 = NULL;
	MPID_Group *group_ptr2 = NULL;
#   ifdef HAVE_ERROR_CHECKING
	int size1;
#   endif
	int i, g2_idx, l1_pid, l2_pid;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_TRANSLATE_RANKS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	/* The routines that setup the group data structures must be executed
	   within a mutex.  As most of the group routines are not performance
	   critical, we simple run these routines within the SINGLE_CS */
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_TRANSLATE_RANKS);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group1, mpi_errno);
			MPIR_ERRTEST_GROUP(group2, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group1, group_ptr1 );
	MPID_Group_get_ptr( group2, group_ptr2 );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr1, mpi_errno );
			MPID_Group_valid_ptr( group_ptr2, mpi_errno );
			/* If either group_ptr is not valid, it will be reset to null */

			MPIR_ERRTEST_ARGNEG(n,L"n",mpi_errno);
			if (group_ptr1)
			{
				/* Check that the rank entries are valid */
				size1 = group_ptr1->size;
				for (i=0; i<n; i++)
				{
					if ( (ranks1[i] < 0 && ranks1[i] != MPI_PROC_NULL) || ranks1[i] >= size1)
					{
						mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_RANK, L"**rank", L"**rank %d %d", ranks1[i], size1 );
						MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"ranks1", i, ranks1[i]);
						break;
					}
				}
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_translate_ranks %G %d %p %G %p", group1, n, ranks1, group2, ranks2); }
	
	/* Initialize the output ranks */
	for (i=0; i<n; i++)
	{
		ranks2[i] = MPI_UNDEFINED;
	}

	/* We may want to optimize for the special case of group2 is 
	   a dup of MPI_COMM_WORLD, or more generally, has rank == lpid 
	   for everything within the size of group2.  NOT DONE YET */
	g2_idx = group_ptr2->idx_of_first_lpid;
	if (g2_idx < 0)
	{
		MPIR_Group_setup_lpid_list( group_ptr2 );
		g2_idx = group_ptr2->idx_of_first_lpid;
	}
	if (g2_idx >= 0)
	{
		/* g2_idx can be < 0 if the g2 group is empty */
		l2_pid = group_ptr2->lrank_to_lpid[g2_idx].lpid;
		for (i=0; i<n; i++)
		{
			if (ranks1[i] == MPI_PROC_NULL)
			{
				ranks2[i] = MPI_PROC_NULL;
				continue;
			}
			l1_pid = group_ptr1->lrank_to_lpid[ranks1[i]].lpid;
			/* Search for this l1_pid in group2.  Use the following
			   optimization: start from the last position in the lpid list
			   if possible.  A more sophisticated version could use a 
			   tree based or even hashed search to speed the translation. */
			if (l1_pid < l2_pid || g2_idx < 0)
			{
				/* Start over from the beginning */
				g2_idx = group_ptr2->idx_of_first_lpid;
				l2_pid = group_ptr2->lrank_to_lpid[g2_idx].lpid;
			}
			while (g2_idx >= 0 && l1_pid > l2_pid)
			{
				g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
				if (g2_idx >= 0)
				{
					l2_pid = group_ptr2->lrank_to_lpid[g2_idx].lpid;
				}
				else 
					l2_pid = -1;
			}
			if (l1_pid == l2_pid)
			{
				ranks2[i] = group_ptr2->lrank_to_lpid[g2_idx].lrank;
			}
		}
	}
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_TRANSLATE_RANKS);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_translate_ranks",
			L"**mpi_group_translate_ranks %G %d %p %G %p", group1, n, ranks1, group2, ranks2);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_size
#define MPI_Group_size PMPI_Group_size

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_size

/*@

MPI_Group_size - Returns the size of a group

Input Parameters:
+ group - group (handle) 
Output Parameter:
- size - number of processes in the group (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_size(MPI_Group group, int *size)
{
	static const wchar_t FCNAME[] = L"MPI_Group_size";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_SIZE);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group, group_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr, mpi_errno );
			/* If group_ptr is not value, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_size %G %p", group, size); }
	
	*size = group_ptr->size;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_SIZE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_size",
			L"**mpi_group_size %G %p", group, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_rank
#define MPI_Group_rank PMPI_Group_rank

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_rank

/*@

MPI_Group_rank - Returns the rank of this process in the given group

Input Parameters:
. group - group (handle) 

Output Parameter:
. rank - rank of the calling process in group, or 'MPI_UNDEFINED'  if the 
process is not a member (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_rank(MPI_Group group, int *rank)
{
	static const wchar_t FCNAME[] = L"MPI_Group_rank";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_RANK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_RANK);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group, group_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr, mpi_errno );
			/* If group_ptr is not value, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_rank %G %p", group, rank); }
	
	*rank = group_ptr->rank;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_RANK);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE,FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_rank",
			L"**mpi_group_rank %G %p", group, rank);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_range_incl
#define MPI_Group_range_incl PMPI_Group_range_incl

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_range_incl

/*@

MPI_Group_range_incl - Creates a new group from ranges of ranks in an 
		existing group

Input Parameters:
+ group - group (handle) 
. n - number of triplets in array  'ranges' (integer) 
- ranges - a one-dimensional array of integer triplets, of the 
form (first rank, last rank, stride) indicating ranks in
'group'  or processes to be included in 'newgroup'.

Output Parameter:
. newgroup - new group derived from above, in the 
order defined by  'ranges' (handle)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_ARG
.N MPI_ERR_RANK

.seealso: MPI_Group_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_range_incl(MPI_Group group, int n, int ranges[][3], 
						 MPI_Group *newgroup)
{
	static const wchar_t FCNAME[] = L"MPI_Group_range_incl";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr = NULL, *new_group_ptr;
	int first, last, stride, nnew, i, j, k;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_RANGE_INCL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_RANGE_INCL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group, group_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr, mpi_errno );
			/* If group_ptr is not value, it will be reset to null */
			if (group_ptr)
			{
				mpi_errno = MPIR_Group_check_valid_ranges( group_ptr, ranges, n );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_range_incl %G %d %p %p", group, n, ranges, newgroup); }
	
	/* Compute size, assuming that included ranks are valid (and distinct) */
	nnew = 0;
	for (i=0; i<n; i++)
	{
		first = ranges[i][0]; last = ranges[i][1]; stride = ranges[i][2];
		/* works for stride of either sign.  Error checking above has already guaranteed stride != 0 */
		nnew += 1 + (last - first) / stride;
	}

	if (nnew == 0)
	{
		*newgroup = MPI_GROUP_EMPTY;
		goto fn_exit;
	}
		
	/* Allocate a new group and lrank_to_lpid array */
	mpi_errno = MPIR_Group_create( nnew, &new_group_ptr );
	if (mpi_errno) goto fn_fail;
	new_group_ptr->rank = MPI_UNDEFINED;

	/* Group members taken in order specified by the range array */
	/* This could be integrated with the error checking, but since this
	   is a low-usage routine, we haven't taken that optimization */
	k = 0;
	for (i=0; i<n; i++)
	{
		first = ranges[i][0]; last = ranges[i][1]; stride = ranges[i][2];
		if (stride > 0)
		{
			for (j=first; j<=last; j += stride)
			{
				new_group_ptr->lrank_to_lpid[k].lrank = k;
				new_group_ptr->lrank_to_lpid[k].lpid = group_ptr->lrank_to_lpid[j].lpid;
				if (j == group_ptr->rank) 
					new_group_ptr->rank = k;
				k++;
			}
		}
		else
		{
			for (j=first; j>=last; j += stride)
			{
				new_group_ptr->lrank_to_lpid[k].lrank = k;
				new_group_ptr->lrank_to_lpid[k].lpid = group_ptr->lrank_to_lpid[j].lpid;
				if (j == group_ptr->rank) 
					new_group_ptr->rank = k;
				k++;
			}
		}
	}
	*newgroup = new_group_ptr->handle;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_RANGE_INCL);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_range_incl",
			L"**mpi_group_range_incl %G %d %p %p", group, n, ranges, newgroup);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_range_excl
#define MPI_Group_range_excl PMPI_Group_range_excl

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_range_excl

/*@

MPI_Group_range_excl - Produces a group by excluding ranges of processes from
	   an existing group

Input Parameters:
+ group - group (handle) 
. n - number of elements in array 'ranks' (integer) 
- ranges - a one-dimensional array of integer triplets of the
form (first rank, last rank, stride), indicating the ranks in
'group'  of processes to be excluded from the output group 'newgroup' .  

Output Parameter:
. newgroup - new group derived from above, preserving the 
order in 'group'  (handle) 

Note:  
The MPI standard requires that each of the ranks to be excluded must be
a valid rank in the group and all elements must be distinct or the
function is erroneous.  

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_RANK
.N MPI_ERR_ARG

.seealso: MPI_Group_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_range_excl(MPI_Group group, int n, int ranges[][3], 
						 MPI_Group *newgroup)
{
	static const wchar_t FCNAME[] = L"MPI_Group_range_excl";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr = NULL, *new_group_ptr;
	int size, i, j, k, nnew, first, last, stride;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_RANGE_EXCL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_RANGE_EXCL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group, group_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr, mpi_errno );
			/* If group_ptr is not valid, it will be reset to null */

			/* Check the exclusion array.  Ensure that all ranges are
			   valid and that the specified exclusions are unique */
			if (group_ptr)
			{
				mpi_errno = MPIR_Group_check_valid_ranges( group_ptr, ranges, n );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_range_excl %G %d %p %p", group, n, ranges, newgroup); }
	
	/* Compute size, assuming that included ranks are valid (and distinct) */
	size = group_ptr->size;
	nnew = 0;
	for (i=0; i<n; i++)
	{
		first = ranges[i][0]; last = ranges[i][1]; stride = ranges[i][2];
		/* works for stride of either sign.  Error checking above 
		   has already guaranteed stride != 0 */
		nnew += 1 + (last - first) / stride;
	}
	nnew = size - nnew;

	if (nnew == 0)
	{
		*newgroup = MPI_GROUP_EMPTY;
		goto fn_exit;
	}

	/* Allocate a new group and lrank_to_lpid array */
	mpi_errno = MPIR_Group_create( nnew, &new_group_ptr );
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno)
	{
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */
	new_group_ptr->rank = MPI_UNDEFINED;

	/* Group members are taken in rank order from the original group,
	   with the specified members removed. Use the flag array for that
	   purpose.  If this was a critical routine, we could use the 
	   flag values set in the error checking part, if the error checking 
	   was enabled *and* we are not MPI_THREAD_MULTIPLE, but since this
	   is a low-usage routine, we haven't taken that optimization.  */

	/* First, mark the members to exclude */
	for (i=0; i<size; i++) 
		group_ptr->lrank_to_lpid[i].flag = 0;
	
	for (i=0; i<n; i++)
	{
		first = ranges[i][0]; last = ranges[i][1]; stride = ranges[i][2];
		if (stride > 0)
		{
			for (j=first; j<=last; j += stride)
			{
				group_ptr->lrank_to_lpid[j].flag = 1;
			}
		}
		else
		{
			for (j=first; j>=last; j += stride)
			{
				group_ptr->lrank_to_lpid[j].flag = 1;
			}
		}
	}
	/* Now, run through the group and pick up the members that were not excluded */
	k = 0;
	for (i=0; i<size; i++)
	{
		if (!group_ptr->lrank_to_lpid[i].flag)
		{
			new_group_ptr->lrank_to_lpid[k].lrank = k;
			new_group_ptr->lrank_to_lpid[k].lpid = group_ptr->lrank_to_lpid[i].lpid;
			if (group_ptr->rank == i)
			{
				new_group_ptr->rank = k;
			}
			k++;
		}
	}
	
	*newgroup = new_group_ptr->handle;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_RANGE_EXCL);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
	mpi_errno = MPIR_Err_create_code(
		mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_range_excl",
		L"**mpi_group_range_excl %G %d %p %p", group, n, ranges, newgroup);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_intersection
#define MPI_Group_intersection PMPI_Group_intersection

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_intersection

/*@

MPI_Group_intersection - Produces a group as the intersection of two existing
						 groups

Input Parameters:
+ group1 - first group (handle) 
- group2 - second group (handle) 

Output Parameter:
. newgroup - intersection group (handle) 

Notes:
The output group contains those processes that are in both 'group1' and 
'group2'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Group_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_intersection(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup)
{
	static const wchar_t FCNAME[] = L"MPI_Group_intersection";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr1 = NULL;
	MPID_Group *group_ptr2 = NULL;
	MPID_Group *new_group_ptr;
	int size1, size2, i, k, g1_idx, g2_idx, l1_pid, l2_pid, nnew;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_INTERSECTION);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_INTERSECTION);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group1, mpi_errno);
			MPIR_ERRTEST_GROUP(group2, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group1, group_ptr1 );
	MPID_Group_get_ptr( group2, group_ptr2 );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr1, mpi_errno );
			MPID_Group_valid_ptr( group_ptr2, mpi_errno );
			/* If either group_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_intersection %G %G %p", group1, group2, newgroup); }
	
	/* Return a group consisting of the members of group1 that are 
	   in group2 */
	size1 = group_ptr1->size;
	size2 = group_ptr2->size;
	/* Insure that the lpid lists are setup */
	MPIR_Group_setup_lpid_pairs( group_ptr1, group_ptr2 );

	for (i=0; i<size1; i++)
	{
		group_ptr1->lrank_to_lpid[i].flag = 0;
	}
	g1_idx = group_ptr1->idx_of_first_lpid;
	g2_idx = group_ptr2->idx_of_first_lpid;
	
	nnew = 0;
	while (g1_idx >= 0 && g2_idx >= 0)
	{
		l1_pid = group_ptr1->lrank_to_lpid[g1_idx].lpid;
		l2_pid = group_ptr2->lrank_to_lpid[g2_idx].lpid;
		if (l1_pid < l2_pid)
		{
			g1_idx = group_ptr1->lrank_to_lpid[g1_idx].next_lpid;
		}
		else if (l1_pid > l2_pid)
		{
			g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
		}
		else
		{
			/* Equal */
			group_ptr1->lrank_to_lpid[g1_idx].flag = 1;
			g1_idx = group_ptr1->lrank_to_lpid[g1_idx].next_lpid;
			g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
			nnew ++;
		}
	}
	/* Create the group.  Handle the trivial case first */
	if (nnew == 0)
	{
		*newgroup = MPI_GROUP_EMPTY;
		goto fn_exit;
	}
	
	mpi_errno = MPIR_Group_create( nnew, &new_group_ptr );
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno)
	{
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */
	new_group_ptr->rank = MPI_UNDEFINED;
	k = 0;
	for (i=0; i<size1; i++)
	{
		if (group_ptr1->lrank_to_lpid[i].flag)
		{
			new_group_ptr->lrank_to_lpid[k].lrank = k;
			new_group_ptr->lrank_to_lpid[k].lpid = group_ptr1->lrank_to_lpid[i].lpid;
			if (i == group_ptr1->rank) 
				new_group_ptr->rank = k;
			k++;
		}
	}

	*newgroup = new_group_ptr->handle;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_INTERSECTION);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_intersection",
			L"**mpi_group_intersection %G %G %p", group1, group2, newgroup);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_incl
#define MPI_Group_incl PMPI_Group_incl

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_incl

/*@

MPI_Group_incl - Produces a group by reordering an existing group and taking
		only listed members

Input Parameters:
+ group - group (handle) 
. n - number of elements in array 'ranks' (and size of newgroup ) (integer) 
- ranks - ranks of processes in 'group' to appear in 'newgroup' (array of 
integers) 

Output Parameter:
. newgroup - new group derived from above, in the order defined by 'ranks' 
(handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_ARG
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_RANK

.seealso: MPI_Group_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_incl(MPI_Group group, int n, int *ranks, MPI_Group *newgroup)
{
	static const wchar_t FCNAME[] = L"MPI_Group_incl";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr = NULL, *new_group_ptr = NULL;
	int i;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_INCL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_INCL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group, mpi_errno);
			MPIR_ERRTEST_ARGNEG(n,"n",mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group, group_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr, mpi_errno );
			/* If group_ptr is not valid, it will be reset to null */
			if (group_ptr)
			{
				mpi_errno = MPIR_Group_check_valid_ranks( group_ptr, ranks, n );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_incl %G %d %p %p", group, n, ranks, newgroup); }
	
	if (n == 0)
	{
		*newgroup = MPI_GROUP_EMPTY;
		goto fn_exit;
	}

	/* Allocate a new group and lrank_to_lpid array */
	mpi_errno = MPIR_Group_create( n, &new_group_ptr );
	if (mpi_errno) goto fn_fail;
	
	new_group_ptr->rank = MPI_UNDEFINED;
	for (i=0; i<n; i++)
	{
		new_group_ptr->lrank_to_lpid[i].lrank = i;
		new_group_ptr->lrank_to_lpid[i].lpid = group_ptr->lrank_to_lpid[ranks[i]].lpid;
		if (ranks[i] == group_ptr->rank) new_group_ptr->rank = i;
	}
	new_group_ptr->size = n;
	new_group_ptr->idx_of_first_lpid = -1;

	*newgroup = new_group_ptr->handle;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_INCL);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_incl",
			L"**mpi_group_incl %G %d %p %p", group, n, ranks, newgroup);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_free
#define MPI_Group_free PMPI_Group_free

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_free

/*@

MPI_Group_free - Frees a group

Input Parameter:
. group - group to free (handle) 

Notes:
On output, group is set to 'MPI_GROUP_NULL'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_PERM_GROUP
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_free(MPI_Group *group)
{
	static const wchar_t FCNAME[] = L"MPI_Group_free";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr = NULL;

	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_FREE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_FREE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(*group, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( *group, group_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr, mpi_errno );
			/* If group_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_free %p", group); }
	
	/* Do not free MPI_GROUP_EMPTY */
	if (*group != MPI_GROUP_EMPTY)
	{ 
		mpi_errno = MPIR_Group_release(group_ptr);
	}

	*group = MPI_GROUP_NULL;
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_FREE);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_free", L"**mpi_group_free %p", group);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_excl
#define MPI_Group_excl PMPI_Group_excl

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_excl

/*@

MPI_Group_excl - Produces a group by reordering an existing group and taking
		only unlisted members

Input Parameters:
+ group - group (handle) 
. n - number of elements in array 'ranks' (integer) 
- ranks - array of integer ranks in 'group' not to appear in 'newgroup' 

Output Parameter:
. newgroup - new group derived from above, preserving the order defined by 
 'group' (handle) 

Note:  
The MPI standard requires that each of the ranks to excluded must be
a valid rank in the group and all elements must be distinct or the
function is erroneous.  

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_ARG
.N MPI_ERR_RANK

.seealso: MPI_Group_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_excl(MPI_Group group, int n, int *ranks, MPI_Group *newgroup)
{
	static const wchar_t FCNAME[] = L"MPI_Group_excl";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr = NULL, *new_group_ptr;
	int size, i, newi;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_EXCL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_EXCL);
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group, mpi_errno);
			MPIR_ERRTEST_ARGNEG(n,"n",mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group, group_ptr );

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr, mpi_errno );
			/* If group_ptr is not valid, it will be reset to null */
			if (group_ptr)
			{
				mpi_errno = MPIR_Group_check_valid_ranks( group_ptr, ranks, n );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_excl %G %d %p %p", group, n, ranks, newgroup); }
	
	/* Allocate a new group and lrank_to_lpid array */
	size = group_ptr->size;
	if (size == n)
	{
		*newgroup = MPI_GROUP_EMPTY;
		goto fn_exit;
	}
	mpi_errno = MPIR_Group_create( size - n, &new_group_ptr );
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno)
	{
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */
	new_group_ptr->rank = MPI_UNDEFINED;
	/* Use flag fields to mark the members to *exclude* . */

	for (i=0; i<size; i++)
	{
		group_ptr->lrank_to_lpid[i].flag = 0;
	}
	for (i=0; i<n; i++)
	{
		group_ptr->lrank_to_lpid[ranks[i]].flag = 1;
	}
	
	newi = 0;
	for (i=0; i<size; i++)
	{
		if (group_ptr->lrank_to_lpid[i].flag == 0)
		{
			new_group_ptr->lrank_to_lpid[newi].lrank = newi;
			new_group_ptr->lrank_to_lpid[newi].lpid = group_ptr->lrank_to_lpid[i].lpid;
			if (group_ptr->rank == i) new_group_ptr->rank = newi;
			newi++;
		}
	}

	new_group_ptr->size = size - n;
	new_group_ptr->idx_of_first_lpid = -1;

	*newgroup = new_group_ptr->handle;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_EXCL);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_excl",
			L"**mpi_group_excl %G %d %p %p", group, n, ranks, newgroup);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_difference
#define MPI_Group_difference PMPI_Group_difference

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_difference

/*@

MPI_Group_difference - Makes a group from the difference of two groups

Input Parameters:
+ group1 - first group (handle) 
- group2 - second group (handle) 

Output Parameter:
. newgroup - difference group (handle) 

Notes:
The generated group containc the members of 'group1' that are not in 'group2'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Group_free
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_difference(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup)
{
	static const wchar_t FCNAME[] = L"MPI_Group_difference";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr1 = NULL;
	MPID_Group *group_ptr2 = NULL;
	MPID_Group *new_group_ptr;
	int size1, size2, i, k, g1_idx, g2_idx, l1_pid, l2_pid, nnew;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_DIFFERENCE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_DIFFERENCE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_GROUP(group1, mpi_errno);
			MPIR_ERRTEST_GROUP(group2, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group1, group_ptr1 );
	MPID_Group_get_ptr( group2, group_ptr2 );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr1, mpi_errno );
			MPID_Group_valid_ptr( group_ptr2, mpi_errno );
			/* If either group_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_difference %G %G %p", group1, group2, newgroup); }
	
	/* Return a group consisting of the members of group1 that are *not*
	   in group2 */
	size1 = group_ptr1->size;
	size2 = group_ptr2->size;
	/* Insure that the lpid lists are setup */
	MPIR_Group_setup_lpid_pairs( group_ptr1, group_ptr2 );

	for (i=0; i<size1; i++)
	{
		group_ptr1->lrank_to_lpid[i].flag = 0;
	}
	g1_idx = group_ptr1->idx_of_first_lpid;
	g2_idx = group_ptr2->idx_of_first_lpid;
	
	nnew = size1;
	while (g1_idx >= 0 && g2_idx >= 0)
	{
		l1_pid = group_ptr1->lrank_to_lpid[g1_idx].lpid;
		l2_pid = group_ptr2->lrank_to_lpid[g2_idx].lpid;
		if (l1_pid < l2_pid)
		{
			g1_idx = group_ptr1->lrank_to_lpid[g1_idx].next_lpid;
		}
		else if (l1_pid > l2_pid)
		{
			g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
		}
		else
		{
			/* Equal */
			group_ptr1->lrank_to_lpid[g1_idx].flag = 1;
			g1_idx = group_ptr1->lrank_to_lpid[g1_idx].next_lpid;
			g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
			nnew --;
		}
	}
	/* Create the group */
	if (nnew == 0)
	{
		/* See 5.3.2, Group Constructors.  For many group routines,
		   the standard explicitly says to return MPI_GROUP_EMPTY; 
		   for others it is implied */
		*newgroup = MPI_GROUP_EMPTY;
		goto fn_exit;
	}
	else
	{
		mpi_errno = MPIR_Group_create( nnew, &new_group_ptr );
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno)
		{
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
		new_group_ptr->rank = MPI_UNDEFINED;
		k = 0;
		for (i=0; i<size1; i++)
		{
			if (!group_ptr1->lrank_to_lpid[i].flag)
			{
				new_group_ptr->lrank_to_lpid[k].lrank = k;
				new_group_ptr->lrank_to_lpid[k].lpid = group_ptr1->lrank_to_lpid[i].lpid;
				if (i == group_ptr1->rank) 
					new_group_ptr->rank = k;
				k++;
			}
		}
	}

	*newgroup = new_group_ptr->handle;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_DIFFERENCE);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_difference",
			L"**mpi_group_difference %G %G %p", group1, group2, newgroup);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Group_compare
#define MPI_Group_compare PMPI_Group_compare

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Group_compare

/*@

MPI_Group_compare - Compares two groups

Input Parameters:
+ group1 - group1 (handle) 
- group2 - group2 (handle) 

Output Parameter:
. result - integer which is 'MPI_IDENT' if the order and members of
the two groups are the same, 'MPI_SIMILAR' if only the members are the same,
and 'MPI_UNEQUAL' otherwise

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Group_compare(MPI_Group group1, MPI_Group group2, int *result)
{
	static const wchar_t FCNAME[] = L"MPI_Group_compare";
	int mpi_errno = MPI_SUCCESS;
	MPID_Group *group_ptr1 = NULL;
	MPID_Group *group_ptr2 = NULL;
	int g1_idx, g2_idx, size, i;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GROUP_COMPARE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	/* The routines that setup the group data structures must be executed
	   within a mutex.  As most of the group routines are not performance
	   critical, we simple run these routines within the SINGLE_CS */
	MPIU_THREAD_SINGLE_CS_ENTER("group");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GROUP_COMPARE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL( result, "result", mpi_errno );
			MPIR_ERRTEST_GROUP(group1, mpi_errno);
			MPIR_ERRTEST_GROUP(group2, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Group_get_ptr( group1, group_ptr1 );
	MPID_Group_get_ptr( group2, group_ptr2 );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate group_ptr */
			MPID_Group_valid_ptr( group_ptr1, mpi_errno );
			MPID_Group_valid_ptr( group_ptr2, mpi_errno );
			/* If group_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_group_compare %G %G %p", group1, group2, result); }
	
	/* See if their sizes are equal */
	if (group_ptr1->size != group_ptr2->size)
	{
		*result = MPI_UNEQUAL;
		goto fn_exit;
	}

	/* Run through the lrank to lpid lists of each group in lpid order
	   to see if the same processes are involved */
	g1_idx = group_ptr1->idx_of_first_lpid;
	g2_idx = group_ptr2->idx_of_first_lpid;
	/* If the lpid list hasn't been created, do it now */
	if (g1_idx < 0)
	{ 
		MPIR_Group_setup_lpid_list( group_ptr1 ); 
		g1_idx = group_ptr1->idx_of_first_lpid;
	}
	if (g2_idx < 0)
	{ 
		MPIR_Group_setup_lpid_list( group_ptr2 ); 
		g2_idx = group_ptr2->idx_of_first_lpid;
	}
	while (g1_idx >= 0 && g2_idx >= 0)
	{
		if (group_ptr1->lrank_to_lpid[g1_idx].lpid != group_ptr2->lrank_to_lpid[g2_idx].lpid)
		{
			*result = MPI_UNEQUAL;
			goto fn_exit;
		}
		g1_idx = group_ptr1->lrank_to_lpid[g1_idx].next_lpid;
		g2_idx = group_ptr2->lrank_to_lpid[g2_idx].next_lpid;
	}

	/* See if the processes are in the same order by rank */
	size = group_ptr1->size;
	for (i=0; i<size; i++)
	{
		if (group_ptr1->lrank_to_lpid[i].lpid != 
			group_ptr2->lrank_to_lpid[i].lpid)
		{
			*result = MPI_SIMILAR;
			goto fn_exit;
		}
	}
		
	/* If we reach here, the groups are identical */
	*result = MPI_IDENT;
   
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GROUP_COMPARE);
	MPIU_THREAD_SINGLE_CS_EXIT("group");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_group_compare",
			L"**mpi_group_compare %G %G %p", group1, group2, result);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
