/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#include "mpi_init.h"
#include <stdlib.h>

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_InitA
#undef MPI_InitW
#define MPI_InitA PMPI_InitA
#define MPI_InitW PMPI_InitW

/* Fortran logical values. extern'd in mpiimpl.h */
/* MPI_Fint MPIR_F_TRUE, MPIR_F_FALSE; */

/* Any internal routines can go here.  Make them static if possible */
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Init

/*@
   MPI_Init - Initialize the MPI execution environment

   Input Parameters:
+  argc - Pointer to the number of arguments 
-  argv - Pointer to the argument vector

Thread and Signal Safety:
This routine must be called by one thread only.  That thread is called
the `main thread` and must be the thread that calls 'MPI_Finalize'.

Notes:
   The MPI standard does not say what a program can do before an 'MPI_INIT' or
   after an 'MPI_FINALIZE'.  In the MPICH implementation, you should do
   as little as possible.  In particular, avoid anything that changes the
   external state of the program, such as opening files, reading standard
   input or writing to standard output.

Notes for Fortran:
The Fortran binding for 'MPI_Init' has only the error return
.vb
	subroutine MPI_INIT( ierr )
	integer ierr
.ve

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INIT

.seealso: MPI_Init_thread, MPI_Finalize
@*/
MPI_DLL_SPEC int MPI_CALL MPI_InitA( int *argc, char ***argv )
{
	static const wchar_t FCNAME[] = L"MPI_InitA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t **wargv = NULL;
	int i;
	MPID_MPI_INIT_STATE_DECL(MPID_STATE_MPI_INITA);

	MPID_MPI_INIT_FUNC_ENTER(MPID_STATE_MPI_INITA);

	if (argv != NULL && argc != NULL)
	{
		wargv = (wchar_t**)MPIU_Malloc(((*argc) + 1) * sizeof(wchar_t*));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, wargv, ((*argc) + 1)*sizeof(wchar_t), "wargv");
		for (i=0; i<((*argc) + 1); i++) wargv[i] = NULL; /* Set them all to NULL first so errors can be handled at any time */
		for (i=0; i<((*argc) + 1); i++)
		{
			if ((*argv)[i] != NULL)
			{
				mpi_errno = MPIU_Mbstowcs((*argv)[i], &wargv[i]);
				MPIU_ERR_CHK_POP(mpi_errno);
			}
		}
	}

	mpi_errno = MPI_InitW(argc, &wargv);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wargv != NULL)
	{
		for (i=0; i<((*argc) + 1); i++)
		{
			if (wargv[i] != NULL)
			{
				MPIU_Free(wargv[i]);
			}
		}
		MPIU_Free(wargv);
	}
	MPID_MPI_INIT_FUNC_EXIT(MPID_STATE_MPI_INITA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_InitW( int *argc, wchar_t ***argv )
{
	static const wchar_t FCNAME[] = L"MPI_Init";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_INIT_STATE_DECL(MPID_STATE_MPI_INIT);

	MPID_CS_INITIALIZE();
	MPIU_THREAD_SINGLE_CS_ENTER("init");
	MPID_MPI_INIT_FUNC_ENTER(MPID_STATE_MPI_INIT);
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			if (MPIR_Process.initialized != MPICH_PRE_INIT)
			{
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**inittwice", NULL );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	MPIU_Init_call_history(argc, argv);
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_init %p %p", argc, argv); }
	
	mpi_errno = MPIR_Init_thread( argc, argv, MPI_THREAD_SINGLE, (int *)0 );
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */
	
	MPID_MPI_INIT_FUNC_EXIT(MPID_STATE_MPI_INIT);
	MPIU_THREAD_SINGLE_CS_EXIT("init");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_HANDLING
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_init", L"**mpi_init %p %p", argc, argv);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	MPIU_THREAD_SINGLE_CS_EXIT("init");
	MPID_CS_FINALIZE();
	return mpi_errno;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Initialized
#define MPI_Initialized PMPI_Initialized
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Initialized

/*@
   MPI_Initialized - Indicates whether 'MPI_Init' has been called.

Output Argument:
. flag - Flag is true if 'MPI_Init' or 'MPI_Init_thread' has been called and 
		 false otherwise.  

   Notes:

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Initialized( int *flag )
{
	static const wchar_t FCNAME[] = L"MPI_Initialized";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INITIALIZED);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INITIALIZED);
	
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Should check that flag is not null */
			if (flag == NULL)
			{
				mpi_errno = MPI_ERR_ARG;
				goto fn_fail;
			}
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	MPIU_Init_call_history(NULL, NULL);
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_initialized %p", flag); }
	
	*flag = (MPIR_Process.initialized >= MPICH_WITHIN_MPI);
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INITIALIZED);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	if (MPIR_Process.initialized == MPICH_WITHIN_MPI)
	{ 
#	   ifdef HAVE_ERROR_CHECKING
		{
			mpi_errno = MPIR_Err_create_code(
				mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_initialized",
				L"**mpi_initialized %p", flag);
		}
#	   endif
		
		mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	}
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Finalize
#define MPI_Finalize PMPI_Finalize

/* Any internal routines can go here.  Make them static if possible */

/* The following routines provide a callback facility for modules that need 
   some code called on exit.  This method allows us to avoid forcing 
   MPI_Finalize to know the routine names a priori.  Any module that wants to 
   have a callback calls MPIR_Add_finalize( routine, extra, priority ).
   
 */
PMPI_LOCAL void MPIR_Call_finalize_callbacks( int, int );
typedef struct Finalize_func_t
{
	int (*f)( void * );	  /* The function to call */
	void *extra_data;		/* Data for the function */
	int  priority;		   /* priority is used to control the order
								in which the callbacks are invoked */
} Finalize_func_t;
#define MAX_FINALIZE_FUNC 16
static Finalize_func_t fstack[MAX_FINALIZE_FUNC];
static int fstack_sp = 0;
static int fstack_max_priority = 0;

#undef FUNCNAME
#define FUNCNAME MPIR_Add_finalize
void MPIR_Add_finalize( int (*f)( void * ), void *extra_data, int priority )
{
	static const wchar_t FCNAME[] = L"MPIR_Add_finalize";
	MPIDI_STATE_DECL(MPID_STATE_MPIR_ADD_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_ADD_FINALIZE);
	/* --BEGIN ERROR HANDLING-- */
	if (fstack_sp >= MAX_FINALIZE_FUNC)
	{
		/* This is a little tricky.  We may want to check the state of
		   MPIR_Process.initialized to decide how to signal the error */
		(void)MPIU_Internal_error_printf( L"overflow in finalize stack!\n" );
		if (MPIR_Process.initialized == MPICH_WITHIN_MPI)
		{
			MPID_Abort( NULL, MPI_SUCCESS, 13, NULL );
		}
		else
		{
			ExitProcess(1);
		}
	}
	/* --END ERROR HANDLING-- */
	fstack[fstack_sp].f			= f;
	fstack[fstack_sp].priority	 = priority;
	fstack[fstack_sp++].extra_data = extra_data;

	if (priority > fstack_max_priority) 
		fstack_max_priority = priority;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_ADD_FINALIZE);
}

/* Invoke the registered callbacks */
#undef FUNCNAME
#define FUNCNAME MPIR_Call_finalize_callbacks
PMPI_LOCAL void MPIR_Call_finalize_callbacks( int min_prio, int max_prio )
{
	static const wchar_t FCNAME[] = L"MPIR_Call_finalize_callbacks";
	int i, j;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_CALL_FINALIZE_CALLBACKS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_CALL_FINALIZE_CALLBACKS);

	MPIU_UNREFERENCED_ARG(max_prio);

	for (j=fstack_max_priority; j>=min_prio; j--)
	{
		for (i=fstack_sp-1; i>=0; i--)
		{
			if (fstack[i].f && fstack[i].priority == j)
			{
				fstack[i].f( fstack[i].extra_data );
				fstack[i].f = 0;
			}
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_CALL_FINALIZE_CALLBACKS);
}
#else
#ifndef USE_WEAK_SYMBOLS
PMPI_LOCAL void MPIR_Call_finalize_callbacks( int, int );
#endif
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Finalize

/*@
   MPI_Finalize - Terminates MPI execution environment

   Notes:
   All processes must call this routine before exiting.  The number of
   processes running `after` this routine is called is undefined; 
   it is best not to perform much more than a 'return rc' after calling
   'MPI_Finalize'.

Thread and Signal Safety:
The MPI standard requires that 'MPI_Finalize' be called `only` by the same 
thread that initialized MPI with either 'MPI_Init' or 'MPI_Init_thread'.

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Finalize( void )
{
	static const wchar_t FCNAME[] = L"MPI_Finalize";
	int mpi_errno = MPI_SUCCESS;
#if defined(HAVE_USLEEP) && defined(USE_COVERAGE)
	int rank=0;
#endif
	MPID_MPI_FINALIZE_STATE_DECL(MPID_STATE_MPI_FINALIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("init");
	MPID_MPI_FINALIZE_FUNC_ENTER(MPID_STATE_MPI_FINALIZE);
	
	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"MPI_Finalize()"); }
	
#if defined(HAVE_USLEEP) && defined(USE_COVERAGE)
	/* We need to get the rank before freeing MPI_COMM_WORLD */
	rank = MPIR_Process.comm_world->rank;
#endif	

	/* Remove the attributes, executing the attribute delete routine.
	   Do this only if the attribute functions are defined. */ 
	/* The standard (MPI-2, section 4.8) says that the attributes on 
	   MPI_COMM_SELF are deleted before almost anything else happens */
	if (MPIR_Process.attr_free && MPIR_Process.comm_self->attributes)
	{
		mpi_errno = MPIR_Process.attr_free( MPI_COMM_SELF, MPIR_Process.comm_self->attributes);
	}
	if (MPIR_Process.attr_free && MPIR_Process.comm_world->attributes)
	{
		mpi_errno = MPIR_Process.attr_free( MPI_COMM_WORLD, MPIR_Process.comm_world->attributes);
	}

	/* FIXME: Why is this not one of the finalize callbacks?.  Do we need
	   pre and post MPID_Finalize callbacks? */
	MPIU_Timer_finalize();

	/* Call the high-priority callbacks */
	MPIR_Call_finalize_callbacks( MPIR_FINALIZE_CALLBACK_PRIO+1, MPIR_FINALIZE_CALLBACK_MAX_PRIO );

	mpi_errno = MPID_Finalize();
	if (mpi_errno)
	{
		MPIU_ERR_POP(mpi_errno);
	}
	
	/* delete local and remote groups on comm_world and comm_self if
	   they had been created (should we use a function pointer here
	   as well to avoid loading the group code?) */
	if (MPIR_Process.comm_world->local_group)
		MPIR_Group_release(MPIR_Process.comm_world->local_group);
	if (MPIR_Process.comm_world->remote_group)
		MPIR_Group_release(MPIR_Process.comm_world->remote_group);
	if (MPIR_Process.comm_self->local_group)
		MPIR_Group_release(MPIR_Process.comm_self->local_group);
	if (MPIR_Process.comm_self->remote_group)
		MPIR_Group_release(MPIR_Process.comm_self->remote_group);

	/* Call the low-priority (post Finalize) callbacks */
	MPIR_Call_finalize_callbacks( 0, MPIR_FINALIZE_CALLBACK_PRIO-1 );

	/* FIXME: Both the memory tracing and debug nesting code blocks should
	   be finalize callbacks */
	/* If memory debugging is enabled, check the memory here, after all
	   finalize callbacks */
#ifdef USE_MEMORY_TRACING
	/* FIXME: the (1) in the if test should be replaced by a 
	   parameter call */
	/* FIXME: We'd like to arrange for the mem dump output to
	   go to separate files or to be sorted by rank (note that
	   the rank is at the head of the line) */
	if (1)
	{
		/* The second argument is the min id to print; memory allocated 
		   after MPI_Init is given an id of one.  This allows us to
		   ignore, if desired, memory leaks in the MPID_Init call */
		MPIU_trdump( (void *)0, -1 );
	}
#endif
#ifdef MPICH_DEBUG_NESTING
	/* FIXME: the (1) in the if test should be replaced by a 
	   parameter call */
	if (1)
	{
		MPIU_THREADPRIV_DECL;

		MPIU_THREADPRIV_GET;
		/* Check for an error in the nesting level */
		if (MPIR_Nest_value())
		{
			int i,n;
			n = MPIR_Nest_value();
			fwprintf( stderr, L"Unexpected value for nesting level = %d\n", n );
			fwprintf( stderr, L"Nest stack is:\n" );
			for (i=n-1; i>=0; i--)
			{
				fwprintf( stderr, L"\t[%d] %s:%d\n", i, 
						 MPIU_THREADPRIV_FIELD(nestinfo[i].file), 
						 MPIU_THREADPRIV_FIELD(nestinfo[i].line) );
			}
		}
	}
#endif

	MPIR_Process.initialized = MPICH_POST_FINALIZED;

#if defined(HAVE_USLEEP) && defined(USE_COVERAGE)
	/* If performing coverage analysis, make each process sleep for
	   rank * 100 ms, to give time for the coverage tool to write out
	   any files */
	usleep( rank * 100000 );
#endif
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FINALIZE_FUNC_EXIT(MPID_STATE_MPI_FINALIZE);
	MPIU_THREAD_SINGLE_CS_EXIT("init");
	MPIU_THREAD_SINGLE_CS_FINALIZE;
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_finalize", 0);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Finalized
#define MPI_Finalized PMPI_Finalized
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Finalized

/*@
   MPI_Finalized - Indicates whether 'MPI_Finalize' has been called.

Output Parameter:
. flag - Flag is true if 'MPI_Finalize' has been called and false otherwise. 
	 (logical)

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Finalized( int *flag )
{
	static const wchar_t FCNAME[] = L"MPI_Finalized";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FINALIZED);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FINALIZED);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Should check that flag is not null */
			if (flag == NULL)
			{
				mpi_errno = MPI_ERR_ARG;
				goto fn_fail;
			}
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_finalized %p", flag); }
	
	*flag = (MPIR_Process.initialized >= MPICH_POST_FINALIZED);
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FINALIZED);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	if (MPIR_Process.initialized == MPICH_WITHIN_MPI)
	{ 
#	   ifdef HAVE_ERROR_CHECKING
		{
			mpi_errno = MPIR_Err_create_code(
				mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_finalized",
				L"**mpi_finalized %p", flag);
		}
#	   endif
		
		mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	}
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Is_thread_main
#define MPI_Is_thread_main PMPI_Is_thread_main
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Is_thread_main

/*@
   MPI_Is_thread_main - Returns a flag indicating whether this thread called 
						'MPI_Init' or 'MPI_Init_thread'

   Output Parameter:
. flag - Flag is true if 'MPI_Init' or 'MPI_Init_thread' has been called by 
		 this thread and false otherwise.  (logical)

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Is_thread_main( int *flag )
{
	static const wchar_t FCNAME[] = L"MPI_Is_thread_main";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_IS_THREAD_MAIN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(flag,L"flag",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_IS_THREAD_MAIN);
	
	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_is_thread_main %p", flag); }
#   if MPICH_THREAD_LEVEL <= MPI_THREAD_FUNNELED
	{
		*flag = TRUE;
	}
#   else
	{
		MPID_Thread_id_t my_thread_id;

		MPID_Thread_self(&my_thread_id);
		MPID_Thread_same(&MPIR_Process.master_thread, &my_thread_id, flag);
	}
#   endif
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_IS_THREAD_MAIN);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, 
			MPI_ERR_OTHER, L"**mpi_is_thread_main",
			L"**mpi_is_thread_main %p", flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Query_thread
#define MPI_Query_thread PMPI_Query_thread
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Query_thread

/*@
   MPI_Query_thread - Return the level of thread support provided by the MPI 
	library

   Output Parameter:
.  provided - Level of thread support provided.  This is the same value
   that was returned in the 'provided' argument in 'MPI_Init_thread'.

   Notes:
   The valid values for the level of thread support are\:
+ MPI_THREAD_SINGLE - Only one thread will execute. 
. MPI_THREAD_FUNNELED - The process may be multi-threaded, but only the main 
  thread will make MPI calls (all MPI calls are funneled to the 
   main thread). 
. MPI_THREAD_SERIALIZED - The process may be multi-threaded, and multiple 
  threads may make MPI calls, but only one at a time: MPI calls are not 
  made concurrently from two distinct threads (all MPI calls are serialized). 
- MPI_THREAD_MULTIPLE - Multiple threads may call MPI, with no restrictions. 

   If 'MPI_Init' was called instead of 'MPI_Init_thread', the level of
   thread support is defined by the implementation.  This routine allows
   you to find out the provided level.  It is also useful for library 
   routines that discover that MPI has already been initialized and
   wish to determine what level of thread support is available.

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Query_thread( int *provided )
{
	static const wchar_t FCNAME[] = L"MPI_Query_thread";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_QUERY_THREAD);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_QUERY_THREAD);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(provided,L"provided",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_query_thread %p", provided); }
	*provided = MPIR_Process.thread_provided;
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_QUERY_THREAD);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_query_thread",
			L"**mpi_query_thread %p", provided);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm(NULL, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Abort
#define MPI_Abort PMPI_Abort

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Abort

/*@
   MPI_Abort - Terminates MPI execution environment

Input Parameters:
+ comm - communicator of tasks to abort 
- errorcode - error code to return to invoking environment 

Notes:
Terminates all MPI processes associated with the communicator 'comm'; in
most systems (all to date), terminates `all` processes.

.N NotThreadSafe
Because the 'MPI_Abort' routine is intended to ensure that an MPI
process (and possibly an entire job), it cannot wait for a thread to 
release a lock or other mechanism for atomic access.

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Abort(MPI_Comm comm, int errorcode)
{
	static const wchar_t FCNAME[] = L"MPI_Abort";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	/* FIXME: 100 is arbitrary and may not be long enough */
	wchar_t abort_str[100], comm_name[MPI_MAX_NAME_STRING];
	int len = MPI_MAX_NAME_STRING;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ABORT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	/* FIXME: It isn't clear that Abort should wait for a critical section,
	   since that could result in the Abort hanging if another routine is
	   hung holding the critical section.  Also note the "not thread-safe"
	   comment in the description of MPI_Abort above. */
	MPIU_THREAD_SINGLE_CS_ENTER("init");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ABORT);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Get handles to MPI objects. */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_abort %C %d", comm, errorcode); }
	
	if (!comm_ptr)
	{
		/* Use comm world if the communicator is not valid */
		comm_ptr = MPIR_Process.comm_world;
	}

	NMPI_Comm_get_name(mpi_errno, (comm, comm_name, &len));
	if (len == 0)
	{
		MPIU_Snprintf(comm_name, MPI_MAX_NAME_STRING, _TRUNCATE, L"comm=0x%X", comm);
	}
	/* FIXME: This is not internationalized */
	MPIU_Snprintf(abort_str, 100, _TRUNCATE, L"application called MPI_Abort(%s, %d) - process %d", comm_name, errorcode, comm_ptr->rank);
	mpi_errno = MPID_Abort( comm_ptr, mpi_errno, errorcode, abort_str );
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	/* MPID_Abort() should never return MPI_SUCCESS */
	MPIU_Assert(0);
	/* --END ERROR HANDLING-- */
	/* ... end of body of routine ... */
	
  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ABORT);
	MPIU_THREAD_SINGLE_CS_EXIT("init");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	/* It is not clear that doing aborting abort makes sense.  We may
	   want to specify that erroneous arguments to MPI_Abort will
	   cause an immediate abort. */
	   
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_abort",
			L"**mpi_abort %C %d", comm, errorcode);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm(comm_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
