/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"
#include "datatype.h"
#include "mpi_init.h"
#include "pmi.h"
#ifdef HAVE_CRTDBG_H
#include <crtdbg.h>
#endif

/* -- Begin Profiling Symbol Block for routine MPI_Init_thread */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Init_thread = PMPI_Init_thread
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Init_thread  MPI_Init_thread
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Init_thread as PMPI_Init_thread
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Init_threadA
#undef MPI_Init_threadW
#define MPI_Init_threadA PMPI_Init_threadA
#define MPI_Init_threadW PMPI_Init_threadW

/* Any internal routines can go here.  Make them static if possible */

/* Global variables can be initialized here */
MPICH_PerProcess_t MPIR_Process = { MPICH_PRE_INIT }; /* all others are irelevant */

/* These are initialized as null (avoids making these into common symbols).
   If the Fortran binding is supported, these can be initialized to 
   their Fortran values (MPI only requires that they be valid between
   MPI_Init and MPI_Finalize) */
MPI_DLL_SPEC MPI_Fint *MPI_F_STATUS_IGNORE = 0;
MPI_DLL_SPEC MPI_Fint *MPI_F_STATUSES_IGNORE = 0;

/* User-defined abort hook function.  Exiting here will prevent the system from
 * bringing up an error dialog box.
 */
static int assert_hook( int reportType, char *message, int *returnValue )
{
	MPIU_UNREFERENCED_ARG(reportType);
	fprintf(stderr, "%s", message);
	if (returnValue != NULL)
		ExitProcess((UINT)(*returnValue));
	ExitProcess((UINT)(-1));
	return TRUE;
}

/* DeinoMPI dll entry point */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpReserved)
{
	BOOL result = TRUE;
	hinstDLL;
	lpReserved;

	switch (fdwReason)
	{
		case DLL_PROCESS_ATTACH:
			break;

		case DLL_THREAD_ATTACH:
			/* allocate thread specific data */
			break;

		case DLL_THREAD_DETACH:
			/* free thread specific data */
			break;

		case DLL_PROCESS_DETACH:
			/*printf("process %d dll detaching.\n", MPIR_Process.comm_world->rank);*/
			break;
	}
	return result;
}

#if !defined(MPICH_IS_THREADED)
/* If single threaded, we preallocate this.  Otherwise, we create it */
MPICH_PerThread_t  MPIR_Thread = { 0 };
#elif defined(HAVE_RUNTIME_THREADCHECK)
/* If we may be single threaded, we need a preallocated version to use
   if we are single threaded case */
MPICH_PerThread_t  MPIR_ThreadSingle = { 0 };
#endif

static int *my_clique_ranks = NULL;
static int my_clique_size = 1;

int compare_ints(const void *p1, const void *p2)
{
	if ((*(int*)p1) > (*(int*)p2))
		return 1;
	if ((*(int*)p1) < (*(int*)p2))
		return -1;
	return 0;
}

/* Return the lowest rank on the same node as this process */
#undef FUNCNAME
#define FUNCNAME getNodeRootRank
static int getNodeRootRank(int *root_rank)
{
	static const wchar_t FCNAME[] = L"getNodeRootRank";
	int num_ranks;
	int pmi_errno;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_GETNODEROOTRANK);

	MPIDI_FUNC_ENTER(MPID_STATE_GETNODEROOTRANK);

	pmi_errno = PMI_Get_clique_size(&num_ranks);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER,
							 L"**pmi_get_clique_size", 
							 L"**pmi_get_clique_size %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	if (num_ranks < 1)
	{
		/* PMI_Get_clique_size returned an invalid size */
		MPIU_ERR_SETANDJUMP1(mpi_errno, MPI_ERR_OTHER,
			L"**pmi_invalid_clique_size",
			L"**pmi_invalid_clique_size %d", num_ranks);
	}
	my_clique_size = num_ranks;

	my_clique_ranks = MPIU_Malloc(num_ranks * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, my_clique_ranks, num_ranks*sizeof(int), L"my_clique_ranks");

	pmi_errno = PMI_Get_clique_ranks(my_clique_ranks, num_ranks);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER,
							 L"**pmi_get_clique_ranks", 
							 L"**pmi_get_clique_ranks %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	qsort(my_clique_ranks, num_ranks, sizeof(int), compare_ints);
	*root_rank = my_clique_ranks[0];

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_GETNODEROOTRANK);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME MPIR_Init_thread
int MPIR_Init_thread(int * argc, wchar_t ***argv, int required,
					 int * provided)
{
	static const wchar_t FCNAME[] = L"MPIR_Init_thread";
	int mpi_errno = MPI_SUCCESS;
	int has_args;
	int has_env;
	int thread_provided;
	MPIU_THREADPRIV_DECL;

	/* prevent the process from bringing up an error message window if mpich asserts */
	_CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ASSERT, _CRTDBG_FILE_STDERR );
	_CrtSetReportHook2(_CRT_RPTHOOK_INSTALL, assert_hook);
#ifdef _WIN64
	{
	/* FIXME: This severly degrades performance but fixes alignment issues 
	   with the datatype code. */
	/* Prevent misaligned faults on Win64 machines */
	UINT mode, old_mode;
	
	old_mode = SetErrorMode(SEM_NOALIGNMENTFAULTEXCEPT);
	mode = old_mode | SEM_NOALIGNMENTFAULTEXCEPT;
	SetErrorMode(mode);
	}
#endif

	/* We need this inorder to implement IS_THREAD_MAIN */
#   if (MPICH_THREAD_LEVEL >= MPI_THREAD_SERIALIZED)
	{
		MPID_Thread_self(&MPIR_Process.master_thread);
	}
#   endif
	
#ifdef HAVE_ERROR_CHECKING
	/* Eventually this will support commandline and environment options
	 for controlling error checks.  It will use the routine 
	 MPIR_Err_init, which does as little as possible (e.g., it only 
	 determines the value of do_error_checks) */
	MPIR_Process.do_error_checks = 1;
#else
	MPIR_Process.do_error_checks = 0;
#endif

	/* Initialize necessary subsystems and setup the predefined attribute
	   values.  Subsystems may change these values. */
	MPIR_Process.attrs.appnum		  = -1;
	MPIR_Process.attrs.host			= 0;
	MPIR_Process.attrs.io			  = 0;
	MPIR_Process.attrs.lastusedcode	= MPI_ERR_LASTCODE;
	MPIR_Process.attrs.tag_ub		  = 0;
	MPIR_Process.attrs.universe		= MPIR_UNIVERSE_SIZE_NOT_SET;
	MPIR_Process.attrs.wtime_is_global = 0;

	/* Set the functions used to duplicate attributes.  These are 
	   when the first corresponding keyval is created */
	MPIR_Process.attr_dup  = 0;
	MPIR_Process.attr_free = 0;

#ifdef HAVE_CXX_BINDING
	/* Set the functions used to call functions in the C++ binding 
	   for reductions and attribute operations.  These are null
	   until a C++ operation is defined.  This allows the C code
	   that implements these operations to not invoke a C++ code
	   directly, which may force the inclusion of symbols known only
	   to the C++ compiler (e.g., under more non-GNU compilers, including
	   Solaris and IRIX). */
	MPIR_Process.cxx_call_op_fn = 0;
	MPIR_Process.cxx_call_delfn = 0;

#endif
	/* This allows the device to select an alternative function for 
	   dimsCreate */
	MPIR_Process.dimsCreate	 = 0;

	/* "Allocate" from the reserved space for builtin communicators and
	   (partially) initialize predefined communicators.  comm_parent is
	   intially NULL and will be allocated by the device if the process group
	   was started using one of the MPI_Comm_spawn functions. */
	MPIR_Process.comm_world				  = MPID_Comm_builtin + 0;
	MPIR_Process.comm_world->handle		  = MPI_COMM_WORLD;
	MPIU_Object_set_ref( MPIR_Process.comm_world, 1 );
    MPIR_Process.comm_world->context_id	    = 0 << MPID_CONTEXT_PREFIX_SHIFT;
    MPIR_Process.comm_world->recvcontext_id = 0 << MPID_CONTEXT_PREFIX_SHIFT;
	MPIR_Process.comm_world->attributes		  = NULL;
	MPIR_Process.comm_world->local_group  = NULL;
	MPIR_Process.comm_world->remote_group = NULL;
	MPIR_Process.comm_world->comm_kind		  = MPID_INTRACOMM;
	/* This initialization of the comm name could be done only when 
	   comm_get_name is called */
	MPIU_Wcsncpy(MPIR_Process.comm_world->name, L"MPI_COMM_WORLD", MPI_MAX_OBJECT_NAME);
	MPIR_Process.comm_world->errhandler		  = NULL; /* XXX */
	MPIR_Process.comm_world->coll_fns		  = NULL; /* XXX */
	MPIR_Process.comm_world->topo_fns	 = NULL; /* XXX */
	MPIR_Process.comm_world->comm_net	 = MPI_COMM_NULL;
	MPIR_Process.comm_world->comm_node	= MPI_COMM_NULL;
	MPIR_Process.comm_world->sub_comms_initialized = 0;
	
	MPIR_Process.comm_self				 = MPID_Comm_builtin + 1;
	MPIR_Process.comm_self->handle		 = MPI_COMM_SELF;
	MPIU_Object_set_ref( MPIR_Process.comm_self, 1 );
    MPIR_Process.comm_self->context_id	    = 1 << MPID_CONTEXT_PREFIX_SHIFT;
    MPIR_Process.comm_self->recvcontext_id  = 1 << MPID_CONTEXT_PREFIX_SHIFT;
	MPIR_Process.comm_self->attributes		 = NULL;
	MPIR_Process.comm_self->local_group		 = NULL;
	MPIR_Process.comm_self->remote_group = NULL;
	MPIR_Process.comm_self->comm_kind		 = MPID_INTRACOMM;
	MPIU_Wcsncpy(MPIR_Process.comm_self->name, L"MPI_COMM_SELF", MPI_MAX_OBJECT_NAME);
	MPIR_Process.comm_self->errhandler		 = NULL; /* XXX */
	MPIR_Process.comm_self->coll_fns		 = NULL; /* XXX */
	MPIR_Process.comm_self->topo_fns	 = NULL; /* XXX */
	MPIR_Process.comm_self->comm_net	 = MPI_COMM_NULL;
	MPIR_Process.comm_self->comm_node	= MPI_COMM_NULL;
	MPIR_Process.comm_self->sub_comms_initialized = 0;

#ifdef MPID_NEEDS_ICOMM_WORLD
    MPIR_Process.icomm_world		    = MPID_Comm_builtin + 2;
    MPIR_Process.icomm_world->handle	    = MPIR_ICOMM_WORLD;
    MPIU_Object_set_ref( MPIR_Process.icomm_world, 1 );
    MPIR_Process.icomm_world->context_id    = 2 << MPID_CONTEXT_PREFIX_SHIFT;
    MPIR_Process.icomm_world->recvcontext_id= 2 << MPID_CONTEXT_PREFIX_SHIFT;
    MPIR_Process.icomm_world->attributes    = NULL;
    MPIR_Process.icomm_world->local_group   = NULL;
    MPIR_Process.icomm_world->remote_group  = NULL;
    MPIR_Process.icomm_world->comm_kind	    = MPID_INTRACOMM;
    /* This initialization of the comm name could be done only when comm_get_name is called */
    MPIU_Strncpy(MPIR_Process.icomm_world->name, "MPI_ICOMM_WORLD", MPI_MAX_OBJECT_NAME);
    MPIR_Process.icomm_world->errhandler    = NULL; /* XXX */
    MPIR_Process.icomm_world->coll_fns	    = NULL; /* XXX */
    MPIR_Process.icomm_world->topo_fns	    = NULL; /* XXX */

    /* Note that these communicators are not ready for use - MPID_Init 
       will setup self and world, and icomm_world if it desires it. */
#endif

	MPIR_Process.comm_parent = NULL;

	/* FIXME: set all the fields for MPI_FILE_NULL */
	MPIR_Process.file_null				= MPID_File_builtin + 0;
	MPIR_Process.file_null->handle		= MPI_FILE_NULL;
	MPID_Errhandler_get_ptr( MPI_ERRORS_RETURN, MPIR_Process.file_null->errhandler);

	/* Setup the initial communicator list in case we have 
	   enabled the debugger message-queue interface */
	MPIR_COMML_REMEMBER( MPIR_Process.comm_world );
	MPIR_COMML_REMEMBER( MPIR_Process.comm_self );

	/* Call any and all MPID_Init type functions */
	MPIU_Init_call_history(argc, argv);
	/* FIXME: The call to err init should be within an ifdef
	   HAVE_ ERROR_CHECKING block (as must all uses of Err_create_code) */
	MPIR_Err_init();
	MPID_Wtime_init();
	MPIR_Datatype_init();

	MPIU_THREADPRIV_GET;

	MPIR_Nest_init();
	/* MPIU_Timer_pre_init(); */

	/* define MPI as initialized so that we can use MPI functions within 
	   MPID_Init if necessary */
	MPIR_Process.initialized = MPICH_WITHIN_MPI;

	/* For any code in the device that wants to check for runtime 
	   decisions on the value of isThreaded, set a provisional
	   value here. We could let the MPID_Init routine override this */
#ifdef HAVE_RUNTIME_THREADCHECK
	MPIR_Process.isThreaded = required == MPI_THREAD_MULTIPLE;
#endif
	mpi_errno = MPID_Init(argc, argv, required, &thread_provided, &has_args, &has_env);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, L"MPIR_Init_thread", __LINE__, MPI_ERR_OTHER, L"**init", 0);
		/* FIXME: the default behavior for all MPI routines is to abort.  
		   This isn't always convenient, because there's no other way to 
		   get this routine to simply return.  But we should provide some
		   sort of control for that and follow the default defined 
		   by the standard */
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */

	/* Capture the level of thread support provided */
	MPIR_Process.thread_provided = thread_provided;
	if (provided) *provided = thread_provided;
	/* FIXME: Rationalize this with the above */
#ifdef HAVE_RUNTIME_THREADCHECK
	MPIR_Process.isThreaded = required == MPI_THREAD_MULTIPLE;
	if (provided) *provided = required;
#if 0
	/* Preallocated MPIR_Thread if we're single-threaded */
	if (provided < MPI_THREAD_MULTIPLE) {
		MPIR_Thread = (MPICH_PerThread_t *) 
			MPIU_Calloc(1, sizeof(MPICH_PerThread_t));
	}
#endif
#endif

	/* FIXME: Define these in the interface.  Does Timer init belong here? */
	MPIU_dbg_init(MPIR_Process.comm_world->rank);
	MPIU_Timer_init(MPIR_Process.comm_world->rank,
					MPIR_Process.comm_world->local_size);
#ifdef USE_MEMORY_TRACING
	MPIU_trinit( MPIR_Process.comm_world->rank );
	/* Indicate that we are near the end of the init step; memory 
	   allocated already will have an id of zero; this helps 
	   separate memory leaks in the initialization code from 
	   leaks in the L"active" code */
	/* Uncomment this code to leave out any of the MPID_Init/etc 
	   memory allocations from the memory leak testing */
	/* MPIU_trid( 1 ); */
#endif
#ifdef USE_DBG_LOGGING
	MPIU_DBG_Init( argc, argv, has_args, has_env, MPIR_Process.comm_world->rank );
#endif

	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
		MPIR_Process.initialized = MPICH_PRE_INIT;
	/* --END ERROR HANDLING-- */

	if (mpi_errno == MPI_SUCCESS && MPIR_Process.comm_world->local_size > 1)
	{
		int color, key, node_root_rank, size/*, in_order*/;
		int use_subcomms = FALSE; /* Default to NOT use sub-communicators */

		MPIU_GetEnvBool(L"DeinoMPI_USE_SUB_COMMUNICATORS", &use_subcomms);
		MPIU_GetEnvBool(L"DeinoMPI_USE_SMP_OPTIMIZATIONS", &use_subcomms);

		/* First set all the defaults if use_subcomms has been turned on */
		if (use_subcomms)
		{
			MPIR_Process.sub_bcast = TRUE;
			MPIR_Process.sub_barrier = TRUE;
			MPIR_Process.sub_allreduce = TRUE;
			MPIR_Process.sub_gather = TRUE;
		}

		/* Control each function individually */
		MPIU_GetEnvBool(L"DeinoMPI_SUB_BCAST", &MPIR_Process.sub_bcast);
		MPIU_GetEnvBool(L"DeinoMPI_SMP_BCAST", &MPIR_Process.sub_bcast);

		MPIU_GetEnvBool(L"DeinoMPI_SUB_BARRIER", &MPIR_Process.sub_barrier);
		MPIU_GetEnvBool(L"DeinoMPI_SMP_BARRIER", &MPIR_Process.sub_barrier);

		MPIU_GetEnvBool(L"DeinoMPI_SUB_ALLREDUCE", &MPIR_Process.sub_allreduce);
		MPIU_GetEnvBool(L"DeinoMPI_SMP_ALLREDUCE", &MPIR_Process.sub_allreduce);

		MPIU_GetEnvBool(L"DeinoMPI_SUB_GATHER", &MPIR_Process.sub_gather);
		MPIU_GetEnvBool(L"DeinoMPI_SMP_GATHER", &MPIR_Process.sub_gather);

		if (MPIR_Process.sub_bcast		== TRUE || 
			MPIR_Process.sub_barrier	== TRUE ||
			MPIR_Process.sub_allreduce	== TRUE ||
			MPIR_Process.sub_gather		== TRUE)
		{
			use_subcomms = TRUE;
		}

		if (use_subcomms == TRUE)
		{
			mpi_errno = getNodeRootRank(&node_root_rank);
			/*printf("[%d] node root = %d\n", MPIR_Process.comm_world->rank, node_root_rank);fflush(stdout);*/
			MPIU_ERR_CHK_POP(mpi_errno);

			/* create a comm_node communicator consisting of the processes on each node only */
			color = node_root_rank;
			key = MPIR_Process.comm_world->rank; /* maintain the same relative rank ordering */
			/*printf("[%d] comm_node with color %d and key %d\n", MPIR_Process.comm_world->rank, color, key);fflush(stdout);*/
			NMPI_Comm_split(mpi_errno, (MPI_COMM_WORLD, color, key, &MPIR_Process.comm_world->comm_node));
			MPIU_ERR_CHK_POP(mpi_errno);
			/*printf("[%d] comm_node created\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
			NMPI_Comm_size(mpi_errno, (MPIR_Process.comm_world->comm_node, &size));
			MPIU_ERR_CHK_POP(mpi_errno);
			if (size == 1)
			{
				/* A node communicator is not needed if there is only one process on the node. */
				NMPI_Comm_free(mpi_errno, (&MPIR_Process.comm_world->comm_node));
				/*printf("[%d] comm_node freed\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
				MPIU_ERR_CHK_POP(mpi_errno);
			}
			/* Save the root node rank.  This can be used by the collective operations when a task is broken up into comm_net
			 * and comm_node steps but the data doesn't reside on a node that is a member of comm_net.  The rank with the data
			 * can send it to this rank first ensuring that a member of comm_net has the data before the two step operation begins.
			 */
			/*MPIR_Process.comm_world->comm_node_root_rank = node_root_rank;*/
			/* Save the root node ranks for all processes.  Knowing your own root allows you to send a message but in order for
			 * the root ranks to know who to receive from, a mapping for all processes must be known.
			 */
			MPIR_Process.comm_world->rank_to_comm_node_root_map = (int*)MPIU_Malloc(MPIR_Process.comm_world->local_size * sizeof(int));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, MPIR_Process.comm_world->rank_to_comm_node_root_map, MPIR_Process.comm_world->local_size * sizeof(int), L"rank_to_comm_node_root_map");
			/* Save my root rank in my rank slot */
			MPIR_Process.comm_world->rank_to_comm_node_root_map[MPIR_Process.comm_world->rank] = node_root_rank;
			/* Allgather the root ranks from all other ranks */
			/*NMPI_Allgather(mpi_errno, (MPI_IN_PLACE, 0, MPI_INT, MPIR_Process.comm_world->rank_to_comm_node_root_map, 1, MPI_INT, MPI_COMM_WORLD));*/
			mpi_errno = MPIR_Allgather(MPI_IN_PLACE, 0, MPI_INT, MPIR_Process.comm_world->rank_to_comm_node_root_map, 1, MPI_INT, MPIR_Process.comm_world);
			MPIU_ERR_CHK_POP(mpi_errno);
			/* Debugging output
			{
				char result[1000] = "";
				char tmp[10];
				int i;
				for (i=0; i<MPIR_Process.comm_world->local_size; i++)
				{
					sprintf(tmp, "%d,", MPIR_Process.comm_world->rank_to_comm_node_root_map[i]);
					strcat(result, tmp);
				}
				printf("[%d] <%s> rank_to_comm_node_root_map\n", MPIR_Process.comm_world->rank, result);fflush(stdout);
			}
			*/
#if 0
			/* Check to see if the ranks are mapped to nodes in order like this: (0 1) (2 3) (4 5) */
			/* If the mapping is round robin like this: (0 2) (1 3) then some optimizations won't work */
			{
				int i;
				in_order = 1;
				for (i=0; i<MPIR_Process.comm_world->local_size-1; i++)
				{
					if (MPIR_Process.comm_world->rank_to_comm_node_root_map[i] > MPIR_Process.comm_world->rank_to_comm_node_root_map[i+1])
					{
						in_order = 0;
						break;
					}
				}
				if (!in_order)
				{
					/* Remove optimizations that rely on ranks on each node being in order */
					MPIR_Process.sub_gather = FALSE;
				}
			}
#endif

			/* Save a mapping from ranks to comm_net ranks for the root of the comm_node that each rank belongs to */
			MPIR_Process.comm_world->root_to_comm_net_rank_map = (int*)MPIU_Malloc(MPIR_Process.comm_world->local_size * sizeof(int));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, MPIR_Process.comm_world->root_to_comm_net_rank_map, MPIR_Process.comm_world->local_size * sizeof(int), L"root_to_comm_net_rank_map");
			{
				int i, j;
				int *net_rank_to_comm_rank = (int*)MPIU_Malloc(MPIR_Process.comm_world->local_size * sizeof(int));
				MPIU_ERR_MEM_CHK_POP(mpi_errno, net_rank_to_comm_rank, MPIR_Process.comm_world->local_size * sizeof(int), L"net_rank_to_comm_rank");
				for (i=0; i<MPIR_Process.comm_world->local_size; i++)
					net_rank_to_comm_rank[i] = -1;
				for (i=0; i<MPIR_Process.comm_world->local_size; i++)
				{
					for (j=0; j<MPIR_Process.comm_world->local_size; j++)
					{
						if (net_rank_to_comm_rank[j] == -1)
						{
							net_rank_to_comm_rank[j] = MPIR_Process.comm_world->rank_to_comm_node_root_map[i];
							MPIR_Process.comm_world->root_to_comm_net_rank_map[i] = j;
							break;
						}
						if (MPIR_Process.comm_world->rank_to_comm_node_root_map[i] == net_rank_to_comm_rank[j])
						{
							MPIR_Process.comm_world->root_to_comm_net_rank_map[i] = j;
							break;
						}
					}
				}
				MPIU_Free(net_rank_to_comm_rank);
			}
			/* Debugging output
			{
				char result[1000] = "";
				char tmp[10];
				int i;
				for (i=0; i<MPIR_Process.comm_world->local_size; i++)
				{
					sprintf(tmp, "%d,", MPIR_Process.comm_world->root_to_comm_net_rank_map[i]);
					strcat(result, tmp);
				}
				printf("[%d] <%s> root_to_comm_net_rank_map\n", MPIR_Process.comm_world->rank, result);fflush(stdout);
			}
			*/

			/* Create a comm_net communicator consisting of the root processes on each node only. */
			/* The rest of the processes end up with MPI_COMM_NULL. */
			color = node_root_rank == MPIR_Process.comm_world->rank ? 0 : MPI_UNDEFINED;
			key = MPIR_Process.comm_world->rank; /* maintain the same relative rank ordering */
			/*printf("[%d] comm_net with color %d and key %d\n", MPIR_Process.comm_world->rank, color, key);fflush(stdout);*/
			NMPI_Comm_split(mpi_errno, (MPI_COMM_WORLD, color, key, &MPIR_Process.comm_world->comm_net));
			MPIU_ERR_CHK_POP(mpi_errno);
			/*printf("[%d] comm_net created\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
			size = 0;
			if (MPIR_Process.comm_world->comm_net != MPI_COMM_NULL)
			{
				NMPI_Comm_size(mpi_errno, (MPIR_Process.comm_world->comm_net, &size));
				MPIU_ERR_CHK_POP(mpi_errno);
				if (size == 1 || size == MPIR_Process.comm_world->local_size)
				{
					/* A net communicator is not needed if:
					 * 1) all processes are on a single node (size == 1)
					 * 2) there is only one process per node (size == comm_world.size) */
					NMPI_Comm_free(mpi_errno, (&MPIR_Process.comm_world->comm_net));
					/*printf("[%d] comm_net freed\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
					MPIU_ERR_CHK_POP(mpi_errno);
				}
				else
				{
					int net_size, net_rank, node_size;
					/*MPID_Comm *comm_net_ptr;*/

					NMPI_Comm_size(mpi_errno, (MPIR_Process.comm_world->comm_net, &net_size));
					MPIU_ERR_CHK_POP(mpi_errno);
					NMPI_Comm_rank(mpi_errno, (MPIR_Process.comm_world->comm_net, &net_rank));
					MPIU_ERR_CHK_POP(mpi_errno);
					if (MPIR_Process.comm_world->comm_node != MPI_COMM_NULL)
					{
						NMPI_Comm_size(mpi_errno, (MPIR_Process.comm_world->comm_node, &node_size));
						MPIU_ERR_CHK_POP(mpi_errno);
					}
					else
					{
						node_size = 1;
					}
					MPIR_Process.comm_world->comm_node_sizes = (int*)MPIU_Malloc(net_size * sizeof(int));
					MPIU_ERR_MEM_CHK_POP(mpi_errno, MPIR_Process.comm_world->comm_node_sizes, net_size*sizeof(int), "comm_node_sizes");
					/*
					MPIR_Process.comm_world->comm_node_displacements = (int*)MPIU_Malloc(net_size * sizeof(int));
					MPIU_ERR_MEM_CHK_POP(mpi_errno, MPIR_Process.comm_world->comm_node_displacements, net_size*sizeof(int), "comm_node_displacements");
					*/
					MPIR_Process.comm_world->comm_node_sizes[net_rank] = node_size;

					/* call the full allgather function */
					NMPI_Allgather(mpi_errno, (MPI_IN_PLACE, 0, MPI_INT, MPIR_Process.comm_world->comm_node_sizes, 1, MPI_INT, MPIR_Process.comm_world->comm_net));
					MPIU_ERR_CHK_POP(mpi_errno);
					/* or call the MPIR version
					MPID_Comm_get_ptr(MPIR_Process.comm_world->comm_net, comm_net_ptr);
					mpi_errno = MPIR_Allgather(MPI_IN_PLACE, 0, MPI_INT, MPIR_Process.comm_world->comm_node_sizes, 1, MPI_INT, comm_net_ptr);
					MPIU_ERR_CHK_POP(mpi_errno);
					*/
					/* calculate the displacements from the comm_node sizes */
					/*
					{
						int i, cur_displacement = 0;
						for (i=0; i<net_size; i++)
						{
							MPIR_Process.comm_world->comm_node_displacements[i] = cur_displacement;
							cur_displacement += MPIR_Process.comm_world->comm_node_sizes[i];
						}
					}
					*/
					/* Debugging output
					{
						char buffer[1024] = "comm_net [";
						char number[20];
						int i;
						for (i=0; i<net_size; i++)
						{
							sprintf(number, "%d, ", MPIR_Process.comm_world->comm_node_sizes[i]);
							strcat(buffer, number);
						}
						printf("%s]\n", buffer);
						fflush(stdout);
					}
					*/
				}
			}
			if (MPIR_Process.comm_world->comm_node != MPI_COMM_NULL)
			{
				/* broadcast the size of comm_net to the comm_node members */
				NMPI_Bcast(mpi_errno, (&size, 1, MPI_INT, 0, MPIR_Process.comm_world->comm_node));
				MPIU_ERR_CHK_POP(mpi_errno);
				/* If the size of comm_net is 1 then there is only one node so there is no need for comm_node */
				if (size == 1)
				{
					NMPI_Comm_free(mpi_errno, (&MPIR_Process.comm_world->comm_node));
					MPIU_ERR_CHK_POP(mpi_errno);
				}
			}

			MPIR_Process.comm_world->node_rank_to_comm_rank = NULL;
			if (MPIR_Process.comm_world->comm_net != MPI_COMM_NULL)
			{
				int *offsets, i, j, cur_offset = 0;
				/* allocate the array */
				MPIR_Process.comm_world->node_rank_to_comm_rank = (int*)MPIU_Malloc(MPIR_Process.comm_world->local_size * sizeof(int));
				MPIU_ERR_MEM_CHK_POP(mpi_errno, MPIR_Process.comm_world->node_rank_to_comm_rank, MPIR_Process.comm_world->local_size * sizeof(int), L"node_rank_to_comm_rank");
				MPIR_Process.comm_world->node_block_reordering = (int*)MPIU_Malloc(MPIR_Process.comm_world->local_size * sizeof(int));
				MPIU_ERR_MEM_CHK_POP(mpi_errno, MPIR_Process.comm_world->node_block_reordering, MPIR_Process.comm_world->local_size * sizeof(int), L"node_block_reordering");
				NMPI_Comm_size(mpi_errno, (MPIR_Process.comm_world->comm_net, &size));
				offsets = (int*)MPIU_Malloc(size * sizeof(int));
				MPIU_ERR_MEM_CHK_POP(mpi_errno, offsets, size*sizeof(int), L"offsets");
				for (i=0; i<size; i++)
				{
					offsets[i] = cur_offset;
					cur_offset += MPIR_Process.comm_world->comm_node_sizes[i];
				}
				/* gather all the comm_node rank maps to the comm_net root */
				NMPI_Gatherv(mpi_errno, (my_clique_ranks, my_clique_size, MPI_INT, MPIR_Process.comm_world->node_rank_to_comm_rank, MPIR_Process.comm_world->comm_node_sizes, offsets, MPI_INT, 0, MPIR_Process.comm_world->comm_net));
				MPIU_ERR_CHK_POP(mpi_errno);
				/* broadcast the map to all the processes */
				NMPI_Bcast(mpi_errno, (MPIR_Process.comm_world->node_rank_to_comm_rank, MPIR_Process.comm_world->local_size, MPI_INT, 0, MPI_COMM_WORLD));
				MPIU_ERR_CHK_POP(mpi_errno);
				for (i=0; i<MPIR_Process.comm_world->local_size; i++)
				{
					for (j=0; j<MPIR_Process.comm_world->local_size; j++)
					{
						if (MPIR_Process.comm_world->node_rank_to_comm_rank[j] == i)
						{
							MPIR_Process.comm_world->node_block_reordering[i] = j;
							break;
						}
					}
				}
			}
			else if (MPIR_Process.comm_world->comm_node != MPI_COMM_NULL)
			{
				int i, j;
				/* allocate the array */
				MPIR_Process.comm_world->node_rank_to_comm_rank = (int*)MPIU_Malloc(MPIR_Process.comm_world->local_size * sizeof(int));
				MPIU_ERR_MEM_CHK_POP(mpi_errno, MPIR_Process.comm_world->node_rank_to_comm_rank, MPIR_Process.comm_world->local_size * sizeof(int), L"node_rank_to_comm_rank");
				MPIR_Process.comm_world->node_block_reordering = (int*)MPIU_Malloc(MPIR_Process.comm_world->local_size * sizeof(int));
				MPIU_ERR_MEM_CHK_POP(mpi_errno, MPIR_Process.comm_world->node_block_reordering, MPIR_Process.comm_world->local_size * sizeof(int), L"node_block_reordering");
				/* broadcast the map to all the processes */
				NMPI_Bcast(mpi_errno, (MPIR_Process.comm_world->node_rank_to_comm_rank, MPIR_Process.comm_world->local_size, MPI_INT, 0, MPI_COMM_WORLD));
				MPIU_ERR_CHK_POP(mpi_errno);
				for (i=0; i<MPIR_Process.comm_world->local_size; i++)
				{
					for (j=0; j<MPIR_Process.comm_world->local_size; j++)
					{
						if (MPIR_Process.comm_world->node_rank_to_comm_rank[j] == i)
						{
							MPIR_Process.comm_world->node_block_reordering[i] = j;
							break;
						}
					}
				}
			}
			/* Debugging output
			if (MPIR_Process.comm_world->node_rank_to_comm_rank != NULL)
			{
				{
					char result[1000] = "";
					char tmp[10];
					int i;
					for (i=0; i<MPIR_Process.comm_world->local_size; i++)
					{
						sprintf(tmp, "%d,", MPIR_Process.comm_world->node_rank_to_comm_rank[i]);
						strcat(result, tmp);
					}
					printf("[%d] <%s> node_rank_to_comm_rank\n", MPIR_Process.comm_world->rank, result);fflush(stdout);
				}
			}
			if (MPIR_Process.comm_world->node_block_reordering != NULL)
			{
				{
					char result[1000] = "";
					char tmp[10];
					int i;
					for (i=0; i<MPIR_Process.comm_world->local_size; i++)
					{
						sprintf(tmp, "%d,", MPIR_Process.comm_world->node_block_reordering[i]);
						strcat(result, tmp);
					}
					printf("[%d] <%s> node_block_reordering\n", MPIR_Process.comm_world->rank, result);fflush(stdout);
				}
			}
			*/

			/*
			if ((MPIR_Process.comm_world->comm_net != MPI_COMM_NULL) || (MPIR_Process.comm_world->comm_node != MPI_COMM_NULL))
			{
				int net_rank = -1, net_size = -1;
				int node_rank = -1, node_size = -1;
				if (MPIR_Process.comm_world->comm_net != MPI_COMM_NULL)
				{
					NMPI_Comm_rank(mpi_errno, (MPIR_Process.comm_world->comm_net, &net_rank));
					NMPI_Comm_size(mpi_errno, (MPIR_Process.comm_world->comm_net, &net_size));
				}
				if (MPIR_Process.comm_world->comm_node != MPI_COMM_NULL)
				{
					NMPI_Comm_rank(mpi_errno, (MPIR_Process.comm_world->comm_node, &node_rank));
					NMPI_Comm_size(mpi_errno, (MPIR_Process.comm_world->comm_node, &node_size));
				}
				printf("[%d] NET [%2d:%2d] NODE [%2d:%2d]\n", MPIR_Process.comm_world->rank,
					net_rank, net_size,
					node_rank, node_size);fflush(stdout);
			}
			*/

			MPIR_Process.comm_world->sub_comms_initialized = 1;
		}
	}

#ifdef HAVE_DEBUGGER_SUPPORT
	MPIR_WaitForDebugger();
#endif

fn_exit:
	return mpi_errno;
fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Init_thread

/*@
   MPI_Init_thread - Initialize the MPI execution environment

   Input Parameters:
+  argc - Pointer to the number of arguments 
.  argv - Pointer to the argument vector
-  required - Level of desired thread support

   Output Parameter:
.  provided - Level of provided thread support

   Command line arguments:
   MPI specifies no command-line arguments but does allow an MPI 
   implementation to make use of them.  See 'MPI_INIT' for a description of 
   the command line arguments supported by 'MPI_INIT' and 'MPI_INIT_THREAD'.

   Notes:
   The valid values for the level of thread support are\:
+ MPI_THREAD_SINGLE - Only one thread will execute. 
. MPI_THREAD_FUNNELED - The process may be multi-threaded, but only the main 
  thread will make MPI calls (all MPI calls are funneled to the 
   main thread). 
. MPI_THREAD_SERIALIZED - The process may be multi-threaded, and multiple 
  threads may make MPI calls, but only one at a time: MPI calls are not 
  made concurrently from two distinct threads (all MPI calls are serialized). 
- MPI_THREAD_MULTIPLE - Multiple threads may call MPI, with no restrictions. 

Notes for Fortran:
   Note that the Fortran binding for this routine does not have the 'argc' and
   'argv' arguments. ('MPI_INIT_THREAD(required, provided, ierror)')


.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER

.seealso: MPI_Init, MPI_Finalize
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Init_threadA( int *argc, char ***argv, int required, int *provided )
{
	static const wchar_t FCNAME[] = L"MPI_Init_threadA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t **wargv = NULL;
	int i;
	MPID_MPI_INIT_STATE_DECL(MPID_STATE_MPI_INIT_THREADA);

	MPID_MPI_INIT_FUNC_ENTER(MPID_STATE_MPI_INIT_THREADA);

	if (argv != NULL && argc != NULL)
	{
		wargv = (wchar_t**)MPIU_Malloc(((*argc) + 1) * sizeof(wchar_t*));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, wargv, ((*argc) + 1)*sizeof(wchar_t), "wargv");
		for (i=0; i<((*argc) + 1); i++) wargv[i] = NULL;
		for (i=0; i<((*argc) + 1); i++)
		{
			if ((*argv)[i] != NULL)
			{
				mpi_errno = MPIU_Mbstowcs((*argv)[i], &wargv[i]);
				MPIU_ERR_CHK_POP(mpi_errno);
			}
		}
	}

	mpi_errno = MPI_Init_threadW(argc, &wargv, required, provided);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wargv != NULL)
	{
		for (i=0; i<((*argc) + 1); i++)
		{
			if (wargv[i] != NULL)
			{
				MPIU_Free(wargv[i]);
			}
		}
		MPIU_Free(wargv);
	}
	MPID_MPI_INIT_FUNC_EXIT(MPID_STATE_MPI_INIT_THREADA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Init_threadW( int *argc, wchar_t ***argv, int required, int *provided )
{
	static const wchar_t FCNAME[] = L"MPI_Init_thread";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_INIT_STATE_DECL(MPID_STATE_MPI_INIT_THREAD);

	MPID_CS_INITIALIZE();
	MPIU_THREAD_SINGLE_CS_ENTER("init");

	/* Create the thread-private region if necessary and go ahead 
	   and initialize it */
	MPIU_THREADPRIV_INITKEY;
	MPIU_THREADPRIV_INIT;

	MPID_MPI_INIT_FUNC_ENTER(MPID_STATE_MPI_INIT_THREAD);
	
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			if (MPIR_Process.initialized != MPICH_PRE_INIT) {
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, L"MPI_Init_thread", __LINE__, MPI_ERR_OTHER,
												  L"**inittwice", 0 );
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	MPIU_Init_call_history(argc, argv);
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_init_thread %p %p %d %p", argc, argv, required, provided); }
	
	mpi_errno = MPIR_Init_thread( argc, argv, required, provided );
	if (mpi_errno != MPI_SUCCESS) goto fn_fail; 

	/* ... end of body of routine ... */
	
	MPID_MPI_INIT_FUNC_EXIT(MPID_STATE_MPI_INIT_THREAD);
	MPIU_THREAD_SINGLE_CS_EXIT("init");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_HANDLING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_init_thread",
			L"**mpi_init_thread %p %p %d %p", argc, argv, required, provided);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	MPID_MPI_INIT_FUNC_EXIT(MPID_STATE_MPI_INIT_THREAD);
	MPIU_THREAD_SINGLE_CS_EXIT("init");
	MPID_CS_FINALIZE();
	return mpi_errno;
	/* --END ERROR HANDLING-- */
}
