/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS

#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_File_delete = PMPI_File_delete
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_File_delete MPI_File_delete
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_File_delete as PMPI_File_delete
/* end of weak pragmas */
#endif

#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_deleteA PMPI_File_deleteA
#define MPI_File_deleteW PMPI_File_deleteW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_delete

extern int ADIO_Init_keyval;

/*@
	MPI_File_delete - Deletes a file

Input Parameters:
. filename - name of file to delete (string)
. info - info object (handle)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_deleteA(char *filename, MPI_Info info)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_deleteA";
	wchar_t *wfilename = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_DELETEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_DELETEA);

	if (filename != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(filename, &wfilename);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = MPI_File_deleteW(wfilename, info);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wfilename != NULL)
		MPIU_Free(wfilename);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_DELETEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_File_deleteW(wchar_t *filename, MPI_Info info)
{
	int mpi_errno = MPI_SUCCESS, file_system;
	wchar_t *tmp;
	ADIOI_Fns *fsops;
	static const wchar_t FCNAME[] = L"MPI_File_delete";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_DELETE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_UNREFERENCED_ARG(info);

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_DELETE);
	MPIU_THREADPRIV_GET;

	/* Validate parameters and objects */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO_OR_NULL(info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_delete %s %I", filename, info); }

	/* first check if ADIO has been initialized. If not, initialize it */
	if (ADIO_Init_keyval == MPI_KEYVAL_INVALID)
	{
		NMPI_Keyval_create(mpi_errno, (MPI_NULL_COPY_FN, ADIOI_End_call, &ADIO_Init_keyval, (void *) 0));
		MPIU_ERR_CHK_POP(mpi_errno);

		/* put a dummy attribute on MPI_COMM_WORLD, because we want the delete
		   function to be called when MPI_COMM_WORLD is freed. Hopefully the
		   MPI library frees MPI_COMM_WORLD when MPI_Finalize is called,
		   though the standard does not mandate this. */

		NMPI_Attr_put(mpi_errno, (MPI_COMM_WORLD, ADIO_Init_keyval, (void *) 0));
		MPIU_ERR_CHK_POP(mpi_errno);

		/* initialize ADIO */
		mpi_errno = ADIO_Init( (int *)0, (wchar_t ***)0);
		MPIU_ERR_CHK_POP(mpi_errno);
	}


	/* resolve file system type from file name; this is a collective call */
	mpi_errno = ADIO_ResolveFileType(MPI_COMM_SELF, filename, &file_system, &fsops);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* skip prefixes on file names if they have more than one character;
	 * single-character prefixes are assumed to be windows drive
	 * specifications (e.g. c:\foo) and are left alone.
	 */
	tmp = wcschr(filename, L':');
	if (tmp > filename + 1)
		filename = tmp + 1;

	/* call the fs-specific delete function */
	mpi_errno = (fsops->ADIOI_xxx_Delete)(filename);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_DELETE);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_delete",
			L"**mpi_file_delete %s %I", filename, info);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(NULL, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
