/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_set_viewA PMPI_File_set_viewA
#define MPI_File_set_viewW PMPI_File_set_viewW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_set_view

/*@
	MPI_File_set_view - Sets the file view

Input Parameters:
. fh - file handle (handle)
. disp - displacement (nonnegative integer)
. etype - elementary datatype (handle)
. filetype - filetype (handle)
. datarep - data representation (string)
. info - info object (handle)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_set_viewA(MPI_File file, MPI_Offset disp, MPI_Datatype etype, MPI_Datatype filetype, char *datarep, MPI_Info info)
{
	static const wchar_t FCNAME[] = L"MPI_File_set_viewA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *tmp_datarep;
	size_t len, result;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SET_VIEWA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SET_VIEWA);

	if (datarep != NULL)
	{
		len = strlen(datarep) + 1;
		tmp_datarep = MPIU_Malloc(len * sizeof(wchar_t));
		error = mbstowcs_s(&result, tmp_datarep, len, datarep, _TRUNCATE);
		if (error != 0)
		{
			if (result < 0 || result >= len)
			{
				/* Invalid datarep string */
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, L"MPI_Unpack_externalA", __LINE__, MPI_ERR_INTERN, L"**fail", L"**argrange %s %d %d", L"num_chars_converted", (int)result, (int)len);
			}
			else
			{
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			}
			goto fn_fail;
		}
		if (result > 0 && result <= len)
		{
			mpi_errno = MPI_File_set_viewW(file, disp, etype, filetype, tmp_datarep, info);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			/* Invalid datarep string */
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, L"MPI_Unpack_externalA", __LINE__, MPI_ERR_INTERN, L"**fail", L"**argrange %s %d %d", L"num_chars_converted", (int)result, (int)len);
		}
	}
	else
	{
		/* Invalid datarep string */
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, L"MPI_Unpack_externalA", __LINE__, MPI_ERR_INTERN, L"**fail", 0);
		MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
	}

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SET_VIEWA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_File_set_viewW(MPI_File file, MPI_Offset disp, MPI_Datatype etype, MPI_Datatype filetype, wchar_t *datarep, MPI_Info info)
{
	int filetype_size, etype_size, mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_set_view";
	MPI_Offset shared_fp, byte_off;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SET_VIEW);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SET_VIEW);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if ((disp < 0) && (disp != MPI_DISPLACEMENT_CURRENT))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iobaddisp", 0);
		goto fn_fail;
	}

	/* rudimentary checks for incorrect etype/filetype.*/
	if (etype == MPI_DATATYPE_NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**ioetype", 0);
		goto fn_fail;
	}

	if (filetype == MPI_DATATYPE_NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iofiletype", 0);
		goto fn_fail;
	}

	if ((file_ptr->access_mode & MPI_MODE_SEQUENTIAL) &&
		(disp != MPI_DISPLACEMENT_CURRENT))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iodispifseq", 0);
		goto fn_fail;
	}

	if ((disp == MPI_DISPLACEMENT_CURRENT) &&
		!(file_ptr->access_mode & MPI_MODE_SEQUENTIAL))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iodispifseq", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	NMPI_Type_size(mpi_errno, (filetype, &filetype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_size(mpi_errno, (etype, &etype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	if (filetype_size % etype_size != 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iofiletype", 0);
		goto fn_fail;
	}

	/*if (wcscmp(datarep, L"native") && wcscmp(datarep, L"NATIVE"))*/
	if (_wcsicmp(datarep, L"native"))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_UNSUPPORTED_DATAREP, L"**unsupporteddatarep",0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_set_view %F %d %D %D %s %I", file, disp, etype, filetype, datarep, info); }

	if (disp == MPI_DISPLACEMENT_CURRENT)
	{
		NMPI_Barrier(mpi_errno, (file_ptr->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIO_Get_shared_fp(file_ptr, 0, &shared_fp);
		MPIU_ERR_CHK_POP(mpi_errno);

		NMPI_Barrier(mpi_errno, (file_ptr->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Get_byte_offset(file_ptr, shared_fp, &byte_off);
		MPIU_ERR_CHK_POP(mpi_errno);

		disp = byte_off;
	}

	mpi_errno = ADIO_Set_view(file_ptr, disp, etype, filetype, info);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* reset shared file pointer to zero */
	if ((file_ptr->file_system != ADIO_PIOFS) &&
		(file_ptr->file_system != ADIO_PVFS) &&
		(file_ptr->file_system != ADIO_PVFS2) && 
		(file_ptr->shared_fp_fd != NULL))
	{
		/* only one process needs to set it to zero, but I don't want to 
		   create the shared-file-pointer file if shared file pointers have 
		   not been used so far. Therefore, every process that has already 
		   opened the shared-file-pointer file sets the shared file pointer 
		   to zero. If the file was not opened, the value is automatically 
		   zero. Note that shared file pointer is stored as no. of etypes
		   relative to the current view, whereas indiv. file pointer is
		   stored in bytes. */

		mpi_errno = ADIO_Set_shared_fp(file_ptr, 0);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	if ((file_ptr->file_system != ADIO_PIOFS) &&
		(file_ptr->file_system != ADIO_PVFS) &&
		(file_ptr->file_system != ADIO_PVFS2 ))
	{
		NMPI_Barrier(mpi_errno, (file_ptr->comm)); /* for above to work correctly */
		MPIU_ERR_CHK_POP(mpi_errno);
	}

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SET_VIEW);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_set_view",
			L"**mpi_file_set_view %F %d %D %D %s %I", file, disp, etype, filetype, datarep, info);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_set_size PMPI_File_set_size
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_set_size

/*@
	MPI_File_set_size - Sets the file size

Input Parameters:
. fh - file handle (handle)
. size - size to truncate or expand file (nonnegative integer)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_set_size(MPI_File file, MPI_Offset size)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_set_size";
	MPI_Offset tmp_sz;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SET_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SET_SIZE);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (size < 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iobadsize", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	tmp_sz = size;
	NMPI_Bcast(mpi_errno, (&tmp_sz, 1, ADIO_OFFSET, 0, file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	if (tmp_sz != size)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**notsame", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_set_size %F %d", file, size); }

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	mpi_errno = ADIO_Resize(file_ptr, size);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SET_SIZE);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_set_size",
			L"**mpi_file_set_size %F %d", file, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_set_info PMPI_File_set_info
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_set_info

/*@
	MPI_File_set_info - Sets new values for the hints associated with a file

Input Parameters:
. fh - file handle (handle)
. info - info object (handle)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_set_info(MPI_File file, MPI_Info info)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_set_info";
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SET_INFO);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SET_INFO);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_set_info %F %I", file, info); }

	/* set new info */
	mpi_errno = ADIO_SetInfo(file_ptr, info);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SET_INFO);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_set_info",
			L"**mpi_file_set_info %F %I", file, info);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_set_atomicity PMPI_File_set_atomicity
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_set_atomicity

/*@
	MPI_File_set_atomicity - Sets the atomicity mode

Input Parameters:
. fh - file handle (handle)
. flag - true to set atomic mode, false to set nonatomic mode (logical)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_set_atomicity(MPI_File file, int flag)
{
	int mpi_errno = MPI_SUCCESS, tmp_flag;
	static const wchar_t FCNAME[] = L"MPI_File_set_atomicity";
	ADIO_Fcntl_t *fcntl_struct;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SET_ATOMICITY);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SET_ATOMICITY);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_set_atomicity %F %d", file, flag); }

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	if (flag) flag = 1;  /* take care of non-one values! */

/* check if flag is the same on all processes */
	tmp_flag = flag;
	NMPI_Bcast(mpi_errno, (&tmp_flag, 1, MPI_INT, 0, file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	if (tmp_flag != flag) {
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**notsame", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	if (file_ptr->atomicity == flag)
	{
		goto fn_exit;
	}

	fcntl_struct = (ADIO_Fcntl_t *) MPIU_Malloc(sizeof(ADIO_Fcntl_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, fcntl_struct, sizeof(ADIO_Fcntl_t), L"fcntl_struct");
	fcntl_struct->atomicity = flag;
	mpi_errno = ADIO_Fcntl(file_ptr, ADIO_FCNTL_SET_ATOMICITY, fcntl_struct);
	MPIU_Free(fcntl_struct);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SET_ATOMICITY);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_set_atomicity",
			L"**mpi_file_set_atomicity %F %d", file, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_viewA PMPI_File_get_viewA
#define MPI_File_get_viewW PMPI_File_get_viewW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_view

/*@
	MPI_File_get_view - Returns the file view

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. disp - displacement (nonnegative integer)
. etype - elementary datatype (handle)
. filetype - filetype (handle)
. datarep - data representation (string)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_viewA(MPI_File file, MPI_Offset *disp, MPI_Datatype *etype, MPI_Datatype *filetype, char *datarep)
{
	static const wchar_t FCNAME[] = L"MPI_File_get_viewA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t wdatarep[MPI_MAX_DATAREP_STRING] = L"";
	wchar_t *wdatarep_ptr = NULL;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_VIEWA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_VIEWA);

	if (datarep != NULL)
		wdatarep_ptr = wdatarep;

	mpi_errno = MPI_File_get_viewW(file, disp, etype, filetype, wdatarep_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (datarep != NULL)
	{
		datarep[0] = '\0';
		error = wcstombs_s(&num_converted, datarep, MPI_MAX_DATAREP_STRING, wdatarep, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= MPI_MAX_DATAREP_STRING)
			{
				/* truncation error */
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"datarep", (int)num_converted);
				goto fn_fail;
			}
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
	}

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_VIEWA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_File_get_viewW(MPI_File file, MPI_Offset *disp, MPI_Datatype *etype, MPI_Datatype *filetype, wchar_t *datarep)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_get_view";
	int i, j, k, combiner;
	MPI_Datatype copy_etype, copy_filetype;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_VIEW);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_VIEW);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(disp, L"disp", mpi_errno);
			MPIR_ERRTEST_ARGNULL(etype, L"etype", mpi_errno);
			MPIR_ERRTEST_ARGNULL(filetype, L"filetype", mpi_errno);
			MPIR_ERRTEST_ARGNULL(datarep, L"datarep", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_view %F %p %p %p %s", file, disp, etype, filetype, datarep); }

	*disp = file_ptr->disp;
	MPIU_Wcsncpy(datarep, L"native", MPI_MAX_DATAREP_STRING);

	NMPI_Type_get_envelope(mpi_errno, (file_ptr->etype, &i, &j, &k, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (combiner == MPI_COMBINER_NAMED)
	{
		*etype = file_ptr->etype;
	}
	else
	{
		NMPI_Type_contiguous(mpi_errno, (1, file_ptr->etype, &copy_etype));
		MPIU_ERR_CHK_POP(mpi_errno);

		NMPI_Type_commit(mpi_errno, (&copy_etype));
		MPIU_ERR_CHK_POP(mpi_errno);
		*etype = copy_etype;
	}

	NMPI_Type_get_envelope(mpi_errno, (file_ptr->filetype, &i, &j, &k, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	
	if (combiner == MPI_COMBINER_NAMED)
	{
		*filetype = file_ptr->filetype;
	}
	else
	{
		NMPI_Type_contiguous(mpi_errno, (1, file_ptr->filetype, &copy_filetype));
		MPIU_ERR_CHK_POP(mpi_errno);

		NMPI_Type_commit(mpi_errno, (&copy_filetype));
		MPIU_ERR_CHK_POP(mpi_errno);
		*filetype = copy_filetype;
	}

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_VIEW);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_view",
			L"**mpi_file_get_view %F %p %p %p %s", file, disp, etype, filetype, datarep);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_size PMPI_File_get_size
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_size

/*@
	MPI_File_get_size - Returns the file size

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. size - size of the file in bytes (nonnegative integer)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_size(MPI_File file, MPI_Offset *size)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	ADIO_Fcntl_t *fcntl_struct;
	static const wchar_t FCNAME[] = L"MPI_File_get_size";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_SIZE);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(size, L"size", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_size %F %p", file, size); }

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	fcntl_struct = (ADIO_Fcntl_t *) MPIU_Malloc(sizeof(ADIO_Fcntl_t));
	mpi_errno = ADIO_Fcntl(file_ptr, ADIO_FCNTL_GET_FSIZE, fcntl_struct);
	MPIU_ERR_CHK_POP(mpi_errno);
	*size = fcntl_struct->fsize;
	MPIU_Free(fcntl_struct);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_SIZE);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_size",
			L"**mpi_file_get_size %F %p", file, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_position_shared PMPI_File_get_position_shared
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_position_shared

/*@
	MPI_File_get_position_shared - Returns the current position of the 
			   shared file pointer in etype units relative to the current view

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. offset - offset of shared file pointer (nonnegative integer)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_position_shared(MPI_File file, MPI_Offset *offset)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_get_position_shared";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_POSITION_SHARED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_POSITION_SHARED);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(offset, L"offset", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_position_shared %F %p", file, offset); }

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_FS_SUPPORTS_SHARED(file_ptr, mpi_errno);
	/* --END ERROR HANDLING-- */

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	mpi_errno = ADIO_Get_shared_fp(file_ptr, 0, offset);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_POSITION_SHARED);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_position_shared",
			L"**mpi_file_get_position_shared %F %p", file, offset);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_position PMPI_File_get_position
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_position

/*@
	MPI_File_get_position - Returns the current position of the 
				individual file pointer in etype units relative to
				the current view

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. offset - offset of individual file pointer (nonnegative integer)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_position(MPI_File file, MPI_Offset *offset)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_get_position";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_POSITION);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_POSITION);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(offset, L"offset", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_position %F %p", file, offset); }

	mpi_errno = ADIOI_Get_position(file_ptr, offset);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_POSITION);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_position",
			L"**mpi_file_get_position %F %p", file, offset);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_info PMPI_File_get_info
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_info

/*@
	MPI_File_get_info - Returns the hints for a file that are actually being used by MPI

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. info_used - info object (handle)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_info(MPI_File file, MPI_Info *info_used)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_get_info";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_INFO);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_INFO);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(info_used, L"info_used", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_info %F %p", file, info_used); }

	NMPI_Info_dup(mpi_errno, (file_ptr->info, info_used));
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_INFO);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_info",
			L"**mpi_file_get_info %F %p", file, info_used);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_group PMPI_File_get_group
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_group

/*@
	MPI_File_get_group - Returns the group of processes that 
						 opened the file

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. group - group that opened the file (handle)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_group(MPI_File file, MPI_Group *group)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_get_group";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_GROUP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_GROUP);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(group, L"group", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_group %F %p", file, group); }

	/* note: this will return the group of processes that called open, but
	 * with deferred open this might not be the group of processes that
	 * actually opened the file from the file system's perspective
	 */
	NMPI_Comm_group(mpi_errno, (file_ptr->comm, group));
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_GROUP);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_group",
			L"**mpi_file_get_group %F %p", file, group);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* returns the current position of the individual file pointer in etype units relative to the current view. */

#undef FUNCNAME
#define FUNCNAME ADIOI_Get_position
int ADIOI_Get_position(MPID_File_ptr fd, MPI_Offset *offset)
{
	static const wchar_t FCNAME[] = L"ADIOI_Get_postion";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Flatlist_node *flat_file;
	int i, n_filetypes, flag, frd_size;
	int filetype_size, etype_size, filetype_is_contig;
	MPI_Aint filetype_extent;
	MPI_Offset disp, byte_offset, sum=0, size_in_file;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GET_POSITION);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GET_POSITION);

	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	if (filetype_is_contig)
	{
		*offset = (fd->fp_ind - fd->disp)/etype_size;
	}
	else
	{
		/* filetype already flattened in ADIO_Open */
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype)
		{
			flat_file = flat_file->next;
		}

		NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
		MPIU_ERR_CHK_POP(mpi_errno);

		disp = fd->disp;
		byte_offset = fd->fp_ind;
		n_filetypes = -1;
		flag = 0;
		while (!flag)
		{
			sum = 0;
			n_filetypes++;
			for (i=0; i<flat_file->count; i++)
			{
				sum += flat_file->blocklens[i];
				if (disp + flat_file->indices[i] + 
			 			(MPI_Offset) n_filetypes*filetype_extent + flat_file->blocklens[i] >= byte_offset)
				{
					frd_size = (int) (disp + flat_file->indices[i] + 
						(MPI_Offset) n_filetypes*filetype_extent
						+ flat_file->blocklens[i] - byte_offset);
					sum -= frd_size;
					flag = 1;
					break;
				}
			}
		}
		size_in_file = (MPI_Offset) n_filetypes*filetype_size + sum;
		*offset = size_in_file/etype_size;
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GET_POSITION);
	return mpi_errno;
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_type_extent PMPI_File_get_type_extent
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_type_extent

/*@
	MPI_File_get_type_extent - Returns the extent of datatype in the file

Input Parameters:
. fh - file handle (handle)
. datatype - datatype (handle)

Output Parameters:
. extent - extent of the datatype (nonnegative integer)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_type_extent(MPI_File file, MPI_Datatype datatype, 
							 MPI_Aint *extent)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_get_type_extent";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_TYPE_EXTENT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_TYPE_EXTENT);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(extent, L"extent", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_type_extent %F %D %p", file, datatype, extent); }

	/* FIXME: handle other file data representations */

	NMPI_Type_extent(mpi_errno, (datatype, extent));
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_TYPE_EXTENT);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_type_extent",
			L"**mpi_file_get_type_extent %F %D %p", file, datatype, extent);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_byte_offset PMPI_File_get_byte_offset
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_byte_offset

/*@
	MPI_File_get_byte_offset - Returns the absolute byte position in 
				the file corresponding to "offset" etypes relative to
				the current view

Input Parameters:
. fh - file handle (handle)
. offset - offset (nonnegative integer)

Output Parameters:
. disp - absolute byte position of offset (nonnegative integer)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_byte_offset(MPI_File file,
							 MPI_Offset offset,
							 MPI_Offset *disp)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_get_byte_offset";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_BYTE_OFFSET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_BYTE_OFFSET);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(disp, L"disp", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_byte_offset %F %o %p", file, offset, disp); }

	if (offset < 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iobadoffset", 0);
		goto fn_fail;
	}

	mpi_errno = ADIOI_Get_byte_offset(file_ptr, offset, disp);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_BYTE_OFFSET);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_byte_offset",
			L"**mpi_file_get_byte_offset %F %o %p", file, offset, disp);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_atomicity PMPI_File_get_atomicity
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_atomicity

/*@
	MPI_File_get_atomicity - Returns the atomicity mode

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. flag - true if atomic mode, false if nonatomic mode (logical)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_atomicity(MPI_File file, int *flag)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_get_atomicity";
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_ATOMICITY);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_ATOMICITY);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(flag, L"flag", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_atomicity %F %p", file, flag); }

	*flag = file_ptr->atomicity;

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_ATOMICITY);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_atomicity",
			L"**mpi_file_get_atomicity %F %p", file, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_get_amode PMPI_File_get_amode
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_get_amode

/*@
	MPI_File_get_amode - Returns the file access mode

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. amode - access mode (integer)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_get_amode(MPI_File file, int *amode)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_get_amode";
	MPID_File *file_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_GET_AMODE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_GET_AMODE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			MPIR_ERRTEST_ARGNULL(amode, L"amode", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_get_amode %F %p", file, amode); }

	*amode = file_ptr->access_mode;

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_GET_AMODE);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_get_amode",
			L"**mpi_file_get_amode %F %p", file, amode);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
