/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS

#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_File_open = PMPI_File_open
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_File_open MPI_File_open
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_File_open as PMPI_File_open
/* end of weak pragmas */
#endif

#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_openA PMPI_File_openA
#define MPI_File_openW PMPI_File_openW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_open

extern int ADIO_Init_keyval;

/*@
	MPI_File_open - Opens a file

Input Parameters:
. comm - communicator (handle)
. filename - name of file to open (string)
. amode - file access mode (integer)
. info - info object (handle)

Output Parameters:
. fh - file handle (handle)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_openA(MPI_Comm comm, char *filename, int amode, MPI_Info info, MPI_File *fh)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_openA";
	wchar_t *wfilename = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_OPENA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_OPENA);

	if (filename != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(filename, &wfilename);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = MPI_File_openW(comm, wfilename, amode, info, fh);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wfilename != NULL)
		MPIU_Free(wfilename);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_OPENA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_File_openW(MPI_Comm comm, wchar_t *filename, int amode, MPI_Info info, MPI_File *fh)
{
	int mpi_errno = MPI_SUCCESS, file_system, flag, /* tmp_amode, */rank;
	wchar_t *tmp;
	MPI_Comm dupcomm = MPI_COMM_NULL;
	ADIOI_Fns *fsops;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_open";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_OPEN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_OPEN);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(fh, L"fh", mpi_errno);
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	NMPI_Comm_test_inter(mpi_errno, (comm, &flag));
	MPIU_ERR_CHK_POP(mpi_errno);
	/* --BEGIN ERROR HANDLING-- */
	if (flag)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_COMM, 
										  L"**commnotintra", 0);
		goto fn_fail;
	}

	if ( ((amode&MPI_MODE_RDONLY)?1:0) + ((amode&MPI_MODE_RDWR)?1:0) +
		 ((amode&MPI_MODE_WRONLY)?1:0) != 1 )
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_AMODE, 
										  L"**fileamodeone", 0);
		goto fn_fail;
	}

	if ((amode & MPI_MODE_RDONLY) && 
			((amode & MPI_MODE_CREATE) || (amode & MPI_MODE_EXCL)))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_AMODE, 
										  L"**fileamoderead", 0);
		goto fn_fail;
	}

	if ((amode & MPI_MODE_RDWR) && (amode & MPI_MODE_SEQUENTIAL))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_AMODE, 
										  L"**fileamodeseq", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_open %C %s %M %I %p", comm, filename, amode, info, fh); }

	/* check if amode is the same on all processes */
	NMPI_Comm_dup(mpi_errno, (comm, &dupcomm));
	MPIU_ERR_CHK_POP(mpi_errno);

/*  
	Removed this check because broadcast is too expensive. 
	tmp_amode = amode;
	NMPI_Bcast(mpi_errno, (&tmp_amode, 1, MPI_INT, 0, dupcomm));
	if (amode != tmp_amode) {
		FPRINTF(stderr, L"MPI_File_open: amode must be the same on all processes\n");
		MPI_Abort(MPI_COMM_WORLD, 1);
	}
*/

	/* check if ADIO has been initialized. If not, initialize it */
	if (ADIO_Init_keyval == MPI_KEYVAL_INVALID)
	{
		NMPI_Keyval_create(mpi_errno, (MPI_NULL_COPY_FN, ADIOI_End_call, &ADIO_Init_keyval, (void *) 0));
		MPIU_ERR_CHK_POP(mpi_errno);

		/* put a dummy attribute on MPI_COMM_WORLD, because we want the delete
		   function to be called when MPI_COMM_WORLD is freed. Hopefully the
		   MPI library frees MPI_COMM_WORLD when MPI_Finalize is called,
		   though the standard does not mandate this. */

		NMPI_Attr_put(mpi_errno, (MPI_COMM_WORLD, ADIO_Init_keyval, (void *) 0));
		MPIU_ERR_CHK_POP(mpi_errno);

		/* initialize ADIO */

		mpi_errno = ADIO_Init( (int *)0, (wchar_t ***)0);
		MPIU_ERR_CHK_POP(mpi_errno);
	}


	file_system = -1;

	/* resolve file system type from file name; this is a collective call */
	mpi_errno = ADIO_ResolveFileType(dupcomm, filename, &file_system, &fsops);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* Test for invalid flags in amode.
	 *
	 * eventually we should allow the ADIO implementations to test for 
	 * invalid flags through some functional interface rather than having
	 *  these tests here. -- Rob, 06/06/2001
	 */
	if (((file_system == ADIO_PIOFS) ||
		 (file_system == ADIO_PVFS) ||
		 (file_system == ADIO_PVFS2) ||
		 (file_system == ADIO_GRIDFTP)) && 
		(amode & MPI_MODE_SEQUENTIAL))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__,
										  MPI_ERR_UNSUPPORTED_OPERATION, 
										  L"**iosequnsupported", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	/* strip off prefix if there is one, but only skip prefixes
	 * if they are greater than length one to allow for windows
	 * drive specifications (e.g. c:\...) */

	tmp = wcschr(filename, L':');
	if (tmp > filename + 1)
	{
		filename = tmp + 1;
	}

	/* use default values for disp, etype, filetype */
	mpi_errno = ADIO_Open(comm, dupcomm, filename, file_system, fsops, amode, 0,
					MPI_BYTE, MPI_BYTE, info, ADIO_PERM_NULL, &file_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* determine name of file that will hold the shared file pointer */
	/* can't support shared file pointers on a file system that doesn't
	   support file locking. */
	if ((file_ptr->file_system != ADIO_PIOFS)
		  && (file_ptr->file_system != ADIO_PVFS) 
		  && (file_ptr->file_system != ADIO_PVFS2)
		  && (file_ptr->file_system != ADIO_GRIDFTP) )
	{
		NMPI_Comm_rank(mpi_errno, (dupcomm, &rank));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Shfp_fname(file_ptr, rank);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* if MPI_MODE_APPEND, set the shared file pointer to end of file.
		   indiv. file pointer already set to end of file in ADIO_Open. 
		   Here file view is just bytes. */
		if (file_ptr->access_mode & MPI_MODE_APPEND)
		{
			if (rank == file_ptr->hints->ranklist[0])
			{
				/* only one person need set the sharedfp */
				mpi_errno = ADIO_Set_shared_fp(file_ptr, file_ptr->fp_ind);
				MPIU_ERR_CHK_POP(mpi_errno);
			}
			NMPI_Barrier(mpi_errno, (dupcomm));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
	}

	*fh = file_ptr->handle;

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_OPEN);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	if (dupcomm != MPI_COMM_NULL)
	{
		int temp_mpi_errno;
		NMPI_Comm_free(temp_mpi_errno, (&dupcomm));
	}
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_open",
			L"**mpi_file_open %C %s %M %I %p", comm, filename, amode, info, fh);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(NULL, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
