/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS

#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_File_preallocate = PMPI_File_preallocate
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_File_preallocate MPI_File_preallocate
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_File_preallocate as PMPI_File_preallocate
/* end of weak pragmas */
#endif

#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_preallocate PMPI_File_preallocate
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_preallocate

/*@
	MPI_File_preallocate - Preallocates storage space for a file

Input Parameters:
. fh - file handle (handle)
. size - size to preallocate (nonnegative integer)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_preallocate(MPI_File file, MPI_Offset size)
{
	ADIO_Fcntl_t *fcntl_struct;
	int mpi_errno = MPI_SUCCESS, mynod=0;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_preallocate";
	MPI_Offset tmp_sz;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_PREALLOCATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_PREALLOCATE);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_preallocate %F %d", file, size); }

	if (size < 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_ARG,
										  L"**iobadsize", 0);
		goto fn_fail;
	}

	tmp_sz = size;
	NMPI_Bcast(mpi_errno, (&tmp_sz, 1, ADIO_OFFSET, 0, file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (tmp_sz != size)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_ARG,
										  L"**notsame", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	if (size == 0)
	{
		goto fn_exit;
	}

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	NMPI_Comm_rank(mpi_errno, (file_ptr->comm, &mynod));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (!mynod)
	{
		fcntl_struct = (ADIO_Fcntl_t *) MPIU_Malloc(sizeof(ADIO_Fcntl_t));
		fcntl_struct->diskspace = size;
		mpi_errno = ADIO_Fcntl(file_ptr, ADIO_FCNTL_SET_DISKSPACE, fcntl_struct);
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(fcntl_struct);
	}
	NMPI_Barrier(mpi_errno, (file_ptr->comm));

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_PREALLOCATE);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_preallocate",
			L"**mpi_file_preallocate %F %d", file, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
