/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read PMPI_File_read
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read

/* status object not filled currently */

/*@
	MPI_File_read - Read using individual file pointer

Input Parameters:
. fh - file handle (handle)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read(MPI_File file, void *buf, int count, MPI_Datatype datatype, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_read";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read %F %p %d %D %p", file, buf, count, datatype, status); }

	MPIU_THREADPRIV_GET;
	MPIR_Nest_incr();
	mpi_errno = MPIOI_File_read(file, (MPI_Offset) 0, ADIO_INDIVIDUAL, buf, count, datatype, status);
	MPIR_Nest_decr();
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read",
			L"**mpi_file_read %F %p %d %D %p", file, buf, count, datatype, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* prevent multiple definitions of this routine */
#ifdef MPIO_BUILD_PROFILING
int MPIOI_File_read(MPI_File file,
					MPI_Offset offset,
					int file_ptr_type,
					void *buf,
					int count,
					MPI_Datatype datatype,
					MPI_Status *status)
{
	static const wchar_t FCNAME[] = L"MPIOI_File_read";
	int mpi_errno = MPI_SUCCESS, bufsize, buftype_is_contig, filetype_is_contig;
	int datatype_size;
	MPID_File *file_ptr = NULL;
	MPI_Offset off;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPIOI_FILE_READ);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPIOI_FILE_READ);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (file_ptr_type == ADIO_EXPLICIT_OFFSET && offset < 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iobadoffset", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (count * datatype_size == 0)
	{
#ifdef HAVE_STATUS_SET_BYTES
	   MPIR_Status_set_bytes(status, datatype, 0);
#endif
		goto fn_exit;
	}

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_INTEGRAL_ETYPE(file_ptr, count, datatype_size, mpi_errno);
	MPIR_CHECK_READABLE(file_ptr, mpi_errno);
	/* --END ERROR HANDLING-- */

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = ADIOI_Datatype_iscontig(file_ptr->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	if (buftype_is_contig && filetype_is_contig)
	{
		/* convert count and offset to bytes */
		bufsize = datatype_size * count;
		if (file_ptr_type == ADIO_EXPLICIT_OFFSET)
		{
			off = file_ptr->disp + file_ptr->etype_size * offset;
		}
		else
		{
			/* ADIO_INDIVIDUAL */
			off = file_ptr->fp_ind;
		}

		/* if atomic mode requested, lock (exclusive) the region, because
		   there could be a concurrent noncontiguous request. Locking doesn't
		   work on PIOFS and PVFS, and on NFS it is done in the
		   ADIO_ReadContig.
		 */

		if ((file_ptr->atomicity) && (file_ptr->file_system != ADIO_PIOFS) && 
			(file_ptr->file_system != ADIO_NFS) && (file_ptr->file_system != ADIO_PVFS) && 
		   		 (file_ptr->file_system != ADIO_PVFS2))
			ADIOI_WRITE_LOCK(file_ptr, off, SEEK_SET, bufsize);

		mpi_errno = ADIO_ReadContig(file_ptr, buf, count, datatype, file_ptr_type, off, status); 
		MPIU_ERR_CHK_POP(mpi_errno);

		if ((file_ptr->atomicity) && (file_ptr->file_system != ADIO_PIOFS) && 
			(file_ptr->file_system != ADIO_NFS) && (file_ptr->file_system != ADIO_PVFS) &&
					(file_ptr->file_system != ADIO_PVFS2))
			ADIOI_UNLOCK(file_ptr, off, SEEK_SET, bufsize);
	}
	else
	{
		mpi_errno = ADIO_ReadStrided(file_ptr, buf, count, datatype, file_ptr_type, offset, status);
		MPIU_ERR_CHK_POP(mpi_errno);
		/* For strided and atomic mode, locking is done in ADIO_ReadStrided */
	}

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPIOI_FILE_READ);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpioi_file_read",
			L"**mpioi_file_read %F %o %d %p %d %D %p", file, offset, file_ptr_type, buf, count, datatype, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_iread_shared PMPI_File_iread_shared
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_iread_shared

/*@
	MPI_File_iread_shared - Nonblocking read using shared file pointer

Input Parameters:
. fh - file handle (handle)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. request - request object (handle)

.N fortran
@*/
#ifdef HAVE_MPI_GREQUEST
/*#include "mpiu_greq.h"*/

#ifdef USE_WIN_THREADED_IO
typedef struct iread_shared_args
{
	MPI_File file;
	void *buf;
	int count;
	MPI_Datatype datatype;
	MPI_Request request;
	MPI_Status *status;
} iread_shared_args;

static DWORD WINAPI iread_shared_thread(LPVOID lpParameter)
{
	int mpi_errno = MPI_SUCCESS;
	iread_shared_args *args = (iread_shared_args *)lpParameter;

	NMPI_File_read_shared(mpi_errno, (args->file, args->buf, args->count, args->datatype, args->status));
	/* ROMIO-1 doesn't do anything with status.MPI_ERROR */
	args->status->MPI_ERROR = mpi_errno;

	NMPI_Grequest_complete(mpi_errno, (args->request));
	MPIU_Free(args);
	return 0;
}
#endif

MPI_DLL_SPEC int MPI_CALL MPI_File_iread_shared(MPI_File file, void *buf, int count, MPI_Datatype datatype, MPI_Request *request)
{
	static const wchar_t FCNAME[] = L"MPI_File_iread_shared";
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	MPI_Status *status;
#ifdef USE_WIN_THREADED_IO
	iread_shared_args *args;
	HANDLE hThread;
#endif
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_IREAD_SHARED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_IREAD_SHARED);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_ARGNULL(request, L"request", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_iread_shared %F %p %d %D %p", file, buf, count, datatype, request); }

	status = (MPI_Status *) MPIU_Malloc(sizeof(MPI_Status));

#ifdef USE_WIN_THREADED_IO
	/* kick off the request */
	NMPI_Grequest_start(mpi_errno, (MPIU_Greq_query_fn, MPIU_Greq_free_fn, 
		MPIU_Greq_cancel_fn, status, request));

	args = (iread_shared_args*) MPIU_Malloc(sizeof(iread_shared_args));
	args->file = file;
	args->buf = buf;
	args->count = count;
	args->datatype = datatype;
	args->status = status;
	args->request = *request;
	hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)iread_shared_thread, args, 0, NULL);
	if (hThread == NULL)
	{
		mpi_errno = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
			L"MPI_File_iread_shared", __LINE__, MPI_ERR_OTHER,
			L"**fail", L"**fail %d", mpi_errno);
		goto fn_fail;
	}
	CloseHandle(hThread);

#else

	/* for now, no threads or anything fancy. 
	* just call the blocking version */
	NMPI_File_read_shared(mpi_errno, (file, buf, count, datatype, status));
	/* ROMIO-1 doesn't do anything with status.MPI_ERROR */
	status->MPI_ERROR = mpi_errno;

	/* kick off the request */
	NMPI_Grequest_start(mpi_errno, (MPIU_Greq_query_fn, MPIU_Greq_free_fn, MPIU_Greq_cancel_fn, status, request));

	/* but we did all the work already */
	NMPI_Grequest_complete(mpi_errno, (*request));
	/* passed the buck to the blocking version...*/
#endif

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_IREAD_SHARED);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_iread_shared",
			L"**mpi_file_iread_shared %F %p %d %D %p", file, buf, count, datatype, request);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#else
MPI_DLL_SPEC int MPI_CALL MPI_File_iread_shared(MPI_File file, void *buf, int count, 
						  MPI_Datatype datatype, MPI_Request *request)
{
	int mpi_errno = MPI_SUCCESS, bufsize, buftype_is_contig, filetype_is_contig;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_iread_shared";
	int datatype_size, incr;
	MPI_Status status;
	MPI_Offset off, shared_fp;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_IREAD_SHARED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_IREAD_SHARED);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_INTEGRAL_ETYPE(file_ptr, count, datatype_size, mpi_errno);
	MPIR_CHECK_FS_SUPPORTS_SHARED(file_ptr, mpi_errno);
	/* --END ERROR HANDLING-- */

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = ADIOI_Datatype_iscontig(file_ptr->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	incr = (count*datatype_size)/file_ptr->etype_size;
	mpi_errno = ADIO_Get_shared_fp(file_ptr, incr, &shared_fp);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (buftype_is_contig && filetype_is_contig)
	{
		/* convert count and shared_fp to bytes */
		bufsize = datatype_size * count;
		off = file_ptr->disp + file_ptr->etype_size * shared_fp;
		if (!(file_ptr->atomicity))
		{
			mpi_errno = ADIO_IreadContig(fh, buf, count, datatype, ADIO_EXPLICIT_OFFSET,
						off, request);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			/* to maintain strict atomicity semantics with other concurrent
			  operations, lock (exclusive) and call blocking routine */

			*request = ADIOI_Malloc_request();
			(*request)->optype = ADIOI_READ;
			(*request)->fd = file_ptr;
			(*request)->datatype = datatype;
			(*request)->queued = 0;
			(*request)->handle = 0;

			if (file_ptr->file_system != ADIO_NFS)
			{
				ADIOI_WRITE_LOCK(file_ptr, off, SEEK_SET, bufsize);
			}

			mpi_errno = ADIO_ReadContig(file_ptr, buf, count, datatype, ADIO_EXPLICIT_OFFSET,
							off, &status);
			MPIU_ERR_CHK_POP(mpi_errno);

			if (file_ptr->file_system != ADIO_NFS)
			{
				ADIOI_UNLOCK(file_ptr, off, SEEK_SET, bufsize);
			}

			file_ptr->async_count++;
			/* status info. must be linked to the request structure, so that
			   it can be accessed later from a wait */
		}
	}
	else
	{
		mpi_errno = ADIO_IreadStrided(file_ptr, buf, count, datatype, ADIO_EXPLICIT_OFFSET,
						   shared_fp, request);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_IREAD_SHARED);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_iread_shared",
			L"**mpi_file_iread_shared %F %p %d %D %p", file, buf, count, datatype, request);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_iread_at PMPI_File_iread_at
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_iread_at

/*@
	MPI_File_iread_at - Nonblocking read using explict offset

Input Parameters:
. fh - file handle (handle)
. offset - file offset (nonnegative integer)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. request - request object (handle)

.N fortran
@*/
#ifdef HAVE_MPI_GREQUEST
/*#include "mpiu_greq.h"*/

#ifdef USE_WIN_THREADED_IO
typedef struct iread_at_args
{
	MPI_File file;
	MPI_Offset offset;
	void *buf;
	int count;
	MPI_Datatype datatype;
	MPIO_Request request;
	MPI_Status *status;
} iread_at_args;

static DWORD WINAPI iread_at_thread(LPVOID lpParameter)
{
	int mpi_errno = MPI_SUCCESS;
	iread_at_args *args = (iread_at_args *)lpParameter;

	NMPI_File_read_at(mpi_errno, (args->file, args->offset, args->buf, args->count, args->datatype, args->status));
	/* ROMIO-1 doesn't do anything with status.MPI_ERROR */
	args->status->MPI_ERROR = mpi_errno;

	NMPI_Grequest_complete(mpi_errno, (args->request));
	MPIU_Free(args);
	return 0;
}
#endif

MPI_DLL_SPEC int MPI_CALL MPI_File_iread_at(MPI_File file, MPI_Offset offset, void *buf, int count, MPI_Datatype datatype, MPI_Request *request)
{
	static const wchar_t FCNAME[] = L"MPI_File_iread_at";
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	MPI_Status *status;
#ifdef USE_WIN_THREADED_IO
	iread_at_args *args;
	HANDLE hThread;
#endif
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_IREAD_AT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_IREAD_AT);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_ARGNULL(request, L"request", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_iread_at %F %o %p %d %D %p", file, offset, buf, count, datatype, request); }

	status = (MPI_Status *) MPIU_Malloc(sizeof(MPI_Status));

#ifdef USE_WIN_THREADED_IO
	/* kick off the request */
	NMPI_Grequest_start(mpi_errno, (MPIU_Greq_query_fn, MPIU_Greq_free_fn, 
		MPIU_Greq_cancel_fn, status, request));

	args = (iread_at_args*) MPIU_Malloc(sizeof(iread_at_args));
	args->file = file;
	args->offset = offset;
	args->buf = buf;
	args->count = count;
	args->datatype = datatype;
	args->status = status;
	args->request = *request;
	hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)iread_at_thread, args, 0, NULL);
	if (hThread == NULL)
	{
		mpi_errno = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
			L"MPI_File_iread_at", __LINE__, MPI_ERR_OTHER,
			L"**fail", L"**fail %d", mpi_errno);
		goto fn_fail;
	}
	CloseHandle(hThread);

#else

	/* for now, no threads or anything fancy. just call the blocking version */
	NMPI_File_read_at(mpi_errno, (file, offset, buf, count, datatype, status));
	/* ROMIO-1 doesn't do anything with status.MPI_ERROR */
	status->MPI_ERROR = mpi_errno;

	/* kick off the request */
	NMPI_Grequest_start(mpi_errno, (MPIU_Greq_query_fn, MPIU_Greq_free_fn, MPIU_Greq_cancel_fn, status, request));
	/* but we did all the work already */
	NMPI_Grequest_complete(mpi_errno, (*request));
	/* passed the buck to the blocking version...*/
#endif

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_IREAD_AT);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_iread_at",
			L"**mpi_file_iread_at %F %o %p %d %D %p", file, offset, buf, count, datatype, request);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#else
MPI_DLL_SPEC int MPI_CALL MPI_File_iread_at(MPI_File file, MPI_Offset offset, void *buf, int count, MPI_Datatype datatype, MPIO_Request *request)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr;
	static const wchar_t FCNAME[] = L"MPI_File_iread_at";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_IREAD_AT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_IREAD_AT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_ARGNULL(request, L"request", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	MPIU_THREADPRIV_GET;
	MPIR_Nest_incr();
	mpi_errno = MPIOI_File_iread(file, offset, ADIO_EXPLICIT_OFFSET, buf,
								  count, datatype, FCNAME, request);
	MPIR_Nest_decr();
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_IREAD_AT);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_iread_at",
			L"**mpi_file_iread_at %F %o %p %d %D %p", file, offset, buf, count, datatype, request);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_iread PMPI_File_iread
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_iread
/*@
	MPI_File_iread - Nonblocking read using individual file pointer

Input Parameters:
. fh - file handle (handle)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. request - request object (handle)

.N fortran
@*/
#ifdef HAVE_MPI_GREQUEST
/*#include "mpiu_greq.h"*/

#ifdef USE_WIN_THREADED_IO
typedef struct iread_args
{
	MPI_File file;
	void *buf;
	int count;
	MPI_Datatype datatype;
	MPI_Request request;
	MPI_Status *status;
} iread_args;

static DWORD WINAPI iread_thread(LPVOID lpParameter)
{
	int mpi_errno = MPI_SUCCESS;
	iread_args *args = (iread_args *)lpParameter;

	NMPI_File_read(mpi_errno, (args->file, args->buf, args->count, args->datatype, args->status));
	/* ROMIO-1 doesn't do anything with status.MPI_ERROR */
	args->status->MPI_ERROR = mpi_errno;

	NMPI_Grequest_complete(mpi_errno, (args->request));
	MPIU_Free(args);
	return 0;
}
#endif

MPI_DLL_SPEC int MPI_CALL MPI_File_iread(MPI_File file, void *buf, int count, MPI_Datatype datatype, MPI_Request *request)
{
	static const wchar_t FCNAME[] = L"MPI_File_iread";
	int mpi_errno = MPI_SUCCESS;
	MPI_Status *status;
	MPID_File *file_ptr = NULL;
#ifdef USE_WIN_THREADED_IO
	iread_args *args;
	HANDLE hThread;
#endif
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_IREAD);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_IREAD);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(request, L"request", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_iread %F %p %d %D %p", file, buf, count, datatype, request); }

	status = (MPI_Status *) MPIU_Malloc(sizeof(MPI_Status));

#ifdef USE_WIN_THREADED_IO
	/* kick off the request */
	NMPI_Grequest_start(mpi_errno, (MPIU_Greq_query_fn, MPIU_Greq_free_fn, 
		MPIU_Greq_cancel_fn, status, request));

	args = (iread_args*) MPIU_Malloc(sizeof(iread_args));
	args->file = file;
	args->buf = buf;
	args->count = count;
	args->datatype = datatype;
	args->status = status;
	args->request = *request;
	hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)iread_thread, args, 0, NULL);
	if (hThread == NULL)
	{
		mpi_errno = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
			L"MPI_File_iread", __LINE__, MPI_ERR_OTHER,
			L"**fail", L"**fail %d", mpi_errno);
		goto fn_fail;
	}
	CloseHandle(hThread);

#else

	/* for now, no threads or anything fancy. 
	* just call the blocking version */
	NMPI_File_read(mpi_errno, (file, buf, count, datatype, status));
	/* ROMIO-1 doesn't do anything with status.MPI_ERROR */
	status->MPI_ERROR = mpi_errno;

	/* kick off the request */
	NMPI_Grequest_start(mpi_errno, (MPIU_Greq_query_fn, MPIU_Greq_free_fn, 
		MPIU_Greq_cancel_fn, status, request));
	/* but we did all the work already */
	NMPI_Grequest_complete(mpi_errno, (*request));

	/* passed the buck to the blocking version...*/
#endif

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_IREAD);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_iread",
			L"**mpi_file_iread %F %p %d %D %p", file, buf, count, datatype, request);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#else
MPI_DLL_SPEC int MPI_CALL MPI_File_iread(MPI_File file, void *buf, int count, MPI_Datatype datatype, MPI_Request *request)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_iread";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_IREAD);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_IREAD);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(request, L"request", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	MPIU_THREADPRIV_GET;
	MPIR_Nest_incr();
	mpi_errno = MPIOI_File_iread(file, (MPI_Offset) 0, ADIO_INDIVIDUAL,
								  buf, count, datatype, FCNAME, request);
	MPIR_Nest_decr();
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_IREAD);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_iread",
			L"**mpi_file_iread %F %p %d %D %p", file, buf, count, datatype, request);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

#ifndef HAVE_MPI_GREQUEST
/* prevent multiple definitions of this routine */
#ifdef MPIO_BUILD_PROFILING
int MPIOI_File_iread(MPI_File file,
					 MPI_Offset offset,
					 int file_ptr_type,
					 void *buf,
					 int count,
					 MPI_Datatype datatype,
					 MPI_Request *request)
{
	static const wchar_t FCNAME[] = L"MPIOI_File_iread";
	int mpi_errno = MPI_SUCCESS, bufsize, buftype_is_contig, filetype_is_contig;
	int datatype_size;
	MPI_Status status;
	MPID_File *file_ptr = NULL;
	MPI_Offset off;
	MPID_MPI_STATE_DECL(MPID_STATE_MPIOI_FILE_IREAD);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPIOI_FILE_IREAD);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(request, L"request", mpi_errno);
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (file_ptr_type == ADIO_EXPLICIT_OFFSET && offset < 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_ARG,
										  L"**iobadoffset", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_INTEGRAL_ETYPE(fh, count, datatype_size, mpi_errno);
	MPIR_CHECK_READABLE(fh, mpi_errno);
	/* --END ERROR HANDLING-- */

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = ADIOI_Datatype_iscontig(file_ptr->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	if (buftype_is_contig && filetype_is_contig)
	{
		/* convert count and offset to bytes */
		bufsize = datatype_size * count;

		if (file_ptr_type == ADIO_EXPLICIT_OFFSET)
		{
			off = file_ptr->disp + file_ptr->etype_size * offset;
		}
		else
		{
			off = file_ptr->fp_ind;
		}

		if (!(file_ptr->atomicity))
		{
			mpi_errno = ADIO_IreadContig(file_ptr, buf, count, datatype, file_ptr_type,
						off, request);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			/* to maintain strict atomicity semantics with other concurrent
			  operations, lock (exclusive) and call blocking routine */

			*request = ADIOI_Malloc_request();
			(*request)->optype = ADIOI_READ;
			(*request)->fd = file_ptr;
			(*request)->datatype = datatype;
			(*request)->queued = 0;
			(*request)->handle = 0;

			if ((file_ptr->file_system != ADIO_PIOFS) && 
			  (file_ptr->file_system != ADIO_NFS) && (file_ptr->file_system != ADIO_PVFS)
			  && (file_ptr->file_system != ADIO_PVFS2))
			{
				ADIOI_WRITE_LOCK(file_ptr, off, SEEK_SET, bufsize);
			}

			mpi_errno = ADIO_ReadContig(file_ptr, buf, count, datatype, file_ptr_type, 
							off, &status);
			MPIU_ERR_CHK_POP(mpi_errno);

			if ((file_ptr->file_system != ADIO_PIOFS) && 
			   (file_ptr->file_system != ADIO_NFS) && (file_ptr->file_system != ADIO_PVFS)
			   && (file_ptr->file_system != ADIO_PVFS2))
			{
				ADIOI_UNLOCK(file_ptr, off, SEEK_SET, bufsize);
			}

			file_ptr->async_count++;
			/* status info. must be linked to the request structure, so that it
			   can be accessed later from a wait */
		}
	}
	else
	{
		mpi_errno = ADIO_IreadStrided(file_ptr, buf, count, datatype, file_ptr_type,
						   offset, request);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPIOI_FILE_IREAD);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpioi_file_iread",
			L"**mpioi_file_iread %F %o %d %p %d %D %p", file, offset, file_ptr_type, buf, count, datatype, request);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_shared PMPI_File_read_shared
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_shared

/* status object not filled currently */

/*@
	MPI_File_read_shared - Read using shared file pointer

Input Parameters:
. fh - file handle (handle)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_shared(MPI_File file, void *buf, int count, MPI_Datatype datatype, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS, bufsize, buftype_is_contig, filetype_is_contig;
	static const wchar_t FCNAME[] = L"MPI_File_read_shared";
	int datatype_size, incr;
	MPI_Offset off, shared_fp;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_SHARED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_CS_ENTER();
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_SHARED);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_shared %F %p %d %D %p", file, buf, count, datatype, status); }

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (count * datatype_size == 0)
	{
#ifdef HAVE_STATUS_SET_BYTES
		MPIR_Status_set_bytes(status, datatype, 0);
#endif
		goto fn_exit;
	}

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_INTEGRAL_ETYPE(file_ptr, count, datatype_size, mpi_errno);
	MPIR_CHECK_READABLE(file_ptr, mpi_errno);
	MPIR_CHECK_FS_SUPPORTS_SHARED(file_ptr, mpi_errno);
	/* --END ERROR HANDLING-- */

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = ADIOI_Datatype_iscontig(file_ptr->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	incr = (count * datatype_size) / file_ptr->etype_size;

	mpi_errno = ADIO_Get_shared_fp(file_ptr, incr, &shared_fp);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* contiguous or strided? */
	if (buftype_is_contig && filetype_is_contig)
	{
		/* convert count and shared_fp to bytes */
		bufsize = datatype_size * count;
		off = file_ptr->disp + file_ptr->etype_size * shared_fp;

		/* if atomic mode requested, lock (exclusive) the region, because there
		   could be a concurrent noncontiguous request. On NFS, locking 
		   is done in the ADIO_ReadContig.*/

		if ((file_ptr->atomicity) && (file_ptr->file_system != ADIO_NFS))
			ADIOI_WRITE_LOCK(file_ptr, off, SEEK_SET, bufsize);

		mpi_errno = ADIO_ReadContig(file_ptr, buf, count, datatype, ADIO_EXPLICIT_OFFSET, off, status);
		MPIU_ERR_CHK_POP(mpi_errno);

		if ((file_ptr->atomicity) && (file_ptr->file_system != ADIO_NFS))
			ADIOI_UNLOCK(file_ptr, off, SEEK_SET, bufsize);
	}
	else
	{
		mpi_errno = ADIO_ReadStrided(file_ptr, buf, count, datatype, ADIO_EXPLICIT_OFFSET, shared_fp, status);
		MPIU_ERR_CHK_POP(mpi_errno);
		/* For strided and atomic mode, locking is done in ADIO_ReadStrided */
	}

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_SHARED);
	MPID_CS_EXIT();
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_shared",
			L"**mpi_file_read_shared %F %p %d %D %p", file, buf, count, datatype, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_ordered_end PMPI_File_read_ordered_end
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_ordered_end

/*@
	MPI_File_read_ordered_end - Complete a split collective read using shared file pointer

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_ordered_end(MPI_File file, void *buf, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_read_ordered_end";
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_ORDERED_END);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_UNREFERENCED_ARG(buf);

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_ORDERED_END);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (!(file_ptr->split_coll_count))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iosplitcollnone", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_ordered_end %F %p %p", file, buf, status); }

#ifdef HAVE_STATUS_SET_BYTES
	if (status != MPI_STATUS_IGNORE)
	   *status = file_ptr->split_status;
#endif
	file_ptr->split_coll_count = 0;

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_ORDERED_END);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_ordered_end",
			L"**mpi_file_read_ordered_end %F %p %p", file, buf, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_ordered_begin PMPI_File_read_ordered_begin
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_ordered_begin

/*@
	MPI_File_read_ordered_begin - Begin a split collective read using shared file pointer

Input Parameters:
. fh - file handle (handle)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_ordered_begin(MPI_File file, void *buf, int count, MPI_Datatype datatype)
{
	int mpi_errno = MPI_SUCCESS, datatype_size, nprocs, myrank, incr;
	int source, dest;
	MPI_Offset shared_fp;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_read_ordered_begin";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_ORDERED_BEGIN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_ORDERED_BEGIN);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (file_ptr->split_coll_count)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iosplitcoll", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_ordered_begin %F %p %d %D", file, buf, count, datatype); }

	file_ptr->split_coll_count = 1;

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_INTEGRAL_ETYPE(file_ptr, count, datatype_size, mpi_errno);
	MPIR_CHECK_FS_SUPPORTS_SHARED(file_ptr, mpi_errno);
	/* --END ERROR HANDLING-- */

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	NMPI_Comm_size(mpi_errno, (file_ptr->comm, &nprocs));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Comm_rank(mpi_errno, (file_ptr->comm, &myrank));
	MPIU_ERR_CHK_POP(mpi_errno);

	incr = (count * datatype_size) / file_ptr->etype_size;
	/* Use a message as a L'token' to order the operations */
	source = myrank - 1;
	dest   = myrank + 1;
	if (source < 0) source = MPI_PROC_NULL;
	if (dest >= nprocs) dest = MPI_PROC_NULL;
	NMPI_Recv(mpi_errno, (NULL, 0, MPI_BYTE, source, 0, file_ptr->comm, MPI_STATUS_IGNORE));
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = ADIO_Get_shared_fp(file_ptr, incr, &shared_fp);
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Send(mpi_errno, (NULL, 0, MPI_BYTE, dest, 0, file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = ADIO_ReadStridedColl(file_ptr, buf, count, datatype, ADIO_EXPLICIT_OFFSET, shared_fp, &file_ptr->split_status);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_ORDERED_BEGIN);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_ordered_begin",
			L"**mpi_file_read_ordered_begin %F %p %d %D", file, buf, count, datatype);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_ordered PMPI_File_read_ordered
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_ordered

/* status object not filled currently */

/*@
	MPI_File_read_ordered - Collective read using shared file pointer

Input Parameters:
. fh - file handle (handle)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_ordered(MPI_File file, void *buf, int count, MPI_Datatype datatype, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS, datatype_size, nprocs, myrank, incr;
	int source, dest;
	static const wchar_t FCNAME[] = L"MPI_File_read_ordered";
	MPI_Offset shared_fp=0;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_ORDERED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_ORDERED);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_ordered %F %p %d %D %p", file, buf, count, datatype, status); }

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_INTEGRAL_ETYPE(file_ptr, count, datatype_size, mpi_errno);
	MPIR_CHECK_FS_SUPPORTS_SHARED(file_ptr, mpi_errno);
	/* --END ERROR HANDLING-- */

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	NMPI_Comm_size(mpi_errno, (file_ptr->comm, &nprocs));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Comm_rank(mpi_errno, (file_ptr->comm, &myrank));
	MPIU_ERR_CHK_POP(mpi_errno);

	incr = (count * datatype_size) / file_ptr->etype_size;

	/* Use a message as a L'token' to order the operations */
	source = myrank - 1;
	dest   = myrank + 1;
	if (source < 0) source = MPI_PROC_NULL;
	if (dest >= nprocs) dest = MPI_PROC_NULL;
	NMPI_Recv(mpi_errno, (NULL, 0, MPI_BYTE, source, 0, file_ptr->comm, MPI_STATUS_IGNORE));
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = ADIO_Get_shared_fp(file_ptr, incr, &shared_fp);
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Send(mpi_errno, (NULL, 0, MPI_BYTE, dest, 0, file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = ADIO_ReadStridedColl(file_ptr, buf, count, datatype, ADIO_EXPLICIT_OFFSET, shared_fp, status);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_ORDERED);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_ordered",
			L"**mpi_file_read_ordered %F %p %d %D %p", file, buf, count, datatype, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_at_all PMPI_File_read_at_all
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_at_all

/* status object not filled currently */

/*@
	MPI_File_read_at_all - Collective read using explict offset

Input Parameters:
. fh - file handle (handle)
. offset - file offset (nonnegative integer)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_at_all(MPI_File file, MPI_Offset offset, void *buf, int count, MPI_Datatype datatype, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_iread_at";
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_AT_ALL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_AT_ALL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_at_all %F %o %p %d %D %p", file, offset, buf, count, datatype, status); }

	mpi_errno = MPIOI_File_read_all(file, offset, ADIO_EXPLICIT_OFFSET, buf, count, datatype, status);
	MPIU_ERR_CHK_POP(mpi_errno);

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_AT_ALL);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_at_all",
			L"**mpi_file_read_at_all %F %o %p %d %D %p", file, offset, buf, count, datatype, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_at PMPI_File_read_at
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_at

/* status object not filled currently */

/*@
	MPI_File_read_at - Read using explict offset

Input Parameters:
. fh - file handle (handle)
. offset - file offset (nonnegative integer)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_at(MPI_File file, MPI_Offset offset, void *buf, int count, MPI_Datatype datatype, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_read_at";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_AT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_AT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_at %F %o %p %d %D %p", file, offset, buf, count, datatype, status); }

	/* ADIOI_File_read() defined in mpi-io/read.c */
	MPIU_THREADPRIV_GET;
	MPIR_Nest_incr();
	mpi_errno = MPIOI_File_read(file, offset, ADIO_EXPLICIT_OFFSET, buf, count, datatype, status);
	MPIR_Nest_decr();
	MPIU_ERR_CHK_POP(mpi_errno);

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_AT);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_at",
			L"**mpi_file_read_at %F %o %p %d %D %p", file, offset, buf, count, datatype, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_all_end PMPI_File_read_all_end
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_all_end

/*@
	MPI_File_read_all_end - Complete a split collective read using
	individual file pointer

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_all_end(MPI_File file, void *buf, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_iread";
	MPID_File *file_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_ALL_END);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_ALL_END);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_all_end %F %p %p", file, buf, status); }

	mpi_errno = MPIOI_File_read_all_end(file, buf, status);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_ALL_END);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_all_end",
			L"**mpi_file_read_all_end %F %p %p", file, buf, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* prevent multiple definitions of this routine */
#ifdef MPIO_BUILD_PROFILING
int MPIOI_File_read_all_end(MPI_File file, void *buf, MPI_Status *status)
{
	static const wchar_t FCNAME[] = L"MPIOI_File_read_all_end";
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPIOI_FILE_READ_ALL_END);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_UNREFERENCED_ARG(buf);

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPIOI_FILE_READ_ALL_END);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (!(file_ptr->split_coll_count))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iosplitcollnone", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

#ifdef HAVE_STATUS_SET_BYTES
	if (status != MPI_STATUS_IGNORE)
	   *status = file_ptr->split_status;
#endif
	file_ptr->split_coll_count = 0;

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPIOI_FILE_READ_ALL_END);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpioi_file_read_all_end",
			L"**mpioi_file_read_all_end %F %p %p", file, buf, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_all_begin PMPI_File_read_all_begin
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_all_begin

/*@
	MPI_File_read_all_begin - Begin a split collective read using individual file pointer

Input Parameters:
. fh - file handle (handle)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_all_begin(MPI_File file, void *buf, int count, MPI_Datatype datatype)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_read_all_begin";
	MPID_File *file_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_ALL_BEGIN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_ALL_BEGIN);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_all_begin %F %p %d %D", file, buf, count, datatype); }

	mpi_errno = MPIOI_File_read_all_begin(file, (MPI_Offset) 0, ADIO_INDIVIDUAL, buf, count, datatype);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_ALL_BEGIN);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_all_begin",
			L"**mpi_file_read_all_begin %F %p %d %D", file, buf, count, datatype);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* prevent multiple definitions of this routine */
#ifdef MPIO_BUILD_PROFILING
int MPIOI_File_read_all_begin(MPI_File file,
							  MPI_Offset offset,
							  int file_ptr_type,
							  void *buf,
							  int count,
							  MPI_Datatype datatype)
{
	static const wchar_t FCNAME[] = L"MPIOI_File_read_all_begin";
	int mpi_errno = MPI_SUCCESS, datatype_size;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPIOI_FILE_READ_ALL_BEGIN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPIOI_FILE_READ_ALL_BEGIN);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (file_ptr_type == ADIO_EXPLICIT_OFFSET && offset < 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iobadoffset", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_INTEGRAL_ETYPE(file_ptr, count, datatype_size, mpi_errno);
	MPIR_CHECK_READABLE(file_ptr, mpi_errno);

	if (file_ptr->split_coll_count)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iosplitcoll", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	file_ptr->split_coll_count = 1;

	mpi_errno = ADIO_ReadStridedColl(file_ptr, buf, count, datatype, file_ptr_type, offset, &file_ptr->split_status);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPIOI_FILE_READ_ALL_BEGIN);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpioi_file_read_all_begin",
			L"**mpioi_file_read_all_begin %F %o %d %p %d %D", file, offset, file_ptr_type, buf, count, datatype);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_all PMPI_File_read_all
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_all

/* status object not filled currently */

/*@
	MPI_File_read_all - Collective read using individual file pointer

Input Parameters:
. fh - file handle (handle)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_all(MPI_File file, void *buf, int count, MPI_Datatype datatype, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"MPI_File_read_all";
	MPID_File *file_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_ALL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_ALL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_all %F %p %d %D %p", file, buf, count, datatype, status); }

	mpi_errno = MPIOI_File_read_all(file, (MPI_Offset) 0, ADIO_INDIVIDUAL, buf, count, datatype, status);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_ALL);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_all",
			L"**mpi_file_read_all %F %p %d %D %p", file, buf, count, datatype, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Note: MPIOI_File_read_all also used by MPI_File_read_at_all */
/* prevent multiple definitions of this routine */
#ifdef MPIO_BUILD_PROFILING
int MPIOI_File_read_all(MPI_File file,
						MPI_Offset offset,
						int file_ptr_type,
						void *buf,
						int count,
						MPI_Datatype datatype,
						MPI_Status *status)
{
	static const wchar_t FCNAME[] = L"MPIOI_File_read_all";
	int mpi_errno = MPI_SUCCESS, datatype_size;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPIOI_FILE_READ_ALL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPIOI_FILE_READ_ALL);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			/* Validate datatype handle */
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (file_ptr_type == ADIO_EXPLICIT_OFFSET && offset < 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**iobadoffset", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_INTEGRAL_ETYPE(file_ptr, count, datatype_size, mpi_errno);
	MPIR_CHECK_READABLE(file_ptr, mpi_errno);
	/* --END ERROR HANDLING-- */

	mpi_errno = ADIO_ReadStridedColl(file_ptr, buf, count, datatype, file_ptr_type, offset, status);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPIOI_FILE_READ_ALL);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpioi_file_read_all",
			L"**mpioi_file_read_all %F %o %d %p %d %D %p", file, offset, file_ptr_type, buf, count, datatype, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_at_all_end PMPI_File_read_at_all_end
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_at_all_end

/*@
	MPI_File_read_at_all_end - Complete a split collective read using
	explict offset

Input Parameters:
. fh - file handle (handle)

Output Parameters:
. buf - initial address of buffer (choice)
. status - status object (Status)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_at_all_end(MPI_File file, void *buf, MPI_Status *status)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_read_at_all_end";
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_AT_ALL_END);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_AT_ALL_END);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_at_all_end %F %p %p", file, buf, status); }

	mpi_errno = MPIOI_File_read_all_end(file, buf, status);
	MPIU_ERR_CHK_POP(mpi_errno);

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_AT_ALL_END);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_at_all_end",
			L"**mpi_file_read_at_all_end %F %p %p", file, buf, status);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_read_at_all_begin PMPI_File_read_at_all_begin
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_read_at_all_begin

/*@
	MPI_File_read_at_all_begin - Begin a split collective read using explict offset

Input Parameters:
. fh - file handle (handle)
. offset - file offset (nonnegative integer)
. count - number of elements in buffer (nonnegative integer)
. datatype - datatype of each buffer element (handle)

Output Parameters:
. buf - initial address of buffer (choice)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_read_at_all_begin(MPI_File file, MPI_Offset offset, void *buf, int count, MPI_Datatype datatype)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_read_at_all_begin";
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_READ_AT_ALL_BEGIN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_READ_AT_ALL_BEGIN);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			MPIR_ERRTEST_USERBUFFER(buf, count, datatype, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_read_at_all_begin %F %o %p %d %D", file, offset, buf, count, datatype); }

	mpi_errno = MPIOI_File_read_all_begin(file, offset, ADIO_EXPLICIT_OFFSET, buf, count, datatype);
	MPIU_ERR_CHK_POP(mpi_errno);

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_READ_AT_ALL_BEGIN);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_read_at_all_begin",
			L"**mpi_file_read_at_all_begin %F %o %p %d %D", file, offset, buf, count, datatype);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
