/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS

#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_File_seek = PMPI_File_seek
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_File_seek MPI_File_seek
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_File_seek as PMPI_File_seek
/* end of weak pragmas */
#endif

#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_seek PMPI_File_seek
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_seek

/*@
	MPI_File_seek - Updates the individual file pointer

Input Parameters:
. fh - file handle (handle)
. offset - file offset (integer)
. whence - update mode (state)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_seek(MPI_File file, MPI_Offset offset, int whence)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_seek";
	MPI_Offset curr_offset, eof_offset, result_offset;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SEEK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SEEK);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_seek %F %o %d", file, offset, whence); }

	switch(whence)
	{
	case MPI_SEEK_SET:
		/* --BEGIN ERROR HANDLING-- */
		if (offset < 0) {
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
											  MPIR_ERR_RECOVERABLE, FCNAME,
											  __LINE__, MPI_ERR_ARG,
											  L"**iobadoffset", 0);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
		break;
	case MPI_SEEK_CUR:
		/* find offset corr. to current location of file pointer */
		ADIOI_Get_position(file_ptr, &curr_offset);
		offset += curr_offset;

		/* --BEGIN ERROR HANDLING-- */
		if (offset < 0)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
											  MPIR_ERR_RECOVERABLE, FCNAME,
											  __LINE__, MPI_ERR_ARG,
											  L"**ionegoffset", 0);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */

		break;
	case MPI_SEEK_END:
		/* we can in many cases do seeks w/o a file actually opened, but not in
		 * the MPI_SEEK_END case */
		MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

		/* find offset corr. to end of file */
		mpi_errno = ADIOI_Get_eof_offset(file_ptr, &eof_offset);
		MPIU_ERR_CHK_POP(mpi_errno);
		offset += eof_offset;

		/* --BEGIN ERROR HANDLING-- */
		if (offset < 0)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
											  MPIR_ERR_RECOVERABLE, FCNAME,
											  __LINE__, MPI_ERR_ARG,
											  L"**ionegoffset", 0);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */

		break;
	default:
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_ARG,
										  L"**iobadwhence", 0);
		goto fn_fail;
		/* --END ERROR HANDLING-- */
	}

	mpi_errno = ADIO_SeekIndividual(file_ptr, offset, ADIO_SEEK_SET, &result_offset);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SEEK);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_seek",
			L"**mpi_file_seek %F %o %d", file, offset, whence);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
