/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS

#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_File_seek_shared = PMPI_File_seek_shared
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_File_seek_shared MPI_File_seek_shared
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_File_seek_shared as PMPI_File_seek_shared
/* end of weak pragmas */
#endif

#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_seek_shared PMPI_File_seek_shared
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_seek_shared

/*@
	MPI_File_seek_shared - Updates the shared file pointer

Input Parameters:
. fh - file handle (handle)
. offset - file offset (integer)
. whence - update mode (state)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_seek_shared(MPI_File file, MPI_Offset offset, int whence)
{
	int mpi_errno = MPI_SUCCESS, tmp_whence, myrank;
	static const wchar_t FCNAME[] = L"MPI_File_seek_shared";
	MPI_Offset curr_offset, eof_offset, tmp_offset;
	MPID_File *file_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SEEK_SHARED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SEEK_SHARED);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			MPIR_ERRTEST_FILE_NOT_SEQUENTIAL_MODE(file_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */


	/* --BEGIN ERROR HANDLING-- */
	MPIR_CHECK_FS_SUPPORTS_SHARED(file_ptr, mpi_errno);
	/* --END ERROR HANDLING-- */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_seek_shared %F %o %d", file, offset, whence); }

	tmp_offset = offset;
	NMPI_Bcast(mpi_errno, (&tmp_offset, 1, ADIO_OFFSET, 0, file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);
	/* --BEGIN ERROR HANDLING-- */
	if (tmp_offset != offset)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_ARG,
										  L"**notsame", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	tmp_whence = whence;
	NMPI_Bcast(mpi_errno, (&tmp_whence, 1, MPI_INT, 0, file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);
	/* --BEGIN ERROR HANDLING-- */
	if (tmp_whence != whence)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_ARG,
										  L"**iobadwhence", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	NMPI_Comm_rank(mpi_errno, (file_ptr->comm, &myrank));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (!myrank)
	{
		switch(whence)
		{
		case MPI_SEEK_SET:
			/* --BEGIN ERROR HANDLING-- */
			if (offset < 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
												  MPIR_ERR_RECOVERABLE,
												  FCNAME, __LINE__,
												  MPI_ERR_ARG,
												  L"**iobadoffset", 0);
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			break;
		case MPI_SEEK_CUR:
			/* get current location of shared file pointer */
			mpi_errno = ADIO_Get_shared_fp(file_ptr, 0, &curr_offset);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
												  MPIR_ERR_FATAL,
												  FCNAME, __LINE__,
												  MPI_ERR_INTERN, 
												  L"**iosharedfailed", 0);
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			offset += curr_offset;
			/* --BEGIN ERROR HANDLING-- */
			if (offset < 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
												  MPIR_ERR_RECOVERABLE,
												  FCNAME, __LINE__,
												  MPI_ERR_ARG,
												  L"**ionegoffset", 0);
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			break;
		case MPI_SEEK_END:
			/* find offset corr. to end of file */
			mpi_errno = ADIOI_Get_eof_offset(file_ptr, &eof_offset);
			MPIU_ERR_CHK_POP(mpi_errno);
			offset += eof_offset;
			/* --BEGIN ERROR HANDLING-- */
			if (offset < 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
												  MPIR_ERR_RECOVERABLE,
												  FCNAME, __LINE__,
												  MPI_ERR_ARG,
												  L"**ionegoffset", 0);
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			break;
		default:
			/* --BEGIN ERROR HANDLING-- */
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
											  MPIR_ERR_RECOVERABLE,
											  FCNAME, __LINE__, MPI_ERR_ARG,
											  L"**iobadwhence", 0);
			goto fn_fail;
			/* --END ERROR HANDLING-- */
		}

		mpi_errno = ADIO_Set_shared_fp(file_ptr, offset);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
											  MPIR_ERR_FATAL,
											  FCNAME, __LINE__,
											  MPI_ERR_INTERN, 
											  L"**iosharedfailed", 0);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */

	}

	/* FIXME: explain why the barrier is necessary */
	NMPI_Barrier(mpi_errno, (file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SEEK_SHARED);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_seek_shared",
			L"**mpi_file_seek_shared %F %o %d", file, offset, whence);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
