/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

/* -- Begin Profiling Symbol Block for routine MPI_Get_processor_name */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Get_processor_name = PMPI_Get_processor_name
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Get_processor_name  MPI_Get_processor_name
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Get_processor_name as PMPI_Get_processor_name
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines.  You can use USE_WEAK_SYMBOLS to see if MPICH is
   using weak symbols to implement the MPI routines. */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Get_processor_nameA
#undef MPI_Get_processor_nameW
#define MPI_Get_processor_nameA PMPI_Get_processor_nameA
#define MPI_Get_processor_nameW PMPI_Get_processor_nameW

/* Any internal routines can go here.  Make them static if possible.  If they
   are used by both the MPI and PMPI versions, use PMPI_LOCAL instead of 
   static; this macro expands into "static" if weak symbols are supported and
   into nothing otherwise. */
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Get_processor_name

/*@
  MPI_Get_processor_name - Gets the name of the processor

  Output Parameters:
+ name - A unique specifier for the actual (as opposed to virtual) node. This
  must be an array of size at least 'MPI_MAX_PROCESSOR_NAME'.
- resultlen - Length (in characters) of the name 

  Notes:
  The name returned should identify a particular piece of hardware; 
  the exact format is implementation defined.  This name may or may not
  be the same as might be returned by 'gethostname', 'uname', or 'sysinfo'.

.N ThreadSafe

.N Fortran

 In Fortran, the character argument should be declared as a character string
 of 'MPI_MAX_PROCESSOR_NAME' rather than an array of dimension 
 'MPI_MAX_PROCESSOR_NAME'.  That is, 
.vb
   character*(MPI_MAX_PROCESSOR_NAME) name
.ve
 rather than
.vb
   character name(MPI_MAX_PROCESSOR_NAME)
.ve
 The two 

.N FortranString

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Get_processor_nameA( char *name, int *resultlen )
{
	static const wchar_t FCNAME[] = L"MPI_Get_processor_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t wname[MPI_MAX_PROCESSOR_NAME];
	int wname_len;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GET_PROCESSOR_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GET_PROCESSOR_NAMEA);

	mpi_errno = MPI_Get_processor_nameW(wname, &wname_len);
	MPIU_ERR_CHK_POP(mpi_errno);

	error = wcstombs_s(&num_converted, name, MPI_MAX_PROCESSOR_NAME, wname, _TRUNCATE);
	if (error != 0)
	{
		if (num_converted >= MPI_MAX_PROCESSOR_NAME)
		{
			/* truncation error */
			MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"name", (int)num_converted);
			goto fn_fail;
		}
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}
	else if (num_converted == (size_t)-1)
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
		goto fn_fail;
	}
	*resultlen = (int)num_converted - 1;

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GET_PROCESSOR_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Get_processor_nameW( wchar_t *name, int *resultlen )
{
	static const wchar_t FCNAME[] = L"MPI_Get_processor_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GET_PROCESSOR_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GET_PROCESSOR_NAME);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(name,L"name",mpi_errno);
			MPIR_ERRTEST_ARGNULL(resultlen,L"resultlen",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_get_processor_name %p %p", name, resultlen); }
	
	mpi_errno = MPID_Get_processor_name( name, MPI_MAX_PROCESSOR_NAME, 
										 resultlen );
	
	/* ... end of body of routine ... */

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GET_PROCESSOR_NAME);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_get_processor_name",
			L"**mpi_get_processor_name %p %p", name, resultlen);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
