/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_create
#define MPI_Win_create PMPI_Win_create
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_create

/*@
   MPI_Win_create - Create an MPI Window object for one-sided communication

   Input Parameters:
+ base - initial address of window (choice) 
. size - size of window in bytes (nonnegative integer) 
. disp_unit - local unit size for displacements, in bytes (positive integer) 
. info - info argument (handle) 
- comm - communicator (handle) 

  Output Parameter:
. win - window object returned by the call (handle) 

.N ThreadSafe
.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_INFO
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_create(void *base, MPI_Aint size, int disp_unit, MPI_Info info, MPI_Comm comm, MPI_Win *win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_create";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_Comm *comm_ptr = NULL;
	MPID_Info *info_ptr = NULL;

	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_WIN_CREATE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Info_get_ptr( info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate pointers */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			if (size < 0)
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_SIZE,L"**rmasize",L"**rmasize %d", size);
			if (disp_unit <= 0)
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**arg", L"**arg %s", L"disp_unit must be positive");
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_create %p %d %d %I %C %p", base, size, disp_unit, info, comm, win); }
	
	mpi_errno = MPID_Win_create(base, size, disp_unit, info_ptr, comm_ptr, &win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* return the handle of the window object to the user */
	*win = win_ptr->handle;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_WIN_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_create", 
			L"**mpi_win_create %p %d %d %I %C %p", base, size, disp_unit, info, comm, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_complete
#define MPI_Win_complete PMPI_Win_complete

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_complete

/*@
   MPI_Win_complete - Completes an RMA operations begun after an MPI_Win_start.

   Input Parameter:
. win - window object (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_complete(MPI_Win win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_complete";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_COMPLETE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_WIN_COMPLETE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_complete %W", win); }
	
	mpi_errno = MPID_Win_complete(win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_WIN_COMPLETE);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_complete", L"**mpi_win_complete %W", win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Put
#define MPI_Put PMPI_Put

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Put

/*@
   MPI_Put - Put data into a memory window on a remote process

   Input Parameters:
+ origin_addr -initial address of origin buffer (choice) 
. origin_count -number of entries in origin buffer (nonnegative integer) 
. origin_datatype -datatype of each entry in origin buffer (handle) 
. target_rank -rank of target (nonnegative integer) 
. target_disp -displacement from start of window to target buffer (nonnegative integer) 
. target_count -number of entries in target buffer (nonnegative integer) 
. target_datatype -datatype of each entry in target buffer (handle) 

- win - window object used for communication (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_COUNT
.N MPI_ERR_RANK
.N MPI_ERR_TYPE
.N MPI_ERR_WIN
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Put(void *origin_addr, int origin_count, MPI_Datatype
			origin_datatype, int target_rank, MPI_Aint target_disp,
			int target_count, MPI_Datatype target_datatype, MPI_Win
			win)
{
	static const wchar_t FCNAME[] = L"MPI_Put";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PUT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_PUT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Comm * comm_ptr;

			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;

			MPIR_ERRTEST_COUNT(origin_count, mpi_errno);
			MPIR_ERRTEST_DATATYPE(origin_datatype, L"origin_datatype", 
								  mpi_errno);
			MPIR_ERRTEST_COUNT(target_count, mpi_errno);
			MPIR_ERRTEST_DATATYPE(target_datatype, L"target_datatype",
								  mpi_errno);
			MPIR_ERRTEST_DISP(target_disp, mpi_errno);

			if (HANDLE_GET_KIND(origin_datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype *datatype_ptr = NULL;
				
				MPID_Datatype_get_ptr(origin_datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}

			if (HANDLE_GET_KIND(target_datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype *datatype_ptr = NULL;
				
				MPID_Datatype_get_ptr(target_datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}

			MPID_Comm_get_ptr(win_ptr->comm, comm_ptr);
			MPIR_ERRTEST_SEND_RANK(comm_ptr, target_rank, mpi_errno);
			
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_put %p %d %D %d %d %d %D %W",
		origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, win); }
	
	if (target_rank == MPI_PROC_NULL) goto fn_exit;

	mpi_errno = MPID_Put(origin_addr, origin_count, origin_datatype,
						 target_rank, target_disp, target_count,
						 target_datatype, win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_PUT);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_put", L"**mpi_put %p %d %D %d %d %d %D %W",
			origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Get
#define MPI_Get PMPI_Get

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Get

/*@
   MPI_Get - Get data from a memory window on a remote process

 Input Parameters:
+ origin_addr - Address of the buffer in which to receive the data
. origin_count - number of entries in origin buffer (nonnegative integer) 
. origin_datatype - datatype of each entry in origin buffer (handle) 
. target_rank - rank of target (nonnegative integer) 
. target_disp - displacement from window start to the beginning of the 
  target buffer (nonnegative integer) 
. target_count - number of entries in target buffer (nonnegative integer) 
. target_datatype - datatype of each entry in target buffer (handle) 
- win - window object used for communication (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_COUNT
.N MPI_ERR_RANK
.N MPI_ERR_TYPE
.N MPI_ERR_WIN
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Get(void *origin_addr, int origin_count, MPI_Datatype
			origin_datatype, int target_rank, MPI_Aint target_disp,
			int target_count, MPI_Datatype target_datatype, MPI_Win
			win)
{
	static const wchar_t FCNAME[] = L"MPI_Get";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_GET);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Comm * comm_ptr;
			
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;

			MPIR_ERRTEST_COUNT(origin_count, mpi_errno);
			MPIR_ERRTEST_DATATYPE(origin_datatype, L"origin_datatype",
								  mpi_errno);
			MPIR_ERRTEST_COUNT(target_count, mpi_errno);
			MPIR_ERRTEST_DATATYPE(target_datatype, L"target_datatype",
								  mpi_errno);
			MPIR_ERRTEST_DISP(target_disp, mpi_errno);

			if (HANDLE_GET_KIND(origin_datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype *datatype_ptr = NULL;
				
				MPID_Datatype_get_ptr(origin_datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}

			if (HANDLE_GET_KIND(target_datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype *datatype_ptr = NULL;
				
				MPID_Datatype_get_ptr(target_datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}

			MPID_Comm_get_ptr(win_ptr->comm, comm_ptr);
			MPIR_ERRTEST_SEND_RANK(comm_ptr, target_rank, mpi_errno);
			
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_get %p %d %D %d %d %d %D %W",
		origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, win); }
	
	if (target_rank == MPI_PROC_NULL) goto fn_exit;

	mpi_errno = MPID_Get(origin_addr, origin_count, origin_datatype,
						 target_rank, target_disp, target_count,
						 target_datatype, win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_GET);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_get", L"**mpi_get %p %d %D %d %d %d %D %W",
			origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Free_mem
#define MPI_Free_mem PMPI_Free_mem

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Free_mem

/*@
   MPI_Free_mem - Free memory allocated with MPI_Alloc_mem

   Input Parameter:
.  base - initial address of memory segment allocated by 'MPI_ALLOC_MEM' 
	   (choice) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Free_mem(void *base)
{
	static const wchar_t FCNAME[] = L"MPI_Free_mem";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FREE_MEM);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FREE_MEM);

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_free_mem %p", base); }
	
	mpi_errno = MPID_Free_mem(base);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FREE_MEM);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_free_mem", L"**mpi_free_mem %p", base);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Alloc_mem
#define MPI_Alloc_mem PMPI_Alloc_mem

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Alloc_mem

/*@
   MPI_Alloc_mem - Allocate memory for message passing and RMA

  Input Parameters:
+ size - size of memory segment in bytes (nonnegative integer) 
- info - info argument (handle) 

  Output Parameter:
. baseptr - pointer to beginning of memory segment allocated 

   Notes:
 Using this routine from Fortran requires that the Fortran compiler accept
 a common pointer extension.  See Section 4.11 (Memory Allocation) in the
 MPI-2 standard for more information and examples.

   Also note that while 'baseptr' is a 'void *' type, this is 
   simply to allow easy use of any pointer object for this parameter.  
   In fact, this argument is really a 'void **' type, that is, a 
   pointer to a pointer. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INFO
.N MPI_ERR_ARG
.N MPI_ERR_NO_MEM
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Alloc_mem(MPI_Aint size, MPI_Info info, void *baseptr)
{
	static const wchar_t FCNAME[] = L"MPI_Alloc_mem";
	int mpi_errno = MPI_SUCCESS;
	void *ap;
	MPID_Info *info_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ALLOC_MEM);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ALLOC_MEM);
	
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNEG(size, L"size", mpi_errno);
			MPIR_ERRTEST_ARGNULL(baseptr, L"baseptr", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_alloc_mem %d %I %p", size, info, baseptr); }
	
	MPID_Info_get_ptr( info, info_ptr );

	ap = MPID_Alloc_mem(size, info_ptr);

	/* --BEGIN ERROR HANDLING-- */
	if (!ap)
	{
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_NO_MEM, L"**allocmem", 0 );
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	*(void **)baseptr = ap;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ALLOC_MEM);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_alloc_mem",
			L"**mpi_alloc_mem %d %I %p", size, info, baseptr);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Accumulate
#define MPI_Accumulate PMPI_Accumulate

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Accumulate

/*@
   MPI_Accumulate - Accumulate data into the target process using remote 
   memory access 

   Input Parameters:
+ origin_addr - initial address of buffer (choice) 
. origin_count - number of entries in buffer (nonnegative integer) 
. origin_datatype - datatype of each buffer entry (handle) 
. target_rank - rank of target (nonnegative integer) 
. target_disp - displacement from start of window to beginning of target 
  buffer (nonnegative integer)  
. target_count - number of entries in target buffer (nonnegative integer) 
. target_datatype - datatype of each entry in target buffer (handle) 
. op - predefined reduce operation (handle) 
- win - window object (handle) 

   Notes:
The basic components of both the origin and target datatype must be the same 
predefined datatype (e.g., all 'MPI_INT' or all 'MPI_DOUBLE_PRECISION').

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_COUNT
.N MPI_ERR_RANK
.N MPI_ERR_TYPE
.N MPI_ERR_WIN
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Accumulate(void *origin_addr, int origin_count, MPI_Datatype
				   origin_datatype, int target_rank, MPI_Aint
				   target_disp, int target_count, MPI_Datatype
				   target_datatype, MPI_Op op, MPI_Win win) 
{
	static const wchar_t FCNAME[] = L"MPI_Accumulate";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ACCUMULATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_ACCUMULATE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Comm * comm_ptr;
			
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;

			MPIR_ERRTEST_COUNT(origin_count, mpi_errno);
			MPIR_ERRTEST_DATATYPE(origin_datatype, L"origin_datatype",
								  mpi_errno);
			MPIR_ERRTEST_COUNT(target_count, mpi_errno);
			MPIR_ERRTEST_DATATYPE(target_datatype, L"target_datatype",
								  mpi_errno);
			MPIR_ERRTEST_DISP(target_disp, mpi_errno);

			if (HANDLE_GET_KIND(origin_datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype *datatype_ptr = NULL;
				
				MPID_Datatype_get_ptr(origin_datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}

			if (HANDLE_GET_KIND(target_datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype *datatype_ptr = NULL;
				
				MPID_Datatype_get_ptr(target_datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}

			MPID_Comm_get_ptr(win_ptr->comm, comm_ptr);
			MPIR_ERRTEST_SEND_RANK(comm_ptr, target_rank, mpi_errno);
			
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_accumulate %p %d %D %d %d %d %D %O %W", origin_addr, origin_count, origin_datatype,
		target_rank, target_disp, target_count, target_datatype, op, win); }
	
	if (target_rank == MPI_PROC_NULL) goto fn_exit;

	mpi_errno = MPID_Accumulate(origin_addr, origin_count, origin_datatype,
								target_rank, target_disp, target_count,
								target_datatype, op, win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_ACCUMULATE);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_accumulate",
			L"**mpi_accumulate %p %d %D %d %d %d %D %O %W", origin_addr, origin_count, origin_datatype,
			target_rank, target_disp, target_count, target_datatype, op, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_wait
#define MPI_Win_wait PMPI_Win_wait

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_wait

/*@
   MPI_Win_wait - Completes an RMA exposure epoch begun with MPI_Win_post

   Input Parameter:
. win - window object (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_wait(MPI_Win win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_wait";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_WAIT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_WIN_WAIT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_wait %W", win); }
	
	mpi_errno = MPID_Win_wait(win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_WIN_WAIT);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_wait", 
			L"**mpi_win_wait %W", win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_unlock
#define MPI_Win_unlock PMPI_Win_unlock

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_unlock

/*@
   MPI_Win_unlock - Completes an RMA access epoch at the target process

   Input Parameters:
+ rank - rank of window (nonnegative integer) 
- win - window object (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_RANK
.N MPI_ERR_WIN
.N MPI_ERR_OTHER

.seealso: MPI_Win_lock
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_unlock(int rank, MPI_Win win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_unlock";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_UNLOCK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_UNLOCK);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Comm * comm_ptr;
			
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;

			MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
			MPIR_ERRTEST_SEND_RANK(comm_ptr, rank, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_unlock %d %W", rank, win); }
	
	mpi_errno = MPID_Win_unlock(rank, win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_UNLOCK);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_unlock", 
			L"**mpi_win_unlock %d %W", rank, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_test
#define MPI_Win_test PMPI_Win_test

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_test

/*@
   MPI_Win_test - Test whether an RMA exposure epoch has completed

   Input Parameter:
. win - window object (handle) 

   Output Parameter:
. flag - success flag (logical) 

   Notes:
   This is the nonblocking version of 'MPI_Win_wait'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER
.N MPI_ERR_ARG

.seealso: MPI_Win_wait, MPI_Win_post
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_test(MPI_Win win, int *flag)
{
	static const wchar_t FCNAME[] = L"MPI_Win_test";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_TEST);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_TEST);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */

			MPIR_ERRTEST_ARGNULL(flag, L"flag", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_test %W %p", win, flag); }
	
	mpi_errno = MPID_Win_test(win_ptr, flag);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
 
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_TEST);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_test", 
			L"**mpi_win_test %W %p", win, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_start
#define MPI_Win_start PMPI_Win_start

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_start

/*@
   MPI_Win_start - Start an RMA access epoch for MPI

  Input Parameters:
+ group - group of target processes (handle) 
. assert - Used to optimize this call; zero may be used as a default.
  See notes. (integer) 
- win - window object (handle) 

   Notes:
   The 'assert' argument is used to indicate special conditions for the
   fence that an implementation may use to optimize the 'MPI_Win_start' 
   operation.  The value zero is always correct.  Other assertion values
   may be or''ed together.  Assertions tha are valid for 'MPI_Win_start' are\:

. MPI_MODE_NOCHECK - the matching calls to 'MPI_WIN_POST' have already 
  completed on all target processes when the call to 'MPI_WIN_START' is made. 
  The nocheck option can be specified in a start call if and only if it is 
  specified in each matching post call. This is similar to the optimization 
  of ready-send that may save a handshake when the handshake is implicit in 
  the code. (However, ready-send is matched by a regular receive, whereas 
  both start and post must specify the nocheck option.) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_start(MPI_Group group, int assert, MPI_Win win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_start";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_Group *group_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_START);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_WIN_START);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			MPIR_ERRTEST_GROUP(group, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Get handles to MPI objects. */
	MPID_Win_get_ptr( win, win_ptr );
	MPID_Group_get_ptr(group, group_ptr);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;

			MPID_Group_valid_ptr(group_ptr, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_start %G %A %W", group, assert, win); }

	mpi_errno = MPID_Win_start(group_ptr, assert, win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_WIN_START);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_start", 
			L"**mpi_win_start %G %A %W", group, assert, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_set_nameA
#undef MPI_Win_set_nameW
#define MPI_Win_set_nameA PMPI_Win_set_nameA
#define MPI_Win_set_nameW PMPI_Win_set_nameW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_set_name

/*@
   MPI_Win_set_name - Set the print name for an MPI RMA window

   Input Parameters:
+ win - window whose identifier is to be set (handle) 
- win_name - the character string which is remembered as the name (string) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_set_nameA(MPI_Win win, char *win_name)
{
	static const wchar_t FCNAME[] = L"MPI_Win_set_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t name[MPI_MAX_OBJECT_NAME];
	wchar_t *name_ptr = L"";
	size_t len = 0;
	errno_t error;

	/* convert the ASCII string to a wide character string and call the wide character version */

	if (win_name != NULL)
	{
		error = mbstowcs_s(&len, name, MPI_MAX_OBJECT_NAME, win_name, _TRUNCATE);
		if (error != 0)
		{
			if (len >= MPI_MAX_OBJECT_NAME)
			{
				/* truncation error */
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"win_name", (int)len);
				goto fn_fail;
			}
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
		if (len >= MPI_MAX_OBJECT_NAME)
		{
			/* truncation error */
			win_name[MPI_MAX_OBJECT_NAME-1] = '\0';
			MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"win_name", (int)len);
			goto fn_fail;
		}
		if (len > 0 && len < MPI_MAX_OBJECT_NAME)
			name_ptr = name;
	}
	else
	{
		name_ptr = NULL;
	}

	mpi_errno = MPI_Win_set_nameW(win, name_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Win_set_nameW(MPI_Win win, wchar_t *win_name)
{
	static const wchar_t FCNAME[] = L"MPI_Win_set_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_SET_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_SET_NAME);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */

			MPIR_ERRTEST_ARGNULL(win_name, L"win_name", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_set_name %W %s", win, win_name); }
	
	/*MPIU_Wcsncpy( win_ptr->name, win_name, MPI_MAX_OBJECT_NAME );*/
	MPIU_Wcsncpy( win_ptr->name, win_name, MPI_MAX_OBJECT_NAME );
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_SET_NAME);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_set_name", 
			L"**mpi_win_set_name %W %s", win, win_name);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}


/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_post
#define MPI_Win_post PMPI_Win_post

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_post

/*@
   MPI_Win_post - Start an RMA exposure epoch

   Input parameters:
+ group - group of origin processes (handle) 
. assert - Used to optimize this call; zero may be used as a default.
  See notes. (integer) 
- win - window object (handle) 

   Notes:
   The 'assert' argument is used to indicate special conditions for the
   fence that an implementation may use to optimize the 'MPI_Win_post' 
   operation.  The value zero is always correct.  Other assertion values
   may be or''ed together.  Assertions that are valid for 'MPI_Win_post' are\:

+ MPI_MODE_NOCHECK - the matching calls to 'MPI_WIN_START' have not yet 
  occurred on any origin processes when the call to 'MPI_WIN_POST' is made. 
  The nocheck option can be specified by a post call if and only if it is 
  specified by each matching start call. 
. MPI_MODE_NOSTORE - the local window was not updated by local stores (or 
  local get or receive calls) since last synchronization. This may avoid 
  the need for cache synchronization at the post call. 
- MPI_MODE_NOPUT - the local window will not be updated by put or accumulate 
  calls after the post call, until the ensuing (wait) synchronization. This 
  may avoid the need for cache synchronization at the wait call. 

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_post(MPI_Group group, int assert, MPI_Win win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_post";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_Group *group_ptr=NULL;

	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_POST);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_WIN_POST);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			MPIR_ERRTEST_GROUP(group, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	MPID_Group_get_ptr(group, group_ptr);
			
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;

			MPID_Group_valid_ptr(group_ptr, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_post %G %A %W", group, assert, win); }
	
	mpi_errno = MPID_Win_post(group_ptr, assert, win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_WIN_POST);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_post",
			L"**mpi_win_post %G %A %W", group, assert, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_lock
#define MPI_Win_lock PMPI_Win_lock

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_lock

/*@
   MPI_Win_lock - Begin an RMA access epoch at the target process.

   Input Parameters:
+ lock_type - Indicates whether other processes may access the target 
   window at the same time (if 'MPI_LOCK_SHARED') or not ('MPI_LOCK_EXCLUSIVE')
. rank - rank of locked window (nonnegative integer) 
. assert - Used to optimize this call; zero may be used as a default.
  See notes. (integer) 
- win - window object (handle) 

   Notes:

   The name of this routine is misleading.  In particular, this
   routine need not block, except when the target process is the calling 
   process.  

   Implementations may restrict the use of RMA communication that is 
   synchronized
   by lock calls to windows in memory allocated by 'MPI_Alloc_mem'. Locks can 
   be used portably only in such memory. 

   The 'assert' argument is used to indicate special conditions for the
   fence that an implementation may use to optimize the 'MPI_Win_fence' 
   operation.  The value zero is always correct.  Other assertion values
   may be or''ed together.  Assertions that are valid for 'MPI_Win_fence' are\:

. MPI_MODE_NOCHECK - no other process holds, or will attempt to acquire a 
  conflicting lock, while the caller holds the window lock. This is useful 
  when mutual exclusion is achieved by other means, but the coherence 
  operations that may be attached to the lock and unlock calls are still 
  required. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_RANK
.N MPI_ERR_WIN
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_lock(int lock_type, int rank, int assert, MPI_Win win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_lock";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_LOCK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_LOCK);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Comm *comm_ptr;

			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not value, it will be reset to null */
			if (mpi_errno) goto fn_fail;
			
			if (lock_type != MPI_LOCK_SHARED && lock_type != MPI_LOCK_EXCLUSIVE)
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**locktype", 0 );

			MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
			MPIR_ERRTEST_SEND_RANK(comm_ptr, rank, mpi_errno);

			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_lock %d %d %A %W", lock_type, rank, assert, win); }
	
	mpi_errno = MPID_Win_lock(lock_type, rank, assert, win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_LOCK);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_lock", 
			L"**mpi_win_lock %d %d %A %W", lock_type, rank, assert, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_get_nameA
#undef MPI_Win_get_nameW
#define MPI_Win_get_nameA PMPI_Win_get_nameA
#define MPI_Win_get_nameW PMPI_Win_get_nameW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_get_name

/*@
   MPI_Win_get_name - Get the print name associated with the MPI RMA window

   Input Parameter:
. win - window whose name is to be returned (handle) 

   Output Parameters:
+ win_name - the name previously stored on the window, or a empty string if 
  no such name exists (string) 
- resultlen - length of returned name (integer) 

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_nameA(MPI_Win win, char *win_name, int *resultlen)
{
	static const wchar_t FCNAME[] = L"MPI_Win_get_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t name[MPI_MAX_OBJECT_NAME] = L"";
	int len;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_GET_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_GET_NAMEA);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(win_name, L"win_name", mpi_errno);
			MPIR_ERRTEST_ARGNULL(resultlen, L"resultlen", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	mpi_errno = MPI_Win_get_nameW(win, name, &len);
	if (mpi_errno == MPI_SUCCESS)
	{
		if (len == 0)
		{
			*win_name = '\0';
			*resultlen = 0;
			MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_GET_NAMEA);
			return mpi_errno;
		}

		error = wcstombs_s(&num_converted, win_name, MPI_MAX_OBJECT_NAME, name, _TRUNCATE);
		if (error != 0)
		{
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
		if (num_converted > 0)
		{
			*resultlen = (int)num_converted - 1;
		}
		else
		{
			/* conversion error */
			MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
		}
	}

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_GET_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Win_get_nameW(MPI_Win win, wchar_t *win_name, int *resultlen)
{
	static const wchar_t FCNAME[] = L"MPI_Win_get_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_GET_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_WIN_GET_NAME);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */

			MPIR_ERRTEST_ARGNULL(win_name, L"win_name", mpi_errno);
			MPIR_ERRTEST_ARGNULL(resultlen, L"resultlen", mpi_errno);

			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_get_name %W %p %p", win, win_name, resultlen); }

	/*
	MPIU_Wcsncpy( win_name, win_ptr->name, MPI_MAX_OBJECT_NAME );
	*resultlen = (int)wcslen( win_name );
	*/
	MPIU_Wcsncpy( win_name, win_ptr->name, MPI_MAX_OBJECT_NAME );
	*resultlen = (int)wcslen( win_name );

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_WIN_GET_NAME);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_get_name", 
			L"**mpi_win_get_name %W %p %p", win, win_name, resultlen);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_get_group
#define MPI_Win_get_group PMPI_Win_get_group

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_get_group

/*@
   MPI_Win_get_group - Get the MPI Group of the window object

   Input Parameter:
. win - window object (handle) 

   Output Parameter:
. group - group of processes which share access to the window (handle) 

   Notes:
   The group is a duplicate of the group from the communicator used to 
   create the MPI window, and should be freed with 'MPI_Group_free' when
   it is no longer needed.  This group can be used to form the group of 
   neighbors for the routines 'MPI_Win_post' and 'MPI_Win_start'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_ARG
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_group(MPI_Win win, MPI_Group *group)
{
	static const wchar_t FCNAME[] = L"MPI_Win_get_group";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_GET_GROUP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_WIN_GET_GROUP);

	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not value, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_get_group %W %p", win, group); }
	
	NMPI_Comm_group(mpi_errno, (win_ptr->comm, group));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_WIN_GET_GROUP);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_get_group", 
			L"**mpi_win_get_group %W %p", win, group);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_free
#define MPI_Win_free PMPI_Win_free

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_free

/*@
   MPI_Win_free - Free an MPI RMA window

   Input Parameter:
. win - window object (handle) 

   Notes:
   If successfully freed, 'win' is set to 'MPI_WIN_NULL'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_free(MPI_Win *win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_free";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_FREE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_WIN_FREE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(*win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( *win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			/* If win_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_free %p", win); }
	
	if (MPIR_Process.attr_free && win_ptr->attributes)
	{
		mpi_errno = MPIR_Process.attr_free( win_ptr->handle, win_ptr->attributes );
	}
	/*
	 * If the user attribute free function returns an error, 
	 * then do not free the window
	 */
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	
	mpi_errno = MPID_Win_free(&win_ptr);
	*win = MPI_WIN_NULL;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_WIN_FREE);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_free", L"**mpi_win_free %p", win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Win_fence
#define MPI_Win_fence PMPI_Win_fence

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Win_fence

/*@
   MPI_Win_fence - Perform an MPI fence synchronization on a MPI window

   Input Parameters:
+ assert - program assertion (integer) 
- win - window object (handle) 

   Notes:
   The 'assert' argument is used to indicate special conditions for the
   fence that an implementation may use to optimize the 'MPI_Win_fence' 
   operation.  The value zero is always correct.  Other assertion values
   may be or''ed together.  Assertions that are valid for 'MPI_Win_fence' are\:

+ MPI_MODE_NOSTORE - the local window was not updated by local stores 
  (or local get or receive calls) since last synchronization. 
. MPI_MODE_NOPUT - the local window will not be updated by put or accumulate 
  calls after the fence call, until the ensuing (fence) synchronization. 
. MPI_MODE_NOPRECEDE - the fence does not complete any sequence of locally 
  issued RMA calls. If this assertion is given by any process in the window 
  group, then it must be given by all processes in the group. 
- MPI_MODE_NOSUCCEED - the fence does not start any sequence of locally 
  issued RMA calls. If the assertion is given by any process in the window 
  group, then it must be given by all processes in the group. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
.N MPI_ERR_WIN
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Win_fence(int assert, MPI_Win win)
{
	static const wchar_t FCNAME[] = L"MPI_Win_fence";
	int mpi_errno = MPI_SUCCESS;
	MPID_Win *win_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_WIN_FENCE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("rma");
	MPID_MPI_RMA_FUNC_ENTER(MPID_STATE_MPI_WIN_FENCE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_WIN(win, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Win_get_ptr( win, win_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate win_ptr */
			MPID_Win_valid_ptr( win_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_win_fence %A %W", assert, win); }
	
	mpi_errno = MPID_Win_fence(assert, win_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_RMA_FUNC_EXIT(MPID_STATE_MPI_WIN_FENCE);
	MPIU_THREAD_SINGLE_CS_EXIT("rma");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_win_fence",
			L"**mpi_win_fence %A %W", assert, win);
	}
#   endif
	mpi_errno = MPIR_Err_return_win( win_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
