/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#include "namepub.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_connectA
#undef MPI_Comm_connectW
#define MPI_Comm_connectA PMPI_Comm_connectA
#define MPI_Comm_connectW PMPI_Comm_connectW

/* Any internal routines can go here.  Make them static if possible */
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_connect

/*@
   MPI_Comm_connect - Make a request to form a new intercommunicator

 Input Parameters:
+ port_name - network address (string, used only on root) 
. info - implementation-dependent information (handle, used only on root) 
. root - rank in comm of root node (integer) 
- comm - intracommunicator over which call is collective (handle) 

 Output Parameter:
. newcomm - intercommunicator with server as remote group (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_INFO
.N MPI_ERR_PORT
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_connectA(char *port_name, MPI_Info info, int root, MPI_Comm comm, MPI_Comm *newcomm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_connectA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wport_name = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_CONNECTA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_CONNECTA);

	if (port_name != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(port_name, &wport_name);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = MPI_Comm_connectW(wport_name, info, root, comm, newcomm);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wport_name != NULL)
	{
		MPIU_Free(wport_name);
	}
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_CONNECTA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Comm_connectW(wchar_t *port_name, MPI_Info info, int root, MPI_Comm comm, MPI_Comm *newcomm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_connect";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_Comm *newcomm_ptr = NULL;
	MPID_Info *info_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_CONNECT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_CONNECT);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			MPIR_ERRTEST_INFO_OR_NULL(info, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Info_get_ptr( info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_connect %s %I %d %C %p", port_name, info, root, comm, newcomm); }
	
	mpi_errno = MPID_Comm_connect(port_name, info_ptr, root, comm_ptr, &newcomm_ptr); 
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	*newcomm = newcomm_ptr->handle;
		
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_CONNECT);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_connect", 
			L"**mpi_comm_connect %s %I %d %C %p", port_name, info, root, comm, newcomm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_acceptA
#undef MPI_Comm_acceptW
#define MPI_Comm_acceptA PMPI_Comm_acceptA
#define MPI_Comm_acceptW PMPI_Comm_acceptW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_accept

/*@
   MPI_Comm_accept - Accept a request to form a new intercommunicator

 Input Parameters:
+ port_name - port name (string, used only on root) 
. info - implementation-dependent information (handle, used only on root) 
. root - rank in comm of root node (integer) 
- IN - comm intracommunicator over which call is collective (handle) 

 Output Parameter:
. newcomm - intercommunicator with client as remote group (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INFO
.N MPI_ERR_COMM
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_acceptA(char *port_name, MPI_Info info, int root, MPI_Comm comm, MPI_Comm *newcomm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_acceptA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wport_name = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_ACCEPTA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_ACCEPTA);

	if (port_name != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(port_name, &wport_name);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = MPI_Comm_acceptW(wport_name, info, root, comm, newcomm);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wport_name != NULL)
	{
		MPIU_Free(wport_name);
	}
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_ACCEPTA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Comm_acceptW(wchar_t *port_name, MPI_Info info, int root, MPI_Comm comm, MPI_Comm *newcomm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_accept";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_Comm *newcomm_ptr = NULL;
	MPID_Info *info_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_ACCEPT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_ACCEPT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			MPIR_ERRTEST_INFO_OR_NULL(info, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Info_get_ptr( info, info_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_accept %s %I %d %C %p", port_name, info, root, comm, newcomm); }
	
	mpi_errno = MPID_Comm_accept(port_name, info_ptr, root, comm_ptr, &newcomm_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	*newcomm = newcomm_ptr->handle;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_ACCEPT);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_accept",
			L"**mpi_comm_accept %s %I %d %C %p", port_name, info, root, comm, newcomm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Close_portA
#undef MPI_Close_portW
#define MPI_Close_portA PMPI_Close_portA
#define MPI_Close_portW PMPI_Close_portW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Close_port

/*@
   MPI_Close_port - close port

   Input Parameter:
.  port_name - a port name (string)

.N NotThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Close_portA(char *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Close_portA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wport_name = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CLOSE_PORTA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CLOSE_PORTA);

	if (port_name != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(port_name, &wport_name);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = MPI_Close_portW(wport_name);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wport_name != NULL)
	{
		MPIU_Free(wport_name);
	}
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CLOSE_PORTA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Close_portW(wchar_t *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Close_port";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CLOSE_PORT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CLOSE_PORT);

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_close_port %s", port_name); }
	
	mpi_errno = MPID_Close_port(port_name);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CLOSE_PORT);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_close_port",
			L"**mpi_close_port %s", port_name);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Unpublish_nameA
#undef MPI_Unpublish_nameW
#define MPI_Unpublish_nameA PMPI_Unpublish_nameA
#define MPI_Unpublish_nameW PMPI_Unpublish_nameW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Unpublish_name

/*@
   MPI_Unpublish_name - Unpublish a service name published with 
   MPI_Publish_name

 Input Parameters:
+ service_name - a service name (string) 
. info - implementation-specific information (handle) 
- port_name - a port name (string) 

.N ThreadSafeNoUpdate

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INFO
.N MPI_ERR_ARG
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Unpublish_nameA(char *service_name, MPI_Info info, char *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Unpublish_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wservice_name = NULL;
	wchar_t *wport_name = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_UNPUBLISH_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_UNPUBLISH_NAMEA);

	if (service_name != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(service_name, &wservice_name);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	if (port_name != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(port_name, &wport_name);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = MPI_Unpublish_nameW(wservice_name, info, wport_name);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wservice_name != NULL)
		MPIU_Free(wservice_name);
	if (wport_name != NULL)
		MPIU_Free(wport_name);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_UNPUBLISH_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Unpublish_nameW(wchar_t *service_name, MPI_Info info, wchar_t *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Unpublish_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Info *info_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_UNPUBLISH_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");  
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_UNPUBLISH_NAME);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO_OR_NULL(info, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate info_ptr (only if not null) */
			if (info_ptr) 
				MPID_Info_valid_ptr( info_ptr, mpi_errno );
			MPIR_ERRTEST_ARGNULL( service_name, L"service_name", mpi_errno );
			MPIR_ERRTEST_ARGNULL( port_name, L"port_name", mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_unpublish_name %s %I %s", service_name, info, port_name); }
	
#   ifdef HAVE_NAMEPUB_SERVICE
	{
		/* The standard leaves explicitly undefined what happens if the code 
		   attempts to unpublish a name that is not published.  In this case, 
		   MPI_Unpublish_name could be called before a name service structure
		   is allocated. */
		if (!MPIR_NS_Initialized)
		{
			mpi_errno = MPID_NS_Create( info_ptr, &MPIR_Namepub );
			MPIU_ERR_CHKANDJUMP((mpi_errno != MPI_SUCCESS), mpi_errno, MPI_ERR_OTHER, L"**fail");
			MPIR_NS_Initialized = 1;
			MPIR_Add_finalize( (int (*)(void*))MPID_NS_Free, &MPIR_Namepub, 9 );
		}

		mpi_errno = MPID_NS_Unpublish( MPIR_Namepub, info_ptr, (const wchar_t *)service_name );
		if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	}
#   else
	{
		/* No name publishing service available */
		MPIU_ERR_SETANDJUMP(mpi_errno, MPI_ERR_OTHER, L"**nonamepub");
	}
#   endif

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_UNPUBLISH_NAME);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_unpublish_name",
			L"**mpi_unpublish_name %s %I %s", service_name, info, port_name);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Publish_nameA
#undef MPI_Publish_nameW
#define MPI_Publish_nameA PMPI_Publish_nameA
#define MPI_Publish_nameW PMPI_Publish_nameW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Publish_name

/*@
   MPI_Publish_name - Publish a service name for use with MPI_Comm_connect

 Input Parameters:
+ service_name - a service name to associate with the port (string) 
. info - implementation-specific information (handle) 
- port_name - a port name (string) 

Notes:
The maximum size string that may be supplied for 'port_name' is
'MPI_MAX_PORT_NAME'. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_INFO
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Publish_nameA(char *service_name, MPI_Info info, char *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Publish_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wservice_name = NULL;
	wchar_t *wport_name = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PUBLISH_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_PUBLISH_NAMEA);

	if (service_name != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(service_name, &wservice_name);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	if (port_name != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(port_name, &wport_name);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = MPI_Publish_nameW(wservice_name, info, wport_name);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wservice_name != NULL)
		MPIU_Free(wservice_name);
	if (wport_name != NULL)
		MPIU_Free(wport_name);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_PUBLISH_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Publish_nameW(wchar_t *service_name, MPI_Info info, wchar_t *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Publish_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Info *info_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PUBLISH_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_PUBLISH_NAME);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO_OR_NULL(info, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Get handles to MPI objects. */
	MPID_Info_get_ptr( info, info_ptr );
	
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate info_ptr (only if not null) */
			if (info_ptr)
				MPID_Info_valid_ptr( info_ptr, mpi_errno );
			/* Validate character pointers */
			MPIR_ERRTEST_ARGNULL( service_name, L"service_name", mpi_errno );
			MPIR_ERRTEST_ARGNULL( port_name, L"port_name", mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_publish_name %s %I %s", service_name, info, port_name); }
	
#   ifdef HAVE_NAMEPUB_SERVICE
	{
		if (!MPIR_NS_Initialized)
		{
			mpi_errno = MPID_NS_Create( info_ptr, &MPIR_Namepub );
			MPIU_ERR_CHKANDJUMP((mpi_errno != MPI_SUCCESS), mpi_errno, MPI_ERR_OTHER, L"**fail");
			MPIR_NS_Initialized = 1;
			MPIR_Add_finalize( (int (*)(void*))MPID_NS_Free, &MPIR_Namepub, 9 );
		}

		mpi_errno = MPID_NS_Publish( MPIR_Namepub, info_ptr, (const wchar_t *)service_name, (const wchar_t *)port_name );
		if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	}
#   else
	{
		/* No name publishing service available */
		MPIU_ERR_SETANDJUMP(mpi_errno, MPI_ERR_OTHER, L"**nonamepub");
	}
#   endif

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_PUBLISH_NAME);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_publish_name",
			L"**mpi_publish_name %s %I %s", service_name, info, port_name);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Open_portA
#undef MPI_Open_portW
#define MPI_Open_portA PMPI_Open_portA
#define MPI_Open_portW PMPI_Open_portW

/* Any internal routines can go here.  Make them static if possible */
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Open_port

/*@
   MPI_Open_port - Establish an address that can be used to establish 
   connections between groups of MPI processes

 Input Parameter:
. info - implementation-specific information on how to establish a 
   port for 'MPI_Comm_accept' (handle) 

 Output Parameter:
. port_name - newly established port (string) 

Notes:
MPI copies a system-supplied port name into 'port_name'. 'port_name' identifies
the newly opened port and can be used by a client to contact the server. 
The maximum size string that may be supplied by the system is 
'MPI_MAX_PORT_NAME'. 

 Reserved Info Key Values:
+ ip_port - Value contains IP port number at which to establish a port. 
- ip_address - Value contains IP address at which to establish a port.
 If the address is not a valid IP address of the host on which the
 'MPI_OPEN_PORT' call is made, the results are undefined. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Open_portA(MPI_Info info, char *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Open_portA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t wport_name[MPI_MAX_PORT_NAME];
	wchar_t *wport_name_ptr = NULL;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_OPEN_PORTA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_OPEN_PORTA);

	if (port_name != NULL)
		wport_name_ptr = wport_name;

	mpi_errno = MPI_Open_portW(info, wport_name_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (port_name != NULL)
	{
		port_name[0] = '\0';
		error = wcstombs_s(&num_converted, port_name, MPI_MAX_PORT_NAME, wport_name, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= MPI_MAX_PORT_NAME)
			{
				/* truncation error */
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"port_name", (int)num_converted);
				goto fn_fail;
			}
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
	}

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_OPEN_PORTA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Open_portW(MPI_Info info, wchar_t *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Open_port";
	int mpi_errno = MPI_SUCCESS;
	MPID_Info *info_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_OPEN_PORT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_OPEN_PORT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Note that a NULL info is allowed */
			MPIR_ERRTEST_INFO_OR_NULL(info, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* FIXME: If info_ptr is non-null, we should validate it */
			MPIR_ERRTEST_ARGNULL(port_name,L"port_name",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_open_port %I %p", info, port_name); }
	
	mpi_errno = MPID_Open_port(info_ptr, port_name);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_OPEN_PORT);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_open_port",
			L"**mpi_open_port %I %p", info, port_name);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Lookup_nameA
#undef MPI_Lookup_nameW
#define MPI_Lookup_nameA PMPI_Lookup_nameA
#define MPI_Lookup_nameW PMPI_Lookup_nameW

/* One of these routines needs to define the global handle.  Since
   Most routines will use lookup (if they use any of the name publishing
   interface at all), we place this in MPI_Lookup_name. 
*/
MPID_NS_Handle MPIR_Namepub = 0;
int MPIR_NS_Initialized = 0;

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Lookup_name

/*@
   MPI_Lookup_name - Lookup a port given a service name

   Input Parameters:
+ service_name - a service name (string) 
- info - implementation-specific information (handle) 


   Output Parameter:
.  port_name - a port name (string) 

Notes:
If the 'service_name' is found, MPI copies the associated value into 
'port_name'.  The maximum size string that may be supplied by the system is 
'MPI_MAX_PORT_NAME'. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INFO
.N MPI_ERR_OTHER
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Lookup_nameA(char *service_name, MPI_Info info, char *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Lookup_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wservice_name = NULL;
	wchar_t wport_name[MPI_MAX_PORT_NAME];
	wchar_t *wport_name_ptr = NULL;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_LOOKUP_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_LOOKUP_NAMEA);

	if (service_name != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(service_name, &wservice_name);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	if (port_name != NULL)
	{
		wport_name_ptr = wport_name;
	}

	mpi_errno = MPI_Lookup_nameW(wservice_name, info, wport_name_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (port_name != NULL)
	{
		port_name[0] = '\0';
		error = wcstombs_s(&num_converted, port_name, MPI_MAX_PORT_NAME, wport_name, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= MPI_MAX_PORT_NAME)
			{
				/* truncation error */
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"port_name", (int)num_converted);
				goto fn_fail;
			}
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
	}

fn_fail:
	if (wservice_name != NULL)
		MPIU_Free(wservice_name);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_LOOKUP_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Lookup_nameW(wchar_t *service_name, MPI_Info info, wchar_t *port_name)
{
	static const wchar_t FCNAME[] = L"MPI_Lookup_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Info *info_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_LOOKUP_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_LOOKUP_NAME);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO_OR_NULL(info, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate info_ptr (only if not null) */
			if (info_ptr) 
				MPID_Info_valid_ptr( info_ptr, mpi_errno );
			/* Validate character pointers */
			MPIR_ERRTEST_ARGNULL( service_name, L"service_name", mpi_errno );
			MPIR_ERRTEST_ARGNULL( port_name, L"port_name", mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_lookup_name %s %I %p", service_name, info, port_name); }
	
#   ifdef HAVE_NAMEPUB_SERVICE
	{
		if (!MPIR_NS_Initialized)
		{
			mpi_errno = MPID_NS_Create( info_ptr, &MPIR_Namepub );
			/* FIXME: change **fail to something more meaningful */
			MPIU_ERR_CHKANDJUMP((mpi_errno != MPI_SUCCESS), mpi_errno, MPI_ERR_OTHER, L"**fail");
			MPIR_NS_Initialized = 1;
			MPIR_Add_finalize( (int (*)(void*))MPID_NS_Free, &MPIR_Namepub, 9 );
		}

		mpi_errno = MPID_NS_Lookup( MPIR_Namepub, info_ptr, (const wchar_t *)service_name, port_name );
		/* FIXME: change **fail to something more meaningful */
		MPIU_ERR_CHKANDJUMP((mpi_errno != MPI_SUCCESS && MPIR_ERR_GET_CLASS(mpi_errno) != MPI_ERR_NAME), mpi_errno, MPI_ERR_OTHER, L"**fail");
	}
#   else
	{
		/* No name publishing service available */
		MPIU_ERR_SETANDJUMP(mpi_errno, MPI_ERR_OTHER, L"**nonamepub");
	}
#   endif
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_LOOKUP_NAME);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_lookup_name",
			L"**mpi_lookup_name %s %I %p", service_name, info, port_name);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_spawn_multipleA
#undef MPI_Comm_spawn_multipleW
#define MPI_Comm_spawn_multipleA PMPI_Comm_spawn_multipleA
#define MPI_Comm_spawn_multipleW PMPI_Comm_spawn_multipleW

/* Any internal routines can go here.  Make them static if possible */
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_spawn_multiple

/*@
   MPI_Comm_spawn_multiple - short description

   Input Parameters:
+ count - number of commands (positive integer, significant to MPI only at 
  root 
. array_of_commands - programs to be executed (array of strings, significant 
  only at root) 
. array_of_argv - arguments for commands (array of array of strings, 
  significant only at root) 
. array_of_maxprocs - maximum number of processes to start for each command 
 (array of integer, significant only at root) 
. array_of_info - info objects telling the runtime system where and how to 
  start processes (array of handles, significant only at root) 
. root - rank of process in which previous arguments are examined (integer) 
- comm - intracommunicator containing group of spawning processes (handle) 

  Output Parameters:
+ intercomm - intercommunicator between original group and newly spawned group
  (handle) 
- array_of_errcodes - one error code per process (array of integer) 

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
.N MPI_ERR_INFO
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_spawn_multipleA(int count, char *array_of_commands[], char* *array_of_argv[], int array_of_maxprocs[], MPI_Info array_of_info[], int root, MPI_Comm comm, MPI_Comm *intercomm, int array_of_errcodes[]) 
{
	static const wchar_t FCNAME[] = L"MPI_Comm_spawn_multipleA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t **warray_of_commands = NULL, ***warray_of_argv = NULL;
	int i, j, k;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SPAWN_MULTIPLEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SPAWN_MULTIPLEA);

	if (array_of_commands != NULL)
	{
		warray_of_commands = (wchar_t**)MPIU_Malloc(count * sizeof(wchar_t*));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, warray_of_commands, count * sizeof(wchar_t*), L"warray_of_commands");
		for (i=0; i<count; i++) warray_of_commands[i] = NULL;
		for (i=0; i<count; i++)
		{
			if (array_of_commands[i] != NULL)
			{
				mpi_errno = MPIU_Mbstowcs(array_of_commands[i], &warray_of_commands[i]);
				MPIU_ERR_CHK_POP_ARRAY_PTR(mpi_errno, L"array_of_commands", i, array_of_commands[i]);
			}
		}
	}

	if (array_of_argv != NULL)
	{
		warray_of_argv = (wchar_t***)MPIU_Malloc(count * sizeof(wchar_t**));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, warray_of_argv, count * sizeof(wchar_t**), L"warray_of_argv");
		for (i=0; i<count; i++) warray_of_argv[i] = NULL;
		for (i=0; i<count; i++)
		{
			char **iter;
			j=1;
			if (array_of_argv[i] != NULL)
			{
				iter = &array_of_argv[i][0];
				while (*iter != NULL)
				{
					j++;
					iter++;
				}
				warray_of_argv[i] = (wchar_t**)MPIU_Malloc(j*sizeof(wchar_t*));
				MPIU_ERR_MEM_CHK_POP(mpi_errno, warray_of_argv[i], j*sizeof(wchar_t*), L"warray_of_argv[i]");
				for (k=0; k<j; k++) warray_of_argv[i][k] = NULL;
				for (k=0; k<j; k++)
				{
					if (array_of_argv[i][k] != NULL)
					{
						mpi_errno = MPIU_Mbstowcs(array_of_argv[i][k], &warray_of_argv[i][k]);
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**argarray", L"**argarray %s %d %p", L"array_of_argv", i, array_of_argv[i]);
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**argarray", L"**argarray %s %d %p", L"array_of_argv[i]", k, array_of_argv[i][k]);
							goto fn_fail;
						}
					}
				}
			}
		}
	}

	mpi_errno = MPI_Comm_spawn_multipleW(count, warray_of_commands, warray_of_argv, array_of_maxprocs, array_of_info, root, comm, intercomm, array_of_errcodes);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (warray_of_commands != NULL)
	{
		for (i=0; i<count; i++)
		{
			if (warray_of_commands[i] != NULL)
				MPIU_Free(warray_of_commands[i]);
		}
		MPIU_Free(warray_of_commands);
	}
	if (warray_of_argv != NULL)
	{
		for (i=0; i<count; i++)
		{
			wchar_t **iter;
			if (warray_of_argv[i] != NULL)
			{
				iter = &warray_of_argv[i][0];
				while (*iter != NULL)
				{
					MPIU_Free(*iter);
					iter++;
				}
				if (warray_of_argv[i] != NULL)
				{
					MPIU_Free(warray_of_argv[i]);
				}
			}
		}
		MPIU_Free(warray_of_argv);
	}
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SPAWN_MULTIPLEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Comm_spawn_multipleW(int count, wchar_t *array_of_commands[], wchar_t* *array_of_argv[], int array_of_maxprocs[], MPI_Info array_of_info[], int root, MPI_Comm comm, MPI_Comm *intercomm, int array_of_errcodes[]) 
{
	static const wchar_t FCNAME[] = L"MPI_Comm_spawn_multiple";
	int mpi_errno = MPI_SUCCESS, i;
	MPID_Comm *comm_ptr = NULL;
	MPID_Comm *intercomm_ptr = NULL;
	MPID_Info **array_of_info_ptrs;
	MPIU_CHKLMEM_DECL(1);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SPAWN_MULTIPLE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SPAWN_MULTIPLE);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			for (i = 0; i < count; i++)
			{
				MPIR_ERRTEST_INFO_OR_NULL(array_of_info[i], mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_info", i, array_of_info[i]);
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_spawn_multiple %d %p %p %p %p %d %C %p %p", count, array_of_commands, array_of_argv, array_of_maxprocs,
		array_of_info, root, comm, intercomm, array_of_errcodes); }
	
	MPIU_CHKLMEM_MALLOC(array_of_info_ptrs, MPID_Info **, count * sizeof(MPID_Info*), mpi_errno, L"array of info pointers");
	for (i=0; i<count; i++)
	{
		MPID_Info_get_ptr(array_of_info[i], array_of_info_ptrs[i]);
	}

	/* TODO: add error check to see if this collective function is
	   being called from multiple threads. */
	mpi_errno = MPID_Comm_spawn_multiple(count, array_of_commands, array_of_argv, array_of_maxprocs, array_of_info_ptrs, root, comm_ptr, &intercomm_ptr, array_of_errcodes);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	*intercomm = intercomm_ptr->handle;

	/* ... end of body of routine ... */

  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SPAWN_MULTIPLE);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_spawn_multiple",
			L"**mpi_comm_spawn_multiple %d %p %p %p %p %d %C %p %p", count, array_of_commands, array_of_argv, array_of_maxprocs,
			array_of_info, root, comm, intercomm, array_of_errcodes);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_spawnA
#undef MPI_Comm_spawnW
#define MPI_Comm_spawnA PMPI_Comm_spawnA
#define MPI_Comm_spawnW PMPI_Comm_spawnW

/* Any internal routines can go here.  Make them static if possible */
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_spawn

/*@
   MPI_Comm_spawn - Spawn up to maxprocs instances of a single MPI application

   Input Parameters:
+ command - name of program to be spawned (string, significant only at root) 
. argv - arguments to command (array of strings, significant only at root) 
. maxprocs - maximum number of processes to start (integer, significant only 
  at root) 
. info - a set of key-value pairs telling the runtime system where and how 
   to start the processes (handle, significant only at root) 
. root - rank of process in which previous arguments are examined (integer) 
- comm - intracommunicator containing group of spawning processes (handle) 

   Output Parameters:
+ intercomm - intercommunicator between original group and the 
   newly spawned group (handle) 
- array_of_errcodes - one code per process (array of integer) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
.N MPI_ERR_INFO
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_spawnA(char *command, char *argv[], int maxprocs, MPI_Info info, int root, MPI_Comm comm, MPI_Comm *intercomm, int array_of_errcodes[])
{
	static const wchar_t FCNAME[] = L"MPI_Comm_spawnA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wcommand = NULL, **wargv = NULL;
	int i, j;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SPAWNA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SPAWNA);

	if (command != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(command, &wcommand);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	if (argv != NULL)
	{
		char **iter;
		iter = &argv[0];
		j = 1;
		while (*iter != NULL)
		{
			j++;
			iter++;
		}
		wargv = (wchar_t**)MPIU_Malloc(j*sizeof(wchar_t*));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, wargv, j*sizeof(wchar_t*), "wargv");
		for (i=0; i<j; i++) wargv[i] = NULL;
		for (i=0; i<j; i++)
		{
			if (argv[i] != NULL)
			{
				mpi_errno = MPIU_Mbstowcs(argv[i], &wargv[i]);
				MPIU_ERR_CHK_POP_ARRAY_PTR(mpi_errno, L"argv", i, argv[i]);
			}
		}
	}

	mpi_errno = MPI_Comm_spawnW(wcommand, wargv, maxprocs, info, root, comm, intercomm, array_of_errcodes);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wcommand != NULL)
	{
		MPIU_Free(wcommand);
	}
	if (wargv != NULL)
	{
		wchar_t **iter;
		iter = &wargv[0];
		while (*iter != NULL)
		{
			MPIU_Free(*iter);
			iter++;
		}
		MPIU_Free(wargv);
	}
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SPAWNA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Comm_spawnW(wchar_t *command, wchar_t *argv[], int maxprocs, MPI_Info info, int root, MPI_Comm comm, MPI_Comm *intercomm, int array_of_errcodes[])
{
	static const wchar_t FCNAME[] = L"MPI_Comm_spawn";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL, *intercomm_ptr;
	MPID_Info *info_ptr=NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SPAWN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SPAWN);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			MPIR_ERRTEST_INFO_OR_NULL(info, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Info_get_ptr( info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_spawn %s %p %d %I %d %C %p %p", command, argv, maxprocs, info, root, comm, intercomm, array_of_errcodes); }
	
	/* TODO: add error check to see if this collective function is being called from multiple threads. */
	mpi_errno = MPID_Comm_spawn_multiple(1, &command, &argv, &maxprocs, &info_ptr, root, comm_ptr, &intercomm_ptr, array_of_errcodes); 
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	*intercomm = intercomm_ptr->handle;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SPAWN);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_spawn",
			L"**mpi_comm_spawn %s %p %d %I %d %C %p %p", command, argv, maxprocs, info, root, comm, intercomm, array_of_errcodes);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Prototypes for local functions */
PMPI_LOCAL int MPIR_fd_send(int, void *, int);
PMPI_LOCAL int MPIR_fd_recv(int, void *, int);

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_join
#define MPI_Comm_join PMPI_Comm_join

#ifdef HAVE_ERRNO_H
#include <errno.h> /* needed for read/write error codes */
#endif

#define SOCKET_EINTR			WSAEINTR

#undef FUNCNAME
#define FUNCNAME MPIR_fd_send
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
PMPI_LOCAL int MPIR_fd_send(int fd, void *buffer, int length)
{
	int result, num_bytes;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_FD_SEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_FD_SEND);

	while (length)
	{
		num_bytes = send(fd, buffer, length, 0);
		/* --BEGIN ERROR HANDLING-- */
		if (num_bytes == -1)
		{
			result = WSAGetLastError();
			if (result == SOCKET_EINTR)
				continue;
			else
			{
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_FD_SEND);
				return result;
			}
		}
		/* --END ERROR HANDLING-- */
		else
		{
			length -= num_bytes;
			buffer = (char*)buffer + num_bytes;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_FD_SEND);
	return 0;
}

#undef FUNCNAME
#define FUNCNAME MPIR_fd_recv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
PMPI_LOCAL int MPIR_fd_recv(int fd, void *buffer, int length)
{
	int result, num_bytes;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_FD_RECV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_FD_RECV);

	while (length)
	{
		num_bytes = recv(fd, buffer, length, 0);
		/* --BEGIN ERROR HANDLING-- */
		if (num_bytes == -1)
		{
			result = WSAGetLastError();
			if (result == SOCKET_EINTR)
				continue;
			else
			{
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_FD_RECV);
				return result;
			}
		}
		/* --END ERROR HANDLING-- */
		else
		{
			length -= num_bytes;
			buffer = (char*)buffer + num_bytes;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_FD_RECV);
	return 0;
}

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_join

/*@
   MPI_Comm_join - Create a communicator by joining two processes connected by 
	 a socket.

   Input Parameter:
. fd - socket file descriptor 

   Output Parameter:
. intercomm - new intercommunicator (handle) 

 Notes:
  The socket must be quiescent before 'MPI_COMM_JOIN' is called and after 
  'MPI_COMM_JOIN' returns. More specifically, on entry to 'MPI_COMM_JOIN', a 
  read on the socket will not read any data that was written to the socket 
  before the remote process called 'MPI_COMM_JOIN'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_join(int fd, MPI_Comm *intercomm)
{
#undef FCNAME
	static const wchar_t FCNAME[] = L"MPI_Comm_join";
	int mpi_errno = MPI_SUCCESS, err;
	wchar_t *local_port, *remote_port;
	//MPIU_THREADPRIV_DECL;
	MPIU_CHKLMEM_DECL(2);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_JOIN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_JOIN);

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_join %d %p", fd, intercomm); }

	//MPIU_THREADPRIV_GET;

	MPIU_CHKLMEM_MALLOC(local_port, wchar_t *, MPI_MAX_PORT_NAME*sizeof(wchar_t), mpi_errno, L"local port name");
	MPIU_CHKLMEM_MALLOC(remote_port, wchar_t *, MPI_MAX_PORT_NAME*sizeof(wchar_t), mpi_errno, L"remote port name");
	
	NMPI_Open_port(mpi_errno, (MPI_INFO_NULL, local_port));
	MPIU_ERR_CHKANDJUMP((mpi_errno != MPI_SUCCESS), mpi_errno, MPI_ERR_OTHER, L"**openportfailed");

	err = MPIR_fd_send(fd, local_port, MPI_MAX_PORT_NAME*sizeof(wchar_t));
	MPIU_ERR_CHKANDJUMP1((err != 0), mpi_errno, MPI_ERR_INTERN, L"**join_send", L"**join_send %d", err);

	err = MPIR_fd_recv(fd, remote_port, MPI_MAX_PORT_NAME*sizeof(wchar_t));
	MPIU_ERR_CHKANDJUMP1((err != 0), mpi_errno, MPI_ERR_INTERN, L"**join_recv", L"**join_recv %d", err);

	MPIU_ERR_CHKANDJUMP2((wcscmp(local_port, remote_port) == 0), mpi_errno, MPI_ERR_INTERN, L"**join_portname", L"**join_portname %s %s", local_port, remote_port);

	if (wcscmp(local_port, remote_port) < 0)
	{
		NMPI_Comm_accept(mpi_errno, (local_port, MPI_INFO_NULL, 0, MPI_COMM_SELF, intercomm));
	}
	else
	{
		NMPI_Comm_connect(mpi_errno, (remote_port, MPI_INFO_NULL, 0, MPI_COMM_SELF, intercomm));
	}
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	NMPI_Close_port(mpi_errno, (local_port));
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_JOIN);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_join",
			L"**mpi_comm_join %d %p", fd, intercomm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_get_parent
#define MPI_Comm_get_parent PMPI_Comm_get_parent

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_get_parent

/*@
   MPI_Comm_get_parent - Return the parent communicator for this process

   Output Parameter:
. parent - the parent communicator (handle) 

   Notes:

 If a process was started with 'MPI_Comm_spawn' or 'MPI_Comm_spawn_multiple', 
 'MPI_Comm_get_parent' returns the parent intercommunicator of the current 
  process. This parent intercommunicator is created implicitly inside of 
 'MPI_Init' and is the same intercommunicator returned by 'MPI_Comm_spawn'
  in the parents. 

  If the process was not spawned, 'MPI_Comm_get_parent' returns 
  'MPI_COMM_NULL'.

  After the parent communicator is freed or disconnected, 'MPI_Comm_get_parent'
  returns 'MPI_COMM_NULL'. 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_parent(MPI_Comm *parent)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_get_parent";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_GET_PARENT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_GET_PARENT);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(parent, L"parent", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_get_parent", L"**mpi_comm_get_parent %p", parent); }
	
	/* Note that MPIU_DBG_OpenFile also uses this code (so as to avoid
	   calling an MPI routine while logging it */
	*parent = (MPIR_Process.comm_parent == NULL) ? MPI_COMM_NULL : (MPIR_Process.comm_parent)->handle;  

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_GET_PARENT);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_comm_get_parent", L"**mpi_comm_get_parent %p", parent);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_disconnect
#define MPI_Comm_disconnect PMPI_Comm_disconnect

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_disconnect

/*@
   MPI_Comm_disconnect - Disconnect from a communicator

   Input Parameter:
.  comm - communicator (handle) 

Notes:
This routine waits for all pending communication to complete, then frees the
communicator and sets 'comm' to 'MPI_COMM_NULL'.  It may not be called 
with 'MPI_COMM_WORLD' or 'MPI_COMM_SELF'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS

.seealso MPI_Comm_connect, MPI_Comm_join
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Comm_disconnect(MPI_Comm * comm)
{
	static const wchar_t FCNAME[] = L"MPI_Comm_disconnect";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_DISCONNECT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("spawn");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_DISCONNECT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(*comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( *comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno)
			{
				goto fn_fail;
			}
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_comm_disconnect %C", *comm); }
	
	/*
	 * Since outstanding I/O bumps the reference count on the communicator, 
	 * we wait until we hold the last reference count to
	 * ensure that all communication has completed.  The reference count
	 * is 1 when the communicator is created, and it is incremented
	 * only for pending communication operations (and decremented when
	 * those complete).
	 */
	if (comm_ptr->ref_count > 1)
	{
		MPID_Progress_state progress_state;
		
		MPID_Progress_start(&progress_state);
		while (comm_ptr->ref_count > 1)
		{
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
		}
		MPID_Progress_end(&progress_state);
	}
	
	mpi_errno = MPID_Comm_disconnect(comm_ptr);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	
	*comm = MPI_COMM_NULL;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_DISCONNECT);
	MPIU_THREAD_SINGLE_CS_EXIT("spawn");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_comm_disconnect",
			L"**mpi_comm_disconnect %C", *comm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
