/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#include "topo.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Topo_test
#define MPI_Topo_test PMPI_Topo_test

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Topo_test

/*@

MPI_Topo_test - Determines the type of topology (if any) associated with a 
				communicator

Input Parameter:
. comm - communicator (handle) 

Output Parameter:
. top_type - topology type of communicator 'comm' (integer).  If the 
  communicator has no associated topology, returns 'MPI_UNDEFINED'.

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG

.seealso: MPI_Graph_create, MPI_Cart_create
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Topo_test(MPI_Comm comm, int *topo_type)
{
	static const wchar_t FCNAME[] = L"MPI_Topo_test";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *topo_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TOPO_TEST);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TOPO_TEST);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(topo_type, L"topo_type", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_topo_test %C %p", comm, topo_type); }
	
	topo_ptr = MPIR_Topology_get( comm_ptr );
	if (topo_ptr)
	{ 
		*topo_type = (int)(topo_ptr->kind);
	}
	else
	{ 
		*topo_type = MPI_UNDEFINED;
	}
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TOPO_TEST);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_topo_test",
			L"**mpi_topo_test %C %p", comm, topo_type);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Graph_neighbors_count
#define MPI_Graph_neighbors_count PMPI_Graph_neighbors_count

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Graph_neighbors_count

/*@
MPI_Graph_neighbors_count - Returns the number of neighbors of a node
							associated with a graph topology

Input Parameters:
+ comm - communicator with graph topology (handle) 
- rank - rank of process in group of 'comm' (integer) 

Output Parameter:
. nneighbors - number of neighbors of specified process (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
.N MPI_ERR_RANK
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Graph_neighbors_count(MPI_Comm comm, int rank, int *nneighbors)
{
	static const wchar_t FCNAME[] = L"MPI_Graph_neighbors_count";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *graph_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GRAPH_NEIGHBORS_COUNT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("topo");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GRAPH_NEIGHBORS_COUNT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			MPIR_ERRTEST_ARGNULL(nneighbors, L"nneighbors", mpi_errno);

			/* If comm_ptr is not value, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_graph_neighbors_count %C %d %p", comm, rank, nneighbors); }
	
	graph_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!graph_ptr || graph_ptr->kind != MPI_GRAPH), mpi_errno, MPI_ERR_TOPOLOGY, L"**notgraphtopo");
	MPIU_ERR_CHKANDJUMP2((rank < 0 || rank >= graph_ptr->topo.graph.nnodes), mpi_errno, MPI_ERR_RANK, L"**rank", L"**rank %d %d", rank, graph_ptr->topo.graph.nnodes );

	if ( rank == 0 ) 
		*nneighbors = graph_ptr->topo.graph.index[rank];
	else
		*nneighbors = graph_ptr->topo.graph.index[rank] - 
			graph_ptr->topo.graph.index[rank-1];

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GRAPH_NEIGHBORS_COUNT);
	MPIU_THREAD_SINGLE_CS_EXIT("topo");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_graph_neighbors_count",
			L"**mpi_graph_neighbors_count %C %d %p", comm, rank, nneighbors);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Graphdims_get
#define MPI_Graphdims_get PMPI_Graphdims_get

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Graphdims_get

/*@

MPI_Graphdims_get - Retrieves graph topology information associated with a 
					communicator

Input Parameter:
. comm - communicator for group with graph structure (handle) 

Output Parameters:
+ nnodes - number of nodes in graph (integer) 
- nedges - number of edges in graph (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Graphdims_get(MPI_Comm comm, int *nnodes, int *nedges)
{
	static const wchar_t FCNAME[] = L"MPI_Graphdims_get";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *topo_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GRAPHDIMS_GET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GRAPHDIMS_GET);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(nnodes, L"nnodes", mpi_errno );
			MPIR_ERRTEST_ARGNULL(nedges, L"nedges", mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_graphdims_get %C %p %p", comm, nnodes, nedges); }
	
	topo_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!topo_ptr || topo_ptr->kind != MPI_GRAPH), mpi_errno, MPI_ERR_TOPOLOGY, L"**notgraphtopo");

	/* Set nnodes */
	*nnodes = topo_ptr->topo.graph.nnodes;

	/* Set nedges */
	*nedges = topo_ptr->topo.graph.nedges;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GRAPHDIMS_GET);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_graphdims_get",
			L"**mpi_graphdims_get %C %p %p", comm, nnodes, nedges);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Graph_create
#define MPI_Graph_create PMPI_Graph_create

#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef FUNCNAME
#define FUNCNAME MPIR_Graph_create
int MPIR_Graph_create( const MPID_Comm *comm_ptr, int nnodes, const int index[], const int edges[], int reorder, MPI_Comm *comm_graph)
{
	static const wchar_t FCNAME[] = L"MPIR_Graph_create";
	int mpi_errno = MPI_SUCCESS;
	int i, nedges;
	MPID_Comm *newcomm_ptr = NULL;
	MPIR_Topology *graph_ptr = NULL;
	MPIU_CHKPMEM_DECL(3);
	MPIDI_STATE_DECL(MPID_STATE_MPIR_GRAPH_CREATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_GRAPH_CREATE);

	/* Create a new communicator */
	if (reorder)
	{
		int nrank;
		MPI_Comm ncomm;
		MPIU_THREADPRIV_DECL;

		MPIU_THREADPRIV_GET;

		/* Allow the cart map routine to remap the assignment of ranks to 
		   processes */
		NMPI_Graph_map(mpi_errno, ( comm_ptr->handle, nnodes, (int *)index, (int *)edges, &nrank ));
		/* Create the new communicator with split, since we need to reorder
		   the ranks (including the related internals, such as the connection
		   tables */
		if (mpi_errno == 0)
		{
			NMPI_Comm_split(mpi_errno, ( comm_ptr->handle, nrank == MPI_UNDEFINED ? MPI_UNDEFINED : 1, nrank, &ncomm ));
			if (!mpi_errno)
			{
				MPID_Comm_get_ptr( ncomm, newcomm_ptr );
			}
		}
	}
	else
	{
		/* Just use the first nnodes processes in the communicator */
		mpi_errno = MPIR_Comm_copy( (MPID_Comm *)comm_ptr, nnodes, &newcomm_ptr );
	}
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;


	/* If this process is not in the resulting communicator, return a null communicator and exit */
	if (!newcomm_ptr)
	{
		*comm_graph = MPI_COMM_NULL;
		goto fn_exit;
	}

	nedges = index[nnodes-1];
	MPIU_CHKPMEM_MALLOC(graph_ptr,MPIR_Topology*,sizeof(MPIR_Topology), mpi_errno,L"graph_ptr");
	
	graph_ptr->kind = MPI_GRAPH;
	graph_ptr->topo.graph.nnodes = nnodes;
	graph_ptr->topo.graph.nedges = nedges;
	MPIU_CHKPMEM_MALLOC(graph_ptr->topo.graph.index,int*, nnodes*sizeof(int),mpi_errno,L"graph.index");
	MPIU_CHKPMEM_MALLOC(graph_ptr->topo.graph.edges,int*, nedges*sizeof(int),mpi_errno,L"graph.edges");
	for (i=0; i<nnodes; i++) 
		graph_ptr->topo.graph.index[i] = index[i];
	for (i=0; i<nedges; i++) 
		graph_ptr->topo.graph.edges[i] = edges[i];

	/* Finally, place the topology onto the new communicator and return the handle */
	mpi_errno = MPIR_Topology_put( newcomm_ptr, graph_ptr );
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	
	*comm_graph = newcomm_ptr->handle;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_GRAPH_CREATE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	MPIU_CHKPMEM_REAP();
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_graph_create",
			L"**mpi_graph_create %C %d %p %p %d %p", comm_ptr->handle, 
			nnodes, index, 
			edges, reorder, comm_graph);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( (MPID_Comm*)comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Graph_create

/*@
MPI_Graph_create - Makes a new communicator to which topology information
				 has been attached

Input Parameters:
+ comm_old - input communicator without topology (handle) 
. nnodes - number of nodes in graph (integer) 
. index - array of integers describing node degrees (see below) 
. edges - array of integers describing graph edges (see below) 
- reorder - ranking may be reordered (true) or not (false) (logical) 

Output Parameter:
. comm_graph - communicator with graph topology added (handle) 

Notes:
Each process must provide a description of the entire graph, not just the
neigbors of the calling process.  

Algorithm:
We ignore the 'reorder' info currently.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG

@*/
MPI_DLL_SPEC int MPI_CALL MPI_Graph_create(MPI_Comm comm_old, int nnodes, int *index, int *edges, int reorder, MPI_Comm *comm_graph)
{
	static const wchar_t FCNAME[] = L"MPI_Graph_create";
	int mpi_errno = MPI_SUCCESS;
	int i;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GRAPH_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("topo");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GRAPH_CREATE);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm_old, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm_old, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (comm_ptr)
			{
				MPIR_ERRTEST_COMM_INTRA(comm_ptr,mpi_errno);
			}
			MPIR_ERRTEST_ARGNEG(nnodes,L"nnodes",mpi_errno);
			if (nnodes > 0)
			{
				MPIR_ERRTEST_ARGNULL(index,L"index",mpi_errno);
				MPIR_ERRTEST_ARGNULL(edges,L"edges",mpi_errno);
			}
			MPIR_ERRTEST_ARGNULL(comm_graph,L"comm_graph",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			int j;
			int comm_size = comm_ptr->remote_size;

			/* Check that the communicator is large enough */
			if (nnodes > comm_size)
			{
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**topotoolarge", L"**topotoolarge %d %d", nnodes, comm_size );
			}
			
			/* Perform the remaining tests only if nnodes is valid.  
			   This avoids SEGVs from accessing invalid parts of the
			   edges or index arrays */
			if (mpi_errno) goto fn_fail;
			
			/* Check that index is monotone nondecreasing */
			/* Use ERR_ARG instead of ERR_TOPOLOGY because there is no
			   topology yet */
			for (i=0; i<nnodes; i++)
			{
				if (index[i] < 0)
				{
					mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**indexneg", L"**indexneg %d %d", i, index[i] );
				}
				if (i+1<nnodes && index[i] > index[i+1])
				{
					mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**indexnonmonotone", L"**indexnonmonotone %d %d %d", i, index[i], index[i+1] );
				}
			}

			/* Check that edge number is in range. Note that the 
			   edges refer to a rank in the communicator, and can 
			   be greater than nnodes */
			if (nnodes > 0)
			{
				for (i=0; i<index[nnodes-1]; i++)
				{
					if (edges[i] > comm_size || edges[i] < 0)
					{
						mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**edgeoutrange", L"**edgeoutrange %d %d %d", i, edges[i], comm_size );
					}
				}
			}
			/* We could also check that no edge is from a node to itself.
			   This gives us an excuse to run over the entire arrays. 
			   This test could be combined with the above to make the code
			   shorter.
			*/
			if (!mpi_errno)
			{
				j = 0;
				for (i=0; i<nnodes && !mpi_errno; i++)
				{
					for (;j<index[i]; j++)
					{
						if (edges[j] == i)
						{
							mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**nulledge", L"**nulledge %d %d", i, j );
							break;
						}
					}
				}
			}
			
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_graph_create %C %d %p %p %d %p", comm_old, nnodes, index, edges, reorder, comm_graph); }
	
	/* Test for empty communicator */
	if (nnodes == 0)
	{
		*comm_graph = MPI_COMM_NULL;
		goto fn_exit;
	}

	if (comm_ptr->topo_fns != NULL && 
		comm_ptr->topo_fns->graphCreate != NULL)
	{
		mpi_errno = comm_ptr->topo_fns->graphCreate( comm_ptr, nnodes, (const int *)index, (const int *)edges, reorder, comm_graph );
	}		
	else
	{
		mpi_errno = MPIR_Graph_create( comm_ptr, nnodes, (const int *)index, (const int *)edges, reorder, comm_graph );
	}
	if (mpi_errno) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GRAPH_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT("topo");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_graph_create",
			L"**mpi_graph_create %C %d %p %p %d %p", comm_old, nnodes, index, 
			edges, reorder, comm_graph);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Graph_neighbors
#define MPI_Graph_neighbors PMPI_Graph_neighbors

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Graph_neighbors

/*@
 MPI_Graph_neighbors - Returns the neighbors of a node associated 
					   with a graph topology

Input Parameters:
+ comm - communicator with graph topology (handle) 
. rank - rank of process in group of comm (integer) 
- maxneighbors - size of array neighbors (integer) 

Output Parameters:
. neighbors - ranks of processes that are neighbors to specified process
 (array of integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
.N MPI_ERR_RANK
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Graph_neighbors(MPI_Comm comm, int rank, int maxneighbors, int *neighbors)
{
	static const wchar_t FCNAME[] = L"MPI_Graph_neighbors";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *graph_ptr;
	int i, is, ie;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GRAPH_NEIGHBORS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GRAPH_NEIGHBORS);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(neighbors,L"neighbors",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_graph_neighbors %C %d %d %p", comm, rank, maxneighbors, neighbors); }
	
	graph_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!graph_ptr || graph_ptr->kind != MPI_GRAPH), mpi_errno, MPI_ERR_TOPOLOGY, L"**notgraphtopo");
	MPIU_ERR_CHKANDJUMP2((rank < 0 || rank >= graph_ptr->topo.graph.nnodes), mpi_errno, MPI_ERR_RANK, L"**rank", L"**rank %d %d", rank, graph_ptr->topo.graph.nnodes );

	/* Get location in edges array of the neighbors of the specified rank */
	if (rank == 0) is = 0;
	else		   is = graph_ptr->topo.graph.index[rank-1];
	ie = graph_ptr->topo.graph.index[rank];

	/* Get neighbors */
	for (i=is; i<ie; i++) 
		*neighbors++ = graph_ptr->topo.graph.edges[i];
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GRAPH_NEIGHBORS);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_graph_neighbors",
			L"**mpi_graph_neighbors %C %d %d %p", comm, rank, maxneighbors, neighbors);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Graph_map
#define MPI_Graph_map PMPI_Graph_map
#undef FUNCNAME
#define FUNCNAME MPIR_Graph_map
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Graph_map( const MPID_Comm *comm_ptr, int nnodes, const int index[], const int edges[], int *newrank )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIR_GRAPH_MAP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_GRAPH_MAP);
	MPIU_UNREFERENCED_ARG(index);
	MPIU_UNREFERENCED_ARG(edges);

	/* This is the trivial version that does not remap any processes. */
	if (comm_ptr->rank < nnodes)
	{
		*newrank = comm_ptr->rank;
	}
	else
	{
		*newrank = MPI_UNDEFINED;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_GRAPH_MAP);
	return MPI_SUCCESS;
}
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Graph_map

/*@
MPI_Graph_map - Maps process to graph topology information

Input Parameters:
+ comm - input communicator (handle) 
. nnodes - number of graph nodes (integer) 
. index - integer array specifying the graph structure, see 'MPI_GRAPH_CREATE' 
- edges - integer array specifying the graph structure 

Output Parameter:
. newrank - reordered rank of the calling process; 'MPI_UNDEFINED' if the 
calling process does not belong to graph (integer) 

.N SignalSafe
 
.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Graph_map(MPI_Comm comm_old, int nnodes, int *index, int *edges, int *newrank)
{
#undef FCNAME
	static const wchar_t FCNAME[] = L"MPI_Graph_map";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GRAPH_MAP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GRAPH_MAP);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm_old, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm_old, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(newrank,L"newrank",mpi_errno);
			MPIR_ERRTEST_ARGNULL(index,L"index",mpi_errno);
			MPIR_ERRTEST_ARGNULL(edges,L"edges",mpi_errno);
			MPIR_ERRTEST_ARGNONPOS(nnodes,L"nnodes",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_graph_map %C %d %p %p %p", comm_old, nnodes, index, edges, newrank); }
	
	MPIU_ERR_CHKANDJUMP(comm_ptr->local_size < nnodes,mpi_errno,MPI_ERR_ARG, L"**graphnnodes");
	
	if (comm_ptr->topo_fns != NULL && comm_ptr->topo_fns->graphMap != NULL)
	{
		mpi_errno = comm_ptr->topo_fns->graphMap( comm_ptr, nnodes, (const int*) index, (const int*) edges, newrank );
	}
	else
	{
		mpi_errno = MPIR_Graph_map( comm_ptr, nnodes, (const int*) index, (const int*) edges, newrank );
	}
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GRAPH_MAP);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_graph_map",
			L"**mpi_graph_map %C %d %p %p %p", comm_old, nnodes, index, edges, 
			newrank);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Graph_get
#define MPI_Graph_get PMPI_Graph_get

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Graph_get

/*@

MPI_Graph_get - Retrieves graph topology information associated with a communicator

Input Parameters:
+ comm - communicator with graph structure (handle) 
. maxindex - length of vector 'index' in the calling program  (integer) 
- maxedges - length of vector 'edges' in the calling program  (integer) 

Output Parameters:
+ index - array of integers containing the graph structure (for details see the definition of 'MPI_GRAPH_CREATE') 
- edges - array of integers containing the graph structure 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Graph_get(MPI_Comm comm, int maxindex, int maxedges, int *index, int *edges)
{
	static const wchar_t FCNAME[] = L"MPI_Graph_get";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *topo_ptr;
	int i, n, *vals;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GRAPH_GET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GRAPH_GET);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			
			MPIR_ERRTEST_ARGNULL( edges, L"edges", mpi_errno );
			MPIR_ERRTEST_ARGNULL( index,  L"index", mpi_errno );

			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_graph_get %C %d %d %p %p", comm, maxindex, maxedges, index, edges); }
	
	topo_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!topo_ptr || topo_ptr->kind != MPI_GRAPH), mpi_errno, MPI_ERR_TOPOLOGY, L"**notgraphtopo");
	MPIU_ERR_CHKANDJUMP3((topo_ptr->topo.graph.nnodes > maxindex), mpi_errno, MPI_ERR_ARG, L"**argrange", L"**argrange %s %d %d", L"maxindex", maxindex, topo_ptr->topo.graph.nnodes);
	MPIU_ERR_CHKANDJUMP3((topo_ptr->topo.graph.nedges > maxedges), mpi_errno, MPI_ERR_ARG, L"**argrange", L"**argrange %s %d %d", L"maxedges", maxedges, topo_ptr->topo.graph.nedges);

	/* Get index */
	n = topo_ptr->topo.graph.nnodes;
	vals = topo_ptr->topo.graph.index;
	for ( i=0; i<n; i++ )
		*index++ = *vals++;

	/* Get edges */
	n = topo_ptr->topo.graph.nedges;
	vals = topo_ptr->topo.graph.edges;
	for ( i=0; i<n; i++ )
		*edges++ = *vals++;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GRAPH_GET);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_graph_get",
			L"**mpi_graph_get %C %d %d %p %p", comm, maxindex, maxedges, index, edges);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Because we store factors with their multiplicities, a small array can
   store all of the factors for a large number (grows *faster* than n factorial). */
#define MAX_FACTORS 10
/* 2^20 is a millon */
#define MAX_DIMS	20

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines.  You can use USE_WEAK_SYMBOLS to see if MPICH is
   using weak symbols to implement the MPI routines. */
typedef struct Factors { int val, cnt; } Factors;
/* This routine may be global if we are not using weak symbols */
PMPI_LOCAL int MPIR_Factor( int, Factors [], int * );
PMPI_LOCAL int MPIR_ChooseFactors( int, Factors [], int, int, int [] );

#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Dims_create
#define MPI_Dims_create PMPI_Dims_create

/* Return the factors of n and their multiplicity in factors; the number of 
   distinct factors is the return value and the total number of factors,
   including multiplicities, is returned in ndivisors */
#define NUM_PRIMES 168
  static int primes[NUM_PRIMES] = 
		   {2,	3,	5,	7,   11,   13,   17,   19,   23,   29, 
		   31,   37,   41,   43,   47,   53,   59,   61,   67,   71, 
		   73,   79,   83,   89,   97,  101,  103,  107,  109,  113, 
		  127,  131,  137,  139,  149,  151,  157,  163,  167,  173, 
		  179,  181,  191,  193,  197,  199,  211,  223,  227,  229, 
		  233,  239,  241,  251,  257,  263,  269,  271,  277,  281, 
		  283,  293,  307,  311,  313,  317,  331,  337,  347,  349, 
		  353,  359,  367,  373,  379,  383,  389,  397,  401,  409, 
		  419,  421,  431,  433,  439,  443,  449,  457,  461,  463, 
		  467,  479,  487,  491,  499,  503,  509,  521,  523,  541, 
		  547,  557,  563,  569,  571,  577,  587,  593,  599,  601, 
		  607,  613,  617,  619,  631,  641,  643,  647,  653,  659, 
		  661,  673,  677,  683,  691,  701,  709,  719,  727,  733, 
		  739,  743,  751,  757,  761,  769,  773,  787,  797,  809, 
		  811,  821,  823,  827,  829,  839,  853,  857,  859,  863, 
		  877,  881,  883,  887,  907,  911,  919,  929,  937,  941, 
		  947,  953,  967,  971,  977,  983,  991,  997};

#undef FUNCNAME
#define FUNCNAME MPIR_Factor
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
PMPI_LOCAL int MPIR_Factor( int n, Factors factors[], int *ndivisors )
{
	int n_tmp, n_root;
	int i, nfactors=0, nall=0;
	int cnt;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_FACTOR);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_FACTOR);

	/* Start from an approximate of the square root of n, by first finding
	   the power of 2 at least as large as n.  The approximate root is then
	   2 to the 1/2 this power */
	n_tmp  = n;
	n_root = 0;
	while (n_tmp)
	{
		n_root ++;
		n_tmp >>= 1;
	}
	n_root = 1 << (n_root / 2);

	/* Find the prime number that less than that value and try dividing out the primes.  */
	for (i=0; i<NUM_PRIMES; i++)
	{
		if (primes[i] > n_root) break;
	}

	/* For each prime, divide out as many as possible */
	for (;i>=0;i--)
	{
		cnt = 0;
		while ( (n %  primes[i]) == 0)
		{
			cnt ++;
			n = n / primes[i];
		}
		if (cnt > 0)
		{
			/* --BEGIN ERROR HANDLING-- */
			if (nfactors + 1 == MAX_FACTORS)
			{
				/* Time to panic.  This should not happen, since the
				   smallest number that could exceed this would
				   be the product of the first 10 primes that are
				   greater than one, which is 6469693230 */
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_FACTOR);
				return nfactors;
			}
			/* --END ERROR HANDLING-- */
			factors[nfactors].val = primes[i];
			factors[nfactors++].cnt = cnt;
			nall += cnt;
		}
	}
	/* If nfactors == 0, n was a prime, so return that */
	if (nfactors == 0)
	{
		nfactors = 1;
		nall = 1;
		factors[0].val = n;
		factors[0].cnt = 1;
	}
	else if (n > 1)
	{
		/* We need one more factor (a single prime > n_root) */
		factors[nfactors].val   = n;
		factors[nfactors++].cnt = 1;
		nall++;
	}
	*ndivisors = nall;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_FACTOR);
	return nfactors;
}

/* 
   Given a collection of factors from the factors routine and a number of
   required values, combine the elements in factors into "needed" elements
   of the array chosen.  These are non-increasing and so can be used directly
   in setting values in the dims array in MPIR_Dims_create.

   Algorithm (very simple)

   target_size = nnodes / ndims needed.
   Accumulate factors, starting from the bottom,
   until the target size is met or exceeded.
   Put all of the remaining factors into the last dimension
   (recompute target_size with each step, since we may
   miss the target by a wide margin.
   
   A much more sophisticated code would try to balance
   the number of nodes assigned to each dimension, possibly
   in concert with guidelines from the device about "good"
   sizes

 */
#undef FUNCNAME
#define FUNCNAME MPIR_ChooseFactors
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
PMPI_LOCAL int MPIR_ChooseFactors( int nfactors, Factors factors[], int nnodes, int needed, int chosen[] )
{
	int nodes_needed = nnodes;
	int target_size = nodes_needed / needed;
	int factor;
	int i, j;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_CHOOSEFACTORS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_CHOOSEFACTORS);

	/* First, distribute the factors into the chosen array */
	j = 0;
	for (i=0; i<needed; i++)
	{
		if (j >= nfactors) break;
		if (i == needed-1)
		{
			/* Dump all of the remaining factors into this entry */
			factor = 1;
			while (j < nfactors)
			{
				factor *= factors[j].val;
				if (--factors[j].cnt == 0) j++;
			}
		}
		else
		{
			/* Get the current target size */
			factor = 1;
			while (j < nfactors && factor < target_size)
			{
				factor *= factors[j].val;
				if (--factors[j].cnt == 0) j++;
			}
		}
		chosen[i] = factor;
		nodes_needed /= factor;
		target_size = nodes_needed / (needed - i);
	}
	/* finish up */
	for (; i<needed; i++) chosen[i] = 1;

	/* Second, sort the chosen array in non-increasing order.  Use
	   a simple bubble sort because the number of elements is always small */
	for (i=0; i<needed-1; i++)
	{
		for (j=i+1; j<needed; j++)
		{
			if (chosen[j] > chosen[i])
			{
				int tmp = chosen[i];
				chosen[i] = chosen[j];
				chosen[j] = tmp;
			}
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_CHOOSEFACTORS);
	return 0;
}

#undef FUNCNAME
#define FUNCNAME MPIR_Dims_create
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Dims_create( int nnodes, int ndims, int *dims )
{
	Factors factors[MAX_FACTORS];
	int chosen[MAX_DIMS];
	int i, j, mpi_errno;
	int dims_needed, dims_product, nfactors, ndivisors;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_DIMS_CREATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_DIMS_CREATE);
	
	/* Find the number of unspecified dimensions in dims and the product of the positive values in dims */
	dims_needed  = 0;
	dims_product = 1;
	for (i=0; i<ndims; i++)
	{
		if (dims[i] < 0)
		{
			mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, L"MPIR_Dims_create", __LINE__, MPI_ERR_DIMS, L"**argarrayneg", L"**argarrayneg %s %d %d", L"dims", i, dims[i]);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIR_DIMS_CREATE);
			return mpi_errno;
		}
		if (dims[i] == 0) dims_needed ++;
		else dims_product *= dims[i];
	}

	/* Can we factor nnodes by dims_product? */
	if ((nnodes / dims_product ) * dims_product != nnodes )
	{
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, L"MPIR_Dims_create", __LINE__, MPI_ERR_DIMS, L"**dimspartition", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_DIMS_CREATE);
		return mpi_errno;
	}

	if (!dims_needed)
	{
		/* Special case - all dimensions provided */
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_DIMS_CREATE);
		return MPI_SUCCESS;
	}
	
	if (dims_needed > MAX_DIMS)
	{
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, L"MPIR_Dims_create", __LINE__,  MPI_ERR_DIMS, L"**dimsmany", L"**dimsmany %d %d", dims_needed, MAX_DIMS );
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_DIMS_CREATE);
		return mpi_errno;
		/* --END ERROR HANDLING-- */
	}

	nnodes /= dims_product;

	/* Now, factor nnodes into dims_needed components.  We'd like these
	   to match the underlying machine topology as much as possible.  In the
	   absence of information about the machine topology, we can try to 
	   make the factors a close to each other as possible.  

	   The MPICH 1 version used donated code that was quite sophisticated
	   and complex.  However, since it didn't take the system topology
	   into account, it was more sophisticated that was perhaps warranted.
	   In addition, useful values of nnodes for most MPI programs will be
	   of the order 10-10000, and powers of two will be common.
	*/

	/* Get the factors */
	nfactors = MPIR_Factor( nnodes, factors, &ndivisors );

	/* Divide into 3 major cases:
	   1. Fewer divisors than needed dimensions.  Just use all of the
		  factors up, setting the remaining dimensions to 1
	   2. Only one distinct factor (typically 2) but with greater 
		  multiplicity.  Give each dimension a nearly equal size
	   3. Other.  There are enough factors to divide among the dimensions.
		  This is done in an ad hoc fashion
	*/

/* DEBUG 
	wprintf( L"factors are (%d of them) with %d divisors\n", nfactors, ndivisors );
	for (j=0; j<nfactors; j++) {
		wprintf( L"val = %d repeated %d\n", factors[j].val, factors[j].cnt );
	}
*/
	/* The MPI spec requires that the values that are set be in nonincreasing
	   order (MPI-1, section 6.5).  */

	/* Distribute the factors among the dimensions */
	if (ndivisors <= dims_needed)
	{
		/* Just use the factors as needed.  */
		MPIR_ChooseFactors( nfactors, factors, nnodes, dims_needed, chosen );
		j = 0;
		for (i=0; i<ndims; i++)
		{
			if (dims[i] == 0)
			{
				dims[i] = chosen[j++];
			}
		}
#if 0
		/* Any remaining unset dims are set to one */
		for (i++;i<ndims; i++)
		{
			if (dims[i] == 0) 
				dims[i] = 1;
		}
#endif
	}
	else
	{
		/* We must combine some of the factors */
		/* This is what the fancy code is for in the MPICH-1 code.
		   If the number of distinct factors is 1 (e.g., a power of 2),
		   then this code can be much simpler */
		/* NOT DONE */
		/* FIXME */
		if (nfactors == 1)
		{
			/* Special case for k**n, such as powers of 2 */
			int factor = factors[0].val;
			int cnt	= factors[0].cnt; /* Numver of factors left */
			int cnteach = ( cnt + dims_needed - 1 ) / dims_needed;
			int factor_each;
			
			factor_each = factor;
			for (i=1; i<cnteach; i++) factor_each *= factor;

			for (i=0; i<ndims; i++)
			{
				if (dims[i] == 0)
				{
					if (cnt > cnteach)
					{
						dims[i] = factor_each;
						cnt -= cnteach;
					}
					else if (cnt > 0)
					{
						factor_each = factor;
						for (j=1; j<cnt; j++) 
							factor_each *= factor;
						dims[i] = factor_each;
						cnt = 0;
					}
					else
					{
						dims[i] = 1;
					}
				}
			}
		}			
		else
		{
			/* Here is the general case.  */
			MPIR_ChooseFactors( nfactors, factors, nnodes, dims_needed, chosen );
			j = 0;
			for (i=0; i<ndims; i++)
			{
				if (dims[i] == 0)
				{
					dims[i] = chosen[j++];
				}
			}
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_DIMS_CREATE);
	return MPI_SUCCESS;
}

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Dims_create

/*@
	MPI_Dims_create - Creates a division of processors in a cartesian grid

 Input Parameters:
+ nnodes - number of nodes in a grid (integer) 
- ndims - number of cartesian dimensions (integer) 

 In/Out Parameter:   
. dims - integer array of size  'ndims' specifying the number of nodes in each 
 dimension.  A value of 0 indicates that 'MPI_Dims_create' should fill in a
 suitable value.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Dims_create(int nnodes, int ndims, int *dims)
{
#undef FCNAME
	static const wchar_t FCNAME[] = L"MPI_Dims_create";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_DIMS_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_DIMS_CREATE);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNEG(nnodes,L"nnodes",mpi_errno);
			MPIR_ERRTEST_ARGNEG(ndims,L"ndims",mpi_errno);
			MPIR_ERRTEST_ARGNULL(dims,L"dims",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_dims_create %d %d %p", nnodes, ndims, dims); }

	if (MPIR_Process.dimsCreate != NULL)
	{
		mpi_errno = MPIR_Process.dimsCreate( nnodes, ndims, dims );
	}
	else
	{
		mpi_errno = MPIR_Dims_create( nnodes, ndims, dims );
	}
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_DIMS_CREATE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_dims_create",
			L"**mpi_dims_create %d %d %p", nnodes, ndims, dims);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Cartdim_get
#define MPI_Cartdim_get PMPI_Cartdim_get

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Cartdim_get

/*@

MPI_Cartdim_get - Retrieves Cartesian topology information associated with a 
				  communicator

Input Parameter:
. comm - communicator with cartesian structure (handle) 

Output Parameter:
. ndims - number of dimensions of the cartesian structure (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Cartdim_get(MPI_Comm comm, int *ndims)
{
	static const wchar_t FCNAME[] = L"MPI_Cartdim_get";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *cart_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CARTDIM_GET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CARTDIM_GET);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(ndims,L"ndims",mpi_errno);
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_cartdim_get %C %p", comm, ndims); }
	
	cart_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!cart_ptr || cart_ptr->kind != MPI_CART), mpi_errno, MPI_ERR_TOPOLOGY, L"**notcarttopo");

	*ndims = cart_ptr->topo.cart.ndims;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CARTDIM_GET);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_cartdim_get",
			L"**mpi_cartdim_get %C %p", comm, ndims);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Cart_sub
#define MPI_Cart_sub PMPI_Cart_sub

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Cart_sub

/*@

MPI_Cart_sub - Partitions a communicator into subgroups which 
			   form lower-dimensional cartesian subgrids

Input Parameters:
+ comm - communicator with cartesian structure (handle) 
- remain_dims - the  'i'th entry of remain_dims specifies whether the 'i'th 
dimension is kept in the subgrid (true) or is dropped (false) (logical 
vector) 

Output Parameter:
. newcomm - communicator containing the subgrid that includes the calling 
process (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Cart_sub(MPI_Comm comm, int *remain_dims, MPI_Comm *comm_new)
{
	static const wchar_t FCNAME[] = L"MPI_Cart_sub";
	int mpi_errno = MPI_SUCCESS;
	int ndims, key, color, ndims_in_subcomm, nnodes_in_subcomm, i, j, rank;
	MPID_Comm *comm_ptr = NULL, *newcomm_ptr;
	MPIR_Topology *topo_ptr, *toponew_ptr;
	MPIU_CHKPMEM_DECL(4);
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CART_SUB);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("topo");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CART_SUB);

	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_cart_sub %C %p %p", comm, remain_dims, comm_new); }
	
	/* Check that the communicator already has a Cartesian topology */
	topo_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP(!topo_ptr,mpi_errno,MPI_ERR_TOPOLOGY,L"**notopology");
	MPIU_ERR_CHKANDJUMP(topo_ptr->kind != MPI_CART,mpi_errno,MPI_ERR_TOPOLOGY, L"**notcarttopo");

	/* Determine the number of remaining dimensions */
	ndims = topo_ptr->topo.cart.ndims;
	ndims_in_subcomm = 0;
	nnodes_in_subcomm = 1;
	for (i=0; i<ndims; i++)
	{
		if (remain_dims[i])
		{
			ndims_in_subcomm ++;
			nnodes_in_subcomm *= topo_ptr->topo.cart.dims[i];
		}
	}

	/* Split this communicator.  Do this even if there are no remaining
	   dimensions so that the topology information is attached */
	key   = 0;
	color = 0;
	for (i=0; i<ndims; i++)
	{
		if (remain_dims[i])
		{
			key = (key * topo_ptr->topo.cart.dims[i]) + topo_ptr->topo.cart.position[i];
		}
		else
		{
			color = (color * topo_ptr->topo.cart.dims[i]) + topo_ptr->topo.cart.position[i];
		}
	}
	NMPI_Comm_split(mpi_errno, ( comm, color, key, comm_new ));
	if (mpi_errno) goto fn_fail;
	
	/* Save the topology of this new communicator */
	MPIU_CHKPMEM_MALLOC(toponew_ptr,MPIR_Topology*,sizeof(MPIR_Topology), mpi_errno,L"toponew_ptr");
		
	toponew_ptr->kind				  = MPI_CART;
	toponew_ptr->topo.cart.ndims  = ndims_in_subcomm;
	toponew_ptr->topo.cart.nnodes = nnodes_in_subcomm;
	if (ndims_in_subcomm)
	{
		MPIU_CHKPMEM_MALLOC(toponew_ptr->topo.cart.dims,int*, ndims_in_subcomm*sizeof(int),mpi_errno,L"cart.dims");
		MPIU_CHKPMEM_MALLOC(toponew_ptr->topo.cart.periodic,int*, ndims_in_subcomm*sizeof(int),mpi_errno,L"cart.periodic");
		MPIU_CHKPMEM_MALLOC(toponew_ptr->topo.cart.position,int*, ndims_in_subcomm*sizeof(int),mpi_errno,L"cart.position");
	}
	else
	{
		toponew_ptr->topo.cart.dims	 = 0;
		toponew_ptr->topo.cart.periodic = 0;
		toponew_ptr->topo.cart.position = 0;
	}

	j = 0;
	for (i=0; i<ndims; i++)
	{
		if (remain_dims[i])
		{
			toponew_ptr->topo.cart.dims[j] = topo_ptr->topo.cart.dims[i];
			toponew_ptr->topo.cart.periodic[j] = topo_ptr->topo.cart.periodic[i];
			j++;
		}
	}

	MPID_Comm_get_ptr( *comm_new, newcomm_ptr );
	/* Compute the position of this process in the new communicator */
	rank = newcomm_ptr->rank;
	for (i=0; i<ndims_in_subcomm; i++)
	{
		nnodes_in_subcomm /= toponew_ptr->topo.cart.dims[i];
		toponew_ptr->topo.cart.position[i] = rank / nnodes_in_subcomm;
		rank = rank % nnodes_in_subcomm;
	}

	mpi_errno = MPIR_Topology_put( newcomm_ptr, toponew_ptr );
	if (mpi_errno) goto fn_fail;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CART_SUB);
	MPIU_THREAD_SINGLE_CS_EXIT("topo");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	MPIU_CHKPMEM_REAP();
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_cart_sub",
			L"**mpi_cart_sub %C %p %p", comm, remain_dims, comm_new);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Cart_shift
#define MPI_Cart_shift PMPI_Cart_shift

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Cart_shift

/*@
MPI_Cart_shift - Returns the shifted source and destination ranks, given a 
				 shift direction and amount

Input Parameters:
+ comm - communicator with cartesian structure (handle) 
. direction - coordinate dimension of shift (integer) 
- displ - displacement (> 0: upwards shift, < 0: downwards shift) (integer) 

Output Parameters:
+ source - rank of source process (integer) 
- dest - rank of destination process (integer) 

Notes:
The 'direction' argument is in the range '[0,n-1]' for an n-dimensional 
Cartesian mesh.

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Cart_shift(MPI_Comm comm, int direction, int displ, int *source, int *dest)
{
	static const wchar_t FCNAME[] = L"MPI_Cart_shift";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *cart_ptr;
	int i;
	int pos[MAX_CART_DIM];
	int rank;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CART_SHIFT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CART_SHIFT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */

			MPIR_ERRTEST_ARGNULL( source, L"source", mpi_errno );
			MPIR_ERRTEST_ARGNULL( dest, L"dest", mpi_errno );
			MPIR_ERRTEST_ARGNEG( direction, L"direction", mpi_errno );
/* Nothing in the standard indicates that a zero displacement is not valid
			if (displ == 0) {
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG,
												  L"**cartshiftzero", 0 );
			}
*/
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_cart_shift %C %d %d %p %p", comm, direction, displ, source, dest); }
	
	cart_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!cart_ptr || cart_ptr->kind != MPI_CART), mpi_errno, MPI_ERR_TOPOLOGY, L"**notcarttopo");
	MPIU_ERR_CHKANDJUMP2((direction >= cart_ptr->topo.cart.ndims), mpi_errno, MPI_ERR_ARG, L"**dimsmany", L"**dimsmany %d %d", cart_ptr->topo.cart.ndims, direction);

	/* Check for the case of a 0 displacement */
	rank = comm_ptr->rank;
	if (displ == 0)
	{
		*source = *dest = rank;
	}
	else
	{
		MPIU_THREADPRIV_DECL;

		MPIU_THREADPRIV_GET;

		/* To support advanced implementations that support MPI_Cart_create,
		   we compute the new position and call PMPI_Cart_rank to get the
		   source and destination.  We could bypass that step if we know that
		   the mapping is trivial.  Copy the current position. */
		for (i=0; i<cart_ptr->topo.cart.ndims; i++)
		{
			pos[i] = cart_ptr->topo.cart.position[i];
		}
		/* We must return MPI_PROC_NULL if shifted over the edge of a 
		   non-periodic mesh */
		pos[direction] += displ;
		if (!cart_ptr->topo.cart.periodic[direction] &&
			(pos[direction] >= cart_ptr->topo.cart.dims[direction] ||
			 pos[direction] < 0))
		{
			*dest = MPI_PROC_NULL;
		}
		else
		{
			NMPI_Cart_rank(mpi_errno, ( comm, pos, dest ));
		}

		pos[direction] = cart_ptr->topo.cart.position[direction] - displ;
		if (!cart_ptr->topo.cart.periodic[direction] &&
			(pos[direction] >= cart_ptr->topo.cart.dims[direction] ||
			 pos[direction] < 0))
		{
			*source = MPI_PROC_NULL;
		}
		else
		{
			NMPI_Cart_rank(mpi_errno, ( comm, pos, source ));
		}
	}

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CART_SHIFT);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_cart_shift",
			L"**mpi_cart_shift %C %d %d %p %p", comm, direction, displ, source, dest);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Cart_rank
#define MPI_Cart_rank PMPI_Cart_rank

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Cart_rank

/*@
MPI_Cart_rank - Determines process rank in communicator given Cartesian
				location

Input Parameters:
+ comm - communicator with cartesian structure (handle) 
- coords - integer array (of size 'ndims', the number of dimensions of
	the Cartesian topology associated with 'comm') specifying the cartesian 
  coordinates of a process 

Output Parameter:
. rank - rank of specified process (integer) 

Notes:
 Out-of-range coordinates are erroneous for non-periodic dimensions.  
 Versions of MPICH before 1.2.2 returned 'MPI_PROC_NULL' for the rank in this 
 case.

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_RANK
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Cart_rank(MPI_Comm comm, int *coords, int *rank)
{
	static const wchar_t FCNAME[] = L"MPI_Cart_rank";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *cart_ptr;
	int i, ndims, coord, multiplier;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CART_RANK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CART_RANK);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(coords,L"coords",mpi_errno);
			MPIR_ERRTEST_ARGNULL(rank,L"rank",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_cart_rank %C %p %p", comm, coords, rank); }
	
	cart_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!cart_ptr || cart_ptr->kind != MPI_CART), mpi_errno, MPI_ERR_TOPOLOGY, L"**notcarttopo");

	/* Validate coordinates */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			ndims = cart_ptr->topo.cart.ndims;
			for (i=0; i<ndims; i++)
			{
				if (!cart_ptr->topo.cart.periodic[i])
				{
					coord = coords[i];
					MPIU_ERR_CHKANDJUMP3((coord < 0 || coord >= cart_ptr->topo.cart.dims[i] ), mpi_errno, MPI_ERR_ARG, L"**cartcoordinvalid", L"**cartcoordinvalid %d %d %d",i, coords[i], cart_ptr->topo.cart.dims[i]-1 )
				}
			}
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	ndims = cart_ptr->topo.cart.ndims;
	*rank = 0;
	multiplier = 1;
	for ( i=ndims-1; i >=0; i-- )
	{
		coord = coords[i];
		if ( cart_ptr->topo.cart.periodic[i] )
		{
			if (coord >= cart_ptr->topo.cart.dims[i])
				coord = coord % cart_ptr->topo.cart.dims[i];
			else if (coord <  0)
			{
				coord = coord % cart_ptr->topo.cart.dims[i];
				if (coord) coord = cart_ptr->topo.cart.dims[i] + coord;
			}
		}
		*rank += multiplier * coord;
		multiplier *= cart_ptr->topo.cart.dims[i];
	}
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CART_RANK);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_cart_rank",
			L"**mpi_cart_rank %C %p %p", comm, coords, rank);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Cart_map
#define MPI_Cart_map PMPI_Cart_map

#undef FUNCNAME
#define FUNCNAME MPIR_Cart_map
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Cart_map( const MPID_Comm *comm_ptr, int ndims, const int dims[], const int periodic[], int *newrank )
{
	int rank, nranks, i, size, mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_CART_MAP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_CART_MAP);

	MPIU_UNREFERENCED_ARG(periodic);

	/* Determine number of processes needed for topology */
	nranks = dims[0];
	for ( i=1; i<ndims; i++ )
		nranks *= dims[i];
	
	size = comm_ptr->remote_size;
	
	/* Test that the communicator is large enough */
	if (size < nranks)
	{
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, L"MPIR_Cart_map", __LINE__, MPI_ERR_DIMS, L"**topotoolarge", L"**topotoolarge %d %d", size, nranks );
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_CART_MAP);
		return mpi_errno;
	}

	/* Am I in this range? */
	rank = comm_ptr->rank;
	if ( rank < nranks )
		/* This relies on the ranks *not* being reordered by the current Cartesian routines */
		*newrank = rank;
	else
		*newrank = MPI_UNDEFINED;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_CART_MAP);
	return MPI_SUCCESS;
}
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Cart_map

/*@
MPI_Cart_map - Maps process to Cartesian topology information 

Input Parameters:
+ comm - input communicator (handle) 
. ndims - number of dimensions of Cartesian structure (integer) 
. dims - integer array of size 'ndims' specifying the number of processes in 
  each coordinate direction 
- periods - logical array of size 'ndims' specifying the periodicity 
  specification in each coordinate direction 

Output Parameter:
. newrank - reordered rank of the calling process; 'MPI_UNDEFINED' if 
  calling process does not belong to grid (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_DIMS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Cart_map(MPI_Comm comm_old, int ndims, int *dims, int *periods, int *newrank)
{
#undef FCNAME
	static const wchar_t FCNAME[] = L"MPI_Cart_map";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CART_MAP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CART_MAP);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm_old, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm_old, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(newrank,L"newrank",mpi_errno);
			MPIR_ERRTEST_ARGNULL(dims,L"dims",mpi_errno);
			if (ndims < 1)
			{
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_DIMS, L"**dims", L"**dims %d", ndims );
				}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_cart_map %C %d %p %p %p", comm_old, ndims, dims, periods, newrank); }

	if (comm_ptr->topo_fns != NULL && comm_ptr->topo_fns->cartMap != NULL)
	{
		mpi_errno = comm_ptr->topo_fns->cartMap( comm_ptr, ndims, (const int*) dims, (const int*) periods, newrank );
	}
	else
	{
		mpi_errno = MPIR_Cart_map( comm_ptr, ndims, (const int*) dims, (const int*) periods, newrank );
	}
	MPIU_ERR_CHK_POP(mpi_errno);
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CART_MAP);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_cart_map",
			L"**mpi_cart_map %C %d %p %p %p", comm_old, ndims, dims, periods, 
			newrank);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Cart_get
#define MPI_Cart_get PMPI_Cart_get

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Cart_get

/*@

MPI_Cart_get - Retrieves Cartesian topology information associated with a 
			   communicator

Input Parameters:
+ comm - communicator with cartesian structure (handle) 
- maxdims - length of vectors  'dims', 'periods', and 'coords'
in the calling program (integer) 

Output Parameters:
+ dims - number of processes for each cartesian dimension (array of integer) 
. periods - periodicity (true/false) for each cartesian dimension 
(array of logical) 
- coords - coordinates of calling process in cartesian structure 
(array of integer) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Cart_get(MPI_Comm comm, int maxdims, int *dims, int *periods, int *coords)
{
	static const wchar_t FCNAME[] = L"MPI_Cart_get";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *cart_ptr;
	int i, n, *vals;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CART_GET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CART_GET);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */

			MPIR_ERRTEST_ARGNULL( dims, L"dims", mpi_errno );
			MPIR_ERRTEST_ARGNULL( periods, L"periods", mpi_errno );
			MPIR_ERRTEST_ARGNULL( coords, L"coords", mpi_errno ); 

			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_cart_get %C %d %p %p %p", comm, maxdims, dims, periods, coords); }
	
	cart_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!cart_ptr || cart_ptr->kind != MPI_CART), mpi_errno, MPI_ERR_TOPOLOGY, L"**notcarttopo");
	MPIU_ERR_CHKANDJUMP2((cart_ptr->topo.cart.ndims > maxdims), mpi_errno, MPI_ERR_ARG, L"**dimsmany", L"**dimsmany %d %d", cart_ptr->topo.cart.ndims, maxdims);

	n = cart_ptr->topo.cart.ndims;
	vals = cart_ptr->topo.cart.dims;
	for ( i=0; i<n; i++ )
		*dims++ = *vals++;
	
	/* Get periods */
	n = cart_ptr->topo.cart.ndims;
	vals = cart_ptr->topo.cart.periodic;
	for ( i=0; i<n; i++ )
		*periods++ = *vals++;
	
	/* Get coords */
	n = cart_ptr->topo.cart.ndims;
	vals = cart_ptr->topo.cart.position;
	for ( i=0; i<n; i++ )
		*coords++ = *vals++;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CART_GET);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_cart_get",
			L"**mpi_cart_get %C %d %p %p %p", comm, maxdims, dims, periods, coords);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Cart_create
#define MPI_Cart_create PMPI_Cart_create

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Cart_create

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef FUNCNAME
#define FUNCNAME MPIR_Cart_create
int MPIR_Cart_create( const MPID_Comm *comm_ptr, int ndims, const int dims[], const int periods[], int reorder, MPI_Comm *comm_cart )
{
	static const wchar_t FCNAME[] = L"MPIR_Cart_create";
	int	   i, newsize, rank, nranks, mpi_errno = MPI_SUCCESS;
	MPID_Comm *newcomm_ptr = NULL;
	MPIR_Topology *cart_ptr = NULL;
	MPIU_CHKPMEM_DECL(4);
	MPIDI_STATE_DECL(MPID_STATE_MPIR_CART_CREATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_CART_CREATE);

	/* Check for invalid arguments */
	newsize = 1;
	for (i=0; i<ndims; i++) 
		newsize *= dims[i];

	/* Use ERR_ARG instead of ERR_TOPOLOGY because there is no topology yet */
	MPIU_ERR_CHKANDJUMP2((newsize > comm_ptr->remote_size), mpi_errno, MPI_ERR_ARG, L"**cartdim", L"**cartdim %d %d", comm_ptr->remote_size, newsize);

	/* Create a new communicator as a duplicate of the input communicator (but do not duplicate the attributes) */
	if (reorder)
	{
		MPI_Comm ncomm;
		MPIU_THREADPRIV_DECL;

		/* Allow the cart map routine to remap the assignment of ranks to processes */
		MPIU_THREADPRIV_GET;
		NMPI_Cart_map(mpi_errno, ( comm_ptr->handle, ndims, (int *)dims, (int *)periods, &rank ));
		/* Create the new communicator with split, since we need to reorder
		   the ranks (including the related internals, such as the connection
		   tables */
		if (mpi_errno == 0)
		{
			NMPI_Comm_split(mpi_errno, ( comm_ptr->handle, rank == MPI_UNDEFINED ? MPI_UNDEFINED : 1, rank, &ncomm ));
			if (!mpi_errno)
			{
				MPID_Comm_get_ptr( ncomm, newcomm_ptr );
			}
		}
	}
	else
	{
		mpi_errno = MPIR_Comm_copy( (MPID_Comm *)comm_ptr, newsize, &newcomm_ptr );
		rank   = comm_ptr->rank;
	}
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* If this process is not in the resulting communicator, return a null communicator and exit */
	if (rank >= newsize || rank == MPI_UNDEFINED)
	{
		*comm_cart = MPI_COMM_NULL;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_CART_CREATE);
		return MPI_SUCCESS;
	}

	/* Create the topololgy structure */
	MPIU_CHKPMEM_MALLOC(cart_ptr,MPIR_Topology*,sizeof(MPIR_Topology), mpi_errno, L"cart_ptr" );

	cart_ptr->kind			   = MPI_CART;
	cart_ptr->topo.cart.nnodes   = newsize;
	cart_ptr->topo.cart.ndims	= ndims;
	MPIU_CHKPMEM_MALLOC(cart_ptr->topo.cart.dims,int*,ndims*sizeof(int), mpi_errno, L"cart.dims");
	MPIU_CHKPMEM_MALLOC(cart_ptr->topo.cart.periodic,int*,ndims*sizeof(int), mpi_errno, L"cart.periodic");
	MPIU_CHKPMEM_MALLOC(cart_ptr->topo.cart.position,int*,ndims*sizeof(int), mpi_errno, L"cart.position");
	nranks = newsize;
	for (i=0; i<ndims; i++)
	{
		cart_ptr->topo.cart.dims[i]	 = dims[i];
		cart_ptr->topo.cart.periodic[i] = periods[i];
		nranks = nranks / dims[i];
		/* FIXME: nranks could be zero (?) */
		cart_ptr->topo.cart.position[i] = rank / nranks;
		rank = rank % nranks;
	}

	/* Place this topology onto the communicator */
	mpi_errno = MPIR_Topology_put( newcomm_ptr, cart_ptr );
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	*comm_cart = newcomm_ptr->handle;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_CART_CREATE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	MPIU_CHKPMEM_REAP();
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, L"MPIR_Cart_create", __LINE__, 
			MPI_ERR_OTHER, L"**fail", 0);
	}
#   endif
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_CART_CREATE);
	return mpi_errno;
	/* --END ERROR HANDLING-- */
}
#endif

/*@

MPI_Cart_create - Makes a new communicator to which topology information
				  has been attached

Input Parameters:
+ comm_old - input communicator (handle) 
. ndims - number of dimensions of cartesian grid (integer) 
. dims - integer array of size ndims specifying the number of processes in 
  each dimension 
. periods - logical array of size ndims specifying whether the grid is 
  periodic (true) or not (false) in each dimension 
- reorder - ranking may be reordered (true) or not (false) (logical) 

Output Parameter:
. comm_cart - communicator with new cartesian topology (handle) 

Algorithm:
We ignore 'reorder' info currently.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_DIMS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Cart_create(MPI_Comm comm_old, int ndims, int *dims, int *periods, int reorder, MPI_Comm *comm_cart)
{
	static const wchar_t FCNAME[] = L"MPI_Cart_create";
	int	   mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CART_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("topo");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CART_CREATE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm_old, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm_old, comm_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL( dims, L"dims", mpi_errno );
			MPIR_ERRTEST_ARGNULL( periods, L"periods", mpi_errno );
			MPIR_ERRTEST_ARGNULL( comm_cart, L"comm_cart", mpi_errno );
			if (ndims <= 0)
			{
				/* Must have a positive number of dimensions */
				mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_DIMS, L"**dims",  L"**dims %d", 0 );
			}
			MPIR_ERRTEST_ARGNEG( ndims, L"ndims", mpi_errno );
			if (comm_ptr)
			{
				MPIR_ERRTEST_COMM_INTRA( comm_ptr, mpi_errno );
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_cart_create %C %d %p %p %d %p", comm_old, ndims, dims, periods, reorder, comm_cart); }

	if (comm_ptr->topo_fns != NULL && comm_ptr->topo_fns->cartCreate != NULL)
	{
		mpi_errno = comm_ptr->topo_fns->cartCreate( comm_ptr, ndims, (const int*) dims, (const int*) periods, reorder, comm_cart );
	}
	else
	{
		mpi_errno = MPIR_Cart_create( comm_ptr, ndims, (const int*) dims, (const int*) periods, reorder, comm_cart );
	}
	MPIU_ERR_CHK_POP(mpi_errno);
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CART_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT("topo");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
			L"**mpi_cart_create",
			L"**mpi_cart_create %C %d %p %p %d %p", comm_old, ndims, dims, 
			periods, reorder, comm_cart);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Cart_coords
#define MPI_Cart_coords PMPI_Cart_coords

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Cart_coords

/*@
MPI_Cart_coords - Determines process coords in cartesian topology given
				  rank in group

Input Parameters:
+ comm - communicator with cartesian structure (handle) 
. rank - rank of a process within group of 'comm' (integer) 
- maxdims - length of vector 'coords' in the calling program (integer) 

Output Parameter:
. coords - integer array (of size 'ndims') containing the Cartesian 
  coordinates of specified process (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_RANK
.N MPI_ERR_DIMS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Cart_coords(MPI_Comm comm, int rank, int maxdims, int *coords)
{
	static const wchar_t FCNAME[] = L"MPI_Cart_coords";
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *comm_ptr = NULL;
	MPIR_Topology *cart_ptr;
	int i, nnodes;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_CART_COORDS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_CART_COORDS);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(coords,L"coords",mpi_errno);
			if (mpi_errno) goto fn_fail;
			
			MPIR_ERRTEST_RANK(comm_ptr, rank, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_cart_coords %C %d %d %p", comm, rank, maxdims, coords); }
	
	cart_ptr = MPIR_Topology_get( comm_ptr );

	MPIU_ERR_CHKANDJUMP((!cart_ptr || cart_ptr->kind != MPI_CART), mpi_errno, MPI_ERR_TOPOLOGY, L"**notcarttopo");
	MPIU_ERR_CHKANDJUMP2((cart_ptr->topo.cart.ndims > maxdims), mpi_errno, MPI_ERR_ARG, L"**dimsmany", L"**dimsmany %d %d", cart_ptr->topo.cart.ndims, maxdims);

	/* Calculate coords */
	nnodes = cart_ptr->topo.cart.nnodes;
	for ( i=0; i < cart_ptr->topo.cart.ndims; i++ )
	{
		nnodes	= nnodes / cart_ptr->topo.cart.dims[i];
		coords[i] = rank / nnodes;
		rank	  = rank % nnodes;
	}

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_CART_COORDS);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_cart_coords",
			L"**mpi_cart_coords %C %d %d %p", comm, rank, maxdims, coords);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
