/* (C) 2009 by Deino Software. */

#include "mpidi_ch3_impl.h"
#include "pmi.h"
#include <stdio.h>

/* prototypes */
static int InitSharedProcesses(MPIDI_PG_t *pg, int nRank);
static int shmMPIDI_CH3I_Progress(int is_blocking, MPID_Progress_state *state);
static int shmMPIDI_CH3_VC_Init( MPIDI_VC_t *vc );
static int MPIDI_CH3I_SHM_wait(MPIDI_VC_t *vc, int millisecond_timeout, MPIDI_VC_t **vc_pptr, int *num_bytes_ptr, shm_wait_t *shm_out);

/* Because packets are interpreted in-place in the shm queue, posting a pkt read requires only setting a flag */
#define post_pkt_recv(vc) vc->ch.shm_reading_pkt = TRUE

typedef int SHM_STATE;
#define ZERO_RANK 0x12345678
#undef USE_SYNCHRONIZE_SHMAPPING

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_Finalize(void)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_FINALIZE);

	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));

	/* FIXME: This should be used in abort as well */
	/* Free resources allocated in CH3_Init() */
	if (MPIDI_PG_Get_size(MPIDI_Process.my_pg) > 1)
		mpi_errno = MPIDI_CH3I_SHM_Release_mem(MPIDI_Process.my_pg, TRUE);
	else
		mpi_errno = MPIDI_CH3I_SHM_Release_mem(MPIDI_Process.my_pg, FALSE);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**finalize_release_mem");
	}

	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_FINALIZE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_Init(int has_parent, MPIDI_PG_t * pg, int pg_rank )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_VC_t * vc;
	int pg_size;
	int p;
	wchar_t * key = NULL;
	wchar_t * val = NULL;
	wchar_t *kvsname = NULL;
	int key_max_sz;
	int val_max_sz;
	wchar_t shmemkey[MPIDI_MAX_SHM_NAME_LENGTH];
	int i, j, k;
	int shm_block;
	wchar_t local_host[100];
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_INIT);

	MPIU_UNREFERENCED_ARG(has_parent);

	/*
	 * Extract process group related information from PMI and initialize
	 * structures that track the process group connections, MPI_COMM_WORLD, and
	 * MPI_COMM_SELF
	 */
	/* MPID_Init in mpid_init.c handles the process group initialization. */

	MPIDI_CH3I_Eager_max_msg_size = 10000;

	/* Get the kvsname associated with MPI_COMM_WORLD */
	MPIDI_PG_GetConnKVSname( &kvsname );

	/* set the global variable defaults */
	pg->ch.nShmEagerLimit = MPIDI_SHM_EAGER_LIMIT;
	pg->ch.nShmRndvLimit = MPIDI_SHM_RNDV_LIMIT;
	pg->ch.pSharedProcessHandles = NULL;
	pg->ch.addr = NULL;
	pg->ch.key[0] = L'\0';
	pg->ch.id = NULL;
	pg->ch.nShmWaitSpinCount = MPIDI_CH3I_SPIN_COUNT_DEFAULT;
	pg->ch.nShmWaitYieldCount = MPIDI_CH3I_YIELD_COUNT_DEFAULT;

	/* Initialize the VC table associated with this process group (and thus COMM_WORLD) */
	pg_size = pg->size;

	for (p = 0; p < pg_size; p++)
	{
		/* FIXME: the vc's must be set to active for the close protocol to work in the shm channel */
		pg->vct[p].state = MPIDI_VC_STATE_ACTIVE;
		shmMPIDI_CH3_VC_Init( &pg->vct[p] );
		pg->vct[p].ch.req = (MPID_Request*)MPIU_Malloc(sizeof(MPID_Request));
		/* FIXME: Should these also be set in the VC_Init, or 
		   is VC_Init moot (never called because this channel does not
		   support dynamic processes?) */
		pg->vct[p].ch.shm_reading_pkt = TRUE;
#ifdef USE_SHM_UNEX
		pg->vct[p].ch.unex_finished_next = NULL;
		pg->vct[p].ch.unex_list = NULL;
#endif
	}
	
	/* save my vc_ptr for easy access */
	/* MPIDI_PG_Get_vcr(pg, pg_rank, &MPIDI_CH3I_Process.vc); */
	/* FIXME: Figure out whether this is a common feature of process 
	   groups (and thus make it part of the general PG_Init) or 
	   something else.  Avoid a "get" routine because of the danger in
	   using "get" where "dup" is required. */
	MPIDI_CH3I_Process.vc = &pg->vct[pg_rank];

	/* Allocate space for pmi keys and values */
	mpi_errno = PMI_KVS_Get_key_length_max(&key_max_sz);
	if (mpi_errno != PMI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_get_key_length_max", L"**pmi_kvs_get_key_length_max %d %s", mpi_errno, PMI_Get_last_error_msgW());
		goto fn_fail;
	}
	key_max_sz++;
	key = MPIU_Malloc(key_max_sz*sizeof(wchar_t));
	if (key == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		goto fn_fail;
	}
	mpi_errno = PMI_KVS_Get_value_length_max(&val_max_sz);
	if (mpi_errno != PMI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_get_value_length_max", L"**pmi_kvs_get_value_length_max %d %s", mpi_errno, PMI_Get_last_error_msgW());
		goto fn_fail;
	}
	val_max_sz++;
	val = MPIU_Malloc(val_max_sz*sizeof(wchar_t));
	if (val == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		goto fn_fail;
	}

	/* initialize the shared memory */
	shm_block = sizeof(MPIDI_CH3I_SHM_Queue_t) * pg_size; 

	if (pg_size > 1)
	{
		if (pg_rank == 0)
		{
			/* Put the shared memory key */
			MPIDI_CH3U_Generate_shm_string(shmemkey);
			if (MPIU_Wcsncpy(key, L"SHMEMKEY", key_max_sz))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIU_Wcsncpy", 0);
				goto fn_fail;
			}
			if (MPIU_Wcsncpy(val, shmemkey, val_max_sz))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIU_Wcsncpy", 0);
				goto fn_fail;
			}
			mpi_errno = PMI_KVS_PutW(kvsname, key, val);
			if (mpi_errno != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_put", L"**pmi_kvs_put %d %s", mpi_errno, PMI_Get_last_error_msgW());
				goto fn_fail;
			}

			/* Put the hostname to make sure everyone is on the same host */
			if (MPIU_Wcsncpy(key, L"SHMHOST", key_max_sz))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIU_Wcsncpy", 0);
				goto fn_fail;
			}
			MPID_Get_processor_name( val, val_max_sz, 0 );
			mpi_errno = PMI_KVS_PutW(kvsname, key, val);
			if (mpi_errno != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_put", L"**pmi_kvs_put %d %s", mpi_errno, PMI_Get_last_error_msgW());
				goto fn_fail;
			}

			mpi_errno = PMI_KVS_CommitW(kvsname);
			if (mpi_errno != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_commit", L"**pmi_kvs_commit %d %s", mpi_errno, PMI_Get_last_error_msgW());
				goto fn_fail;
			}
			mpi_errno = PMI_Barrier();
			if (mpi_errno != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_barrier", L"**pmi_barrier %d %s", mpi_errno, PMI_Get_last_error_msgW());
				goto fn_fail;
			}
		}
		else
		{
			/* Get the shared memory key */
			if (MPIU_Wcsncpy(key, L"SHMEMKEY", key_max_sz))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIU_Wcsncpy", 0);
				goto fn_fail;
			}
			mpi_errno = PMI_Barrier();
			if (mpi_errno != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_barrier", L"**pmi_barrier %d %s", mpi_errno, PMI_Get_last_error_msgW());
				goto fn_fail;
			}
			mpi_errno = PMI_KVS_GetW(kvsname, key, val, val_max_sz);
			if (mpi_errno != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_get", L"**pmi_kvs_get %d %s", mpi_errno, PMI_Get_last_error_msgW());
				goto fn_fail;
			}
			if (MPIU_Wcsncpy(shmemkey, val, val_max_sz))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIU_Wcsncpy", 0);
				goto fn_fail;
			}
			/* Get the root host and make sure local process is on the same node */
			if (MPIU_Wcsncpy(key, L"SHMHOST", key_max_sz))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIU_Wcsncpy", 0);
				goto fn_fail;
			}
			mpi_errno = PMI_KVS_GetW(kvsname, key, val, val_max_sz);
			if (mpi_errno != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_get", L"**pmi_kvs_get %d %s", mpi_errno, PMI_Get_last_error_msgW());
				goto fn_fail;
			}
			MPID_Get_processor_name( local_host, sizeof(local_host) / sizeof(wchar_t), NULL );
			if (wcscmp(val, local_host))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmhost", L"**shmhost %s %s", local_host, val);
				goto fn_fail;
			}
		}

		MPIU_DBG_PRINTF((L"KEY = %s\n", shmemkey));
		if (MPIU_Wcsncpy(pg->ch.key, shmemkey, MPIDI_MAX_SHM_NAME_LENGTH))
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIU_Wcsncpy", 0);
			goto fn_fail;
		}

		mpi_errno = MPIDI_CH3I_SHM_Get_mem( pg, pg_size * shm_block, pg_rank, pg_size, TRUE );
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmgetmem", 0);
			goto fn_fail;
		}
		InitSharedProcesses(pg, pg_rank);
	}
	else
	{
		mpi_errno = MPIDI_CH3I_SHM_Get_mem( pg, shm_block, 0, 1, FALSE );
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmgetmem", 0);
			goto fn_fail;
		}
	}

	/* initialize each shared memory queue */
	for (i=0; i<pg_size; i++)
	{
		/* MPIDI_PG_Get_vcr(pg, i, &vc); */
		/* FIXME: Move this code to the general init pg for shared
		   memory */
		vc = &pg->vct[i];
		if (pg->ch.pSharedProcessHandles)
			vc->ch.hSharedProcessHandle = pg->ch.pSharedProcessHandles[i];
		if (i == pg_rank)
		{
			vc->ch.shm = (MPIDI_CH3I_SHM_Queue_t*)((char*)pg->ch.addr + (shm_block * i));
			for (j=0; j<pg_size; j++)
			{
				vc->ch.shm[j].head_index = 0;
				vc->ch.shm[j].tail_index = 0;
				for (k=0; k<MPIDI_CH3I_NUM_PACKETS; k++)
				{
					vc->ch.shm[j].packet[k].offset = 0;
					vc->ch.shm[j].packet[k].avail = MPIDI_CH3I_PKT_AVAILABLE;
				}
			}
		}
		else
		{
			/*vc->ch.shm += pg_rank;*/
			vc->ch.shm = NULL;
			vc->ch.write_shmq = (MPIDI_CH3I_SHM_Queue_t*)((char*)pg->ch.addr + (shm_block * i)) + pg_rank;
			vc->ch.read_shmq = (MPIDI_CH3I_SHM_Queue_t*)((char*)pg->ch.addr + (shm_block * pg_rank)) + i;
			/* post a read of the first packet header */
			vc->ch.shm_reading_pkt = TRUE;
		}
	}

	{
		/* if you know the number of processors, calculate the spin count relative to that number */
		SYSTEM_INFO info;
		GetSystemInfo(&info);
		if (info.dwNumberOfProcessors == 1)
			pg->ch.nShmWaitSpinCount = 1;
		else if (info.dwNumberOfProcessors < (DWORD) pg_size)
			pg->ch.nShmWaitSpinCount = ( MPIDI_CH3I_SPIN_COUNT_DEFAULT * info.dwNumberOfProcessors ) / pg_size;
	}

	mpi_errno = PMI_KVS_CommitW(kvsname);
	if (mpi_errno != 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_commit", L"**pmi_kvs_commit %d %s", mpi_errno, PMI_Get_last_error_msgW());
		goto fn_fail;
	}
	mpi_errno = PMI_Barrier();
	if (mpi_errno != 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_barrier", L"**pmi_barrier %d %s", mpi_errno, PMI_Get_last_error_msgW());
		goto fn_fail;
	}

fn_fail:
	if (val != NULL)
	{ 
		MPIU_Free(val);
	}
	if (key != NULL)
	{ 
		MPIU_Free(key);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_INIT);
	return mpi_errno;
}

/* Perform the channel-specific vc initialization.  This routine is used
   in MPIDI_CH3_Init and in routines that create and initialize connections */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_VC_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_VC_Init( MPIDI_VC_t *vc )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_VC_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_VC_INIT);
	vc->ch.sendq_head		 = NULL;
	vc->ch.sendq_tail		 = NULL;

	/* Which of these do we need? */
	vc->ch.recv_active		= NULL;
	vc->ch.send_active		= NULL;
	vc->ch.req				= NULL;
	vc->ch.read_shmq		= NULL;
	vc->ch.write_shmq		= NULL;
	vc->ch.shm				= NULL;
	vc->ch.shm_state		= 0;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_VC_INIT);
	return MPI_SUCCESS;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iSend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_iSend(MPIDI_VC_t * vc, MPID_Request * sreq, void * pkt, MPIDI_msg_sz_t pkt_sz)
{
	int mpi_errno = MPI_SUCCESS;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISEND);

	MPIU_DBG_PRINTF((L"ch3_isend\n"));
	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (pkt_sz > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
		return mpi_errno;
	}
#endif

	/* The SHM implementation uses a fixed length header, the size of which is the maximum of all possible packet headers */
	pkt_sz = sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)pkt);
	
	if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
	{
		int nb;
		
		MPIDI_DBG_PRINTF((55, FCNAME, L"send queue empty, attempting to write"));
		
		/* MT: need some signalling to lock down our right to use the channel, thus insuring that the progress engine does also try to write */
		
		MPIU_DBG_PRINTF((L"shm_write(%d bytes)\n", pkt_sz));
		mpi_errno = MPIDI_CH3I_SHM_write(vc, pkt, (int)pkt_sz, &nb);
		if (mpi_errno == MPI_SUCCESS)
		{		
			MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));

			if (nb == pkt_sz)
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"write complete, calling MPIDI_CH3U_Handle_send_req()"));
				MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
				if (!complete)
				{
					sreq->ch.iov_offset = 0;
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					vc->ch.send_active = sreq;
				}
				else
				{
					vc->ch.send_active = MPIDI_CH3I_SendQ_head(vc);
				}
			}
			else
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"partial write, enqueuing at head"));
#ifdef MPICH_DBG_OUTPUT
				if (pkt_sz != sizeof(MPIDI_CH3_Pkt_t))
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
					MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST);
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
					return mpi_errno;
				}
#endif
				sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
				sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt + nb);
				sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz - nb);
				sreq->dev.iov_count = 1;
				sreq->ch.iov_offset = 0;
				MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
				vc->ch.send_active = sreq;
			}
		}
		else
		{
			/*vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;*/
			/* TODO: Create an appropriate error message based on the return value (rc) */
			sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmwrite", 0);
			/* MT -CH3U_Request_complete() performs write barrier */
			MPIDI_CH3U_Request_complete(sreq);
		}
	}
	else
	{
		MPIDI_DBG_PRINTF((55, FCNAME, L"send queue not empty, enqueuing"));
#ifdef MPICH_DBG_OUTPUT
		if (pkt_sz != sizeof(MPIDI_CH3_Pkt_t))
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
			return mpi_errno;
		}
#endif
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt);
		sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz);
		sreq->dev.iov_count = 1;
		sreq->ch.iov_offset = 0;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}

	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iSendv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_iSendv(MPIDI_VC_t * vc, MPID_Request * sreq, MPID_IOV * iov, int n_iov)
{
	int mpi_errno = MPI_SUCCESS;
	int complete;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISENDV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISENDV);
	MPIU_DBG_PRINTF((L"ch3_isendv\n"));
	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (n_iov > MPID_IOV_LIMIT)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
		return mpi_errno;
	}
	if (iov[0].MPID_IOV_LEN > (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
		return mpi_errno;
	}
#endif
	/* The SHM implementation uses a fixed length header, the size of which is the maximum of all possible packet headers */
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF);
	
	/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
	if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
	{
		int nb;
		
		MPIDI_DBG_PRINTF((55, FCNAME, L"send queue empty, attempting to write"));
		
		/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does also try to write */
		
		/* FIXME: the current code only agressively writes the first IOV.  Eventually it should be changed to agressively write
		   as much as possible.  Ideally, the code would be shared between the send routines and the progress engine. */
		
		mpi_errno = (n_iov > 1) ?
			MPIDI_CH3I_SHM_writev(vc, iov, n_iov, &nb) :
			MPIDI_CH3I_SHM_write(vc, iov->MPID_IOV_BUF, iov->MPID_IOV_LEN, &nb);
		if (mpi_errno == MPI_SUCCESS)
		{
			int offset = 0;
			
			MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));
			
			while (offset < n_iov)
			{
				if ((int)iov[offset].MPID_IOV_LEN <= nb)
				{
					nb -= (int)iov[offset].MPID_IOV_LEN;
					offset++;
				}
				else
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"partial write, enqueuing at head"));
					for (i = 0; i < n_iov; i++)
					{
						sreq->dev.iov[i] = iov[i];
					}
					if (offset == 0)
					{
#ifdef MPICH_DBG_OUTPUT
						if (iov[0].MPID_IOV_LEN != (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
						{
							mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
							return mpi_errno;
						}
#endif
						sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
						sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->ch.pkt;
					}
					sreq->dev.iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[offset].MPID_IOV_BUF + nb);
					sreq->dev.iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
					sreq->dev.iov_count = n_iov;
					sreq->ch.iov_offset = offset;
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					vc->ch.send_active = sreq;
					break;
				}
			}
			if (offset == n_iov)
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"write complete, calling MPIDI_CH3U_Handle_send_req()"));
				MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
				if (!complete)
				{
					sreq->ch.iov_offset = 0;
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					vc->ch.send_active = sreq;
				}
				else
				{
					vc->ch.send_active = MPIDI_CH3I_SendQ_head(vc);
				}
			}
		}
		else
		{
			/* Connection just failed.  Mark the request complete and return an error. */
			/*vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;*/
			/* TODO: Create an appropriate error message based on the value of errno */
			sreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmwrite", 0);
			/* MT - CH3U_Request_complete performs write barrier */
			MPIDI_CH3U_Request_complete(sreq);
		}
	}
	else
	{
		MPIDI_DBG_PRINTF((55, FCNAME, L"send queue not empty, enqueuing"));
		for (i = 0; i < n_iov; i++)
		{
			sreq->dev.iov[i] = iov[i];
		}
#ifdef MPICH_DBG_OUTPUT
		if (iov[0].MPID_IOV_LEN != (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
			return mpi_errno;
		}
#endif
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[0].MPID_IOV_BUF);
		sreq->dev.iov_count = n_iov;
		sreq->ch.iov_offset = 0;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	
	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPID_STATE_CREATE_REQUEST
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int create_request_istartmsg(MPID_Request *sreq, void *pkt, MPIDI_msg_sz_t pkt_sz, int nb)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CREATE_REQUEST);
	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_REQUEST);
	sreq = MPID_Request_create();
	/*MPIU_Assert(sreq != NULL);*/
	if (sreq == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST);
		return mpi_errno;
	}
	MPIU_Object_set_ref(sreq, 2);
	sreq->kind = MPID_REQUEST_SEND;
	MPIU_Assert(pkt_sz == sizeof(MPIDI_CH3_Pkt_t));
	sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
	sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt + nb);
	sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz - nb);
	sreq->dev.iov_count = 1;
	sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST);
	return mpi_errno;
}

/*
 * MPIDI_CH3_iStartMsg() attempts to send the message immediately.  If the
 * entire message is successfully sent, then NULL is returned.  Otherwise a
 * request is allocated, the header is copied into the request, and a pointer
 * to the request is returned.  An error condition also results in a request be
 * allocated and the errror being returned in the status field of the
 * request.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_iStartMsg(MPIDI_VC_t * vc, void * pkt, MPIDI_msg_sz_t pkt_sz, MPID_Request **sreq_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * sreq = NULL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTMSG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTMSG);

	MPIU_DBG_PRINTF((L"ch3_istartmsg\n"));
	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (pkt_sz > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
		return mpi_errno;
	}
#endif

	/* The SHM implementation uses a fixed length header, the size of which is the maximum of all possible packet headers */
	pkt_sz = sizeof(MPIDI_CH3_Pkt_t);
	
	/* If send queue is empty attempt to send data, queuing any unsent data. */
	if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
	{
		int nb;
		
		/* MT - need some signalling to lock down our right to use the
		   channel, thus insuring that the progress engine does also try to
		   write */
		
		mpi_errno = MPIDI_CH3I_SHM_write(vc, pkt, (int)pkt_sz, &nb);
		if (mpi_errno == MPI_SUCCESS)
		{
			if (nb == pkt_sz)
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"data sent immediately"));
				/* done.  get us out of here as quickly as possible. */
			}
			else
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"send delayed, request enqueued"));
				mpi_errno = create_request_istartmsg(sreq, pkt, pkt_sz, nb);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
					return mpi_errno;
				}
				MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
				vc->ch.send_active = sreq;
			}
		}
		else
		{
			sreq = MPID_Request_create();
			if (sreq == NULL)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
				return mpi_errno;
			}
			sreq->kind = MPID_REQUEST_SEND;
			sreq->cc = 0;
			/* TODO: Create an appropriate error message based on the value of errno */
			sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmwrite", 0);
		}
	}
	else
	{
		MPIDI_DBG_PRINTF((55, FCNAME, L"send in progress, request enqueued"));
		mpi_errno = create_request_istartmsg(sreq, pkt, pkt_sz, 0);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
			return mpi_errno;
		}
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}

	*sreq_ptr = sreq;

	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPID_STATE_CREATE_REQUEST
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int create_request_istartmsgv(MPID_Request *sreq, MPID_IOV *iov, int count, int offset, int nb) \
{
	int mpi_errno = MPI_SUCCESS;
	/*int i;*/
	MPIDI_STATE_DECL(MPID_STATE_CREATE_REQUEST);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);
	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_REQUEST);
	sreq = MPID_Request_create();
	/*MPIU_Assert(sreq != NULL);*/
	if (sreq == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST);
		return mpi_errno;
	}
	MPIU_Object_set_ref(sreq, 2);
	sreq->kind = MPID_REQUEST_SEND;
	MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
	memcpy(sreq->dev.iov, iov, count * sizeof(MPID_IOV));
	MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
	/*for (i = 0; i < count; i++) { sreq->dev.iov[i] = iov[i]; }*/
	if (offset == 0)
	{
		/* memcpy(&sreq->ch.pkt, iov[0].MPID_IOV_BUF, iov[0].MPID_IOV_LEN); */
		MPIU_Assert(iov[0].MPID_IOV_LEN == sizeof(MPIDI_CH3_Pkt_t));
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->ch.pkt;
	}
	sreq->dev.iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[offset].MPID_IOV_BUF + nb);
	sreq->dev.iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
	sreq->ch.iov_offset = offset;
	sreq->dev.iov_count = count;
	sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST);
	return mpi_errno;
}

/*
 * MPIDI_CH3_iStartMsgv() attempts to send the message immediately.  If the
 * entire message is successfully sent, then NULL is returned.  Otherwise a
 * request is allocated, the iovec and the first buffer pointed to by the
 * iovec (which is assumed to be a MPIDI_CH3_Pkt_t) are copied into the
 * request, and a pointer to the request is returned.  An error condition also
 * results in a request be allocated and the errror being returned in the
 * status field of the request.
 */

/* XXX - What do we do if MPID_Request_create() returns NULL???  If
   MPIDI_CH3_iStartMsgv() returns NULL, the calling code assumes the request
   completely successfully, but the reality is that we couldn't allocate the
   memory for a request.  This seems like a flaw in the CH3 API. */

/* NOTE - The completion action associated with a request created by
   CH3_iStartMsgv() is alway MPIDI_CH3_CA_COMPLETE.  This implies that
   CH3_iStartMsgv() can only be used when the entire message can be described
   by a single iovec of size MPID_IOV_LIMIT. */
	
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartMsgv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_iStartMsgv(MPIDI_VC_t * vc, MPID_IOV * iov, int n_iov, MPID_Request **sreq_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * sreq = NULL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTMSGV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTMSGV);

	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (n_iov > MPID_IOV_LIMIT)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
		return mpi_errno;
	}
	if (iov[0].MPID_IOV_LEN > (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
		return mpi_errno;
	}
#endif

	/* The SHM implementation uses a fixed length header, the size of which is the maximum of all possible packet headers */
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF);
	
	/* If send queue is empty attempt to send data, queuing any unsent data. */
	if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
	{
		int error;
		int nb;
		
		/* MT - need some signalling to lock down our right to use the
		channel, thus insuring that the progress engine does also try to
		write */

		error = (n_iov > 1) ?
			MPIDI_CH3I_SHM_writev(vc, iov, n_iov, &nb) :
			MPIDI_CH3I_SHM_write(vc, iov->MPID_IOV_BUF, iov->MPID_IOV_LEN, &nb);
		if (error == MPI_SUCCESS)
		{
			int offset = 0;

			MPIU_DBG_PRINTF((L"ch3_istartmsgv: shm_writev returned %d bytes\n", nb));

			while (offset < n_iov)
			{
				if (nb >= (int)iov[offset].MPID_IOV_LEN)
				{
					nb -= (int)iov[offset].MPID_IOV_LEN;
					offset++;
				}
				else
				{
					MPIU_DBG_PRINTF((L"ch3_istartmsgv: shm_writev did not complete the send, allocating request\n"));
					mpi_errno = create_request_istartmsgv(sreq, iov, n_iov, offset, nb);
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
						return mpi_errno;
					}
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					vc->ch.send_active = sreq;
					break;
				}
			}
		}
		else
		{
			sreq = MPID_Request_create();
			if (sreq == NULL)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
				return mpi_errno;
			}
			sreq->kind = MPID_REQUEST_SEND;
			sreq->cc = 0;
			/* TODO: Create an appropriate error message based on the value of errno */
			sreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmwrite", 0);
		}
	}
	else
	{
		mpi_errno = create_request_istartmsgv(sreq, iov, n_iov, 0, 0);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
			return mpi_errno;
		}
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}

	*sreq_ptr = sreq;

	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_iStartRndvMsg(MPIDI_VC_t * vc, MPID_Request * sreq, MPIDI_CH3_Pkt_t * rts_pkt)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * rts_sreq;
#ifdef MPICH_DBG_OUTPUT
	int i;
#endif
#ifdef USE_SHM_RDMA_GET
	MPIDI_CH3_Pkt_t pkt;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);

#ifdef USE_SHM_RDMA_GET
	MPIDI_Pkt_init(&pkt.rts_iov, MPIDI_CH3_PKT_RTS_IOV);
	pkt.rts_iov.sreq = sreq->handle;
	pkt.rts_iov.iov_len = sreq->dev.iov_count;

	sreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
	sreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	sreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)sreq->dev.iov;
	sreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.iov_count * sizeof(MPID_IOV));
	/* Save a copy of the rts packet in case it is not completely written by the iStartMsgv call */
	/*sreq->dev.rdma_iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rts_pkt;*/
	sreq->ch.pkt = *rts_pkt;
	sreq->dev.rdma_iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->ch.pkt;
	sreq->dev.rdma_iov[2].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<sreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvMsg: send buf[%d] = %p, len = %d\n", i, sreq->dev.iov[i].MPID_IOV_BUF, sreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = shmMPIDI_CH3_iStartMsgv(vc, sreq->dev.rdma_iov, 3, &rts_sreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(sreq, 0);
		MPIDI_CH3_Request_destroy(sreq);
		sreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (rts_sreq != NULL)
	{
		/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
		MPID_Request_release(rts_sreq);
	}
#else
#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<sreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvMsg: send buf[%d] = %p, len = %d\n", i, sreq->dev.iov[i].MPID_IOV_BUF, sreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = shmMPIDI_CH3_iStartMsg(vc, rts_pkt, sizeof(*rts_pkt), &rts_sreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(sreq, 0);
		MPIDI_CH3_Request_destroy(sreq);
		sreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (rts_sreq != NULL)
	{
		/* The sender doesn't need to know when the packet has been sent.  So release the request immediately */
		MPID_Request_release(rts_sreq);
	}
#endif

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvTransfer
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_iStartRndvTransfer(MPIDI_VC_t * vc, MPID_Request * rreq)
{
	int mpi_errno = MPI_SUCCESS;
#ifndef USE_SHM_RDMA_GET
	MPIDI_CH3_Pkt_t pkt;
	MPID_Request *request_ptr;
#endif
#ifdef MPICH_DBG_OUTPUT
	int i;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);

#ifdef USE_SHM_RDMA_GET

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<rreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: recv buf[%d] = %p, len = %d\n", i, rreq->dev.iov[i].MPID_IOV_BUF, rreq->dev.iov[i].MPID_IOV_LEN));
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: send buf[%d] = %p, len = %d\n", i, rreq->dev.rdma_iov[i].MPID_IOV_BUF, rreq->dev.rdma_iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = MPIDI_CH3I_SHM_rdma_readv(vc, rreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

#else

	MPIDI_Pkt_init(&pkt.cts_iov, MPIDI_CH3_PKT_CTS_IOV);
	pkt.cts_iov.sreq = rreq->dev.sender_req_id;
	pkt.cts_iov.rreq = rreq->handle;
	pkt.cts_iov.iov_len = rreq->dev.iov_count;

	rreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
	rreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	rreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rreq->dev.iov;
	rreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.iov_count * sizeof(MPID_IOV));

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<rreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: recv buf[%d] = %p, len = %d\n", i, rreq->dev.iov[i].MPID_IOV_BUF, rreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = shmMPIDI_CH3_iStartMsgv(vc, rreq->dev.rdma_iov, 2, &request_ptr);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(rreq, 0);
		MPIDI_CH3_Request_destroy(rreq);
		rreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (request_ptr != NULL)
	{
		/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
		MPID_Request_release(request_ptr);
	}

#endif

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);
	return mpi_errno;
}

/* This function initializes the table of routines used for 
   implementing the MPI Port-related functions.  This channel does not
   support those operations, so all functions are set to NULL */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_PortFnsInit
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_PortFnsInit( MPIDI_PortFns *portFns )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PORTFNSINIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PORTFNSINIT);
	portFns->OpenPort	 = 0;
	portFns->ClosePort	 = 0;
	portFns->CommAccept	 = 0;
	portFns->CommConnect = 0;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PORTFNSINIT);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Connection_terminate
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3_Connection_terminate(MPIDI_VC_t * vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);
	/* There is no post_close for shm connections so handle them as closed immediately. */
	mpi_errno = MPIDI_CH3U_Handle_connection(vc, MPIDI_VC_EVENT_TERMINATED);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME handle_read
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline int handle_read(MPIDI_VC_t *vc, int nb)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * req;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_HANDLE_READ);

	MPIDI_FUNC_ENTER(MPID_STATE_HANDLE_READ);
	
	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));

	req = vc->ch.recv_active;
	if (req == NULL)
	{
		MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
		MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_READ);
		return MPI_SUCCESS;
	}

	if (nb > 0)
	{
		if (MPIDI_CH3I_Request_adjust_iov(req, nb))
		{
			/* Read operation complete */
			mpi_errno = MPIDI_CH3U_Handle_recv_req(vc, req, &complete);
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			}
			if (complete)
			{
				post_pkt_recv(vc);
			}
			else
			{
				vc->ch.recv_active = req;
				mpi_errno = MPIDI_CH3I_SHM_post_readv(vc, req->dev.iov, req->dev.iov_count, NULL);
			}
		}
	}
	else
	{
		MPIDI_DBG_PRINTF((65, FCNAME, L"Read args were iov=%x, count=%d",
			req->dev.iov, req->dev.iov_count));
	}
	
	MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_READ);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME handle_written
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline int handle_written(MPIDI_VC_t * vc)
{
	int mpi_errno = MPI_SUCCESS;
	int nb;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_HANDLE_WRITTEN);

	MPIDI_FUNC_ENTER(MPID_STATE_HANDLE_WRITTEN);
	
	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));
	while (vc->ch.send_active != NULL)
	{
		MPID_Request * req = vc->ch.send_active;

		/*MPIDI_DBG_PRINTF((60, FCNAME, L"calling shm_writev"));*/
		mpi_errno = MPIDI_CH3I_SHM_writev(vc, req->dev.iov + req->ch.iov_offset, req->dev.iov_count - req->ch.iov_offset, &nb);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**handle_written", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_WRITTEN);
			return mpi_errno;
		}
		MPIDI_DBG_PRINTF((60, FCNAME, L"shm_writev returned %d", nb));

		if (nb > 0)
		{
			if (MPIDI_CH3I_Request_adjust_iov(req, nb))
			{
				/* Write operation complete */
				mpi_errno = MPIDI_CH3U_Handle_send_req(vc, req, &complete);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
					MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_WRITTEN);
					return mpi_errno;
				}
				if (complete)
				{
					MPIDI_CH3I_SendQ_dequeue(vc);
				}
				vc->ch.send_active = MPIDI_CH3I_SendQ_head(vc);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((65, FCNAME, L"shm_post_writev returned %d bytes", nb));
			break;
		}
	}

	MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));

	MPIDI_FUNC_EXIT(MPID_STATE_HANDLE_WRITTEN);
	return mpi_errno;
}


static int shmMPIDI_CH3_Progress_test(void)
{
	return shmMPIDI_CH3I_Progress(FALSE, NULL);
}

static int shmMPIDI_CH3_Progress_wait(MPID_Progress_state *state)
{
	return shmMPIDI_CH3I_Progress(TRUE, state);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmMPIDI_CH3I_Progress(int is_blocking, MPID_Progress_state *state)
{
	int mpi_errno = MPI_SUCCESS;
	int i;
	MPIDI_VC_t *vc_ptr;
	int num_bytes;
	shm_wait_t wait_result;
#ifdef MPICH_DBG_OUTPUT
	unsigned register count;
#endif
	int spin_count = 1;
	unsigned completions = MPIDI_CH3I_progress_completion_count;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS);
#ifdef USE_SLEEP_YIELD
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_SLEEP_YIELD);
#else
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_YIELD);
#endif

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS);

	MPIU_UNREFERENCED_ARG(state);

	MPIDI_DBG_PRINTF((50, FCNAME, L"entering, blocking=%s", is_blocking ? L"true" : L"false"));
	do
	{
		mpi_errno = MPIDI_CH3I_SHM_wait(MPIDI_CH3I_Process.vc, 0, &vc_ptr, &num_bytes, &wait_result);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shm_wait", 0);
			goto fn_exit;
		}
		switch (wait_result)
		{
		case SHM_WAIT_TIMEOUT:
			if (spin_count >= MPIDI_Process.my_pg->ch.nShmWaitSpinCount)
			{
#ifdef USE_SLEEP_YIELD
				if (spin_count >= MPIDI_Process.my_pg->ch.nShmWaitYieldCount)
				{
					MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_SLEEP_YIELD);
					MPIDU_Sleep_yield();
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_SLEEP_YIELD);
				}
				else
				{
					MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
					MPIDU_Yield();
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
				}
#else
				MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
				MPIDU_Yield();
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
#endif
			}
			spin_count++;
			break;
		case SHM_WAIT_READ:
			MPIDI_DBG_PRINTF((50, FCNAME, L"MPIDI_CH3I_SHM_wait reported %d bytes read", num_bytes));
			spin_count = 1;
#ifdef USE_SLEEP_YIELD
			MPIDI_Sleep_yield_count = 0;
#endif
			mpi_errno = handle_read(vc_ptr, num_bytes);
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
				goto fn_exit;
			}
			break;
		case SHM_WAIT_WRITE:
			MPIDI_DBG_PRINTF((50, FCNAME, L"MPIDI_CH3I_SHM_wait reported %d bytes written", num_bytes));
			spin_count = 1;
#ifdef USE_SLEEP_YIELD
			MPIDI_Sleep_yield_count = 0;
#endif
			mpi_errno = handle_written(vc_ptr);
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
				goto fn_exit;
			}
			break;
		case SHM_WAIT_WAKEUP:
			break;
		default:
			/*MPIDI_err_printf(FCNAME, L"MPIDI_CH3I_SHM_wait returned an unknown operation code\n");*/
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shm_op", L"**shm_op %d", wait_result);
			goto fn_exit;
			/*
			MPID_Abort(MPIR_Process.comm_world, MPI_SUCCESS, 13, NULL);
			break;
			*/
		}

		/* pound on the write queues since shm_wait currently does not return SHM_WAIT_WRITE */
		for (i=0; i<MPIDI_PG_Get_size(MPIDI_CH3I_Process.vc->pg); i++)
		{
			MPIDI_PG_Get_vc(MPIDI_CH3I_Process.vc->pg, i, &vc_ptr);
			if (/*MPIDI_Process.my_pg->ch.vc_table[i].*/vc_ptr->ch.send_active != NULL)
			{
				mpi_errno = handle_written(vc_ptr/*&MPIDI_Process.my_pg->ch.vc_table[i]*/);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS);
					return mpi_errno;
				}
			}
		}
	}
	while (completions == MPIDI_CH3I_progress_completion_count && is_blocking);

fn_exit:
#ifdef MPICH_DBG_OUTPUT
	count = MPIDI_CH3I_progress_completion_count - completions;
	if (is_blocking)
	{
		MPIDI_DBG_PRINTF((50, FCNAME, L"exiting, count=%d", count));
	}
	else
	{
		if (count > 0)
		{
			MPIDI_DBG_PRINTF((50, FCNAME, L"exiting (non-blocking), count=%d", count));
		}
	}
#endif
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS);
	return mpi_errno;
}


/* shm functions */

#ifdef USE_SHM_UNEX

#undef FUNCNAME
#define FUNCNAME shmi_buffer_unex_read
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmi_buffer_unex_read(MPIDI_VC_t *vc_ptr, MPIDI_CH3I_SHM_Packet_t *pkt_ptr, void *mem_ptr, unsigned int offset, unsigned int num_bytes)
{
	MPIDI_CH3I_SHM_Unex_read_t *p;
	MPIDI_STATE_DECL(MPID_STATE_SHMI_BUFFER_UNEX_READ);

	MPIDI_FUNC_ENTER(MPID_STATE_SHMI_BUFFER_UNEX_READ);

	MPIDI_DBG_PRINTF((60, FCNAME, L"%d bytes\n", num_bytes));

	p = (MPIDI_CH3I_SHM_Unex_read_t *)MPIU_Malloc(sizeof(MPIDI_CH3I_SHM_Unex_read_t));
	p->pkt_ptr = pkt_ptr;
	p->buf = (unsigned char *)mem_ptr + offset;
	p->length = num_bytes;
	p->next = vc_ptr->ch.unex_list;
	vc_ptr->ch.unex_list = p;

	MPIDI_FUNC_EXIT(MPID_STATE_SHMI_BUFFER_UNEX_READ);
	return 0;
}

#undef FUNCNAME
#define FUNCNAME shmi_read_unex
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int shmi_read_unex(MPIDI_VC_t *vc_ptr)
{
	unsigned int len;
	MPIDI_CH3I_SHM_Unex_read_t *temp;
	MPIDI_STATE_DECL(MPID_STATE_SHMI_READ_UNEX);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_SHMI_READ_UNEX);

	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));
	MPIU_Assert(vc_ptr->ch.unex_list);

	/* copy the received data */
	while (vc_ptr->ch.unex_list)
	{
		len = min(vc_ptr->ch.unex_list->length, vc_ptr->ch.read.bufflen);
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(vc_ptr->ch.read.buffer, vc_ptr->ch.unex_list->buf, len);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		/* advance the user pointer */
		vc_ptr->ch.read.buffer = (char*)(vc_ptr->ch.read.buffer) + len;
		vc_ptr->ch.read.bufflen -= len;
		vc_ptr->ch.read.total += len;
		if (len != vc_ptr->ch.unex_list->length)
		{
			vc_ptr->ch.unex_list->length -= len;
			vc_ptr->ch.unex_list->buf += len;
		}
		else
		{
			/* put the receive packet back in the pool */
			MPIU_Assert(vc_ptr->ch.unex_list->pkt_ptr != NULL);
			vc_ptr->ch.unex_list->pkt_ptr->cur_pos = 
				vc_ptr->ch.unex_list->pkt_ptr->data;
			vc_ptr->ch.unex_list->pkt_ptr->avail = MPIDI_CH3I_PKT_AVAILABLE;
			/* MPIU_Free the unexpected data node */
			temp = vc_ptr->ch.unex_list;
			vc_ptr->ch.unex_list = vc_ptr->ch.unex_list->next;
			MPIU_Free(temp);
		}
		/* check to see if the entire message was received */
		if (vc_ptr->ch.read.bufflen == 0)
		{
			/* place this vc_ptr in the finished list so it will be 
			   completed by shm_wait */
			vc_ptr->ch.shm_state &= ~SHM_READING_BIT;
			vc_ptr->ch.unex_finished_next = MPIDI_CH3I_Process.unex_finished_list;
			MPIDI_CH3I_Process.unex_finished_list = vc_ptr;
			MPIDI_FUNC_EXIT(MPID_STATE_SHMI_READ_UNEX);
			return SHM_SUCCESS;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_SHMI_READ_UNEX);
	return SHM_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME shmi_readv_unex
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int shmi_readv_unex(MPIDI_VC_t *vc_ptr)
{
	unsigned int num_bytes;
	MPIDI_CH3I_SHM_Unex_read_t *temp;
	MPIDI_STATE_DECL(MPID_STATE_SHMI_READV_UNEX);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_SHMI_READV_UNEX);

	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));

	while (vc_ptr->ch.unex_list)
	{
		while (vc_ptr->ch.unex_list->length && vc_ptr->ch.read.iovlen)
		{
			num_bytes = min(vc_ptr->ch.unex_list->length, 
							vc_ptr->ch.read.iov[vc_ptr->ch.read.index].MPID_IOV_LEN);
			MPIDI_DBG_PRINTF((60, FCNAME, L"copying %d bytes\n", num_bytes));
			/* copy the received data */
			MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
			memcpy(vc_ptr->ch.read.iov[vc_ptr->ch.read.index].MPID_IOV_BUF, vc_ptr->ch.unex_list->buf, num_bytes);
			MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
			vc_ptr->ch.read.total += num_bytes;
			vc_ptr->ch.unex_list->buf += num_bytes;
			vc_ptr->ch.unex_list->length -= num_bytes;
			/* update the iov */
			vc_ptr->ch.read.iov[vc_ptr->ch.read.index].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)num_bytes;
			vc_ptr->ch.read.iov[vc_ptr->ch.read.index].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)(
				(char*)(vc_ptr->ch.read.iov[vc_ptr->ch.read.index].MPID_IOV_BUF) + num_bytes);
			if (vc_ptr->ch.read.iov[vc_ptr->ch.read.index].MPID_IOV_LEN == 0)
			{
				vc_ptr->ch.read.index++;
				vc_ptr->ch.read.iovlen--;
			}
		}

		if (vc_ptr->ch.unex_list->length == 0)
		{
			/* put the receive packet back in the pool */
			MPIU_Assert(vc_ptr->ch.unex_list->pkt_ptr != NULL);
			vc_ptr->ch.unex_list->pkt_ptr->cur_pos = vc_ptr->ch.unex_list->pkt_ptr->data;
			vc_ptr->ch.unex_list->pkt_ptr->avail = MPIDI_CH3I_PKT_AVAILABLE;
			/* MPIU_Free the unexpected data node */
			temp = vc_ptr->ch.unex_list;
			vc_ptr->ch.unex_list = vc_ptr->ch.unex_list->next;
			MPIU_Free(temp);
		}
		
		if (vc_ptr->ch.read.iovlen == 0)
		{
			vc_ptr->ch.shm_state &= ~SHM_READING_BIT;
			vc_ptr->ch.unex_finished_next = MPIDI_CH3I_Process.unex_finished_list;
			MPIDI_CH3I_Process.unex_finished_list = vc_ptr;
			MPIDI_DBG_PRINTF((60, FCNAME, L"finished read saved in MPIDI_CH3I_Process.unex_finished_list\n"));
			MPIDI_FUNC_EXIT(MPID_STATE_SHMI_READV_UNEX);
			return SHM_SUCCESS;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_SHMI_READV_UNEX);
	return SHM_SUCCESS;
}

#endif /* USE_SHM_UNEX */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_SHM_wait(MPIDI_VC_t *vc, int millisecond_timeout, MPIDI_VC_t **vc_pptr, int *num_bytes_ptr, shm_wait_t *shm_out)
{
	int mpi_errno = MPI_SUCCESS;
	void *mem_ptr;
	char *iter_ptr;
	int num_bytes;
	unsigned int offset;
	MPIDI_VC_t *recv_vc_ptr;
	MPIDI_CH3I_SHM_Packet_t *pkt_ptr;
	int i;
	register int index, working;
#ifdef USE_SHM_UNEX
	MPIDI_VC_t *temp_vc_ptr;
#endif
	MPID_Request *sreq, *rreq;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_WAIT);

	for (;;) 
	{
#ifdef USE_SHM_UNEX
		if (MPIDI_CH3I_Process.unex_finished_list)
		{
			MPIDI_DBG_PRINTF((60, FCNAME, L"returning previously received %d bytes", MPIDI_CH3I_Process.unex_finished_list->ch.read.total));

			*num_bytes_ptr = MPIDI_CH3I_Process.unex_finished_list->ch.read.total;
			*vc_pptr = MPIDI_CH3I_Process.unex_finished_list;
			/* remove this vc from the finished list */
			temp_vc_ptr = MPIDI_CH3I_Process.unex_finished_list;
			MPIDI_CH3I_Process.unex_finished_list = MPIDI_CH3I_Process.unex_finished_list->ch.unex_finished_next;
			temp_vc_ptr->ch.unex_finished_next = NULL;

			*shm_out = SHM_WAIT_READ;
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
			return MPI_SUCCESS;
		}
#endif /* USE_SHM_UNEX */

		working = FALSE;

		for (i=0; i<MPIDI_PG_Get_size(vc->pg); i++)
		{
			/* skip over the vc to myself */
			if (vc->pg_rank == i)
				continue;

			index = vc->ch.shm[i].head_index;

			/* if the packet at the head index is available, the queue is empty */
			if (vc->ch.shm[i].packet[index].avail == MPIDI_CH3I_PKT_AVAILABLE)
				continue;
			MPID_READ_BARRIER(); /* no loads after this line can occur before the avail flag has been read */

			working = TRUE;

			pkt_ptr = &vc->ch.shm[i].packet[index];
			mem_ptr = (void*)(pkt_ptr->data + pkt_ptr->offset);
			/*mem_ptr = (void*)vc->ch.shm[i].packet[index].cur_pos;*/
			num_bytes = vc->ch.shm[i].packet[index].num_bytes;
			MPIDI_PG_Get_vc(vc->pg, i, &recv_vc_ptr);

			if (recv_vc_ptr->ch.shm_reading_pkt)
			{
				MPIU_DBG_PRINTF((L"shm_read_progress: reading %d byte header from shm packet %d offset %d size %d\n", sizeof(MPIDI_CH3_Pkt_t), index, pkt_ptr->offset, num_bytes));
				if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type > MPIDI_CH3_PKT_END_CH3)
				{
					if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3_PKT_RTS_IOV)
					{
						/*wprintf(L"received rts packet.\n");fflush(stdout);*/
						rreq = MPID_Request_create();
						if (rreq == NULL)
						{
							mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
							return mpi_errno;
						}
						MPIU_Object_set_ref(rreq, 1);
						rreq->kind = MPIDI_CH3I_RTS_IOV_READ_REQUEST;
						rreq->dev.rdma_request = ((MPIDI_CH3_Pkt_rdma_rts_iov_t*)mem_ptr)->sreq;
						rreq->dev.rdma_iov_count = ((MPIDI_CH3_Pkt_rdma_rts_iov_t*)mem_ptr)->iov_len;
						rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&rreq->dev.rdma_iov;
						rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.rdma_iov_count * sizeof(MPID_IOV));
						rreq->dev.iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&rreq->ch.pkt;
						rreq->dev.iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
						rreq->dev.iov_count = 2;
						rreq->ch.req = NULL;
						recv_vc_ptr->ch.recv_active = rreq;
					}
					else if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3_PKT_CTS_IOV)
					{
						/*wprintf(L"received cts packet.\n");fflush(stdout);*/
						MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_cts_iov_t*)mem_ptr)->sreq, sreq);
						sreq->dev.rdma_request = ((MPIDI_CH3_Pkt_rdma_cts_iov_t*)mem_ptr)->rreq;
						sreq->dev.rdma_iov_count = ((MPIDI_CH3_Pkt_rdma_cts_iov_t*)mem_ptr)->iov_len;
						rreq = MPID_Request_create();
						if (rreq == NULL)
						{
							mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
							return mpi_errno;
						}
						MPIU_Object_set_ref(rreq, 1);
						rreq->kind = MPIDI_CH3I_IOV_WRITE_REQUEST;
						rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->dev.rdma_iov;
						rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.rdma_iov_count * sizeof(MPID_IOV));
						rreq->dev.iov_count = 1;
						rreq->ch.req = sreq;
						recv_vc_ptr->ch.recv_active = rreq;
						/*MPIDI_CH3I_SHM_post_read(recv_vc_ptr, &sreq->ch.rdma_iov, sreq->ch.rdma_iov_count * sizeof(MPID_IOV), NULL);*/
					}
					else if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3_PKT_IOV)
					{
						if ( ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->send_recv == MPIDI_CH3_PKT_RELOAD_SEND )
						{
							/*wprintf(L"received sender's iov packet, posting a read of %d iovs.\n", ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->iov_len);fflush(stdout);*/
							MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->req, sreq);
							sreq->dev.rdma_iov_count = ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->iov_len;
							rreq = MPID_Request_create();
							if (rreq == NULL)
							{
								mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
								return mpi_errno;
							}
							MPIU_Object_set_ref(rreq, 1);
							rreq->kind = MPIDI_CH3I_IOV_READ_REQUEST;
							rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->dev.rdma_iov;
							rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.rdma_iov_count * sizeof(MPID_IOV));
							rreq->dev.iov_count = 1;
							rreq->ch.req = sreq;
							recv_vc_ptr->ch.recv_active = rreq;
						}
						else if ( ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->send_recv == MPIDI_CH3_PKT_RELOAD_RECV )
						{
							/*wprintf(L"received receiver's iov packet, posting a read of %d iovs.\n", ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->iov_len);fflush(stdout);*/
							MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->req, rreq);
							rreq->dev.rdma_iov_count = ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->iov_len;
							sreq = MPID_Request_create();
							if (sreq == NULL)
							{
								mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
								return mpi_errno;
							}
							MPIU_Object_set_ref(sreq, 1);
							sreq->kind = MPIDI_CH3I_IOV_WRITE_REQUEST;
							sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&rreq->dev.rdma_iov;
							sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.rdma_iov_count * sizeof(MPID_IOV));
							sreq->dev.iov_count = 1;
							sreq->ch.req = rreq;
							recv_vc_ptr->ch.recv_active = sreq;
						}
						else
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"received invalid MPIDI_CH3_PKT_IOV packet");
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
							return mpi_errno;
						}
					}
					else if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3_PKT_RELOAD)
					{
						if (((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->send_recv == MPIDI_CH3_PKT_RELOAD_SEND)
						{
							/*wprintf(L"received reload send packet.\n");fflush(stdout);*/
							MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->sreq, sreq);
							mpi_errno = MPIDI_CH3U_Handle_send_req(recv_vc_ptr, sreq, &complete);
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"unable to update send request after receiving a reload packet");
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
								return mpi_errno;
							}
							if (!complete)
							{
								/* send a new iov */
								MPID_Request * rts_sreq;
								MPIDI_CH3_Pkt_t pkt;

								/*wprintf(L"sending reloaded send iov of length %d\n", sreq->dev.iov_count);fflush(stdout);*/
								MPIDI_Pkt_init(&pkt.iov, MPIDI_CH3_PKT_IOV);
								pkt.iov.send_recv = MPIDI_CH3_PKT_RELOAD_SEND;
								pkt.iov.req = ((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->rreq;
								pkt.iov.iov_len = sreq->dev.iov_count;

								sreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
								sreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
								sreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)sreq->dev.iov;
								sreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.iov_count * sizeof(MPID_IOV));

								mpi_errno = shmMPIDI_CH3_iStartMsgv(recv_vc_ptr, sreq->dev.rdma_iov, 2, &rts_sreq);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									MPIU_Object_set_ref(sreq, 0);
									MPIDI_CH3_Request_destroy(sreq);
									sreq = NULL;
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
									MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
									return mpi_errno;
								}
								/* --END ERROR HANDLING-- */
								if (rts_sreq != NULL)
								{
									/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
									MPID_Request_release(rts_sreq);
								}
							}
						}
						else if (((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->send_recv == MPIDI_CH3_PKT_RELOAD_RECV)
						{
							/*wprintf(L"received reload recv packet.\n");fflush(stdout);*/
							MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->rreq, rreq);
							mpi_errno = MPIDI_CH3U_Handle_recv_req(recv_vc_ptr, rreq, &complete);
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"unable to update request after receiving a reload packet");
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
								return mpi_errno;
							}
							if (!complete)
							{
								/* send a new iov */
								MPID_Request * cts_sreq;
								MPIDI_CH3_Pkt_t pkt;

								/*wprintf(L"sending reloaded recv iov of length %d\n", rreq->dev.iov_count);fflush(stdout);*/
								MPIDI_Pkt_init(&pkt.iov, MPIDI_CH3_PKT_IOV);
								pkt.iov.send_recv = MPIDI_CH3_PKT_RELOAD_RECV;
								pkt.iov.req = ((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->sreq;
								pkt.iov.iov_len = rreq->dev.iov_count;

								rreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
								rreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
								rreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rreq->dev.iov;
								rreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.iov_count * sizeof(MPID_IOV));

								mpi_errno = shmMPIDI_CH3_iStartMsgv(recv_vc_ptr, rreq->dev.rdma_iov, 2, &cts_sreq);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									/* This destruction probably isn't correct. */
									/* I think it needs to save the error in the request, complete the request and return */
									MPIU_Object_set_ref(rreq, 0);
									MPIDI_CH3_Request_destroy(rreq);
									rreq = NULL;
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
									MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
									return mpi_errno;
								}
								/* --END ERROR HANDLING-- */
								if (cts_sreq != NULL)
								{
									/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
									MPID_Request_release(cts_sreq);
								}
							}
						}
						else
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
							return mpi_errno;
						}

						recv_vc_ptr->ch.recv_active = NULL;
						recv_vc_ptr->ch.shm_reading_pkt = TRUE;
						if (num_bytes > sizeof(MPIDI_CH3_Pkt_t))
						{
							pkt_ptr->offset += sizeof(MPIDI_CH3_Pkt_t);
							num_bytes -= sizeof(MPIDI_CH3_Pkt_t);
							pkt_ptr->num_bytes = num_bytes;
							mem_ptr = (char*)mem_ptr + sizeof(MPIDI_CH3_Pkt_t);
						}
						else
						{
							pkt_ptr->offset = 0;
							MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
							pkt_ptr->avail = MPIDI_CH3I_PKT_AVAILABLE;
							vc->ch.shm[i].head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
						}
						/* return from the wait */
						*num_bytes_ptr = 0;
						*vc_pptr = recv_vc_ptr;
						*shm_out = SHM_WAIT_WAKEUP;
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
						return MPI_SUCCESS;
					}
					else
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"shared memory read progress unable to handle unknown rdma packet");
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
						return mpi_errno;
					}
				}
				else
				{
					mpi_errno = MPIDI_CH3U_Handle_recv_pkt(recv_vc_ptr, (MPIDI_CH3_Pkt_t*)mem_ptr, &recv_vc_ptr->ch.recv_active);
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"shared memory read progress unable to handle incoming packet");
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
						return mpi_errno;
					}
				}

				if (recv_vc_ptr->ch.recv_active == NULL)
				{
					recv_vc_ptr->ch.shm_reading_pkt = TRUE;
				}
				else
				{
					mpi_errno = MPIDI_CH3I_SHM_post_readv(recv_vc_ptr, recv_vc_ptr->ch.recv_active->dev.iov, recv_vc_ptr->ch.recv_active->dev.iov_count, NULL);
				}

				if (num_bytes > sizeof(MPIDI_CH3_Pkt_t))
				{
					pkt_ptr->offset += sizeof(MPIDI_CH3_Pkt_t);
					num_bytes -= sizeof(MPIDI_CH3_Pkt_t);
					pkt_ptr->num_bytes = num_bytes;
					mem_ptr = (char*)mem_ptr + sizeof(MPIDI_CH3_Pkt_t);
				}
				else
				{
					pkt_ptr->offset = 0;
					MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
					pkt_ptr->avail = MPIDI_CH3I_PKT_AVAILABLE;
					vc->ch.shm[i].head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
					continue;
				}
				if (recv_vc_ptr->ch.recv_active == NULL)
					continue;
			}

			MPIDI_DBG_PRINTF((60, FCNAME, L"read %d bytes\n", num_bytes));
			/*MPIDI_DBG_PRINTF((60, FCNAME, L"shm_wait(recv finished %d bytes)", num_bytes));*/
			if (!(recv_vc_ptr->ch.shm_state & SHM_READING_BIT))
			{
#ifdef USE_SHM_UNEX
				/* Should we buffer unexpected messages or leave them in the shmem queue? */
				/*shmi_buffer_unex_read(recv_vc_ptr, pkt_ptr, mem_ptr, 0, num_bytes);*/
#endif
				continue;
			}
			MPIDI_DBG_PRINTF((60, FCNAME, L"read update, total = %d + %d = %d\n", recv_vc_ptr->ch.read.total, num_bytes, recv_vc_ptr->ch.read.total + num_bytes));
			if (recv_vc_ptr->ch.read.use_iov)
			{
				iter_ptr = mem_ptr;
				while (num_bytes && recv_vc_ptr->ch.read.iovlen > 0)
				{
					if ((int)recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN <= num_bytes)
					{
						/* copy the received data */
						MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
						memcpy(recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_BUF, iter_ptr,
							recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN);
						MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
						MPIU_DBG_PRINTF((L"a:shm_read_progress: %d bytes read from packet %d offset %d\n",
							recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN, index,
							pkt_ptr->offset + (int)((char*)iter_ptr - (char*)mem_ptr)));
						iter_ptr += recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN;
						/* update the iov */
						num_bytes -= (int)recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN;
						recv_vc_ptr->ch.read.index++;
						recv_vc_ptr->ch.read.iovlen--;
					}
					else
					{
						/* copy the received data */
						MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
						memcpy(recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_BUF, iter_ptr, num_bytes);
						MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
						MPIU_DBG_PRINTF((L"b:shm_read_progress: %d bytes read from packet %d offset %d\n", num_bytes, index,
							pkt_ptr->offset + (int)((char*)iter_ptr - (char*)mem_ptr)));
						iter_ptr += num_bytes;
						/* update the iov */
						recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)num_bytes;
						recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)(
							(char*)(recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_BUF) + num_bytes);
						num_bytes = 0;
					}
				}
				offset = (unsigned int)((unsigned char*)iter_ptr - (unsigned char*)mem_ptr);
				recv_vc_ptr->ch.read.total += offset;
				if (num_bytes == 0)
				{
					/* put the shm buffer back in the queue */
					vc->ch.shm[i].packet[index].offset = 0;
					MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
					vc->ch.shm[i].packet[index].avail = MPIDI_CH3I_PKT_AVAILABLE;
					vc->ch.shm[i].head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
				}
				else
				{
					/* save the unused but received data */
					/*shmi_buffer_unex_read(recv_vc_ptr, pkt_ptr, mem_ptr, offset, num_bytes);*/
					/* OR */
					/* update the head of the shmem queue */
					pkt_ptr->offset += (pkt_ptr->num_bytes - num_bytes);
					pkt_ptr->num_bytes = num_bytes;
				}
				if (recv_vc_ptr->ch.read.iovlen == 0)
				{
					if (recv_vc_ptr->ch.recv_active->kind < MPID_LAST_REQUEST_KIND)
					{
						recv_vc_ptr->ch.shm_state &= ~SHM_READING_BIT;
						*num_bytes_ptr = recv_vc_ptr->ch.read.total;
						*vc_pptr = recv_vc_ptr;
						*shm_out = SHM_WAIT_READ;
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
						return MPI_SUCCESS;
					}
					else if (recv_vc_ptr->ch.recv_active->kind == MPIDI_CH3I_RTS_IOV_READ_REQUEST)
					{
						int found;
						/*wprintf(L"received rts iov_read.\n");fflush(stdout);*/

						mpi_errno = MPIDI_CH3U_Handle_recv_rndv_pkt(recv_vc_ptr,
																	&recv_vc_ptr->ch.recv_active->ch.pkt,
																	&rreq, &found);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"shared memory read progress unable to handle incoming rts(get) packet");
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
							return mpi_errno;
						}
						/* --END ERROR HANDLING-- */

						for (i=0; i<recv_vc_ptr->ch.recv_active->dev.rdma_iov_count; i++)
						{
							rreq->dev.rdma_iov[i].MPID_IOV_BUF = recv_vc_ptr->ch.recv_active->dev.rdma_iov[i].MPID_IOV_BUF;
							rreq->dev.rdma_iov[i].MPID_IOV_LEN = recv_vc_ptr->ch.recv_active->dev.rdma_iov[i].MPID_IOV_LEN;
						}
						rreq->dev.rdma_iov_count = recv_vc_ptr->ch.recv_active->dev.rdma_iov_count;

						if (found)
						{
							mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, &rreq);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code (mpi_errno, MPIR_ERR_FATAL,
									FCNAME, __LINE__,
									MPI_ERR_OTHER,
									L"**ch3|postrecv",
									L"**ch3|postrecv %s",
									L"MPIDI_CH3_PKT_RNDV_REQ_TO_SEND");
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
								return mpi_errno;
							}
							/* --END ERROR HANDLING-- */
							mpi_errno = shmMPIDI_CH3_iStartRndvTransfer(recv_vc_ptr, rreq);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code (mpi_errno, MPIR_ERR_FATAL,
									FCNAME, __LINE__,
									MPI_ERR_OTHER,
									L"**ch3|ctspkt", 0);
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
								return mpi_errno;
							}
							/* --END ERROR HANDLING-- */
						}

						rreq = recv_vc_ptr->ch.recv_active;
						/* free the request used to receive the rts packet and iov data */
						MPIU_Object_set_ref(rreq, 0);
						MPIDI_CH3_Request_destroy(rreq);

						recv_vc_ptr->ch.recv_active = NULL;
						recv_vc_ptr->ch.shm_reading_pkt = TRUE;
					}
					else if (recv_vc_ptr->ch.recv_active->kind == MPIDI_CH3I_IOV_READ_REQUEST)
					{
						/*wprintf(L"received iov_read.\n");fflush(stdout);*/
						rreq = recv_vc_ptr->ch.recv_active;

						mpi_errno = shmMPIDI_CH3_iStartRndvTransfer(recv_vc_ptr, rreq->ch.req);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"shared memory read progress unable to handle incoming rts(get) iov");
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
							return mpi_errno;
						}
						/* --END ERROR HANDLING-- */

						recv_vc_ptr->ch.recv_active = NULL;
						recv_vc_ptr->ch.shm_reading_pkt = TRUE;

						/* free the request used to receive the iov data */
						MPIU_Object_set_ref(rreq, 0);
						MPIDI_CH3_Request_destroy(rreq);
					}
					else if (recv_vc_ptr->ch.recv_active->kind == MPIDI_CH3I_IOV_WRITE_REQUEST)
					{
						/*wprintf(L"received iov_write.\n");fflush(stdout);*/
						mpi_errno = MPIDI_CH3I_SHM_rdma_writev(recv_vc_ptr, recv_vc_ptr->ch.recv_active->ch.req);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
							return mpi_errno;
						}
						/* --END ERROR HANDLING-- */

						/* return from the wait */
						MPID_Request_release(recv_vc_ptr->ch.recv_active);
						recv_vc_ptr->ch.recv_active = NULL;
						recv_vc_ptr->ch.shm_reading_pkt = TRUE;
						*num_bytes_ptr = 0;
						*vc_pptr = recv_vc_ptr;
						*shm_out = SHM_WAIT_WAKEUP;
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
						return MPI_SUCCESS;
					}
					else
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s %d", L"invalid request type", recv_vc_ptr->ch.recv_active->kind);
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
						return mpi_errno;
					}
				}
			}
			else
			{
				if ((unsigned int)num_bytes > recv_vc_ptr->ch.read.bufflen)
				{
					/* copy the received data */
					MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
					memcpy(recv_vc_ptr->ch.read.buffer, mem_ptr, recv_vc_ptr->ch.read.bufflen);
					MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
					MPIU_DBG_PRINTF((L"c:shm_read_progress: %d bytes read from packet %d offset %d\n", recv_vc_ptr->ch.read.bufflen, index, pkt_ptr->offset));
					recv_vc_ptr->ch.read.total = recv_vc_ptr->ch.read.bufflen;
					/*shmi_buffer_unex_read(recv_vc_ptr, pkt_ptr, mem_ptr, recv_vc_ptr->ch.read.bufflen, num_bytes - recv_vc_ptr->ch.read.bufflen);*/
					pkt_ptr->offset += recv_vc_ptr->ch.read.bufflen;
					pkt_ptr->num_bytes = num_bytes - recv_vc_ptr->ch.read.bufflen;
					recv_vc_ptr->ch.read.bufflen = 0;
				}
				else
				{
					/* copy the received data */
					MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
					memcpy(recv_vc_ptr->ch.read.buffer, mem_ptr, num_bytes);
					MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
					MPIU_DBG_PRINTF((L"d:shm_read_progress: %d bytes read from packet %d offset %d\n", num_bytes, index, pkt_ptr->offset));
					recv_vc_ptr->ch.read.total += num_bytes;
					/* advance the user pointer */
					recv_vc_ptr->ch.read.buffer = (char*)(recv_vc_ptr->ch.read.buffer) + num_bytes;
					recv_vc_ptr->ch.read.bufflen -= num_bytes;
					/* put the shm buffer back in the queue */
					vc->ch.shm[i].packet[index].offset = 0;
					MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
					vc->ch.shm[i].packet[index].avail = MPIDI_CH3I_PKT_AVAILABLE;
					vc->ch.shm[i].head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
				}
				if (recv_vc_ptr->ch.read.bufflen == 0)
				{
					MPIU_Assert(recv_vc_ptr->ch.recv_active->kind < MPID_LAST_REQUEST_KIND);

					recv_vc_ptr->ch.shm_state &= ~SHM_READING_BIT;
					*num_bytes_ptr = recv_vc_ptr->ch.read.total;
					*vc_pptr = recv_vc_ptr;
					*shm_out = SHM_WAIT_READ;
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
					return MPI_SUCCESS;
				}
			}
		}

		if (millisecond_timeout == 0 && !working)
		{
			*shm_out = SHM_WAIT_TIMEOUT;
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
			return MPI_SUCCESS;
		}
	}

	MPIU_UNREACHABLE_CODE();

	/*
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WAIT);
	return MPI_SUCCESS;
	*/
}

#undef FUNCNAME
#define FUNCNAME InitSharedProcesses
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int InitSharedProcesses(MPIDI_PG_t *pg, int nRank)
{
	int mpi_errno;
	int i;
	MPIDI_CH3I_Shared_process_t *pSharedProcess;
	int nProc;
	MPIDI_STATE_DECL(MPID_STATE_INITSHAREDPROCESSES);

	MPIDI_FUNC_ENTER(MPID_STATE_INITSHAREDPROCESSES);

	nProc = MPIDI_PG_Get_size(pg);

	/* initialize arrays */
	pg->ch.pSharedProcessHandles = (HANDLE*)MPIU_Malloc(sizeof(HANDLE) * nProc);

	pSharedProcess = pg->ch.pSHP;

	pSharedProcess[nRank].nPid = GetCurrentProcessId();
	pSharedProcess[nRank].bFinished = FALSE;
	if (nRank == 0)
		pSharedProcess[nRank].nRank = ZERO_RANK;
	else
		pSharedProcess[nRank].nRank = nRank;

	for (i=0; i<nProc; i++)
	{
		if (i != nRank)
		{
			if (i == 0)
			{
				while (pSharedProcess[i].nRank != ZERO_RANK)
					MPIDU_Yield();
			}
			else
			{
				while (pSharedProcess[i].nRank != i)
					MPIDU_Yield();
			}
			/*MPIU_DBG_PRINTF((L"Opening process[%d]: %d\n", i, pSharedProcess[i].nPid));*/
			pg->ch.pSharedProcessHandles[i] =
				OpenProcess(STANDARD_RIGHTS_REQUIRED | PROCESS_VM_READ | PROCESS_VM_WRITE | PROCESS_VM_OPERATION, 
							FALSE, pSharedProcess[i].nPid);
			if (pg->ch.pSharedProcessHandles[i] == NULL)
			{
				int err = GetLastError();
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**OpenProcess", L"**OpenProcess %d %d", i, err); /*L"unable to open process %d, error %d\n", i, err);*/
				MPIDI_FUNC_EXIT(MPID_STATE_INITSHAREDPROCESSES);
				return mpi_errno;
			}
		}
		else
		{
			pg->ch.pSharedProcessHandles[i] = NULL;
		}
	}
	if (nRank == 0)
	{
		for (i=1; i<nProc; i++)
		{
			while (pSharedProcess[i].bFinished != TRUE)
				MPIDU_Yield();
		}
		/* Why are the fields erased here? */
		for (i=1; i<nProc; i++)
		{
			pSharedProcess[i].nPid = (DWORD)-1;
			pSharedProcess[i].bFinished = (BOOL)-1;
			pSharedProcess[i].nRank = -1;
		}
		pSharedProcess[0].nPid = (DWORD)-1;
		pSharedProcess[0].nRank = -1;
		pSharedProcess[0].bFinished = TRUE;
	}
	else
	{
		pSharedProcess[nRank].bFinished = TRUE;
		while (pSharedProcess[0].bFinished == FALSE)
			MPIDU_Yield();
	}
	MPIDI_FUNC_EXIT(MPID_STATE_INITSHAREDPROCESSES);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME shmMPIDI_CH3_Init_fns
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int shmMPIDI_CH3_Init_fns(MPIDI_CH3_Functions_t *function_table)
{
	function_table->Init = shmMPIDI_CH3_Init;
	function_table->PortFnsInit = shmMPIDI_CH3_PortFnsInit;
	function_table->RMAFnsInit = MPIDI_CH3U_RMAFnsInit;
	function_table->Finalize = shmMPIDI_CH3_Finalize;
	function_table->VC_Init = shmMPIDI_CH3_VC_Init;
	function_table->PG_Init = NULL;
	function_table->iStartMsg = shmMPIDI_CH3_iStartMsg;
	function_table->iStartMsgv = shmMPIDI_CH3_iStartMsgv;
	function_table->iSend = shmMPIDI_CH3_iSend;
	function_table->iSendv = shmMPIDI_CH3_iSendv;
	function_table->Progress_wait = shmMPIDI_CH3_Progress_wait;
	function_table->Progress_test = shmMPIDI_CH3_Progress_test;
	function_table->Connection_terminate = shmMPIDI_CH3_Connection_terminate;
	//function_table->Connect_to_root = shmMPIDI_CH3_Connect_to_root;
	function_table->Complete_unidirectional_connection_connect = MPIDI_CH3_Complete_unidirectional_connection_connect;
	function_table->Complete_unidirectional_connection_accept = MPIDI_CH3_Complete_unidirectional_connection_accept;
	function_table->Cleanup_after_connection = MPIDI_CH3_Cleanup_after_connection;
	function_table->iStartRndvMsg = shmMPIDI_CH3_iStartRndvMsg;
	function_table->iStartRndvTransfer = shmMPIDI_CH3_iStartRndvTransfer;
	//function_table->Get_business_card = shmMPIDI_CH3I_Get_business_card;
	function_table->VC_NeedsSendClose = MPIDI_CH3U_VC_NeedsSendClose;
	function_table->VC_SendClose = MPIDI_CH3U_VC_SendClose;
	return MPI_SUCCESS;
}
