/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"
#include "mpid_dataloop.h"
#include <stdlib.h>
#include <limits.h>

/* MPI datatype debugging helper routines.
 * 
 * The one you want to call is:
 *   MPIDU_Datatype_debug(MPI_Datatype type, int array_ct)
 *
 * The "array_ct" value tells the call how many array values to print
 * for struct, indexed, and blockindexed types.
 *
 */


void MPIDI_Datatype_dot_printf(MPI_Datatype type, int depth, int header);
void MPIDI_Dataloop_dot_printf(MPID_Dataloop *loop_p, int depth, int header);
void MPIDI_Datatype_contents_printf(MPI_Datatype type, int depth, int acount);
static wchar_t *MPIDI_Datatype_depth_spacing(int depth);

/* note: this isn't really "error handling" per se, but leave these comments
 * because Bill uses them for coverage analysis.
 */

/* --BEGIN ERROR HANDLING-- */
#undef FUNCNAME
#define FUNCNAME MPIDI_Datatype_dot_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_Datatype_dot_printf(MPI_Datatype type,
							   int depth,
							   int header)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_DATATYPE_DOT_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_DATATYPE_DOT_PRINTF);
	if (HANDLE_GET_KIND(type) == HANDLE_KIND_BUILTIN) {
		MPIU_DBG_OUT(DATATYPE, L"MPIDI_Datatype_dot_printf: type is a basic");
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_DOT_PRINTF);
		return;
	}
	else {
#ifdef USE_DBG_LOGGING /* With MPIU_DBG_XXX macros turned off MPIDI_Dataloop_dot_printf does nothing and the compiler complains about a function call with no effect. */
		MPID_Datatype *dt_p;
		MPID_Dataloop *loop_p;

		MPID_Datatype_get_ptr(type, dt_p);
		loop_p = dt_p->dataloop;

		MPIDI_Dataloop_dot_printf(loop_p, depth, header);
#else
		MPIU_UNREFERENCED_ARG(depth);
		MPIU_UNREFERENCED_ARG(header);
#endif
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_DOT_PRINTF);
		return;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_DOT_PRINTF);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Dataloop_dot_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_Dataloop_dot_printf(MPID_Dataloop *loop_p,
							   int depth,
							   int header)
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_DATALOOP_DOT_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_DATALOOP_DOT_PRINTF);

	if (header) {
		MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
								   L"digraph %p {   {", loop_p));
	}

	switch (loop_p->kind & DLOOP_KIND_MASK) {
		case DLOOP_KIND_CONTIG:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
			L"	  dl%d [shape = record, label = \"contig |{ ct = %d; el_sz = %d; el_ext = %d }\"];",
							depth,
							(int) loop_p->loop_params.c_t.count,
							(int) loop_p->el_size,
							(int) loop_p->el_extent));
			break;
		case DLOOP_KIND_VECTOR:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
			L"	  dl%d [shape = record, label = \"vector |{ ct = %d; blk = %d; str = %d; el_sz = %d; el_ext = %d }\"];",
							depth,
							(int) loop_p->loop_params.v_t.count,
							(int) loop_p->loop_params.v_t.blocksize,
							(int) loop_p->loop_params.v_t.stride,
							(int) loop_p->el_size,
							(int) loop_p->el_extent));
			break;
		case DLOOP_KIND_INDEXED:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
			L"	  dl%d [shape = record, label = \"indexed |{ ct = %d; tot_blks = %d; regions = ",
							depth,
							(int) loop_p->loop_params.i_t.count,
							(int) loop_p->loop_params.i_t.total_blocks));
			
			/* 3 picked as arbitrary cutoff */
			for (i=0; i < 3 && i < loop_p->loop_params.i_t.count; i++) {
				if (i + 1 < loop_p->loop_params.i_t.count) {
					/* more regions after this one */
					MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
					L"(%d, %d), ",
						  (int) loop_p->loop_params.i_t.offset_array[i],
						  (int) loop_p->loop_params.i_t.blocksize_array[i]));
				}
				else {
					MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
						   L"(%d, %d); ",
						   (int) loop_p->loop_params.i_t.offset_array[i],
						   (int) loop_p->loop_params.i_t.blocksize_array[i]));
				}
			}
			if (i < loop_p->loop_params.i_t.count) {
				MPIU_DBG_OUT(DATATYPE, L"...; ");
			}

			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
									   L"el_sz = %d; el_ext = %d }\"];\n",
									   (int) loop_p->el_size,
									   (int) loop_p->el_extent));
			break;
		case DLOOP_KIND_BLOCKINDEXED:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
			L"	  dl%d [shape = record, label = \"blockindexed |{ ct = %d; blk = %d; disps = ",
							depth,
							(int) loop_p->loop_params.bi_t.count,
							(int) loop_p->loop_params.bi_t.blocksize));
			
			/* 3 picked as arbitrary cutoff */
			for (i=0; i < 3 && i < loop_p->loop_params.bi_t.count; i++) {
				if (i + 1 < loop_p->loop_params.bi_t.count) {
					/* more regions after this one */
					MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
						L"%d, ",
						(int) loop_p->loop_params.bi_t.offset_array[i]));
				}
				else {
					MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
						 L"%d; ",
						 (int) loop_p->loop_params.bi_t.offset_array[i]));
				}
			}
			if (i < loop_p->loop_params.bi_t.count) {
				MPIU_DBG_OUT(DATATYPE, L"...; ");
			}

			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
									  L"el_sz = %d; el_ext = %d }\"];",
									   (int) loop_p->el_size,
									   (int) loop_p->el_extent));
			break;
		case DLOOP_KIND_STRUCT:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
			L"	  dl%d [shape = record, label = \"struct | {ct = %d; blks = ",
							depth,
							(int) loop_p->loop_params.s_t.count));
			for (i=0; i < 3 && i < loop_p->loop_params.s_t.count; i++) {
				if (i + 1 < loop_p->loop_params.s_t.count) {
					MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"%d, ",
							(int) loop_p->loop_params.s_t.blocksize_array[i]));
				}
				else {
					MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"%d; ",
							(int) loop_p->loop_params.s_t.blocksize_array[i]));
				}
			}
			if (i < loop_p->loop_params.s_t.count) {
				MPIU_DBG_OUT(DATATYPE, L"...; disps = ");
			}
			else {
				MPIU_DBG_OUT(DATATYPE, L"disps = ");
			}

			for (i=0; i < 3 && i < loop_p->loop_params.s_t.count; i++) {
				if (i + 1 < loop_p->loop_params.s_t.count) {
					MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"%d, ",
							(int) loop_p->loop_params.s_t.offset_array[i]));
				}
				else {
					MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"%d; ",
							(int) loop_p->loop_params.s_t.offset_array[i]));
				}
			}
			if (i < loop_p->loop_params.s_t.count) {
				MPIU_DBG_OUT(DATATYPE, L"... }\"];");
			}
			else {
				MPIU_DBG_OUT(DATATYPE, L"}\"];");
			}
			break;
		default:
			MPIU_Assert(0);
	}

	if (!(loop_p->kind & DLOOP_FINAL_MASK)) {
		/* more loops to go; recurse */
		MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
								   L"	  dl%d -> dl%d;\n", depth, depth + 1));
		switch (loop_p->kind & DLOOP_KIND_MASK) {
			case DLOOP_KIND_CONTIG:
				MPIDI_Dataloop_dot_printf(loop_p->loop_params.c_t.dataloop, depth + 1, 0);
				break;
			case DLOOP_KIND_VECTOR:
				MPIDI_Dataloop_dot_printf(loop_p->loop_params.v_t.dataloop, depth + 1, 0);
				break;
			case DLOOP_KIND_INDEXED:
				MPIDI_Dataloop_dot_printf(loop_p->loop_params.i_t.dataloop, depth + 1, 0);
				break;
			case DLOOP_KIND_BLOCKINDEXED:
				MPIDI_Dataloop_dot_printf(loop_p->loop_params.bi_t.dataloop, depth + 1, 0);
				break;
			case DLOOP_KIND_STRUCT:
				for (i=0; i < loop_p->loop_params.s_t.count; i++) {
					MPIDI_Dataloop_dot_printf(loop_p->loop_params.s_t.dataloop_array[i],
											  depth + 1, 0);
				}
				break;
			default:
				MPIU_DBG_OUT(DATATYPE, L"	  < unsupported type >");
		}
	}


	if (header) {
		MPIU_DBG_OUT(DATATYPE, L"   }\n}");
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATALOOP_DOT_PRINTF);
	return;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Datatype_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_Datatype_printf(MPI_Datatype type,
						   int depth,
						   MPI_Aint displacement,
						   int blocklength,
						   int header)
{
	int mpi_errno = MPI_SUCCESS;
	wchar_t *string;
	int size;
	MPI_Aint extent, true_lb, true_ub, lb, ub, sticky_lb, sticky_ub;
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_DATATYPE_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_DATATYPE_PRINTF);
	
	MPIU_THREADPRIV_GET;

	MPIU_UNREFERENCED_ARG(blocklength);
	MPIU_UNREFERENCED_ARG(depth);
	MPIU_UNREFERENCED_ARG(displacement);

	if (HANDLE_GET_KIND(type) == HANDLE_KIND_BUILTIN) {
		string = MPIDU_Datatype_builtin_to_string(type);
		if (type == MPI_LB) sticky_lb = 1;
		else sticky_lb = 0;
		if (type == MPI_UB) sticky_ub = 1;
		else sticky_ub = 0;
	}
	else {
		MPID_Datatype *type_ptr;

		MPID_Datatype_get_ptr(type, type_ptr);
		string = MPIDU_Datatype_combiner_to_string(type_ptr->contents->combiner);
		sticky_lb = type_ptr->has_sticky_lb;
		sticky_ub = type_ptr->has_sticky_ub;
	}

	NMPI_Type_size(mpi_errno, (type, &size));
	NMPI_Type_get_true_extent(mpi_errno, (type, &true_lb, &extent));
	true_ub = extent + true_lb;
	NMPI_Type_get_extent(mpi_errno, (type, &lb, &extent));
	ub = extent + lb;

	if (header == 1) {
		/*			   012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789 */
		MPIU_DBG_OUT(DATATYPE, L"------------------------------------------------------------------------------------------------------------------------------------------\n");
		MPIU_DBG_OUT(DATATYPE, L"depth				   type		 size	   extent	  true_lb	  true_ub		   lb(s)		   ub(s)		 disp	   blklen\n");
		MPIU_DBG_OUT(DATATYPE, L"------------------------------------------------------------------------------------------------------------------------------------------\n");
	}
	MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"%5d  %21s  %11d  %11d  %11d  %11d  %11d(%1d)  %11d(%1d)  %11d  %11d",
					depth,
					string,
					(int) size,
					(int) extent,
					(int) true_lb,
					(int) true_ub,
					(int) lb,
					(int) sticky_lb,
					(int) ub,
					(int) sticky_ub,
					(int) displacement,
					(int) blocklength));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_PRINTF);
	return;
}
/* --END ERROR HANDLING-- */

/* longest string is 21 characters */
wchar_t *MPIDU_Datatype_builtin_to_string(MPI_Datatype type)
{
	static wchar_t t_char[]			 = L"MPI_CHAR";
	static wchar_t t_uchar[]			= L"MPI_UNSIGNED_CHAR";
	static wchar_t t_byte[]			 = L"MPI_BYTE";
	static wchar_t t_wchar_t[]		  = L"MPI_WCHAR";
	static wchar_t t_short[]			= L"MPI_SHORT";
	static wchar_t t_ushort[]		   = L"MPI_UNSIGNED_SHORT";
	static wchar_t t_int[]			  = L"MPI_INT";
	static wchar_t t_uint[]			 = L"MPI_UNSIGNED";
	static wchar_t t_long[]			 = L"MPI_LONG";
	static wchar_t t_ulong[]			= L"MPI_UNSIGNED_LONG";
	static wchar_t t_float[]			= L"MPI_FLOAT";
	static wchar_t t_double[]		   = L"MPI_DOUBLE";
	static wchar_t t_longdouble[]	   = L"MPI_LONG_DOUBLE";
	static wchar_t t_longlongint[]	  = L"MPI_LONG_LONG_INT";
	static wchar_t t_longlong[]		 = L"MPI_LONG_LONG";
	static wchar_t t_ulonglong[]		= L"MPI_UNSIGNED_LONG_LONG";
	static wchar_t t_schar[]			= L"MPI_SIGNED_CHAR";

	static wchar_t t_packed[]		   = L"MPI_PACKED";
	static wchar_t t_lb[]			   = L"MPI_LB";
	static wchar_t t_ub[]			   = L"MPI_UB";

	static wchar_t t_floatint[]		 = L"MPI_FLOAT_INT";
	static wchar_t t_doubleint[]		= L"MPI_DOUBLE_INT";
	static wchar_t t_longint[]		  = L"MPI_LONG_INT";
	static wchar_t t_shortint[]		 = L"MPI_SHORT_INT";
	static wchar_t t_2int[]			 = L"MPI_2INT";
	static wchar_t t_longdoubleint[]	= L"MPI_LONG_DOUBLE_INT";

	static wchar_t t_complex[]		  = L"MPI_COMPLEX";
	static wchar_t t_doublecomplex[]	= L"MPI_DOUBLE_COMPLEX";
	static wchar_t t_logical[]		  = L"MPI_LOGICAL";
	static wchar_t t_real[]			 = L"MPI_REAL";
	static wchar_t t_doubleprecision[]  = L"MPI_DOUBLE_PRECISION";
	static wchar_t t_integer[]		  = L"MPI_INTEGER";
	static wchar_t t_2integer[]		 = L"MPI_2INTEGER";
	static wchar_t t_2complex[]		 = L"MPI_2COMPLEX";
	static wchar_t t_2doublecomplex[]   = L"MPI_2DOUBLE_COMPLEX";
	static wchar_t t_2real[]			= L"MPI_2REAL";
	static wchar_t t_2doubleprecision[] = L"MPI_2DOUBLE_PRECISION";
	static wchar_t t_character[]		= L"MPI_CHARACTER";

	if (type == MPI_CHAR)			  return t_char;
	if (type == MPI_UNSIGNED_CHAR)	 return t_uchar;
	if (type == MPI_SIGNED_CHAR)	   return t_schar;
	if (type == MPI_BYTE)			  return t_byte;
	if (type == MPI_WCHAR)			 return t_wchar_t;
	if (type == MPI_SHORT)			 return t_short;
	if (type == MPI_UNSIGNED_SHORT)	return t_ushort;
	if (type == MPI_INT)			   return t_int;
	if (type == MPI_UNSIGNED)		  return t_uint;
	if (type == MPI_LONG)			  return t_long;
	if (type == MPI_UNSIGNED_LONG)	 return t_ulong;
	if (type == MPI_FLOAT)			 return t_float;
	if (type == MPI_DOUBLE)			return t_double;
	if (type == MPI_LONG_DOUBLE)	   return t_longdouble;
	if (type == MPI_LONG_LONG_INT)	 return t_longlongint;
	if (type == MPI_LONG_LONG)		 return t_longlong;
	if (type == MPI_UNSIGNED_LONG_LONG) return t_ulonglong;
		
	if (type == MPI_PACKED)			return t_packed;
	if (type == MPI_LB)				return t_lb;
	if (type == MPI_UB)				return t_ub;
		
	if (type == MPI_FLOAT_INT)		 return t_floatint;
	if (type == MPI_DOUBLE_INT)		return t_doubleint;
	if (type == MPI_LONG_INT)		  return t_longint;
	if (type == MPI_SHORT_INT)		 return t_shortint;
	if (type == MPI_2INT)			  return t_2int;
	if (type == MPI_LONG_DOUBLE_INT)   return t_longdoubleint;
		
	if (type == MPI_COMPLEX)		   return t_complex;
	if (type == MPI_DOUBLE_COMPLEX)	return t_doublecomplex;
	if (type == MPI_LOGICAL)		   return t_logical;
	if (type == MPI_REAL)			  return t_real;
	if (type == MPI_DOUBLE_PRECISION)  return t_doubleprecision;
	if (type == MPI_INTEGER)		   return t_integer;
	if (type == MPI_2INTEGER)		  return t_2integer;
	if (type == MPI_2COMPLEX)		  return t_2complex;
	if (type == MPI_2DOUBLE_COMPLEX)   return t_2doublecomplex;
	if (type == MPI_2REAL)			 return t_2real;
	if (type == MPI_2DOUBLE_PRECISION) return t_2doubleprecision;
	if (type == MPI_CHARACTER)		 return t_character;
	
	return NULL;
}

/* MPIDU_Datatype_combiner_to_string(combiner)
 *
 * Converts a numeric combiner into a pointer to a string used for printing.
 *
 * longest string is 16 characters.
 */
#undef FUNCNAME
#define FUNCNAME MPIDU_Datatype_combiner_to_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
wchar_t *MPIDU_Datatype_combiner_to_string(int combiner)
{
	static wchar_t c_named[]	= L"named";
	static wchar_t c_contig[]   = L"contig";
	static wchar_t c_vector[]   = L"vector";
	static wchar_t c_hvector[]  = L"hvector";
	static wchar_t c_indexed[]  = L"indexed";
	static wchar_t c_hindexed[] = L"hindexed";
	static wchar_t c_struct[]   = L"struct";
	static wchar_t c_dup[]			  = L"dup";
	static wchar_t c_hvector_integer[]  = L"hvector_integer";
	static wchar_t c_hindexed_integer[] = L"hindexed_integer";
	static wchar_t c_indexed_block[]	= L"indexed_block";
	static wchar_t c_struct_integer[]   = L"struct_integer";
	static wchar_t c_subarray[]		 = L"subarray";
	static wchar_t c_darray[]		   = L"darray";
	static wchar_t c_f90_real[]		 = L"f90_real";
	static wchar_t c_f90_complex[]	  = L"f90_complex";
	static wchar_t c_f90_integer[]	  = L"f90_integer";
	static wchar_t c_resized[]		  = L"resized";

	if (combiner == MPI_COMBINER_NAMED)	  return c_named;
	if (combiner == MPI_COMBINER_CONTIGUOUS) return c_contig;
	if (combiner == MPI_COMBINER_VECTOR)	 return c_vector;
	if (combiner == MPI_COMBINER_HVECTOR)	return c_hvector;
	if (combiner == MPI_COMBINER_INDEXED)	return c_indexed;
	if (combiner == MPI_COMBINER_HINDEXED)   return c_hindexed;
	if (combiner == MPI_COMBINER_STRUCT)	 return c_struct;
	if (combiner == MPI_COMBINER_DUP)			  return c_dup;
	if (combiner == MPI_COMBINER_HVECTOR_INTEGER)  return c_hvector_integer;
	if (combiner == MPI_COMBINER_HINDEXED_INTEGER) return c_hindexed_integer;
	if (combiner == MPI_COMBINER_INDEXED_BLOCK)	return c_indexed_block;
	if (combiner == MPI_COMBINER_STRUCT_INTEGER)   return c_struct_integer;
	if (combiner == MPI_COMBINER_SUBARRAY)		 return c_subarray;
	if (combiner == MPI_COMBINER_DARRAY)		   return c_darray;
	if (combiner == MPI_COMBINER_F90_REAL)		 return c_f90_real;
	if (combiner == MPI_COMBINER_F90_COMPLEX)	  return c_f90_complex;
	if (combiner == MPI_COMBINER_F90_INTEGER)	  return c_f90_integer;
	if (combiner == MPI_COMBINER_RESIZED)		  return c_resized;
	
	return NULL;
}

/* --BEGIN ERROR HANDLING-- */
#undef FUNCNAME
#define FUNCNAME MPIDU_Datatype_debug
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDU_Datatype_debug(MPI_Datatype type,
						  int array_ct)
{
	int is_builtin;
	MPID_Datatype *dtp;
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_DATATYPE_DEBUG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_DATATYPE_DEBUG);

	is_builtin = (HANDLE_GET_KIND(type) == HANDLE_KIND_BUILTIN);

	MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
			   L"# MPIU_Datatype_debug: MPI_Datatype = 0x%0x (%s)", type,
			   (is_builtin) ? MPIDU_Datatype_builtin_to_string(type) :
				L"derived"));

	if (is_builtin)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_DATATYPE_DEBUG);
		return;
	}

	MPID_Datatype_get_ptr(type, dtp);

	MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
	  L"# Size = %d, Extent = %d, LB = %d%s, UB = %d%s, Extent = %d, Element Size = %d (%s), %s",
					(int) dtp->size,
					(int) dtp->extent,
					(int) dtp->lb,
					(dtp->has_sticky_lb) ? L"(sticky)" : L"",
					(int) dtp->ub,
					(dtp->has_sticky_ub) ? L"(sticky)" : L"",
					(int) dtp->extent,
					(int) dtp->element_size,
					dtp->element_size == -1 ? L"multiple types" :
					MPIDU_Datatype_builtin_to_string(dtp->eltype),
					dtp->is_contig ? L"is N contig" : L"is not N contig"));

	MPIU_DBG_OUT(DATATYPE, L"# Contents:");
	MPIDI_Datatype_contents_printf(type, 0, array_ct);

	MPIU_DBG_OUT(DATATYPE, L"# Dataloop:");
	MPIDI_Datatype_dot_printf(type, 0, 1);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_DATATYPE_DEBUG);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Datatype_depth_spacing
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static wchar_t *MPIDI_Datatype_depth_spacing(int depth)
{
	static wchar_t d0[] = L"";
	static wchar_t d1[] = L"  ";
	static wchar_t d2[] = L"	";
	static wchar_t d3[] = L"	  ";
	static wchar_t d4[] = L"		";
	static wchar_t d5[] = L"		  ";

	switch (depth) {
		case 0: return d0;
		case 1: return d1;
		case 2: return d2;
		case 3: return d3;
		case 4: return d4;
		default: return d5;
	}
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Datatype_contents_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_Datatype_contents_printf(MPI_Datatype type,
									int depth,
									int acount)
{
	int i;
	MPID_Datatype *dtp;
	MPID_Datatype_contents *cp;
	MPI_Aint *aints;
	MPI_Datatype *types;
	int *ints;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);

	if (HANDLE_GET_KIND(type) == HANDLE_KIND_BUILTIN) {
		MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"# %stype: %s\n",
						MPIDI_Datatype_depth_spacing(depth),
						MPIDU_Datatype_builtin_to_string(type)));
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
		return;
	}

	MPID_Datatype_get_ptr(type, dtp);
	cp = dtp->contents;

	types = (MPI_Datatype *) (((char *) cp) +
							  sizeof(MPID_Datatype_contents));
	ints  = (int *) (((char *) types) +
					 cp->nr_types * sizeof(MPI_Datatype));
	aints = (MPI_Aint *) (((char *) ints) +
						  cp->nr_ints * sizeof(int));

	MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"# %scombiner: %s",
					MPIDI_Datatype_depth_spacing(depth),
					MPIDU_Datatype_combiner_to_string(cp->combiner)));

	switch (cp->combiner) {
		case MPI_COMBINER_NAMED:
		case MPI_COMBINER_DUP:
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
		case MPI_COMBINER_RESIZED:
			/* not done */
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
		case MPI_COMBINER_CONTIGUOUS:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"# %scontig ct = %d\n", 
							MPIDI_Datatype_depth_spacing(depth),
									   *ints));
			MPIDI_Datatype_contents_printf(*types,
										   depth + 1,
										   acount);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
		case MPI_COMBINER_VECTOR:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
						L"# %svector ct = %d, blk = %d, str = %d\n",
						MPIDI_Datatype_depth_spacing(depth), 
							ints[0],
							ints[1],
							ints[2]));
			MPIDI_Datatype_contents_printf(*types,
										   depth + 1,
										   acount);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
		case MPI_COMBINER_HVECTOR:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
						  L"# %shvector ct = %d, blk = %d, str = %d\n",
							MPIDI_Datatype_depth_spacing(depth),
							ints[0],
							ints[1],
							(int) aints[0]));
			MPIDI_Datatype_contents_printf(*types,
										   depth + 1,
										   acount);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
		case MPI_COMBINER_INDEXED:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"# %sindexed ct = %d:",
							MPIDI_Datatype_depth_spacing(depth),
							ints[0]));
			for (i=0; i < acount && i < ints[0]; i++) {
				MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
						 L"# %s  indexed [%d]: blk = %d, disp = %d\n",
								MPIDI_Datatype_depth_spacing(depth),
								i,
								ints[i+1],
								ints[i+(cp->nr_ints/2)+1]));
				MPIDI_Datatype_contents_printf(*types,
											   depth + 1,
											   acount);
			}
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
		case MPI_COMBINER_HINDEXED:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"# %shindexed ct = %d:",
							MPIDI_Datatype_depth_spacing(depth),
							ints[0]));
			for (i=0; i < acount && i < ints[0]; i++) {
				MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
							L"# %s  hindexed [%d]: blk = %d, disp = %d\n",
								MPIDI_Datatype_depth_spacing(depth),
								i,
								(int) ints[i+1],
								(int) aints[i]));
				MPIDI_Datatype_contents_printf(*types,
											   depth + 1,
											   acount);
			}
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
		case MPI_COMBINER_STRUCT:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"# %sstruct ct = %d:",
							MPIDI_Datatype_depth_spacing(depth),
							(int) ints[0]));
			for (i=0; i < acount && i < ints[0]; i++) {
				MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST,
						   L"# %s  struct[%d]: blk = %d, disp = %d\n",
								MPIDI_Datatype_depth_spacing(depth),
								i,
								(int) ints[i+1],
								(int) aints[i]));
				MPIDI_Datatype_contents_printf(types[i],
											   depth + 1,
											   acount);
			}
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
		default:
			MPIU_DBG_OUT_FMT(DATATYPE,(MPIU_DBG_FDEST, L"# %sunhandled combiner",
						MPIDI_Datatype_depth_spacing(depth)));
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
			return;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DATATYPE_CONTENTS_PRINTF);
}
/* --END ERROR HANDLING-- */
