// This is the main DLL file.

#include "stdafx.h"
#include "DeinoLP.h"
#include <stdio.h>
#include <userenv.h>

#define MAX_ERROR_LENGTH 8192
#define MAX_WIN_ERROR_LENGTH 1024

namespace DeinoLP
{

ProcessCreation::ProcessCreation()
{
    hProcess = INVALID_HANDLE_VALUE;
    hRead = INVALID_HANDLE_VALUE;
    hWrite = INVALID_HANDLE_VALUE;
    output_list = NULL;
}

ProcessCreation::~ProcessCreation()
{
    if (hRead != INVALID_HANDLE_VALUE)
    {
	CloseHandle(hRead);
	hRead = INVALID_HANDLE_VALUE;
    }
    if (hWrite != INVALID_HANDLE_VALUE)
    {
	CloseHandle(hWrite);
	hWrite = INVALID_HANDLE_VALUE;
    }
    while (output_list != NULL)
    {
	BufferNode *node = output_list;
	output_list = output_list->next;
	if (node->buffer != NULL)
	    delete [] node->buffer;
	delete node;
    }
    if (hProcess != INVALID_HANDLE_VALUE)
    {
	TerminateProcess(hProcess, -1);
	hProcess = INVALID_HANDLE_VALUE;
    }
}

static wchar_t *encode_handle(wchar_t *str, int len, HANDLE h)
{
    _snwprintf_s(str, len, len, L"%p", h);
    return str;
}

static HANDLE decode_handle(wchar_t *str)
{
    HANDLE p;
    swscanf_s(str, L"%p", &p);
    return p;
}

static wchar_t *DupString(System::String ^str)
{
    wchar_t *result;
    size_t len;

    if (str == nullptr)
	return NULL;

    len = str->Length;
    result = new wchar_t[str->Length + 1];
    for (size_t i=0; i<len; i++)
    {
	result[(int)i] = str[(int)i];
    }
    result[len] = L'\0';
    return result;
}

static int get_user_handle(wchar_t *account, wchar_t *domain, wchar_t *password, HANDLE *handle_ptr)
{
    HANDLE hUser;
    int error;
    int num_tries = 3;

    if (domain != NULL)
    {
	wprintf(L"LogonUser(%s\\%s)\n", domain, account);
    }
    else
    {
	wprintf(L"LogonUser(%s)\n", account);
    }

    /* logon the user */
    while (!LogonUser(
	account,
	domain,
	password,
	LOGON32_LOGON_INTERACTIVE,
	LOGON32_PROVIDER_DEFAULT,
	&hUser))
    {
	error = GetLastError();
	if (error == ERROR_NO_LOGON_SERVERS)
	{
	    if (num_tries)
		Sleep(250);
	    else
	    {
		*handle_ptr = INVALID_HANDLE_VALUE;
		return error;
	    }
	    num_tries--;
	}
	else
	{
	    *handle_ptr = INVALID_HANDLE_VALUE;
	    return error;
	}
    }

    *handle_ptr = hUser;
    return 0;
}

static int get_hostname(wchar_t *host, int length)
{
    DWORD len = length;
    if (!GetComputerNameEx(ComputerNameDnsFullyQualified, host, &len))
    {
	return GetLastError();
    }
    return 0;
}

static void translate_win_error(int error, wchar_t *msg, int maxlen)
{
    HLOCAL str;
    int num_characters;

    num_characters = FormatMessage(
	FORMAT_MESSAGE_FROM_SYSTEM |
	FORMAT_MESSAGE_ALLOCATE_BUFFER,
	0,
	error,
	MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
	(LPTSTR) &str,
	0,0);
    memcpy(msg, str, min(maxlen, num_characters+1)*sizeof(wchar_t));
    LocalFree(str);
}

void ProcessCreation::lock()
{
    DWORD result;
    wchar_t err_msg[MAX_ERROR_LENGTH];

    if (hLPLock == NULL)
    {
	hLPLock = CreateMutex(NULL, FALSE, L"DeinoLP launch mutex");
    }
    result = WaitForSingleObject(hLPLock, INFINITE);
    if (result == WAIT_FAILED)
    {
	result = GetLastError();
	translate_win_error(result, err_msg, MAX_ERROR_LENGTH);
	wprintf(L"Error waiting for hLPLock %d, %s\n", result, err_msg);
	return;
    }
    if (result == WAIT_OBJECT_0)
	return;
    wprintf(L"WaitForSingleObject(\"DeinoLP launch mutex\") failed: %d\n", result);
}

void ProcessCreation::unlock()
{
    if (!ReleaseMutex(hLPLock))
    {
	DWORD error = GetLastError();
	wchar_t msg[1024];
	translate_win_error(error, msg, 1024);
	wprintf(L"ReleaseMutex(\"DeinoLP launch mutex\") failed: %s\n", msg);
    }
}

//#define USE_REDIRECTED_OUTPUT
#ifdef USE_REDIRECTED_OUTPUT
int ProcessCreation::launch_process(wchar_t *cmd, wchar_t *domain, wchar_t *user, wchar_t *password)
{
    HANDLE hStdin = INVALID_HANDLE_VALUE, hStdout = INVALID_HANDLE_VALUE, hStderr = INVALID_HANDLE_VALUE;
    HANDLE hInTemp = INVALID_HANDLE_VALUE, hOutTemp = INVALID_HANDLE_VALUE, hErrTemp = INVALID_HANDLE_VALUE;
    HANDLE hPipeStdinR = NULL, hPipeStdinW = NULL;
    HANDLE hPipeStdoutR = NULL, hPipeStdoutW = NULL;
    HANDLE hPipeStderrR = NULL, hPipeStderrW = NULL;
    STARTUPINFO saInfo;
    PROCESS_INFORMATION psInfo = { 0 };
    DWORD launch_flag = 0;
    int nError = 0, result = 0;
    SECURITY_ATTRIBUTES saAttr;
    wchar_t str[MAX_ERROR_LENGTH] = L"";
    BOOL bSuccess = TRUE;
    wchar_t err_msg[MAX_ERROR_LENGTH] = L"";
    DWORD resume_count;
    HANDLE user_handle;

    /* Initialize the psInfo structure in case there is an error and we jump to CLEANUP before psInfo is set */
    psInfo.hProcess = INVALID_HANDLE_VALUE;

    /* Save stdin, stdout, and stderr */
    lock();
    hStdin = GetStdHandle(STD_INPUT_HANDLE);
    if (hStdin == INVALID_HANDLE_VALUE)
    {
	nError = GetLastError();
	unlock();
	translate_win_error(nError, err_msg, MAX_ERROR_LENGTH);
	swprintf_s(str, MAX_ERROR_LENGTH, L"GetStdHandle(STD_INPUT_HANDLE) failed, error %d, %s\n", nError, err_msg);
	throw gcnew System::Exception(gcnew System::String(str));
	return -1;
    }
    hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
    if (hStdout == INVALID_HANDLE_VALUE)
    {
	nError = GetLastError();
	unlock();
	translate_win_error(nError, err_msg, MAX_ERROR_LENGTH);
	swprintf_s(str, MAX_ERROR_LENGTH, L"GetStdHandle(STD_OUTPUT_HANDLE) failed, error %d, %s\n", nError, err_msg);
	throw gcnew System::Exception(gcnew System::String(str));
	return -1;
    }
    hStderr = GetStdHandle(STD_ERROR_HANDLE);
    if (hStderr == INVALID_HANDLE_VALUE)
    {
	nError = GetLastError();
	unlock();
	translate_win_error(nError, err_msg, MAX_ERROR_LENGTH);
	swprintf_s(str, MAX_ERROR_LENGTH, L"GetStdHandle(STD_ERROR_HANDLE) failed, error %d, %s\n", nError, err_msg);
	throw gcnew System::Exception(gcnew System::String(str));
	return -1;
    }

    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.lpSecurityDescriptor = NULL;
    saAttr.bInheritHandle = TRUE;

    /* Create the pipes for stdout, stderr */
    if (!CreatePipe(&hPipeStdinR, &hPipeStdinW, &saAttr, 0))
    {
	swprintf_s(str, MAX_ERROR_LENGTH, L"CreatePipe(stdin) failed, error %d\n", GetLastError());
	goto CLEANUP;
    }
    if (!CreatePipe(&hPipeStdoutR, &hPipeStdoutW, &saAttr, 0))
    {
	swprintf_s(str, MAX_ERROR_LENGTH, L"CreatePipe(stdout) failed, error %d\n", GetLastError());
	goto CLEANUP;
    }
    /*
    if (!CreatePipe(&hPipeStderrR, &hPipeStderrW, &saAttr, 0))
    {
	swprintf_s(str, MAX_ERROR_LENGTH, L"CreatePipe(stderr) failed, error %d\n", GetLastError());
	goto CLEANUP;
    }
    */

    /* Make the ends of the pipes that this process will use not inheritable */
    if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hPipeStdinW, GetCurrentProcess(), &hInTemp, 
	0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
    {
	nError = GetLastError();
	swprintf_s(str, MAX_ERROR_LENGTH, L"DuplicateHandle failed, error %d\n", nError);
	goto CLEANUP;
    }
    hWrite = hInTemp;
    if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hPipeStdoutR, GetCurrentProcess(), &hOutTemp, 
	0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
    {
	nError = GetLastError();
	swprintf_s(str, MAX_ERROR_LENGTH, L"DuplicateHandle failed, error %d\n", nError);
	goto CLEANUP;
    }
    hRead = hOutTemp;
    /*
    if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hPipeStderrR, GetCurrentProcess(), &hErrTemp, 
	0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
    {
	nError = GetLastError();
	swprintf_s(str, MAX_ERROR_LENGTH, L"DuplicateHandle failed, error %d\n", nError);
	goto CLEANUP;
    }
    hErr = hErrTemp;
    */

    /* Set stdin, stdout, and stderr to the ends of the pipe the created process will use */
    if (!SetStdHandle(STD_INPUT_HANDLE, (HANDLE)hPipeStdinR))
    {
	nError = GetLastError();
	swprintf_s(str, MAX_ERROR_LENGTH, L"SetStdHandle failed, error %d\n", nError);
	goto CLEANUP;
    }
    if (!SetStdHandle(STD_OUTPUT_HANDLE, (HANDLE)hPipeStdoutW))
    {
	nError = GetLastError();
	swprintf_s(str, MAX_ERROR_LENGTH, L"SetStdHandle failed, error %d\n", nError);
	goto RESTORE_CLEANUP;
    }
    if (!SetStdHandle(STD_ERROR_HANDLE, (HANDLE)hPipeStdoutW/*hPipeStderrW*/))
    {
	nError = GetLastError();
	swprintf_s(str, MAX_ERROR_LENGTH, L"SetStdHandle failed, error %d\n", nError);
	goto RESTORE_CLEANUP;
    }

    /* Create the process */
    memset(&saInfo, 0, sizeof(STARTUPINFO));
    saInfo.cb = sizeof(STARTUPINFO);
    saInfo.hStdInput = (HANDLE)hPipeStdinR;
    saInfo.hStdOutput = (HANDLE)hPipeStdoutW;
    saInfo.hStdError = (HANDLE)hPipeStdoutW/*hPipeStderrW*/;
    saInfo.dwFlags = STARTF_USESTDHANDLES;

    launch_flag = 
	CREATE_SUSPENDED | /*CREATE_NEW_CONSOLE*/ /*DETACHED_PROCESS*/ CREATE_NO_WINDOW | /*priorityClass*/NORMAL_PRIORITY_CLASS | CREATE_UNICODE_ENVIRONMENT;
    //if (dbg) launch_flag = launch_flag | DEBUG_PROCESS;

    if (user != NULL)
    {
	result = get_user_handle(user, domain, password, &user_handle);
	if (user_handle == INVALID_HANDLE_VALUE)
	{
	    swprintf_s(str, MAX_ERROR_LENGTH, L"get_user_handle failed, error %d.\n", result);
	    goto RESTORE_CLEANUP;
	}

	result = 0;
	if (!ImpersonateLoggedOnUser(user_handle))
	{
	    result = GetLastError();
	    swprintf_s(str, MAX_ERROR_LENGTH, L"ImpersonateLoggedOnUser failed, error %d\n", result);
	    goto RESTORE_CLEANUP;
	}

	wprintf(L"CreateProcessAsUser(%s)\n", cmd);
	psInfo.hProcess = INVALID_HANDLE_VALUE;
	if (CreateProcessAsUser(
	    user_handle,
	    NULL,
	    cmd,
	    NULL, NULL, TRUE,
	    launch_flag,
	    NULL,
	    NULL,
	    &saInfo, &psInfo))
	{
	    //SetThreadPriority(psInfo.hThread, THREAD_PRIORITY_NORMAL);
	    //process->pid = psInfo.dwProcessId;
	}
	else
	{
	    nError = GetLastError();
	    translate_win_error(nError, str, MAX_ERROR_LENGTH);
	    swprintf_s(str, MAX_ERROR_LENGTH, L"CreateProcessAsUser('%s') failed, error %d - %s\n", cmd, nError, str);
	    psInfo.hProcess = INVALID_HANDLE_VALUE;
	    bSuccess = FALSE;
	}

	RevertToSelf();
    }
    else
    {
	wprintf(L"CreateProcess(%s)\n", cmd);
	psInfo.hProcess = INVALID_HANDLE_VALUE;
	if (CreateProcess(
	    NULL,
	    cmd,
	    NULL, NULL, TRUE,
	    launch_flag,
	    NULL,
	    NULL,
	    &saInfo, &psInfo))
	{
	    //SetThreadPriority(psInfo.hThread, THREAD_PRIORITY_NORMAL);
	    //process->pid = psInfo.dwProcessId;
	}
	else
	{
	    nError = GetLastError();
	    translate_win_error(nError, str, MAX_ERROR_LENGTH);
	    swprintf_s(str, MAX_ERROR_LENGTH, L"CreateProcess('%s') failed, error %d - %s\n", cmd, nError, str);
	    psInfo.hProcess = INVALID_HANDLE_VALUE;
	    bSuccess = FALSE;
	}
    }

    if (!bSuccess)
    {
	/* close all the handles allocated in this function */
	CloseHandle(hRead);
	CloseHandle(hWrite);
	//CloseHandle(hErr);
    }

RESTORE_CLEANUP:
    /* Restore stdin, stdout, stderr */
    SetStdHandle(STD_INPUT_HANDLE, hStdin);
    SetStdHandle(STD_OUTPUT_HANDLE, hStdout);
    SetStdHandle(STD_ERROR_HANDLE, hStderr);

CLEANUP:
    unlock();
    CloseHandle((HANDLE)hPipeStdinR);
    CloseHandle((HANDLE)hPipeStdoutW);
    //CloseHandle((HANDLE)hPipeStderrW);

    if (psInfo.hProcess != INVALID_HANDLE_VALUE)
    {
	hProcess = psInfo.hProcess;
	hThread = psInfo.hThread;
	resume_count = ResumeThread(psInfo.hThread);
	if (resume_count == (DWORD)-1)
	{
	    swprintf_s(str, MAX_ERROR_LENGTH, L"unable to resume the main thread after CreateProcess.  Error %d\n", GetLastError());
	    throw gcnew System::Exception(gcnew System::String(str));
	}
	if (resume_count > 1)
	    wprintf(L"Resume count = %d after ResumeThread.\n", resume_count);
	return 0;
    }

    throw gcnew System::Exception(gcnew System::String(str));
    return -1;
}
#else
int ProcessCreation::launch_process(wchar_t *cmd_orig, wchar_t *domain, wchar_t *user, wchar_t *password)
{
    wchar_t cmd[8192];
    wchar_t read_handle_str[20], write_handle_str[20];
    HANDLE hReadTemp, hWriteTemp, hReadRemote, hWriteRemote;
    STARTUPINFO saInfo;
    PROCESS_INFORMATION psInfo = { 0 };
    DWORD launch_flag = 0;
    int nError = 0, result = 0;
    SECURITY_ATTRIBUTES saAttr;
    wchar_t str[MAX_ERROR_LENGTH] = L"";
    wchar_t win_error_str[MAX_WIN_ERROR_LENGTH] = L"";
    BOOL bSuccess = TRUE;
    wchar_t err_msg[MAX_ERROR_LENGTH] = L"";
    DWORD resume_count;
    HANDLE user_handle;
    PROFILEINFO info;
    LPVOID pEnv;

    /* Initialize the psInfo structure in case there is an error and we jump to CLEANUP before psInfo is set */
    psInfo.hProcess = INVALID_HANDLE_VALUE;

    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.lpSecurityDescriptor = NULL;
    saAttr.bInheritHandle = TRUE;

    /* create a pipe to send data to the manager through */
    if (!CreatePipe(&hReadTemp, &hWriteRemote, &saAttr, 0))
    {
	nError = GetLastError();
	swprintf_s(str, MAX_ERROR_LENGTH, L"CreatePipe failed, error %d\n", nError);
	goto CLEANUP;
    }
    /* prevent the local read end of the pipe from being inherited */
    if (!DuplicateHandle(GetCurrentProcess(), hReadTemp, GetCurrentProcess(), &hReadTemp, 0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
    {
	nError = GetLastError();
	CloseHandle(hReadTemp);
	CloseHandle(hWriteRemote);
	swprintf_s(str, MAX_ERROR_LENGTH, L"Unable to duplicate the read end of the pipe, error %d\n", nError);
	goto CLEANUP;
    }
    hRead = hReadTemp;
    /* create a pipe to read data from the manager through */
    if (!CreatePipe(&hReadRemote, &hWriteTemp, &saAttr, 0))
    {
	nError = GetLastError();
	CloseHandle(hReadTemp);
	CloseHandle(hWriteRemote);
	swprintf_s(str, MAX_ERROR_LENGTH, L"CreatePipe failed, error %d\n", nError);
	goto CLEANUP;
    }
    /* prevent the local write end of the pipe from being inherited */
    if (!DuplicateHandle(GetCurrentProcess(), hWriteTemp, GetCurrentProcess(), &hWriteTemp, 0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
    {
	nError = GetLastError();
	CloseHandle(hReadTemp);
	CloseHandle(hWriteTemp);
	CloseHandle(hReadRemote);
	CloseHandle(hWriteRemote);
	swprintf_s(str, MAX_ERROR_LENGTH, L"Unable to duplicate the read end of the pipe, error %d\n", nError);
	goto CLEANUP;
    }
    hWrite = hWriteTemp;

    _snwprintf_s(cmd, 8192, 8192, L"%s %s %s", cmd_orig, encode_handle(read_handle_str, 20, hReadRemote), encode_handle(write_handle_str, 20, hWriteRemote));

    /* Create the process */
    GetStartupInfo(&saInfo);

    launch_flag = CREATE_SUSPENDED | CREATE_NO_WINDOW | NORMAL_PRIORITY_CLASS | CREATE_UNICODE_ENVIRONMENT;

    if (user != NULL)
    {
	result = get_user_handle(user, domain, password, &user_handle);
	if (user_handle == INVALID_HANDLE_VALUE)
	{
	    translate_win_error(result, win_error_str, MAX_WIN_ERROR_LENGTH);
	    swprintf_s(str, MAX_ERROR_LENGTH, L"get_user_handle failed, error %d: %s.\n", result, win_error_str);
	    goto CLEANUP;
	}

	/* Try to load the user's profile */
	memset(&info, 0, sizeof(PROFILEINFO));
	info.dwSize = sizeof(PROFILEINFO);
	info.lpUserName = user;
	info.dwFlags = PI_NOUI;
	LoadUserProfile(user_handle, &info);

	if (!CreateEnvironmentBlock(&pEnv, user_handle, FALSE))
	{
	    pEnv = NULL;
	}

	result = 0;
	if (!ImpersonateLoggedOnUser(user_handle))
	{
	    result = GetLastError();
	    translate_win_error(result, win_error_str, MAX_WIN_ERROR_LENGTH);
	    swprintf_s(str, MAX_ERROR_LENGTH, L"ImpersonateLoggedOnUser failed, error %d: %s.\n", result, win_error_str);
	    goto CLEANUP;
	}

	wprintf(L"CreateProcessAsUser(%s)\n", cmd);
	psInfo.hProcess = INVALID_HANDLE_VALUE;
	if (!CreateProcessAsUser(
	    user_handle,
	    NULL,
	    cmd,
	    NULL, NULL, TRUE,
	    launch_flag,
	    pEnv,
	    NULL,
	    &saInfo, &psInfo))
	{
	    nError = GetLastError();
	    translate_win_error(nError, win_error_str, MAX_WIN_ERROR_LENGTH);
	    swprintf_s(str, MAX_ERROR_LENGTH, L"CreateProcessAsUser('%s') failed, error %d - %s\n", cmd, nError, win_error_str);
	    psInfo.hProcess = INVALID_HANDLE_VALUE;
	    bSuccess = FALSE;
	}

	RevertToSelf();
    }
    else
    {
	wprintf(L"CreateProcess(%s)\n", cmd);
	psInfo.hProcess = INVALID_HANDLE_VALUE;
	if (!CreateProcess(
	    NULL,
	    cmd,
	    NULL, NULL, TRUE,
	    launch_flag,
	    NULL,
	    NULL,
	    &saInfo, &psInfo))
	{
	    nError = GetLastError();
	    translate_win_error(nError, win_error_str, MAX_WIN_ERROR_LENGTH);
	    swprintf_s(str, MAX_ERROR_LENGTH, L"CreateProcess('%s') failed, error %d - %s\n", cmd, nError, win_error_str);
	    psInfo.hProcess = INVALID_HANDLE_VALUE;
	    bSuccess = FALSE;
	}
    }

    if (!bSuccess)
    {
	/* close all the handles allocated in this function */
	CloseHandle(hRead);
	CloseHandle(hWrite);
	CloseHandle(hReadRemote);
	CloseHandle(hWriteRemote);
    }

CLEANUP:
    CloseHandle(hReadRemote);
    CloseHandle(hWriteRemote);

    if (psInfo.hProcess != INVALID_HANDLE_VALUE)
    {
	hProcess = psInfo.hProcess;
	hThread = psInfo.hThread;
	resume_count = ResumeThread(psInfo.hThread);
	if (resume_count == (DWORD)-1)
	{
	    nError = GetLastError();
	    translate_win_error(nError, win_error_str, MAX_WIN_ERROR_LENGTH);
	    swprintf_s(str, MAX_ERROR_LENGTH, L"Unable to resume the main thread after CreateProcess.  Error %d: %s.\n", nError, win_error_str);
	    throw gcnew System::Exception(gcnew System::String(str));
	}
	if (resume_count > 1)
	    wprintf(L"Resume count = %d after ResumeThread.\n", resume_count);
	return 0;
    }

    throw gcnew System::Exception(gcnew System::String(str));
    return -1;
}
#endif

void ProcessCreation::DeinoLaunchProcess(System::String ^cmd_str, System::String ^domain_str, System::String ^user_str, System::String ^password_str)
{
    wchar_t *cmd = NULL, *domain = NULL, *user = NULL, *password = NULL;
    if (cmd_str == nullptr)
	return;

    cmd = DupString(cmd_str);
    domain = DupString(domain_str);
    user = DupString(user_str);
    password = DupString(password_str);
    //wprintf(L"cmd: '%s'\ndomain: '%s'\nuser: '%s'\n\n", cmd, domain, user);

    int result = launch_process(cmd, domain, user, password);

    if (cmd != NULL)
	delete [] cmd;
    if (domain != NULL)
	delete [] domain;
    if (user != NULL)
	delete [] user;
    if (password != NULL)
	delete [] password;

    if (result != 0)
    {
	throw gcnew System::Exception(System::String::Format(L"Unable to start the process using the provided credentials: {0}", user_str));
    }
}

void WriteToLog(wchar_t *str)
{
    /*
    FILE *fout;
    int error = _wfopen_s(&fout, L"C:\\temp\\manager.log", L"a+");
    if (error == 0)
    {
	fwprintf(fout, L"%s\n", str);
	fclose(fout);
    }
    */
}

bool ProcessCreation::ReadNextBuffer()
{
    wchar_t line[8192];
    DWORD num_read = 0;
    int i;
    BufferNode *node = NULL;
    DWORD error;
    //int length;

    /*
    if (ReadFile(hRead, &length, sizeof(int), &num_read, NULL))
    {
	wchar_t msg[100];
	swprintf_s(msg, 100, L"Read length: %d", length);
	WriteToLog(msg);
	*/
	if (!ReadFile(hRead, line, 8192 * sizeof(wchar_t), &num_read, NULL))
	//if (!ReadFile(hRead, line, length, &num_read, NULL))
	{
	    error = GetLastError();
	    /* There was an error but some bytes were read */
	    if (num_read > 0)
	    {
		wchar_t msg[100];
		swprintf_s(msg, 100, L"Read %d bytes", num_read);
		WriteToLog(msg);
		WriteToLog(line);
	    }
	    else
	    {
		wprintf(L"ReadFile failed, error %d\n", error);
		swprintf_s(line, 8192, L"ReadFile failed, error %d", error);
		throw gcnew System::Exception(gcnew System::String(line));
	    }
	}
	else if (num_read < 1)
	{
	    wprintf(L"ReadFile returned %d bytes\n", num_read);
	    return false;
	}
	else
	{
	    wchar_t msg[100];
	    swprintf_s(msg, 100, L"Read %d bytes", num_read);
	    WriteToLog(msg);
	    WriteToLog(line);
	}
	/*
    }
    else
    {
	error = GetLastError();
	wprintf(L"ReadFile failed, error %d\n", error);
	swprintf_s(line, 8192, L"ReadFile failed, error %d", error);
	throw gcnew System::Exception(gcnew System::String(line));
    }
    */

    // Is the data read from ReadFile ASCII or UNICODE?
    num_read = num_read / sizeof(wchar_t);

    node = new BufferNode();
    node->buffer = new wchar_t[num_read];
    node->curpos = 0;
    node->length = num_read;
    node->next = NULL;
    for (i=0; i<(int)num_read; i++)
	node->buffer[i] = line[i];

    if (output_list == NULL)
    {
	output_list = node;
    }
    else
    {
	BufferNode *iter;
	iter = output_list;
	while (iter->next != NULL)
	    iter = iter->next;
	iter->next = node;
    }
    return true;
}

System::String^ ProcessCreation::DeinoLPReadLine()
{
    int i, count, curpos;
    bool bCrLfFound = false;
    BufferNode *iter, *dead;

    do
    {
	iter = output_list;
	while (iter != NULL)
	{
	    for (i=iter->curpos; i<iter->length; i++)
	    {
		if (iter->buffer[i] == L'\n')
		{
		    bCrLfFound = true;
		    WriteToLog(L"Found CrLf");
		    break;
		}
	    }
	    if (bCrLfFound)
	    {
		WriteToLog(L"Breaking out of the iter loop");
		break;
	    }
	    iter = iter->next;
	}
	if (!bCrLfFound)
	{
	    if (!ReadNextBuffer())
		break;
	}
    } while (!bCrLfFound);

    if (!bCrLfFound)
    {
	if (output_list != NULL)
	{
	    // return the buffered data
	}
	WriteToLog(L"returning nullptr");
	return nullptr;
    }

    count = 0;
    iter = output_list;
    bCrLfFound = false;
    for (;;)
    {
	for (i=iter->curpos; i<iter->length; i++)
	{
	    if (iter->buffer[i] == L'\n')
	    {
		bCrLfFound = true;
		break;
	    }
	    count++;
	}
	if (bCrLfFound)
	    break;
	iter = iter->next;
    }

    wchar_t msg[100];
    swprintf_s(msg, 100, L"count = %d", count);
    WriteToLog(msg);

    wchar_t *line = new wchar_t[count+1];
    curpos = 0;
    iter = output_list;
    bCrLfFound = false;
    for (;;)
    {
	for (i=iter->curpos; i<iter->length; i++)
	{
	    if (iter->buffer[i] == L'\n')
	    {
		iter->curpos++;
		if (iter->curpos == iter->length)
		{
		    output_list = iter->next;
		    delete [] iter->buffer;
		    delete iter;
		}
		bCrLfFound = true;
		line[curpos] = L'\0';
		break;
	    }
	    line[curpos++] = iter->buffer[i];
	    iter->curpos++;
	}
	if (bCrLfFound)
	    break;
	dead = iter;
	iter = iter->next;
	delete [] dead->buffer;
	delete dead;
    }

    WriteToLog(L"Returning string:");
    WriteToLog(line);
    System::String^ str = gcnew System::String(line, 0, count);
    delete [] line;
    return str;
}

System::String^ ProcessCreation::DeinoLPReadLine(System::String^ read_handle)
{
    if (hRead == INVALID_HANDLE_VALUE)
    {
	wchar_t *str = DupString(read_handle);
	if (str != NULL)
	{
	    hRead = decode_handle(str);
	    delete [] str;
	}
    }

    return DeinoLPReadLine();
}

void ProcessCreation::DeinoLPWriteLine(System::String ^str)
{
    wchar_t err_msg[1024];
    DWORD num_written = 0;
    wchar_t *line;
    str = str + "\n";
    int length = str->Length;
    line = DupString(str);
    length = length * sizeof(wchar_t);
    /*
    if (WriteFile(hWrite, &length, sizeof(int), &num_written, NULL))
    {
	wchar_t msg[100];
	swprintf_s(msg, 100, L"Wrote length: %d", length);
	WriteToLog(msg);
	*/
	if (WriteFile(hWrite, line, length, &num_written, NULL))
	{
	    FlushFileBuffers(hWrite);
	    wchar_t msg[100];
	    swprintf_s(msg, 100, L"Wrote %d bytes", num_written);
	    WriteToLog(msg);
	    WriteToLog(line);
	}
	else
	{
	    swprintf_s(err_msg, 1024, L"Unable to write to the process input: Error %d\n", GetLastError());
	    throw gcnew System::Exception(gcnew System::String(err_msg));
	}
	/*
    }
    else
    {
	swprintf_s(err_msg, 1024, L"Unable to write the length integer: Error %d", GetLastError());
	throw gcnew System::Exception(gcnew System::String(err_msg));
    }
    */
}

void ProcessCreation::DeinoLPWriteLine(System::String ^write_handle, System::String ^str)
{
    if (hWrite == INVALID_HANDLE_VALUE)
    {
	wchar_t *wstr = DupString(write_handle);
	if (wstr != NULL)
	{
	    hWrite = decode_handle(wstr);
	    delete [] wstr;
	}
    }

    return DeinoLPWriteLine(str);
}

void ProcessCreation::DeinoWaitForExit()
{
    if (hProcess != INVALID_HANDLE_VALUE)
    {
	if (WaitForSingleObject(hProcess, INFINITE) == WAIT_OBJECT_0)
	{
	    hProcess = INVALID_HANDLE_VALUE;
	}
    }
}

void ProcessCreation::Kill()
{
    if (hProcess != INVALID_HANDLE_VALUE)
    {
	TerminateProcess(hProcess, (UINT)-1);
    }
}

int DriveMapping::MapDrive(System::Char drive_letter, System::String ^mapping, System::String ^username, System::String ^password)
{
    wchar_t pszDrive[3];
    wchar_t *pszMapping=NULL, *pszAccount=NULL, *pszPassword=NULL;
    NETRESOURCE net;

    pszDrive[0] = drive_letter;
    pszDrive[1] = L':';
    pszDrive[2] = L'\0';

    pszMapping = DupString(mapping);
    pszAccount = DupString(username);
    pszPassword = DupString(password);
    if (pszAccount != NULL && pszAccount[0] == L'\0')
    {
	delete [] pszAccount;
	pszAccount = NULL;
	if (pszPassword != NULL)
	{
	    delete [] pszPassword;
	    pszPassword = NULL;
	}
    }

    memset(&net, 0, sizeof(NETRESOURCE));
    net.lpLocalName = pszDrive;
    net.lpRemoteName = pszMapping;
    net.dwType = RESOURCETYPE_DISK;
    net.lpProvider = NULL;

    int result = WNetAddConnection2(&net, pszPassword, pszAccount, CONNECT_REDIRECT);
    return result;
}

int DriveMapping::UnMapDrive(System::Char drive_letter)
{
    wchar_t pszDrive[3];
    pszDrive[0] = drive_letter;
    pszDrive[1] = L':';
    pszDrive[2] = L'\0';
    
    int result = WNetCancelConnection2(pszDrive,
	CONNECT_UPDATE_PROFILE, /* This option makes sure that the connection is not re-established at the next user logon */
	TRUE);
    return result;
}

}
