using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Win32;
using System.Collections;
using System.Xml;
using System.IO;
using System.Windows.Forms;

namespace DeinoMPI
{
    public class CredentialStore
    {
        private SecurityKeys keys;
        private ArrayList credentials;
        private bool use_registry;
        private string filename;

        #region Getters and Setters
        /// <summary>
        /// Location of the file containing the Credential Store
        /// </summary>
        public string FileName
        {
            get
            {
                return filename;
            }
        }

        /// <summary>
        /// Boolean value specifying whether or not the Credential Store is saved in the Windows Registry
        /// </summary>
        public bool UseRegistry
        {
            get
            {
                return use_registry;
            }
        }

        /// <summary>
        /// SecurityKeys used to protect the SecureCredentials in the store
        /// </summary>
        public SecurityKeys Keys
        {
            get
            {
                return keys;
            }
        }
        /// <summary>
        /// Number of credentials in the store
        /// </summary>
        public int Count
        {
            get
            {
                return credentials.Count;
            }
        }
        /// <summary>
        /// The first credential in the store or null if the store is empty
        /// </summary>
        public SecureCredential DefaultCredential
        {
            get
            {
                if (credentials.Count > 0)
                    return (SecureCredential)credentials[0];
                return null;
            }
        }
        #endregion

        #region Constructors
        #region New Store creation
        /// <summary>
        /// Create a new Credential Store in the Windows Registry using the provided SecurityKeys
        /// </summary>
        /// <param name="keys"></param>
        public CredentialStore(SecurityKeys keys)
        {
            this.keys = keys;
            credentials = new ArrayList();
            use_registry = true;
            WriteToRegistry();
        }

        /// <summary>
        /// Create a new Credential Store in the specified file using the provided SecurityKeys
        /// </summary>
        /// <param name="keys"></param>
        /// <param name="filename"></param>
        public CredentialStore(SecurityKeys keys, string filename)
        {
            this.keys = keys;
            this.filename = filename;
            credentials = new ArrayList();
            Write(filename);
        }
        #endregion

        #region Open existing stores
        /// <summary>
        /// Location of the current user's Credential store
        /// </summary>
        /// <returns>"registry" or the file name</returns>
        public static string GetLocation()
        {
            RegistryKey key;
            key = Registry.CurrentUser.OpenSubKey(@"Software\Deino\CredentialStore", false);
            if (key != null)
            {
                string location;
                location = (string)key.GetValue("location", null);
                key.Close();
                return location;
            }
            return null;
        }

        /// <summary>
        /// Set the location of the current user's Credential store
        /// </summary>
        /// <param name="path">"registry" or the file name</param>
        /// <returns></returns>
        public static bool SetLocation(string path)
        {
            try
            {
                RegistryKey key;
                key = Registry.CurrentUser.CreateSubKey(@"Software\Deino\CredentialStore");
                if (key == null)
                {
                    return false;
                }
                if (path.ToLower() == "registry")
                {
                    key.SetValue("location", "registry");
                }
                else
                {
                    key.SetValue("location", path);
                }
            }
            catch
            {
                return false;
            }
            return true;
        }

        /// <summary>
        /// private constructor to enable the static Open function to return a new store
        /// </summary>
        private CredentialStore()
        {
            use_registry = false;
            credentials = new ArrayList();
        }

        /// <summary>
        /// Open an existing Credential Store protected by the provided password
        /// </summary>
        /// <param name="password"></param>
        public CredentialStore(string password)
        {
            use_registry = false;
            credentials = new ArrayList();
            /*
            try
            {
            */
                    string location;
                    location = GetLocation();
                    if (location == null || location == "registry")
                    {
                        use_registry = true;
                        ReadFromRegistry(password);
                    }
                    else
                    {
                        Read(location, password);
                    }
            /*
            }
            catch (Exception)
            {
            }
            */
        }

        /// <summary>
        /// Open an existing Credential Store
        /// </summary>
        /// <param name="password">password protecting the store</param>
        /// <param name="ignore_location">ignore the location and read from the registry</param>
        public CredentialStore(string password, bool ignore_location)
        {
            use_registry = false;
            credentials = new ArrayList();
            /*
            try
            {
            */
            string location;
            location = GetLocation();
            if (ignore_location || location == null || location == "registry")
            {
                use_registry = true;
                ReadFromRegistry(password);
                SaveLocation();
            }
            else
            {
                Read(location, password);
            }
            /*
            }
            catch (Exception)
            {
            }
            */
        }

        /// <summary>
        /// Open the current user's Credential store
        /// </summary>
        /// <param name="password">password protecting the store</param>
        /// <returns></returns>
        static public CredentialStore Open(string password)
        {
            CredentialStore store = new CredentialStore();
            try
            {
                string location;
                location = GetLocation();
                if (location == null || location == "registry")
                {
                    store.ReadFromRegistry(password);
                }
                else
                {
                    store.Read(location, password);
                }
                return store;
            }
            catch (Exception)
            {
            }
            return null;
        }

        /// <summary>
        /// Open an existing Credential Store saved in the specified file and proteced by the provided password
        /// </summary>
        /// <param name="filename"></param>
        /// <param name="password"></param>
        public CredentialStore(string filename, string password)
        {
            credentials = new ArrayList();
            use_registry = false;
            Read(filename, password);
            SaveLocation();
        }
        #endregion
        #endregion

        /// <summary>
        /// Check if the current user's Credential store requires a password
        /// </summary>
        /// <returns>true or false</returns>
        static public bool RequiresPassword()
        {
            string location = GetLocation();
            if (location == null || location == "registry")
            {
                return ReadNeedsPassword();
            }
            return ReadNeedsPassword(location);
        }

        /// <summary>
        /// Check if the current user has a Credential store
        /// </summary>
        /// <returns></returns>
        static public bool Exists()
        {
            string location = GetLocation();
            if (location == null)
                return false;
            if (location == "registry")
                return true;
            if (File.Exists(location))
                return true;
            return false;
        }

        /// <summary>
        /// Check if the Credential store is stored in a removable media device
        /// </summary>
        /// <returns></returns>
        static public bool IsRemovableMedia()
        {
            string location = GetLocation();
            if (location == null)
                return false;
            if (location == "registry")
                return false;
            string root_drive = Path.GetPathRoot(location);
            if (root_drive == null || root_drive.Length < 1)
                return false;
            foreach (DriveInfo drive in System.IO.DriveInfo.GetDrives())
            {
                if (drive.Name[0].ToString().ToLower() == root_drive[0].ToString().ToLower())
                {
                    if (drive.DriveType == DriveType.Removable || drive.DriveType == DriveType.CDRom)
                        return true;
                    return false;
                }
            }
            return false;
        }

        /// <summary>
        /// Check if the current user's Registry Credential store requires a password
        /// </summary>
        /// <returns>true or false</returns>
        static public bool RegistryStoreRequiresPassword()
        {
            return ReadNeedsPassword();
        }

        /// <summary>
        /// Check if the location of the current user's Credential store is ready to be read
        /// </summary>
        /// <returns>true or false</returns>
        static public bool LocationReady()
        {
            string location = GetLocation();
            if (location != null && location.Length > 2)
            {
                if (location[1] == ':')
                {
                    DriveInfo info = new DriveInfo(location[0].ToString());
                    return info.IsReady;
                }
            }
            return true;
        }

        /// <summary>
        /// Check if the Credential store requires a password
        /// </summary>
        /// <param name="filename">file name of the Credential store</param>
        /// <returns>true or false</returns>
        static public bool RequiresPassword(string filename)
        {
            return ReadNeedsPassword(filename);
        }

        /// <summary>
        /// Add a credential to the store
        /// </summary>
        /// <param name="credential">SecureCredential to be added to the store</param>
        public void AddCredential(SecureCredential credential)
        {
            if (credentials == null)
                credentials = new ArrayList();

            // Search for and remove a credential with the same name
            SecureCredential found = null;
            foreach (SecureCredential cred in credentials)
            {
                if (cred.Account == credential.Account)
                    found = cred;
            }
            if (found != null)
                credentials.Remove(found);

            // Then add the new credential to the list
            credentials.Add(credential);

            // Save the store immediately.
            // Another approach could be to modify the credentials in memory only and then
            // require the user to call another function like SaveCredentialStore() to save any changes.
            if (UseRegistry)
            {
                WriteToRegistry();
            }
            else
            {
                Write(filename);
            }
        }

        /// <summary>
        /// Remove the account from the Credential store
        /// </summary>
        /// <param name="account">account to remove</param>
        public void RemoveCredential(string account)
        {
            SecureCredential found = null;
            foreach (SecureCredential cred in credentials)
            {
                if (cred.Account == account)
                {
                    found = cred;
                    break;
                }
            }
            if (found != null)
                credentials.Remove(found);

            // Save the store immediately.
            // Another approach could be to modify the credentials in memory only and then
            // require the user to call another function like SaveCredentialStore() to save any changes.
            if (UseRegistry)
            {
                WriteToRegistry();
            }
            else
            {
                Write(filename);
            }
        }

        /// <summary>
        /// Retreive the secure credential with the provided account name
        /// </summary>
        /// <param name="account">Account name</param>
        /// <returns>SecureCredential or null if not found</returns>
        public SecureCredential GetCredential(string account)
        {
            if (credentials == null)
                return null;
            foreach (SecureCredential cred in credentials)
            {
                if (String.Compare(cred.Account, account, true) == 0)
                {
                    return cred;
                }
            }
            return null;
        }

        /// <summary>
        /// Save the location of the Credential store
        /// </summary>
        private void SaveLocation()
        {
            try
            {
                RegistryKey key;
                key = Registry.CurrentUser.CreateSubKey(@"Software\Deino\CredentialStore");
                if (key != null)
                {
                    if (UseRegistry)
                    {
                        key.SetValue("location", "registry");
                    }
                    else
                    {
                        key.SetValue("location", FileName);
                    }
                    key.Close();
                }
            }
            catch
            {
            }
        }

        /// <summary>
        /// Read the Credential store from the specified file
        /// </summary>
        /// <param name="filename">file name of the store</param>
        /// <param name="password">password protecting the store</param>
        private void Read(string filename, string password)
        {
            XmlReaderSettings settings = new XmlReaderSettings();
            settings.IgnoreComments = true;
            settings.IgnoreWhitespace = true;
            using (StreamReader sr = new StreamReader(filename))
            {
                using (XmlReader rd = XmlReader.Create(sr, settings))
                {
                    rd.ReadStartElement("CredentialStore");
                    string private_key = rd.ReadElementString("SecurityKeys");
                    keys = SecurityKeys.Create(private_key, password);
                    if (keys == null)
                    {
                        throw new Exception("Unable to create the security keys with the given private key and password.");
                    }
                    while (rd.IsStartElement())
                    {
                        if (rd.Name == "SecureCredential")
                        {
                            string cred_str = rd.ReadElementString("SecureCredential");
                            SecureCredential cred = new SecureCredential(keys, cred_str);
                            credentials.Add(cred);
                        }
                        else
                        {
                            throw new Exception(String.Format("Unexpected element: {0}", rd.Name));
                        }
                    }
                    rd.ReadEndElement(); // CredentialStore
                    rd.Close();
                    this.filename = filename;
                }
                sr.Close();
            }
        }

        /// <summary>
        /// Determine if the Credential store is password protected
        /// </summary>
        /// <param name="filename">file name of the store</param>
        /// <returns></returns>
        static private bool ReadNeedsPassword(string filename)
        {
            bool result = true;
            try
            {
                XmlReaderSettings settings = new XmlReaderSettings();
                settings.IgnoreComments = true;
                settings.IgnoreWhitespace = true;
                using (StreamReader sr = new StreamReader(filename))
                {
                    using (XmlReader rd = XmlReader.Create(sr, settings))
                    {
                        rd.ReadStartElement("CredentialStore");
                        string private_key = rd.ReadElementString("SecurityKeys");
                        result = SecurityKeys.RequiresPassword(private_key);
                        rd.Close();
                    }
                    sr.Close();
                }
            }
            catch
            {
            }
            return result;
        }

        /// <summary>
        /// Determine if the Credential store in the Registry is password protected
        /// </summary>
        /// <returns></returns>
        static private bool ReadNeedsPassword()
        {
            bool result = true;
            try
            {
                RegistryKey key;
                key = Registry.CurrentUser.OpenSubKey(@"Software\Deino\CredentialStore");
                if (key != null)
                {
                    string private_key = (string)key.GetValue("SecurityKeys", "");
                    result = SecurityKeys.RequiresPassword(private_key);
                    key.Close();
                }
            }
            catch
            {
            }
            return result;
        }

        /// <summary>
        /// Read the Credential store from the Registry
        /// </summary>
        /// <param name="password">password protecting the store</param>
        private void ReadFromRegistry(string password)
        {
            use_registry = true;
            try
            {
                RegistryKey key;
                credentials = new ArrayList();
                key = Registry.CurrentUser.OpenSubKey(@"Software\Deino\CredentialStore");
                if (key != null)
                {
                    string private_key = (string)key.GetValue("SecurityKeys", "");
                    try
                    {
                        keys = new SecurityKeys(private_key, password);
                    }
                    catch
                    {
                        key.Close();
                        return;
                    }
                    string[] names = key.GetValueNames();
                    if (names != null)
                    {
                        foreach (string name in names)
                        {
                            if (name != "location" && name != "SecurityKeys")
                            {
                                SecureCredential cred = SecureCredential.FromSecureXml(keys, (string)key.GetValue(name));
                                if (cred != null)
                                    credentials.Add(cred);
                            }
                        }
                    }
                    key.Close();
                }
            }
            catch
            {
            }
        }

        /// <summary>
        /// Write the store to a file
        /// </summary>
        /// <param name="filename">file name to write the store to</param>
        private void Write(string filename)
        {
            this.filename = filename;
            using (XmlWriter wr = XmlWriter.Create(filename))
            {
                wr.WriteStartDocument();
                wr.WriteStartElement("CredentialStore");
                wr.WriteElementString("SecurityKeys", keys.PrivateKey);
                if (credentials != null)
                {
                    foreach (SecureCredential cred in credentials)
                    {
                        string secure_xml = cred.SecureXml;
                        if (secure_xml != null)
                            wr.WriteElementString("SecureCredential", secure_xml);
                    }
                }
                wr.WriteEndElement(); // CredentialStore
                wr.WriteEndDocument();
                wr.Close();
                SaveLocation();
            }
        }

        /// <summary>
        /// Write the store to the current user's Registry hive
        /// </summary>
        private void WriteToRegistry()
        {
            try
            {
                RegistryKey key;
                key = Registry.CurrentUser.CreateSubKey(@"Software\Deino\CredentialStore");
                if (key != null)
                {
                    // erase all old entries
                    string[] names = key.GetValueNames();
                    foreach (string name in names)
                        key.DeleteValue(name);

                    // save the new entries
                    key.SetValue("location", "registry");
                    key.SetValue("SecurityKeys", keys.PrivateKey);
                    if (credentials != null)
                    {
                        foreach (SecureCredential cred in credentials)
                        {
                            if (cred.Account != null)
                            {
                                string val_name = cred.Account.Replace('\\', '*'); // If you have \ characters in the name then it creates subkeys
                                string val_value = cred.SecureXml;
                                if (val_name != null && val_value != null)
                                    key.SetValue(val_name, val_value);
                            }
                        }
                    }
                    key.Close();
                }
            }
            catch //(Exception e)
            {
                //MessageBox.Show("WriteToRegistry failed: " + e.Message);
            }
        }

        public IEnumerator GetEnumerator()
        {
            return credentials.GetEnumerator();
        }
    }
}
