using System;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.IO;
using System.Text;
using Microsoft.Win32;
using System.Xml;
using System.Security.Cryptography;
using System.Net.Security;
using System.Diagnostics;
using System.ComponentModel;
using DeinoLP;

namespace DeinoMPI
{
    #region PMI classes

    public class PMI_Keyval
    {
        private string _key;
        private string _val;
        public string key
        {
            get { return _key; }
            set { _key = value; }
        }
        public string val
        {
            get { return _val; }
            set { _val = value; }
        }
    }

    public class KVS
    {
        //private Mpiexec _mpiexec;
        private string name;
        private Hashtable data;
        private ManualResetEvent barrier_in_event, barrier_out_event;
        private volatile int barrier_count;
        private volatile int barrier_incount;
        private Queue iter;

        public KVS()
        {
            name = System.Guid.NewGuid().ToString();
            data = new Hashtable();
            barrier_in_event = new ManualResetEvent(false);
            barrier_out_event = new ManualResetEvent(false);
            barrier_count = -1;
            barrier_incount = 0;
        }

        public void Barrier(int count)
        {
            lock (this)
            {
                //_mpiexec.WriteLine("barrier in: {0} -> {1}", barrier_incount, barrier_incount+1);
                if (barrier_count == -1)
                    barrier_count = count;
                barrier_incount++;
                if (barrier_incount >= barrier_count)
                {
                    //_mpiexec.WriteLine("barrier in Set");
                    barrier_out_event.Reset();
                    barrier_in_event.Set();
                }
            }
            barrier_in_event.WaitOne();
            lock (this)
            {
                //_mpiexec.WriteLine("barrier out");
                barrier_incount--;
                if (barrier_incount == 0)
                {
                    //_mpiexec.WriteLine("barrier out Reset");
                    barrier_count = -1;
                    barrier_in_event.Reset();
                    barrier_out_event.Set();
                }
            }
            barrier_out_event.WaitOne();
            //_mpiexec.WriteLine("leaving barrier");
        }

        public void Put(string key, string val)
        {
            //_mpiexec.WriteLine("put: {0}:{1}", key, val);
            data[key] = val;
        }

        public string Get(string key)
        {
            //_mpiexec.WriteLine("get: {0}:{1}", key, data[key]);
            return (string)data[key];
        }

        public void Remove(string key)
        {
            //_mpiexec.WriteLine("removing: {0}", key);
            string val = (key == null ) ? null : (string)data[key];
            if (val == null)
                throw new Exception("Invalid key name");
            data.Remove(key);
        }

        public PMI_Keyval Iter_first()
        {
            iter = new Queue();
            foreach (string key in data.Keys)
            {
                PMI_Keyval kv = new PMI_Keyval();
                kv.key = key;
                kv.val = (string)data[key];
                iter.Enqueue(kv);
            }
            if (iter.Count == 0)
                return null;
            return (PMI_Keyval)iter.Dequeue();
        }

        public PMI_Keyval Iter_next()
        {
            if (iter == null)
                return null;
            if (iter.Count == 0)
                return null;
            return (PMI_Keyval)iter.Dequeue();
        }

        #region getters and setters
        public string Name
        {
            get
            {
                return name;
            }
        }
        #endregion
    }

    public class PMI
    {
        public const int ID_MAXLEN = 100;
        public const int NAME_MAXLEN = 100;
        public const int KEY_MAXLEN = 256;
        public const int VALUE_MAXLEN = 32768;
        private static KVS _domain = new KVS();
        private KVS _id;
        private Hashtable kvs;
        private int _nproc;
        private ManualResetEvent barrier_in_event, barrier_out_event;
        private volatile int barrier_incount;
        public Mpiexec mpiexec;

        public PMI()
        {
            _id = new KVS();
            kvs = new Hashtable();
            kvs[_domain.Name] = _domain;
            kvs[_id.Name] = _id;
            barrier_in_event = new ManualResetEvent(false);
            barrier_out_event = new ManualResetEvent(false);
            barrier_incount = 0;
        }

        #region getters and setters
        public string domain
        {
            get { return _domain.Name; }
        }
        public string id
        {
            get { return _id.Name; }
        }
        public int nproc
        {
            get { return _nproc; }
            set { _nproc = value; }
        }
        #endregion

        public void Barrier()
        {
            lock (typeof(PMI))
            {
                //mpiexec.WriteLine("barrier in: {0} -> {1}", barrier_incount, barrier_incount+1);
                barrier_incount++;
                if (barrier_incount >= nproc)
                {
                    //mpiexec.WriteLine("barrier in Set");
                    barrier_out_event.Reset();
                    barrier_in_event.Set();
                }
            }
            barrier_in_event.WaitOne();
            lock (typeof(PMI))
            {
                //mpiexec.WriteLine("barrier out");
                barrier_incount--;
                if (barrier_incount == 0)
                {
                    //mpiexec.WriteLine("barrier out Reset");
                    barrier_in_event.Reset();
                    barrier_out_event.Set();
                }
            }
            barrier_out_event.WaitOne();
            //mpiexec.WriteLine("leaving barrier");
        }

        public void KVS_Barrier(string kvsname, int count)
        {
            KVS k;
            lock (kvs)
            {
                k = (KVS)kvs[kvsname];
            }
            k.Barrier(count);
        }

        public string KVS_Create()
        {
            KVS k = new KVS();
            lock (kvs)
            {
                kvs[k.Name] = k;
            }
            return k.Name;
        }

        public void KVS_Destroy(string kvsname)
        {
            lock (kvs)
            {
                KVS k = (KVS)kvs[kvsname];
                if (k == null)
                    throw new Exception("Invalid KVS name");
                kvs.Remove(k);
            }
        }

        public void KVS_Put(string kvs_name, string key, string val)
        {
            //mpiexec.WriteLine("Put: {0}:{1}:{2}", kvs_name, key, val);
            lock (kvs)
            {
                KVS k = (KVS)kvs[kvs_name];
                if (k == null)
                    throw new Exception("Invalid KVS name");
                k.Put(key, val);
            }
        }

        public string KVS_Get(string kvs_name, string key)
        {
            KVS k;
            //mpiexec.WriteLine("Get: {0}:{1}", kvs_name, key);
            lock (kvs)
            {
                k = (KVS)kvs[kvs_name];
                if (k == null)
                    throw new Exception("Invalid KVS name");
            }
            //mpiexec.WriteLine("Got: {0}", k.Get(key));
            return k.Get(key);
        }

        private class SpawnCommandThreadArg
        {
            public ArrayList nodes;
            public PMI_Keyval[] preput;
            public string[] errors;
            public bool quiet;
            public ManualResetEvent ready_event;
        }

        public void SpawnCommandThread(object param)
        {
            SpawnCommandThreadArg sparg = (SpawnCommandThreadArg)param;
            ArrayList nodes = sparg.nodes;
            PMI_Keyval[] preput = sparg.preput;
            string[] errors = sparg.errors;
            bool quiet = sparg.quiet;
            ManualResetEvent ready_event = sparg.ready_event;
            sparg = null;
            ProcessGroup pg = new ProcessGroup(mpiexec);
            pg.QuietAbort = quiet;
            if (preput != null && preput.Length > 0)
            {
                for (int i = 0; i < preput.Length; i++)
                {
                    pg.pmi.KVS_Put(pg.pmi.id, preput[i].key, preput[i].val);
                    //mpiexec.WriteLine("pre-put: kvs = {0}, key = {1}, val = {2}", pg.pmi.id, preput[i].key, preput[i].val);
                }
            }
            preput = null;
            if (pg.StartGroup(nodes))
            {
                for (int i = 0; i < errors.Length; i++)
                    errors[i] = null;
                pg.QuietAbort = false; // No errors occurred starting the processes.  Turn off the quiet flag so run-time abort messages will be printed.
                ready_event.Set();
            }
            else
            {
                pg.GetStartErrors(errors);
                ready_event.Set();
                pg.Kill();
            }
            pg.WaitForGroupExit();
            if (mpiexec.global_options.exitcodes)
            {
                pg.PrintExitCodes();
            }
            mpiexec.process_groups.Remove(pg);
        }

        public void Spawn_multiple(string[] commands, string[][] argvs, int[] maxprocs, PMI_Keyval[][] keyvals, PMI_Keyval[] preput, string[] errors, bool quiet)
        {
            if (commands != null)
            {
                ArrayList nodes = null;
                try
                {
                    nodes = mpiexec.CreateLaunchNodes(commands, argvs, maxprocs, keyvals);
                }
                catch (Exception e)
                {
                    //mpiexec.WriteLine("CreateLaunchNodes failed.");
                    //mpiexec.WriteLine(e.Message);
                    throw new Exception("CreateLaunchNodes failed: " + e.Message);
                }
                if (nodes != null)
                {
                    if (nodes.Count > 0)
                    {
                        int result = mpiexec.GetStoreAndCredential(nodes);
                        //mpiexec.WriteLine("past GetStoreAndCredentail");
                        if (result != 0)
                        {
                            throw new Exception("Unable to get the user credentials to spawn new processes.");
                        }

                        SpawnCommandThreadArg sparg = new SpawnCommandThreadArg();
                        sparg.nodes = nodes;
                        sparg.preput = preput;
                        sparg.errors = errors;
                        sparg.quiet = quiet;
                        sparg.ready_event = new ManualResetEvent(false);
                        ParameterizedThreadStart ts = new ParameterizedThreadStart(this.SpawnCommandThread);
                        Thread spawn_thread = new Thread(ts);
                        spawn_thread.Start(sparg);
                        sparg.ready_event.WaitOne();
                        //mpiexec.WriteLine("past SpawnCommandThread start");
                    }
                }
            }
            else
            {
                //mpiexec.WriteLine("No commands");
            }
        }

        public void Publish_name(string service_name, string port)
        {
            lock (_domain)
            {
                //mpiexec.WriteLine("{0} publish {1}={2}", _domain.Name, service_name, port);
                _domain.Put(service_name, port);
            }
        }

        public void Unpublish_name(string service_name)
        {
            lock (_domain)
            {
                //mpiexec.WriteLine("{0} unpublish {1}", _domain.Name, service_name);
                _domain.Remove(service_name);
                //mpiexec.WriteLine("{0} after unpublish {1}={2}", _domain.Name, service_name, _domain.Get(service_name));
            }
        }

        public string Lookup_name(string service_name)
        {
            string port;
            lock (_domain)
            {
                port = _domain.Get(service_name);
                //mpiexec.WriteLine("{0} lookup {1}={2}", _domain.Name, service_name, port);
            }
            return port;
        }

        public void Init(string kvsname, int rank, int size)
        {
        }

        public void Finalize(string kvsname, int rank)
        {
        }

        public PMI_Keyval Iter_first()
        {
            return _id.Iter_first();
        }

        public PMI_Keyval Iter_next()
        {
            return _id.Iter_next();
        }
    }

    public class ActivePMI_Client
    {
        string _host;
        int _port;
        TcpClient _client;
        TcpListener _listener;
        ServerSession _session;
        //Mpiexec _mpiexec;

        public ActivePMI_Client()
        {
            _host = null;
            _port = 0;
            _client = null;
            _listener = null;
            _session = null;
            //_mpiexec = new Mpiexec();
            //_mpiexec.WriteLine("ActivePMI_Client constructor");
        }

        public string host
        {
            get { return _host; }
        }
        public int port
        {
            get { return _port; }
        }

        public ServerSession session
        {
            get { return _session; }
        }

        public void Listen()
        {
            //_mpiexec.WriteLine("ActivePMI_Client Listen");
            _listener = new TcpListener(IPAddress.Any/*IPAddress.IPv6Any*/, 0);
            _listener.Start();
            _port = ((IPEndPoint)_listener.LocalEndpoint).Port;
            _host = System.Net.Dns.GetHostName();
        }

        public void Accept()
        {
            //_mpiexec.WriteLine("ActivePMI_Client Accept");
            _client = _listener.AcceptTcpClient();
            _client.NoDelay = true;
            //_mpiexec.WriteLine("ActivePMI_Client new ServerSession");
            _session = new ServerSession(_client, null);
            _listener.Stop();
        }
    }

    public class PMI_Client
    {
        private int _size;
        private int _rank;
        private int _universe_size;
        private int _appnum;
        private string _id;
        private string _domain_id;
        private int _id_length_max;
        private int _clique_size;
        private int[] _clique_ranks;
        private string _my_name;
        private int _name_length_max;
        private int _key_length_max;
        private int _value_length_max;
        private bool _initialized;
        private bool _spawned;
        private Mpiexec _mpiexec;
        private ClientSession client;
        private bool singleton;
        private PMI singleton_pmi;

        public PMI_Client()
        {
            _size = -1;
            _rank = -1;
            _universe_size = -1;
            _appnum = -1;
            _id = null;
            _domain_id = null;
            _id_length_max = -1;
            _clique_size = -1;
            _clique_ranks = null;
            _my_name = null;
            _name_length_max = -1;
            _key_length_max = -1;
            _value_length_max = -1;
            _initialized = false;
            _spawned = false;
            _mpiexec = new Mpiexec();
        }

        public void SendMessage(ActiveMessage message)
        {
            client.SendMessage(message);
        }

        public bool Init(ActivePMI_Client active)
        {
            string env;
            string host;
            int port;
            string key;

            if (_initialized)
                return _spawned;

            env = Environment.GetEnvironmentVariable("PMI_SPAWN");
            if (env != null)
            {
                if (Convert.ToInt32(env) != 0)
                {
                    _spawned = true;
                }
            }
            env = Environment.GetEnvironmentVariable("PMI_APPNUM");
            if (env != null)
            {
                _appnum = Convert.ToInt32(env);
                if (_appnum < 0)
                    _appnum = 0;
            }
            _id = Environment.GetEnvironmentVariable("PMI_KVS");
            _domain_id = Environment.GetEnvironmentVariable("PMI_DOMAIN");
            env = Environment.GetEnvironmentVariable("PMI_RANK");
            if (env != null)
            {
                _rank = Convert.ToInt32(env);
            }
            env = Environment.GetEnvironmentVariable("PMI_SIZE");
            if (env != null)
            {
                _size = Convert.ToInt32(env);
            }
            env = Environment.GetEnvironmentVariable("PMI_CLIQUE");
            if (env != null)
            {
                //Mpiexec.WriteLine("Parsing PMI_CLIQUE: <{0}>", env);
                _clique_ranks = ProcessGroup.ParseCliqueInts(env);
                _clique_size = _clique_ranks.Length;
                /*
                string[] ranks = ProcessGroup.ParseClique(env);
                _clique_size = ranks.Length;
                _clique_ranks = new int[_clique_size];
                for (int i = 0; i < _clique_size; i++)
                {
                    _clique_ranks[i] = Convert.ToInt32(ranks[i]);
                }
                */
                //Mpiexec.WriteLine("PMI_CLIQUE length = {0}", _clique_size);
            }
            key = Environment.GetEnvironmentVariable("PMI_KEY");
            if (key == null)
            {
                key = "anyprocess";
            }
            host = Environment.GetEnvironmentVariable("PMI_WHOST");
            if (host != null)
            {
                env = Environment.GetEnvironmentVariable("PMI_WPORT");
                if (env != null)
                {
                    port = Convert.ToInt32(env);
                    //Mpiexec.WriteLine("Establishing client session with {0}:{1}", host, port);
                    client = new ClientSession(host, port);
                    PMIMessage message = new PMIMessage();
                    message.pmi_type = "client";
                    message.name = "init";
                    message.data["key"] = key;
                    message.data["active_host"] = active.host;
                    message.data["active_port"] = active.port.ToString();
                    //Mpiexec.WriteLine("Sending init message.");
                    client.SendMessage(message);
                    //Mpiexec.WriteLine("init message sent.");
                    message = (PMIMessage)client.NextMessage();
                    if ((string)message.data["success"] == "true")
                    {
                        //Mpiexec.WriteLine("init pmi command returned true");
                    }
                    else
                    {
                        /*
                        string execption_message = (string)message.data["message"];
                        Mpiexec.WriteLine("init pmi command failed: {0}", execption_message);
                        Exception e2 = new Exception(execption_message);
                        Mpiexec.WriteLine(String.Format("e2.message = {0}", e2.Message));
                        throw e2;
                        */
                        throw new Exception((string)message.data["message"]);
                    }
                    singleton = false;
                }
                else
                {
                    throw new Exception("No PMI_WPORT environment variable found.");
                }
            }
            else
            {
                singleton = true;
                singleton_pmi = new PMI();
                singleton_pmi.mpiexec = new Mpiexec();
                _mpiexec.global_options.dir = Environment.CurrentDirectory;
                singleton_pmi.nproc = 1;
                _rank = 0;
                _size = 1;
            }
            _initialized = true;
            return _spawned;
        }
        public bool Initialized
        {
            get { return _initialized; }
            set { _initialized = value; }
        }
        public void pmiFinalize() // Finalize is an invalid function name
        {
            if (singleton)
            {
                if (singleton_pmi != null)
                    singleton_pmi.mpiexec.WaitForAllExit();
                else
                    _mpiexec.WaitForAllExit();
                return;
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "finalize";
            message.data["id"] = _id;
            message.data["rank"] = _rank.ToString();
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("finalize pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("finalize pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("finalize pmi command failed: {0}", (string)message.data["message"]);
            }
        }
        public int size
        {
            get
            {
                if (_size == -1)
                {
                    if (singleton)
                    {
                        _size = 1;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "size";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _size = Convert.ToInt32((string)message.data["size"]);
                            //_mpiexec.WriteLine("size pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("size pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("size pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _size;
            }
            set { _size = value; }
        }
        public int rank
        {
            get
            {
                if (_rank == -1)
                {
                    if (singleton)
                    {
                        _rank = 0;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "rank";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _rank = Convert.ToInt32((string)message.data["rank"]);
                            //_mpiexec.WriteLine("rank pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("rank pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("rank pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _rank;
            }
            set { _rank = value; }
        }
        public int universe_size
        {
            get
            {
                if (_universe_size == -1)
                {
                    if (singleton)
                    {
                        _universe_size = -1;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "universe";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _universe_size = Convert.ToInt32((string)message.data["universe"]);
                            //_mpiexec.WriteLine("universe pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("universe pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("universe pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _universe_size;
            }
            set { _universe_size = value; }
        }
        public int appnum
        {
            get
            {
                if (_appnum == -1)
                {
                    if (singleton)
                    {
                        _appnum = 0;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "appnum";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _appnum = Convert.ToInt32((string)message.data["appnum"]);
                            //_mpiexec.WriteLine("appnum pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("appnum pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("appnum pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _appnum;
            }
            set { _appnum = value; }
        }
        public void Publish_name(string service_name, string port)
        {
            if (singleton)
            {
                singleton_pmi.Publish_name(service_name, port);
                return;
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "publish";
            message.data["name"] = service_name;
            message.data["port"] = port;
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("publish pmi command returned true");
            }
            else
            {
                _mpiexec.WriteLine("publish pmi command failed: {0}", (string)message.data["message"]);
                throw new Exception(String.Format("publish pmi command failed: {0}", (string)message.data["message"]));
            }
        }
        public void Unpublish_name(string service_name)
        {
            if (singleton)
            {
                singleton_pmi.Unpublish_name(service_name);
                return;
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "unpublish";
            message.data["name"] = service_name;
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("unpublish pmi command returned true");
            }
            else
            {
                _mpiexec.WriteLine("unpublish pmi command failed: {0}", (string)message.data["message"]);
                throw new Exception(String.Format("unpublish pmi command failed: {0}", (string)message.data["message"]));
            }
        }
        public string Lookup_name(string service_name)
        {
            if (singleton)
            {
                return singleton_pmi.Lookup_name(service_name);
            }
            String port = null;
            PMIMessage message = new PMIMessage();
            //_mpiexec.WriteLine("allocated PMIMessage");
            message.pmi_type = "client";
            message.name = "lookup";
            message.data["name"] = service_name;
            //_mpiexec.WriteLine("about to send pmi message");
            client.SendMessage(message);
            //_mpiexec.WriteLine("pmi message sent, waiting for result");
            message = (PMIMessage)client.NextMessage();
            //_mpiexec.WriteLine("pmi result message received");
            if ((string)message.data["success"] == "true")
            {
                port = (string)message.data["port"];
                //_mpiexec.WriteLine("lookup pmi command returned true: port={0}", port);
            }
            else
            {
                _mpiexec.WriteLine("lookup pmi command failed: {0}", (string)message.data["message"]);
                throw new Exception(String.Format("lookup pmi command failed: {0}", (string)message.data["message"]));
            }
            return port;
        }
        public string id
        {
            get
            {
                if (_id == null)
                {
                    if (singleton)
                    {
                        _id = singleton_pmi.id;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "id";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _id = (string)message.data["id"];
                            //_mpiexec.WriteLine("id pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("id pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("id pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _id;
            }
            set { _id = value; }
        }
        public string domain_id
        {
            get
            {
                if (_domain_id == null)
                {
                    if (singleton)
                    {
                        _domain_id = singleton_pmi.domain;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "domain";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _domain_id = (string)message.data["id"];
                            //_mpiexec.WriteLine("domain pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("domain pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("domain pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _domain_id;
            }
            set { _domain_id = value; }
        }
        public int id_length_max
        {
            get
            {
                if (_id_length_max == -1)
                {
                    if (singleton)
                    {
                        _id_length_max = PMI.ID_MAXLEN;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "id_maxlen";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _id_length_max = Convert.ToInt32((string)message.data["id_maxlen"]);
                            //_mpiexec.WriteLine("id_maxlen pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("id_maxlen pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("id_maxlen pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _id_length_max;
            }
            set { _id_length_max = value; }
        }
        public void Barrier()
        {
            if (singleton)
                return;
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "barrier";
            message.data["id"] = id;
            message.data["rank"] = rank.ToString();
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("barrier pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("barrier pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("barrier pmi command failed: {0}", (string)message.data["message"]);
            }
        }
        public int clique_size
        {
            get
            {
                if (_clique_size == -1)
                {
                    if (singleton)
                    {
                        _clique_size = 1;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "clique_size";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _clique_size = Convert.ToInt32((string)message.data["clique_size"]);
                            //_mpiexec.WriteLine("clique_size pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("clique_size pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("clique_size pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _clique_size;
            }
            set { _clique_size = value; }
        }
        public int[] clique_ranks
        {
            get
            {
                if (_clique_ranks == null)
                {
                    if (singleton)
                    {
                        _clique_ranks = new int[1];
                        _clique_ranks[0] = 0;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "clique";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _clique_size = (int)message.data["num_ranks"];
                            _clique_ranks = new int[_clique_size];
                            for (int i = 0; i < _clique_size; i++)
                            {
                                _clique_ranks[i] = Convert.ToInt32((string)message.data[String.Format("rank{0}", i)]);
                            }
                            //_mpiexec.WriteLine("clique pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("clique pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("clique pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _clique_ranks;
            }
            set { _clique_ranks = value; }
        }
        public void Abort(int exit_code, string error_message)
        {
            if (singleton)
            {
                // Kill any spawned groups
                foreach (ProcessGroup pg in _mpiexec.process_groups)
                {
                    pg.Kill();
                    pg.WaitForGroupExit();
                }
                // Then exit yourself
                _mpiexec.WriteLine();
                _mpiexec.WriteLine("job aborted:");
                _mpiexec.WriteLine("process: node: exit code: error message:");
                Console.Write("0: localhost: {0}", exit_code);
                if (error_message != null)
                {
                    Console.Write(": {0}", error_message);
                }
                _mpiexec.WriteLine();
                System.Environment.Exit(exit_code);
                return;
            }
            //_mpiexec.WriteLine("PMI_Client::Abort({0}, {1}) called", exit_code, error_message);
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "abort";
            message.data["id"] = id;
            message.data["rank"] = rank.ToString();
            message.data["exit code"] = exit_code.ToString();
            message.data["message"] = error_message != null ? error_message : "";
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("abort pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("abort pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("abort pmi command failed: {0}", (string)message.data["message"]);
            }
        }
        public string my_name
        {
            get
            {
                if (_my_name == null)
                {
                    _my_name = id;
                }
                return _my_name;
            }
            set { _my_name = value; }
        }
        public int name_length_max
        {
            get
            {
                if (_name_length_max == -1)
                {
                    if (singleton)
                    {
                        _name_length_max = PMI.NAME_MAXLEN;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "name_maxlen";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _name_length_max = Convert.ToInt32((string)message.data["name_maxlen"]);
                            //_mpiexec.WriteLine("name_maxlen pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("name_maxlen pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("name_maxlen pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _name_length_max;
            }
            set { _name_length_max = value; }
        }
        public int key_length_max
        {
            get
            {
                if (_key_length_max == -1)
                {
                    if (singleton)
                    {
                        _key_length_max = PMI.KEY_MAXLEN;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "key_maxlen";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _key_length_max = Convert.ToInt32((string)message.data["key_maxlen"]);
                            //_mpiexec.WriteLine("key_maxlen pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("key_maxlen pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("key_maxlen pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _key_length_max;
            }
            set { _key_length_max = value; }
        }
        public int value_length_max
        {
            get
            {
                if (_value_length_max == -1)
                {
                    if (singleton)
                    {
                        _value_length_max = PMI.VALUE_MAXLEN;
                    }
                    else
                    {
                        PMIMessage message = new PMIMessage();
                        message.pmi_type = "client";
                        message.name = "value_maxlen";
                        client.SendMessage(message);
                        message = (PMIMessage)client.NextMessage();
                        if ((string)message.data["success"] == "true")
                        {
                            _value_length_max = Convert.ToInt32((string)message.data["value_maxlen"]);
                            //_mpiexec.WriteLine("value_maxlen pmi command returned true");
                        }
                        else
                        {
                            throw new Exception(String.Format("value_maxlen pmi command failed: {0}", (string)message.data["message"]));
                            //_mpiexec.WriteLine("value_maxlen pmi command failed: {0}", (string)message.data["message"]);
                        }
                    }
                }
                return _value_length_max;
            }
            set { _value_length_max = value; }
        }
        public string KVS_Create()
        {
            string kvs = null;
            if (singleton)
            {
                return singleton_pmi.KVS_Create();
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "create";
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                kvs = (string)message.data["name"];
                //_mpiexec.WriteLine("create pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("create pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("create pmi command failed: {0}", (string)message.data["message"]);
            }
            return kvs;
        }
        public void KVS_Destroy(string kvsname)
        {
            if (singleton)
            {
                singleton_pmi.KVS_Destroy(kvsname);
                return;
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "destroy";
            message.data["name"] = kvsname;
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("destroy pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("destroy pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("destroy pmi command failed: {0}", (string)message.data["message"]);
            }
        }
        public void KVS_Put(string kvsname, string key, string val)
        {
            if (singleton)
            {
                singleton_pmi.KVS_Put(kvsname, key, val);
                return;
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "put";
            message.data["kvs"] = kvsname;
            message.data["key"] = key;
            message.data["value"] = val;
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("put pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("put pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("put pmi command failed: {0}", (string)message.data["message"]);
            }
        }
        public void KVS_Commit(string kvsname)
        {
            if (singleton)
                return;
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "commit";
            message.data["kvs"] = kvsname;
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("commit pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("commit pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("commit pmi command failed: {0}", (string)message.data["message"]);
            }
        }
        public string KVS_Get(string kvsname, string key)
        {
            if (singleton)
            {
                return singleton_pmi.KVS_Get(kvsname, key);
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "get";
            message.data["kvs"] = kvsname;
            message.data["key"] = key;
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("get pmi command returned true");
                return (string)message.data["value"];
            }
            throw new Exception(String.Format("get pmi command failed: {0}", (string)message.data["message"]));
            //_mpiexec.WriteLine("get pmi command failed: {0}", (string)message.data["message"]);
            //return null;
        }
        public void KVS_Iter_first(string kvsname, ref string key, ref string val)
        {
            if (singleton)
            {
                PMI_Keyval kv;
                kv = singleton_pmi.Iter_first();
                if (kv != null)
                {
                    key = kv.key;
                    val = kv.val;
                }
                else
                {
                    key = null;
                    val = null;
                }
                return;
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "iter_first";
            message.data["kvs"] = kvsname;
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                key = (string)message.data["key"];
                val = (string)message.data["value"];
                //_mpiexec.WriteLine("iter_first pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("iter_first pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("iter_first pmi command failed: {0}", (string)message.data["message"]);
            }
        }
        public void KVS_Iter_next(string kvsname, ref string key, ref string val)
        {
            if (singleton)
            {
                PMI_Keyval kv;
                kv = singleton_pmi.Iter_next();
                if (kv != null)
                {
                    key = kv.key;
                    val = kv.val;
                }
                else
                {
                    key = null;
                    val = null;
                }
                return;
            }
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "iter";
            message.data["kvs"] = kvsname;
            client.SendMessage(message);
            message = (PMIMessage)client.NextMessage();
            if ((string)message.data["success"] == "true")
            {
                key = (string)message.data["key"];
                val = (string)message.data["value"];
                //_mpiexec.WriteLine("iter pmi command returned true");
            }
            else
            {
                throw new Exception(String.Format("iter pmi command failed: {0}", (string)message.data["message"]));
                //_mpiexec.WriteLine("iter pmi command failed: {0}", (string)message.data["message"]);
            }
        }
        public void Spawn_multiple(string[] commands, string[][] argvs, int[] maxprocs, PMI_Keyval[][] keyvals, PMI_Keyval[] preput, string[] errors, bool quiet)
        {
            if (singleton)
            {
                singleton_pmi.Spawn_multiple(commands, argvs, maxprocs, keyvals, preput, errors, quiet);
                return;
            }
            
            int i;
            PMIMessage message = new PMIMessage();
            message.pmi_type = "client";
            message.name = "spawn";

            if (commands != null)
            {
                message.data["commands_count"] = commands.Length;
                for (i=0; i<commands.Length; i++)
                {
                    string cmd = commands[i];
                    if (commands[i].Length > 1 && commands[i][0] == '.' && commands[i][1] == '/')
                        cmd = ".\\" + commands[i].Substring(2);
                    message.data[String.Format("exe{0}", i)] = cmd;
                    if (maxprocs != null && maxprocs.Length > i)
                        message.data[String.Format("maxproc{0}", i)] = maxprocs[i];
                    else
                        message.data[String.Format("maxproc{0}", i)] = 1;
                    if (argvs != null && argvs.Length > i && argvs[i] != null)
                    {
                        message.data[String.Format("args{0}_count", i)] = argvs[i].Length;
                        for (int j = 0; j < argvs[i].Length; j++)
                        {
                            message.data[String.Format("args{0}_arg{1}", i, j)] = argvs[i][j];
                        }
                    }
                    else
                    {
                        message.data[String.Format("args{0}_count", i)] = 0;
                    }
                    if (keyvals != null && keyvals.Length > i && keyvals[i] != null)
                    {
                        message.data[String.Format("keyvals{0}_count", i)] = keyvals[i].Length;
                        for (int j = 0; j < keyvals[i].Length; j++)
                        {
                            message.data[String.Format("keyvals{0}_key{1}", i, j)] = keyvals[i][j].key;
                            message.data[String.Format("keyvals{0}_value{1}", i, j)] = keyvals[i][j].val;
                        }
                    }
                    else
                    {
                        message.data[String.Format("keyvals{0}_count", i)] = 0;
                    }
                }
            }
            else
            {
                message.data["commands_count"] = 0;
            }
            if (preput != null && preput.Length > 0)
            {
                message.data["preput_count"] = preput.Length;
                for (i=0; i<preput.Length; i++)
                {
                    message.data[String.Format("preput_key{0}", i)] = preput[i].key;
                    message.data[String.Format("preput_value{0}", i)] = preput[i].val;
                }
            }
            else
            {
                message.data["preput_count"] = 0;
            }
            message.data["quiet"] = quiet;

            try
            {
                //_mpiexec.WriteLine("Sending spawn pmi command.");
                client.SendMessage(message);
                //_mpiexec.WriteLine("spawn message sent, waiting for result");
                message = (PMIMessage)client.NextMessage();
            }
            catch (Exception ex)
            {
                /*
                _mpiexec.WriteLine("Exception caught trying to send the spawn command and receive the result");
                _mpiexec.WriteLine("Error: {0}", ex.Message);
                */
                throw new Exception(String.Format("Exception caught trying to send the spawn command and receive the result: {0}", ex.Message));
            }
            if ((string)message.data["success"] == "true")
            {
                //_mpiexec.WriteLine("spawn pmi command returned true");
                ProcessGroup.SpawnMessageStringToErrorStrings((string)message.data["message"], errors);
            }
            else
            {
                //_mpiexec.WriteLine("Spawn PMI command failed: {0}", (string)message.data["message"]);
                throw new Exception((string)message.data["message"]);
            }
        }
    }

    public class SmpdPmi
    {
        private Mpiexec _mpiexec;
        public const char QUOTE_CHAR = '"';
        public const char DELIM_CHAR = '=';
        public const string DELIM_STR = "=";
        public const char ESCAPE_CHAR = '\\';
        public const string ESCAPE_STR = "\\";
        public const char HIDE_CHAR = '*';
        public const char SEPAR_CHAR = ' ';

        public Hashtable data;

        public SmpdPmi()
        {
            _mpiexec = new Mpiexec();
            data = new Hashtable();
        }

        public SmpdPmi(Mpiexec mpiexec)
        {
            _mpiexec = mpiexec;
            if (_mpiexec == null)
                _mpiexec = new Mpiexec();
            data = new Hashtable();
        }

        public void Print()
        {
            _mpiexec.WriteLine("smpd command:");
            foreach (string key in data.Keys)
            {
                _mpiexec.WriteLine(" {0} = {1}", key, data[key]);
            }
            if (data.Keys.Count == 0)
            {
                _mpiexec.WriteLine(" empty");
            }
        }

        // The QuoteString, ReadString and ReadFullString functions do not quote strings correctly but this buggy code is required
        // since this is the format that existing MPICH2 programs provide the data.
        // In quoted strings only the " character is escaped.  The escape character is not escaped.
        // So this input:            "\"a \\""
        // will produce this output: "a \"
        // The error in not escaping the escape character is that a string cannot end in the escape character.
        // This string cannot be quoted: c:\temp\
        // The encoding becomes: "c:\temp\"
        // The encoded string cannot be decoded because the last escape character is mistakenly used to escape the end quote character.
        // The workaround is to do what MPICH2 mistakenly does and that is to always end strings with a space.
        static public string QuoteString(string str)
        {
            if (str.EndsWith(ESCAPE_STR))
                str = str + " ";

            if (str.IndexOf(SEPAR_CHAR) != -1 ||
                str.IndexOf(DELIM_CHAR) != -1 ||
                str.IndexOf(QUOTE_CHAR) != -1 /*||
                str.EndsWith(ESCAPE_STR)*/)
            {
                // quote
                StringBuilder sb = new StringBuilder();
                sb.Append(QUOTE_CHAR);
                foreach (char ch in str)
                {
                    if (ch == QUOTE_CHAR)
                        sb.Append(ESCAPE_CHAR);
                    /*
                    if (ch == ESCAPE_CHAR)
                        sb.Append(ESCAPE_CHAR);
                    */
                    sb.Append(ch);
                }
                sb.Append(QUOTE_CHAR);
                return sb.ToString();
            }
            // no need to quote
            return str;
        }

        /// <summary>
        /// Read a string with MPICH2 escaping
        /// </summary>
        /// <param name="str"></param>
        /// <param name="start_index"></param>
        /// <returns></returns>
        static public string ReadString(string str, ref int start_index)
        {
            StringBuilder token;
            int index = start_index;

            token = new StringBuilder();
            // skip over separator characters
            while (index < str.Length && str[index] == SEPAR_CHAR)
            {
                index++;
            }
            if (index == str.Length)
                return null;
            if (str[index] == DELIM_CHAR)
            {
                start_index = index + 1;
                return DELIM_STR;
            }
            if (str[index] == QUOTE_CHAR)
            {
                // read a quoted string
                index++;
                while (index < str.Length && str[index] != QUOTE_CHAR)
                {
                    if (str[index] == ESCAPE_CHAR)
                    {
                        index++;
                        if (index == str.Length)
                        {
                            token.Append(ESCAPE_CHAR);
                            // error, truncated string?
                        }
                        else if (str[index] == QUOTE_CHAR)
                        {
                            token.Append(QUOTE_CHAR);
                            index++;
                        }
                            /*
                        else if (str[index] == ESCAPE_CHAR)
                        {
                            token.Append(ESCAPE_CHAR);
                            index++;
                        }
                            */
                        else
                        {
                            token.Append(ESCAPE_CHAR);
                            // error, empty escape character?
                        }
                    }
                    else
                    {
                        token.Append(str[index]);
                        index++;
                    }
                }
                // move over the ending quote
                if (index < str.Length)
                    index++;
            }
            else
            {
                // read a literal string
                while (index < str.Length && str[index] != SEPAR_CHAR && str[index] != DELIM_CHAR)
                {
                    token.Append(str[index]);
                    index++;
                }
            }
            start_index = index;
            return token.ToString();
        }

        /// <summary>
        /// Read a string with correct escaping
        /// </summary>
        /// <param name="str"></param>
        /// <param name="start_index"></param>
        /// <returns></returns>
        static public string ReadStringTrue(string str, ref int start_index)
        {
            StringBuilder token;
            int index = start_index;

            token = new StringBuilder();
            // skip over separator characters
            while (index < str.Length && str[index] == SEPAR_CHAR)
            {
                index++;
            }
            if (index == str.Length)
                return null;
            if (str[index] == DELIM_CHAR)
            {
                start_index = index + 1;
                return DELIM_STR;
            }
            if (str[index] == QUOTE_CHAR)
            {
                // read a quoted string
                index++;
                while (index < str.Length && str[index] != QUOTE_CHAR)
                {
                    if (str[index] == ESCAPE_CHAR)
                    {
                        index++;
                        if (index == str.Length)
                        {
                            token.Append(ESCAPE_CHAR);
                            // error, truncated string?
                        }
                        else if (str[index] == QUOTE_CHAR)
                        {
                            token.Append(QUOTE_CHAR);
                            index++;
                        }
                        else if (str[index] == ESCAPE_CHAR)
                        {
                            token.Append(ESCAPE_CHAR);
                            index++;
                        }
                        else
                        {
                            token.Append(ESCAPE_CHAR);
                            // error, empty escape character?
                        }
                    }
                    else
                    {
                        token.Append(str[index]);
                        index++;
                    }
                }
                // move over the ending quote
                if (index < str.Length)
                    index++;
            }
            else
            {
                // read a literal string
                while (index < str.Length && str[index] != SEPAR_CHAR && str[index] != DELIM_CHAR)
                {
                    token.Append(str[index]);
                    index++;
                }
            }
            start_index = index;
            return token.ToString();
        }

        static public string ReadFullString(string str, ref int start_index)
        {
            StringBuilder token;
            int index = start_index;

            token = new StringBuilder();
            // skip over separator characters
            while (index < str.Length && str[index] == SEPAR_CHAR)
            {
                index++;
            }
            if (index == str.Length)
                return null;
            if (str[index] == QUOTE_CHAR)
            {
                // read a quoted string
                index++;
                while (index < str.Length && str[index] != QUOTE_CHAR)
                {
                    if (str[index] == ESCAPE_CHAR)
                    {
                        index++;
                        if (index == str.Length)
                        {
                            token.Append(ESCAPE_CHAR);
                            // error, truncated string?
                        }
                        else if (str[index] == QUOTE_CHAR)
                        {
                            token.Append(QUOTE_CHAR);
                            index++;
                        }
                            /*
                        else if (str[index] == ESCAPE_CHAR)
                        {
                            token.Append(ESCAPE_CHAR);
                            index++;
                        }
                            */
                        else
                        {
                            token.Append(ESCAPE_CHAR);
                            // error, empty escape character?
                        }
                    }
                    else
                    {
                        token.Append(str[index]);
                        index++;
                    }
                }
                // move over the ending quote
                if (index < str.Length)
                    index++;
            }
            else
            {
                // read a literal string
                while (index < str.Length && str[index] != SEPAR_CHAR)
                {
                    token.Append(str[index]);
                    index++;
                }
            }
            start_index = index;
            return token.ToString();
        }

        public void ParseCommand(string str)
        {
            string key, equals, val;
            int index = 0;

            data.Clear();
            while (true)
            {
                key = ReadString(str, ref index);
                if (key == null)
                    return;
                equals = ReadString(str, ref index);
                if (equals != DELIM_STR)
                    return;
                val = ReadString(str, ref index);
                if (val == null)
                    return;
                data[key] = val;
            }
        }

        static public Hashtable ParseLine(string str)
        {
            string key, equals, val;
            int index = 0;
            Hashtable data = new Hashtable();
            while (true)
            {
                key = ReadString(str, ref index);
                if (key == null)
                    return data;
                equals = ReadString(str, ref index);
                if (equals != DELIM_STR)
                    return data;
                val = ReadString(str, ref index);
                if (val == null)
                    return data;
                data[key] = val;
            }
        }

        static public Hashtable ParseLineTrue(string str)
        {
            string key, equals, val;
            int index = 0;
            Hashtable data = new Hashtable();
            while (true)
            {
                key = ReadStringTrue(str, ref index);
                if (key == null)
                    return data;
                equals = ReadStringTrue(str, ref index);
                if (equals != DELIM_STR)
                    return data;
                val = ReadStringTrue(str, ref index);
                if (val == null)
                    return data;
                data[key] = val;
            }
        }

        public bool NextMessage(TcpClient client)
        {
            byte[] buffer;
            string length_str, message_str;
            int length;
            BinaryReader br = new BinaryReader(client.GetStream(), Encoding.ASCII);

            buffer = br.ReadBytes(13);
            length_str = Encoding.ASCII.GetString(buffer);
            //_mpiexec.WriteLine("smpd header: '{0}'", length_str);
            length = Convert.ToInt32(length_str);
            buffer = br.ReadBytes(length);
            message_str = Encoding.ASCII.GetString(buffer);
            //_mpiexec.WriteLine("smpd command: '{0}'", message_str);
            ParseCommand(message_str);
            return true;
        }

        private void SendPMIMessage(TcpClient client, string str)
        {
            //_mpiexec.WriteLine("sending pmi message: {0}", str);
            byte[] message = Encoding.ASCII.GetBytes(str);
            byte[] header = Encoding.ASCII.GetBytes(message.Length.ToString());
            client.GetStream().Write(header, 0, header.Length);
            if (header.Length < 13)
            {
                byte[] padding = new byte[13 - header.Length];
                for (int i = 0; i < padding.Length; i++)
                    padding[i] = 0;
                client.GetStream().Write(padding, 0, padding.Length);
            }
            client.GetStream().Write(message, 0, message.Length);
        }

        public bool SendResultMessage(TcpClient client, string common, bool success)
        {
            StringBuilder str = new StringBuilder();
            str.Append("cmd=result ");
            if (success)
            {
                str.Append("result=SUCCESS ");
            }
            else
            {
                str.Append("result=FAIL ");
            }
            if (common != null)
            {
                str.Append(common);
            }
            SendPMIMessage(client, str.ToString());
            return true;
        }

        public bool SendDBResultMessage(TcpClient client, string common, bool success)
        {
            StringBuilder str = new StringBuilder();
            str.Append("cmd=result ");
            if (success)
            {
                str.Append("result=DBS_SUCCESS ");
            }
            else
            {
                str.Append("result=DBS_FAIL ");
            }
            if (common != null)
            {
                str.Append(common);
            }
            SendPMIMessage(client, str.ToString());
            return true;
        }

        public bool SendDBResultMessage(TcpClient client, string common, bool success, string name, string key, string val)
        {
            StringBuilder str = new StringBuilder();
            str.Append("cmd=result ");
            if (success)
            {
                str.Append("result=DBS_SUCCESS ");
            }
            else
            {
                str.Append("result=DBS_FAIL ");
            }
            if (common != null)
            {
                str.Append(common);
            }
            if (name != null)
            {
                str.AppendFormat("name={0} ", QuoteString(name));
            }
            if (key != null)
            {
                str.AppendFormat("key={0} ", QuoteString(key));
            }
            if (val != null)
            {
                str.AppendFormat("value={0}", QuoteString(val));
            }
            SendPMIMessage(client, str.ToString());
            return true;
        }

        internal static string[] ReadStrings(string line)
        {
            if (line == null)
                return null;
            int index = 0;
            ArrayList strings = new ArrayList();
            string str = ReadFullString(line, ref index);
            while (str != null)
            {
                strings.Add(str);
                str = ReadFullString(line, ref index);
            }
            if (strings.Count < 1)
                return null;
            string[] result = new string[strings.Count];
            for (index = 0; index < strings.Count; index++)
                result[index] = (string)strings[index];
            return result;
        }

        internal static string[] ReadKeyvals(string line)
        {
            if (line == null)
                return null;
            int index = 0;
            ArrayList strings = new ArrayList();
            string str = ReadString(line, ref index); // read the key
            str = ReadString(line, ref index); // read the =
            str = ReadString(line, ref index); // read the value
            while (str != null)
            {
                strings.Add(str);
                ReadString(line, ref index); // read the key
                str = ReadString(line, ref index); // read the =
                str = ReadString(line, ref index); // read the value
            }
            if (strings.Count < 1)
                return null;
            string[] result = new string[strings.Count];
            for (index = 0; index < strings.Count; index++)
                result[index] = (string)strings[index];
            return result;
        }
    }

    #endregion // PMI classes

    #region Session classes

    public class SecretData
    {
        public const string default_passphrase = "DeinoMPIte";
        public byte[] data;

        public SecretData()
        {
            data = Encoding.UTF8.GetBytes(default_passphrase);
        }

        public SecretData(string str)
        {
            if (str != null)
                data = Encoding.UTF8.GetBytes(str);
            else
                data = Encoding.UTF8.GetBytes(default_passphrase);
        }

        public string DataString
        {
            get
            {
                if (data != null)
                    return Encoding.UTF8.GetString(data);
                return "";
            }
            set
            {
                data = Encoding.UTF8.GetBytes(value);
            }
        }
    }

    public abstract class Session
    {
        protected Mpiexec _mpiexec;

        public const int DEINOPM_PORT = 62741;
        // Version information
        private const int _version_major = 2;
        private const int _version_minor = 0;
        private const int _version_sub_minor = 1;
        private const string _version_extra = "DeinoMPI";

        private SecretData _secret;
        protected TcpClient _client;
        protected BinaryReader _binary_reader;
        protected BinaryWriter _binary_writer;
        protected SecureMessageStream _secure_message_stream;

        public enum ImpersonationType { None, Identify, Impersonate, Delegate, Password };

        public Session()
        {
            _mpiexec = null;
            _client = null;
            _secret = null;
            _secure_message_stream = null;
        }

        public Session(TcpClient client)
        {
            _mpiexec = new Mpiexec();
            _secure_message_stream = null;
            _client = client;
            Initialize(null);
        }

        protected void Initialize(string secret)
        {
            _binary_reader = new BinaryReader(client.GetStream());
            _binary_writer = new BinaryWriter(client.GetStream());
            _secret = new SecretData(secret);
            VersionMessage version = new VersionMessage(_version_major, _version_minor, _version_sub_minor, _version_extra);
            // This assumes that you can buffer an entire version message in the send buffers
            // because both sides execute the following code at the same time:
            Message.SendMessage(_binary_writer, version);
            version = (VersionMessage)Message.NextMessage(_binary_reader);
            if (version.major != _version_major || version.minor != _version_minor || version.sub_minor != _version_sub_minor)
            {
                throw new Exception(String.Format("Version mismatch: {0}.{1}.{2} does not match {3}.{4}.{5}", version.major, version.minor, version.sub_minor, _version_major, _version_minor, _version_sub_minor));
            }
        }

        protected void Initialize(string secret, ref string version_str)
        {
            _binary_reader = new BinaryReader(client.GetStream());
            _binary_writer = new BinaryWriter(client.GetStream());
            _secret = new SecretData(secret);
            VersionMessage version = new VersionMessage(_version_major, _version_minor, _version_sub_minor, _version_extra);
            // This assumes that you can buffer an entire version message in the send buffers
            // because both sides execute the following code at the same time:
            Message.SendMessage(_binary_writer, version);
            version = (VersionMessage)Message.NextMessage(_binary_reader);
            /*
            if (version.major != _version_major || version.minor != _version_minor || version.sub_minor != _version_sub_minor)
            {
                throw new Exception(String.Format("Version mismatch: {0}.{1}.{2} does not match {3}.{4}.{5}", version.major, version.minor, version.sub_minor, _version_major, _version_minor, _version_sub_minor));
            }
            */
            version_str = String.Format("{0}.{1}.{2}", version.major, version.minor, version.sub_minor);
        }

        public Message NextMessage()
        {
            if (_secure_message_stream != null)
                return Message.NextMessage(_secure_message_stream);
            if (_binary_reader != null)
                return Message.NextMessage(_binary_reader);
            throw new Exception("No stream available for this session.");
        }

        public void SendMessage(Message message)
        {
            if (_secure_message_stream != null)
                Message.SendMessage(_secure_message_stream, message);
            else if (_binary_writer != null)
                Message.SendMessage(_binary_writer, message);
            else
            {
                throw new Exception("No stream available for this session.");
            }
        }

        #region getters and setters
        public TcpClient client
        {
            get { return _client; }
            set { _client = value; }
        }
        static public string Version
        {
            get { return String.Format("{0}.{1}.{2}", _version_major, _version_minor, _version_sub_minor); }
        }
        static public int version_major
        {
            get { return _version_major; }
        }
        static public int version_minor
        {
            get { return _version_minor; }
        }
        static public int version_sub_minor
        {
            get { return _version_sub_minor; }
        }
        static public string version_extra
        {
            get { return _version_extra; }
        }
        #endregion
    }

    public class ClientSession : Session
    {
        public ClientSession(TcpClient client, SecurityKeys private_key)
            : base(client)
        {
            //_mpiexec.WriteLine("client authenticating");
            Authenticate();
            SecureStream(private_key);
        }
        public ClientSession(string host, SecurityKeys private_key)
        {
            Random r = null;
            for (int i = 0; i < 5; i++)
            {
                try
                {
                    client = new TcpClient(host, DEINOPM_PORT);
                    break;
                }
                catch //(System.Net.Sockets.SocketException)
                {
                    if (r == null)
                        r = new Random();
                    int time = (int)(r.NextDouble() * 300.0);
                    //_mpiexec.WriteLine("Sleeping {0} milliseconds", time);
                    System.Threading.Thread.Sleep(time);
                }
            }
            client.NoDelay = true;
            //client.ReceiveTimeout = 30000;
            //_mpiexec.WriteLine("client initializing");
            Initialize(null);
            //_mpiexec.WriteLine("client authenticating");
            Authenticate();
            SecureStream(private_key);
        }
        public ClientSession(int port, SecurityKeys private_key)
        {
            Random r = null;
            string host = System.Net.Dns.GetHostName();
            for (int i = 0; i < 5; i++)
            {
                try
                {
                    client = new TcpClient(host, port);
                    break;
                }
                catch //(System.Net.Sockets.SocketException)
                {
                    if (r == null)
                        r = new Random();
                    int time = (int)(r.NextDouble() * 300.0);
                    //_mpiexec.WriteLine("Sleeping {0} milliseconds", time);
                    System.Threading.Thread.Sleep(time);
                }
            }
            client.NoDelay = true;
            //client.ReceiveTimeout = 30000;
            //_mpiexec.WriteLine("client initializing");
            Initialize(null);
            /* Authentication requires access to the Registry to read the passphrase which may not be possible because the process is now running under user credentials which may be limited.
            Authenticate();
            */
            //SecureStream(private_key);
            _secure_message_stream = SecureMessageStream.FromPrivateKeyNoSendPublicKey(private_key, client.GetStream());
        }
        public ClientSession(string host, int port, string phrase, SecurityKeys private_key, SecureCredential credential)
        {
            if (port == -1)
                port = DEINOPM_PORT;
            Random r = null;
            for (int i = 0; i < 5; i++)
            {
                try
                {
                    client = new TcpClient(host, port);
                    break;
                }
                catch (Exception e)
                {
                    if (i == 4)
                        throw e;
                    if (r == null)
                        r = new Random();
                    int time = (int)(r.NextDouble() * 300.0);
                    //_mpiexec.WriteLine("Sleeping {0} milliseconds", time);
                    System.Threading.Thread.Sleep(time);
                }
            }
            client.NoDelay = true;
            //client.ReceiveTimeout = 30000;
            //_mpiexec.WriteLine("rtimeout = {0}, stimeout = {1}", client.ReceiveTimeout, client.SendTimeout);
            //_mpiexec.WriteLine("client initializing");
            Initialize(phrase);
            //_mpiexec.WriteLine("client authenticating");
            Authenticate();
            SecureStream(private_key);
            Impersonate(phrase, private_key, credential);
        }
        public ClientSession(string host, int port)
        {
            if (port == -1)
                port = DEINOPM_PORT;
            Random r = null;
            for (int i = 0; i < 5; i++)
            {
                try
                {
                    client = new TcpClient(host, port);
                    break;
                }
                catch //(System.Net.Sockets.SocketException)
                {
                    if (r == null)
                        r = new Random();
                    int time = (int)(r.NextDouble() * 300.0);
                    //_mpiexec.WriteLine("Sleeping {0} milliseconds", time);
                    System.Threading.Thread.Sleep(time);
                }
            }
            client.NoDelay = true;
            //client.ReceiveTimeout = 30000;
            //_mpiexec.WriteLine("client initializing");
            Initialize(null);
            //_mpiexec.WriteLine("client authenticating");
            Authenticate();
        }

        public ClientSession(string host, int port, bool bAuthenticate)
        {
            if (port == -1)
                port = DEINOPM_PORT;
            Random r = null;
            for (int i = 0; i < 5; i++)
            {
                try
                {
                    client = new TcpClient(host, port);
                    break;
                }
                catch //(System.Net.Sockets.SocketException)
                {
                    if (r == null)
                        r = new Random();
                    int time = (int)(r.NextDouble() * 300.0);
                    //_mpiexec.WriteLine("Sleeping {0} milliseconds", time);
                    System.Threading.Thread.Sleep(time);
                }
            }
            client.NoDelay = true;
            //client.ReceiveTimeout = 30000;
            //_mpiexec.WriteLine("client initializing");
            Initialize(null);
            if (bAuthenticate)
            {
                //_mpiexec.WriteLine("client authenticating");
                Authenticate();
            }
        }

        public ClientSession(string host, int port, ref string version)
        {
            if (port == -1)
                port = DEINOPM_PORT;
            Random r = null;
            for (int i = 0; i < 5; i++)
            {
                try
                {
                    client = new TcpClient(host, port);
                    break;
                }
                catch //(System.Net.Sockets.SocketException)
                {
                    if (r == null)
                        r = new Random();
                    int time = (int)(r.NextDouble() * 300.0);
                    //_mpiexec.WriteLine("Sleeping {0} milliseconds", time);
                    System.Threading.Thread.Sleep(time);
                }
            }
            client.NoDelay = true;
            Initialize(null, ref version);
            client.Close();
        }

        private void Authenticate()
        {
            ResultMessage result;
            // receive the challenge
            //_mpiexec.WriteLine("receiving challenge");
            ChallengeMessage cn = (ChallengeMessage)NextMessage();
            //_mpiexec.WriteLine("received challenge: {0}", Convert.ToBase64String(cn.challenge));

            // compute the hash response
            ResponseMessage rn = new ResponseMessage();
            HashAlgorithm sha = new SHA1CryptoServiceProvider();
            SecretData secret = new SecretData();
            rn.response = new byte[cn.challenge.Length + secret.data.Length];
            Array.Copy(cn.challenge, rn.response, cn.challenge.Length);
            Array.Copy(secret.data, 0, rn.response, cn.challenge.Length, secret.data.Length);
            rn.response = sha.ComputeHash(rn.response);
            //_mpiexec.WriteLine("response hash: {0}", Convert.ToBase64String(rn.response));

            // send the response
            //_mpiexec.WriteLine("sending response");
            SendMessage(rn);

            // receive the result
            //_mpiexec.WriteLine("receiving result");
            result = (ResultMessage)NextMessage();
            if (!result.success)
            {
                throw new Exception("Challenge response authentication failed");
            }
            //_mpiexec.WriteLine("challenge-response succeeded");

        }

        private void SecureStream(SecurityKeys private_key)
        {
            //_mpiexec.WriteLine("Securing stream: providing private key");
            _secure_message_stream = SecureMessageStream.FromPrivateKey(private_key, _client.GetStream());
        }

        public void Impersonate(string phrase, SecurityKeys private_key, SecureCredential credential)
        {
            //_mpiexec.WriteLine("Impersonating user: {0}", credential.Account);
            ImpersonationMessage message = new ImpersonationMessage(credential.Account, credential.Password);
            SendMessage(message);
            ResultMessage result = (ResultMessage)NextMessage();
            if (result.success)
            {
                ReconnectRequestMessage rmessage = (ReconnectRequestMessage)NextMessage();
                //_mpiexec.WriteLine("Received reconnect request: {0}:{1}", rmessage.host, rmessage.port);

                // reconnect
                TcpClient rclient = new TcpClient(rmessage.host, rmessage.port);
                SecureMessageStream old_stream = _secure_message_stream;
                rclient.NoDelay = true;
                //rclient.ReceiveTimeout = 30000;
                _client = rclient;
                //_mpiexec.WriteLine("connected to {0}:{1}, initializing", rmessage.host, rmessage.port);
                Initialize(phrase);
                /* Authentication requires access to the Registry to read the passphrase which may not be possible because the process is now running under user credentials which may be limited.
                //_mpiexec.WriteLine("Initialization finished, authenticating");
                Authenticate();
                //_mpiexec.WriteLine("Authentication finished, securing");
                */
                //_mpiexec.WriteLine("Initialization finished, securing");
                _secure_message_stream = SecureMessageStream.FromPrivateKeyNoSendPublicKey(private_key, rclient.GetStream());
                //_mpiexec.WriteLine("Securing finished, sending result.");

                // send a success message
                result = new ResultMessage();
                result.success = true;
                result.message = "Successful reconnect to manager.";
                Message.SendMessage(old_stream, result);
                //_mpiexec.WriteLine("Result sent, closing old connection");
                old_stream.Close();
            }
            else
            {
                throw new Exception(result.message);
            }
        }

        public void Impersonate(ImpersonationType type)
        {
        }
    }

    public class ServerSession : Session
    {
        public static bool use_csharp_process = false;

        public ServerSession(TcpClient client)
            : base(client)
        {
            //_mpiexec.WriteLine("server authenticating");
            Authenticate();
            //_mpiexec.WriteLine("server securing connection");
            SecureStream();
            //_mpiexec.WriteLine("server impersonating");
            Impersonate();
            //_mpiexec.WriteLine("server constructed");
        }

        public ServerSession(TcpClient client, bool authenticate_only)
            : base(client)
        {
            //_mpiexec.WriteLine("server authenticating");
            Authenticate();
            //SecureStream();
            //if (!authenticate_only)
            //    Impersonate();
        }

        public ServerSession(TcpClient client, string PublicKeyString)
            : base(client)
        {
            // Authentication requires access to the Registry to read the secret word which may not be available once the process is under the user credentials
            //Authenticate();

            // PublicKeyString will be valid for mpiexec re-connect initializations
            // PublicKeyString will be null for active pmi connections
            if (PublicKeyString != null)
            {
                _secure_message_stream = SecureMessageStream.FromPublicKeyString(_client.GetStream(), PublicKeyString);
            }
        }

        private void Authenticate()
        {
            // send random challenge data
            ChallengeMessage cn = new ChallengeMessage();
            cn.challenge = new byte[100];
            Random r = new Random();
            r.NextBytes(cn.challenge);
            //_mpiexec.WriteLine("sending challenge");
            SendMessage(cn);

            // hash the data
            HashAlgorithm sha = new SHA1CryptoServiceProvider();
            SecretData secret = new SecretData();
            byte[] hash = new byte[cn.challenge.Length + secret.data.Length];
            Array.Copy(cn.challenge, hash, cn.challenge.Length);
            Array.Copy(secret.data, 0, hash, cn.challenge.Length, secret.data.Length);
            hash = sha.ComputeHash(hash);
            //_mpiexec.WriteLine("challenge hash: {0}", Convert.ToBase64String(hash));

            // receive the response
            //_mpiexec.WriteLine("receiving response");
            ResponseMessage rn = (ResponseMessage)NextMessage();

            // compare the results
            ResultMessage result = new ResultMessage();
            //_mpiexec.WriteLine("hashed-challenge: {0}", Convert.ToBase64String(hash));
            //_mpiexec.WriteLine("hashed-response : {0}", Convert.ToBase64String(rn.response));
            result.success = (Convert.ToBase64String(hash) == Convert.ToBase64String(rn.response));
            //_mpiexec.WriteLine("sending result: {0}", result.success);
            SendMessage(result);
            if (!result.success)
            {
                _mpiexec.WriteLine("challenge response failed");
                _mpiexec.WriteLine("client closing");
                client.Close();
                client = null;
                return;
            }
            //_mpiexec.WriteLine("client authenticated");

        }

        private void SecureStream()
        {
            //_mpiexec.WriteLine("Securing stream: accepting public key");
            if (_client == null)
                throw new Exception("Attempting to secure a stream that does not exist: _client = null.");
            _secure_message_stream = SecureMessageStream.FromPublicKey(_client.GetStream());
        }

        public class MyReadOutputThread
        {
            public Mpiexec mpiexec;
            public StreamReader output;
            public Queue queue;

            public void Run()
            {
                try
                {
                    string line = output.ReadLine();
                    while (line != null)
                    {
                        //mpiexec.WriteLine("DeinoPM: {0}", line);
                        queue.Enqueue(line);
                        line = output.ReadLine();
                    }
                }
                catch (Exception e)
                {
                    if (mpiexec != null)
                        mpiexec.WriteLine("Exception caught reading output: {0}", e.Message);
                }
                queue.Enqueue(queue);
            }
        }

        private void Impersonate()
        {
            ImpersonationMessage message = (ImpersonationMessage)NextMessage();
            //_mpiexec.WriteLine("impersonation message received: {0}", message.account);
            Process p = new Process();
            ProcessStartInfo pinfo = new ProcessStartInfo();
            pinfo.ErrorDialog = false;
            pinfo.CreateNoWindow = true;
            pinfo.FileName = Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location) + @"\DeinoPM.exe";
            pinfo.LoadUserProfile = true;
            pinfo.Password = new System.Security.SecureString();
            for (int i = 0; i < message.password.Length; i++)
                pinfo.Password.AppendChar(message.password[i]);
            if (message.account.Contains("\\"))
            {
                int index = message.account.IndexOf("\\");
                pinfo.Domain = message.account.Substring(0, index);
                pinfo.UserName = message.account.Substring(index + 1, message.account.Length - index - 1);
            }
            else
            {
                pinfo.UserName = message.account;
            }
            pinfo.RedirectStandardOutput = true;
            pinfo.RedirectStandardInput = true;
            pinfo.RedirectStandardError = true;
            string single_arg_encoding = Convert.ToBase64String(Encoding.Unicode.GetBytes(_secure_message_stream.PublicKey.PublicKey));
            pinfo.Arguments = "/public_key";
            pinfo.UseShellExecute = false;
            p.StartInfo = pinfo;

            ProcessCreation pc = new ProcessCreation();

            string port_str = "";
            ResultMessage start_result = new ResultMessage();
            start_result.success = true;
            Random r = null;
            for (int i = 0; i < 5; i++)
            {
                try
                {
                    _mpiexec.WriteLine("Starting DeinoPM manager for user {0}", message.account);
                    if (use_csharp_process)
                    {
                        /*
                        // For testing remove the username stuff so I can attach and debug the processes.
                        pinfo.Domain = null;
                        pinfo.UserName = null;
                        pinfo.Password = null;
                        */
                        p.Start();
                        p.StandardInput.WriteLine(single_arg_encoding);
                        port_str = p.StandardOutput.ReadLine();
                    }
                    else
                    {
                        pc.DeinoLaunchProcess(String.Format("\"{0}\" /public_key", pinfo.FileName), pinfo.Domain, pinfo.UserName, message.password);
                        //_mpiexec.WriteLine("PublicKey encoding: {0}", single_arg_encoding);
                        pc.DeinoLPWriteLine(single_arg_encoding);
                        //_mpiexec.WriteLine("Wrote PublicKey, reading manager listener port");
                        port_str = pc.DeinoLPReadLine();
                        //_mpiexec.WriteLine("Read port from the manager: {0}", port_str);
                    }
                    //_mpiexec.WriteLine("DeinoPM manager started, waiting for reconnect on port {0}", port_str);
                    break;
                }
                catch (Win32Exception e)
                {
                    if ((e.NativeErrorCode == 1723) && (i < 4))
                    {
                        // RPC Server busy error so wait a second and try again
                        if (r == null)
                            r = new Random();
                        int time = (int)(r.NextDouble() * 300.0);
                        //_mpiexec.WriteLine("Sleeping {0} milliseconds", time);
                        System.Threading.Thread.Sleep(time);
                    }
                    else
                    {
                        start_result.message = String.Format("Starting '{0}' failed: {1}, error #{2}", pinfo.FileName, e.Message, e.NativeErrorCode);
                        start_result.success = false;
                    }
                }
                catch (Exception e)
                {
                    start_result.message = e.Message;
                    start_result.success = false;
                }
                if (!start_result.success)
                    break;
            }

            bool killed_or_exited = false;
            try
            {
                SendMessage(start_result);
                if (start_result.success)
                {

                    int port = Convert.ToInt32(port_str);
                    //_mpiexec.WriteLine("port = {0}", port);
                    ReconnectRequestMessage rmessage = new ReconnectRequestMessage(System.Net.Dns.GetHostName(), port);
                    SendMessage(rmessage);

                    if (use_csharp_process)
                    {
                        // Create two threads to collect the output from the manager process
                        Queue q = Queue.Synchronized(new Queue());

                        MyReadOutputThread othread_data = new MyReadOutputThread();
                        othread_data.mpiexec = _mpiexec;
                        othread_data.output = p.StandardOutput;
                        othread_data.queue = q;
                        ThreadStart ts = new ThreadStart(othread_data.Run);
                        Thread othread = new Thread(ts);
                        othread.Start();

                        MyReadOutputThread ethread_data = new MyReadOutputThread();
                        ethread_data.mpiexec = _mpiexec;
                        ethread_data.output = p.StandardError;
                        ethread_data.queue = q;
                        ThreadStart ts2 = new ThreadStart(ethread_data.Run);
                        Thread ethread = new Thread(ts2);
                        ethread.Start();

                        // Receive the result message from mpiexec to see if it was able to re-connect to the manager process
                        ResultMessage result = (ResultMessage)NextMessage();
                        //_mpiexec.WriteLine("result: {0} {1}", result.success.ToString(), result.message);
                        if (result.success == false)
                        {
                            _mpiexec.WriteLine("Unable to launch a DeinoPM manager for user {0}", message.account);
                            p.Kill();
                            killed_or_exited = true;
                        }

                        // Read all the output from the manager process
                        int exited_count = 2;
                        while (exited_count > 0)
                        {
                            object o;
                            while (q.Count == 0)
                                Thread.Sleep(100);
                            o = q.Dequeue();
                            if (o == q)
                            {
                                //_mpiexec.WriteLine("MyOutputThread exited");
                                exited_count--;
                            }
                            else
                            {
                                string s = (string)o;
                                _mpiexec.WriteLine("DeinoPM.{0}: {1}", message.account, s);
                            }
                        }
                        killed_or_exited = true;
                    }
                    else
                    {
                        // Receive the result message from mpiexec to see if it was able to re-connect to the manager process
                        ResultMessage result = (ResultMessage)NextMessage();
                        //_mpiexec.WriteLine("result: {0} {1}", result.success.ToString(), result.message);
                        if (result.success == false)
                        {
                            _mpiexec.WriteLine("Unable to launch a DeinoPM manager for user {0}", message.account);
                            pc.Kill();
                            killed_or_exited = true;
                        }
                        pc.DeinoWaitForExit();
                        killed_or_exited = true;
                        /* Since DeinoLaunchProcess does not redirect stdin/out/err we cannot simply call ReadLine until the process exits.
                        string s = pc.DeinoLPReadLine();
                        while (s != null)
                        {
                            _mpiexec.WriteLine("DeinoPM.{0}: {1}", message.account, s);
                            s = pc.DeinoLPReadLine();
                        }
                        */
                    }
                }
                else
                {
                    // Make sure no manager processes remain
                    try
                    {
                        if (use_csharp_process)
                            p.Kill();
                        else
                            pc.Kill();
                    }
                    catch
                    {
                        // It is ok for the Kill to fail since the process may not have been started.
                    }
                }
            }
            catch (Exception ex)
            {
                if (!killed_or_exited)
                {
                    try
                    {
                        if (use_csharp_process)
                            p.Kill();
                        else
                            pc.Kill();
                    }
                    catch
                    {
                        // It is ok for the Kill to fail since the process may not have been started.
                    }
                }
                throw ex;
            }
        }
    }


    #endregion // Session classes

    #region Message classes

    public abstract class Message
    {
        //protected Mpiexec _mpiexec;
        protected string type;
        public string MessageType
        {
            get
            {
                return type;
            }
        }
        protected void StartXmlWriter(XmlTextWriter xwriter, string type)
        {
            xwriter.Formatting = Formatting.Indented;
            xwriter.WriteStartDocument();
            xwriter.WriteStartElement("deinopm_message");
            xwriter.WriteElementString("type", type);
        }
        protected void FinishXmlWriter(XmlTextWriter xwriter)
        {
            xwriter.WriteEndElement();
            xwriter.WriteEndDocument();
        }
        protected static string StartXmlReader(XmlTextReader xreader)
        {
            xreader.ReadStartElement("deinopm_message");
            return xreader.ReadElementString("type");
        }
        protected static void FinishXmlReader(XmlTextReader xreader)
        {
            xreader.ReadEndElement();
        }
        public abstract byte[] CreateXmlBuffer();
        public static Message FromXmlBuffer(byte[] buffer)
        {
            Message message;
            string type;
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            type = StartXmlReader(xreader);
            switch (type)
            {
                case InputOutputMessage.TypeString:
                    message = new InputOutputMessage(xreader);
                    break;
                case VersionMessage.TypeString:
                    message = new VersionMessage(xreader);
                    break;
                case LaunchMessage.TypeString:
                    message = new LaunchMessage(xreader);
                    break;
                case ChallengeMessage.TypeString:
                    message = new ChallengeMessage(xreader);
                    break;
                case ResponseMessage.TypeString:
                    message = new ResponseMessage(xreader);
                    break;
                case ResultMessage.TypeString:
                    message = new ResultMessage(xreader);
                    break;
                case ProcessExitMessage.TypeString:
                    message = new ProcessExitMessage(xreader);
                    break;
                case BufferMessage.TypeString:
                    message = new BufferMessage(xreader);
                    break;
                case QuitMessage.TypeString:
                    message = new QuitMessage(xreader);
                    break;
                case AbortProcessMessage.TypeString:
                    message = new AbortProcessMessage(xreader);
                    break;
                case PMIMessage.TypeString:
                    message = new PMIMessage(xreader);
                    //((PMIMessage)message).Print();
                    break;
                case ImpersonationMessage.TypeString:
                    message = new ImpersonationMessage(xreader);
                    break;
                case ReconnectRequestMessage.TypeString:
                    message = new ReconnectRequestMessage(xreader);
                    break;
                case ActiveMessage.TypeString:
                    message = new ActiveMessage(xreader);
                    break;
                case HostInfoMessage.TypeString:
                    message = new HostInfoMessage(xreader);
                    break;
                case InfoRequestMessage.TypeString:
                    message = new InfoRequestMessage(xreader);
                    break;
                case LaunchInfoMessage.TypeString:
                    message = new LaunchInfoMessage(xreader);
                    break;
                default:
                    throw new Exception(String.Format("Invalid Message type: '{0}'", type));
            }
            FinishXmlReader(xreader);
            xreader.Close();
            //_mpiexec.WriteLine("received {0} message", message.type);
            return message;
        }

        public static Message NextMessage(BinaryReader binary_reader)
        {
            Message message;
            int length;
            byte[] buffer;
            lock (binary_reader)
            {
                length = binary_reader.ReadInt32();
                buffer = binary_reader.ReadBytes(length);
            }
            //_mpiexec.WriteLine("message xml: {0}", Encoding.UTF8.GetString(buffer));
            message = Message.FromXmlBuffer(buffer);
            //_mpiexec.WriteLine("received {0} message", message.type);
            return message;
        }

        public static Message NextMessage(SecureMessageStream secure_reader)
        {
            Message message;
            byte[] buffer;
            buffer = secure_reader.Read();
            //_mpiexec.WriteLine("message xml: {0}", Encoding.UTF8.GetString(buffer));
            try
            {
                message = Message.FromXmlBuffer(buffer);
            }
            catch (Exception e)
            {
                /*
                _mpiexec.WriteLine("Exception caught interpreting xml:");
                _mpiexec.WriteLine("Length: {0}", buffer.Length);
                _mpiexec.WriteLine("Data:");
                _mpiexec.WriteLine(Encoding.UTF8.GetString(buffer));
                */
                throw e;
            }
            //_mpiexec.WriteLine("received {0} message", message.type);
            return message;
        }

        public static void SendMessage(BinaryWriter binary_writer, Message message)
        {
            byte[] buffer = message.CreateXmlBuffer();

            MemoryStream ms = new MemoryStream();
            BinaryWriter bw = new BinaryWriter(ms);
            bw.Write(buffer.Length);
            bw.Write(buffer);
            bw.Flush();
            byte[] single_message = ms.ToArray();
            lock (binary_writer)
            {
                binary_writer.Write(single_message);
            }
            /*
            lock (binary_writer)
            {
                //_mpiexec.WriteLine("sending {0} message: length {1}", message.type, buffer.Length);
                binary_writer.Write(buffer.Length);
                binary_writer.Write(buffer);
            }
            */
            //_mpiexec.WriteLine("sent {0} message", message.type);
        }

        public static void SendMessage(SecureMessageStream secure_writer, Message message)
        {
            byte[] buffer = message.CreateXmlBuffer();
            secure_writer.Write(buffer);
            //_mpiexec.WriteLine("sent {0} message", message.type);
        }
    }

    public class VersionMessage : Message
    {
        public const string TypeString = "version";

        private int _major;
        private int _minor;
        private int _sub_minor;
        private string _extra;

        public VersionMessage()
        {
            type = TypeString;
        }
        public VersionMessage(int major, int minor, int sub_minor, string extra)
        {
            type = TypeString;
            _major = major;
            _minor = minor;
            _sub_minor = sub_minor;
            _extra = extra;
        }
        public VersionMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadVersionXml(xreader);
        }

        #region getters and setters
        public int major
        {
            get { return _major; }
            set { _major = value; }
        }
        public int minor
        {
            get { return _minor; }
            set { _minor = value; }
        }
        public int sub_minor
        {
            get { return _sub_minor; }
            set { _sub_minor = value; }
        }
        public string extra
        {
            get { return _extra; }
            set { _extra = value; }
        }
        #endregion

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateVersionXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadVersionXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateVersionXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, VersionMessage.TypeString);
            xwriter.WriteStartElement("version");
            xwriter.WriteElementString("major", XmlConvert.ToString(_major));
            xwriter.WriteElementString("minor", XmlConvert.ToString(_minor));
            xwriter.WriteElementString("sub_minor", XmlConvert.ToString(_sub_minor));
            xwriter.WriteElementString("extra", _extra);
            xwriter.WriteEndElement();
            FinishXmlWriter(xwriter);
        }

        public void ReadVersionXml(XmlTextReader xreader)
        {
            xreader.ReadStartElement("version");
            _major = XmlConvert.ToInt32(xreader.ReadElementString("major"));
            _minor = XmlConvert.ToInt32(xreader.ReadElementString("minor"));
            _sub_minor = XmlConvert.ToInt32(xreader.ReadElementString("sub_minor"));
            _extra = xreader.ReadElementString("extra");
            xreader.ReadEndElement();
        }
    }

    public class ImpersonationMessage : Message
    {
        public const string TypeString = "impersonation";

        private Session.ImpersonationType _impersonation_type;
        private string _account, _password;

        public ImpersonationMessage()
        {
            type = TypeString;
            _impersonation_type = Session.ImpersonationType.None;
        }
        public ImpersonationMessage(string account, string password)
        {
            type = TypeString;
            _impersonation_type = Session.ImpersonationType.Password;
            _account = account;
            _password = password;
        }
        public ImpersonationMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadImpersonationXml(xreader);
        }

        #region getters and setters
        public Session.ImpersonationType impersonation_type
        {
            get { return _impersonation_type; }
            set { _impersonation_type = value; }
        }
        public string account
        {
            get { return _account; }
            set { _account = value; }
        }
        public string password
        {
            get { return _password; }
            set { _password = value; }
        }
        #endregion

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateImpersonationXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadImpersonationXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateImpersonationXml(XmlTextWriter xwriter)
        {
            string temp = "None";
            StartXmlWriter(xwriter, ImpersonationMessage.TypeString);
            xwriter.WriteStartElement(TypeString);
            switch (_impersonation_type)
            {
                case Session.ImpersonationType.None:
                    temp = "None";
                    break;
                case Session.ImpersonationType.Delegate:
                    temp = "Delegate";
                    break;
                case Session.ImpersonationType.Identify:
                    temp = "Identify";
                    break;
                case Session.ImpersonationType.Impersonate:
                    temp = "Impersonate";
                    break;
                case Session.ImpersonationType.Password:
                    temp = "Password";
                    break;
            }
            xwriter.WriteElementString("impersonation_type", temp);
            if (_impersonation_type == Session.ImpersonationType.Password)
            {
                xwriter.WriteElementString("account", _account);
                xwriter.WriteElementString("password", _password);
            }
            xwriter.WriteEndElement();
            FinishXmlWriter(xwriter);
        }

        public void ReadImpersonationXml(XmlTextReader xreader)
        {
            xreader.ReadStartElement(TypeString);
            string temp = xreader.ReadElementString("impersonation_type");
            switch (temp)
            {
                case "None":
                    _impersonation_type = Session.ImpersonationType.None;
                    break;
                case "Delegate":
                    _impersonation_type = Session.ImpersonationType.Delegate;
                    break;
                case "Identify":
                    _impersonation_type = Session.ImpersonationType.Identify;
                    break;
                case "Impersonate":
                    _impersonation_type = Session.ImpersonationType.Impersonate;
                    break;
                case "Password":
                    _impersonation_type = Session.ImpersonationType.Password;
                    break;
            }
            if (_impersonation_type == Session.ImpersonationType.Password)
            {
                _account = xreader.ReadElementString("account");
                _password = xreader.ReadElementString("password");
            }
            xreader.ReadEndElement();
        }
    }

    public class ReconnectRequestMessage : Message
    {
        public const string TypeString = "reconnect";

        private string _host;
        private int _port;

        public ReconnectRequestMessage()
        {
            type = TypeString;
        }
        public ReconnectRequestMessage(string host, int port)
        {
            type = TypeString;
            _host = host;
            _port = port;
        }
        public ReconnectRequestMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadReconnectRequestXml(xreader);
        }

        #region getters and setters
        public string host
        {
            get { return _host; }
            set { _host = value; }
        }
        public int port
        {
            get { return _port; }
            set { _port = value; }
        }
        #endregion

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateReconnectRequestXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadReconnectRequestXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateReconnectRequestXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, ReconnectRequestMessage.TypeString);
            xwriter.WriteStartElement(TypeString);
            xwriter.WriteElementString("host", _host);
            xwriter.WriteElementString("port", XmlConvert.ToString(_port));
            xwriter.WriteEndElement();
            FinishXmlWriter(xwriter);
        }

        public void ReadReconnectRequestXml(XmlTextReader xreader)
        {
            xreader.ReadStartElement(TypeString);
            _host = xreader.ReadElementString("host");
            _port = XmlConvert.ToInt32(xreader.ReadElementString("port"));
            xreader.ReadEndElement();
        }
    }

    public class LaunchMessage : Message
    {
        public const string TypeString = "launch";
        public string path, exe;
        public ArrayList args;
        public string dir;
        public Hashtable env;
        public Hashtable map;
        public int priority_class, priority_thread;

        private void Initialize()
        {
            priority_class = 2;
            priority_thread = 3;
            args = new ArrayList();
            env = new Hashtable();
            map = new Hashtable();
        }

        public LaunchMessage()
        {
            type = TypeString;
            Initialize();
        }
        public LaunchMessage(XmlTextReader xreader)
        {
            type = TypeString;
            Initialize();
            ReadLaunchXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateLaunchXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadLaunchXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateLaunchXml(XmlTextWriter xwriter)
        {
            /*
            xwriter.Formatting = Formatting.Indented;
            xwriter.WriteStartDocument();
            xwriter.WriteStartElement("deinopm_message");
            */
            StartXmlWriter(xwriter, LaunchMessage.TypeString);
            xwriter.WriteStartElement("launch");
            //xwriter.WriteAttributeString("type", "launch");
            xwriter.WriteElementString("path", path);
            xwriter.WriteElementString("executable", exe);
            xwriter.WriteStartElement("args");
            xwriter.WriteElementString("count", XmlConvert.ToString(args.Count));
            //xwriter.WriteAttributeString("count", XmlConvert.ToString(args.Count));
            foreach (string arg in args)
            {
                xwriter.WriteElementString("arg", arg);
            }
            xwriter.WriteEndElement();
            xwriter.WriteElementString("wdir", dir);
            xwriter.WriteStartElement("environment");
            xwriter.WriteElementString("count", XmlConvert.ToString(env.Count));
            //xwriter.WriteAttributeString("count", XmlConvert.ToString(env.Count));
            foreach (string key in env.Keys)
            {
                xwriter.WriteStartElement("env");
                xwriter.WriteElementString("name", key);
                xwriter.WriteElementString("value", (string)env[key]);
                xwriter.WriteEndElement();
            }
            xwriter.WriteEndElement();
            xwriter.WriteStartElement("drive_mappings");
            xwriter.WriteElementString("count", XmlConvert.ToString(map.Count));
            //xwriter.WriteAttributeString("count", XmlConvert.ToString(map.Count));
            foreach (string drive in map.Keys)
            {
                xwriter.WriteStartElement("map");
                xwriter.WriteElementString("drive", drive);
                xwriter.WriteElementString("share", (string)map[drive]);
                xwriter.WriteEndElement();
            }
            xwriter.WriteEndElement();
            xwriter.WriteElementString("priority_class", XmlConvert.ToString(priority_class));
            xwriter.WriteElementString("priority_thread", XmlConvert.ToString(priority_thread));
            xwriter.WriteEndElement();
            /*
            xwriter.WriteEndElement();
            xwriter.WriteEndDocument();
            */
            FinishXmlWriter(xwriter);
        }

        public void ReadLaunchXml(XmlTextReader xreader)
        {
            int count;
            //string type;
            Initialize();
            //xreader.ReadStartElement("deinopm_message");
            //StartXmlReader(xreader);
            /*
            type = xreader.GetAttribute("type");
            if (type != "launch")
            {
                throw new Exception("Invalid deinopm_message type: '" + type + "'");
            }
            */
            xreader.ReadStartElement("launch");
            path = xreader.ReadElementString("path");
            //xreader.ReadStartElement("path");
            //path = xreader.ReadString();
            //xreader.ReadEndElement();
            //_mpiexec.WriteLine("path = {0}", path);
            exe = xreader.ReadElementString("executable");
            //xreader.ReadStartElement("executable");
            //exe = xreader.ReadString();
            //xreader.ReadEndElement();
            //_mpiexec.WriteLine("exe = {0}", exe);
            xreader.ReadStartElement("args");
            //count = XmlConvert.ToInt32(xreader.GetAttribute("count"));
            count = XmlConvert.ToInt32(xreader.ReadElementString("count"));
            //_mpiexec.WriteLine("count = {0}", count);
            for (int i = 0; i < count; i++)
            {
                args.Add(xreader.ReadElementString("arg"));
                //_mpiexec.WriteLine("arg = {0}", args[args.Count-1]);
            }
            xreader.ReadEndElement();
            dir = xreader.ReadElementString("wdir");
            //_mpiexec.WriteLine("dir = {0}", dir);
            xreader.ReadStartElement("environment");
            //count = XmlConvert.ToInt32(xreader.GetAttribute("count"));
            count = XmlConvert.ToInt32(xreader.ReadElementString("count"));
            //_mpiexec.WriteLine("count = {0}", count);
            for (int i = 0; i < count; i++)
            {
                string env_name, env_value;
                xreader.ReadStartElement("env");
                env_name = xreader.ReadElementString("name");
                //_mpiexec.WriteLine("env_name = {0}", env_name);
                env_value = xreader.ReadElementString("value");
                //_mpiexec.WriteLine("env_value = {0}", env_value);
                env[env_name] = env_value;
                xreader.ReadEndElement();
            }
            xreader.ReadEndElement();
            xreader.ReadStartElement("drive_mappings");
            //count = XmlConvert.ToInt32(xreader.GetAttribute("count"));
            count = XmlConvert.ToInt32(xreader.ReadElementString("count"));
            //_mpiexec.WriteLine("count = {0}", count);
            for (int i = 0; i < count; i++)
            {
                string drive, share;
                xreader.ReadStartElement("map");
                drive = xreader.ReadElementString("drive");
                //_mpiexec.WriteLine("drive = {0}", drive);
                share = xreader.ReadElementString("share");
                //_mpiexec.WriteLine("share = {0}", share);
                map[drive] = share;
                xreader.ReadEndElement();
            }
            xreader.ReadEndElement();
            priority_class = XmlConvert.ToInt32(xreader.ReadElementString("priority_class"));
            //_mpiexec.WriteLine("priority_class = {0}", priority_class);
            priority_thread = XmlConvert.ToInt32(xreader.ReadElementString("priority_thread"));
            //_mpiexec.WriteLine("priority_thread = {0}", priority_thread);
            xreader.ReadEndElement();
            //xreader.ReadEndElement();
            //FinishXmlReader(xreader);
        }

        public void Print(Mpiexec mpiexec)
        {
            lock (typeof(LaunchMessage))
            {
                mpiexec.WriteLine("Launch:");
                if (exe != null)
                    mpiexec.WriteLine(" exe: {0}", exe);
                if (args != null)
                {
                    mpiexec.Write(" args: ");
                    foreach (string str in args)
                    {
                        mpiexec.Write("{0}, ", str);
                    }
                    mpiexec.WriteLine();
                }
                if (path != null)
                    mpiexec.WriteLine(" path: {0}", path);
                if (dir != null)
                    mpiexec.WriteLine(" dir: {0}", dir);
                foreach (string key in env.Keys)
                {
                    mpiexec.WriteLine(" env: {0} = {1}", key, env[key]);
                }
                foreach (string drive in map.Keys)
                {
                    mpiexec.WriteLine(" map: {0}:{1}", drive, map[drive]);
                }
                mpiexec.WriteLine(" priority: {0}:{1}", priority_class, priority_thread);
            }
        }
    }

    public class QuitMessage : Message
    {
        public const string TypeString = "quit";
        public QuitMessage()
        {
            type = TypeString;
        }
        public QuitMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadQuitXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateQuitXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadQuitXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateQuitXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, QuitMessage.TypeString);
            FinishXmlWriter(xwriter);
        }

        public void ReadQuitXml(XmlTextReader xreader)
        {
        }
    }

    public class AbortProcessMessage : Message
    {
        public const string TypeString = "abort_process";
        public int exit_code;

        public AbortProcessMessage()
        {
            type = TypeString;
        }
        public AbortProcessMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadAbortProcessXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateAbortProcessXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadAbortProcessXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateAbortProcessXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, AbortProcessMessage.TypeString);
            xwriter.WriteElementString("exit_code", (string)XmlConvert.ToString(exit_code));
            FinishXmlWriter(xwriter);
        }

        public void ReadAbortProcessXml(XmlTextReader xreader)
        {
            exit_code = XmlConvert.ToInt32(xreader.ReadElementString("exit_code"));
        }
    }

    public class ChallengeMessage : Message
    {
        public const string TypeString = "challenge";
        public string version;
        public byte[] challenge;

        public ChallengeMessage()
        {
            type = TypeString;
        }
        public ChallengeMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadChallengeXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateChallengeXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadChallengeXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateChallengeXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, ChallengeMessage.TypeString);
            //xwriter.WriteElementString("length", XmlConvert.ToString(challenge.Length));
            xwriter.WriteStartElement("data");
            xwriter.WriteBase64(challenge, 0, challenge.Length);
            xwriter.WriteEndElement();
            FinishXmlWriter(xwriter);
        }

        public void ReadChallengeXml(XmlTextReader xreader)
        {
            //int length = XmlConvert.ToInt32(xreader.ReadElementString("length"));
            challenge = Convert.FromBase64String(xreader.ReadElementString("data"));
            //_mpiexec.WriteLine("challenge string: {0} bytes", challenge.Length);
        }
    }

    public class ResponseMessage : Message
    {
        public const string TypeString = "response";
        public byte[] response;

        public ResponseMessage()
        {
            type = TypeString;
        }
        public ResponseMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadResponseXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateResponseXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadResponseXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateResponseXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, ResponseMessage.TypeString);
            //xwriter.WriteElementString("length", XmlConvert.ToString(response.Length));
            xwriter.WriteStartElement("data");
            xwriter.WriteBase64(response, 0, response.Length);
            xwriter.WriteEndElement();
            FinishXmlWriter(xwriter);
        }

        public void ReadResponseXml(XmlTextReader xreader)
        {
            response = Convert.FromBase64String(xreader.ReadElementString("data"));
            /*
            int length = XmlConvert.ToInt32(xreader.ReadElementString("length"));
            xreader.ReadStartElement("data");
            response = new byte[length];
            xreader.ReadBase64(response, 0, length);
            xreader.ReadEndElement();
            */
        }
    }

    public class ProcessExitMessage : Message
    {
        public const string TypeString = "exit";
        public string message;
        public int exit_code;

        public ProcessExitMessage()
        {
            type = TypeString;
        }
        public ProcessExitMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadProcessExitXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateProcessExitXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadProcessExitXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateProcessExitXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, ProcessExitMessage.TypeString);
            xwriter.WriteElementString("exit_code", XmlConvert.ToString(exit_code));
            if (message == null) message = "";
            xwriter.WriteElementString("message", message);
            FinishXmlWriter(xwriter);
        }

        public void ReadProcessExitXml(XmlTextReader xreader)
        {
            exit_code = XmlConvert.ToInt32(xreader.ReadElementString("exit_code"));
            message = xreader.ReadElementString("message");
        }
    }

    public class InputOutputMessage : Message
    {
        public const string TypeString = "input_output";
        public enum IO_Type { INPUT, OUTPUT, ERROR };
        public IO_Type io_type;
        private byte[] buffer;

        public InputOutputMessage()
        {
            type = TypeString;
            io_type = IO_Type.OUTPUT;
        }
        public InputOutputMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadInputOutputXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateInputOutputXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadInputOutputXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateInputOutputXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, InputOutputMessage.TypeString);
            switch (io_type)
            {
                case IO_Type.INPUT:
                    xwriter.WriteElementString("io_type", "INPUT");
                    break;
                case IO_Type.OUTPUT:
                    xwriter.WriteElementString("io_type", "OUTPUT");
                    break;
                case IO_Type.ERROR:
                    xwriter.WriteElementString("io_type", "ERROR");
                    break;
            }
            //xwriter.WriteElementString("length", XmlConvert.ToString(buffer.Length));
            xwriter.WriteStartElement("data");
            xwriter.WriteBase64(buffer, 0, buffer.Length);
            xwriter.WriteEndElement();
            FinishXmlWriter(xwriter);
        }

        public void ReadInputOutputXml(XmlTextReader xreader)
        {
            /*
            int length = XmlConvert.ToInt32(xreader.ReadElementString("length"));
            xreader.ReadStartElement("data");
            buffer = new byte[length];
            xreader.ReadBase64(buffer, 0, length);
            xreader.ReadEndElement();
            */
            string io_type_str = xreader.ReadElementString("io_type");
            switch (io_type_str)
            {
                case "INPUT":
                    io_type = IO_Type.INPUT;
                    break;
                case "OUTPUT":
                    io_type = IO_Type.OUTPUT;
                    break;
                case "ERROR":
                    io_type = IO_Type.ERROR;
                    break;
            }
            buffer = Convert.FromBase64String(xreader.ReadElementString("data"));
        }

        public void SetBuffer(char[] input, int start_index, int length)
        {
            char[] temp_buffer = new char[length];
            Array.Copy(input, start_index, temp_buffer, 0, length);
            Buffer = temp_buffer;
        }

        public void SetBuffer(byte[] input, int start_index, int length)
        {
            byte[] temp_buffer = new byte[length];
            Array.Copy(input, start_index, temp_buffer, 0, length);
            buffer = temp_buffer;
        }

        public void SetBuffer(char[] input)
        {
            Buffer = input;
        }

        public char[] Buffer
        {
            get
            {
                return Encoding.UTF8.GetChars(buffer);
            }
            set
            {
                buffer = Encoding.UTF8.GetBytes(value);
            }
        }
    }

    public class ResultMessage : Message
    {
        public const string TypeString = "result";
        public bool success;
        public string message;

        public ResultMessage()
        {
            type = TypeString;
            success = true;
        }
        public ResultMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadResultXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateResultXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadResultXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateResultXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, ResultMessage.TypeString);
            xwriter.WriteElementString("success", XmlConvert.ToString(success));
            xwriter.WriteElementString("message", message);
            FinishXmlWriter(xwriter);
        }

        public void ReadResultXml(XmlTextReader xreader)
        {
            success = XmlConvert.ToBoolean(xreader.ReadElementString("success"));
            message = xreader.ReadElementString("message");
        }
    }

    public class BufferMessage : Message
    {
        public const string TypeString = "buffer";
        public byte[] buffer;

        public BufferMessage()
        {
            type = TypeString;
            buffer = null;
        }
        public BufferMessage(XmlTextReader xreader)
        {
            type = TypeString;
            buffer = null;
            ReadBufferXml(xreader);
        }

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateBufferXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadBufferXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateBufferXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, BufferMessage.TypeString);
            /*
            xwriter.WriteStartElement("data");
            xwriter.WriteBase64(buffer, 0, buffer.Length);
            xwriter.WriteEndElement();
            */
            xwriter.WriteElementString("data", Base64String);
            FinishXmlWriter(xwriter);
        }

        public void ReadBufferXml(XmlTextReader xreader)
        {
            string str = xreader.ReadElementString("data");
            if (str == null || str == "")
                buffer = null;
            else
                buffer = Convert.FromBase64String(str);
        }

        public int Length
        {
            get
            {
                if (buffer != null)
                    return buffer.Length;
                return 0;
            }
        }

        public string Base64String
        {
            get
            {
                if (buffer != null)
                    return Convert.ToBase64String(buffer);
                return "";
            }
        }
    }

    public class PMIMessage : Message
    {
        public const string TypeString = "pmi";

        private string _pmi_type;
        private string _name;
        private Hashtable _data;

        public PMIMessage()
        {
            type = TypeString;
            _data = new Hashtable();
        }
        public PMIMessage(XmlTextReader xreader)
        {
            type = TypeString;
            _data = new Hashtable();
            ReadPMIXml(xreader);
        }

        #region getters and setters
        public string pmi_type
        {
            get { return _pmi_type; }
            set { _pmi_type = value; }
        }
        public string name
        {
            get { return _name; }
            set { _name = value; }
        }
        public Hashtable data
        {
            get { return _data; }
            set { _data = value; }
        }
        #endregion

        #region Print to the console
        public void Print(Mpiexec mpiexec)
        {
            lock (typeof(PMIMessage))
            {
                mpiexec.WriteLine("PMI:");
                mpiexec.WriteLine(" pmi_type: {0}", pmi_type);
                mpiexec.WriteLine(" name: {0}", name);
                if (_pmi_type == "client")
                {
                    switch (_name)
                    {
                        case "init":
                            mpiexec.WriteLine(" key: {0}", (string)_data["key"]);
                            break;
                        case "finalize":
                            mpiexec.WriteLine(" id: {0}", (string)_data["id"]);
                            mpiexec.WriteLine(" rank: {0}", (string)_data["rank"]);
                            break;
                        case "publish":
                            mpiexec.WriteLine(" name: {0}", (string)_data["name"]);
                            mpiexec.WriteLine(" port: {0}", (string)_data["port"]);
                            break;
                        case "unpublish":
                            mpiexec.WriteLine(" name: {0}", (string)_data["name"]);
                            break;
                        case "lookup":
                            mpiexec.WriteLine(" name: {0}", (string)_data["name"]);
                            break;
                        case "barrier":
                            mpiexec.WriteLine(" id: {0}", (string)_data["id"]);
                            break;
                        case "abort":
                            mpiexec.WriteLine(" id: {0}", (string)_data["id"]);
                            mpiexec.WriteLine(" rank: {0}", (string)_data["rank"]);
                            mpiexec.WriteLine(" message: {0}", (string)_data["message"]);
                            break;
                        case "destroy":
                            mpiexec.WriteLine(" name: {0}", (string)_data["name"]);
                            break;
                        case "put":
                            mpiexec.WriteLine(" kvs: {0}", (string)_data["kvs"]);
                            mpiexec.WriteLine(" key: {0}", (string)_data["key"]);
                            mpiexec.WriteLine(" value: {0}", (string)_data["value"]);
                            break;
                        case "commit":
                            mpiexec.WriteLine(" kvs: {0}", (string)_data["kvs"]);
                            break;
                        case "get":
                            mpiexec.WriteLine(" kvs: {0}", (string)_data["kvs"]);
                            mpiexec.WriteLine(" key: {0}", (string)_data["key"]);
                            break;
                        case "iter_first":
                            mpiexec.WriteLine(" kvs: {0}", (string)_data["kvs"]);
                            break;
                        case "iter":
                            mpiexec.WriteLine(" kvs: {0}", (string)_data["kvs"]);
                            break;
                        case "spawn":
                            int count;
                            count = (int)_data["preput_count"];
                            mpiexec.WriteLine(" preput_count: {0}", count);
                            for (int i = 0; i < count; i++)
                            {
                                string key_str = String.Format("preput_key{0}", i);
                                string value_str = String.Format("preput_value{0}", i);
                                mpiexec.WriteLine(" {0}: {1}", key_str, (string)_data[key_str]);
                                mpiexec.WriteLine(" {0}: {1}", value_str, (string)_data[value_str]);
                            }
                            count = (int)_data["commands_count"];
                            mpiexec.WriteLine(" count: {0}", count);
                            for (int i = 0; i < count; i++)
                            {
                                string max_proc_str = String.Format("maxproc{0}", i);
                                mpiexec.WriteLine(" {0}: {1}", max_proc_str, (int)_data[max_proc_str]);
                                mpiexec.WriteLine(" exe{0}: {1}", i, (string)_data[String.Format("exe{0}", i)]);
                                int args_count = (int)_data[String.Format("args{0}_count", i)];
                                mpiexec.WriteLine(" args{0}_count: {1}", i, args_count);
                                for (int j = 0; j < args_count; j++)
                                {
                                    mpiexec.WriteLine(" args{0}_arg{1}: {2}", i, j, (string)_data[String.Format("args{0}_arg{1}", i, j)]);
                                }
                                int keyvals_count = (int)_data[String.Format("keyvals{0}_count", i)];
                                mpiexec.WriteLine(" keyvals{0}_count: {1}", i, keyvals_count);
                                for (int j = 0; j < keyvals_count; j++)
                                {
                                    mpiexec.WriteLine(" keyvals{0}_key{1}: {2}", i, j, (string)_data[String.Format("keyvals{0}_key{1}", i, j)]);
                                    mpiexec.WriteLine(" keyvals{0}_value{1}: {2}", i, j, (string)_data[String.Format("keyvals{0}_value{1}", i, j)]);
                                }
                            }
                            break;
                    }
                }
                else
                {
                    switch (_name)
                    {
                        case "init":
                            mpiexec.WriteLine(" spawned: {0}", (string)_data["spawned"]);
                            mpiexec.WriteLine(" id: {0}", (string)_data["id"]);
                            mpiexec.WriteLine(" domain: {0}", (string)_data["domain"]);
                            mpiexec.WriteLine(" rank: {0}", (string)_data["rank"]);
                            mpiexec.WriteLine(" size: {0}", (string)_data["size"]);
                            mpiexec.WriteLine(" appnum: {0}", (string)_data["appnum"]);
                            break;
                        case "initialized":
                            mpiexec.WriteLine(" initialized: {0}", (string)_data["initialized"]);
                            break;
                        case "size":
                            mpiexec.WriteLine(" size: {0}", (string)_data["size"]);
                            break;
                        case "rank":
                            mpiexec.WriteLine(" rank: {0}", (string)_data["rank"]);
                            break;
                        case "universe":
                            mpiexec.WriteLine(" universe: {0}", (string)_data["universe"]);
                            break;
                        case "appnum":
                            mpiexec.WriteLine(" appnum: {0}", (string)_data["appnum"]);
                            break;
                        case "lookup":
                            mpiexec.WriteLine(" port: {0}", (string)_data["port"]);
                            break;
                        case "id":
                            mpiexec.WriteLine(" id: {0}", (string)_data["id"]);
                            break;
                        case "domain":
                            mpiexec.WriteLine(" domain: {0}", (string)_data["domain"]);
                            break;
                        case "id_maxlen":
                            mpiexec.WriteLine(" id_maxlen: {0}", (string)_data["id_maxlen"]);
                            break;
                        case "clique_size":
                            mpiexec.WriteLine(" clique_size: {0}", (string)_data["clique_size"]);
                            break;
                        case "clique":
                            int num_ranks;
                            num_ranks = (int)_data["num_ranks"];
                            mpiexec.WriteLine(" num_ranks: {0}", num_ranks);
                            for (int i = 0; i < num_ranks; i++)
                            {
                                string rank_str = String.Format("rank{0}", i);
                                mpiexec.WriteLine(" {0}: {1}", rank_str, (string)_data[rank_str]);
                            }
                            break;
                        case "name":
                            mpiexec.WriteLine(" name: {0}", (string)_data["name"]);
                            break;
                        case "name_maxlen":
                            mpiexec.WriteLine(" name_maxlen: {0}", (string)_data["name_maxlen"]);
                            break;
                        case "key_maxlen":
                            mpiexec.WriteLine(" key_maxlen: {0}", (string)_data["key_maxlen"]);
                            break;
                        case "value_maxlen":
                            mpiexec.WriteLine(" value_maxlen: {0}", (string)_data["value_maxlen"]);
                            break;
                        case "created":
                            mpiexec.WriteLine(" name: {0}", (string)_data["name"]);
                            break;
                        case "get":
                            mpiexec.WriteLine(" value: {0}", (string)_data["value"]);
                            break;
                        case "iter_first":
                            mpiexec.WriteLine(" key: {0}", (string)_data["key"]);
                            mpiexec.WriteLine(" value: {0}", (string)_data["value"]);
                            break;
                        case "iter":
                            mpiexec.WriteLine(" key: {0}", (string)_data["key"]);
                            mpiexec.WriteLine(" value: {0}", (string)_data["value"]);
                            break;
                        case "result":
                            mpiexec.WriteLine(" success: {0}", (string)_data["success"]);
                            mpiexec.WriteLine(" message: {0}", (string)_data["message"]);
                            break;
                    }
                }
            }
        }

        public void PrintRaw(Mpiexec mpiexec)
        {
            lock (typeof(PMIMessage))
            {
                mpiexec.WriteLine("PMI:");
                mpiexec.WriteLine(" pmi_type: {0}", pmi_type);
                mpiexec.WriteLine(" name: {0}", name);
                foreach (string key in data.Keys)
                {
                    mpiexec.WriteLine(" {0}: {1}", key, data[key]);
                }
            }
        }
        #endregion

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreatePMIXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadPMIXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreatePMIXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, PMIMessage.TypeString);
            xwriter.WriteStartElement("pmi");
            xwriter.WriteElementString("type", _pmi_type);
            if (_pmi_type == "client")
            {
                xwriter.WriteStartElement("request");
                xwriter.WriteElementString("name", _name);
                switch (_name)
                {
                    case "init":
                        xwriter.WriteElementString("key", (string)_data["key"]);
                        xwriter.WriteElementString("active_host", (string)_data["host"]);
                        xwriter.WriteElementString("active_port", (string)_data["active_port"]);
                        break;
                    case "initialized":
                        break;
                    case "finalize":
                        xwriter.WriteElementString("id", (string)_data["id"]);
                        xwriter.WriteElementString("rank", (string)_data["rank"]);
                        break;
                    case "size":
                        break;
                    case "rank":
                        break;
                    case "universe":
                        break;
                    case "appnum":
                        break;
                    case "publish":
                        xwriter.WriteElementString("name", (string)_data["name"]);
                        xwriter.WriteElementString("port", (string)_data["port"]);
                        break;
                    case "unpublish":
                        xwriter.WriteElementString("name", (string)_data["name"]);
                        break;
                    case "lookup":
                        xwriter.WriteElementString("name", (string)_data["name"]);
                        break;
                    case "id":
                        break;
                    case "domain":
                        break;
                    case "id_maxlen":
                        break;
                    case "barrier":
                        xwriter.WriteElementString("id", (string)_data["id"]);
                        xwriter.WriteElementString("rank", (string)_data["rank"]);
                        break;
                    case "clique_size":
                        break;
                    case "clique":
                        break;
                    case "abort":
                        xwriter.WriteElementString("id", (string)_data["id"]);
                        xwriter.WriteElementString("rank", (string)_data["rank"]);
                        xwriter.WriteElementString("exit_code", (string)_data["exit code"]);
                        xwriter.WriteElementString("message", (string)_data["message"]);
                        break;
                    case "name":
                        break;
                    case "name_maxlen":
                        break;
                    case "key_maxlen":
                        break;
                    case "value_maxlen":
                        break;
                    case "create":
                        break;
                    case "destroy":
                        xwriter.WriteElementString("name", (string)_data["name"]);
                        break;
                    case "put":
                        xwriter.WriteElementString("kvs", (string)_data["kvs"]);
                        xwriter.WriteElementString("key", (string)_data["key"]);
                        xwriter.WriteElementString("value", (string)_data["value"]);
                        break;
                    case "commit":
                        xwriter.WriteElementString("kvs", (string)_data["kvs"]);
                        break;
                    case "get":
                        xwriter.WriteElementString("kvs", (string)_data["kvs"]);
                        xwriter.WriteElementString("key", (string)_data["key"]);
                        break;
                    case "iter_first":
                        xwriter.WriteElementString("kvs", (string)_data["kvs"]);
                        break;
                    case "iter":
                        xwriter.WriteElementString("kvs", (string)_data["kvs"]);
                        break;
                    case "spawn":
                        //PrintRaw();
                        int count;
                        count = (int)_data["preput_count"];
                        xwriter.WriteStartElement("preput");
                        xwriter.WriteElementString("count", XmlConvert.ToString(count));
                        for (int i = 0; i < count; i++)
                        {
                            xwriter.WriteStartElement("keyval");
                            xwriter.WriteElementString("key", (string)_data[String.Format("preput_key{0}", i)]);
                            xwriter.WriteElementString("value", (string)_data[String.Format("preput_value{0}", i)]);
                            xwriter.WriteEndElement();
                        }
                        xwriter.WriteEndElement(); // preput
                        xwriter.WriteStartElement("commands");
                        count = (int)_data["commands_count"];
                        xwriter.WriteElementString("count", XmlConvert.ToString(count));
                        for (int i = 0; i < count; i++)
                        {
                            xwriter.WriteStartElement("command");
                            xwriter.WriteElementString("maxproc", XmlConvert.ToString((int)_data[String.Format("maxproc{0}", i)]));
                            xwriter.WriteElementString("exe", (string)_data[String.Format("exe{0}", i)]);
                            xwriter.WriteStartElement("args");
                            int args_count = (int)_data[String.Format("args{0}_count", i)];
                            xwriter.WriteElementString("count", XmlConvert.ToString(args_count));
                            for (int j = 0; j < args_count; j++)
                            {
                                xwriter.WriteElementString("arg", (string)_data[String.Format("args{0}_arg{1}", i, j)]);
                            }
                            xwriter.WriteEndElement(); // args
                            xwriter.WriteStartElement("keyvals");
                            int keyvals_count = (int)_data[String.Format("keyvals{0}_count", i)];
                            xwriter.WriteElementString("count", XmlConvert.ToString(keyvals_count));
                            for (int j = 0; j < keyvals_count; j++)
                            {
                                xwriter.WriteStartElement("keyval");
                                xwriter.WriteElementString("key", (string)_data[String.Format("keyvals{0}_key{1}", i, j)]);
                                xwriter.WriteElementString("value", (string)_data[String.Format("keyvals{0}_value{1}", i, j)]);
                                xwriter.WriteEndElement(); // keyval
                            }
                            xwriter.WriteEndElement(); // keyvals
                            xwriter.WriteEndElement(); // command
                        }
                        xwriter.WriteEndElement(); // commands
                        xwriter.WriteElementString("quiet", XmlConvert.ToString((bool)_data["quiet"]));
                        break;
                }
                xwriter.WriteEndElement(); // request
            }
            else
            {
                xwriter.WriteStartElement("result");
                xwriter.WriteElementString("name", _name);
                xwriter.WriteElementString("success", (string)data["success"]);
                if (_data["message"] != null)
                    xwriter.WriteElementString("message", (string)_data["message"]);
                else
                    xwriter.WriteElementString("message", "");
                switch (_name)
                {
                    case "init":
                        xwriter.WriteElementString("spawned", (string)_data["spawned"]);
                        xwriter.WriteElementString("id", (string)_data["id"]);
                        xwriter.WriteElementString("domain", (string)_data["domain"]);
                        xwriter.WriteElementString("rank", (string)_data["rank"]);
                        xwriter.WriteElementString("size", (string)_data["size"]);
                        xwriter.WriteElementString("appnum", (string)_data["appnum"]);
                        break;
                    case "initialized":
                        xwriter.WriteElementString("initialized", (string)_data["initialized"]);
                        break;
                    case "finalize":
                        break;
                    case "size":
                        xwriter.WriteElementString("size", (string)_data["size"]);
                        break;
                    case "rank":
                        xwriter.WriteElementString("rank", (string)_data["rank"]);
                        break;
                    case "universe":
                        xwriter.WriteElementString("universe", (string)_data["universe"]);
                        break;
                    case "appnum":
                        xwriter.WriteElementString("appnum", (string)_data["appnum"]);
                        break;
                    case "publish":
                        break;
                    case "unpublish":
                        break;
                    case "lookup":
                        xwriter.WriteElementString("port", (string)_data["port"]);
                        break;
                    case "id":
                        xwriter.WriteElementString("id", (string)_data["id"]);
                        break;
                    case "domain":
                        xwriter.WriteElementString("domain", (string)_data["domain"]);
                        break;
                    case "id_maxlen":
                        xwriter.WriteElementString("id_maxlen", (string)_data["id_maxlen"]);
                        break;
                    case "barrier":
                        break;
                    case "clique_size":
                        xwriter.WriteElementString("clique_size", (string)_data["clique_size"]);
                        break;
                    case "clique":
                        int num_ranks;
                        num_ranks = (int)_data["num_ranks"];
                        xwriter.WriteElementString("num_ranks", XmlConvert.ToString(num_ranks));
                        for (int i = 0; i < num_ranks; i++)
                        {
                            string rank_str = String.Format("rank{0}", i);
                            xwriter.WriteElementString(rank_str, (string)_data[rank_str]);
                        }
                        break;
                    case "abort":
                        break;
                    case "name":
                        xwriter.WriteElementString("name", (string)_data["name"]);
                        break;
                    case "name_maxlen":
                        xwriter.WriteElementString("name_maxlen", (string)_data["name_maxlen"]);
                        break;
                    case "key_maxlen":
                        xwriter.WriteElementString("key_maxlen", (string)_data["key_maxlen"]);
                        break;
                    case "value_maxlen":
                        xwriter.WriteElementString("value_maxlen", (string)_data["value_maxlen"]);
                        break;
                    case "created":
                        xwriter.WriteElementString("name", (string)_data["name"]);
                        break;
                    case "destroy":
                        break;
                    case "put":
                        break;
                    case "commit":
                        break;
                    case "get":
                        xwriter.WriteElementString("value", (string)_data["value"]);
                        break;
                    case "iter_first":
                        xwriter.WriteElementString("key", (string)_data["key"]);
                        xwriter.WriteElementString("value", (string)_data["value"]);
                        break;
                    case "iter":
                        xwriter.WriteElementString("key", (string)_data["key"]);
                        xwriter.WriteElementString("value", (string)_data["value"]);
                        break;
                    case "spawn":
                        break;
                    case "result":
                        xwriter.WriteElementString("success", (string)_data["success"]);
                        xwriter.WriteElementString("message", (string)_data["message"]);
                        break;
                }
                xwriter.WriteEndElement(); // result
            }
            xwriter.WriteEndElement(); // pmi
            FinishXmlWriter(xwriter);
        }

        public void ReadPMIXml(XmlTextReader xreader)
        {
            _data.Clear();
            xreader.ReadStartElement("pmi");
            _pmi_type = xreader.ReadElementString("type");
            if (_pmi_type == "client")
            {
                xreader.ReadStartElement("request");
                _name = xreader.ReadElementString("name");
                switch (_name)
                {
                    case "init":
                        _data["key"] = xreader.ReadElementString("key");
                        _data["active_host"] = xreader.ReadElementString("active_host");
                        _data["active_port"] = xreader.ReadElementString("active_port");
                        break;
                    case "finalize":
                        _data["id"] = xreader.ReadElementString("id");
                        _data["rank"] = xreader.ReadElementString("rank");
                        break;
                    case "publish":
                        _data["name"] = xreader.ReadElementString("name");
                        _data["port"] = xreader.ReadElementString("port");
                        break;
                    case "unpublish":
                        _data["name"] = xreader.ReadElementString("name");
                        break;
                    case "lookup":
                        _data["name"] = xreader.ReadElementString("name");
                        break;
                    case "barrier":
                        _data["id"] = xreader.ReadElementString("id");
                        _data["rank"] = xreader.ReadElementString("rank");
                        break;
                    case "abort":
                        _data["id"] = xreader.ReadElementString("id");
                        _data["rank"] = xreader.ReadElementString("rank");
                        _data["exit code"] = xreader.ReadElementString("exit_code");
                        _data["message"] = xreader.ReadElementString("message");
                        break;
                    case "destroy":
                        _data["name"] = xreader.ReadElementString("name");
                        break;
                    case "put":
                        _data["kvs"] = xreader.ReadElementString("kvs");
                        _data["key"] = xreader.ReadElementString("key");
                        _data["value"] = xreader.ReadElementString("value");
                        break;
                    case "commit":
                        _data["kvs"] = xreader.ReadElementString("kvs");
                        break;
                    case "get":
                        _data["kvs"] = xreader.ReadElementString("kvs");
                        _data["key"] = xreader.ReadElementString("key");
                        break;
                    case "iter_first":
                        _data["kvs"] = xreader.ReadElementString("kvs");
                        break;
                    case "iter":
                        _data["kvs"] = xreader.ReadElementString("kvs");
                        break;
                    case "spawn":
                        int count;
                        xreader.ReadStartElement("preput");
                        count = XmlConvert.ToInt32(xreader.ReadElementString("count"));
                        _data["preput_count"] = count;
                        for (int i = 0; i < count; i++)
                        {
                            xreader.ReadStartElement("keyval");
                            _data[String.Format("preput_key{0}", i)] = xreader.ReadElementString("key");
                            _data[String.Format("preput_value{0}", i)] = xreader.ReadElementString("value");
                            xreader.ReadEndElement(); // keyval
                        }
                        xreader.ReadEndElement(); // preput
                        xreader.ReadStartElement("commands");
                        count = XmlConvert.ToInt32(xreader.ReadElementString("count"));
                        _data["commands_count"] = count;
                        for (int i = 0; i < count; i++)
                        {
                            xreader.ReadStartElement("command");
                            _data[String.Format("maxproc{0}", i)] = XmlConvert.ToInt32(xreader.ReadElementString("maxproc"));
                            _data[String.Format("exe{0}", i)] = xreader.ReadElementString("exe");
                            xreader.ReadStartElement("args");
                            int args_count;
                            args_count = XmlConvert.ToInt32(xreader.ReadElementString("count"));
                            _data[String.Format("args{0}_count", i)] = args_count;
                            for (int j = 0; j < args_count; j++)
                            {
                                _data[String.Format("args{0}_arg{1}", i, j)] = xreader.ReadElementString("arg");
                            }
                            xreader.ReadEndElement(); // args
                            xreader.ReadStartElement("keyvals");
                            int keyvals_count;
                            keyvals_count = XmlConvert.ToInt32(xreader.ReadElementString("count"));
                            _data[String.Format("keyvals{0}_count", i)] = keyvals_count;
                            for (int j = 0; j < keyvals_count; j++)
                            {
                                xreader.ReadStartElement("keyval");
                                _data[String.Format("keyvals{0}_key{1}", i, j)] = xreader.ReadElementString("key");
                                _data[String.Format("keyvals{0}_value{1}", i, j)] = xreader.ReadElementString("value");
                                xreader.ReadEndElement(); // keyval
                            }
                            xreader.ReadEndElement(); // keyvals
                            xreader.ReadEndElement(); // command
                        }
                        xreader.ReadEndElement(); // commands
                        _data["quiet"] = XmlConvert.ToBoolean(xreader.ReadElementString("quiet"));
                        break;
                }
                xreader.ReadEndElement(); // request
            }
            else
            {
                xreader.ReadStartElement("result");
                _name = xreader.ReadElementString("name");
                _data["success"] = xreader.ReadElementString("success");
                _data["message"] = xreader.ReadElementString("message");
                switch (_name)
                {
                    case "init":
                        _data["spawned"] = xreader.ReadElementString("spawned");
                        _data["id"] = xreader.ReadElementString("id");
                        _data["domain"] = xreader.ReadElementString("domain");
                        _data["rank"] = xreader.ReadElementString("rank");
                        _data["size"] = xreader.ReadElementString("size");
                        _data["appnum"] = xreader.ReadElementString("appnum");
                        break;
                    case "initialized":
                        _data["initialized"] = xreader.ReadElementString("initialized");
                        break;
                    case "size":
                        _data["size"] = xreader.ReadElementString("size");
                        break;
                    case "rank":
                        _data["rank"] = xreader.ReadElementString("rank");
                        break;
                    case "universe":
                        _data["universe"] = xreader.ReadElementString("universe");
                        break;
                    case "appnum":
                        _data["appnum"] = xreader.ReadElementString("appnum");
                        break;
                    case "lookup":
                        _data["port"] = xreader.ReadElementString("port");
                        break;
                    case "id":
                        _data["id"] = xreader.ReadElementString("id");
                        break;
                    case "domain":
                        _data["domain"] = xreader.ReadElementString("domain");
                        break;
                    case "id_maxlen":
                        _data["id_maxlen"] = xreader.ReadElementString("id_maxlen");
                        break;
                    case "clique_size":
                        _data["clique_size"] = xreader.ReadElementString("clique_size");
                        break;
                    case "clique":
                        int num_ranks;
                        num_ranks = XmlConvert.ToInt32(xreader.ReadElementString("num_ranks"));
                        _data["num_ranks"] = num_ranks;
                        for (int i = 0; i < num_ranks; i++)
                        {
                            string rank_str = String.Format("rank{0}", i);
                            _data[rank_str] = xreader.ReadElementString(rank_str);
                        }
                        break;
                    case "name":
                        _data["name"] = xreader.ReadElementString("name");
                        break;
                    case "name_maxlen":
                        _data["name_maxlen"] = xreader.ReadElementString("name_maxlen");
                        break;
                    case "key_maxlen":
                        _data["key_maxlen"] = xreader.ReadElementString("key_maxlen");
                        break;
                    case "value_maxlen":
                        _data["value_maxlen"] = xreader.ReadElementString("value_maxlen");
                        break;
                    case "created":
                        _data["name"] = xreader.ReadElementString("name");
                        break;
                    case "get":
                        _data["value"] = xreader.ReadElementString("value");
                        break;
                    case "iter_first":
                        _data["key"] = xreader.ReadElementString("key");
                        _data["value"] = xreader.ReadElementString("value");
                        break;
                    case "iter":
                        _data["key"] = xreader.ReadElementString("key");
                        _data["value"] = xreader.ReadElementString("value");
                        break;
                    case "result":
                        _data["success"] = xreader.ReadElementString("success");
                        _data["message"] = xreader.ReadElementString("message");
                        break;
                }
                xreader.ReadEndElement(); // result
            }
            xreader.ReadEndElement(); // pmi
        }
    }

    public class ActiveMessage : Message
    {
        public const string TypeString = "active";

        private string _active_type;
        private string _name;
        private Hashtable _data;

        public ActiveMessage()
        {
            type = TypeString;
            _data = new Hashtable();
        }
        public ActiveMessage(XmlTextReader xreader)
        {
            type = TypeString;
            _data = new Hashtable();
            ReadActiveXml(xreader);
        }

        #region getters and setters
        public string active_type
        {
            get { return _active_type; }
            set { _active_type = value; }
        }
        public string name
        {
            get { return _name; }
            set { _name = value; }
        }
        public Hashtable data
        {
            get { return _data; }
            set { _data = value; }
        }
        #endregion

        #region Print to the console
        public void Print(Mpiexec mpiexec)
        {
            lock (typeof(ActiveMessage))
            {
                mpiexec.WriteLine("Active:");
                mpiexec.WriteLine(" active_type: {0}", active_type);
                mpiexec.WriteLine(" name: {0}", name);
                if (_active_type == "client")
                {
                    switch (_name)
                    {
                        case "abort_callback":
                            mpiexec.WriteLine(" exit_code: {0}", (string)_data["exit_code"]);
                            break;
                        case "queues":
                            break;
                        case "history":
                            break;
                        case "exit":
                            mpiexec.WriteLine(" exit_code: {0}", (string)_data["exit_code"]);
                            break;
                    }
                }
                else
                {
                    switch (_name)
                    {
                        case "abort_callback":
                            break;
                        case "queues":
                            mpiexec.WriteLine(" data: {0}", (string)_data["data"]);
                            break;
                        case "history":
                            mpiexec.WriteLine(" data: {0}", (string)_data["data"]);
                            break;
                        case "exit":
                            break;
                    }
                }
            }
        }

        public void PrintRaw(Mpiexec mpiexec)
        {
            lock (typeof(ActiveMessage))
            {
                mpiexec.WriteLine("Active:");
                mpiexec.WriteLine(" active_type: {0}", active_type);
                mpiexec.WriteLine(" name: {0}", name);
                foreach (string key in data.Keys)
                {
                    mpiexec.WriteLine(" {0}: {1}", key, data[key]);
                }
            }
        }
        #endregion

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateActiveXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadActiveXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateActiveXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, ActiveMessage.TypeString);
            xwriter.WriteStartElement("active");
            xwriter.WriteElementString("type", _active_type);
            if (_active_type == "client")
            {
                xwriter.WriteStartElement("request");
                xwriter.WriteElementString("name", _name);
                switch (_name)
                {
                    case "abort_callback":
                        xwriter.WriteElementString("exit_code", (string)_data["exit_code"]);
                        break;
                    case "queues":
                        break;
                    case "history":
                        break;
                    case "exit":
                        xwriter.WriteElementString("exit_code", (string)_data["exit_code"]);
                        break;
                }
                xwriter.WriteEndElement(); // request
            }
            else
            {
                xwriter.WriteStartElement("result");
                xwriter.WriteElementString("name", _name);
                xwriter.WriteElementString("success", (string)data["success"]);
                switch (_name)
                {
                    case "abort_callback":
                        break;
                    case "queues":
                        xwriter.WriteElementString("data", (string)_data["data"]);
                        break;
                    case "history":
                        xwriter.WriteElementString("data", (string)_data["data"]);
                        break;
                    case "exit":
                        // there should be no result of an exit command since the purpose of the command is to cause the process to exit immediately
                        break;
                }
                xwriter.WriteEndElement(); // result
            }
            xwriter.WriteEndElement(); // active
            FinishXmlWriter(xwriter);
        }

        public void ReadActiveXml(XmlTextReader xreader)
        {
            _data.Clear();
            xreader.ReadStartElement("active");
            _active_type = xreader.ReadElementString("type");
            if (_active_type == "client")
            {
                xreader.ReadStartElement("request");
                _name = xreader.ReadElementString("name");
                switch (_name)
                {
                    case "abort_callback":
                        _data["exit_code"] = xreader.ReadElementString("exit_code");
                        break;
                    case "queues":
                        break;
                    case "history":
                        break;
                    case "exit":
                        _data["exit_code"] = xreader.ReadElementString("exit_code");
                        break;
                }
                xreader.ReadEndElement(); // request
            }
            else
            {
                xreader.ReadStartElement("result");
                _name = xreader.ReadElementString("name");
                _data["success"] = xreader.ReadElementString("success");
                switch (_name)
                {
                    case "abort_callback":
                        break;
                    case "queues":
                        _data["data"] = xreader.ReadElementString("data");
                        break;
                    case "history":
                        _data["data"] = xreader.ReadElementString("data");
                        break;
                    case "exit":
                        // there should be no result of an exit command since the purpose of the command is to cause the process to exit immediately
                        break;
                }
                xreader.ReadEndElement(); // result
            }
            xreader.ReadEndElement(); // active
        }
    }

    public class HostInfoMessage : Message
    {
        public const string TypeString = "host_info";

        private HostInfo _info;

        public HostInfoMessage()
        {
            type = TypeString;
            _info = new HostInfo();
        }
        public HostInfoMessage(HostInfo info)
        {
            type = TypeString;
            _info = new HostInfo(info);
        }
        public HostInfoMessage(XmlTextReader xreader)
        {
            type = TypeString;
            _info = new HostInfo();
            ReadHostInfoXml(xreader);
        }

        #region getters and setters
        public HostInfo Info
        {
            get { return _info; }
            set { _info = value; }
        }
        #endregion

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateHostInfoXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadHostInfoXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateHostInfoXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, HostInfoMessage.TypeString);
            _info.Write(xwriter);
            FinishXmlWriter(xwriter);
        }

        public void ReadHostInfoXml(XmlTextReader xreader)
        {
            _info.Read(xreader);
        }
    }

    public class LaunchInfoMessage : Message
    {
        public const string TypeString = "launch_info";
        private string _good_exe_path, _bad_exe_path;
        private string _good_wdir, _bad_wdir;
        private string _missing_dlls, _missing_delayed_dlls;

        public LaunchInfoMessage()
        {
            type = TypeString;
        }
        public LaunchInfoMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadLaunchInfoXml(xreader);
        }

        #region getters and setters
        public string good_exe_path
        {
            get { return _good_exe_path; }
            set { _good_exe_path = value; }
        }
        public string bad_exe_path
        {
            get { return _bad_exe_path; }
            set { _bad_exe_path = value; }
        }
        public string good_wdir
        {
            get { return _good_wdir; }
            set { _good_wdir = value; }
        }
        public string bad_wdir
        {
            get { return _bad_wdir; }
            set { _bad_wdir = value; }
        }
        public string missing_dlls
        {
            get { return _missing_dlls; }
            set { _missing_dlls = value; }
        }
        public string missing_delayed_dlls
        {
            get { return _missing_delayed_dlls; }
            set { _missing_delayed_dlls = value; }
        }
        #endregion

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateLaunchInfoXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadLaunchInfoXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateLaunchInfoXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, LaunchInfoMessage.TypeString);
            xwriter.WriteElementString("good_exe", _good_exe_path);
            xwriter.WriteElementString("bad_exe", _bad_exe_path);
            xwriter.WriteElementString("good_wdir", _good_wdir);
            xwriter.WriteElementString("bad_wdir", _bad_wdir);
            xwriter.WriteElementString("missing_dlls", _missing_dlls);
            xwriter.WriteElementString("missing_delayed_dlls", _missing_delayed_dlls);
            FinishXmlWriter(xwriter);
        }

        public void ReadLaunchInfoXml(XmlTextReader xreader)
        {
            _good_exe_path = xreader.ReadElementString("good_exe");
            _bad_exe_path = xreader.ReadElementString("bad_exe");
            _good_wdir = xreader.ReadElementString("good_wdir");
            _bad_wdir = xreader.ReadElementString("bad_wdir");
            _missing_dlls = xreader.ReadElementString("missing_dlls");
            _missing_delayed_dlls = xreader.ReadElementString("missing_delayed_dlls");
        }
    }

    public class InfoRequestMessage : Message
    {
        public const string TypeString = "info";

        private string _request;

        public InfoRequestMessage()
        {
            type = TypeString;
        }
        public InfoRequestMessage(string request)
        {
            type = TypeString;
            _request = request;
        }
        public InfoRequestMessage(XmlTextReader xreader)
        {
            type = TypeString;
            ReadInfoRequestXml(xreader);
        }

        #region getters and setters
        public string request
        {
            get { return _request; }
            set { _request = value; }
        }
        #endregion

        public override byte[] CreateXmlBuffer()
        {
            MemoryStream mem_stream = new MemoryStream();
            XmlTextWriter xwriter = new XmlTextWriter(mem_stream, Encoding.UTF8);
            CreateInfoRequestXml(xwriter);
            xwriter.Close();
            return mem_stream.GetBuffer();
        }

        public new void FromXmlBuffer(byte[] buffer)
        {
            MemoryStream mem_stream = new MemoryStream(buffer);
            XmlTextReader xreader = new XmlTextReader(mem_stream);
            StartXmlReader(xreader);
            ReadInfoRequestXml(xreader);
            FinishXmlReader(xreader);
            xreader.Close();
        }

        private void CreateInfoRequestXml(XmlTextWriter xwriter)
        {
            StartXmlWriter(xwriter, InfoRequestMessage.TypeString);
            xwriter.WriteElementString("request", _request);
            FinishXmlWriter(xwriter);
        }

        public void ReadInfoRequestXml(XmlTextReader xreader)
        {
            _request = xreader.ReadElementString("request");
        }
    }
    #endregion // Message classes
}
