using System;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.IO;
using System.Text;
using Microsoft.Win32;
using System.Xml;
using System.Security.Cryptography;
using DeinoMPI;
using System.Diagnostics;
using System.ComponentModel;

namespace DeinoMPI
{
    #region Data classes
    public class AbortWatcherArg
    {
        private ClientSession _client;
        private ManualResetEvent _stop_event;
        private int _rank;

        public AbortWatcherArg()
        {
            _client = null;
            _stop_event = new ManualResetEvent(false);
            _rank = -1;
        }
        public AbortWatcherArg(ClientSession client)
        {
            _client = client;
            _stop_event = new ManualResetEvent(false);
            _rank = -1;
        }
        #region getters and setters
        public ClientSession client
        {
            get { return _client; }
            set { _client = value; }
        }
        public ManualResetEvent stop_event
        {
            get { return _stop_event; }
            set { _stop_event = value; }
        }
        public int Rank
        {
            get { return _rank; }
            set { _rank = value; }
        }
        #endregion
    }

    public class PMIClientArg
    {
        private volatile bool _connected;
        private TcpListener _listener;
        private PMI _pmi;

        #region getters and setters
        public bool connected
        {
            get { return _connected; }
            set { _connected = value; }
        }
        public TcpListener listener
        {
            get { return _listener; }
            set { _listener = value; }
        }
        public PMI pmi
        {
            get { return _pmi; }
            set { _pmi = value; }
        }
        #endregion
    }

    public class input_thread_arg
    {
        private ClientSession _client;
        private TextReader _stream;
        private InputOutputMessage.IO_Type _io_type;
        private volatile bool _stop;

        #region getters and setters
        public ClientSession client
        {
            get { return _client; }
            set { _client = value; }
        }
        public TextReader stream
        {
            get { return _stream; }
            set { _stream = value; }
        }
        public InputOutputMessage.IO_Type io_type
        {
            get { return _io_type; }
            set { _io_type = value; }
        }
        public bool stop
        {
            get { return _stop; }
            set { _stop = value; }
        }
        #endregion
    }

    public class HostNode
    {
        private string _hostname, _alt_hostname;
        private int _nproc;

        public HostNode()
        {
        }
        public HostNode(string hostname)
        {
            if (hostname != null)
            {
                string[] tokens = hostname.Split(':');
                if (tokens.Length > 1)
                {
                    if (tokens.Length > 2)
                        throw new Exception(String.Format("Invalid host specification '{0}'", hostname));
                    try
                    {
                        this.nproc = Convert.ToInt32(tokens[1]);
                        if (this._nproc < 1)
                        {
                            throw new Exception("Invalid processor field in host name '" + hostname + "'");
                        }
                    }
                    catch (Exception e)
                    {
                        throw new Exception(String.Format("Invalid host specification '{0}', {1}", hostname, e.Message));
                    }
                }
                else
                {
                    this._nproc = 1;
                }
                this._hostname = tokens[0];
            }
        }
        public HostNode(string hostname, int nproc)
        {
            this._hostname = hostname;
            this._nproc = nproc;
        }
        #region getters and setters
        public string hostname
        {
            get { return _hostname; }
            set { _hostname = value; }
        }
        public string alt_hostname
        {
            get { return _alt_hostname; }
            set { _alt_hostname = value; }
        }
        public int nproc
        {
            get { return _nproc; }
            set { _nproc = value; }
        }
        #endregion
    }

    public class LaunchNode
    {
        // stuff
        private Mpiexec _mpiexec;
        private PMI _pmi;
        private bool _spawned;
        private bool _launch_local;

        // connect information
        private string _hostname;
        private int _port;
        private string _phrase;
        private SecurityKeys _private_key;
        private SecureCredential _credential;
        private string _secure_user;

        // launch data
        private string _path;
        private string _exe;
        private ArrayList _args;
        private string _dir;
        private Hashtable _env;
        private Hashtable _map;
        private string _clique;
        private int _iproc;
        private int _nproc;
        private int _appnum;
        private int _priority_class;
        private int _priority_thread;

        private void Initialize()
        {
            iproc = 0;
            nproc = 1;
            appnum = 0;
            priority_class = 2;
            priority_thread = 3;
            args = new ArrayList();
            env = new Hashtable();
            map = new Hashtable();
            port = -1;
            secure_user = null;
            spawned = false;
            launch_local = false;
            mpiexec = null;
        }

        public LaunchNode()
        {
            Initialize();
        }

        public LaunchNode(Mpiexec mpiexec)
        {
            Initialize();
            this.mpiexec = mpiexec;
        }

        #region getters and setters
        public Mpiexec mpiexec
        {
            get { return _mpiexec; }
            set { _mpiexec = value; }
        }
        public PMI pmi
        {
            get { return _pmi; }
            set { _pmi = value; }
        }
        public string hostname
        {
            get { return _hostname; }
            set { _hostname = value; }
        }
        public int port
        {
            get { return _port; }
            set { _port = value; }
        }
        public string phrase
        {
            get { return _phrase; }
            set { _phrase = value; }
        }
        public string path
        {
            get { return _path; }
            set { _path = value; }
        }
        public string exe
        {
            get { return _exe; }
            set { _exe = value; }
        }
        public ArrayList args
        {
            get { return _args; }
            set { _args = value; }
        }
        public string dir
        {
            get { return _dir; }
            set { _dir = value; }
        }
        public Hashtable env
        {
            get { return _env; }
            set { _env = value; }
        }
        public Hashtable map
        {
            get { return _map; }
            set { _map = value; }
        }
        public string clique
        {
            get { return _clique; }
            set { _clique = value; }
        }
        public int iproc
        {
            get { return _iproc; }
            set { _iproc = value; }
        }
        public int nproc
        {
            get { return _nproc; }
            set { _nproc = value; }
        }
        public int appnum
        {
            get { return _appnum; }
            set { _appnum = value; }
        }
        public int priority_class
        {
            get { return _priority_class; }
            set { _priority_class = value; }
        }
        public int priority_thread
        {
            get { return _priority_thread; }
            set { _priority_thread = value; }
        }
        public SecurityKeys private_key
        {
            get { return _private_key; }
            set { _private_key = value; }
        }
        public SecureCredential credential
        {
            get { return _credential; }
            set { _credential = value; }
        }
        public string secure_user
        {
            get { return _secure_user; }
            set { _secure_user = value; }
        }
        public bool spawned
        {
            get { return _spawned; }
            set { _spawned = value; }
        }
        public bool launch_local
        {
            get { return _launch_local; }
            set { _launch_local = value; }
        }
        #endregion

        public LaunchNode Clone()
        {
            LaunchNode node = new LaunchNode();
            // Shallow copies of all the strings, arrays and hastables is ok because strings are immutable
            node.mpiexec = mpiexec;
            node.pmi = pmi;
            node.spawned = spawned;
            node.launch_local = launch_local;
            node.hostname = hostname;
            node.port = port;
            node.phrase = phrase;
            // WARNING: Shallow copies of the private_key and credential means that the cloned copy must treat them as read only
            // since they are shared between the original and the clone.
            node.private_key = private_key;
            node.credential = credential;
            node.secure_user = secure_user;
            node.path = path;
            node.exe = exe;
            if (args != null)
                node.args = (ArrayList)args.Clone();
            else
                node.args = null;
            node.dir = dir;
            if (env != null)
                node.env = (Hashtable)env.Clone();
            else
                node.env = null;
            if (map != null)
                node.map = (Hashtable)map.Clone();
            else
                node.map = null;
            node.clique = clique;
            node.iproc = iproc;
            node.nproc = nproc;
            node.appnum = appnum;
            node.priority_class = priority_class;
            node.priority_thread = priority_thread;
            return node;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder("Node:\r\n");
            if (hostname != null)
                sb.AppendFormat(" hostname: {0}\r\n", hostname);
            if (launch_local)
                sb.AppendFormat(" launch local: true\r\n");
            if (credential != null)
                sb.AppendFormat(" credential: {0}\r\n", credential.Account);
            if (secure_user != null)
                sb.AppendFormat(" secure user: {0}\r\n", secure_user);
            if (exe != null)
                sb.AppendFormat(" exe: {0}\r\n", exe);
            if (args != null)
            {
                sb.Append(" args: ");
                foreach (string str in args)
                {
                    sb.AppendFormat("{0}, ", str);
                }
                sb.Append("\r\n");
            }
            if (path != null)
                sb.AppendFormat(" path: {0}\r\n", path);
            if (dir != null)
                sb.AppendFormat(" dir: {0}\r\n", dir);
            if (env != null)
            {
                foreach (string key in env.Keys)
                {
                    sb.AppendFormat(" env: {0} = {1}\r\n", key, env[key]);
                }
            }
            if (map != null)
            {
                foreach (string drive in map.Keys)
                {
                    sb.AppendFormat(" map: {0}:{1}\r\n", drive, map[drive]);
                }
            }
            sb.AppendFormat(" rank: {0}\r\n", iproc);
            if (clique != null)
                sb.AppendFormat(" clique: {0}\r\n", clique);
            sb.AppendFormat(" size: {0}\r\n", nproc);
            sb.AppendFormat(" appnum: {0}\r\n", appnum);
            sb.AppendFormat(" priority: {0}:{1}\r\n", priority_class, priority_thread);
            if (pmi != null)
            {
                sb.AppendFormat(" PMI:\r\n");
                sb.AppendFormat("  domain: {0}\r\n", pmi.domain);
                sb.AppendFormat("  id: {0}\r\n", pmi.id);
                sb.AppendFormat("  nproc: {0}\r\n", pmi.nproc);
            }
            return sb.ToString();
        }

        public void Print()
        {
            lock (typeof(LaunchNode))
            {
                if (mpiexec != null)
                {
                    mpiexec.WriteLine("Node:");
                    if (hostname != null)
                        mpiexec.WriteLine(" hostname: {0}", hostname);
                    if (launch_local)
                        mpiexec.WriteLine(" launch local: true");
                    if (credential != null)
                        mpiexec.WriteLine(" credential: {0}", credential.Account);
                    if (secure_user != null)
                        mpiexec.WriteLine(" secure user: {0}", secure_user);
                    if (exe != null)
                        mpiexec.WriteLine(" exe: {0}", exe);
                    if (args != null)
                    {
                        mpiexec.Write(" args: ");
                        foreach (string str in args)
                        {
                            mpiexec.Write("{0}, ", str);
                        }
                        mpiexec.WriteLine();
                    }
                    if (path != null)
                        mpiexec.WriteLine(" path: {0}", path);
                    if (dir != null)
                        mpiexec.WriteLine(" dir: {0}", dir);
                    if (env != null)
                    {
                        foreach (string key in env.Keys)
                        {
                            mpiexec.WriteLine(" env: {0} = {1}", key, env[key]);
                        }
                    }
                    if (map != null)
                    {
                        foreach (string drive in map.Keys)
                        {
                            mpiexec.WriteLine(" map: {0}:{1}", drive, map[drive]);
                        }
                    }
                    mpiexec.WriteLine(" rank: {0}", iproc);
                    if (clique != null)
                        mpiexec.WriteLine(" clique: {0}", clique);
                    mpiexec.WriteLine(" size: {0}", nproc);
                    mpiexec.WriteLine(" appnum: {0}", appnum);
                    mpiexec.WriteLine(" priority: {0}:{1}", priority_class, priority_thread);
                    if (pmi != null)
                    {
                        mpiexec.WriteLine(" PMI:");
                        mpiexec.WriteLine("  domain: {0}", pmi.domain);
                        mpiexec.WriteLine("  id: {0}", pmi.id);
                        mpiexec.WriteLine("  nproc: {0}", pmi.nproc);
                    }
                }
                else
                {
                    Console.WriteLine(ToString());
                }
            }
        }
    }

    public class Options
    {
        private string _exe;
        private string _dir;
        private string _path;
        private ArrayList _args;
        private Hashtable _map;
        private Hashtable _env;
        private string _hostname;
        private int _nproc;
        private int _appnum;
        private int _priority_class;
        private int _priority_thread;
        private bool _localonly;
        private ArrayList _hosts;
        private int _cur_host;
        private int _cur_processor;
        private int _cur_rank;
        private bool _logon;
        private bool _impersonation;
        private bool _delegation;
        private string _username;
        private string _password;
        private bool _nompi;
        private bool _exitcodes;
        private bool _localroot;
        private bool _noprompt;
        private int _port;
        private string _phrase;
        private int _timeout;
        private string _channel;
        private string _secure_user;
        private bool _use_output_queue;
        private Queue _output_queue;
        private bool _validate;
        private string _stage_src, _stage_dest, _stage_wildcards;
        private bool _stage_nopull, _stage_recursive;

        private void Initialize()
        {
            nproc = -1;
            appnum = -1;
            priority_class = 2;
            priority_thread = 3;
            localonly = false;
            args = new ArrayList();
            hosts = new ArrayList();
            cur_host = -1;
            cur_processor = -1;
            cur_rank = 0;
            map = new Hashtable();
            env = new Hashtable();
            logon = false;
            impersonation = false;
            delegation = false;
            nompi = false;
            exitcodes = false;
            localroot = false;
            noprompt = false;
            port = -1;
            secure_user = null;
            _use_output_queue = false;
            _output_queue = Queue.Synchronized(new Queue());
            _timeout = 0;
            _validate = false;
            _stage_src = null;
            _stage_nopull = false;
            _stage_recursive = false;
        }

        public Options()
        {
            Initialize();
        }

        public void Clear()
        {
            Initialize();
        }

        #region getters and setters
        public string exe
        {
            get { return _exe; }
            set { _exe = value; }
        }
        public string dir
        {
            get { return _dir; }
            set { _dir = value; }
        }
        public string path
        {
            get { return _path; }
            set { _path = value; }
        }
        public ArrayList args
        {
            get { return _args; }
            set { _args = value; }
        }
        public Hashtable map
        {
            get { return _map; }
            set { _map = value; }
        }
        public Hashtable env
        {
            get { return _env; }
            set { _env = value; }
        }
        public string hostname
        {
            get { return _hostname; }
            set { _hostname = value; }
        }
        public int nproc
        {
            get { return _nproc; }
            set { _nproc = value; }
        }
        public int appnum
        {
            get { return _appnum; }
            set { _appnum = value; }
        }
        public int priority_class
        {
            get { return _priority_class; }
            set { _priority_class = value; }
        }
        public int priority_thread
        {
            get { return _priority_thread; }
            set { _priority_thread = value; }
        }
        public bool localonly
        {
            get { return _localonly; }
            set { _localonly = value; }
        }
        public ArrayList hosts
        {
            get { return _hosts; }
            set { _hosts = value; }
        }
        public int cur_host
        {
            get { return _cur_host; }
            set { _cur_host = value; }
        }
        public int cur_processor
        {
            get { return _cur_processor; }
            set { _cur_processor = value; }
        }
        public int cur_rank
        {
            get { return _cur_rank; }
            set { _cur_rank = value; }
        }
        public bool logon
        {
            get { return _logon; }
            set { _logon = value; }
        }
        public bool impersonation
        {
            get { return _impersonation; }
            set { _impersonation = value; }
        }
        public bool delegation
        {
            get { return _delegation; }
            set { _delegation = value; }
        }
        public string username
        {
            get { return _username; }
            set { _username = value; }
        }
        public string password
        {
            get { return _password; }
            set { _password = value; }
        }
        public bool nompi
        {
            get { return _nompi; }
            set { _nompi = value; }
        }
        public bool exitcodes
        {
            get { return _exitcodes; }
            set { _exitcodes = value; }
        }
        public bool localroot
        {
            get { return _localroot; }
            set { _localroot = value; }
        }
        public bool noprompt
        {
            get { return _noprompt; }
            set { _noprompt = value; }
        }
        public int port
        {
            get { return _port; }
            set { _port = value; }
        }
        public string phrase
        {
            get { return _phrase; }
            set { _phrase = value; }
        }
        public int timeout
        {
            get { return _timeout; }
            set { _timeout = value; }
        }
        public string channel
        {
            get { return _channel; }
            set { _channel = value; }
        }
        public string secure_user
        {
            get { return _secure_user; }
            set { _secure_user = value; }
        }
        public bool use_output_queue
        {
            get { return _use_output_queue; }
            set { _use_output_queue = value; }
        }
        public Queue output_queue
        {
            get { return _output_queue; }
        }
        public bool validate
        {
            get { return _validate; }
            set { _validate = value; }
        }
        public string stage_src
        {
            get { return _stage_src; }
            set { _stage_src = value; }
        }
        public string stage_dest
        {
            get { return _stage_dest; }
            set { _stage_dest = value; }
        }
        public string stage_wildcards
        {
            get { return _stage_wildcards; }
            set { _stage_wildcards = value; }
        }
        public bool stage_nopull
        {
            get { return _stage_nopull; }
            set { _stage_nopull = value; }
        }
        public bool stage_recursive
        {
            get { return _stage_recursive; }
            set { _stage_recursive = value; }
        }
        #endregion
    }
    #endregion

    public class ProcessGroup
    {
        private Mpiexec mpiexec;
        public int exit_code;
        ArrayList threads;
        public Queue thread_param_queue;
        Queue abort_thread_param_queue;
        Queue pmi_queue;
        byte[] input_buffer;
        public Thread input_thread;
        public PMI pmi;
        public int num_connecting;
        public bool connection_failed;
        public bool job_aborted;
        private bool job_abort_printed;
        public int job_aborted_exit_code;
        public string job_aborted_message;
        public int job_aborted_rank;
        ManualResetEvent connected_event;
        public ManualResetEvent abort_job_event;
        ExitInfo[] exit_info;
        public ManualResetEvent print_queues_event;
        public ManualResetEvent print_queues_exit_event;
        public ArrayList client_list;
        public int timeout;
        private Thread timeout_thread;
        private string job_timedout_message;
        private int num_queue_messages_outstanding;
        private Hashtable queue_messages;
        private int num_histroy_messages_outstanding;
        private Hashtable history_messages;
        private Mutex wait_for_exit_mutex;
        private bool already_waited;

        private class ExitInfo
        {
            public int rank;
            public string host;
            public int exit_code;
            public string message;
            public bool pmi_init_called;
            public bool pmi_finalize_called;
            public bool exited;
            public int error;
            public ManualResetEvent running_or_error_event;

            private void Initialize()
            {
                rank = -1;
                host = null;
                exit_code = -1;
                message = null;
                pmi_init_called = false;
                pmi_finalize_called = false;
                exited = false;
                error = 0;
                running_or_error_event = new ManualResetEvent(false);
            }

            public ExitInfo()
            {
                Initialize();
            }

            public ExitInfo(int rank, string host, int exit_code, string message)
            {
                Initialize();
                this.rank = rank;
                this.host = host;
                this.exit_code = exit_code;
                this.message = message;
            }
        }

        private void Initialize()
        {
            thread_param_queue = Queue.Synchronized(new Queue());
            abort_thread_param_queue = Queue.Synchronized(new Queue());
            pmi_queue = Queue.Synchronized(new Queue());
            input_buffer = new byte[4096];
            input_thread = null;
            pmi = new PMI();
            connection_failed = false;
            job_aborted = false;
            job_abort_printed = false;
            job_aborted_exit_code = -1;
            job_aborted_message = null;
            job_aborted_rank = -1;
            connected_event = new ManualResetEvent(false);
            abort_job_event = new ManualResetEvent(false);
            exit_info = null;
            print_queues_event = new ManualResetEvent(false);
            print_queues_exit_event = new ManualResetEvent(false);
            client_list = new ArrayList();
            timeout = 0;
            timeout_thread = null;
            job_timedout_message = null;
            num_queue_messages_outstanding = 0;
            queue_messages = new Hashtable();
            num_histroy_messages_outstanding = 0;
            history_messages = new Hashtable();
            mpiexec = null;
            wait_for_exit_mutex = new Mutex();
            already_waited = false;
        }

        public ProcessGroup()
        {
            Initialize();
        }

        public ProcessGroup(Mpiexec mpiexec)
        {
            Initialize();
            if (mpiexec != null && mpiexec.global_options != null)
                timeout = mpiexec.global_options.timeout;
            this.mpiexec = mpiexec;
            pmi.mpiexec = mpiexec;
            if (mpiexec != null)
                mpiexec.process_groups.Add(this);
        }

        public bool QuietAbort
        {
            // The logic here is that if the abort messages have been printed then any subsequent aborts will be quiet.
            // If you set job_abort_printed to true then nothing will be printed when an abort really happens.
            get { return job_abort_printed; }
            set { job_abort_printed = value; }
        }

        int CountClique(string clique)
        {
            string[] ranks = ParseClique(clique);
            return ranks.Length;
        }

        static public string[] ParseClique(string clique)
        {
            StringBuilder sb = new StringBuilder();
            int cur_rank, i, j, top;
            i = 0;
            cur_rank = -1;
            while (i < clique.Length)
            {
                if (clique[i] == '(')
                {
                    i++;
                    if (i == clique.Length)
                        break;
                }
                if (clique[i] == '.')
                {
                    i++;
                    if (i == clique.Length)
                        break;
                }
                if (clique[i] == '.' || clique[i] == '-')
                {
                    i++;
                    if (i == clique.Length)
                        break;
                    j = i;
                    while (Char.IsNumber(clique[j]))
                    {
                        j++;
                        if (j == clique.Length)
                            break;
                    }
                    top = Convert.ToInt32(clique.Substring(i, j - i));
                    i = j;
                    for (j = cur_rank; j <= top; j++)
                    {
                        sb.AppendFormat("{0} ", j);
                    }
                    cur_rank = -1;
                    continue;
                }
                if (clique[i] == ',')
                {
                    i++;
                    if (i == clique.Length)
                        break;
                    j = i;
                    while (Char.IsNumber(clique[j]))
                    {
                        j++;
                        if (j == clique.Length)
                            break;
                    }
                    cur_rank = Convert.ToInt32(clique.Substring(i, j - i));
                    sb.AppendFormat("{0} ", cur_rank);
                    cur_rank = -1;
                    i = j;
                    continue;
                }
                if (Char.IsNumber(clique[i]))
                {
                    j = i;
                    while (Char.IsNumber(clique[j]))
                    {
                        j++;
                        if (j == clique.Length)
                            break;
                    }
                    cur_rank = Convert.ToInt32(clique.Substring(i, j - i));
                    sb.AppendFormat("{0} ", cur_rank);
                    cur_rank = -1;
                    i = j;
                    continue;
                }
                if (clique[i] == ')')
                {
                    i++;
                    if (cur_rank != -1)
                    {
                        sb.AppendFormat("{0} ", cur_rank);
                        cur_rank = -1;
                    }
                }
            }
            if (cur_rank != -1)
            {
                sb.AppendFormat("{0} ", cur_rank);
            }
            string str = sb.ToString().TrimEnd(null);
            return str.Split(' ');
        }

        static public int[] ParseCliqueInts(string clique)
        {
            clique = clique.Trim();
            ArrayList ranks = new ArrayList();
            int cur_rank, i, j, top;
            i = 0;
            cur_rank = -1;
            while (i < clique.Length)
            {
                if (clique[i] == '(')
                {
                    i++;
                    cur_rank = -1;
                    if (i == clique.Length)
                        break;
                }
                if (clique[i] == '.')
                {
                    i++;
                    if (i == clique.Length)
                        break;
                }
                if (clique[i] == '.' || clique[i] == '-')
                {
                    i++;
                    if (i == clique.Length)
                        break;
                    j = i;
                    while (Char.IsNumber(clique[j]))
                    {
                        j++;
                        if (j == clique.Length)
                            break;
                    }
                    top = Convert.ToInt32(clique.Substring(i, j - i));
                    i = j;
                    for (j = cur_rank + 1; j <= top; j++)
                    {
                        ranks.Add(j);
                    }
                    cur_rank = -1;
                    continue;
                }
                if (clique[i] == ',')
                {
                    i++;
                    if (i == clique.Length)
                        break;
                    j = i;
                    while (Char.IsNumber(clique[j]))
                    {
                        j++;
                        if (j == clique.Length)
                            break;
                    }
                    cur_rank = Convert.ToInt32(clique.Substring(i, j - i));
                    ranks.Add(cur_rank);
                    i = j;
                    continue;
                }
                if (Char.IsNumber(clique[i]))
                {
                    j = i;
                    while (Char.IsNumber(clique[j]))
                    {
                        j++;
                        if (j == clique.Length)
                            break;
                    }
                    cur_rank = Convert.ToInt32(clique.Substring(i, j - i));
                    ranks.Add(cur_rank);
                    i = j;
                    continue;
                }
                if (clique[i] == ')')
                {
                    i++;
                    cur_rank = -1;
                }
            }
            return (int[])ranks.ToArray(typeof(int));
        }

        void print_queues_watcher_fn()
        {
            try
            {
                WaitHandle[] handles = new WaitHandle[3];
                handles[0] = print_queues_event;
                handles[1] = abort_job_event;
                handles[2] = print_queues_exit_event;
                while (true)
                {
                    int index = WaitHandle.WaitAny(handles);
                    if (index > 0 || job_aborted)
                        return;
                    print_queues_event.Reset();
                    if (client_list != null)
                    {
                        foreach (ClientSession client in client_list)
                        {
                            try
                            {
                                ActiveMessage message = new ActiveMessage();
                                message.active_type = "client";
                                message.name = "queues";
                                //mpiexec.WriteLine("Sending queues message to client");
                                client.SendMessage(message);
                                //Thread.Sleep(1);
                                message = new ActiveMessage();
                                message.active_type = "client";
                                message.name = "history";
                                client.SendMessage(message);
                            }
                            catch
                            {
                            }
                        }
                    }
                }
            }
            catch (Exception e)
            {
                if (!job_aborted)
                {
                    mpiexec.WriteLine("Exception caught in print_queues_watcher_fn: {0}", e.Message);
                    mpiexec.WriteLine("Stack trace: {1}", e.StackTrace);
                }
            }
        }

        void abort_job_watcher_fn()
        {
            try
            {
                AbortWatcherArg arg;
                WaitHandle[] handles = new WaitHandle[2];
                arg = (AbortWatcherArg)abort_thread_param_queue.Dequeue();
                handles[0] = arg.stop_event;
                handles[1] = abort_job_event;
                //mpiexec.WriteLine("abort_job_watcher thread waiting");
                int index = WaitHandle.WaitAny(handles);
                if (job_aborted)
                {
                    /*
                    if (exit_info != null && arg.Rank >= 0 && arg.Rank < exit_info.Length)
                    {
                        if (exit_info[arg.Rank].message == null)
                            exit_info[arg.Rank].message = PMKilledMessage;
                    }
                    */
                    AbortProcessMessage message = new AbortProcessMessage();
                    message.exit_code = job_aborted_exit_code;
                    //mpiexec.WriteLine("Sending AbortProcessMessage");
                    arg.client.SendMessage(message);
                }
                //mpiexec.WriteLine("abort_job_watcher thread exiting");
            }
            catch (Exception e)
            {
                if (!job_aborted)
                {
                    mpiexec.WriteLine("Exception caught in abort_job_watcher_fn: {0}", e.Message);
                    mpiexec.WriteLine("Stack trace: {1}", e.StackTrace);
                }
            }
        }

        /// <summary>
        /// This thread handles the smpd pmi wire protocol
        /// </summary>
        void pmi_thread_fn()
        {
            try
            {
                TcpListener listener;
                TcpClient client;
                SmpdPmi smpd_pmi = new SmpdPmi(mpiexec);
                string name, key, val, common_result;
                PMIClientArg arg;
                int rank = -1;
                PMI pmi;
                int i, j;

                arg = (PMIClientArg)pmi_queue.Dequeue();
                pmi = arg.pmi;
                listener = arg.listener;
                //mpiexec.WriteLine("pmi_thread waiting for connection");
                // race condition here between AcceptTcpClient and setting connected to true
                client = listener.AcceptTcpClient();
                client.NoDelay = true;
                arg.connected = true;
                listener.Stop();

                // read pmi messges
                //mpiexec.WriteLine("pmi_thread waiting for pmi messges");
                while (smpd_pmi.NextMessage(client))
                {
                    //mpiexec.WriteLine("pmi command received:");
                    //smpd_pmi.Print();
                    name = (string)smpd_pmi.data["name"];
                    key = (string)smpd_pmi.data["key"];
                    val = (string)smpd_pmi.data["value"];
                    common_result = String.Format("src=0 dest={0} ctx_id={1} tag=0 cmd_tag={2} ", smpd_pmi.data["src"], smpd_pmi.data["ctx_key"], smpd_pmi.data["tag"]);
                    switch ((string)smpd_pmi.data["cmd"])
                    {
                        case "init":
                            // name = kvsname
                            // key = rank
                            // val = size
                            // returns:
                            // result = success/fail
                            rank = Convert.ToInt32(key);
                            // Check to see if any other process has exited without calling PMI_Init
                            // Perform this check before calling pmi.Init to avoid hanging in that function
                            if (exit_info != null && rank < exit_info.Length)
                            {
                                exit_info[rank].pmi_init_called = true;
                                //mpiexec.WriteLine("exit_info[{0}].pmi_init_called = true", rank);
                                for (i = 0; i < pmi.nproc; i++)
                                {
                                    if (exit_info[i].exited && !exit_info[i].pmi_init_called)
                                    {
                                        job_aborted_exit_code = -1;
                                        exit_info[rank].message = job_aborted_message = String.Format("Aborted because process {0} exited without calling PMI_Init()", i);
                                        if (exit_info[i].message == null)
                                            exit_info[i].message = "Process exited without calling PMI_Init";
                                        job_aborted_rank = rank;
                                        job_aborted = true; //mpiexec.WriteLine("job_aborted - init not called");
                                        abort_job_event.Set();
                                        break;
                                    }
                                }
                            }
                            //mpiexec.WriteLine("[{0}] init", rank);
                            pmi.Init(name, rank, Convert.ToInt32(val));
                            smpd_pmi.SendResultMessage(client, common_result, true);
                            break;
                        case "finalize":
                            // name = kvsname
                            // key = rank
                            // returns:
                            // result = success/fail
                            //mpiexec.WriteLine("[{0}] finalize", rank);
                            if (exit_info != null && rank < exit_info.Length)
                            {
                                exit_info[rank].pmi_finalize_called = true;
                                //mpiexec.WriteLine("exit_info[{0}].pmi_finalize_called = true", rank);
                            }
                            pmi.Finalize(name, Convert.ToInt32(key));
                            smpd_pmi.SendResultMessage(client, common_result, true);
                            break;
                        case "abort_job":
                            // name = kvsname
                            // rank = rank
                            // error = error message
                            // exit_code = int exitcode
                            job_aborted = true; //mpiexec.WriteLine("job_aborted - abort_job smpd pmi command received");
                            job_aborted_exit_code = Convert.ToInt32(smpd_pmi.data["exit_code"]);
                            job_aborted_message = (string)smpd_pmi.data["error"];
                            job_aborted_rank = Convert.ToInt32(smpd_pmi.data["rank"]);
                            if (exit_info != null && job_aborted_rank < exit_info.Length)
                            {
                                exit_info[job_aborted_rank].exit_code = job_aborted_exit_code;
                                exit_info[job_aborted_rank].message = job_aborted_message;
                            }
                            abort_job_event.Set();
                            break;
                        case "barrier":
                            // name = kvsname
                            // val = count
                            // returns:
                            // result = success/fail
                            //mpiexec.WriteLine("[{0}] barrier in", rank);
                            //pmi.Barrier();
                            pmi.KVS_Barrier(name, Convert.ToInt32(val));
                            //mpiexec.WriteLine("[{0}] barrier out", rank);
                            smpd_pmi.SendDBResultMessage(client, common_result, true);
                            break;
                        case "dbcreate":
                            // returns:
                            // result = success/fail
                            // name = kvsname
                            name = pmi.KVS_Create();
                            //mpiexec.WriteLine("[{0}] created {1}", rank, name);
                            smpd_pmi.SendDBResultMessage(client, common_result, true, name, null, null);
                            break;
                        case "dbdestroy":
                            // name = kvsname
                            // returns:
                            // result = success/fail
                            //mpiexec.WriteLine("[{0}] destroy {1}", rank, name);
                            pmi.KVS_Destroy(name);
                            smpd_pmi.SendDBResultMessage(client, common_result, true);
                            break;
                        case "dbput":
                            // name = kvsname
                            // key = key
                            // val = val
                            // returns:
                            // result = success/fail
                            //mpiexec.WriteLine("[{0}] put {1}:{2}", rank, key, val);
                            pmi.KVS_Put(name, key, val);
                            smpd_pmi.SendDBResultMessage(client, common_result, true);
                            break;
                        case "dbget":
                            // name = kvsname
                            // key = key
                            // returns:
                            // result = success/fail
                            // value = val
                            //mpiexec.WriteLine("[{0}] get {1}", rank, key);
                            val = pmi.KVS_Get(name, key);
                            smpd_pmi.SendDBResultMessage(client, common_result, true, null, null, val);
                            break;
                        case "dbfirst":
                            // name = kvsname
                            // returns:
                            // result = success/fail
                            // key = key or DBS_END_STR
                            // value = val or nothing
                            break;
                        case "dbnext":
                            // name = kvsname
                            // returns:
                            // result = success/fail
                            // key = key or DBS_END_STR
                            // value = val or nothing
                            break;
                        case "spawn":
                            string[] commands;
                            string[][] argvs;
                            int[] maxprocs;
                            PMI_Keyval[][] keyvals;
                            PMI_Keyval[] preput;
                            string[] errors;

                            int count = Convert.ToInt32((string)smpd_pmi.data["ncmds"]);
                            if (count > 0)
                            {
                                commands = new string[count];
                                argvs = new string[count][];

                                maxprocs = new int[count];
                                for (i = 0; i < count; i++)
                                    maxprocs[i] = 1;
                                string[] maxprocs_str = ((string)smpd_pmi.data["maxprocs"]).Split(' ');
                                for (i = 0; i < maxprocs_str.Length; i++)
                                {
                                    if (i < count)
                                        maxprocs[i] = Convert.ToInt32(maxprocs_str[i]);
                                }

                                int[] nkeyvals = new int[count];
                                for (i = 0; i < count; i++)
                                    nkeyvals[i] = 0;
                                string[] nkeyvals_str = ((string)smpd_pmi.data["nkeyvals"]).Split(' ');
                                for (i = 0; i < nkeyvals_str.Length; i++)
                                {
                                    if (i < count)
                                        nkeyvals[i] = Convert.ToInt32(nkeyvals_str[i]);
                                }
                                keyvals = new PMI_Keyval[count][];

                                errors = new string[count];

                                for (i = 0; i < count; i++)
                                {
                                    errors[i] = null;
                                    commands[i] = (string)smpd_pmi.data[String.Format("cmd{0}", i)];
                                    argvs[i] = SmpdPmi.ReadStrings((string)smpd_pmi.data[String.Format("argv{0}", i)]);
                                    int num_args = nkeyvals[i];
                                    if (num_args > 0)
                                    {
                                        keyvals[i] = new PMI_Keyval[num_args];
                                        string[] keyvals_str = SmpdPmi.ReadKeyvals((string)smpd_pmi.data[String.Format("keyvals{0}", i)]);
                                        for (j = 0; j < num_args; j++)
                                        {
                                            keyvals[i][j] = new PMI_Keyval();
                                            int index = 0;
                                            keyvals[i][j].key = SmpdPmi.ReadString(keyvals_str[j], ref index);
                                            SmpdPmi.ReadString(keyvals_str[j], ref index); // throw away the = character
                                            keyvals[i][j].val = SmpdPmi.ReadString(keyvals_str[j], ref index);
                                        }
                                    }
                                    else
                                    {
                                        keyvals[i] = null;
                                    }
                                }

                                count = Convert.ToInt32((string)smpd_pmi.data["npreput"]);
                                if (count > 0)
                                {
                                    preput = new PMI_Keyval[count];
                                    string[] keyvals_str = SmpdPmi.ReadKeyvals((string)smpd_pmi.data["preput"]);
                                    for (i = 0; i < count; i++)
                                    {
                                        preput[i] = new PMI_Keyval();
                                        int index = 0;
                                        preput[i].key = SmpdPmi.ReadString(keyvals_str[i], ref index);
                                        SmpdPmi.ReadString(keyvals_str[i], ref index); // throw away the = character
                                        preput[i].val = SmpdPmi.ReadString(keyvals_str[i], ref index);
                                    }
                                }
                                else
                                {
                                    preput = null;
                                }
                                pmi.Spawn_multiple(commands, argvs, maxprocs, keyvals, preput, errors, false);
                                StringBuilder sb = new StringBuilder();
                                sb.Append("Spawn failed:\n");
                                bool found = false;
                                for (i = 0; i < errors.Length; i++)
                                {
                                    if (errors[i] != null)
                                    {
                                        sb.AppendFormat("spawnee rank {0} failed, error: {1}\n", i, errors[i]);
                                        found = true;
                                    }
                                }
                                if (found)
                                {
                                    // FIXME: Do something with the error string created above
                                    smpd_pmi.SendResultMessage(client, common_result, false);
                                }
                                else
                                {
                                    smpd_pmi.SendResultMessage(client, common_result, true);
                                }
                            }
                            break;
                        case "done":
                            client.Close();
                            //mpiexec.WriteLine("pmi_thread exiting");
                            return;
                        case null:
                            goto case "";
                        case "":
                            mpiexec.WriteLine("Invalid smpd pmi command received");
                            smpd_pmi.Print();
                            break;
                        default:
                            mpiexec.WriteLine("Unknown smpd pmi command received");
                            smpd_pmi.Print();
                            break;
                    }
                }
            }
            catch (Exception)// e)
            {
                //mpiexec.WriteLine("pmi_thread exception: {0}", e.Message);
            }
            //mpiexec.WriteLine("pmi_thread exiting");
        }

        string SpawnErrorsToString(string[] errors)
        {
            if (errors == null)
                return "";
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < errors.Length; i++)
            {
                if ((errors[i] == null) || (errors[i] == ""))
                    sb.Append("null ");
                else
                    sb.AppendFormat("{0} ", Convert.ToBase64String(Encoding.Unicode.GetBytes(errors[i])));
            }
            return sb.ToString().TrimEnd(' ');
        }

        public static void SpawnMessageStringToErrorStrings(string message, string[] errors)
        {
            string[] tokens = message.Split(' ');
            int min_length = tokens.Length;
            if (min_length > errors.Length)
                min_length = errors.Length;
            for (int i = 0; i < min_length; i++)
            {
                if (tokens[i] == "null" || tokens[i] == "")
                    errors[i] = null;
                else
                    errors[i] = Encoding.Unicode.GetString(Convert.FromBase64String(tokens[i]));
            }
            if (min_length < errors.Length)
            {
                for (int i = min_length; i < errors.Length; i++)
                    errors[i] = null;
            }
        }

        void HandlePMICommand(PMI pmi, PMIMessage message, ClientSession session, LaunchNode mynode)
        {
            int i, j;
            PMI_Keyval kv;
            PMIMessage result = new PMIMessage();
            result.data["success"] = "true";
            PMIMessage success_result = new PMIMessage();
            success_result.pmi_type = "server";
            success_result.name = "result";
            success_result.data["success"] = "true";
            success_result.data["message"] = "success";
            result.pmi_type = "server";
            bool abort_job_after_result = false;
            //message.Print();
            try
            {
                switch (message.name)
                {
                    case "init":
                        // Check to see if any other process has exited with calling PMI_Init.
                        // Perform this check before calling pmi.Init to avoid hanging in that function.
                        exit_info[mynode.iproc].pmi_init_called = true;
                        //mpiexec.WriteLine("exit_info[{0}].pmi_init_called = true", mynode.iproc);
                        for (i = 0; i < pmi.nproc; i++)
                        {
                            if (exit_info[i].exited && !exit_info[i].pmi_init_called)
                            {
                                job_aborted_exit_code = -1;
                                exit_info[mynode.iproc].message = job_aborted_message = String.Format("Aborted because process {0} exited without calling PMI_Init()", i);
                                if (exit_info[i].message == null)
                                    exit_info[i].message = "Process exited without calling PMI_Init";
                                job_aborted_rank = mynode.iproc;
                                job_aborted = true; //mpiexec.WriteLine("job_aborted - process did not call init");
                                //abort_job_event.Set();
                                abort_job_after_result = true;
                                result.data["success"] = "false";
                                result.data["message"] = exit_info[mynode.iproc].message;
                                break;
                            }
                        }
                        //mpiexec.WriteLine("init: key = {0}", (string)message.data["key"]);
                        pmi.Init(pmi.id, Convert.ToInt32((string)message.data["key"]), pmi.nproc);
                        //mpiexec.WriteLine("pmi.Init returned");
                        result.name = "init";
                        result.data["size"] = pmi.nproc.ToString();
                        result.data["rank"] = mynode.iproc.ToString();
                        result.data["spawned"] = mynode.spawned.ToString();
                        result.data["id"] = pmi.id;
                        result.data["domain"] = pmi.domain;
                        result.data["appnum"] = mynode.appnum.ToString();
                        break;
                    case "finalize":
                        //mpiexec.WriteLine("finalize: id = {0}, rank = {1}", (string)message.data["id"], (string)message.data["rank"]);
                        exit_info[mynode.iproc].pmi_finalize_called = true;
                        //mpiexec.WriteLine("exit_info[{0}].pmi_finalize_called = true", mynode.iproc);
                        pmi.Finalize(pmi.id, Convert.ToInt32((string)message.data["rank"]));
                        result = success_result;
                        break;
                    case "size":
                        result.name = "size";
                        result.data["size"] = pmi.nproc.ToString();
                        break;
                    case "rank":
                        result.name = "rank";
                        result.data["rank"] = mynode.iproc.ToString();
                        break;
                    case "universe":
                        result.name = "universe";
                        result.data["universe"] = "-1"; // FIXME
                        break;
                    case "appnum":
                        result.name = "appnum";
                        result.data["appnum"] = mynode.appnum.ToString();
                        break;
                    case "publish":
                        //mpiexec.WriteLine("publish: name = {0}, port = {1}", (string)message.data["name"], (string)message.data["port"]);
                        pmi.Publish_name((string)message.data["name"], (string)message.data["port"]);
                        result = success_result;
                        break;
                    case "unpublish":
                        //mpiexec.WriteLine("unpublish: name = {0}", (string)message.data["name"]);
                        pmi.Unpublish_name((string)message.data["name"]);
                        result = success_result;
                        break;
                    case "lookup":
                        //mpiexec.WriteLine("lookup: name = {0}", (string)message.data["name"]);
                        result.name = "lookup";
                        result.data["port"] = pmi.Lookup_name((string)message.data["name"]);
                        break;
                    case "id":
                        result.name = "id";
                        result.data["id"] = pmi.id;
                        break;
                    case "domain":
                        result.name = "domain";
                        result.data["domain"] = pmi.domain;
                        break;
                    case "id_maxlen":
                        result.name = "id_maxlen";
                        result.data["id_maxlen"] = PMI.ID_MAXLEN.ToString();
                        break;
                    case "barrier":
                        //mpiexec.WriteLine("barrier: id = {0}, rank = {1}", (string)message.data["id"], (string)message.data["rank"]);
                        //pmi.KVS_Barrier((string)message.data["id"], pmi.nproc);
                        pmi.Barrier();
                        result = success_result;
                        break;
                    case "clique_size":
                        result.name = "clique_size";
                        result.data["clique_size"] = CountClique(mynode.clique).ToString();
                        break;
                    case "clique":
                        result.name = "clique";
                        string[] ranks = ParseClique(mynode.clique);
                        result.data["num_ranks"] = ranks.Length;
                        for (i = 0; i < ranks.Length; i++)
                        {
                            result.data[String.Format("rank{0}", i)] = ranks[i];
                        }
                        break;
                    case "abort":
                        //mpiexec.WriteLine("abort: id = {0}, rank = {1}, exit code = {2}, message = {3}", (string)message.data["id"], (string)message.data["rank"], (string)message.data["exit code"], (string)message.data["message"]);
                        job_aborted_exit_code = Convert.ToInt32((string)message.data["exit code"]);
                        job_aborted_message = (string)message.data["message"];
                        job_aborted_rank = Convert.ToInt32((string)message.data["rank"]);
                        job_aborted = true; //mpiexec.WriteLine("job_aborted - abort command received");
                        if ((exit_info != null) && (job_aborted_rank < exit_info.Length))
                        {
                            exit_info[job_aborted_rank].exit_code = job_aborted_exit_code;
                            exit_info[job_aborted_rank].message = job_aborted_message;
                        }
                        //abort_job_event.Set();
                        abort_job_after_result = true;
                        break;
                    case "name_maxlen":
                        result.name = "name_maxlen";
                        result.data["name_maxlen"] = PMI.NAME_MAXLEN.ToString();
                        break;
                    case "key_maxlen":
                        result.name = "key_maxlen";
                        result.data["key_maxlen"] = PMI.KEY_MAXLEN.ToString();
                        break;
                    case "value_maxlen":
                        result.name = "value_maxlen";
                        result.data["value_maxlen"] = PMI.VALUE_MAXLEN.ToString();
                        break;
                    case "create":
                        result.name = "created";
                        result.data["name"] = pmi.KVS_Create();
                        break;
                    case "destroy":
                        //mpiexec.WriteLine("destroy: name = {0}", (string)message.data["name"]);
                        pmi.KVS_Destroy((string)message.data["name"]);
                        result = success_result;
                        break;
                    case "put":
                        //mpiexec.WriteLine("put: kvs = {0}, key = {1}, value = {2}", (string)message.data["kvs"], (string)message.data["key"], (string)message.data["value"]);
                        pmi.KVS_Put((string)message.data["kvs"], (string)message.data["key"], (string)message.data["value"]);
                        result = success_result;
                        break;
                    case "commit":
                        //mpiexec.WriteLine("commit: kvs = {0}", (string)message.data["kvs"]);
                        result = success_result;
                        break;
                    case "get":
                        //mpiexec.WriteLine("get: kvs = {0}, key = {1}", (string)message.data["kvs"], (string)message.data["key"]);
                        result.name = "get";
                        result.data["value"] = pmi.KVS_Get((string)message.data["kvs"], (string)message.data["key"]);
                        //mpiexec.WriteLine("got: value = {0}", result.data["value"]);
                        break;
                    case "iter_first":
                        //mpiexec.WriteLine("iter_first: kvs = {0}", (string)message.data["kvs"]);
                        result.name = "iter";
                        kv = pmi.Iter_first();
                        if (kv != null)
                        {
                            result.data["key"] = kv.key;
                            result.data["value"] = kv.val;
                        }
                        else
                        {
                            result.data["key"] = null;
                            result.data["value"] = null;
                        }
                        break;
                    case "iter":
                        //mpiexec.WriteLine("iter: kvs = {0}", (string)message.data["kvs"]);
                        result.name = "iter";
                        kv = pmi.Iter_next();
                        if (kv != null)
                        {
                            result.data["key"] = kv.key;
                            result.data["value"] = kv.val;
                        }
                        else
                        {
                            result.data["key"] = null;
                            result.data["value"] = null;
                        }
                        break;
                    case "spawn":
                        string[] commands;
                        string[][] argvs;
                        int[] maxprocs;
                        PMI_Keyval[][] keyvals;
                        PMI_Keyval[] preput;
                        string[] errors;
                        int total_num_procs = 0;
                        bool quiet;

                        //mpiexec.WriteLine("spawn command received");
                        int count = (int)message.data["commands_count"];
                        if (count > 0)
                        {
                            commands = new string[count];
                            argvs = new string[count][];
                            maxprocs = new int[count];
                            keyvals = new PMI_Keyval[count][];
                            for (i = 0; i < count; i++)
                            {
                                commands[i] = (string)message.data[String.Format("exe{0}", i)];
                                maxprocs[i] = (int)message.data[String.Format("maxproc{0}", i)];
                                total_num_procs += maxprocs[i];
                                int num_args = (int)message.data[String.Format("args{0}_count", i)];
                                if (num_args > 0)
                                {
                                    argvs[i] = new string[num_args];
                                    for (j = 0; j < num_args; j++)
                                    {
                                        argvs[i][j] = (string)message.data[String.Format("args{0}_arg{1}", i, j)];
                                    }
                                }
                                else
                                {
                                    argvs[i] = null;
                                }
                                num_args = (int)message.data[String.Format("keyvals{0}_count", i)];
                                if (num_args > 0)
                                {
                                    keyvals[i] = new PMI_Keyval[num_args];
                                    for (j = 0; j < num_args; j++)
                                    {
                                        keyvals[i][j] = new PMI_Keyval();
                                        keyvals[i][j].key = (string)message.data[String.Format("keyvals{0}_key{1}", i, j)];
                                        keyvals[i][j].val = (string)message.data[String.Format("keyvals{0}_value{1}", i, j)];
                                    }
                                }
                                else
                                {
                                    keyvals[i] = null;
                                }
                            }
                            count = (int)message.data["preput_count"];
                            if (count > 0)
                            {
                                preput = new PMI_Keyval[count];
                                for (i = 0; i < count; i++)
                                {
                                    preput[i] = new PMI_Keyval();
                                    preput[i].key = (string)message.data[String.Format("preput_key{0}", i)];
                                    preput[i].val = (string)message.data[String.Format("preput_value{0}", i)];
                                }
                            }
                            else
                            {
                                preput = null;
                            }
                            quiet = (bool)message.data["quiet"];
                            errors = new string[total_num_procs];
                            for (i = 0; i < total_num_procs; i++)
                                errors[i] = null;
                            //mpiexec.WriteLine("calling PMI.Spawn_multiple");
                            pmi.Spawn_multiple(commands, argvs, maxprocs, keyvals, preput, errors, quiet);
                            //mpiexec.WriteLine("PMI.Spawn_multiple returned");
                            bool found = false;
                            for (i = 0; i < errors.Length; i++)
                            {
                                if (errors[i] != null)
                                {
                                    found = true;
                                }
                            }
                            if (found)
                            {
                                result.data["message"] = SpawnErrorsToString(errors);
                            }
                        }
                        break;
                    default:
                        result.name = "result";
                        result.data["success"] = "false";
                        result.data["message"] = String.Format("unknown pmi command: {0}", message.name);
                        break;
                }
            }
            catch (Exception e)
            {
                result = new PMIMessage();
                result.name = "result";
                result.pmi_type = "server";
                result.data["success"] = "false";
                result.data["message"] = e.Message;
            }
            //mpiexec.WriteLine("Sending result message");
            session.SendMessage(result);
            if (abort_job_after_result)
            {
                //mpiexec.WriteLine("aborting job after handling PMI message");
                abort_job_event.Set();
            }
        }

        // Mpiexec.verify_thread and DeinoMPIWin.verify_thread borrow launching code from deinopm_thread_fn so if it is changed here the verify_thread functions need to be updated also.
        // The more maintainable solution would be to extract common code into several functions and have everyone call the same functions.
        public void deinopm_thread_fn()
        {
            LaunchNode node = null;
            Message message = null;
            try
            {
                string error = null;
                ClientSession client = null;
                LaunchMessage launch;
                ResultMessage result;
                QuitMessage quit_message;
                PMI pmi;
                Thread pmi_thread = null;
                bool smpd_pmi_cancelled = false;

                node = (LaunchNode)thread_param_queue.Dequeue();
                pmi = node.pmi;

                //node.Print();

                // connect to the host
                try
                {
                    if (mpiexec.global_options.localonly || node.launch_local)
                    {
                        int port = DeinoPMUtil.StartLocalPM(mpiexec, node.private_key);
                        client = new ClientSession(port, node.private_key);
                    }
                    else
                    {
                        client = new ClientSession(node.hostname, node.port, node.phrase, node.private_key, node.credential);
                    }
                }
                catch (Exception e)
                {
                    StringBuilder sb = new StringBuilder();
                    sb.Append(e.Message);
                    foreach (string key in e.Data.Keys)
                    {
                        sb.AppendFormat("\r\n{0}={1}", key, e.Data[key]);
                    }
                    //sb.AppendFormat("\r\nStack trace:\r\n{0}", e.StackTrace);
                    error = sb.ToString();
                }

                ///
                /// barrier across all threads until everyone is connected or errored out
                /// 

                client_list.Add(client);
                bool wakeup = false;
                lock (typeof(ProcessGroup))
                {
                    num_connecting--;
                    if (error != null)
                        connection_failed = true;
                    if (num_connecting < 1)
                        wakeup = true;
                }
                if (wakeup)
                    connected_event.Set();
                else
                    connected_event.WaitOne();
                if (error != null)
                {
                    exit_info[node.iproc].message = String.Format("Unable to establish a secure connection to host '{0}' for rank {1} using user account '{2}'\nError: {3}\n", node.hostname, node.iproc, node.credential.Account, error);
                    exit_info[node.iproc].error = -1;
                    exit_info[node.iproc].running_or_error_event.Set();
                    //mpiexec.WriteLine("Unable to establish a secure connection to host '{0}' for rank {1} using user account '{2}'", node.hostname, node.iproc, node.credential.Account);
                    //mpiexec.WriteLine("Error: {0}", error);
                    // let the output queue thread know that this thread is exiting
                    if (mpiexec.global_options.use_output_queue)
                        mpiexec.global_options.output_queue.Enqueue(Thread.CurrentThread);
                    return;
                }
                else
                {
                    if (connection_failed)
                    {
                        //mpiexec.WriteLine("Aborting connection #{0} to {1} due to an error in another connection.", node.iproc, node.hostname);
                        quit_message = new QuitMessage();
                        client.SendMessage(quit_message);
                        // let the output queue thread know that this thread is exiting
                        if (mpiexec.global_options.use_output_queue)
                            mpiexec.global_options.output_queue.Enqueue(Thread.CurrentThread);
                        exit_info[node.iproc].running_or_error_event.Set();
                        return;
                    }
                }

                ///
                /// launch the process
                /// 

                launch = new LaunchMessage();

                // create a launch message from the launch node

                // add the process fields
                launch.path = node.path;
                if (node.map != null)
                {
                    foreach (string drive in node.map.Keys)
                    {
                        launch.map[drive] = node.map[drive];
                    }
                }
                launch.dir = node.dir;
                launch.exe = node.exe;
                if (node.args != null)
                {
                    foreach (string arg in node.args)
                    {
                        launch.args.Add(arg);
                    }
                }
                if (node.env != null)
                {
                    foreach (string key in node.env.Keys)
                    {
                        launch.env[key] = node.env[key];
                    }
                }
                launch.priority_class = node.priority_class;
                launch.priority_thread = node.priority_thread;

                // add the pmi variables
                TcpListener pmi_listener = null;
                pmi_listener = new TcpListener(IPAddress.Any, 0);
                pmi_listener.Start();
                if (node.spawned)
                    launch.env["PMI_SPAWN"] = "1";
                launch.env["PMI_RANK"] = node.iproc.ToString();
                launch.env["PMI_SIZE"] = node.nproc.ToString();
                launch.env["PMI_CLIQUE"] = node.clique;
                launch.env["PMI_APPNUM"] = node.appnum.ToString();
                launch.env["PMI_DOMAIN"] = pmi.domain;
                launch.env["PMI_KVS"] = pmi.id;
                // SMPD variables
                launch.env["PMI_HOST"] = System.Net.Dns.GetHostName();
                launch.env["PMI_PORT"] = ((IPEndPoint)pmi_listener.LocalEndpoint).Port.ToString();
                launch.env["PMI_SMPD_ID"] = node.iproc.ToString();
                launch.env["PMI_SMPD_KEY"] = node.iproc.ToString();
                // DeinoPMI variables
                launch.env["PMI_KEY"] = node.iproc.ToString();

                // start the pmi thread
                PMIClientArg pmi_arg = null;
                pmi_arg = new PMIClientArg();
                pmi_arg.pmi = pmi;
                pmi_arg.listener = pmi_listener;
                pmi_arg.connected = false;
                pmi_queue.Enqueue(pmi_arg);
                ThreadStart pmi_threadProc = new ThreadStart(pmi_thread_fn);
                pmi_thread = new Thread(pmi_threadProc);
                pmi_thread.Start();

                //mpiexec.WriteLine("sending launch");
                client.SendMessage(launch);
                result = (ResultMessage)client.NextMessage();
                if (!result.success)
                {
                    exit_info[node.iproc].message = String.Format("launch failed: {0}\n", result.message);
                    exit_info[node.iproc].error = -1;
                    exit_info[node.iproc].running_or_error_event.Set();
                    //mpiexec.WriteLine("launch failed: {0}", result.message);
                    if (!node.spawned)
                    {
                        job_aborted = true; //mpiexec.WriteLine("job_aborted - launch failed");
                        abort_job_event.Set();
                    }
                    if (!pmi_arg.connected)
                    {
                        pmi_listener.Stop();
                        pmi_thread.Abort();
                    }
                    pmi_thread.Join();
                }
                else
                {
                    // The process was successfully launched.
                    exit_info[node.iproc].running_or_error_event.Set();

                    Thread abort_watcher_thread, print_queues_watcher_thread = null;
                    AbortWatcherArg arg = new AbortWatcherArg(client);
                    arg.Rank = node.iproc;
                    abort_thread_param_queue.Enqueue(arg);
                    ThreadStart threadProc = new ThreadStart(abort_job_watcher_fn);
                    abort_watcher_thread = new Thread(threadProc);
                    abort_watcher_thread.Start();

                    if (node.iproc == 0)
                    {
                        // Make sure the client is added before starting the input so no input is lost
                        mpiexec.input.AddClient(client);
                        if (!mpiexec.input.IsRunning)
                        {
                            if (mpiexec.global_options.use_output_queue)
                            {
                                mpiexec.input.Start();
                            }
                            else
                            {
                                mpiexec.input.Start(Console.In);
                            }
                        }

                        // Start a watcher thread to send print queue messages
                        ThreadStart ts = new ThreadStart(this.print_queues_watcher_fn);
                        print_queues_watcher_thread = new Thread(ts);
                        print_queues_watcher_thread.Start();
                    }

                    bool done = false;
                    while (!done)
                    {
                        message = client.NextMessage();
                        switch (message.MessageType)
                        {
                            case InputOutputMessage.TypeString:
                                {
                                    switch (((InputOutputMessage)message).io_type)
                                    {
                                        case InputOutputMessage.IO_Type.OUTPUT:
                                            mpiexec.Write(((InputOutputMessage)message).Buffer);
                                            break;
                                        case InputOutputMessage.IO_Type.ERROR:
                                            mpiexec.WriteError(((InputOutputMessage)message).Buffer);
                                            break;
                                    }
                                    break;
                                }
                            case ProcessExitMessage.TypeString:
                                exit_info[node.iproc].exit_code = ((ProcessExitMessage)message).exit_code;
                                exit_info[node.iproc].exited = true;
                                if (exit_info[node.iproc].message == null)
                                    exit_info[node.iproc].message = ((ProcessExitMessage)message).message;
                                if (((ProcessExitMessage)message).exit_code != 0)
                                {
                                    // last non-zero exit code wins
                                    exit_code = ((ProcessExitMessage)message).exit_code;
                                }
                                if (!job_aborted)
                                {
                                    // Check for PMI consistency
                                    if (exit_info[node.iproc].pmi_init_called)
                                    {
                                        if (!exit_info[node.iproc].pmi_finalize_called)
                                        {
                                            // Init called without Finalize
                                            job_aborted_exit_code = -1;
                                            exit_info[node.iproc].message = job_aborted_message = "Process exited without calling PMI_Finalize()";
                                            job_aborted_rank = node.iproc;
                                            job_aborted = true; //mpiexec.WriteLine("job_aborted - no pmi_finalize");
                                            abort_job_event.Set();
                                        }
                                        else
                                        {
                                            // Init and Finalize called, check to make sure all other exited processes also called Init and Finalize
                                            for (int i = 0; i < node.nproc; i++)
                                            {
                                                if (exit_info[i].exited && (!exit_info[i].pmi_init_called || !exit_info[i].pmi_finalize_called))
                                                {
                                                    // This process called Init and Finalize but another process exited without calling either
                                                    job_aborted_exit_code = -1;
                                                    exit_info[node.iproc].message = job_aborted_message = String.Format("Process called PMI_Init/Finalize() while process {0} never called PMI_Init", i);
                                                    if (exit_info[i].message == null)
                                                        exit_info[i].message = "Process exited without calling PMI_Init";
                                                    job_aborted_rank = node.iproc;
                                                    job_aborted = true; //mpiexec.WriteLine("job_aborted - proc exited w/o PMI_Init");
                                                    abort_job_event.Set();
                                                    break;
                                                }
                                            }
                                        }
                                    }
                                    else
                                    {
                                        if (exit_info[node.iproc].pmi_finalize_called)
                                        {
                                            // Init NOT called but Finalize called
                                            // This should be impossible
                                            job_aborted_exit_code = -1;
                                            exit_info[node.iproc].message = job_aborted_message = "Process exited after calling PMI_Finalize() without previously calling PMI_Init()";
                                            job_aborted_rank = node.iproc;
                                            job_aborted = true; //mpiexec.WriteLine("job_aborted - finalize w/o init");
                                            abort_job_event.Set();
                                        }
                                        else
                                        {
                                            // This process did not call Init.  Check to see if others did
                                            for (int i = 0; i < node.nproc; i++)
                                            {
                                                if (exit_info[i].pmi_init_called)
                                                {
                                                    job_aborted_exit_code = -1;
                                                    exit_info[node.iproc].message = job_aborted_message = "Process exited without calling PMI_Init()";
                                                    job_aborted_rank = node.iproc;
                                                    job_aborted = true; //mpiexec.WriteLine("job_aborted - no call to PMI_Init");
                                                    abort_job_event.Set();
                                                    break;
                                                }
                                            }
                                        }
                                    }
                                }
                                done = true;
                                break;
                            case PMIMessage.TypeString:
                                // If a PMI messge is received then cancel the SMPD PMI thread.
                                // Only one version of PMI can be selected.
                                if (!smpd_pmi_cancelled)
                                {
                                    //mpiexec.WriteLine("Cancelling smpd pmi thread.");
                                    if (!pmi_arg.connected)
                                    {
                                        // race condition here between when the AcceptTcpClient finishes and the connected field is set to true
                                        pmi_listener.Stop();
                                    }
                                    pmi_thread.Abort();
                                    pmi_thread.Join();
                                    smpd_pmi_cancelled = true;
                                    //mpiexec.WriteLine("smpd pmi thread cancelled.");
                                }
                                HandlePMICommand(pmi, (PMIMessage)message, client, node);
                                break;
                            case ActiveMessage.TypeString:
                                //mpiexec.WriteLine("ActiveMessage result received");
                                ActiveMessage am = (ActiveMessage)message;
                                if (am.name == "queues")
                                {
                                    lock (typeof(ProcessGroup))
                                    {
                                        if (num_queue_messages_outstanding == 0)
                                        {
                                            num_queue_messages_outstanding = node.nproc;
                                        }
                                        num_queue_messages_outstanding--;
                                        if (((string)am.data["success"] == "true") && (am.data["data"] != null))
                                        {
                                            queue_messages[node.iproc] = (string)am.data["data"];
                                        }
                                        else
                                        {
                                            queue_messages[node.iproc] = String.Format("Process {0}: error retreiving queues", node.iproc);
                                        }
                                        if (num_queue_messages_outstanding == 0)
                                        {
                                            mpiexec.WriteLine();
                                            for (int index = 0; index < node.nproc; index++)
                                            {
                                                if (queue_messages[index] != null)
                                                    mpiexec.WriteLine((string)queue_messages[index]);
                                            }
                                            queue_messages.Clear();
                                        }
                                    }
                                    //if (((string)am.data["success"] == "true") && (am.data["data"] != null))
                                    //    mpiexec.WriteLine((string)am.data["data"]);
                                }
                                if (am.name == "history")
                                {
                                    lock (typeof(ProcessGroup))
                                    {
                                        if (num_histroy_messages_outstanding == 0)
                                        {
                                            num_histroy_messages_outstanding = node.nproc;
                                        }
                                        num_histroy_messages_outstanding--;
                                        if (((string)am.data["success"] == "true") && (am.data["data"] != null))
                                        {
                                            history_messages[node.iproc] = (string)am.data["data"];
                                        }
                                        else
                                        {
                                            history_messages[node.iproc] = String.Format("Process {0}: error retreiving history", node.iproc);
                                        }
                                        if (num_histroy_messages_outstanding == 0)
                                        {
                                            mpiexec.WriteLine();
                                            for (int index = 0; index < node.nproc; index++)
                                            {
                                                if (history_messages[index] != null)
                                                    mpiexec.WriteLine((string)history_messages[index]);
                                            }
                                            history_messages.Clear();
                                        }
                                    }
                                }
                                break;
                            default:
                                mpiexec.WriteLine("unexpected {0} message received", message.MessageType);
                                break;
                        }
                    }

                    if (!smpd_pmi_cancelled)
                    {
                        if (!pmi_arg.connected)
                        {
                            // race condition here between when the AcceptTcpClient finishes and the connected field is set to true
                            pmi_listener.Stop();
                            pmi_thread.Abort();
                        }
                        pmi_thread.Join();
                    }

                    if (node.iproc == 0)
                    {
                        mpiexec.input.RemoveClient(client);
                    }

                    // close the abort_watcher thread
                    arg.stop_event.Set();
                    abort_watcher_thread.Join();

                    if (print_queues_watcher_thread != null)
                    {
                        // close the print_queues_watcher_fn thread
                        print_queues_exit_event.Set();
                        print_queues_watcher_thread.Join();
                    }
                }

                ///
                /// shutdown the connection
                /// 

                quit_message = new QuitMessage();
                client.SendMessage(quit_message);
                // get a response to the quit message
                Message message2;
                message2 = client.NextMessage();
                try
                {
                    // Send the last leg of the quit handshake
                    quit_message = new QuitMessage();
                    client.SendMessage(quit_message);
                }
                catch
                {
                    // Ignore errors here since we are guaranteed to be in the last leg of the quit handshake.
                }
                //mpiexec.WriteLine("deinopm_thread exiting");
            }
            catch (Exception ex)
            {
                if (!job_aborted)
                {
                    /*
                    if (node != null)
                        mpiexec.WriteLine("Exception caught in process thread #{0}: {1}", node.iproc, ex.Message);
                    else
                        mpiexec.WriteLine("Exception caught in process thread: {1}", ex.Message);
                    mpiexec.WriteLine("Stack trace:");
                    mpiexec.WriteLine(ex.StackTrace);
                    if (message != null)
                    {
                        mpiexec.WriteLine("Last message type: {0}", message.MessageType);
                    }
                    */
                    if (node != null)
                    {
                        exit_info[node.iproc].message = String.Format("Exception caught in process thread #{0}: {1}\nStack trace:\n{2}", node.iproc, ex.Message, ex.StackTrace);
                        if (message != null)
                            exit_info[node.iproc].message += String.Format("\nLast message type: {0}", message.MessageType);
                        exit_info[node.iproc].error = -1;
                        exit_info[node.iproc].running_or_error_event.Set();
                    }
                    else
                    {
                        // Exception thrown before the node is retrieved.  Hopefully this will never happen.
                        mpiexec.WriteLine("Exception caught in process thread: {1}", ex.Message);
                        mpiexec.WriteLine("Stack trace:");
                        mpiexec.WriteLine(ex.StackTrace);
                        if (message != null)
                        {
                            mpiexec.WriteLine("Last message type: {0}", message.MessageType);
                        }
                        job_aborted = true;
                        abort_job_event.Set();
                    }
                    /*
                    bool tripme = false;
                    mpiexec.WriteLine("Waiting for the debugger to attach...");
                    while (!tripme)
                    {
                        Thread.Sleep(1000);
                    }
                    */
                }
            }

            // Enqueue this thread that is about to exit in the watcher output_queue so that the count of running processes can be decremented.
            if (mpiexec.global_options.use_output_queue && mpiexec.global_options.output_queue != null)
            {
                mpiexec.global_options.output_queue.Enqueue(Thread.CurrentThread);
            }
        }

        public void TimeoutThread()
        {
            try
            {
                Thread.Sleep(timeout * 1000);
                job_timedout_message = String.Format("Timeout of {0} seconds expired", timeout);
                Kill();
                Thread.Sleep(timeout * 500);
                Kill();
                Thread.Sleep(timeout * 250);
                System.Environment.Exit(-1);
            }
            catch //(Exception e)
            {
                /* The timeout thread is always cancelled on successful jobs
                if (!job_aborted)
                {
                    mpiexec.WriteLine("Exception caught in TimeoutThread: {0}", e.Message);
                    mpiexec.WriteLine("Stack trace: {0}", e.StackTrace);
                }
                */
            }
        }

        public void PrintStartErrors()
        {
            if (exit_info != null)
            {
                for (int i = 0; i < exit_info.Length; i++)
                {
                    if (exit_info[i] != null && exit_info[i].error != 0 && exit_info[i].message != null)
                    {
                        string[] lines;
                        lines = exit_info[i].message.Split('\n');
                        foreach (string line in lines)
                        {
                            mpiexec.WriteLine(line);
                        }
                    }
                }
            }
        }

        public void GetStartErrors(string[] errors)
        {
            int min_length = errors.Length;
            if (exit_info.Length < min_length)
                min_length = exit_info.Length;
            for (int i = 0; i < min_length; i++)
            {
                if (exit_info[i].error != 0 && exit_info[i].message != null)
                    errors[i] = exit_info[i].message;
                else
                    errors[i] = null;
            }
            if (min_length < errors.Length)
            {
                for (int i = min_length; i < errors.Length; i++)
                    errors[i] = null;
            }
        }

        /*
        public string[] GetStartErrorStrings()
        {
            if (exit_info == null && exit_info.Length < 1)
                return null;
            string[] results = new string[exit_info.Length];
            for (int i = 0; i < exit_info.Length; i++)
            {
                if (exit_info[i].error != 0 && exit_info[i].message != null)
                    results[i] = exit_info[i].message;
                else
                    results[i] = null;
            }
            return results;
        }
        */

        public bool StartGroup(ArrayList nodes)
        {
            if (nodes == null)
            {
                return true;
            }
            mpiexec.TotalNumProcs += ((LaunchNode)nodes[0]).nproc; // This assumes that nproc processes will be successfully launched.  FIXME: Could this lead to a hang where the output threads are waiting for processes to exit that don't even exist?
            // create a thread for each deinopm connection to launch an individual process
            threads = new ArrayList();
            num_connecting = nodes.Count;
            if (nodes.Count > 0)
            {
                pmi.nproc = ((LaunchNode)nodes[0]).nproc;
                exit_info = new ExitInfo[((LaunchNode)nodes[0]).nproc];
            }
            foreach (LaunchNode node in nodes)
            {
                exit_info[node.iproc] = new ExitInfo();
                exit_info[node.iproc].rank = node.iproc;
                exit_info[node.iproc].host = node.hostname;

                node.pmi = pmi;
                //node.Print();
                Thread thread;
                thread_param_queue.Enqueue(node);
                ThreadStart threadProc = new ThreadStart(this.deinopm_thread_fn);
                thread = new Thread(threadProc);
                threads.Add(thread);
            }
            foreach (Thread thread in threads)
            {
                thread.Start();
            }
            if (nodes.Count > 0 && timeout > 0)
            {
                ThreadStart ts = new ThreadStart(this.TimeoutThread);
                timeout_thread = new Thread(ts);
                timeout_thread.Start();
            }
            bool no_errors = true;
            foreach (LaunchNode node in nodes)
            {
                exit_info[node.iproc].running_or_error_event.WaitOne();
                if (exit_info[node.iproc].error != 0)
                    no_errors = false;
            }
            return no_errors;
        }

        public int WaitForGroupExit()
        {
            // Only one thread should perform the waiting code and the rest should block
            wait_for_exit_mutex.WaitOne();
            if (!already_waited)
            {
                int num_remaining = 0;

                if (threads != null)
                {
                    num_remaining = threads.Count;
                    //mpiexec.WriteLine("Waiting for {0} threads", num_remaining);
                    foreach (Thread thread in threads)
                    {
                        thread.Join();
                        //mpiexec.WriteLine("thread exited, {0} remaining", --num_remaining);
                    }
                    threads = null;
                }
                if (timeout_thread != null)
                {
                    timeout_thread.Abort();
                    timeout_thread = null;
                }
                if (job_aborted)
                {
                    bool doPrint = true;
                    //lock (this)
                    {
                        if (job_abort_printed)
                            doPrint = false;
                        job_abort_printed = true;
                    }
                    if (doPrint)
                    {
                        mpiexec.WriteLine();
                        mpiexec.WriteLine("job aborted:");
                        if (job_timedout_message != null)
                        {
                            mpiexec.WriteLine(job_timedout_message);
                        }
                        else
                        {
                            mpiexec.WriteLine("process: node: exit code[: error message]");
                            for (int i = 0; i < pmi.nproc; i++)
                            {
                                if (exit_info[i] != null)
                                {
                                    if (exit_info[i].message != null)
                                        mpiexec.WriteLine("{0}: {1}: {2}: {3}", exit_info[i].rank, exit_info[i].host, exit_info[i].exit_code, exit_info[i].message);
                                    else
                                        mpiexec.WriteLine("{0}: {1}: {2}", exit_info[i].rank, exit_info[i].host, exit_info[i].exit_code);
                                }
                                else
                                {
                                    mpiexec.WriteLine("{0}: unknown exit status", i);
                                }
                            }
                        }
                    }
                    exit_code = job_aborted_exit_code;
                }
                already_waited = true;
            }
            wait_for_exit_mutex.ReleaseMutex();
            //mpiexec.WriteLine("Process group exiting, code: {0}", exit_code);
            return exit_code;
        }

        public void Kill()
        {
            // Can't do this because it would make enumerating and killing the groups impossible
            //mpiexec.process_groups.Remove(this);

            if (job_aborted)
            {
                if (threads != null)
                {
                    // If the abort job fails then kill all the threads
                    foreach (Thread thread in threads)
                    {
                        if (thread.IsAlive)
                        {
                            try
                            {
                                thread.Abort();
                                thread.Join();
                            }
                            catch
                            {
                            }
                        }
                    }
                }
            }
            else
            {
                // The first time Kill is called, try to abort the job
                job_aborted = true;
                abort_job_event.Set();
            }
        }

        public void PrintExitCodes()
        {
            if (exit_info != null)
            {
                for (int i = 0; i < exit_info.Length; i++)
                {
                    mpiexec.WriteLine("{0}: {1}: {2}", i, exit_info[i].host, exit_info[i].exit_code);
                }
            }
        }
    }

    public class Mpiexec
    {
        public CredentialStore store = null;
        public int mpiexec_exit_code = 0;
        public Options global_options;
        public ArrayList process_groups;
        private int total_nproc;
        public InputHandling input;

        public int TotalNumProcs
        {
            get { return total_nproc; }
            set { total_nproc = value; }
        }

        public Mpiexec()
        {
            total_nproc = 0;
            process_groups = new ArrayList();
            input = new InputHandling();
            global_options = new Options();
            global_options.env["PMI_WDLL_NAME"] = "DeinoPMI.dll";
        }

        public void Kill()
        {
            foreach (ProcessGroup pg in process_groups)
            {
                pg.Kill();
            }
        }

        public void WaitForAllExit()
        {
            ArrayList current_pgs = new ArrayList();
            int max_exit_code = 0, min_exit_code = 0;
            while (process_groups.Count > 0)
            {
                // This has to be done in a loop because new ProcessGroups may be added while waiting for the current
                // groups to exit and you cannot modify the enumeration inside a foreach statement.
                // So what we do is make a copy of the groups, wait for them to all exit, remove them from the global
                // array and then repeat until there are no groups left in the global array.
                current_pgs.Clear();
                foreach (ProcessGroup pg2 in process_groups)
                    current_pgs.Add(pg2);
                foreach (ProcessGroup pg3 in current_pgs)
                {
                    //WriteLine("Waiting for process group: {0}", (pg3.pmi != null) ? pg3.pmi.nproc : -1);
                    int current_exit_code = pg3.WaitForGroupExit();
                    if (current_exit_code != 0)
                    {
                        max_exit_code = (max_exit_code > current_exit_code) ? max_exit_code : current_exit_code;
                        min_exit_code = (min_exit_code < current_exit_code) ? min_exit_code : current_exit_code;
                    }
                    if (global_options.exitcodes)
                    {
                        pg3.PrintExitCodes();
                    }
                    process_groups.Remove(pg3);
                }
            }
            if (max_exit_code != 0)
                mpiexec_exit_code = max_exit_code;
            else if (min_exit_code != 0)
                mpiexec_exit_code = min_exit_code;
            else
                mpiexec_exit_code = 0;
        }

        public bool GenerateStagingNodes(ref ArrayList nodes, ref string error_msg, ArrayList launch_nodes, Options options)
        {
            string src_dir, dest_dir, wildcards;
            bool use_one_node = false;

            if (launch_nodes == null || launch_nodes.Count == 0)
            {
                nodes = null;
                return true;
            }

            // get and validate the source directory
            if (options.stage_src.Trim() == "")
            {
                error_msg = "Please enter a valid path in the source directory field or uncheck the Use Directory Staging check box.";
                return false;
            }
            if (!Directory.Exists(options.stage_src.Trim()))
            {
                error_msg = String.Format("The specified directory does not exist: {0}", options.stage_src);
                return false;
            }
            src_dir = Path.GetFullPath(options.stage_src.Trim());

            // get and validate the destination directory
            if (options.stage_dest != null)
                dest_dir = Path.GetFullPath(options.stage_dest.Trim());
            else
                dest_dir = src_dir;
            if (dest_dir.IndexOfAny(Path.GetInvalidPathChars()) >= 0)
            {
                error_msg = String.Format("The specified destination directory contains invalid path characters: {0}", dest_dir);
                return false;
            }
            if (dest_dir.Length > 0)
            {
                if (dest_dir[0] == '.')
                {
                    error_msg = String.Format("The destination directory must be an absolute path, relative paths are not valid: {0}", dest_dir);
                    return false;
                }
                if (dest_dir.StartsWith(@"\\"))
                {
                    if (dest_dir.ToUpper() == src_dir.ToUpper())
                    {
                        // The source and destination are on a network path and they are the same so staging will have no effect
                        nodes = null;
                        return true;
                    }
                    // Use only one node since multiple nodes will only clobber the files in the shared directory
                    use_one_node = true;
                }
            }
            else
            {
                dest_dir = src_dir;
                if (dest_dir.StartsWith(@"\\"))
                {
                    // The source and destination are on a network path and they are the same so staging will have no effect
                    nodes = null;
                    return true;
                }
            }

            // FIXME: Validate the wildcard string?
            if (options.stage_wildcards != null)
                wildcards = options.stage_wildcards.Trim();
            else
                wildcards = "*.*";
            if (wildcards.Length == 0)
                wildcards = "*.*";

            ArrayList args = new ArrayList();
            args.Add(src_dir);
            args.Add(String.Format("/dest_dir:{0}", dest_dir));
            if (options.stage_recursive)
                args.Add("/recursive");
            args.Add("/ignore_src");
            if (options.stage_nopull)
                args.Add("/nopull");
            if (wildcards != "*.*")
                args.Add(String.Format("/wc:{0}", wildcards));

            LaunchNode src_node = new LaunchNode(this);
            src_node.appnum = 0;
            src_node.args = (ArrayList)args.Clone();
            src_node.clique = "0";
            src_node.credential = null;
            src_node.dir = null;
            src_node.env = null;
            src_node.exe = "deino_stager.exe";
            src_node.hostname = Environment.MachineName;
            src_node.iproc = 0;
            src_node.launch_local = true;
            src_node.map = null;
            src_node.nproc = -1; // to be determined
            src_node.path = "%DEINO_MPI_DIR%";
            src_node.phrase = null;
            src_node.pmi = null;
            src_node.port = -1;
            src_node.private_key = ((LaunchNode)launch_nodes[0]).private_key;
            src_node.secure_user = null;
            src_node.spawned = false;
            nodes = new ArrayList();
            nodes.Add(src_node);
            if (use_one_node)
            {
                src_node.nproc = 2;
                LaunchNode node = ((LaunchNode)launch_nodes[0]).Clone();
                node.appnum = 0;
                node.args = (ArrayList)args.Clone();
                node.clique = "1";
                node.exe = "deino_stager.exe";
                node.iproc = 1;
                node.nproc = 2;
                node.path = "%DEINO_MPI_DIR%";
                node.env = null;
                nodes.Add(node);
            }
            else
            {
                int i, nproc = ((LaunchNode)launch_nodes[0]).nproc;
                int[] node_root_ranks = new int[nproc];
                for (i = 0; i < nproc; i++)
                    node_root_ranks[i] = -1;
                bool all_ranks_covered = false;
                int lowest_not_covered = 0;
                while (!all_ranks_covered)
                {
                    all_ranks_covered = true;
                    for (i = lowest_not_covered; i < nproc; i++)
                    {
                        if (node_root_ranks[i] == -1)
                        {
                            lowest_not_covered = i;
                            all_ranks_covered = false;
                            int[] clique_ranks = ProcessGroup.ParseCliqueInts(((LaunchNode)launch_nodes[i]).clique);
                            if (clique_ranks != null)
                            {
                                foreach (int clique_rank in clique_ranks)
                                {
                                    node_root_ranks[clique_rank] = clique_ranks[0]; // This assumes that the first clique rank is the lowest
                                }
                            }
                            break;
                        }
                    }
                }
                Hashtable node_ranks_hash = new Hashtable();
                for (i = 0; i < nproc; i++)
                {
                    node_ranks_hash[node_root_ranks[i]] = node_root_ranks[i];
                }
                int[] node_ranks = new int[node_ranks_hash.Count];
                i = 0;
                foreach (int node_rank in node_ranks_hash.Keys)
                    node_ranks[i++] = node_rank;
                Array.Sort(node_ranks);

                src_node.nproc = node_ranks.Length + 1;
                for (i = 0; i < node_ranks.Length; i++)
                {
                    LaunchNode node = ((LaunchNode)launch_nodes[node_ranks[i]]).Clone();
                    node.appnum = 0;
                    node.args = (ArrayList)args.Clone();
                    node.clique = String.Format("{0}", i + 1);
                    node.exe = "deino_stager.exe";
                    node.iproc = i + 1;
                    node.nproc = node_ranks.Length + 1;
                    node.path = "%DEINO_MPI_DIR%";
                    node.env = null;
                    nodes.Add(node);
                }
            }

            //MessageBox.Show(String.Format("Staging:\r\n Source: {0}{1}{2}\r\n Destination: {3}", src_dir, src_dir.EndsWith("\\") ? "" : "\\", wildcards, dest_dir));
            return true;
        }

        #region Output functions
        public void PrintExitcodes()
        {
            foreach (ProcessGroup pg in process_groups)
            {
                pg.PrintExitCodes();
            }
        }

        public void WriteLine()
        {
            if (global_options.use_output_queue)
                global_options.output_queue.Enqueue("\n");
            else
                Console.WriteLine();
        }

        public void WriteLine(string str)
        {
            if (global_options.use_output_queue)
            {
                global_options.output_queue.Enqueue(str + "\n");
            }
            else
            {
                Console.WriteLine(str);
            }
        }

        public void WriteLine(string format, params object[] args)
        {
            if (global_options.use_output_queue)
            {
                string str = String.Format(format, args);
                global_options.output_queue.Enqueue(str + "\n");
            }
            else
            {
                Console.WriteLine(format, args);
            }
        }

        public void Write(string format, params object[] args)
        {
            if (global_options.use_output_queue)
            {
                string str = String.Format(format, args);
                global_options.output_queue.Enqueue(str);
            }
            else
            {
                Console.Write(format, args);
            }
        }

        public void Write(string str)
        {
            if (global_options.use_output_queue)
                global_options.output_queue.Enqueue(str);
            else
                Console.Write(str);
        }

        public void Write(char[] char_array)
        {
            if (global_options.use_output_queue)
                global_options.output_queue.Enqueue(new string(char_array));
            else
                Console.Write(char_array);
        }

        public void WriteError(char[] char_array)
        {
            if (global_options.use_output_queue)
                global_options.output_queue.Enqueue(new string(char_array));
            else
                Console.Error.Write(char_array);
        }
        #endregion

        #region Credential store console functions
        string GetCachedPassword()
        {
            // Insert password caching code here:
            return null;
        }

        void DeleteCachedPassword()
        {
        }

        static string GetPassword(string message)
        {
            StringBuilder entered_password = new StringBuilder();
            Console.WriteLine(message);

            try
            {
                // get the first character of the password
                ConsoleKeyInfo nextKey = Console.ReadKey(true);

                while (nextKey.Key != ConsoleKey.Enter)
                {
                    if (nextKey.Key == ConsoleKey.Backspace)
                    {
                        if (entered_password.Length > 0)
                        {
                            entered_password.Remove(entered_password.Length - 1, 1);

                            // erase the last * as well
                            Console.Write(nextKey.KeyChar);
                            Console.Write(" ");
                            Console.Write(nextKey.KeyChar);
                        }
                    }
                    else
                    {
                        entered_password.Append(nextKey.KeyChar);
                        Console.Write("*");
                    }

                    nextKey = Console.ReadKey(true);
                }
            }
            catch (InvalidOperationException)
            {
                // The console has been redirected so read a line from the redirected input
                entered_password = new StringBuilder(Console.In.ReadLine());
            }

            Console.WriteLine();

            return entered_password.ToString();
        }

        public int GetStore(ArrayList nodes)
        {
            if (store == null)
            {
                // setup the Credential store
                if (!CredentialStore.Exists())
                {
                    string location = CredentialStore.GetLocation();
                    if (CredentialStore.IsRemovableMedia())
                    {
                        Console.WriteLine("Error encountered:");
                        Console.WriteLine("Please insert the media containing the Credential store and try again: {0}", location);
                    }
                    else
                    {
                        if (location != null)
                        {
                            Console.WriteLine("Error encountered:");
                            Console.WriteLine("The Credential store is not available: {0}", location);
                            Console.WriteLine("This can be a result of the store being on a removable media device that is not plugged in or it could be that the store has been deleted.");
                            Console.WriteLine("Please select a valid Credential store or create a new one and then try again.");
                        }
                        else
                        {
                            Console.WriteLine("Error encountered:");
                            Console.WriteLine("No Credential store available.  Please set up a Credential store before running mpiexec.");
                        }
                    }
                    mpiexec_exit_code = -1;
                    return mpiexec_exit_code;
                }
                string password = null;
                if (CredentialStore.RequiresPassword())
                {
                    password = GetCachedPassword();
                    if (password != null)
                    {
                        try
                        {
                            store = new CredentialStore(password);
                        }
                        catch
                        {
                            DeleteCachedPassword();
                            store = null;
                        }
                    }
                    if (store == null)
                    {
                        password = GetPassword("Please enter the passphrase used to protect the keys in the Credential Store:");
                    }
                }
                try
                {
                    if (store == null)
                        store = new CredentialStore(password);
                }
                catch (Exception e)
                {
                    Console.WriteLine("Unable to open the Credential Store: {0}", e.Message);
                    mpiexec_exit_code = -1;
                    return mpiexec_exit_code;
                }

                //this.store = store;
            }
            if (nodes != null)
            {
                // use the same private key in all client connections
                foreach (LaunchNode node in nodes)
                {
                    node.private_key = store.Keys;
                }
            }
            return 0;
        }

        public int GetStoreAndCredential(ArrayList nodes)
        {
            return GetStoreAndCredential(store, nodes);
        }

        public int GetStoreAndCredential(CredentialStore store, ArrayList nodes)
        {
            int result = GetStore(nodes);
            if (result != 0)
                return result;
            if (store == null)
                store = this.store;

            SecureCredential login_credential = null;
            if (global_options.logon)
            {
                Console.WriteLine("In order to launch processes a user account and password must be provided.");
                if (global_options.noprompt)
                {
                    Console.WriteLine("Error: -noprompt prevents user input");
                    mpiexec_exit_code = -1;
                    return mpiexec_exit_code;
                }
                Console.WriteLine("Account:");
                login_credential = new SecureCredential(store.Keys);
                login_credential.Account = Console.ReadLine().Trim();
                login_credential.Password = GetPassword("Password:");
            }

            if (nodes != null)
            {
                // use the same private key in all client connections
                foreach (LaunchNode node in nodes)
                {
                    //node.private_key = store.Keys; // already set by GetStore()

                    if (login_credential != null)
                    {
                        node.credential = login_credential;
                    }
                    else if (node.secure_user != null)
                    {
                        //Console.WriteLine("Looking up user '{0}' in the credential store.", node.secure_user);
                        foreach (SecureCredential cred in store)
                        {
                            if (String.Compare(cred.Account, node.secure_user, true) == 0)
                            {
                                //Console.WriteLine("Found user {0}", cred.Account);
                                node.credential = cred;
                                break;
                            }
                        }
                        if (node.credential == null)
                        {
                            Console.WriteLine("Error, account not found in the Credential Store: {0}", node.secure_user);
                            mpiexec_exit_code = -1;
                            return mpiexec_exit_code;
                        }
                    }
                    else if (global_options.username != null)
                    {
                        node.credential = new SecureCredential(store.Keys);
                        node.credential.Account = global_options.username;
                        node.credential.Password = global_options.password;
                    }

                    if (node.credential == null)
                    {
                        node.credential = store.DefaultCredential;
                    }
                }
                SecureCredential temp_credential = null;
                foreach (LaunchNode node in nodes)
                {
                    if (node.credential == null)
                    {
                        if (temp_credential == null)
                        {
                            Console.WriteLine("The Credential store does not contain any user account information.");
                            Console.WriteLine("In order to launch processes a user account and password must be provided.");
                            if (global_options.noprompt)
                            {
                                mpiexec_exit_code = -1;
                                return mpiexec_exit_code;
                            }
                            Console.WriteLine("Account:");
                            temp_credential = new SecureCredential(store.Keys);
                            temp_credential.Account = Console.ReadLine().Trim();
                            temp_credential.Password = GetPassword("Password:");
                        }
                        node.credential = temp_credential;
                    }
                }
            }
            return 0;
        }
        #endregion

        #region MPIEXEC argument parsing section

        public void GetDefaultSettings()
        {
            try
            {
                IDictionary vars = Environment.GetEnvironmentVariables();
                foreach (string env in vars.Keys)
                {
                    switch (env.ToLower())
                    {
                        case "deinompi_timeout":
                            // global only option
                            try
                            {
                                int number = Convert.ToInt32((string)vars[env]);
                                if (number > 0)
                                    global_options.timeout = number;
                            }
                            catch
                            {
                                // Bad number provided. Silently ignore the error?
                            }
                            break;
                        case "mpiexec_timeout":
                            goto case "deinompi_timeout";
                        case "deinompi_channel":
                            global_options.channel = (string)vars[env];
                            break;
                        case "deinompi_localonly":
                            global_options.localonly = true;
                            break;
                        case "mpiexec_localonly":
                            goto case "deinompi_localonly";
                        case "deinompi_noprompt":
                            // global only option
                            global_options.noprompt = true;
                            break;
                        case "mpiexec_noprompt":
                            goto case "deinompi_noprompt";
                        case "deinompi_host":
                            goto case "deinompi_hosts";
                        case "deinompi_hosts":
                            // global only option
                            GetHostsFromString((string)vars[env]);
                            break;
                        case "deinompi_machinefile":
                            // global only option
                            ParseMachinefile((string)vars[env]);
                            break;
                        case "deinompi_priority":
                            string[] tokens;
                            tokens = ((string)vars[env]).Split(':');
                            try
                            {
                                int pc, pt;
                                pc = Convert.ToInt32(tokens[0]);
                                if (tokens.Length > 1)
                                {
                                    pt = Convert.ToInt32(tokens[1]);
                                    if (pt >= 0 && pt <= 5)
                                        global_options.priority_thread = pt;
                                }
                                if (pc >= 0 && pc <= 4)
                                    global_options.priority_class = pc;
                            }
                            catch //(Exception e)
                            {
                            }
                            break;
                    }
                }
            }
            catch
            {
                // Don't let bad environment variables stop jobs from running?
            }
        }

        static public string[] ParseStrings(string str)
        {
            if (str == null)
                return null;
            ArrayList list = new ArrayList();
            int token_end;
            str = str.Trim();
            if (str == null) // does Trim() return null or an empty string?
                return null;
            token_end = 0;
            while (token_end < str.Length)
            {
                if (str[token_end] == '\"')
                {
                    token_end++; // skip over the quote character
                    StringBuilder sb = new StringBuilder();
                    while (token_end < str.Length && str[token_end] != '\"')
                    {
                        if (str[token_end] == '\\' && token_end < (str.Length - 1))
                        {
                            if (str[token_end + 1] == '\"' || str[token_end + 1] == '\\')
                                token_end++;
                        }
                        sb.Append(str[token_end]);
                        token_end++;
                    }
                    list.Add(sb.ToString());
                    token_end++; // skip over the quote character
                    while (token_end < str.Length && str[token_end] == ' ')
                        token_end++; // skip over spaces
                }
                else
                {
                    StringBuilder sb = new StringBuilder();
                    while (token_end < str.Length && str[token_end] != ' ')
                    {
                        sb.Append(str[token_end]);
                        token_end++;
                    }
                    list.Add(sb.ToString());
                    while (token_end < str.Length && str[token_end] == ' ')
                        token_end++; // skip over spaces
                }
            }
            if (list.Count > 0)
            {
                string[] ret_val = new string[list.Count];
                for (int i = 0; i < list.Count; i++)
                    ret_val[i] = (string)list[i];
                return ret_val;
            }
            return null;
        }

        string GetPath(string exe)
        {
            FileInfo info;
            if (exe == null || exe.Length == 0)
                return null;
            if (exe[0] == '\\' || (exe.Length > 1 && exe[1] == ':'))
            {
                exe = DriveMapping.ExeToUNC(exe);
                // full path specified
                info = new FileInfo(exe);
                return info.DirectoryName;
            }
            info = new FileInfo(exe);
            if (info.Exists)
            {
                exe = DriveMapping.ExeToUNC(info.FullName);
                info = new FileInfo(exe);
                return info.DirectoryName;
            }
            string[] extensions = Environment.GetEnvironmentVariable("PATHEXT").Split(';');
            if (extensions == null)
                extensions = new string[] { ".COM", ".EXE", ".BAT", ".CMD", ".VBS", ".VBE", ".JS", ".JSE", ".WSF", ".WSH" };
            foreach (string extension in extensions)
            {
                info = new FileInfo(exe + extension);
                if (info.Exists)
                {
                    exe = DriveMapping.ExeToUNC(info.FullName);
                    info = new FileInfo(exe);
                    return info.DirectoryName;
                }
            }
            if (exe.IndexOf('\\') > 0)
            {
                // relative path specified but file does not exist
                throw new Exception("The specified file does not exist: " + exe);
            }
            string path = Environment.GetEnvironmentVariable("PATH");
            if (path != null)
            {
                bool add_exe = true;
                exe = "\\" + exe;
                int exe_index = exe.IndexOf(".exe");
                if (exe_index == -1)
                    exe_index = exe.IndexOf(".EXE");
                if ((exe_index != -1) && (exe_index == (exe.Length - 4)))
                {
                    add_exe = false;
                }
                string[] dirs = path.Split(';');
                foreach (string cdir in dirs)
                {
                    // Put this in a try block to catch invalid path strings and ignore them
                    try
                    {
                        string dir = cdir.Replace("\"", ""); // FileInfo does not accept quotes in the path
                        info = new FileInfo(dir + exe);
                        if (info.Exists)
                        {
                            exe = DriveMapping.ExeToUNC(info.FullName);
                            info = new FileInfo(exe);
                            return info.DirectoryName;
                        }
                        if (add_exe)
                        {
                            foreach (string extension in extensions)
                            {
                                info = new FileInfo(dir + exe + extension);
                                if (info.Exists)
                                {
                                    exe = DriveMapping.ExeToUNC(info.FullName);
                                    info = new FileInfo(exe);
                                    return info.DirectoryName;
                                }
                            }
                        }
                    }
                    catch
                    {
                    }
                }
            }
            return null;
        }

        string GetExecutable(string exe)
        {
            FileInfo info = new FileInfo(exe);
            //Console.WriteLine("exe: {0}", exe);
            //Console.WriteLine("unc: {0}", DriveMapping.ExeToUNC(exe));
            return info.Name;
        }

        void ParseCliques(ArrayList launch_nodes)
        {
            // a clique is a range of ranks with the same hostname
            int cur_iproc, printed_iproc;
            LaunchNode cur_node, iter;
            for (int i = 0; i < launch_nodes.Count; i++)
            {
                cur_node = (LaunchNode)launch_nodes[i];
                if (cur_node.clique == null)
                {
                    printed_iproc = cur_iproc = cur_node.iproc;
                    StringBuilder str = new StringBuilder();
                    str.AppendFormat("{0}", cur_iproc);
                    for (int j = i + 1; j < launch_nodes.Count; j++)
                    {
                        iter = (LaunchNode)launch_nodes[j];
                        if (iter.hostname == cur_node.hostname)
                        {
                            if (iter.iproc == cur_iproc + 1)
                            {
                                cur_iproc = iter.iproc;
                            }
                            else
                            {
                                if (printed_iproc == cur_iproc)
                                {
                                    str.AppendFormat(",{0}", iter.iproc);
                                }
                                else
                                {
                                    str.AppendFormat("..{0},{1}", cur_iproc, iter.iproc);
                                }
                                printed_iproc = cur_iproc = iter.iproc;
                            }
                        }
                    }
                    if (printed_iproc != cur_iproc)
                    {
                        str.AppendFormat("..{0}", cur_iproc);
                    }
                    cur_node.clique = str.ToString();
                    for (int j = i + 1; j < launch_nodes.Count; j++)
                    {
                        if (((LaunchNode)launch_nodes[j]).hostname == cur_node.hostname)
                            ((LaunchNode)launch_nodes[j]).clique = str.ToString();
                    }
                }
            }
        }

        void SetDefaultHosts(string[] hosts)
        {
            try
            {
                RegistryKey reg;
                string line = String.Join(" ", hosts);
                reg = Registry.LocalMachine.CreateSubKey(@"SOFTWARE\Deino\DeinoPM");
                if (reg != null)
                    reg.SetValue("hosts", line);
            }
            catch (Exception e)
            {
                Console.WriteLine("Unable to save the host list to the machine registry:");
                Console.WriteLine("{0}", e.Message);
            }
        }

        void SetMyDefaultHosts(string[] hosts)
        {
            try
            {
                RegistryKey reg;
                string line = String.Join(" ", hosts);
                reg = Registry.CurrentUser.CreateSubKey(@"SOFTWARE\Deino\DeinoPM");
                if (reg != null)
                    reg.SetValue("hosts", line);
            }
            catch (Exception e)
            {
                Console.WriteLine("Unable to save the host list to the current user's registry:");
                Console.WriteLine("{0}", e.Message);
            }
        }

        string[] GetDefaultHostsStrings()
        {
            return GetMyDefaultHostsStrings();
        }

        string[] GetMyDefaultHostsStrings()
        {
            RegistryKey reg;
            string line;
            try
            {
                // First try to get the current user's setting
                reg = Registry.CurrentUser.OpenSubKey(@"SOFTWARE\Deino\DeinoPM");
                if (reg != null)
                {
                    line = (string)reg.GetValue("hosts");
                    line = line.Trim();
                    if (line.Length < 1)
                        line = null;
                    reg.Close();
                }
                else
                    line = null;

                // If the current user does not have a setting then try to get the machine setting
                if (line == null)
                {
                    reg = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Deino\DeinoPM");
                    if (reg != null)
                    {
                        line = (string)reg.GetValue("hosts");
                        line = line.Trim();
                        if (line.Length < 1)
                            line = null;
                        reg.Close();
                    }
                    else
                        line = null;
                }
            }
            catch
            {
                line = null;
            }
            if (line == null)
            {
                string[] ret_val = new string[1];
                ret_val[0] = Environment.MachineName;
                return ret_val;
            }
            line = line.Trim();
            string[] hosts = line.Split(' ');
            if (hosts == null || hosts.Length < 1)
            {
                string[] ret_val = new string[1];
                ret_val[0] = Environment.MachineName;
                return ret_val;
            }
            return hosts;
        }

        void GetDefaultHosts()
        {
            RegistryKey reg;
            string line;
            try
            {
                // First try to get the current user's setting
                reg = Registry.CurrentUser.OpenSubKey(@"SOFTWARE\Deino\DeinoPM");
                if (reg != null)
                {
                    line = (string)reg.GetValue("hosts");
                    line = line.Trim();
                    if (line.Length < 1)
                        line = null;
                    reg.Close();
                }
                else
                    line = null;

                // If the current user does not have a setting then try to get the machine setting
                if (line == null)
                {
                    reg = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Deino\DeinoPM");
                    if (reg != null)
                    {
                        line = (string)reg.GetValue("hosts");
                        line = line.Trim();
                        if (line.Length < 1)
                            line = null;
                        reg.Close();
                    }
                    else
                        line = null;
                }
            }
            catch
            {
                line = null;
            }
            global_options.cur_host = 0;
            global_options.cur_processor = 0;
            if (line == null)
            {
                global_options.hosts.Add(new HostNode(Environment.MachineName, 1));
                return;
            }
            line = line.Trim();
            string[] hosts = line.Split(' ');
            foreach (string host in hosts)
            {
                string[] parts = host.Split(':');
                if (parts.Length > 1)
                {
                    int n;
                    try
                    {
                        n = Convert.ToInt32(parts[1]);
                    }
                    catch (Exception)
                    {
                        n = 1;
                    }
                    global_options.hosts.Add(new HostNode(parts[0], n));
                }
                else
                {
                    global_options.hosts.Add(new HostNode(host, 1));
                }
            }
            if (global_options.hosts.Count == 0)
            {
                global_options.hosts.Add(new HostNode(Environment.MachineName, 1));
            }
        }

        void GetHostsFromStringOrDefault(string hosts_str)
        {
            if (hosts_str == null)
            {
                GetDefaultHosts();
                return;
            }
            hosts_str = hosts_str.Trim();
            if (hosts_str.Length < 1)
            {
                GetDefaultHosts();
                return;
            }
            global_options.cur_host = 0;
            global_options.cur_processor = 0;
            string[] hosts = hosts_str.Split(' ');
            foreach (string host in hosts)
            {
                string[] parts = host.Split(':');
                if (parts.Length > 1)
                {
                    int n;
                    try
                    {
                        n = Convert.ToInt32(parts[1]);
                    }
                    catch (Exception)
                    {
                        n = 1;
                    }
                    global_options.hosts.Add(new HostNode(parts[0], n));
                }
                else
                {
                    global_options.hosts.Add(new HostNode(host, 1));
                }
            }
            if (global_options.hosts.Count == 0)
            {
                global_options.hosts.Add(new HostNode(Environment.MachineName, 1));
            }
        }

        void GetHostsFromString(string hosts_str)
        {
            if (hosts_str == null)
                return;
            hosts_str = hosts_str.Trim();
            if (hosts_str.Length < 1)
                return;
            ArrayList hosts = new ArrayList();
            string[] hosts_tokens = hosts_str.Split(' ');
            foreach (string host in hosts_tokens)
            {
                string[] parts = host.Split(':');
                if (parts.Length > 1)
                {
                    int n;
                    try
                    {
                        n = Convert.ToInt32(parts[1]);
                        if (n < 1)
                            n = 1;
                    }
                    catch (Exception)
                    {
                        n = 1;
                    }
                    hosts.Add(new HostNode(parts[0], n));
                }
                else
                {
                    hosts.Add(new HostNode(host, 1));
                }
            }
            if (hosts.Count > 0)
            {
                global_options.cur_host = 0;
                global_options.cur_processor = 0;
                for (int i = 0; i < hosts.Count; i++)
                    global_options.hosts.Add(hosts[i]);
            }
        }

        private bool IsNumber(string str)
        {
            int number;
            try
            {
                number = Convert.ToInt32(str);
            }
            catch (Exception)
            {
                return false;
            }
            return true;
        }

        public void PrintOptions(bool extended)
        {
            // Basic help
            WriteLine();
            WriteLine("Usage:");
            WriteLine("mpiexec -n <maxprocs> [options] executable [args ...]");
            WriteLine("mpiexec [options] executable [args ...] : [options] exe [args] : ...");
            WriteLine("mpiexec -configfile <configfile>");
            WriteLine();
            WriteLine("options:");
            WriteLine();
            WriteLine("standard:");
            WriteLine("-n <maxprocs>");
            WriteLine("-wdir <working directory>");
            WriteLine("-configfile <filename> -");
            WriteLine("       each line contains a complete set of mpiexec options");
            WriteLine("       including the executable and arguments");
            WriteLine("-host <hostname>");
            /*
            WriteLine("-soft <Fortran90 triple> - acceptable number of processes up to maxprocs");
            WriteLine("       a or a:b or a:b:c where");
            WriteLine("       1) a = a");
            WriteLine("       2) a:b = a, a+1, a+2, ..., b");
            WriteLine("       3) a:b:c = a, a+c, a+2c, a+3c, ..., a+kc");
            WriteLine("          where a+kc <= b if c>0");
            WriteLine("                a+kc >= b if c<0");
            */
            WriteLine("-path <search path for executable, ; separated>");
            WriteLine();
            WriteLine("extensions:");
            WriteLine("-env <variable value>");
            WriteLine("-hosts <n host1 host2 ... hostn>");
            WriteLine("-hosts <n host1 m1 host2 m2 ... hostn mn>");
            WriteLine("-machinefile <filename> - one host per line, #commented");
            WriteLine("-localonly <numprocs>");
            WriteLine("-exitcodes - print the exit codes of processes as they exit");
            WriteLine("-genvlist <list of env var names a,b,c,...> - pass current values of these vars");
            WriteLine("-g<local arg name> - global version of local options");
            WriteLine("  genv, gwdir, ghost, gpath, gmap");
            //WriteLine("-file <filename> - old mpich1 job configuration file");
            WriteLine();
            WriteLine("examples:");
            WriteLine("mpiexec -n 4 cpi");
            WriteLine("mpiexec -n 1 -host foo master : -n 8 worker");
            WriteLine();

            if (extended)
            {
                // Full option listing
                WriteLine();
                WriteLine("All options to mpiexec:");
                WriteLine();
                WriteLine("-n x");
                WriteLine("-np x");
                WriteLine("  Launch x processes.");
                WriteLine("-localonly x");
                WriteLine("-n x -localonly");
                WriteLine("  Launch x processes on the local machine.");
                WriteLine("-machinefile filename");
                WriteLine("  Use a file to list the names of machines to launch on.");
                WriteLine("-host hostname");
                WriteLine("-hosts n host1 host2 ... hostn");
                WriteLine("-hosts n host1 m1 host2 m2 ... hostn mn");
                WriteLine("  Launch on the specified hosts.");
                WriteLine("  In the second version the number of processes = m1 + m2 + ... + mn.");
                WriteLine("-map drive:\\\\host\\share");
                WriteLine("  Map a drive on all the nodes.");
                WriteLine("  The mapping will be removed when the processes exit.");
                WriteLine("-dir drive:\\my\\working\\directory");
                WriteLine("-wdir drive:\\my\\working\\directory");
                WriteLine("  Launch processes in the specified directory.");
                WriteLine("-env var val");
                WriteLine("  Set an environment variable before launching the processes.");
                WriteLine("-logon");
                WriteLine("  Prompt for user account and password.");
                WriteLine("-pwdfile filename");
                WriteLine("  Read the account and password from the file specified.");
                WriteLine("  Put the account on the first line and the password on the second.");
                WriteLine("-user <user name>");
                WriteLine("  Use the specified user account from the Credential store to launch the job.");
                //WriteLine("-nopopup_debug");
                //WriteLine("  Disable the system popup dialog if the process crashes.");
                WriteLine("-exitcodes");
                WriteLine("  Print the process exit codes when each process exits.");
                WriteLine("-noprompt");
                WriteLine("  Prevent mpiexec from prompting for user credentials.");
                WriteLine("-priority class[:level]");
                WriteLine("  Set the process startup priority class and optionally level.");
                WriteLine("  class = 0,1,2,3,4   = idle, below, normal, above, high");
                WriteLine("  level = 0,1,2,3,4,5 = idle, lowest, below, normal, above, highest");
                WriteLine("  The default is -priority 2:3");
                WriteLine("-localroot");
                WriteLine("  Launch the root process directly from mpiexec if the host is local.");
                WriteLine("  (This allows the root process to create windows and be debugged.)");
                WriteLine("-path search_path");
                WriteLine("  Specify the search path for the executable, ; separated");
                WriteLine("-timeout seconds");
                WriteLine("  Specify the timeout for the job in seconds.");
                /*
                WriteLine("-delegate");
                WriteLine("  Use passwordless delegation to launch processes.");
                WriteLine("-impersonate");
                WriteLine("  Use passwordless authentication to launch processes.");
                */
                WriteLine("-log");
                WriteLine("  Select the MPE logging wrapper to generate a Jumpshot viewable log file.");
                WriteLine("-set_hosts hostA hostB:N hostC ...");
                WriteLine("  Set the default hosts and optionally the number of processes per host.");
                WriteLine("  This is a machine wide setting that affects all DeinoMPI users.");
                WriteLine("-set_myhosts hostA hostB:N hostC ...");
                WriteLine("  Set the default hosts and optionally the number of processes per host.");
                WriteLine("  This setting only affects the current user.");
                WriteLine("-validate");
                WriteLine("  Validate that the job can run without actually starting the job.");
                WriteLine("  Add this flag along with all the rest of the mpiexec command line.");
                WriteLine("  The hosts are contacted to determine if the job can run but executables");
                WriteLine("  are not started.");
                WriteLine("-stage path");
                WriteLine("  Specify the source directory you want to copy out to all the nodes before");
                WriteLine("  the job is started.");
                WriteLine("  Example: -stage c:\\data\\myjob or -stage \\\\myserver\\myshare\\data");
                WriteLine("-stage_dest path");
                WriteLine("  Specify the destination directory where you want the staged directory to be");
                WriteLine("  copied to.  If this option is not specified then the source directory is used");
                WriteLine("  as the destination.");
                WriteLine("  Example: mpiexec -stage \\\\myserver\\myshare\\data -stage_dest c:\\temp\\data ...");
                WriteLine("-stage_recursive");
                WriteLine("  Specify this flag to indicate that you want the source directory to be");
                WriteLine("  copied and all of its sub-directories also.");
                WriteLine("-stage_wildcards pattern");
                WriteLine("  Specify a wild-card pattern to limit which files are copied to the worker");
                WriteLine("  nodes when staging a directory.  Only files that match this pattern will");
                WriteLine("  be copied.");
                WriteLine("  Example: mpiexec -stage_wildcards *.txt -stage c:\\temp\\data ...");
                WriteLine("-stage_nopull");
                WriteLine("  Specify this flag to indicate that you do not want any modified files");
                WriteLine("  copied back to the source directory after the job has finished.  The default");
                WriteLine("  behavior is for new or modified files in the staging destination directory to");
                WriteLine("  be copied back to the source directory after the job has completed.  This");
                WriteLine("  flag overrides that option.");
                WriteLine("-version [hostname]");
                WriteLine("  Print the version of DeinoMPI installed on the specified machine.  If no host");
                WriteLine("  is specified then print the version installed on the local machine.");
            }
            else
            {
                WriteLine("For a list of all mpiexec options, execute 'mpiexec -help2'");
            }
        }

        void ParseMachinefile(string filename)
        {
            using (StreamReader sr = new StreamReader(filename))
            {
                String line;

                while ((line = sr.ReadLine()) != null)
                {
                    // ignore empty lines, all whitespace, or starting with the # character
                    line = line.Trim();
                    if (line.Length > 0 && line[0] != '#')
                    {
                        HostNode node;
                        string[] tokens = line.Split(' ');
                        try
                        {
                            node = new HostNode(tokens[0]);
                        }
                        catch (Exception e)
                        {
                            throw new Exception("Invalid line in the machinefile, '" + line + "', " + e.Message);
                        }
                        if (tokens.Length > 1)
                        {
                            for (int i = 1; i < tokens.Length; i++)
                            {
                                if (tokens[i] == "-ifhn")
                                {
                                    if (i + 1 < tokens.Length)
                                    {
                                        node.alt_hostname = tokens[i + 1];
                                    }
                                }
                                // add future options here
                            }
                        }
                        global_options.hosts.Add(node);
                    }
                }
                sr.Close();
            }
            if (global_options.hosts.Count > 0)
            {
                global_options.cur_host = 0;
                global_options.cur_processor = 0;
            }
            /*
            foreach (HostNode node in global_options.hosts)
            {
                mpiexec.WriteLine("host: {0}\nnproc: {1}", node.hostname, node.nproc);
            }
            */
        }

        void ParseCredentials(string filename)
        {
            using (StreamReader sr = new StreamReader(filename))
            {
                global_options.username = sr.ReadLine();
                global_options.password = sr.ReadLine();
                sr.Close();
            }
        }

        ArrayList ParseConfigfile(string filename)
        {
            ArrayList list = new ArrayList();
            using (StreamReader sr = new StreamReader(filename))
            {
                String line;

                while ((line = sr.ReadLine()) != null)
                {
                    line = line.Trim();
                    if (line.Length > 0 && line[0] != '#')
                    {
                        string[] tokens;
                        tokens = ParseStrings(line);
                        list.Add(tokens);
                    }
                }
                sr.Close();
            }
            return list;
        }

        ArrayList ParseFile(string filename)
        {
            ArrayList list = new ArrayList();
            using (StreamReader sr = new StreamReader(filename))
            {
                String line;

                while ((line = sr.ReadLine()) != null)
                {
                    line = line.Trim();
                    if (line.Length > 0 && line[0] != '#')
                    {
                        // add code here to parse the old configfile format
                        string[] tokens;
                        tokens = ParseStrings(line);
                        list.Add(tokens);
                    }
                }
                sr.Close();
            }
            return list;
        }

        ArrayList ParseSet(string[] args, int appnum)
        {
            ArrayList nodes = new ArrayList();
            Options options = new Options();
            options.appnum = appnum;

            // Set the defaults from the global options that can then be overridden by local flags
            options.localonly = global_options.localonly;
            options.priority_class = global_options.priority_class;
            options.priority_thread = global_options.priority_thread;
            options.channel = global_options.channel;

            for (int i = 0; i < args.Length; i++)
            {
                #region Option parsing section
                if (args[i][0] == '-')
                {
                    string option = args[i].TrimStart('-');
                    switch (option)
                    {
                        case "configfile":
                            throw new Exception("The -configfile option must be the first and only option specified");
                        case "file":
                            throw new Exception("The -file option must be the first and only option specified");
                        case "n":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No number specified after the -{0} option", option));
                            if (options.nproc != -1)
                                throw new Exception("Only one option is allowed to determine the number of processes, -hosts, -n, -np and -localonly x are mutually exclusive");
                            try
                            {
                                options.nproc = Convert.ToInt32(args[i]);
                            }
                            catch (Exception e)
                            {
                                throw new Exception(String.Format("Unable to interpret the size argument after the -{0} option: {1}", option, e.Message));
                            }
                            break;
                        case "np":
                            goto case "n";
                        case "localonly":
                            global_options.localonly = true;
                            options.localonly = true;
                            if (i + 1 != args.Length)
                            {
                                int nproc;
                                try
                                {
                                    nproc = Convert.ToInt32(args[i + 1]);
                                }
                                catch (Exception)
                                {
                                    // ignore exception because -localonly can be a flag or a size option
                                    options.hosts.Clear();
                                    options.hosts.Add(new HostNode(System.Environment.MachineName, -1));
                                    break;
                                }
                                if (options.nproc != -1)
                                    throw new Exception("Only one option is allowed to determine the number of processes, -hosts, -n, -np and -localonly x are mutually exclusive.");
                                if (nproc < 1)
                                    throw new Exception("Invalid size parameter following the -localonly option.");
                                if (options.hosts.Count > 0)
                                    throw new Exception("Only one options is allowed to determine the location of of processes, -localonly and -host are mutually exclusive.");
                                options.nproc = nproc;
                                i++;
                            }
                            options.hosts.Clear();
                            options.hosts.Add(new HostNode(System.Environment.MachineName, -1));
                            break;
                        case "machinefile":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No filename specified after the -machinefile option");
                            ParseMachinefile(args[i]);
                            break;
                        case "map":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No argument specified after the -{0} option", option));
                            if (args[i].Length < 3 || args[i][1] != ':')
                                throw new Exception(String.Format("Invalid argument specified after the -{0} option: {1}", option, args[i]));
                            if (option == "map")
                                options.map[args[i].Substring(0, 1)] = args[i].Remove(0, 2);
                            else
                                global_options.map[args[i].Substring(0, 1)] = args[i].Remove(0, 2);
                            break;
                        case "gmap":
                            goto case "map";
                        case "wdir":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No path specified after the -{0} option", option));
                            options.dir = args[i];
                            break;
                        case "dir":
                            goto case "wdir";
                        case "gwdir":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No path specified after the -{0} option", option));
                            global_options.dir = args[i];
                            break;
                        case "gdir":
                            goto case "gwdir";
                        case "env":
                            if (i + 2 == args.Length)
                                throw new Exception(String.Format("No keyval pair specified after the -{0} option", option));
                            if (option == "env")
                                options.env[args[i + 1]] = args[i + 2];
                            else
                                global_options.env[args[i + 1]] = args[i + 2];
                            i += 2;
                            break;
                        case "genv":
                            goto case "env";
                        case "logon":
                            global_options.logon = true;
                            break;
                        case "login":
                            goto case "logon";
                        case "impersonate":
                            global_options.impersonation = true;
                            break;
                        case "delegate":
                            global_options.delegation = true;
                            break;
                        case "pwdfile":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No filename specified after the -pwdfile option");
                            ParseCredentials(args[i]);
                            break;
                        case "user":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No user name specified after the -user option");
                            options.secure_user = args[i];
                            break;
                        case "guser":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No user name specified after the -guser option");
                            global_options.secure_user = args[i];
                            break;
                        case "host":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No hostname specified after the -host option");
                            if (options.localonly)
                                throw new Exception("Either -hosts or -localonly may be specified but not both.  They are mutually exclusive.");
                            options.hosts.Add(new HostNode(args[i], -1));
                            break;
                        case "ghost":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No hostname specified after the -ghost option");
                            global_options.hostname = args[i];
                            options.hosts.Add(new HostNode(args[i], -1));
                            break;
                        case "hosts":
                            int num_hosts;
                            if (options.nproc != -1)
                                throw new Exception("Only one option is allowed to determine the number of processes, -hosts, -n, -np and -localonly x are mutually exclusive");
                            if (options.localonly)
                                throw new Exception("Either -hosts or -localonly may be specified but not both.  They are mutually exclusive.");
                            i++;
                            if (i == args.Length || !IsNumber(args[i]))
                                throw new Exception("No number of hosts specified after the -hosts option");
                            num_hosts = Convert.ToInt32(args[i]);
                            options.nproc = num_hosts;
                            i++;
                            if (i + num_hosts > args.Length)
                                throw new Exception("Not enough hosts specified after the -hosts option");
                            for (int j = 0; j < num_hosts; j++)
                            {
                                if (i == args.Length)
                                    throw new Exception("Not enough hosts specified after the -hosts option");
                                HostNode host_node = new HostNode();
                                host_node.hostname = args[i];
                                i++;
                                if (i < args.Length && IsNumber(args[i]))
                                {
                                    host_node.nproc = Convert.ToInt32(args[i]);
                                    if (host_node.nproc < 1)
                                        throw new Exception(String.Format("Invalid number of processors argument: \"{0}\"", args[i]));
                                    options.nproc += host_node.nproc - 1;
                                    i++;
                                }
                                options.hosts.Add(host_node);
                            }
                            // It's easier to walk off the end and then back up one than to use [i+1] all over the place.
                            i--;
                            break;
                        case "nompi":
                            global_options.nompi = true;
                            break;
                        case "help":
                            PrintOptions(false);
                            break;
                        case "help2":
                            PrintOptions(true);
                            break;
                        case "exitcodes":
                            global_options.exitcodes = true;
                            break;
                        case "localroot":
                            global_options.localroot = true;
                            break;
                        case "priority":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No number specified after the -priority option");
                            string[] tokens;
                            tokens = args[i].Split(':');
                            try
                            {
                                options.priority_class = Convert.ToInt32(tokens[0]);
                                if (tokens.Length > 1)
                                {
                                    options.priority_thread = Convert.ToInt32(tokens[1]);
                                }
                            }
                            catch (Exception e)
                            {
                                throw new Exception(String.Format("Unable to interpret the priority argument, \"{0}\", after the -priority option: {1}", args[i], e.Message));
                            }
                            break;
                        case "p":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No number specified after the -{0} option", option));
                            try
                            {
                                options.port = Convert.ToInt32(args[i]);
                            }
                            catch (Exception e)
                            {
                                throw new Exception(String.Format("Unable to interpret the port number argument after the -{0} option: {1}", option, e.Message));
                            }
                            break;
                        case "port":
                            goto case "p";
                        case "path":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No path specified after the -path option");
                            options.path = args[i];
                            break;
                        case "gpath":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No path specified after the -gpath option");
                            global_options.path = args[i];
                            break;
                        case "noprompt":
                            global_options.noprompt = true;
                            break;
                        case "phrase":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No phrase specified after the -phrase option");
                            options.phrase = args[i];
                            break;
                        case "timeout":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No number specified after the -{0} option", option));
                            try
                            {
                                global_options.timeout = Convert.ToInt32(args[i]);
                                if (global_options.timeout < 1)
                                    throw new Exception(String.Format("Invalid timeout value specified '{0}' - must be greater than zero.", global_options.timeout));
                            }
                            catch (Exception e)
                            {
                                throw new Exception(String.Format("Unable to interpret the timeout argument after the -{0} {1} option: {2}", option, args[i], e.Message));
                            }
                            break;
                        case "hide_console":
                            break;
                        case "quiet_abort":
                            break;
                        case "channel":
                            i++;
                            if (i == args.Length)
                                throw new Exception("No channel specified after the -channel option");
                            options.channel = args[i];
                            break;
                        case "log":
                            global_options.env["MPI_WRAP_DLL_NAME"] = "mpe";
                            break;
                        case "smpd":
                            break;
                        case "validate":
                            global_options.validate = true;
                            break;
                        case "verify": // Allow for "verify" as a synonym for "validate"
                            global_options.validate = true;
                            break;
                        case "stage":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No directory specified after the -{0} option", option));
                            global_options.stage_src = args[i];
                            break;
                        case "stage_dest":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No directory specified after the -{0} option", option));
                            global_options.stage_dest = args[i];
                            break;
                        case "stage_nopull":
                            global_options.stage_nopull = true;
                            break;
                        case "stage_recursive":
                            global_options.stage_recursive = true;
                            break;
                        case "stage_wildcards":
                            i++;
                            if (i == args.Length)
                                throw new Exception(String.Format("No wildcard specified after the -{0} option", option));
                            global_options.stage_wildcards = args[i];
                            break;
                        default:
                            WriteLine("Unknown option -{0}", option);
                            break;
                    }
                }
                else
                {
                    // No more options, the rest of the args are the executable
                    options.exe = GetExecutable(args[i]);
                    if (options.path == null)
                        options.path = GetPath(args[i]);
                    else
                        options.path = options.path + ";" + GetPath(args[i]);
                    for (int j = i + 1; j < args.Length; j++)
                    {
                        options.args.Add(args[j]);
                    }
                    i = args.Length - 1;
                }
                #endregion
            }

            // validate selected options
            if (options.exe == null)
            {
                throw new Exception("No executable specified in this block: " + String.Join(" ", args));
            }
            if (options.nproc == -1)
            {
                throw new Exception("No number of processes option specified: -n, -np, -hosts, or -localonly x");
            }

            // add global options
            // FIXME: This is wrong. global options should be added last after all sections are parsed in case a global option is specified after this block is parsed.
            if (global_options.hostname != null && (options.hosts == null || options.hosts.Count == 0))
            {
                if (options.hosts == null)
                    options.hosts = new ArrayList();
                options.hosts.Add(new HostNode(global_options.hostname, -1));
            }
            if (options.hosts != null && options.hosts.Count > 0 && ((HostNode)(options.hosts[0])).nproc == -1)
            {
                ((HostNode)(options.hosts[0])).nproc = options.nproc;
            }
            if (options.dir == null)
                options.dir = global_options.dir;
            if (global_options.path != null)
            {
                if (options.path == null)
                    options.path = global_options.path;
                else
                    options.path = options.path + ";" + global_options.path;
            }
            if (options.port == -1)
                options.port = global_options.port;
            if (options.phrase == null)
                options.phrase = global_options.phrase;
            if (options.secure_user == null)
                options.secure_user = global_options.secure_user;

            // create a list of launch nodes
            for (int i = 0; i < options.nproc; i++)
            {
                LaunchNode node = new LaunchNode(this);
                // add the environment variables
                node.env = (Hashtable)global_options.env.Clone();
                foreach (string key in options.env.Keys)
                {
                    node.env[key] = options.env[key];
                }
                if (options.channel != null &&
                    options.channel.Length > 0 &&
                    options.channel.ToLower() != "sock" &&
                    options.channel.ToLower() != "auto") // FIXME: Don't support auto channel selection for now
                    node.env["DeinoMPI_CHANNEL"] = options.channel;
                // add the drive mappings
                node.map = (Hashtable)global_options.map.Clone();
                foreach (string drive in options.map.Keys)
                {
                    node.map[drive] = options.map[drive];
                }
                // add the host, initializing if necessary
                if (options.hosts.Count == 0)
                {
                    if (global_options.cur_host == -1)
                    {
                        GetDefaultHosts();
                    }
                }
                if (options.hosts.Count > 0 && options.cur_host == -1)
                {
                    options.cur_host = 0;
                    options.cur_processor = 0;
                }
                if (options.hosts.Count > 0)
                {
                    HostNode host_node;
                    host_node = (HostNode)(options.hosts[options.cur_host]);
                    node.hostname = host_node.hostname;
                    if (host_node.alt_hostname != null)
                    {
                        node.env["DeinoMPI_INTERFACE_HOSTNAME"] = host_node.alt_hostname;
                    }
                    options.cur_processor++;
                    if (options.cur_processor >= host_node.nproc)
                    {
                        options.cur_processor = 0;
                        options.cur_host++;
                        if (options.cur_host >= options.hosts.Count)
                            options.cur_host = 0;
                    }
                }
                else
                {
                    HostNode host_node;
                    host_node = (HostNode)(global_options.hosts[global_options.cur_host]);
                    node.hostname = host_node.hostname;
                    if (host_node.alt_hostname != null)
                    {
                        node.env["DeinoMPI_INTERFACE_HOSTNAME"] = host_node.alt_hostname;
                    }
                    global_options.cur_processor++;
                    if (global_options.cur_processor >= host_node.nproc)
                    {
                        global_options.cur_processor = 0;
                        global_options.cur_host++;
                        if (global_options.cur_host >= global_options.hosts.Count)
                            global_options.cur_host = 0;
                    }
                }
                // add the rest of the options
                node.iproc = global_options.cur_rank++;
                if (node.iproc == 0 && global_options.localroot)
                {
                    node.launch_local = true;
                    // Now turn off the localroot option to ensure that only the first process of the first process group is launched locally.
                    global_options.localroot = false;
                }
                node.appnum = options.appnum;
                node.priority_class = options.priority_class;
                node.priority_thread = options.priority_thread;
                node.dir = options.dir;
                node.path = options.path;
                node.exe = options.exe;
                node.args = options.args;
                node.port = options.port;
                node.phrase = options.phrase;
                node.secure_user = options.secure_user;
                //node.pmi = pmi;
                nodes.Add(node);
                //node.Print();
            }
            return nodes;
        }

        public ArrayList ParseArgs(string[] args)
        {
            // Check for the special case of -validate before any other options.
            // This must be checked first so it doesn't interfere with --configfile or --file
            if (args != null && args.Length > 0)
            {
                if ((String.Compare(args[0].TrimStart('-'), "validate", true) == 0) ||
                    (String.Compare(args[0].TrimStart('-'), "verify", true) == 0))
                {
                    global_options.validate = true;
                    if (args.Length > 1)
                    {
                        string[] args2 = new string[args.Length - 1];
                        for (int i = 1; i < args.Length; i++)
                            args2[i - 1] = args[i];
                        args = args2;
                    }
                }
            }

            ArrayList nodes = new ArrayList();
            ArrayList colons = new ArrayList();
            // find the position of all the colons in the arguments
            for (int i = 0; i < args.Length; i++)
            {
                if (args[i] == ":")
                {
                    colons.Add(i);
                }
            }
            // terminate the list with a virtual colon just past the last argument
            colons.Add(args.Length);
            // create a list of argument sets
            ArrayList args_set = new ArrayList();
            int last_colon = 0;
            for (int i = 0; i < colons.Count; i++)
            {
                string[] cur_set;
                cur_set = new string[(int)colons[i] - last_colon];
                for (int j = last_colon; j < (int)colons[i]; j++)
                {
                    cur_set[j - last_colon] = args[j];
                }
                if (cur_set.Length > 0 && cur_set[0].Length > 0 && cur_set[0][0] == '-')
                {
                    // two special cases: -configfile and -file
                    // parse the two file options into argument sets
                    if (String.Compare(cur_set[0].TrimStart('-'), "configfile", true) == 0)
                    {
                        if (cur_set.Length < 2)
                        {
                            throw new Exception("No configfile filename specified");
                        }
                        ArrayList list = ParseConfigfile(cur_set[1]);
                        foreach (string[] temp_set in list)
                        {
                            args_set.Add(temp_set);
                        }
                    }
                    else if (String.Compare(cur_set[0].TrimStart('-'), "file", true) == 0)
                    {
                        if (cur_set.Length < 2)
                        {
                            throw new Exception("No filename specified after the -file option");
                        }
                        ArrayList list = ParseFile(cur_set[1]);
                        foreach (string[] temp_set in list)
                        {
                            args_set.Add(temp_set);
                        }
                    }
                    else if (String.Compare(cur_set[0].TrimStart('-'), "set_hosts", true) == 0)
                    {
                        if (cur_set.Length < 2)
                        {
                            throw new Exception("No hosts specified after the -set_hosts option");
                        }
                        string[] hosts = new string[cur_set.Length - 1];
                        for (int j = 1; j < cur_set.Length; j++)
                        {
                            hosts[j - 1] = cur_set[j];
                        }
                        SetDefaultHosts(hosts);
                        return null;
                    }
                    else if (String.Compare(cur_set[0].TrimStart('-'), "set_myhosts", true) == 0)
                    {
                        if (cur_set.Length < 2)
                        {
                            throw new Exception("No hosts specified after the -set_myhosts option");
                        }
                        string[] hosts = new string[cur_set.Length - 1];
                        for (int j = 1; j < cur_set.Length; j++)
                        {
                            hosts[j - 1] = cur_set[j];
                        }
                        SetMyDefaultHosts(hosts);
                        return null;
                    }
                    else if (String.Compare(cur_set[0].TrimStart('-'), "get_hosts", true) == 0)
                    {
                        string[] hosts = GetDefaultHostsStrings();
                        foreach (string host in hosts)
                        {
                            Console.WriteLine("{0}", host);
                        }
                        return null;
                    }
                    else if (String.Compare(cur_set[0].TrimStart('-'), "get_myhosts", true) == 0)
                    {
                        string[] hosts = GetMyDefaultHostsStrings();
                        foreach (string host in hosts)
                        {
                            Console.WriteLine("{0}", host);
                        }
                        return null;
                    }
                    else
                    {
                        args_set.Add(cur_set);
                    }
                }
                else
                {
                    args_set.Add(cur_set);
                }
                last_colon = (int)colons[i] + 1;
            }
            // parse each argument set into lists of launch node structures
            for (int i = 0; i < args_set.Count; i++)
            {
                nodes.AddRange(ParseSet((string[])(args_set[i]), i));
            }
            // set the total number of processes field
            foreach (LaunchNode node in nodes)
            {
                node.nproc = global_options.cur_rank;
            }
            // create the cliques
            ParseCliques(nodes);
            return nodes;
        }

        public ArrayList CreateLaunchNodes(string[] commands, string[][] argvs, int[] maxprocs, PMI_Keyval[][] keyvals)
        {
            int i, j;
            ArrayList nodes = new ArrayList();
            if (commands != null)
            {
                PMI pmi = new PMI();
                pmi.mpiexec = this;
                int iproc = 0;
                pmi.nproc = 0;
                if (maxprocs != null)
                {
                    foreach (int n in maxprocs)
                    {
                        pmi.nproc += n;
                    }
                }
                for (i = 0; i < commands.Length; i++)
                {
                    string exe = GetExecutable(commands[i]);
                    string path = GetPath(commands[i]);
                    ArrayList args = new ArrayList();
                    //mpiexec.WriteLine("cmd: {0}", commands[i]);
                    if (argvs != null && argvs.Length > i && argvs[i] != null)
                    {
                        for (j = 0; j < argvs[i].Length; j++)
                        {
                            args.Add(argvs[i][j]);
                            //mpiexec.WriteLine("arg: {0}", argvs[i][j]);
                        }
                    }
                    else
                    {
                        //mpiexec.WriteLine("no arguments");
                    }

                    LaunchNode default_node = new LaunchNode(this);
                    if (global_options != null)
                        default_node.dir = global_options.dir;
                    if (global_options.env != null)
                        default_node.env = (Hashtable)global_options.env.Clone();
                    if (keyvals != null && keyvals.Length > i && keyvals[i] != null)
                    {
                        for (j = 0; j < keyvals[i].Length; j++)
                        {
                            //mpiexec.WriteLine("key: {0}", keyvals[i][j].key);
                            //mpiexec.WriteLine("val: {0}", keyvals[i][j].val);
                            switch (keyvals[i][j].key)
                            {
                                case "path":
                                    if (path.Length > 0)
                                        path = String.Format("{0};{1}", path, keyvals[i][j].val);
                                    else
                                        path = keyvals[i][j].val;
                                    break;
                                case "env":
                                    Hashtable env_vars = SmpdPmi.ParseLineTrue(keyvals[i][j].val);
                                    foreach (string key in env_vars.Keys)
                                    {
                                        default_node.env[key] = env_vars[key];
                                    }
                                    break;
                                case "host":
                                    default_node.hostname = keyvals[i][j].val;
                                    break;
                                case "hosts":
                                    GetHostsFromString(keyvals[i][j].val);
                                    break;
                                case "log":
                                    default_node.env["MPI_WRAP_DLL_NAME"] = "mpe";
                                    break;
                                case "machinefile":
                                    ParseMachinefile(keyvals[i][j].val);
                                    break;
                                case "wdir":
                                    default_node.dir = keyvals[i][j].val;
                                    break;
                                case "dir":
                                    goto case "wdir";
                                case "map":
                                    /*
                                    if (keyvals[i][j].val.Length > 2)
                                        default_node.map[keyvals[i][j].val.Substring(0, 1)] = keyvals[i][j].val.Remove(0, 2);
                                    */
                                    try
                                    {
                                        string[] tokens = keyvals[i][j].val.Trim().Split(';');
                                        if (tokens != null)
                                        {
                                            foreach (string mapping in tokens)
                                            {
                                                string[] parts = mapping.Split(':');
                                                if (parts.Length > 1)
                                                {
                                                    default_node.map[parts[0]] = parts[1];
                                                }
                                            }
                                        }
                                    }
                                    catch
                                    {
                                    }
                                    break;
                                case "localonly":
                                    default_node.launch_local = true;
                                    break;
                                case "configfile":
                                    break;
                                case "file":
                                    break;
                                case "priority":
                                    try
                                    {
                                        string[] tokens;
                                        tokens = keyvals[i][j].val.Split(':');
                                        default_node.priority_class = Convert.ToInt32(tokens[0]);
                                        if (tokens.Length > 1)
                                        {
                                            default_node.priority_thread = Convert.ToInt32(tokens[1]);
                                        }
                                    }
                                    catch //(Exception e)
                                    {
                                    }
                                    break;
                                case "timeout":
                                    break;
                                case "exitcodes":
                                    global_options.exitcodes = true;
                                    break;
                                case "nompi":
                                    //global_options.nompi = true;
                                    break;
                                default:
                                    break;
                            }
                        }
                    }
                    else
                    {
                        //mpiexec.WriteLine("no keyvals");
                    }

                    if (maxprocs != null)
                    {
                        for (j = 0; j < maxprocs[i]; j++)
                        {
                            LaunchNode node = new LaunchNode(this);
                            node.spawned = true;
                            node.exe = exe;
                            node.path = path;
                            node.env = (Hashtable)default_node.env.Clone();
                            node.dir = default_node.dir;
                            node.iproc = iproc++;
                            node.nproc = pmi.nproc;
                            node.appnum = i;
                            node.args = (ArrayList)args.Clone();
                            if (default_node.hostname == null || default_node.hostname.Length < 1)
                            {
                                HostNode host_node;
                                if (global_options.hosts == null || global_options.cur_host < 0)
                                    GetDefaultHosts();
                                host_node = (HostNode)(global_options.hosts[global_options.cur_host]);
                                node.hostname = host_node.hostname;
                                global_options.cur_processor++;
                                if (global_options.cur_processor >= host_node.nproc)
                                {
                                    global_options.cur_processor = 0;
                                    global_options.cur_host++;
                                    if (global_options.cur_host >= global_options.hosts.Count)
                                        global_options.cur_host = 0;
                                }
                            }
                            else
                            {
                                node.hostname = default_node.hostname;
                            }
                            node.launch_local = default_node.launch_local;
                            node.priority_class = default_node.priority_class;
                            node.priority_thread = default_node.priority_thread;
                            node.map = (Hashtable)default_node.map.Clone();
                            node.pmi = pmi;
                            nodes.Add(node);
                        }
                    }
                }
                ParseCliques(nodes);
            }
            else
            {
                //mpiexec.WriteLine("No commands");
            }
            return nodes;
        }

        public ArrayList CreateLaunchNodes(string command_line, int nproc, string hosts_str, string env_str, string wdir)
        {
            int i;
            Hashtable env_vars;
            if (env_str != null)
                env_vars = SmpdPmi.ParseLineTrue(env_str);
            else
                env_vars = new Hashtable();
            GetHostsFromStringOrDefault(hosts_str);
            if (nproc < 1)
                nproc = 1;
            string[] cmd_tokens = ParseStrings(command_line);
            if (cmd_tokens == null)
                return null;
            ArrayList args = new ArrayList();
            for (i = 1; i < cmd_tokens.Length; i++)
            {
                args.Add(cmd_tokens[i]);
            }
            string exe = GetExecutable(cmd_tokens[0]);
            string path = GetPath(cmd_tokens[0]);
            global_options.cur_host = 0;
            global_options.cur_processor = 0;
            ArrayList nodes = new ArrayList();
            for (i = 0; i < nproc; i++)
            {
                LaunchNode node = new LaunchNode(this);
                node.exe = exe;
                node.path = path;
                node.env = (Hashtable)env_vars.Clone();
                node.dir = wdir;
                node.iproc = i;
                if (node.iproc == 0 && global_options.localroot)
                    node.launch_local = true;
                node.nproc = nproc;
                node.appnum = 0; // FIXME
                node.args = (ArrayList)args.Clone();
                if (global_options.map != null)
                    node.map = (Hashtable)global_options.map.Clone();
                HostNode host_node;
                host_node = (HostNode)(global_options.hosts[global_options.cur_host]);
                node.hostname = host_node.hostname;
                global_options.cur_processor++;
                if (global_options.cur_processor >= host_node.nproc)
                {
                    global_options.cur_processor = 0;
                    global_options.cur_host++;
                    if (global_options.cur_host >= global_options.hosts.Count)
                        global_options.cur_host = 0;
                }
                nodes.Add(node);
            }
            ParseCliques(nodes);
            return nodes;
        }

        #endregion

        private void PrintHostVersion(string host, SecureCredential credential)
        {
            try
            {
                ClientSession client;
                try
                {
                    client = new ClientSession(host, -1, global_options.phrase, store.Keys, credential);
                }
                catch (Exception e)
                {
                    if (e.Message.StartsWith("Version mismatch: "))
                    {
                        string[] tokens = e.Message.Split(' ');
                        Console.WriteLine(tokens[2]);
                        return;
                    }
                    StringBuilder sb = new StringBuilder("ERROR: ");
                    sb.Append(e.Message);
                    foreach (string key in e.Data.Keys)
                    {
                        sb.AppendFormat("\r\n{0}={1}", key, e.Data[key]);
                    }
                    Console.WriteLine(sb.ToString());
                    return;
                }
                InfoRequestMessage message = new InfoRequestMessage("host_info");
                client.SendMessage(message);
                HostInfoMessage hmessage = (HostInfoMessage)client.NextMessage();
                QuitMessage quit_message = new QuitMessage();
                client.SendMessage(quit_message);
                // get a response to the quit message
                DeinoMPI.Message message2;
                message2 = client.NextMessage();
                try
                {
                    // Send the last leg of the quit handshake
                    quit_message = new QuitMessage();
                    client.SendMessage(quit_message);
                }
                catch
                {
                    // Ignore errors here since we are guaranteed to be in the last leg of the quit handshake.
                }
                if (hmessage.Info != null)
                {
                    Console.WriteLine("{0}", hmessage.Info.Version);
                    Console.Out.Flush();
                }
            }
            catch
            {
            }
        }

        public void GetHostVersion(string hostname)
        {
            // create a bogus launch node to hold the credential
            LaunchNode node = new LaunchNode(this);
            ArrayList nodes = new ArrayList();
            nodes.Add(node);
            int result = GetStoreAndCredential(nodes);
            if (result != 0)
            {
                mpiexec_exit_code = result;
                return;
            }
            PrintHostVersion(hostname, node.credential);
            mpiexec_exit_code = 0;
        }

        public void PrintVersion()
        {
            Console.WriteLine(Session.Version);
            mpiexec_exit_code = 0;
        }
    }
}
