using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Security.Cryptography;
using System.Threading;

namespace DeinoMPI
{
    public class SecureMessageStream
    {
        const int PUBLIC_KEY_SUCCESS = 0;
        const int REQUEST_PUBLIC_KEY = 1;
        const int PUBLIC_KEY_REJECTED = 2;

        private BinaryReader reader;
        private BinaryWriter writer;
        private ICryptoTransform encryptor;
        private ICryptoTransform decryptor;
        private SecurityKeys public_key;

        public SecurityKeys PublicKey
        {
            get { return public_key; }
            set { public_key = value; }
        }

        /// <summary>
        /// Create a secure stream using a public key
        /// </summary>
        /// <param name="stream">The stream over which encrypted data is passed</param>
        /// <returns>SecureMessageStream</returns>
        static public SecureMessageStream FromPublicKey(Stream stream)
        {
            SymmetricAlgorithm sa = SymmetricAlgorithm.Create();
            if (sa == null)
                throw new Exception("Unable to create a symmetric encryption algorithm");
            SecureMessageStream ss = new SecureMessageStream();
            ss.writer = new BinaryWriter(stream);
            ss.reader = new BinaryReader(stream);
            int length;
            byte[] hash_bytes;
            // read the public key hash
            length = ss.reader.ReadInt32();
            hash_bytes = ss.reader.ReadBytes(length);
            // lookup the public key
            string hash;
            hash = UnicodeEncoding.Unicode.GetString(hash_bytes);
            ss.PublicKey = PublicKeyStore.GetPublicKey(hash);
            if (ss.PublicKey != null)
            {
                ss.writer.Write(PUBLIC_KEY_SUCCESS); // successfully looked up public key
            }
            else
            {
                if (PublicKeyStore.AutoAcceptPublicKeys())
                {
                    // public key not found, request it to be sent
                    ss.writer.Write(REQUEST_PUBLIC_KEY);
                    length = ss.reader.ReadInt32();
                    byte[] public_key_bytes = ss.reader.ReadBytes(length);
                    try
                    {
                        ss.PublicKey = new SecurityKeys(UnicodeEncoding.Unicode.GetString(public_key_bytes));
                        PublicKeyStore.SavePublicKey(ss.PublicKey);
                    }
                    catch
                    {
                        ss.writer.Close();
                        ss.reader.Close();
                        return null;
                    }
                }
                else
                {
                    ss.writer.Write(PUBLIC_KEY_REJECTED);
                }
            }
            byte[] exchange_blob = ss.PublicKey.GetKeyExchange(sa.Key);
            ss.writer.Write(exchange_blob.Length);
            ss.writer.Write(exchange_blob);
            ss.writer.Write(sa.IV.Length);
            ss.writer.Write(sa.IV);
            ss.encryptor = sa.CreateEncryptor();
            if (ss.encryptor == null)
                throw new Exception("Unable to create an encryptor.");
            ss.decryptor = sa.CreateDecryptor();
            if (ss.decryptor == null)
                throw new Exception("Unable to create a decryptor.");

            return ss;
        }
        /// <summary>
        /// Create a secure stream using the provided public key
        /// </summary>
        /// <param name="stream"></param>
        /// <param name="public_key_string"></param>
        /// <returns></returns>
        static public SecureMessageStream FromPublicKeyString(Stream stream, string public_key_string)
        {
            SymmetricAlgorithm sa = SymmetricAlgorithm.Create();
            if (sa == null)
                throw new Exception("Unable to create a symmetric encryption algorithm");
            SecureMessageStream ss = new SecureMessageStream();
            ss.writer = new BinaryWriter(stream);
            ss.reader = new BinaryReader(stream);
            ss.PublicKey = SecurityKeys.Create(public_key_string);
            byte[] exchange_blob = ss.PublicKey.GetKeyExchange(sa.Key);
            ss.writer.Write(exchange_blob.Length);
            ss.writer.Write(exchange_blob);
            ss.writer.Write(sa.IV.Length);
            ss.writer.Write(sa.IV);
            ss.encryptor = sa.CreateEncryptor();
            if (ss.encryptor == null)
                throw new Exception("Unable to create an encryptor.");
            ss.decryptor = sa.CreateDecryptor();
            if (ss.decryptor == null)
                throw new Exception("Unable to create a decryptor.");

            return ss;
        }

        /// <summary>
        /// Create a secure stream from a private key
        /// </summary>
        /// <param name="private_key">The private key used to decrypt the symmetric key</param>
        /// <param name="stream">The stream over which encrypted data is passed</param>
        /// <returns>SecureMessageStream</returns>
        static public SecureMessageStream FromPrivateKey(SecurityKeys private_key, Stream stream)
        {
            byte[] exchange_blob;
            byte[] key;
            byte[] IV;
            int length, result;

            SecureMessageStream ss = new SecureMessageStream();
            ss.reader = new BinaryReader(stream);
            ss.writer = new BinaryWriter(stream);
            // send the public key hash
            byte[] hash_bytes = UnicodeEncoding.Unicode.GetBytes(private_key.Hash);
            ss.writer.Write(hash_bytes.Length);
            ss.writer.Write(hash_bytes);
            result = ss.reader.ReadInt32();
            switch (result)
            {
                case PUBLIC_KEY_SUCCESS:
                    // public key accepted
                    break;
                case REQUEST_PUBLIC_KEY:
                    // public key not found on remote host
                    byte[] public_key_bytes = UnicodeEncoding.Unicode.GetBytes(private_key.PublicKey);
                    ss.writer.Write(public_key_bytes.Length);
                    ss.writer.Write(public_key_bytes);
                    break;
                case PUBLIC_KEY_REJECTED:
                    // public key not found on remote host and key cannot be sent
                    ss.writer.Close();
                    ss.reader.Close();
                    return null;
                default:
                    break;
            }
            // receive the exchange key and initialization vector
            length = ss.reader.ReadInt32();
            exchange_blob = ss.reader.ReadBytes(length);
            length = ss.reader.ReadInt32();
            IV = ss.reader.ReadBytes(length);

            key = private_key.DecryptKeyExchange(exchange_blob);

            SymmetricAlgorithm sa = SymmetricAlgorithm.Create();
            if (sa == null)
                throw new Exception("Unable to create a symmetric encryption algorithm");
            ss.encryptor = sa.CreateEncryptor(key, IV);
            if (ss.encryptor == null)
                throw new Exception("Unable to create an encryptor.");
            ss.decryptor = sa.CreateDecryptor(key, IV);
            if (ss.decryptor == null)
                throw new Exception("Unable to create a decryptor.");

            return ss;
        }

        /// <summary>
        /// Create a secure stream from a private key where the other side already has the public key
        /// </summary>
        /// <param name="private_key">The private key used to decrypt the symmetric key</param>
        /// <param name="stream">The stream over which encrypted data is passed</param>
        /// <returns>SecureMessageStream</returns>
        static public SecureMessageStream FromPrivateKeyNoSendPublicKey(SecurityKeys private_key, Stream stream)
        {
            byte[] exchange_blob;
            byte[] key;
            byte[] IV;
            int length;

            SecureMessageStream ss = new SecureMessageStream();
            ss.reader = new BinaryReader(stream);
            ss.writer = new BinaryWriter(stream);
            // receive the exchange key and initialization vector
            length = ss.reader.ReadInt32();
            exchange_blob = ss.reader.ReadBytes(length);
            length = ss.reader.ReadInt32();
            IV = ss.reader.ReadBytes(length);

            key = private_key.DecryptKeyExchange(exchange_blob);

            SymmetricAlgorithm sa = SymmetricAlgorithm.Create();
            if (sa == null)
                throw new Exception("Unable to create a symmetric encryption algorithm");
            ss.encryptor = sa.CreateEncryptor(key, IV);
            if (ss.encryptor == null)
                throw new Exception("Unable to create an encryptor.");
            ss.decryptor = sa.CreateDecryptor(key, IV);
            if (ss.decryptor == null)
                throw new Exception("Unable to create a decryptor.");

            return ss;
        }

        /// <summary>
        /// Read the next message on the stream
        /// </summary>
        /// <returns>decrypted message buffer</returns>
        public byte[] Read()
        {
            int length, text_length, num_read;
            byte[] cipher;

            lock (reader)
            {
                text_length = reader.ReadInt32();
                length = reader.ReadInt32();
                cipher = reader.ReadBytes(length);
            }
            MemoryStream ms = new MemoryStream(cipher);
            CryptoStream cs = new CryptoStream(ms, decryptor, CryptoStreamMode.Read);
            byte[] text = new byte[text_length];
            num_read = cs.Read(text, 0, text_length);
            if (num_read != text_length)
                throw new Exception(String.Format("Expected decrypted message length {0} but {1} bytes read", text_length, num_read));
            try
            {
                ms.Close();
                cs.Close();
            }
            catch
            {
                // FIXME: Is it ok to ignore an error after successfully reading the bytes?
            }
            return text;
        }

        /// <summary>
        /// Write a message to the stream
        /// </summary>
        /// <param name="buffer">the buffer to be encrypted and written</param>
        public void Write(byte[] buffer)
        {
            if (buffer == null || buffer.Length == 0)
                return;
            MemoryStream ms = new MemoryStream();
            CryptoStream cs = new CryptoStream(ms, encryptor, CryptoStreamMode.Write);

            // Encrypt the plaintext bytes into the memory stream.
            cs.Write(buffer, 0, buffer.Length);
            cs.FlushFinalBlock();

            // Get the encrypted bytes stored in the memory stream.
            byte[] cipherTextBytes = ms.ToArray();

            ms.Close();
            cs.Close();

            ms = new MemoryStream();
            BinaryWriter bw = new BinaryWriter(ms);
            bw.Write(buffer.Length);
            bw.Write(cipherTextBytes.Length);
            bw.Write(cipherTextBytes);
            bw.Flush();
            byte[] single_message = ms.ToArray();
            lock (writer)
            {
                writer.Write(single_message);
            }
            /*
            lock (writer)
            {
                writer.Write(buffer.Length);
                writer.Write(cipherTextBytes.Length);
                writer.Write(cipherTextBytes);
            }
            */
        }

        /// <summary>
        /// Close the stream
        /// </summary>
        public void Close()
        {
            writer.Close();
            reader.Close();
        }
    }
}
