#include "FileTree.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <windows.h>
#include <shlobj.h>

/*
 *
 * FileTreeFileNode class
 *
 */
FileTreeFileNode::FileTreeFileNode(void)
{
    filename = NULL;
    name_part = NULL;
    next = NULL;
    traversed = false;
}

FileTreeFileNode::~FileTreeFileNode(void)
{
    if (filename != NULL)
	free(filename);
}

void FileTreeFileNode::SetFileName(const wchar_t *filename)
{
    this->filename = _wcsdup(filename);
    if (this->filename != NULL)
    {
	name_part = wcsrchr(this->filename, L'\\');
	if (name_part == NULL)
	    name_part = this->filename;
	else
	    name_part++; // skip over the \ character
    }
    else
    {
	name_part = NULL;
    }
}

void FileTreeFileNode::SetWriteTime(FILETIME write_time)
{
    this->write_time = write_time;
}

const wchar_t *FileTreeFileNode::GetFullFileName()
{
    return filename;
}

const wchar_t *FileTreeFileNode::GetFileName()
{
    return name_part;
}

FILETIME FileTreeFileNode::GetWriteTime()
{
    return write_time;
}

bool FileTreeFileNode::Different(FileTreeFileNode *file)
{
    return CompareFileTime(&write_time, &file->write_time) != 0;
}

void FileTreeFileNode::ResetTraversed()
{
    traversed = false;
    if (next != NULL)
	next->ResetTraversed();
}

void FileTreeFileNode::Print()
{
    SYSTEMTIME stUTC, stLocal;

    wprintf(L"FileName: %s\n", filename);
    FileTimeToSystemTime(&write_time, &stUTC);
    SystemTimeToTzSpecificLocalTime(NULL, &stUTC, &stLocal);
    wprintf(L"Last write time: %02d/%02d/%d  %02d:%02d\n",
        stLocal.wMonth, stLocal.wDay, stLocal.wYear,
        stLocal.wHour, stLocal.wMinute);
}

/*
 *
 * FileTreeDirNode class
 *
 */
FileTreeDirNode::FileTreeDirNode(void)
{
    parent = NULL;
    children = NULL;
    sibling = NULL;
    files = NULL;
    full_path = NULL;
    dir_name = NULL;
    traversed = false;
}

FileTreeDirNode::FileTreeDirNode(const wchar_t *current_dir, const wchar_t *sub_dir)
{
    size_t len, sub_dir_len;
    bool add_slash = false;

    parent = NULL;
    children = NULL;
    sibling = NULL;
    files = NULL;
    traversed = false;
    len = wcslen(current_dir);
    if (current_dir[len-1] != L'\\')
    {
	add_slash = true;
	len++;
    }
    sub_dir_len = wcslen(sub_dir);
    len += sub_dir_len + 1;
    full_path = new wchar_t[len];
    if (full_path == NULL)
    {
	MPI_Abort(MPI_COMM_WORLD, 30);
    }
    if (add_slash)
	swprintf_s(full_path, len, L"%s\\%s", current_dir, sub_dir);
    else
	swprintf_s(full_path, len, L"%s%s", current_dir, sub_dir);
    dir_name = &full_path[len - sub_dir_len - 1];
    //dir_name = _wcsdup(sub_dir);
}

FileTreeDirNode::~FileTreeDirNode(void)
{
    if (full_path != NULL)
    {
	delete [] full_path;
	full_path = NULL;
    }
    //if (dir_name != NULL) free(dir_name);
}

void FileTreeDirNode::Populate(wchar_t *wild_cards)
{
    WIN32_FIND_DATA FindFileData;
    HANDLE hFind = INVALID_HANDLE_VALUE;
    DWORD dwError;
    size_t len, size;
    wchar_t *dir = NULL;

    if ((wild_cards == NULL) || (wild_cards[0] == L'\0'))
	wild_cards = L"*.*";
    len = wcslen(full_path);
    size = len + wcslen(wild_cards);
    if (full_path[len-1] != L'\\')
    {
	size += 2;
	dir = new wchar_t[size];
	if (dir == NULL)
	{
	    MPI_Abort(MPI_COMM_WORLD, 31);
	}
	swprintf_s(dir, size, L"%s\\%s", full_path, wild_cards);
    }
    else
    {
	size++; // one for the NULL terminator
	dir = new wchar_t[size];
	if (dir == NULL)
	{
	    MPI_Abort(MPI_COMM_WORLD, 32);
	}
	swprintf_s(dir, size, L"%s%s", full_path, wild_cards);
    }

    hFind = FindFirstFile(dir, &FindFileData);
    if (hFind == INVALID_HANDLE_VALUE)
    {
	wprintf(L"Invalid file handle. Error is %u\n", GetLastError());
	if (dir != NULL)
	    delete [] dir;
	return;
    }
    else
    {
	if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
	{
	    if (FindFileData.cFileName[0] != L'.')
	    {
		AddChild(FindFileData.cFileName, wild_cards);
	    }
	}
	else
	{
	    AddFile(FindFileData.cFileName, FindFileData.ftLastWriteTime);
	}
	while (FindNextFile(hFind, &FindFileData) != 0)
	{
	    if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
	    {
		if (FindFileData.cFileName[0] != L'.')
		{
		    AddChild(FindFileData.cFileName, wild_cards);
		}
	    }
	    else
	    {
		AddFile(FindFileData.cFileName, FindFileData.ftLastWriteTime);
	    }
	}

	dwError = GetLastError();
	FindClose(hFind);
	if (dwError != ERROR_NO_MORE_FILES)
	{
	    wprintf(L"FindNextFile error. Error is %u\n", dwError);
	    if (dir != NULL)
		delete [] dir;
	    return;
	}
    }
    if (dir != NULL)
	delete [] dir;
}

void FileTreeDirNode::AddChild(wchar_t *dir, wchar_t *wild_cards)
{
    // This function creates a child and populates it
    FileTreeDirNode *child = new FileTreeDirNode(full_path, dir);
    if (child == NULL)
	MPI_Abort(MPI_COMM_WORLD, 33);
    child->parent = this;
    child->sibling = children;
    children = child;
    child->Populate(wild_cards);
}

FileTreeDirNode* FileTreeDirNode::CreateChild(const wchar_t *dir)
{
    // This function creates a child
    //wprintf(L"creating child: path=%s, dir=%s\n", full_path, dir);fflush(stdout);
    FileTreeDirNode *child = new FileTreeDirNode(full_path, dir);
    if (child == NULL)
	MPI_Abort(MPI_COMM_WORLD, 34);
    child->parent = this;
    child->sibling = children;
    children = child;
    return child;
}

FileTreeDirNode* FileTreeDirNode::CreateSibling(const wchar_t *dir)
{
    // This funcion creates a sibling
    FileTreeDirNode *sib;
    if (parent != NULL)
    {
	sib = new FileTreeDirNode(parent->full_path, dir);
	if (sib == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 35);
    }
    else
    {
	wchar_t *parent_path = _wcsdup(full_path);
	wchar_t *last_slash = wcsrchr(full_path, '\\');
	if (last_slash != NULL)
	    *last_slash = L'\0';
	sib = new FileTreeDirNode(parent_path, dir);
	if (sib == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 36);
	if (parent_path != NULL)
	    free(parent_path);
    }
    sib->parent = parent;
    sib->sibling = sibling;
    sibling = sib;
    return sib;
}

void FileTreeDirNode::AddFile(const wchar_t *filename, const FILETIME last_write)
{
    wchar_t *full_filename;
    size_t len;
    bool add_slash = false;
    FileTreeFileNode *node = new FileTreeFileNode();
    if (node == NULL)
	MPI_Abort(MPI_COMM_WORLD, 37);
    //wprintf(L"Adding file: full_path=%s, filename=%s\n", full_path, filename);fflush(stdout);
    len = wcslen(full_path);
    if (full_path[len-1] != L'\\')
    {
	add_slash = true;
	len++;
    }
    len += wcslen(filename);
    full_filename = new wchar_t[len+1];
    if (full_filename == NULL)
    {
	delete node;
	MPI_Abort(MPI_COMM_WORLD, 38);
	return;
    }
    if (add_slash)
	swprintf_s(full_filename, len+1, L"%s\\%s", full_path, filename);
    else
	swprintf_s(full_filename, len+1, L"%s%s", full_path, filename);
    //wprintf(L"Setting file name to: %s\n", full_filename);fflush(stdout);
    node->SetFileName(full_filename);
    node->SetWriteTime(last_write);
    node->next = files;
    files = node;
    delete [] full_filename;
}

FileTreeDirNode *FileTreeDirNode::GetParent()
{
    return parent;
}

FileTreeDirNode *FileTreeDirNode::GetFirstChild()
{
    return children;
}

FileTreeDirNode *FileTreeDirNode::GetNextSibling()
{
    return sibling;
}

FileTreeFileNode *FileTreeDirNode::GetFirstFile()
{
    return files;
}

const wchar_t *FileTreeDirNode::GetDirName()
{
    return dir_name;
}

const wchar_t *FileTreeDirNode::GetFullPath()
{
    return full_path;
}

void FileTreeDirNode::ResetTraversed()
{
    traversed = false;
    if (children != NULL)
	children->ResetTraversed();
    if (sibling != NULL)
	sibling->ResetTraversed();
    if (files != NULL)
	files->ResetTraversed();
}

bool FileTreeDirNode::Empty()
{
    return (children == NULL) && (files == NULL);
}

void FileTreeDirNode::PruneEmptyDirectories()
{
    FileTreeDirNode *temp;
    if (children != NULL)
	children->PruneEmptyDirectories();
    if (sibling != NULL)
	sibling->PruneEmptyDirectories();
    if (sibling != NULL && sibling->Empty())
    {
	temp = sibling;
	sibling = sibling->GetNextSibling();
	delete temp;
    }
    if (children != NULL && children->Empty())
    {
	temp = children;
	children = children->GetNextSibling();
	delete temp;
    }
}

void FileTreeDirNode::Print()
{
    if (dir_name != NULL)
	wprintf(L"%s\n", dir_name);
    FileTreeFileNode *iter = files;
    while (iter != NULL)
    {
	iter->Print();
	iter = iter->next;
    }
    FileTreeDirNode *diter = GetFirstChild();
    if (diter != NULL)
    {
	diter->Print();
	diter = diter->GetNextSibling();
	while (diter != NULL)
	{
	    diter->Print();
	    diter = diter->GetNextSibling();
	}
    }
}

/*
 *
 * FileTree class
 *
 */
FileTree::FileTree(void)
{
    recursive = true;
    src_dir = NULL;
    dest_dir = NULL;
    iter_list = NULL;
    root = NULL;
    src_rank = 0;
}

FileTree::FileTree(wchar_t *src_dir, bool recursive)
{
    this->recursive = recursive;
    this->src_dir = _wcsdup(src_dir);
    dest_dir = NULL;
    iter_list = NULL;
    root = NULL;
    src_rank = 0;
}

FileTree::~FileTree(void)
{
    if (src_dir != NULL)
    {
	free(src_dir);
	src_dir = NULL;
    }
    if (dest_dir != NULL)
    {
	free(dest_dir);
	dest_dir = NULL;
    }
}

void FileTree::SetSrcDir(wchar_t *dir)
{
    size_t len;
    src_dir = _wcsdup(dir);
    if (src_dir != NULL)
    {
	len = wcslen(src_dir);
	if (len > 0)
	{
	    if (src_dir[len-1] == L'\\')
		src_dir[len-1] = L'\0';
	}
    }
}

void FileTree::SetDestDir(wchar_t *dir)
{
    size_t len;
    dest_dir = _wcsdup(dir);
    if (dest_dir != NULL)
    {
	len = wcslen(dest_dir);
	if (len > 0)
	{
	    if (dest_dir[len-1] == L'\\')
		dest_dir[len-1] = L'\0';
	}
    }
}

void FileTree::SetSrcRank(int comm_world_rank)
{
    src_rank = comm_world_rank;
}

void FileTree::CreateDestDir()
{
    int result;
    result = SHCreateDirectoryEx(NULL, dest_dir, NULL);
    switch (result)
    {
    case ERROR_BAD_PATHNAME:
	// The pszPath parameter was set to a relative path. 
	break;
    case ERROR_FILENAME_EXCED_RANGE:
	// The path pointed to by pszPath is too long.
	break;
    case ERROR_PATH_NOT_FOUND:
	// The system cannot find the path pointed to by pszPath. The path may contain an invalid entry.
	break;
    case ERROR_FILE_EXISTS:
	// The directory exists.
	break;
    case ERROR_ALREADY_EXISTS:
	// The directory exists.
	break;
    default:
	break;
    }
}

void FileTree::PopulateTree(wchar_t *src_dir, wchar_t *wild_cards)
{
    if (src_dir == NULL || src_dir[0] == L'\0')
	return;

    WIN32_FIND_DATA FindFileData;
    HANDLE hFind = INVALID_HANDLE_VALUE;
    DWORD dwError;
    size_t len, size;
    wchar_t *dir = NULL;

    if ((wild_cards == NULL) || (wild_cards[0] == L'\0'))
	wild_cards = L"*.*";
    len = wcslen(src_dir);
    size = len + wcslen(wild_cards);
    if (src_dir[len-1] != L'\\')
    {
	size += 2;
	dir = new wchar_t[size];
	if (dir == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 38);
	swprintf_s(dir, size, L"%s\\%s", src_dir, wild_cards);
    }
    else
    {
	size++; // one for the NULL terminator
	dir = new wchar_t[size];
	if (dir == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 39);
	swprintf_s(dir, size, L"%s%s", src_dir, wild_cards);
    }

    hFind = FindFirstFile(dir, &FindFileData);

    if (hFind == INVALID_HANDLE_VALUE)
    {
	wprintf(L"Invalid file handle. Error is %u\n", GetLastError());
	if (dir != NULL)
	    delete [] dir;
	return;
    }
    else
    {
	root = new FileTreeDirNode(src_dir, L"");
	if (root == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 40);
	if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
	{
	    if ((FindFileData.cFileName[0] != L'.') && (recursive))
	    {
		root->AddChild(FindFileData.cFileName, wild_cards);
	    }
	}
	else
	{
	    root->AddFile(FindFileData.cFileName, FindFileData.ftLastWriteTime);
	}
	while (FindNextFile(hFind, &FindFileData) != 0)
	{
	    if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
	    {
		if ((FindFileData.cFileName[0] != L'.') && (recursive))
		{
		    root->AddChild(FindFileData.cFileName, wild_cards);
		}
	    }
	    else
	    {
		root->AddFile(FindFileData.cFileName, FindFileData.ftLastWriteTime);
	    }
	}

	dwError = GetLastError();
	FindClose(hFind);
	if (dwError != ERROR_NO_MORE_FILES)
	{
	    wprintf(L"FindNextFile error. Error is %u\n", dwError);
	    if (dir != NULL)
		delete [] dir;
	    return;
	}
    }
    if (dir != NULL)
	delete [] dir;
}

void FileTree::PopulateTreeFromSrc(wchar_t *wild_cards)
{
    PopulateTree(src_dir, wild_cards);
}

void FileTree::PopulateTreeFromDest(wchar_t *wild_cards)
{
    PopulateTree(dest_dir, wild_cards);
}

static void GetChanges(FileTreeDirNode *start, FileTreeDirNode *end, FileTreeDirNode *diff)
{
    FileTreeDirNode *child, *sibling, *dir_iter, *diff_child, *diff_sibling;
    FileTreeFileNode *cur_file, *file_iter;

    // Traverse the children of start
    if (start != NULL)
    {
	child = start->GetFirstChild();
	while (child != NULL)
	{
	    if (end != NULL)
	    {
		dir_iter = end->GetFirstChild();
		while (dir_iter != NULL)
		{
		    if (wcscmp(dir_iter->GetDirName(), child->GetDirName()) == 0)
		    {
			if (!dir_iter->traversed || !child->traversed)
			{
			    diff_child = diff->CreateChild(child->GetDirName());
			    dir_iter->traversed = true;
			    child->traversed = true;
			    GetChanges(child, dir_iter, diff_child);
			}
		    }
		    dir_iter = dir_iter->GetNextSibling();
		}
	    }
	    if (!child->traversed)
	    {
		child->traversed = true;
		// Do not traverse directories not found in the end tree.  These are deleted directories.
		// FIXME: Should the corresponding directory be deleted at the source rank?
		//diff_child = diff->CreateChild(child->GetDirName());
		//GetChanges(child, NULL, diff_child);
	    }
	    child = child->GetNextSibling();
	}
    }
    // Traverse the children of end that did not match any children of start
    if (end != NULL)
    {
	child = end->GetFirstChild();
	while (child != NULL)
	{
	    if (!child->traversed)
	    {
		diff_child = diff->CreateChild(child->GetDirName());
		GetChanges(NULL, child, diff_child);
	    }
	    child = child->GetNextSibling();
	}
    }
    // Traverse the siblings of start
    if (start != NULL)
    {
	sibling = start->GetNextSibling();
	while (sibling != NULL)
	{
	    if (end != NULL)
	    {
		dir_iter = end->GetNextSibling();
		while (dir_iter != NULL)
		{
		    if (wcscmp(dir_iter->GetDirName(), sibling->GetDirName()) == 0)
		    {
			if (!dir_iter->traversed || !sibling->traversed)
			{
			    dir_iter->traversed = true;
			    sibling->traversed = true;
			    diff_sibling = diff->CreateSibling(sibling->GetDirName());
			    GetChanges(sibling, dir_iter, diff_sibling);
			}
		    }
		    dir_iter = dir_iter->GetNextSibling();
		}
	    }
	    /*
	    Ignore siblings that do not exist in the end list.  These are deleted directories.
	    FIXME: Should the corresponding directory be deleted at the source rank?
	    if (!sibling->traversed)
	    {
		diff_sibling = diff->CreateSibling(sibling->GetDirName());
		GetChanges(sibling, NULL, diff_sibling);
	    }
	    */
	    sibling = sibling->GetNextSibling();
	}
    }
    // Traverse the siblings of end that did not match any siblings of start
    if (end != NULL)
    {
	sibling = end->GetNextSibling();
	while (sibling != NULL)
	{
	    if (!sibling->traversed)
	    {
		sibling->traversed = true;
		diff_sibling = diff->CreateSibling(sibling->GetDirName());
		GetChanges(NULL, sibling, diff_sibling);
	    }
	    sibling = sibling->GetNextSibling();
	}
    }
    // Add files from start that are different from matches in end
    if (start != NULL)
    {
	cur_file = start->GetFirstFile();
	while (cur_file != NULL)
	{
	    if (end != NULL)
	    {
		file_iter = end->GetFirstFile();
		while (file_iter != NULL)
		{
		    if (wcscmp(cur_file->GetFullFileName(), file_iter->GetFullFileName()) == 0)
		    {
			if (!cur_file->traversed || !file_iter->traversed)
			{
			    cur_file->traversed = true;
			    file_iter->traversed = true;
			    if (cur_file->Different(file_iter))
			    {
				//wprintf(L"Adding different file: full=%s, name=%s\n", file_iter->GetFullFileName(), file_iter->GetFileName());fflush(stdout);
				diff->AddFile(file_iter->GetFileName(), file_iter->GetWriteTime());
			    }
			}
		    }
		    file_iter = file_iter->next;
		}
	    }
	    if (!cur_file->traversed)
	    {
		cur_file->traversed = true;
		// Do not add files from the start list that are not in the end list.  These are deleted files.
		// FIXME: Should the corresponding file be deleted at the source rank?
		//wprintf(L"Adding file from untraversed start list: %s\n", cur_file->GetFullFileName());fflush(stdout);
		//diff->AddFile(cur_file->GetFullFileName(), cur_file->GetWriteTime());
	    }
	    cur_file = cur_file->next;
	}
    }
    // Add files in end that did not match any files in start
    if (end != NULL)
    {
	cur_file = end->GetFirstFile();
	while (cur_file != NULL)
	{
	    if (!cur_file->traversed)
	    {
		cur_file->traversed = true;
		//wprintf(L"Adding file from untraversed end list: full=%s, name=%s\n", cur_file->GetFullFileName(), cur_file->GetFileName());fflush(stdout);
		diff->AddFile(cur_file->GetFileName(), cur_file->GetWriteTime());
	    }
	    cur_file = cur_file->next;
	}
    }
}

FileTree* FileTree::GetChangedFiles(FileTree *tree)
{
    // This function only works if src_dir and tree->src_dir are the same
    // It assumes you are comparing two trees with the same root.
    if (tree == NULL)
	return NULL;
    FileTree *diff = new FileTree();
    if (diff == NULL)
	MPI_Abort(MPI_COMM_WORLD, 41);
    diff->SetSrcDir(src_dir);
    if (root != NULL)
	root->ResetTraversed();
    if (tree->root != NULL)
	tree->root->ResetTraversed();
    diff->root = new FileTreeDirNode(src_dir, L"");
    if (diff->root == NULL)
	MPI_Abort(MPI_COMM_WORLD, 42);
    GetChanges(root, tree->root, diff->root);
    if (diff->root != NULL)
    {
	diff->root->PruneEmptyDirectories();
	if (diff->root->Empty())
	{
	    delete diff->root;
	    diff->root = NULL;
	}
    }
    return diff;
}

void FileTree::Print()
{
    if (root != NULL)
	root->Print();
}

void FileTree::SendTree()
{
    SendTree(NULL);
}

const wchar_t *FileTree::GetDestDir()
{
    return dest_dir;
}

#define MAX_BUFFER_SIZE 32*1024*1024
static unsigned char *buffer = NULL;

static void SendTreeRecursive(FileTreeDirNode *dir_node, int src_rank, FileTree *dest_tree)
{
    FileTreeFileNode *file_iter;
    FileTreeDirNode *subdir_iter, *sibling_iter;
    int len;
    const wchar_t *filename;

    if (buffer == NULL)
    {
	buffer = new unsigned char[MAX_BUFFER_SIZE];
	if (buffer == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 2);
    }
    file_iter = dir_node->GetFirstFile();
    while (file_iter != NULL)
    {
	filename = file_iter->GetFullFileName();
	HANDLE hFile, hLocalFile = INVALID_HANDLE_VALUE;
	hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	filename = wcsrchr(filename, L'\\');
	if (filename == NULL)
	    filename = file_iter->GetFullFileName();
	else
	    filename++;
	len = (int)wcslen(filename) + 1;
	if (dest_tree != NULL)
	{
	    hLocalFile = CreateFile(filename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	}
	if (hFile != INVALID_HANDLE_VALUE)
	{
	    // Send file name length
	    //wprintf(L"Sending file name length: %d\n", len);fflush(stdout);
	    MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
	    // Send file name
	    //wprintf(L"Sending file name: %s\n", filename);fflush(stdout);
	    MPI_Bcast((void*)filename, (int)len, MPI_WCHAR, src_rank, MPI_COMM_WORLD);
	    DWORD num_read, num_written;
	    ReadFile(hFile, buffer, MAX_BUFFER_SIZE, &num_read, NULL);
	    while (num_read != 0)
	    {
		// Send number of bytes read
		//wprintf(L"Sending num_read: %d\n", num_read);fflush(stdout);
		MPI_Bcast(&num_read, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
		// Send bytes
		//wprintf(L"Sending bytes\n");fflush(stdout);
		MPI_Bcast(buffer, num_read, MPI_BYTE, src_rank, MPI_COMM_WORLD);
		if (hLocalFile != INVALID_HANDLE_VALUE)
		    WriteFile(hLocalFile, buffer, num_read, &num_written, NULL);
		ReadFile(hFile, buffer, MAX_BUFFER_SIZE, &num_read, NULL);
	    }
	    // Send zero bytes read to indicate the end of the file
	    //wprintf(L"Sending zero num_read to indicate end of file\n");fflush(stdout);
	    MPI_Bcast(&num_read, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
	    CloseHandle(hFile);
	    if (hLocalFile != INVALID_HANDLE_VALUE)
		CloseHandle(hLocalFile);
	}
	file_iter = file_iter->next;
    }
    // Send zero length file name to indicate no more files
    len = 0;
    //wprintf(L"Sending zero file name length to indicate no more files\n");fflush(stdout);
    MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
    subdir_iter = dir_node->GetFirstChild();
    if (subdir_iter != NULL)
    {
	const wchar_t *subdir_name = subdir_iter->GetDirName();
	len = (int)wcslen(subdir_name) + 1;
	// Send the sub-directory name length
	//wprintf(L"Sending sub-dir name length: %d\n", len);fflush(stdout);
	MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
	// Send the sub-directory name
	//wprintf(L"Sending sub-dir name: %s\n", subdir_name);fflush(stdout);
	MPI_Bcast((void*)subdir_name, (int)len, MPI_WCHAR, src_rank, MPI_COMM_WORLD);
	if (dest_tree != NULL)
	{
	    if (!CreateDirectory(subdir_name, NULL))
	    {
		DWORD error = GetLastError();
		if (error != ERROR_ALREADY_EXISTS)
		    MPI_Abort(MPI_COMM_WORLD, error);
	    }
	    if (!SetCurrentDirectory(subdir_name))
	    {
		MPI_Abort(MPI_COMM_WORLD, 15);
	    }
	}
	SendTreeRecursive(subdir_iter, src_rank, dest_tree);
	if (dest_tree != NULL)
	{
	    if (!SetCurrentDirectory(L".."))
	    {
		MPI_Abort(MPI_COMM_WORLD, 16);
	    }
	}
    }
    else
    {
	// Send a zero sub-directory name length to indicate no more sub-directories
	len = 0;
	//wprintf(L"Sending zero sub-dir name length to indicate no more sub directories\n");fflush(stdout);
	MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
    }
    sibling_iter = dir_node->GetNextSibling();
    if (sibling_iter != NULL)
    {
	const wchar_t *sibling_name = sibling_iter->GetDirName();
	len = (int)wcslen(sibling_name) + 1;
	// Send the sibling-directory name length
	//wprintf(L"Sending sibling name length: %d\n", len);fflush(stdout);
	MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
	// Send the sibling-directory name
	//wprintf(L"Sending sibling name: %s\n", sibling_name);fflush(stdout);
	MPI_Bcast((void*)sibling_name, (int)len, MPI_WCHAR, src_rank, MPI_COMM_WORLD);
	if (dest_tree != NULL)
	{
	    if (!SetCurrentDirectory(L".."))
	    {
		MPI_Abort(MPI_COMM_WORLD, 17);
	    }
	    if (!CreateDirectory(sibling_name, NULL))
	    {
		DWORD error = GetLastError();
		if (error != ERROR_ALREADY_EXISTS)
		    MPI_Abort(MPI_COMM_WORLD, error);
	    }
	    if (!SetCurrentDirectory(sibling_name))
	    {
		MPI_Abort(MPI_COMM_WORLD, 18);
	    }
	}
	SendTreeRecursive(sibling_iter, src_rank, dest_tree);
    }
    else
    {
	// Send a zero sibling-directory name length to indicate no more sibling directories
	len = 0;
	//wprintf(L"Sending zero sibling name length to indicate no more siblings\n");fflush(stdout);
	MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
    }
    if (buffer != NULL)
    {
	delete [] buffer;
	buffer = NULL;
    }
}

static void ReceiveTreeRecursive(FileTreeDirNode *dir_node, int src_rank)
{
    int len;
    const wchar_t *filename;

    if (buffer == NULL)
    {
	buffer = new unsigned char[MAX_BUFFER_SIZE];
	if (buffer == NULL)
	{
	    MPI_Abort(MPI_COMM_WORLD, 3);
	}
    }

    // Receive the files for the current node
    //wprintf(L"Receiving the file name length\n");fflush(stdout);
    MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
    while (len != 0)
    {
	filename = new wchar_t[len];
	if (filename == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 43);
	//wprintf(L"Received file name length: %d\nReceiving file name\n", len);fflush(stdout);
	MPI_Bcast((void*)filename, (int)len, MPI_WCHAR, src_rank, MPI_COMM_WORLD);
	//wprintf(L"Received file name: %s\n", filename);fflush(stdout);
	HANDLE hFile = CreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	delete [] filename;
	DWORD num_read = 0, num_written = 0;
	MPI_Bcast(&num_read, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
	while (num_read != 0)
	{
	    MPI_Bcast(buffer, num_read, MPI_BYTE, src_rank, MPI_COMM_WORLD);
	    if (hFile != INVALID_HANDLE_VALUE)
	    {
		WriteFile(hFile, buffer, num_read, &num_written, NULL);
	    }
	    MPI_Bcast(&num_read, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
	}
	if (hFile != INVALID_HANDLE_VALUE)
	    CloseHandle(hFile);
	//wprintf(L"Receiving the file name length\n");fflush(stdout);
	MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
    }

    // Receive the sub-directory name
    //wprintf(L"Receiving the sub-directory name length\n");fflush(stdout);
    MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
    if (len != 0)
    {
	wchar_t *dir_name = new wchar_t[len];
	if (dir_name == NULL)
	{
	    MPI_Abort(MPI_COMM_WORLD, 11);
	}
	//wprintf(L"Received sub-dir name length: %d\nReceiving sub-dir name\n", len);fflush(stdout);
	MPI_Bcast(dir_name, (int)len, MPI_WCHAR, src_rank, MPI_COMM_WORLD);
	//wprintf(L"Received sub-dir name: %s\nCreating sub-directory\n", dir_name);fflush(stdout);
	if (!CreateDirectory(dir_name, NULL))
	{
	    DWORD error = GetLastError();
	    if (error != ERROR_ALREADY_EXISTS)
		MPI_Abort(MPI_COMM_WORLD, error);
	}
	//wprintf(L"Changing into sub-directory\n");fflush(stdout);
	if (!SetCurrentDirectory(dir_name))
	{
	    MPI_Abort(MPI_COMM_WORLD, 5);
	}
	delete [] dir_name;
	dir_name = NULL;
	ReceiveTreeRecursive(NULL, src_rank);
	//wprintf(L"Returning to parent directory\n");fflush(stdout);
	if (!SetCurrentDirectory(L".."))
	{
	    MPI_Abort(MPI_COMM_WORLD, 6);
	}
    }

    // Receive the sibling directory name
    //wprintf(L"Receiving sibling directory name length\n");fflush(stdout);
    MPI_Bcast(&len, 1, MPI_INT, src_rank, MPI_COMM_WORLD);
    if (len != 0)
    {
	wchar_t *dir_name = new wchar_t[len];
	if (dir_name == NULL)
	{
	    MPI_Abort(MPI_COMM_WORLD, 12);
	}
	//wprintf(L"Received sibling dir name length: %d\nReceiving sibling dir name\n", len);fflush(stdout);
	//wprintf(L"Returning to parent directory\n");fflush(stdout);
	if (!SetCurrentDirectory(L".."))
	{
	    MPI_Abort(MPI_COMM_WORLD, 9);
	}
	MPI_Bcast(dir_name, (int)len, MPI_WCHAR, src_rank, MPI_COMM_WORLD);
	//wprintf(L"Received sibling dir name: %s\nCreating sibling directory\n", dir_name);fflush(stdout);
	if (!CreateDirectory(dir_name, NULL))
	{
	    DWORD error = GetLastError();
	    if (error != ERROR_ALREADY_EXISTS)
		MPI_Abort(MPI_COMM_WORLD, error);
	}
	//wprintf(L"Changing into sibling directory\n");fflush(stdout);
	if (!SetCurrentDirectory(dir_name))
	{
	    MPI_Abort(MPI_COMM_WORLD, 8);
	}
	delete [] dir_name;
	dir_name = NULL;
	ReceiveTreeRecursive(NULL, src_rank);
    }

    if (buffer != NULL)
    {
	delete [] buffer;
	buffer = NULL;
    }
}

void FileTree::SendTree(FileTree *dest_tree)
{
    if (dest_tree != NULL)
    {
	dest_tree->CreateDestDir();
	if (!SetCurrentDirectory(dest_tree->GetDestDir()))
	{
	    MPI_Abort(MPI_COMM_WORLD, 14);
	}
    }
    SendTreeRecursive(root, src_rank, dest_tree);
}

void FileTree::ReceiveTree()
{
    if (!SetCurrentDirectory(dest_dir))
    {
	MPI_Abort(MPI_COMM_WORLD, 10);
    }
    ReceiveTreeRecursive(root, src_rank);
}

static void GetFiles(FileTreeDirNode *node, FileTreeFileNodeList *&iter_list)
{
    if (node == NULL)
	return;
    GetFiles(node->GetFirstChild(), iter_list);
    GetFiles(node->GetNextSibling(), iter_list);
    FileTreeFileNode *file_iter = node->GetFirstFile();
    while (file_iter != NULL)
    {
	FileTreeFileNodeList *temp = new FileTreeFileNodeList();
	if (temp == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 44);
	temp->file = file_iter;
	temp->next = iter_list;
	iter_list = temp;
	file_iter = file_iter->next;
    }
}

FileTreeFileNode *FileTree::GetFirstFile()
{
    FileTreeFileNodeList *temp;
    FileTreeFileNode *file;

    // Free any old nodes which can happen if GetFirstFile is called again before GetNextFile returns NULL
    while (iter_list != NULL)
    {
	temp = iter_list;
	iter_list = iter_list->next;
	delete temp;
    }

    GetFiles(root, iter_list);
    if (iter_list == NULL)
	return NULL;
    temp = iter_list;
    iter_list = iter_list->next;
    file = temp->file;
    delete temp;
    return file;
}

FileTreeFileNode *FileTree::GetNextFile()
{
    FileTreeFileNodeList *temp;
    FileTreeFileNode *file;
    if (iter_list == NULL)
	return NULL;
    temp = iter_list;
    iter_list = iter_list->next;
    file = temp->file;
    delete temp;
    return file;
}

FileTreeFileNodeList::FileTreeFileNodeList()
{
    file = NULL;
    next = NULL;
}
