#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <windows.h>
#include "FileTree.h"

#define MAX_BUFFER_SIZE 32*1024*1024

static FileTreeFileNodeList *saved_files = NULL;

static bool AlreadySaved(const wchar_t *filename)
{
    FileTreeFileNodeList *iter;
    iter = saved_files;
    while (iter != NULL)
    {
	if (iter->file != NULL)
	{
	    if (_wcsicmp(iter->file->GetFullFileName(), filename) == 0)
	    {
		return true;
	    }
	}
	iter = iter->next;
    }
    return false;
}

static void AddToSaved(const wchar_t *filename)
{
    FileTreeFileNodeList *node = new FileTreeFileNodeList();
    if (node == NULL)
    {
	// Failure
	MPI_Abort(MPI_COMM_WORLD, 45);
	return;
    }
    node->file = new FileTreeFileNode();
    if (node->file == NULL)
    {
	// Failure
	delete node;
	MPI_Abort(MPI_COMM_WORLD, 46);
	return;
    }
    node->file->SetFileName(filename);
    node->next = saved_files;
    saved_files = node;
}

static void ClearSaved()
{
    FileTreeFileNodeList *temp;
    while (saved_files != NULL)
    {
	temp = saved_files;
	saved_files = saved_files->next;
	if (temp->file != NULL)
	    delete temp->file;
	delete temp;
    }
}

static void GenerateUniqueName(wchar_t *&filename, int rank, int size, SYSTEMTIME job_time)
{
    wchar_t *name, *ext, *unique;
    int len;

    name = _wcsdup(filename);
    ext = wcsrchr(name, L'.');
    if (ext != NULL)
    {
	*ext = L'\0';
	ext++;
    }
    len = (int)wcslen(filename) + 100;
    unique = new wchar_t[len];
    if (unique == NULL)
	MPI_Abort(MPI_COMM_WORLD, 21);

    // First try a name with the rank and size added to it
    if (ext != NULL)
	swprintf_s(unique, len, L"%s.%d.%d.%s", name, rank, size, ext);
    else
	swprintf_s(unique, len, L"%s.%d.%d", name, rank, size);
    HANDLE hFile = CreateFile(unique, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
	// File does not exist so it's a good name
	free(name);
	delete [] filename;
	filename = unique;
	return;
    }
    CloseHandle(hFile);

    // Then try a name with the rank and size and a time stamp added to it
    if (ext != NULL)
	swprintf_s(unique, len, L"%s.%d.%d.%d-%d-%d.%dh%dm%ds.%s", name, rank, size, job_time.wYear, job_time.wMonth, job_time.wDay, job_time.wHour, job_time.wMinute, job_time.wSecond, ext);
    else
	swprintf_s(unique, len, L"%s.%d.%d.%d-%d-%d.%dh%dm%ds", name, rank, size, job_time.wYear, job_time.wMonth, job_time.wDay, job_time.wHour, job_time.wMinute, job_time.wSecond);
    hFile = CreateFile(unique, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
	// File does not exist so it's a good name
	free(name);
	delete [] filename;
	filename = unique;
	return;
    }
    CloseHandle(hFile);

    // Then try names with the rank and size and timestamp and an extra number added to the end of it
    // Continue increasing the last number until a unique name is generated
    int i=0;
    for (;;)
    {
	if (ext != NULL)
	    swprintf_s(unique, len, L"%s.%d.%d.%d-%d-%d.%dh%dm%ds.%d.%s", name, rank, size, job_time.wYear, job_time.wMonth, job_time.wDay, job_time.wHour, job_time.wMinute, job_time.wSecond, i, ext);
	else
	    swprintf_s(unique, len, L"%s.%d.%d.%d-%d-%d.%dh%dm%ds.%d", name, rank, size, job_time.wYear, job_time.wMonth, job_time.wDay, job_time.wHour, job_time.wMinute, job_time.wSecond, i);
	hFile = CreateFile(unique, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
	    // File does not exist so it's a good name
	    free(name);
	    delete [] filename;
	    filename = unique;
	    return;
	}
	CloseHandle(hFile);
    }
}

int wmain(int argc, wchar_t *argv[])
{
    int i, src_rank = 0, dest_rank = 0;
    bool recursive = false;
    bool pull = true;
    bool ignore_src = false;
    wchar_t src_dir[4096] = L"";
    wchar_t dest_dir[4096] = L"";
    int rank, size;
    bool use_wc = false;
    wchar_t wc[256] = L"*.*";
    bool use_regex = false;
    wchar_t regex[256] = L".*";
    size_t len;
    FileTree *src_tree, *dest_tree = NULL, *dest_tree1, *dest_tree2, *diff_tree;
    SYSTEMTIME job_time;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if (argc < 2)
    {
	if (rank == 0)
	{
	    //wprintf(L"Usage: deino_stager.exe src_dir /src:rank /dest:rank /dest_dir:directory /recursive /nopull /wc:pattern /regex:pattern\n");
	    wprintf(L"Usage: deino_stager.exe src_dir [/src:rank] [/dest:rank] [/dest_dir:directory] [/recursive] [/nopull] [/ignore_src] [/wc:pattern]\n");
	    wprintf(L"  /src:rank - the rank in MPI_COMM_WORLD where the source directory resides.\n");
	    wprintf(L"  /dest:rank - the rank where the modified files will be collected.\n");
	    wprintf(L"  /dest_dir:directory - the root directory where the files will be copied to.  If the directory does not exist it will be created.\n");
	    wprintf(L"  /recursive - copy sub-directories and files.\n");
	    wprintf(L"  /nopull - do not copy the modified files back to the source rank.\n");
	    wprintf(L"  /ignore_src - ignore the destination directory at the source.  This is useful for copying data from a node that will not participate in the job.\n");
	    wprintf(L"  /wc:pattern - use the specified wild card pattern to select files to copy.\n");
	    fflush(stdout);
	}
	MPI_Finalize();
	return 0;
    }

    for (i=1; i<argc; i++)
    {
	wchar_t *temp_arg = _wcsdup(argv[i]);
	size_t n;
	len = wcslen(temp_arg);
	for (n=0; n<len; n++)
	{
	    temp_arg[n] = towupper(temp_arg[n]);
	}
	if (wcscmp(temp_arg, L"/RECURSIVE") == 0)
	{
	    recursive = true;
	}
	else if (wcscmp(temp_arg, L"/NOPULL") == 0)
	{
	    pull = false;
	}
	else if (wcscmp(temp_arg, L"/IGNORE_SRC") == 0)
	{
	    ignore_src = true;
	}
	else if ((len > 5) && (temp_arg[0] == L'/') && (temp_arg[1] == L'S') && (temp_arg[2] == L'R') && (temp_arg[3] == L'C') && (temp_arg[4] == L':'))
	{
	    src_rank = _wtol(&temp_arg[5]);
	    if ((src_rank < 0) || (src_rank >= size))
		src_rank = 0;
	}
	else if ((len > 6) && (temp_arg[0] == L'/') && (temp_arg[1] == L'D') && (temp_arg[2] == L'E') && (temp_arg[3] == L'S') && (temp_arg[4] == L'T') && (temp_arg[5] == L':'))
	{
	    dest_rank = _wtol(&temp_arg[6]);
	    if ((dest_rank < 0) || (dest_rank >= size))
		dest_rank = 0;
	}
	else if ((len > 10) && (temp_arg[0] == L'/') && (temp_arg[1] == L'D') && (temp_arg[2] == L'E') && (temp_arg[3] == L'S') && (temp_arg[4] == L'T') && (temp_arg[5] == L'_') && (temp_arg[6] == L'D') && (temp_arg[7] == L'I') && (temp_arg[8] == L'R') && (temp_arg[9] == L':'))
	{
	    wcsncpy_s(dest_dir, 4096, &argv[i][10], _TRUNCATE);
	}
	else if ((len > 4) && (temp_arg[0] == L'/') && (temp_arg[1] == L'W') && (temp_arg[2] == L'C') && (temp_arg[3] == L':'))
	{
	    use_wc = true;
	    use_regex = false;
	    wcsncpy_s(wc, 256, &argv[i][4], _TRUNCATE);
	}
	else if ((len > 7) && (temp_arg[0] == L'/') && (temp_arg[1] == L'R') && (temp_arg[2] == L'E') && (temp_arg[3] == L'G') && (temp_arg[4] == L'E') && (temp_arg[5] == L'X') && (temp_arg[6] == L':'))
	{
	    use_regex = true;
	    use_wc = false;
	    wcsncpy_s(regex, 256, &argv[i][7], _TRUNCATE);
	}
	else if ((len > 0) && (temp_arg[0] == L'/'))
	{
	    wprintf(L"Unknown option: %s\n", argv[i]);
	}
	else
	{
	    wcsncpy_s(src_dir, 4096, argv[i], _TRUNCATE);
	}
	free(temp_arg);
    }

    if (src_dir[0] == L'\0')
    {
	wprintf(L"No source directory specified.\n");
	fflush(stdout);
	MPI_Abort(MPI_COMM_WORLD, 1);
    }

    if (dest_dir[0] == L'\0')
    {
	wcsncpy_s(dest_dir, 4096, src_dir, _TRUNCATE);
    }

    // Create a tree of the source files
    if (rank == src_rank)
    {
	/*
	wprintf(L"Source rank: %d\n", src_rank);
	wprintf(L"Destination rank: %d\n", dest_rank);
	wprintf(L"Recursive: %s\n", recursive ? L"true" : L"false");
	wprintf(L"Pull: %s\n", pull ? L"true" : L"false");
	wprintf(L"Source directory: %s\n", src_dir);
	wprintf(L"Destination directory: %s\n", dest_dir);
	if (use_wc)
	    wprintf(L"Wildcard pattern: %s\n", wc);
	if (use_regex)
	    wprintf(L"Regular expression: %s\n", regex);
	wprintf(L"\n");
	fflush(stdout);
	*/

	src_tree = new FileTree(src_dir, recursive);
	if (src_tree == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 47);
	src_tree->SetSrcRank(src_rank);
	src_tree->SetSrcDir(src_dir);
	src_tree->PopulateTreeFromSrc(wc);
	//src_tree->Print();
	//wprintf(L"\n");
    }

    // Create the destination directory
    if ((!ignore_src) || (rank != src_rank))
    {
	dest_tree = new FileTree(dest_dir, recursive);
	if (dest_tree == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 48);
	dest_tree->SetSrcRank(src_rank);
	dest_tree->SetDestDir(dest_dir);
	dest_tree->CreateDestDir();
    }

    // Send the source files to all the destination directories
    bool copy_at_source = false;
    if (wcscmp(src_dir, dest_dir) != 0)
	copy_at_source = true;
    if (ignore_src)
	copy_at_source = false;
    if (rank == src_rank)
    {
	if (copy_at_source)
	    src_tree->SendTree(dest_tree);
	else
	    src_tree->SendTree();
    }
    else
    {
	dest_tree->ReceiveTree();
    }

    if (pull)
    {
	if (rank != src_rank || !ignore_src)
	{
	    // Save the destination directory files and attributes
	    dest_tree1 = new FileTree(dest_dir, recursive);
	    if (dest_tree1 == NULL)
		MPI_Abort(MPI_COMM_WORLD, 48);
	    dest_tree1->SetSrcRank(src_rank);
	    dest_tree1->SetDestDir(dest_dir);
	    dest_tree1->PopulateTreeFromDest(wc);
	}
	else
	{
	    dest_tree1 = NULL;
	}

	// Ready, wait for Exit
	MPI_Barrier(MPI_COMM_WORLD);
	GetLocalTime(&job_time);
	if (rank == 0)
	{
	    wprintf(L"READY\n");
	    fflush(stdout);
	    wchar_t line[100];
	    if (fgetws(line, 100, stdin) == NULL)
	    {
		MPI_Abort(MPI_COMM_WORLD, 13);
	    }
	}
	// At this point everyone waits in the barrier until process 0 receives the signal to continue
	MPI_Barrier(MPI_COMM_WORLD);

	if (rank != src_rank || !ignore_src)
	{
	    // Get the current state of the destination directory
	    dest_tree2 = new FileTree(dest_dir, recursive);
	    if (dest_tree2 == NULL)
		MPI_Abort(MPI_COMM_WORLD, 49);
	    dest_tree2->SetSrcRank(src_rank);
	    dest_tree2->SetDestDir(dest_dir);
	    dest_tree2->PopulateTreeFromDest(wc);

	    // Get the difference between the original and current destination directory
	    diff_tree = dest_tree1->GetChangedFiles(dest_tree2);
	    /*
	    wprintf(L"dest_tree1\n");
	    dest_tree1->Print();
	    wprintf(L"dest_tree2\n");
	    dest_tree2->Print();
	    wprintf(L"diff_tree1\n");
	    diff_tree->Print();
	    fflush(stdout);
	    */
	}
	else
	{
	    dest_tree2 = NULL;
	    diff_tree = NULL;
	}

	// Send the files that have changed back to the source rank
	if (rank == src_rank)
	{
	    // First check to see if any files have been been changed in place at the source and add them to the saved list
	    if ((!copy_at_source) && (!ignore_src) && (diff_tree != NULL))
	    {
		// Making it inside this if statement means that:
		// 1) src_dir and dest_dir are the same at the source rank.
		// 2) The source rank is not ignored (meaning the source rank participated in the job).
		// 3) Files have been modified.
		// So add all the modified files to the saved list so they won't be overwritten by files from the remote hosts.
		FileTreeFileNode *file_iter;
		file_iter = diff_tree->GetFirstFile();
		while (file_iter != NULL)
		{
		    AddToSaved(file_iter->GetFullFileName());
		    file_iter = diff_tree->GetNextFile();
		}
	    }

	    // Receive changed files
	    int len;
	    size_t dest_dir_len;
	    wchar_t *filename;
	    bool add_separator = false;
	    unsigned char *buffer = new unsigned char[MAX_BUFFER_SIZE];
	    if (buffer == NULL)
		MPI_Abort(MPI_COMM_WORLD, 19);
	    dest_dir_len = wcslen(dest_dir);
	    if (dest_dir[dest_dir_len-1] != L'\\')
		dest_dir_len++; // If the dest_dir does not end in a path separator then add one to the length so the path separator will be skippped
	    len = (int)wcslen(src_dir);
	    if (src_dir[len-1] != L'\\')
		add_separator = true;
	    for (int i=0; i<size; i++)
	    {
		if (i == src_rank)
		{
		    if (copy_at_source)
		    {
			// Copy the destination directory to the source directory
			if (diff_tree != NULL)
			{
			    FileTreeFileNode *file_iter;
			    //wprintf(L"Rank %d difference files:\n", rank);
			    file_iter = diff_tree->GetFirstFile();
			    while (file_iter != NULL)
			    {
				const wchar_t *local_filename2 = file_iter->GetFullFileName();
				//wprintf(L"%s\n", file_iter->GetFullFileName());fflush(stdout);
				if (local_filename2 != NULL)
				{
				    len = (int)wcslen(src_dir) + (int)wcslen(&local_filename2[dest_dir_len]) + 1 + (add_separator ? 1 : 0);
				    wchar_t *local_filename = new wchar_t[len];
				    if (local_filename == NULL)
					MPI_Abort(MPI_COMM_WORLD, 21);
				    swprintf_s(local_filename, len, L"%s%s%s", src_dir, add_separator ? L"\\" : L"", &local_filename2[dest_dir_len]);
				    if (AlreadySaved(local_filename))
				    {
					GenerateUniqueName(local_filename, i, size, job_time);
					CopyFile(local_filename2, local_filename, TRUE);
				    }
				    else
				    {
					AddToSaved(local_filename);
					CopyFile(local_filename2, local_filename, FALSE);
				    }
				    delete [] local_filename;
				}
				file_iter = diff_tree->GetNextFile();
			    }
			}
		    }
		}
		else
		{
		    // Receive modified files from rank i
		    MPI_Status status;
		    MPI_Recv(&len, 1, MPI_INT, i, 100, MPI_COMM_WORLD, &status);
		    //wprintf(L"received file name length: %d\n", len);fflush(stdout);
		    while (len != 0)
		    {
			filename = new wchar_t[len];
			if (filename == NULL)
			    MPI_Abort(MPI_COMM_WORLD, 19);
			MPI_Recv((void*)filename, len, MPI_WCHAR, i, 100, MPI_COMM_WORLD, &status);
			//wprintf(L"Receiving file: %s%s%s\n", src_dir, add_separator ? L"\\" : L"", &filename[dest_dir_len]);fflush(stdout);
			len = (int)wcslen(src_dir) + (int)wcslen(&filename[dest_dir_len]) + 1 + (add_separator ? 1 : 0);
			wchar_t *local_filename = new wchar_t[len];
			if (local_filename == NULL)
			    MPI_Abort(MPI_COMM_WORLD, 20);
			swprintf_s(local_filename, len, L"%s%s%s", src_dir, add_separator ? L"\\" : L"", &filename[dest_dir_len]);
			delete [] filename;
			HANDLE hFile;
			if (AlreadySaved(local_filename))
			{
			    GenerateUniqueName(local_filename, i, size, job_time);
			    hFile = CreateFile(local_filename, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);
			}
			else
			{
			    AddToSaved(local_filename);
			    hFile = CreateFile(local_filename, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
			}
			//wprintf(L"Receiving file: %s\n", local_filename);
			delete [] local_filename;
			int num_read = 0;
			MPI_Recv(&num_read, 1, MPI_INT, i, 100, MPI_COMM_WORLD, &status);
			//wprintf(L"received num_read: %d\n", num_read);fflush(stdout);
			while (num_read != 0)
			{
			    MPI_Recv((void*)buffer, num_read, MPI_BYTE, i, 100, MPI_COMM_WORLD, &status);
			    //wprintf(L"received %d bytes\n", num_read);fflush(stdout);
			    if (hFile != INVALID_HANDLE_VALUE)
			    {
				DWORD num_written = 0;
				WriteFile(hFile, buffer, num_read, &num_written, NULL);
			    }
			    MPI_Recv(&num_read, 1, MPI_INT, i, 100, MPI_COMM_WORLD, &status);
			    //wprintf(L"received num_read: %d\n", num_read);fflush(stdout);
			}
			if (hFile != INVALID_HANDLE_VALUE)
			    CloseHandle(hFile);
			MPI_Recv(&len, 1, MPI_INT, i, 100, MPI_COMM_WORLD, &status);
			//wprintf(L"received file name length: %d\n", len);fflush(stdout);
		    }
		}
	    }
	}
	else
	{
	    int len;
	    const wchar_t *filename;
	    unsigned char *buffer = new unsigned char[MAX_BUFFER_SIZE];
	    if (buffer == NULL)
		MPI_Abort(MPI_COMM_WORLD, 19);
	    // Send modified files to src_rank
	    if (diff_tree != NULL)
	    {
		//wprintf(L"Rank %d difference tree:\n", rank);
		//diff_tree->Print();
		FileTreeFileNode *file_iter;
		//wprintf(L"Rank %d difference files:\n", rank);
		file_iter = diff_tree->GetFirstFile();
		while (file_iter != NULL)
		{
		    filename = file_iter->GetFullFileName();
		    HANDLE hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		    if (hFile != INVALID_HANDLE_VALUE)
		    {
			//wprintf(L"%s\n", filename);fflush(stdout);
			len = (int)wcslen(filename) + 1;
			MPI_Send(&len, 1, MPI_INT, 0, 100, MPI_COMM_WORLD);
			MPI_Send((void*)filename, len, MPI_WCHAR, 0, 100, MPI_COMM_WORLD);
			DWORD num_read;
			ReadFile(hFile, buffer, MAX_BUFFER_SIZE, &num_read, NULL);
			while (num_read != 0)
			{
			    // Send number of bytes read
			    //wprintf(L"Sending num_read: %d\n", num_read);fflush(stdout);
			    MPI_Send(&num_read, 1, MPI_INT, 0, 100, MPI_COMM_WORLD);
			    // Send bytes
			    //wprintf(L"Sending bytes\n");fflush(stdout);
			    MPI_Send(buffer, num_read, MPI_BYTE, 0, 100, MPI_COMM_WORLD);
			    ReadFile(hFile, buffer, MAX_BUFFER_SIZE, &num_read, NULL);
			}
			// Send zero bytes read to indicate the end of the file
			//wprintf(L"Sending zero num_read to indicate end of file\n");fflush(stdout);
			MPI_Send(&num_read, 1, MPI_INT, 0, 100, MPI_COMM_WORLD);
			CloseHandle(hFile);
		    }
		    else
		    {
			wprintf(L"Error opening file %s: %d\n", filename, GetLastError());fflush(stdout);
		    }
		    file_iter = diff_tree->GetNextFile();
		}
		// Send zero length file name to indicate no more files
		len = 0;
		//wprintf(L"Sending zero file name length to indicate no more files\n");fflush(stdout);
		MPI_Send(&len, 1, MPI_INT, 0, 100, MPI_COMM_WORLD);
	    }
	}
    }
    else
    {
	// Add a barrier here just in case MPI_Finalize does not act as a barrier.
	// This shouldn't really be necessary because in the non-pull case the caller should wait for mpiexec to return
	// which would indicate that all the processes have exited and all the files have been pushed out.
	// If mpiexec is allowed to exit before the processes do then there really isn't anything that can be done.
	MPI_Barrier(MPI_COMM_WORLD);
	if (rank == 0)
	{
	    wprintf(L"READY\n");
	    fflush(stdout);
	}
    }

    MPI_Finalize();
    return 0;
}
