using System;
using System.Collections.Generic;
using System.Text;
using DeinoMPI;
using System.Collections;

namespace manage_credentials
{
    class manage_credentials
    {
        static void PrintUsage()
        {
            Console.WriteLine("Usage:");
            Console.WriteLine(" manage_credentials /add [username] [password]");
            Console.WriteLine("    Add a credential to the store.");
            Console.WriteLine("    If you do not enter a username and or password you will be prompted to");
            Console.WriteLine("    enter these parameters.  If a credential with the same username already");
            Console.WriteLine("    exists in the store it will be replaced.");
            Console.WriteLine(" manage_credentials /remove [username]");
            Console.WriteLine("    Remove the credential from the store.");
            Console.WriteLine(" manage_credentials /list");
            Console.WriteLine("    List all the usernames in the store.");
            Console.WriteLine(" manage_credentials /replace_keys");
            Console.WriteLine("    Create a new private and public key and re-encrypt all the existing");
            Console.WriteLine("    credentials in the store.");
            Console.WriteLine(" manage_credentials /set_location [path or \"registry\"]");
            Console.WriteLine("    Point the current user to the Credential store specified.");
            Console.WriteLine("    Provide either the full path to the Credential store file or registry");
            Console.WriteLine("    to refer to the Windows Registry.");
        }

        static string GetPassword(string message)
        {
            StringBuilder entered_password = new StringBuilder();
            Console.WriteLine(message);

            try
            {
                // get the first character of the password
                ConsoleKeyInfo nextKey = Console.ReadKey(true);

                while (nextKey.Key != ConsoleKey.Enter)
                {
                    if (nextKey.Key == ConsoleKey.Backspace)
                    {
                        if (entered_password.Length > 0)
                        {
                            entered_password.Remove(entered_password.Length - 1, 1);

                            // erase the last * as well
                            Console.Write(nextKey.KeyChar);
                            Console.Write(" ");
                            Console.Write(nextKey.KeyChar);
                        }
                    }
                    else
                    {
                        entered_password.Append(nextKey.KeyChar);
                        Console.Write("*");
                    }

                    nextKey = Console.ReadKey(true);
                }
            }
            catch (InvalidOperationException)
            {
                // The console has been redirected so read a line from the redirected input
                entered_password = new StringBuilder(Console.In.ReadLine());
            }

            Console.WriteLine();

            return entered_password.ToString();
        }

        static string GetFlag(string full_flag)
        {
            if (full_flag == null)
                return full_flag;
            if (full_flag.Length > 2 && full_flag[0] == '-' && full_flag[1] == '-')
                return full_flag.Remove(0, 2);
            if (full_flag.Length > 1 && (full_flag[0] == '-' || full_flag[0] == '/'))
                return full_flag.Remove(0, 1);
            return full_flag;
        }

        static void Main(string[] args)
        {
            string flag;
            // Validate the arguments
            if (args.Length < 1)
            {
                PrintUsage();
                return;
            }
            flag = GetFlag(args[0]);
            switch (flag)
            {
                case "add":
                    break;
                case "remove":
                    break;
                case "list":
                    break;
                case "replace_keys":
                    break;
                case "set_location":
                    break;
                case "get_location":
                    break;
                default:
                    PrintUsage();
                    return;
            }

            string password = null;
            CredentialStore store = null;
            if (flag != "set_location" && flag != "get_location")
            {
                if (CredentialStore.RequiresPassword())
                    password = GetPassword("Please enter the passphrase used to protect the keys in the Credential Store:");
                try
                {
                    store = new CredentialStore(password);
                }
                catch (Exception e)
                {
                    Console.WriteLine("Unable to open the Credential Store: {0}", e.Message);
                    return;
                }
            }

            switch (flag)
            {
                case "add":
                    {
                        SecureCredential credential = new SecureCredential(store.Keys);
                        if (args.Length > 1)
                            credential.Account = args[1];
                        else
                        {
                            string current_user = Environment.UserDomainName + "\\" + Environment.UserName;
                            Console.Write("Enter the account name: [{0}] ", current_user);
                            string result = Console.ReadLine().Trim();
                            if (result == "")
                                result = current_user;
                            credential.Account = result;
                        }
                        if (args.Length > 2)
                            credential.Password = args[2];
                        else
                        {
                            string pwd1, pwd2;
                            do
                            {
                                pwd1 = GetPassword("Enter the account password:");
                                pwd2 = GetPassword("Enter the password again to verify:");
                                if (pwd1 != pwd2)
                                    Console.WriteLine("The passwords aren't the same");
                            } while (pwd1 != pwd2);
                            credential.Password = pwd1;
                        }
                        store.AddCredential(credential);
                    }
                    break;
                case "remove":
                    if (args.Length > 1)
                        store.RemoveCredential(args[1]);
                    else
                    {
                        if (store.Count > 0)
                        {
                            Console.WriteLine("Existing accounts in the store:");
                            foreach (SecureCredential cred in store)
                            {
                                Console.WriteLine(" {0}", cred.Account);
                            }
                            Console.Write("Enter an account to remove: ");
                            string account = Console.ReadLine().Trim();
                            store.RemoveCredential(account);
                        }
                        else
                        {
                            Console.WriteLine("The store contains no credentials.");
                        }
                    }
                    break;
                case "list":
                    foreach (SecureCredential credential in store)
                    {
                        Console.WriteLine(credential.Account);
                    }
                    break;
                case "replace_keys":
                    // Save the information describing the current store
                    string filename = null;
                    bool use_registry = store.UseRegistry;
                    if (!use_registry)
                        filename = store.FileName;
                    Hashtable old_credentials = new Hashtable();
                    foreach (SecureCredential cred in store)
                        old_credentials[cred.Account] = cred.Password;

                    Console.WriteLine("This option replaces the public and private keys in the Credential Store.");
                    if (use_registry)
                        Console.WriteLine("The current Credential Store is located in the registry.");
                    else
                    {
                        Console.WriteLine("The current Credential Store is located here:");
                        Console.WriteLine(filename);
                    }
                    Console.WriteLine();
                    Console.WriteLine("Do you want to replace the keys in this store (yes,no)? [no]");
                    string answer = Console.ReadLine().Trim().ToLower();
                    if (answer == "yes")
                    {
                        SecurityKeys keys = new SecurityKeys();
                        keys.GetPassword(true);
                        keys.GetEncryptionMethod(true);
                        keys.CreateKeys();

                        CredentialStore new_store;
                        if (use_registry)
                            new_store = new CredentialStore(keys);
                        else
                            new_store = new CredentialStore(keys, filename);
                        foreach (string key in old_credentials.Keys)
                        {
                            SecureCredential cred = new SecureCredential(keys);
                            cred.Account = key;
                            cred.Password = (string)old_credentials[key];
                            new_store.AddCredential(cred);
                        }
                        Console.WriteLine("The Credential Store has been updated.");
                        Console.WriteLine("The new public key will need to be distributed before this store can be used.");
                    }
                    break;
                case "set_location":
                    if (args.Length < 2)
                    {
                        Console.WriteLine("Please specify the location of the Credential store.  Provide either the full path to the file or \"registry\" to refer to the Window's Registry.");
                        break;
                    }
                    if (CredentialStore.SetLocation(args[1]))
                    {
                        Console.WriteLine("Credential store location set: {0}", args[1]);
                    }
                    else
                    {
                        Console.WriteLine("Unable to set the location of the Credential store.");
                    }
                    break;
                case "get_location":
                    string location = CredentialStore.GetLocation();
                    if (location == null)
                    {
                        Console.WriteLine("There is no Credential Store specified.");
                        Console.WriteLine("Use create_credential_store.exe to create a new Credential Store for the current user.");
                        Console.WriteLine("Or use \"manage_credentials /set_location <path-to-store>\" to specifiy the location of an existing Credential Store.");
                    }
                    else
                        Console.WriteLine(location);
                    break;
            }
        }
    }
}
