using System;
using System.Collections.Generic;
using System.Text;
using DeinoMPI;

namespace manage_public_keys
{
    class Program
    {
        static void PrintUsage()
        {
            Console.WriteLine("Usage:");
            Console.WriteLine(" manage_public_keys /export <filename>");
            Console.WriteLine("    Export the public key from the current user's Credential Store.");
            Console.WriteLine();
            Console.WriteLine(" manage_public_keys /export <filename> <CredentialStore filename>");
            Console.WriteLine("    Export the public key from the specified Credential Store.");
            Console.WriteLine();
            Console.WriteLine(" manage_public_keys /import <filename>");
            Console.WriteLine("    Import the public key from the specified file and save it in the");
            Console.WriteLine("    PublicKey store for the local machine.");
            Console.WriteLine();
            Console.WriteLine(" manage_public_keys /list");
            Console.WriteLine("    List the public keys in the PublicKey store for the local machine.");
            Console.WriteLine();
            Console.WriteLine(" manage_public_keys /remove <public key hash>");
            Console.WriteLine("    Remove the public key in the PublicKey store for the local machine");
            Console.WriteLine("    that corresponds to the specified key hash.");
            Console.WriteLine();
            Console.WriteLine(" manage_public_keys /clear_all");
            Console.WriteLine("    Remove all public keys in the PublicKey store for the local machine.");
            Console.WriteLine();
            Console.WriteLine(" manage_public_keys /auto_keys <true or false>");
            Console.WriteLine("    true - Allow the local machine to accept new public keys automatically.");
            Console.WriteLine("    false - Do not accept new public keys.  Keys must be individually allowed");
            Console.WriteLine("            on each machine using the /import option.");
        }

        static string GetPassword(string message)
        {
            StringBuilder entered_password = new StringBuilder();
            Console.WriteLine(message);

            try
            {
                // get the first character of the password
                ConsoleKeyInfo nextKey = Console.ReadKey(true);

                while (nextKey.Key != ConsoleKey.Enter)
                {
                    if (nextKey.Key == ConsoleKey.Backspace)
                    {
                        if (entered_password.Length > 0)
                        {
                            entered_password.Remove(entered_password.Length - 1, 1);

                            // erase the last * as well
                            Console.Write(nextKey.KeyChar);
                            Console.Write(" ");
                            Console.Write(nextKey.KeyChar);
                        }
                    }
                    else
                    {
                        entered_password.Append(nextKey.KeyChar);
                        Console.Write("*");
                    }

                    nextKey = Console.ReadKey(true);
                }
            }
            catch (InvalidOperationException)
            {
                // The console has been redirected so read a line from the redirected input
                entered_password = new StringBuilder(Console.In.ReadLine());
            }

            Console.WriteLine();

            return entered_password.ToString();
        }

        static string GetFlag(string full_flag)
        {
            if (full_flag == null)
                return full_flag;
            if (full_flag.Length > 2 && full_flag[0] == '-' && full_flag[1] == '-')
                return full_flag.Remove(0, 2);
            if (full_flag.Length > 1 && (full_flag[0] == '-' || full_flag[0] == '/'))
                return full_flag.Remove(0, 1);
            return full_flag;
        }

        static void Main(string[] args)
        {
            string flag;
            // Validate the arguments
            if (args.Length < 1)
            {
                PrintUsage();
                return;
            }
            flag = GetFlag(args[0]);
            switch (flag)
            {
                case "export":
                    if (args.Length < 2)
                    {
                        PrintUsage();
                        return;
                    }
                    break;
                case "import":
                    if (args.Length < 2)
                    {
                        PrintUsage();
                        return;
                    }
                    break;
                case "list":
                    break;
                case "remove":
                    if (args.Length < 2)
                    {
                        PrintUsage();
                        return;
                    }
                    break;
                case "clear_all":
                    break;
                case "auto_keys":
                    if (args.Length < 2)
                    {
                        PrintUsage();
                        return;
                    }
                    break;
                default:
                    PrintUsage();
                    return;
            }

            if (flag == "export")
            {
                string password = null;
                CredentialStore store = null;
                if (args.Length > 2)
                {
                    if (CredentialStore.RequiresPassword(args[2]))
                        password = GetPassword("Please enter the passphrase used to protect the keys in the Credential Store:");
                    try
                    {
                        store = new CredentialStore(password, args[2]);
                    }
                    catch (Exception e)
                    {
                        Console.WriteLine("Unable to open the Credential Store: {0}", e.Message);
                        return;
                    }
                }
                else
                {
                    if (CredentialStore.RequiresPassword())
                        password = GetPassword("Please enter the passphrase used to protect the keys in the Credential Store:");
                    try
                    {
                        store = new CredentialStore(password);
                    }
                    catch (Exception e)
                    {
                        Console.WriteLine("Unable to open the Credential Store: {0}", e.Message);
                        return;
                    }
                }
                if (store.Keys.ToKeyFile(args[1], false))
                {
                    Console.WriteLine("Public key written to file:");
                    Console.WriteLine(args[1]);
                }
                else
                {
                    Console.WriteLine("Unable to save the public key.");
                }
            }
            else if (flag == "import")
            {
                // import
                SecurityKeys keys = new SecurityKeys();
                if (keys.FromKeyFile(args[1]))
                {
                    if (PublicKeyStore.SavePublicKey(keys))
                    {
                        Console.WriteLine("Public key imported: {0}", keys.Hash);
                    }
                    else
                    {
                        Console.WriteLine("Unable to save the public key.");
                        Console.WriteLine("This is most likely due to restricted access to the Registry where the keys are stored.");
                        Console.WriteLine("You may need administrator privileges to save public keys.");
                    }
                }
                else
                {
                    Console.WriteLine("Unable to retrieve a public key from the specified file:");
                    Console.WriteLine(args[1]);
                }
            }
            else if (flag == "list")
            {
                string[] hashes = PublicKeyStore.GetPublicKeyHashes();
                if (hashes != null)
                {
                    foreach (string hash in hashes)
                    {
                        Console.WriteLine(hash);
                    }
                }
                else
                {
                    Console.WriteLine("No public keys in the PublicKey store.");
                }
            }
            else if (flag == "remove")
            {
                if (PublicKeyStore.RemovePublicKey(args[1]))
                {
                    Console.WriteLine("Public key removed.");
                }
                else
                {
                    Console.WriteLine("Unable to remove the public key.");
                }
            }
            else if (flag == "clear_all")
            {
                if (PublicKeyStore.ClearAllKeys())
                {
                    Console.WriteLine("All public keys removed from the PublicKey store.");
                }
                else
                {
                    Console.WriteLine("Unable to clear the public keys from the PublicKey store.");
                }
            }
            else if (flag == "auto_keys")
            {
                string option = args[1].Trim().ToLower();
                if (option == "true" || option == "1" || option == "on" || option == "+")
                {
                    PublicKeyStore.SetAutoAcceptPublicKeys(true);
                }
                else if (option == "false" || option == "0" || option == "off" || option == "-")
                {
                    PublicKeyStore.SetAutoAcceptPublicKeys(false);
                }
                else
                {
                    Console.WriteLine("Invalid option after the auto_keys flag: {0}", args[1]);
                }
            }
        }
    }
}
