using System;
using System.Collections.Generic;
using System.Text;
using DeinoMPI;
using System.Collections;
using System.Threading;
using Microsoft.Win32;
using System.IO;
using System.Diagnostics;
using System.ComponentModel;

namespace mpiexec
{
    class MpiexecProcess
    {
        const string CredentialRequiredMessage = "In order to retrieve host information an account must be selected from the Credential store first.";
        static VerifiedNode[] vnode;
        static Mpiexec mpiexec;

        static void UpdateVerifyNode(int index, LaunchInfoMessage message)
        {
            vnode[index].message = message;
            if (message.bad_exe_path != "" || message.missing_dlls != "")
                vnode[index].state = VerifiedNode.VerificationState.BAD;
            else if (message.bad_wdir != "" || message.missing_delayed_dlls != "")
                vnode[index].state = VerifiedNode.VerificationState.OK;
            else
                vnode[index].state = VerifiedNode.VerificationState.GOOD;
        }

        // code in verify_thread must be synchronized with deinopm_thread_fn and DeinoMPIWin.verify_thread
        static void verify_thread(object o)
        {
            LaunchNode node = null;
            try
            {
                ClientSession client;
                node = (LaunchNode)o;
                try
                {
                    if (mpiexec.global_options.localonly || node.launch_local)
                    {
                        int port = DeinoPMUtil.StartLocalPM(mpiexec, node.private_key);
                        client = new ClientSession(port, node.private_key);
                    }
                    else
                    {
                        client = new ClientSession(node.hostname, -1, mpiexec.global_options.phrase, mpiexec.store.Keys, node.credential);
                    }
                }
                catch (Exception e)
                {
                    StringBuilder sb = new StringBuilder("ERROR: ");
                    sb.Append(e.Message);
                    foreach (string key in e.Data.Keys)
                    {
                        sb.AppendFormat("\r\n{0}={1}", key, e.Data[key]);
                    }
                    vnode[node.iproc].error_msg = sb.ToString();
                    vnode[node.iproc].state = VerifiedNode.VerificationState.BAD;
                    return;
                }
                PMI pmi;
                pmi = node.pmi;
                LaunchMessage launch = new LaunchMessage();

                // create a launch message from the launch node

                // add the process fields
                launch.path = node.path;
                foreach (string drive in node.map.Keys)
                {
                    launch.map[drive] = node.map[drive];
                }
                launch.dir = node.dir;
                launch.exe = node.exe;
                foreach (string arg in node.args)
                {
                    launch.args.Add(arg);
                }
                foreach (string key in node.env.Keys)
                {
                    launch.env[key] = node.env[key];
                }
                launch.priority_class = node.priority_class;
                launch.priority_thread = node.priority_thread;

                // add the pmi variables
                if (node.spawned)
                    launch.env["PMI_SPAWN"] = "1";
                launch.env["PMI_RANK"] = node.iproc.ToString();
                launch.env["PMI_SIZE"] = node.nproc.ToString();
                launch.env["PMI_CLIQUE"] = node.clique;
                launch.env["PMI_APPNUM"] = node.appnum.ToString();
                if (pmi != null)
                {
                    launch.env["PMI_DOMAIN"] = pmi.domain;
                    launch.env["PMI_KVS"] = pmi.id;
                }
                else
                {
                    launch.env["PMI_DOMAIN"] = "any_domain";
                    launch.env["PMI_KVS"] = "any_id";
                }
                // SMPD variables
                launch.env["PMI_HOST"] = System.Net.Dns.GetHostName();
                launch.env["PMI_PORT"] = "-1";
                launch.env["PMI_SMPD_ID"] = node.iproc.ToString();
                launch.env["PMI_SMPD_KEY"] = node.iproc.ToString();
                // DeinoPMI variables
                launch.env["PMI_KEY"] = node.iproc.ToString();

                InfoRequestMessage message = new InfoRequestMessage("launch_info");
                client.SendMessage(message);
                client.SendMessage(launch);
                LaunchInfoMessage lmessage = (LaunchInfoMessage)client.NextMessage();
                QuitMessage quit_message = new QuitMessage();
                client.SendMessage(quit_message);
                // get a response to the quit message
                DeinoMPI.Message message2;
                message2 = client.NextMessage();
                try
                {
                    // Send the last leg of the quit handshake
                    quit_message = new QuitMessage();
                    client.SendMessage(quit_message);
                }
                catch
                {
                    // Ignore errors here since we are guaranteed to be in the last leg of the quit handshake.
                }
                UpdateVerifyNode(node.iproc, lmessage);
            }
            catch (Exception ex)
            {
                StringBuilder sb = new StringBuilder("ERROR: ");
                sb.Append(ex.Message);
                foreach (string key in ex.Data.Keys)
                {
                    sb.AppendFormat("\r\n{0}={1}", key, ex.Data[key]);
                }
                if (node != null)
                {
                    vnode[node.iproc].error_msg = sb.ToString();
                    vnode[node.iproc].state = VerifiedNode.VerificationState.BAD;
                }
                else
                {
                    vnode[node.iproc].error_msg = sb.ToString();
                    vnode[node.iproc].state = VerifiedNode.VerificationState.BAD;
                }
            }
        }

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [MTAThread]
        static int Main(string[] args)
        {
            mpiexec = null;
            try
            {
                mpiexec = new Mpiexec(); // Does this throw an exception?
                if (mpiexec == null)
                    return -1;
                // Handle the help options
                if (args == null || args.Length == 0)
                {
                    mpiexec.PrintOptions(false);
                    return mpiexec.mpiexec_exit_code;
                }
                if (args.Length == 1)
                {
                    if (args[0] == "-help")
                    {
                        mpiexec.PrintOptions(false);
                        return mpiexec.mpiexec_exit_code;
                    }
                    else if (args[0] == "-help2")
                    {
                        mpiexec.PrintOptions(true);
                        return mpiexec.mpiexec_exit_code;
                    }
                }
                // Handle the version option
                if (args.Length == 1 && args[0] == "-version")
                {
                    mpiexec.PrintVersion();
                    return mpiexec.mpiexec_exit_code;
                }
                if (args.Length == 2)
                {
                    if (args[0] == "-version")
                    {
                        mpiexec.GetHostVersion(args[1]);
                        return mpiexec.mpiexec_exit_code;
                    }
                }
                // parse the mpiexec arguments into launch nodes
                mpiexec.global_options.dir = Environment.CurrentDirectory;
                mpiexec.GetDefaultSettings();
                ArrayList nodes = mpiexec.ParseArgs(args);
                if (nodes == null)
                {
                    System.Environment.Exit(mpiexec.mpiexec_exit_code);
                }

                int result = 0;
                if (mpiexec.global_options.localonly)
                {
                    result = mpiexec.GetStore(nodes);
                }
                else
                {
                    result = mpiexec.GetStoreAndCredential(nodes);
                }
                if (result != 0)
                    return result;

                // Check to see if we want to validate the job or run the job
                if (mpiexec.global_options.validate)
                {
                    int return_value = -1;
                    try
                    {
                        int i, num_threads;
                        Thread[] threads;
                        if (nodes == null)
                            return 0;
                        num_threads = nodes.Count;
                        threads = new Thread[num_threads];
                        vnode = new VerifiedNode[num_threads];
                        for (i = 0; i < num_threads; i++)
                        {
                            vnode[i] = new VerifiedNode();
                            vnode[i].node = (LaunchNode)nodes[i];
                            ParameterizedThreadStart ts = new ParameterizedThreadStart(verify_thread);
                            threads[i] = new Thread(ts);
                            threads[i].Start(nodes[i]);
                        }
                        for (i = 0; i < num_threads; i++)
                            threads[i].Join();
                        return_value = 0;
                        for (i = 0; i < num_threads; i++)
                        {
                            if (vnode[i].state == VerifiedNode.VerificationState.BAD)
                                return_value = -1;
                            string line = vnode[i].GetLaunchInfoString();
                            line = line.TrimEnd('\r', '\n');
                            Console.WriteLine(line);
                        }
                    }
                    catch (Exception x)
                    {
                        Console.WriteLine("ERROR: Exception caught: {0}", x.Message);
                    }
                    return return_value;
                }

                ArrayList staging_nodes = null;
                Mpiexec stager_mpiexec = null;
                ProcessGroup stager_pg = null;
                if (mpiexec.global_options.stage_src != null)
                {
                    stager_mpiexec = new Mpiexec();
                    stager_mpiexec.global_options.use_output_queue = true;
                    string error_msg = null;
                    if (!stager_mpiexec.GenerateStagingNodes(ref staging_nodes, ref error_msg, nodes, mpiexec.global_options))
                    {
                        if (error_msg != null)
                            Console.WriteLine(error_msg);
                        return -1;
                    }
                    stager_pg = new ProcessGroup(stager_mpiexec);
                    if (stager_pg.StartGroup(staging_nodes))
                    {
                        while (true)
                        {
                            if (stager_mpiexec.global_options.output_queue.Count > 0)
                            {
                                object o = stager_mpiexec.global_options.output_queue.Dequeue();
                                if (o is Thread)
                                {
                                    ((Thread)o).Join();
                                    // The process exited without any output
                                    stager_mpiexec.Kill();
                                    stager_pg = null;
                                    stager_mpiexec = null;
                                    Console.WriteLine("WARNING: Directory staging failed");
                                    break;
                                }
                                else
                                {
                                    if (o is string)
                                    {
                                        string str = (string)o;
                                        str = str.TrimEnd(new char[] { '\r', '\n', ' ' });
                                        if (str != "READY")
                                        {
                                            stager_mpiexec.Kill();
                                            stager_pg = null;
                                            stager_mpiexec = null;
                                            Console.WriteLine(String.Format("WARNING: Directory staging failed: {0}", str));
                                        }
                                    }
                                    break;
                                }
                            }
                            else
                            {
                                Thread.Sleep(100);
                            }
                        }
                    }
                    else
                    {
                        stager_pg.Kill();
                    }
                }
                ProcessGroup pg = new ProcessGroup(mpiexec);
                if (!pg.StartGroup(nodes))
                {
                    /* Don't print the errors because pg.Kill() will create an abort with the same errors printed out.
                    string[] errors = new string[nodes.Count];
                    pg.GetStartErrors(errors);
                    foreach (string str in errors)
                    {
                        mpiexec.WriteLine(str);
                    }
                    */
                    pg.Kill();
                }
                mpiexec.WaitForAllExit();
                if (stager_mpiexec != null)
                {
                    try
                    {
                        stager_mpiexec.input.EnqueueInput("\n");
                        stager_mpiexec.WaitForAllExit();
                        stager_mpiexec.input.Stop();
                        stager_mpiexec = null;
                        stager_pg = null;
                    }
                    catch
                    {
                    }
                }
                mpiexec.input.Stop();
            }
            catch (Exception e)
            {
                Console.WriteLine("Error: {0}", e.Message);
                /*
                Console.WriteLine("Stack:");
                Console.WriteLine(e.StackTrace);
                */
                if (mpiexec != null)
                    mpiexec.mpiexec_exit_code = -1;
            }
            if (mpiexec != null)
            {
                foreach (ProcessGroup pg in mpiexec.process_groups)
                {
                    pg.Kill();
                }
                System.Environment.Exit(mpiexec.mpiexec_exit_code);
                return mpiexec.mpiexec_exit_code;
            }
            return -1;
        }
    }
}
