/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"

int smpd_getpid()
{
	return GetCurrentProcessId();
}

#undef FCNAME
#define FCNAME L"smpd_gen_authentication_strings"
int smpd_gen_authentication_strings(wchar_t *phrase, wchar_t *append, wchar_t *crypted, size_t maxlen)
{
	int stamp;
	wchar_t hash[SMPD_PASSPHRASE_MAX_LENGTH];
	wchar_t phrase_internal[SMPD_PASSPHRASE_MAX_LENGTH+1];

	smpd_enter_fn(FCNAME);

	stamp = rand();

	_snwprintf_s(phrase_internal, SMPD_PASSPHRASE_MAX_LENGTH, _TRUNCATE, L"%s%s %d", phrase, SMPD_VERSION, stamp);
	_snwprintf_s(append, SMPD_AUTHENTICATION_STR_LEN, _TRUNCATE, L"%s %d", SMPD_VERSION, stamp);

	smpd_hash(phrase_internal, (int)wcslen(phrase_internal), hash, SMPD_PASSPHRASE_MAX_LENGTH);
	if (wcslen(hash) > SMPD_PASSPHRASE_MAX_LENGTH)
	{
		smpd_err_printf(L"internal crypted string too long: %d > %d\n", wcslen(hash), SMPD_PASSPHRASE_MAX_LENGTH);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	wcscpy_s(crypted, maxlen, hash);

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

static const wchar_t * crypt_error(int error)
{
	static wchar_t unknown[100];
	switch (error)
	{
	case ERROR_INVALID_HANDLE:
		return L"One of the parameters specifies an invalid handle.";
	case ERROR_INVALID_PARAMETER:
		return L"One of the parameters contains an invalid value. This is most often an invalid pointer.";
	case NTE_BAD_ALGID:
		return L"The hKey session key specifies an algorithm that this CSP does not support.";
	case NTE_BAD_DATA:
		return L"The data to be decrypted is invalid. For example, when a block cipher is used and the Final flag is FALSE, the value specified by pdwDataLen must be a multiple of the block size. This error can also be returned when the padding is found to be invalid.";
	case NTE_BAD_FLAGS:
		return L"The dwFlags parameter is nonzero.";
	case NTE_BAD_HASH:
		return L"The hHash parameter contains an invalid handle.";
	case NTE_BAD_KEY:
		return L"The hKey parameter does not contain a valid handle to a key.";
	case NTE_BAD_LEN:
		return L"The size of the output buffer is too small to hold the generated plaintext.";
	case NTE_BAD_UID:
		return L"The CSP context that was specified when the key was created cannot be found.";
	case NTE_DOUBLE_ENCRYPT:
		return L"The application attempted to decrypt the same data twice.";
	case NTE_FAIL:
		return L"Failure";
	}
	swprintf_s(unknown, 100, L"Unknown failure %d", error);
	return unknown;
}

#undef FCNAME
#define FCNAME L"smpd_encrypt_data_plaintext"
static int smpd_encrypt_data_plaintext(wchar_t *input, int input_length, wchar_t *output, int output_length)
{
	smpd_enter_fn(FCNAME);
	if (output_length < input_length + 2)
	{
		smpd_err_printf(L"encryption output buffer too small: %d < %d.\n", output_length, input_length + 2);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	*output = SMPD_PLAINTEXT_PREFIX;
	output++;
	output_length--;
	memcpy(output, input, input_length*sizeof(wchar_t));
	output[input_length] = L'\0';
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_encrypt_data"
int smpd_encrypt_data(wchar_t *input, int input_length, wchar_t *output, int output_length)
{
	HCRYPTPROV hCryptProv = 0;
	HCRYPTKEY hKey = 0;
	HCRYPTHASH hHash = 0;
	wchar_t szPassword[SMPD_MAX_PASSWORD_LENGTH] = L"";
	DWORD dwLength;
	char *buffer = NULL;
	int length, result;
	int ret_val = SMPD_SUCCESS;

	smpd_enter_fn(FCNAME);

	if (smpd_process.plaintext)
	{
		ret_val = smpd_encrypt_data_plaintext(input, input_length, output, output_length);
		smpd_exit_fn(FCNAME);
		return ret_val;
	}

	/* Acquire a cryptographic provider context handle. */
	if (!CryptAcquireContext(&hCryptProv, NULL, MS_DEF_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT))
	{
		/* Error creating key container!*/
		smpd_err_printf(L"Error during CryptAcquireContext: %d\n", GetLastError());
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* Create an empty hash object. */
	if (!CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hHash)) 
	{
		smpd_err_printf(L"Error during CryptCreateHash\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* Hash the password string. */
	/*MPIU_Snprintf(szPassword, SMPD_MAX_PASSWORD_LENGTH, _TRUNCATE, L"%s %s", smpd_process.encrypt_prefix, smpd_process.passphrase);*/
	MPIU_Wcsncpy(szPassword, smpd_process.passphrase, SMPD_MAX_PASSWORD_LENGTH);
	/*smpd_dbg_printf(L"phrase to generate the encryption key: '%s'\n", szPassword);*/
	dwLength = (DWORD)wcslen(szPassword)*sizeof(wchar_t);
	if (!CryptHashData(hHash, (BYTE *)szPassword, dwLength, 0)) 
	{
		smpd_err_printf(L"Error during CryptHashData\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* Create a session key based on the hash of the smpd password. */
	if (!CryptDeriveKey(hCryptProv, CALG_RC2, hHash, CRYPT_EXPORTABLE, &hKey)) 
	{
		smpd_err_printf(L"Error during CryptDeriveKey\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* Allocate a temporary buffer to work on */
	buffer = (char*)MPIU_Malloc((input_length * 2)*sizeof(wchar_t));
	if (buffer == NULL)
	{
		smpd_err_printf(L"MPIU_Malloc returned NULL\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}
	/* copy the input data to the temporary buffer and encrypt in-place */
	memcpy(buffer, input, input_length * sizeof(wchar_t));
	dwLength = input_length * sizeof(wchar_t);
	if (!CryptEncrypt(hKey, 0, TRUE, 0, (BYTE*)buffer, &dwLength, (input_length*2)*sizeof(wchar_t)))
	{
		smpd_err_printf(L"Error during CryptEncrypt: %s\n", crypt_error(GetLastError()));
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* encode the binary encrypted blob to the output text buffer */
	*output = SMPD_ENCRYPTED_PREFIX;
	output++;
	output_length--;
	result = smpd_encode_buffer(output, output_length, buffer, dwLength, &length);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"unable to encode the encrypted password\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}
	/*smpd_dbg_printf(L"encrypted and encoded password: %d bytes = '%s'\n", length, output);*/

fn_cleanup:
	if (buffer)
	{
		MPIU_Free(buffer);
	}

	/* Destroy the hash object. */
	if (hHash)
	{
		CryptDestroyHash(hHash);
	}

	/* Destroy the session key. */
	if (hKey)
	{
		CryptDestroyKey(hKey);
	}

	/* Release the provider handle. */
	if (hCryptProv)
	{
		CryptReleaseContext(hCryptProv, 0);
	}

	smpd_exit_fn(FCNAME);
	return ret_val;
}

#undef FCNAME
#define FCNAME L"smpd_decrypt_data_plaintext"
static int smpd_decrypt_data_plaintext(wchar_t *input, int input_length, wchar_t *output, int *output_length)
{
	smpd_enter_fn(FCNAME);
	/* check the prefix character to make sure the data is plaintext */
	if (*input != SMPD_PLAINTEXT_PREFIX)
	{
		smpd_err_printf(L"decryption module not available, please try again with -plaintext.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	/* skip over the prefix character */
	input++;
	input_length--;
	if (*output_length < input_length)
	{
		smpd_err_printf(L"decryption output buffer too small.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	memcpy(output, input, input_length*sizeof(wchar_t));
	*output_length = input_length;
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_decrypt_data"
int smpd_decrypt_data(wchar_t *input, int input_length, wchar_t *output, int *output_length)
{
	HCRYPTPROV hCryptProv = 0;
	HCRYPTKEY hKey = 0;
	HCRYPTHASH hHash = 0;
	wchar_t szPassword[SMPD_MAX_PASSWORD_LENGTH] = L"";
	DWORD dwLength;
	char *buffer = NULL;
	int length, result;
	int ret_val = SMPD_SUCCESS;

	smpd_enter_fn(FCNAME);

	if (*input == SMPD_PLAINTEXT_PREFIX)
	{
		ret_val = smpd_decrypt_data_plaintext(input, input_length, output, output_length);
		smpd_exit_fn(FCNAME);
		return ret_val;
	}

	/* Acquire a cryptographic provider context handle. */
	if (!CryptAcquireContext(&hCryptProv, NULL, MS_DEF_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT))
	{
		/* Error creating key container!*/
		smpd_err_printf(L"Error during CryptAcquireContext: %d\n", GetLastError());
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* Create an empty hash object. */
	if (!CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hHash)) 
	{
		smpd_err_printf(L"Error during CryptCreateHash\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* Hash the password string. */
	/*MPIU_Snprintf(szPassword, SMPD_MAX_PASSWORD_LENGTH, _TRUNCATE, L"%s %s", smpd_process.encrypt_prefix, smpd_process.passphrase);*/
	MPIU_Wcsncpy(szPassword, smpd_process.passphrase, SMPD_MAX_PASSWORD_LENGTH);
	/*smpd_dbg_printf(L"phrase to generate the decryption key: '%s'\n", szPassword);*/
	dwLength = (DWORD)wcslen(szPassword)*sizeof(wchar_t);
	if (!CryptHashData(hHash, (BYTE *)szPassword, dwLength, 0)) 
	{
		smpd_err_printf(L"Error during CryptHashData\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* Create a session key based on the hash of the smpd password. */
	if (!CryptDeriveKey(hCryptProv, CALG_RC2, hHash, CRYPT_EXPORTABLE, &hKey)) 
	{
		smpd_err_printf(L"Error during CryptDeriveKey\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* skip over the prefix character */
	input++;
	input_length--;

	/* Allocate a temporary buffer to work on */
	buffer = (char*)MPIU_Malloc(input_length*sizeof(char));
	if (buffer == NULL)
	{
		smpd_err_printf(L"MPIU_Malloc returned NULL\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* decode the data back to it's binary encrypted format */
	/*smpd_dbg_printf(L"encoded encrypted password: '%s'\n", input);*/
	result = smpd_decode_buffer(input, buffer, input_length, &length);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"unable to decode the encrypted password\n");
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* decrypt the data blob back to plaintext */
	dwLength = length;
	if (!CryptDecrypt(hKey, 0, TRUE, 0, (BYTE*)buffer, &dwLength))
	{
		smpd_err_printf(L"Error during CryptDecrypt: %s\n", crypt_error(GetLastError()));
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}
	if (dwLength >= (DWORD)*output_length)
	{
		smpd_err_printf(L"decrypted buffer is larger than the output buffer supplied: %d >= %d\n", dwLength, *output_length);
		ret_val = SMPD_FAIL;
		goto fn_cleanup;
	}

	/* copy the plaintext to the output buffer */
	memcpy(output, buffer, dwLength);
	*output_length = dwLength / sizeof(wchar_t);

fn_cleanup:
	if (buffer)
	{
		MPIU_Free(buffer);
	}

	/* Destroy the hash object. */
	if (hHash)
	{
		CryptDestroyHash(hHash);
	}

	/* Destroy the session key. */
	if (hKey)
	{
		CryptDestroyKey(hKey);
	}

	/* Release the provider handle. */
	if (hCryptProv)
	{
		CryptReleaseContext(hCryptProv, 0);
	}

	smpd_exit_fn(FCNAME);
	return ret_val;
}
