/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif

#undef FCNAME
#define FCNAME L"first_token"
static const wchar_t * first_token(const wchar_t *str)
{
	smpd_enter_fn(FCNAME);
	if (str == NULL)
	{
		smpd_exit_fn(FCNAME);
		return NULL;
	}
	while (isspace(*str))
		str++;
	if (*str == L'\0')
	{
		smpd_exit_fn(FCNAME);
		return NULL;
	}
	smpd_exit_fn(FCNAME);
	return str;
}

#undef FCNAME
#define FCNAME L"next_token"
static const wchar_t * next_token(const wchar_t *str)
{
	const wchar_t *result;

	smpd_enter_fn(FCNAME);
	if (str == NULL)
	{
		smpd_exit_fn(FCNAME);
		return NULL;
	}
	str = first_token(str);
	if (str == NULL)
	{
		smpd_exit_fn(FCNAME);
		return NULL;
	}
	if (*str == SMPD_QUOTE_CHAR)
	{
		/* move over string */
		str++; /* move over the first quote */
		if (*str == L'\0')
		{
			smpd_exit_fn(FCNAME);
			return NULL;
		}
		while (*str != SMPD_QUOTE_CHAR)
		{
			/* move until the last quote, ignoring escaped quotes */
			if (*str == SMPD_ESCAPE_CHAR)
			{
				str++;
				if (*str == SMPD_QUOTE_CHAR)
					str++;
			}
			else
			{
				str++;
			}
			if (*str == L'\0')
			{
				smpd_exit_fn(FCNAME);
				return NULL;
			}
		}
		str++; /* move over the last quote */
	}
	else
	{
		if (*str == SMPD_DELIM_CHAR)
		{
			/* move over the DELIM token */
			str++;
		}
		else
		{
			/* move over literal */
			while (!isspace(*str) && *str != SMPD_DELIM_CHAR && *str != L'\0')
				str++;
		}
	}
	result = first_token(str);
	smpd_exit_fn(FCNAME);
	return result;
}

#undef FCNAME
#define FCNAME L"compare_token"
static int compare_token(const wchar_t *token, const wchar_t *str)
{
	smpd_enter_fn(FCNAME);

	if (token == NULL || str == NULL)
	{
		smpd_exit_fn(FCNAME);
		return -1;
	}

	if (*token == SMPD_QUOTE_CHAR)
	{
		/* compare quoted strings */
		token++; /* move over the first quote */
		/* compare characters until reaching the end of the string or the end quote character */
		for (;;)
		{
			if (*token == SMPD_ESCAPE_CHAR)
			{
				if (*(token+1) == SMPD_QUOTE_CHAR)
				{
					/* move over the escape character if the next character is a quote character */
					token++;
				}
				if (*token != *str)
					break;
			}
			else
			{
				if (*token != *str || *token == SMPD_QUOTE_CHAR)
					break;
			}
			if (*str == L'\0')
				break;
			token++;
			str++;
		}
		if (*str == L'\0' && *token == SMPD_QUOTE_CHAR)
		{
			smpd_exit_fn(FCNAME);
			return 0;
		}
		if (*token == SMPD_QUOTE_CHAR)
		{
			smpd_exit_fn(FCNAME);
			return 1;
		}
		if (*str < *token)
		{
			smpd_exit_fn(FCNAME);
			return -1;
		}
		smpd_exit_fn(FCNAME);
		return 1;
	}

	/* compare DELIM token */
	if (*token == SMPD_DELIM_CHAR)
	{
		if (*str == SMPD_DELIM_CHAR)
		{
			str++;
			if (*str == L'\0')
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			smpd_exit_fn(FCNAME);
			return 1;
		}
		if (*token < *str)
		{
			smpd_exit_fn(FCNAME);
			return -1;
		}
		smpd_exit_fn(FCNAME);
		return 1;
	}

	/* compare literals */
	while (*token == *str && *str != L'\0' && *token != SMPD_DELIM_CHAR && !isspace(*token))
	{
		token++;
		str++;
	}
	if ( (*str == L'\0') && (*token == SMPD_DELIM_CHAR || isspace(*token) || *token == L'\0') )
	{
		smpd_exit_fn(FCNAME);
		return 0;
	}
	if (*token == SMPD_DELIM_CHAR || isspace(*token) || *token < *str)
	{
		smpd_exit_fn(FCNAME);
		return -1;
	}
	smpd_exit_fn(FCNAME);
	return 1;
}

#undef FCNAME
#define FCNAME L"token_copy"
static void token_copy(const wchar_t *token, wchar_t *str, int maxlen)
{
	smpd_enter_fn(FCNAME);
	/* check parameters */
	if (token == NULL || str == NULL)
	{
		smpd_exit_fn(FCNAME);
		return;
	}

	/* check special buffer lengths */
	if (maxlen < 1)
	{
		smpd_exit_fn(FCNAME);
		return;
	}
	if (maxlen == 1)
	{
		*str = L'\0';
		smpd_exit_fn(FCNAME);
		return;
	}

	/* cosy up to the token */
	token = first_token(token);
	if (token == NULL)
	{
		smpd_exit_fn(FCNAME);
		return;
	}

	if (*token == SMPD_DELIM_CHAR)
	{
		/* copy the special deliminator token */
		str[0] = SMPD_DELIM_CHAR;
		str[1] = L'\0';
		smpd_exit_fn(FCNAME);
		return;
	}

	if (*token == SMPD_QUOTE_CHAR)
	{
		/* quoted copy */
		token++; /* move over the first quote */
		do
		{
			if (*token == SMPD_ESCAPE_CHAR)
			{
				if (*(token+1) == SMPD_QUOTE_CHAR)
					token++;
				*str = *token;
			}
			else
			{
				if (*token == SMPD_QUOTE_CHAR)
				{
					*str = L'\0';
					smpd_exit_fn(FCNAME);
					return;
				}
				*str = *token;
			}
			str++;
			token++;
			maxlen--;
		} while (maxlen);
		/* we've run out of destination characters so back up and null terminate the string */
		str--;
		*str = L'\0';
		smpd_exit_fn(FCNAME);
		return;
	}

	/* literal copy */
	while (*token != SMPD_DELIM_CHAR && !isspace(*token) && *token != L'\0' && maxlen)
	{
		*str = *token;
		str++;
		token++;
		maxlen--;
	}
	if (maxlen)
		*str = L'\0';
	smpd_exit_fn(FCNAME);
}

#undef FCNAME
#define FCNAME L"token_hide"
static void token_hide(wchar_t *token)
{
	smpd_enter_fn(FCNAME);
	/* check parameters */
	if (token == NULL)
	{
		smpd_exit_fn(FCNAME);
		return;
	}

	/* cosy up to the token */
	token = (wchar_t*)first_token(token);
	if (token == NULL)
	{
		smpd_exit_fn(FCNAME);
		return;
	}

	if (*token == SMPD_DELIM_CHAR)
	{
		*token = SMPD_HIDE_CHAR;
		smpd_exit_fn(FCNAME);
		return;
	}

	/* quoted */
	if (*token == SMPD_QUOTE_CHAR)
	{
		*token = SMPD_HIDE_CHAR;
		token++; /* move over the first quote */
		while (*token != L'\0')
		{
			if (*token == SMPD_ESCAPE_CHAR)
			{
				if (*(token+1) == SMPD_QUOTE_CHAR)
				{
					*token = SMPD_HIDE_CHAR;
					token++;
				}
				*token = SMPD_HIDE_CHAR;
			}
			else
			{
				if (*token == SMPD_QUOTE_CHAR)
				{
					*token = SMPD_HIDE_CHAR;
					smpd_exit_fn(FCNAME);
					return;
				}
				*token = SMPD_HIDE_CHAR;
			}
			token++;
		}
		smpd_exit_fn(FCNAME);
		return;
	}

	/* literal */
	while (*token != SMPD_DELIM_CHAR && !isspace(*token) && *token != L'\0')
	{
		*token = SMPD_HIDE_CHAR;
		token++;
	}
	smpd_exit_fn(FCNAME);
}

#undef FCNAME
#define FCNAME L"smpd_get_string_arg"
int smpd_get_string_arg(const wchar_t *str, const wchar_t *flag, wchar_t *val, int maxlen)
{
	smpd_enter_fn(FCNAME);
	if (maxlen < 1)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}

	/* line up with the first token */
	str = first_token(str);
	if (str == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}

	/* This loop will match the first instance of L"flag = value" in the string. */
	do
	{
		if (compare_token(str, flag) == 0)
		{
			str = next_token(str);
			if (compare_token(str, SMPD_DELIM_STR) == 0)
			{
				str = next_token(str);
				if (str == NULL)
				{
					smpd_exit_fn(FCNAME);
					return SMPD_FALSE;
				}
				token_copy(str, val, maxlen);
				smpd_exit_fn(FCNAME);
				return SMPD_TRUE;
			}
		}
		else
		{
			str = next_token(str);
		}
	} while (str);
	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;
}

#undef FCNAME
#define FCNAME L"smpd_hide_string_arg"
int smpd_hide_string_arg(wchar_t *str, const wchar_t *flag)
{
	smpd_enter_fn(FCNAME);
	/* line up with the first token */
	str = (wchar_t*)first_token(str);
	if (str == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	do
	{
		if (compare_token(str, flag) == 0)
		{
			str = (wchar_t*)next_token(str);
			if (compare_token(str, SMPD_DELIM_STR) == 0)
			{
				str = (wchar_t*)next_token(str);
				if (str == NULL)
				{
					smpd_exit_fn(FCNAME);
					return SMPD_SUCCESS;
				}
				token_hide(str);
				smpd_exit_fn(FCNAME);
				return SMPD_SUCCESS;
			}
		}
		else
		{
			str = (wchar_t*)next_token(str);
		}
	} while (str);
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_int_arg"
int smpd_get_int_arg(const wchar_t *str, const wchar_t *flag, int *val_ptr)
{
	wchar_t int_str[12];

	smpd_enter_fn(FCNAME);
	if (smpd_get_string_arg(str, flag, int_str, 12))
	{
		*val_ptr = _wtoi(int_str);
		smpd_exit_fn(FCNAME);
		return SMPD_TRUE;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;
}

#undef FCNAME
#define FCNAME L"quoted_printf"
static int quoted_printf(wchar_t *str, int maxlen, const wchar_t *val)
{
	int count = 0;
	smpd_enter_fn(FCNAME);
	if (maxlen < 1)
	{
		smpd_exit_fn(FCNAME);
		return 0;
	}
	*str = SMPD_QUOTE_CHAR;
	str++;
	maxlen--;
	count++;
	while (maxlen)
	{
		if (*val == L'\0')
			break;
		if (*val == SMPD_QUOTE_CHAR)
		{
			*str = SMPD_ESCAPE_CHAR;
			str++;
			maxlen--;
			count++;
			if (maxlen == 0)
			{
				smpd_exit_fn(FCNAME);
				return count;
			}
		}
		*str = *val;
		str++;
		maxlen--;
		count++;
		val++;
	}
	if (maxlen)
	{
		*str = SMPD_QUOTE_CHAR;
		str++;
		maxlen--;
		count++;
		if (maxlen == 0)
		{
			smpd_exit_fn(FCNAME);
			return count;
		}
		*str = L'\0';
	}
	smpd_exit_fn(FCNAME);
	return count;
}

#undef FCNAME
#define FCNAME L"smpd_add_string"
int smpd_add_string(wchar_t *str, int maxlen, const wchar_t *val)
{
	int num_chars;

	smpd_enter_fn(FCNAME);
	if (wcsstr(val, L" ") || val[0] == SMPD_QUOTE_CHAR)
	{
		num_chars = quoted_printf(str, maxlen, val);
		if (num_chars < maxlen)
		{
			str[num_chars] = SMPD_SEPAR_CHAR;
			str[num_chars+1] = L'\0';
		}
		num_chars++;
	}
	else
	{
		num_chars = _snwprintf_s(str, maxlen, _TRUNCATE, L"%s ", val);
	}
	smpd_exit_fn(FCNAME);
	return num_chars;
}

#undef FCNAME
#define FCNAME L"smpd_get_string"
const wchar_t * smpd_get_string(const wchar_t *str, wchar_t *val, int maxlen, int *num_chars)
{
	smpd_enter_fn(FCNAME);
	if (maxlen < 1)
	{
		*num_chars = 0;
		smpd_exit_fn(FCNAME);
		return NULL;
	}

	/* line up with the first token */
	str = first_token(str);
	if (str == NULL)
	{
		*num_chars = 0;
		smpd_exit_fn(FCNAME);
		return NULL;
	}

	/* copy the token */
	token_copy(str, val, maxlen);
	*num_chars = (int)wcslen(val);

	/* move to the next token */
	str = next_token(str);

	smpd_exit_fn(FCNAME);
	return str;
}

#undef FCNAME
#define FCNAME L"smpd_add_string_arg"
int smpd_add_string_arg(wchar_t **str_ptr, int *maxlen_ptr, const wchar_t *flag, const wchar_t *val)
{
	int num_chars;

	smpd_enter_fn(FCNAME);
	if (*maxlen_ptr < 1)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	/* add the flag */
	if (wcsstr(flag, L" ") || wcsstr(flag, SMPD_DELIM_STR) || flag[0] == SMPD_QUOTE_CHAR)
	{
		num_chars = quoted_printf(*str_ptr, *maxlen_ptr, flag);
	}
	else
	{
		num_chars = _snwprintf_s(*str_ptr, *maxlen_ptr, _TRUNCATE, L"%s", flag);
	}
	*maxlen_ptr = *maxlen_ptr - num_chars;
	if (*maxlen_ptr < 1)
	{
		(*str_ptr)[num_chars-1] = L'\0';
		smpd_dbg_printf(L"partial argument added to string: '%s'\n", *str_ptr);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	*str_ptr = *str_ptr + num_chars;

	/* add the deliminator character */
	**str_ptr = SMPD_DELIM_CHAR;
	*str_ptr = *str_ptr + 1;
	*maxlen_ptr = *maxlen_ptr - 1;

	/* add the value string */
	if (wcsstr(val, L" ") || wcsstr(val, SMPD_DELIM_STR) || val[0] == SMPD_QUOTE_CHAR)
	{
		num_chars = quoted_printf(*str_ptr, *maxlen_ptr, val);
	}
	else
	{
		num_chars = _snwprintf_s(*str_ptr, *maxlen_ptr, _TRUNCATE, L"%s", val);
	}
	*str_ptr = *str_ptr + num_chars;
	*maxlen_ptr = *maxlen_ptr - num_chars;
	if (*maxlen_ptr < 2)
	{
		*str_ptr = *str_ptr - 1;
		**str_ptr = L'\0';
		smpd_dbg_printf(L"partial argument added to string: '%s'\n", *str_ptr);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	
	/* add the trailing space */
	**str_ptr = SMPD_SEPAR_CHAR;
	*str_ptr = *str_ptr + 1;
	**str_ptr = L'\0';
	*maxlen_ptr = *maxlen_ptr - 1;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_add_int_arg"
int smpd_add_int_arg(wchar_t **str_ptr, int *maxlen_ptr, const wchar_t *flag, int val)
{
	int result;
	wchar_t val_str[12];
	smpd_enter_fn(FCNAME);
	swprintf_s(val_str, 12, L"%d", val);
	result = smpd_add_string_arg(str_ptr, maxlen_ptr, flag, val_str);
	smpd_exit_fn(FCNAME);
	return result;
}

#undef FCNAME
#define FCNAME L"smpd_get_opt"
int smpd_get_opt(int *argc, wchar_t ***argv, wchar_t * flag)
{
	int i,j;

	smpd_enter_fn(FCNAME);
	if (flag == NULL)
	{
		smpd_exit_fn(FCNAME);
		return 0;
	}

	for (i=0; i<*argc; i++)
	{
		if (wcscmp((*argv)[i], flag) == 0)
		{
			for (j=i; j<*argc; j++)
			{
				(*argv)[j] = (*argv)[j+1];
			}
			*argc -= 1;
			smpd_exit_fn(FCNAME);
			return 1;
		}
	}
	smpd_exit_fn(FCNAME);
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_get_opt_int"
int smpd_get_opt_int(int *argc, wchar_t ***argv, wchar_t * flag, int *n)
{
	int i,j;

	smpd_enter_fn(FCNAME);
	if (flag == NULL)
	{
		smpd_exit_fn(FCNAME);
		return 0;
	}

	for (i=0; i<*argc; i++)
	{
		if (wcscmp((*argv)[i], flag) == 0)
		{
			if (i+1 == *argc)
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			*n = _wtoi((*argv)[i+1]);
			for (j=i; j<*argc-1; j++)
			{
				(*argv)[j] = (*argv)[j+2];
			}
			*argc -= 2;
			smpd_exit_fn(FCNAME);
			return 1;
		}
	}
	smpd_exit_fn(FCNAME);
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_get_opt_long"
int smpd_get_opt_long(int *argc, wchar_t ***argv, wchar_t * flag, long *n)
{
	int i;

	smpd_enter_fn(FCNAME);
	if (smpd_get_opt_int(argc, argv, flag, &i))
	{
		*n = (long)i;
		smpd_exit_fn(FCNAME);
		return 1;
	}
	smpd_exit_fn(FCNAME);
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_get_opt_double"
int smpd_get_opt_double(int *argc, wchar_t ***argv, wchar_t * flag, double *d)
{
	int i,j;

	smpd_enter_fn(FCNAME);
	if (flag == NULL)
	{
		smpd_exit_fn(FCNAME);
		return 0;
	}

	for (i=0; i<*argc; i++)
	{
		if (wcscmp((*argv)[i], flag) == 0)
		{
			if (i+1 == *argc)
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			*d = _wtof((*argv)[i+1]);
			for (j=i; j<*argc-1; j++)
			{
				(*argv)[j] = (*argv)[j+2];
			}
			*argc -= 2;
			smpd_exit_fn(FCNAME);
			return 1;
		}
	}
	smpd_exit_fn(FCNAME);
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_get_opt_string"
int smpd_get_opt_string(int *argc, wchar_t ***argv, wchar_t * flag, wchar_t *str, int len)
{
	int i,j;

	smpd_enter_fn(FCNAME);
	if (flag == NULL)
	{
		smpd_exit_fn(FCNAME);
		return 0;
	}

	for (i=0; i<*argc; i++)
	{
		if (wcscmp((*argv)[i], flag) == 0)
		{
			if (i+1 == (*argc))
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			if ((*argv)[i+1][0] == L'-')
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			MPIU_Wcsncpy(str, (*argv)[i+1], len);
			str[len-1] = L'\0';
			for (j=i; j<(*argc)-1; j++)
			{
				(*argv)[j] = (*argv)[j+2];
			}
			*argc -= 2;
			smpd_exit_fn(FCNAME);
			return 1;
		}
	}
	smpd_exit_fn(FCNAME);
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_get_opt_two_strings"
int smpd_get_opt_two_strings(int *argc, wchar_t ***argv, wchar_t * flag, wchar_t *str1, int len1, wchar_t *str2, int len2)
{
	int i, j;

	smpd_enter_fn(FCNAME);
	if (flag == NULL)
	{
		smpd_exit_fn(FCNAME);
		return 0;
	}

	for (i=0; i<*argc; i++)
	{
		if (wcscmp((*argv)[i], flag) == 0)
		{
			if (i+1 == (*argc))
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			if ((*argv)[i+1][0] == L'-')
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			if (i+2 == (*argc))
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			if ((*argv)[i+2][0] == L'-')
			{
				smpd_exit_fn(FCNAME);
				return 0;
			}
			MPIU_Wcsncpy(str1, (*argv)[i+1], len1);
			str1[len1-1] = L'\0';
			MPIU_Wcsncpy(str2, (*argv)[i+2], len2);
			str2[len2-1] = L'\0';
			for (j=i; j<(*argc)-2; j++)
			{
				(*argv)[j] = (*argv)[j+3];
			}
			*argc -= 3;
			smpd_exit_fn(FCNAME);
			return 1;
		}
	}
	smpd_exit_fn(FCNAME);
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_get_win_opt_string"
int smpd_get_win_opt_string(int *argc, wchar_t ***argv, wchar_t * flag, wchar_t *str, int len)
{
	int i,j;
	wchar_t *iter;

	smpd_enter_fn(FCNAME);
	if (flag == NULL)
	{
		smpd_exit_fn(FCNAME);
		return 0;
	}

	for (i=0; i<*argc; i++)
	{
		if (wcsncmp((*argv)[i], flag, wcslen(flag)) == 0)
		{
			if (wcslen(flag) == wcslen((*argv)[i]))
			{
				/* option specified with no value: /flag */
				continue;
			}
			iter = &(*argv)[i][wcslen(flag)];
			if ((*iter != L':') && (*iter != L'=')) /* /flag=val or /flag:val */
			{
				/* partial option match: /flag=val == /flagger=val */
				continue;
			}
			iter++;
			MPIU_Wcsncpy(str, iter, len);
			str[len-1] = L'\0';
			for (j=i; j<(*argc); j++)
			{
				(*argv)[j] = (*argv)[j+1];
			}
			*argc -= 1;
			smpd_exit_fn(FCNAME);
			return 1;
		}
	}
	smpd_exit_fn(FCNAME);
	return 0;
}
