/*   Copyright (C) 2009 Deino Software. */

/*
 * This file provides a set of routines that can be used to record debug
 * messages in a ring so that the may be dumped at a later time.  For example,
 * this can be used to record debug messages without printing them; when
 * a special event, such as an error occurs, a call to 
 * MPIU_dump_dbg_memlog( stderr ) will print the contents of the file ring
 * to stderr.
 */
#include "mpiimpl.h"
#include <stdio.h>
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

/* Temporary.  sig values will change */
/* style: allow:vwprintf:3 sig:0 */
/* style: allow:fputws:1 sig:0 */
/* style: allow:wprintf:2 sig:0 */
/* style: allow:fwprintf:7 sig:0 */

#ifdef HAVE_VA_COPY
# define va_copy_end(a) va_end(a)
#else
# ifdef HAVE___VA_COPY
#  define va_copy(a,b) __va_copy(a,b)
#  define va_copy_end(a) 
# else
#  define va_copy(a,b) ((a) = (b))
/* Some writers recommend define va_copy(a,b) memcpy(&a,&b,sizeof(va_list)) */
#  define va_copy_end(a)
# endif
#endif

#if !defined(MPICH_DBG_MEMLOG_NUM_LINES)
#define MPICH_DBG_MEMLOG_NUM_LINES 1024
#endif
#if !defined(MPICH_DBG_MEMLOG_LINE_SIZE)
#define MPICH_DBG_MEMLOG_LINE_SIZE 256
#endif

MPIU_dbg_state_t MPIUI_dbg_state = MPIU_DBG_STATE_UNINIT;
FILE * MPIUI_dbg_fp = NULL;
static int dbg_memlog_num_lines = MPICH_DBG_MEMLOG_NUM_LINES;
static int dbg_memlog_line_size = MPICH_DBG_MEMLOG_LINE_SIZE;
static wchar_t **dbg_memlog = NULL;
static int dbg_memlog_next = 0;
static int dbg_memlog_count = 0;
static int dbg_rank = -1;

static void dbg_init(void);

#undef FUNCNAME
#define FUNCNAME MPIU_dbg_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_dbg_init(int rank)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIU_DBG_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DBG_INIT);

	dbg_rank = rank;
	if (MPIUI_dbg_state == MPIU_DBG_STATE_UNINIT)
	{
		dbg_init();
	}

	/* If file logging is enable, we need to open a file */
	if (MPIUI_dbg_state & MPIU_DBG_STATE_FILE)
	{
		wchar_t fn[128];
		errno_t error;

		/* Only open the file only once in case MPIU_dbg_init is called more 
		   than once */
		if (MPIUI_dbg_fp == NULL)
		{
			MPIU_Snprintf(fn, 128, _TRUNCATE, L"mpich2-dbg-%d.log", dbg_rank);
			error = _wfopen_s(&MPIUI_dbg_fp, fn, L"w");
			if (error == 0)
				setvbuf(MPIUI_dbg_fp, NULL, _IONBF, 0);
		}
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBG_INIT);
	return 0;
}

#undef FUNCNAME
#define FUNCNAME dbg_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void dbg_init(void)
{
	const wchar_t * envstr;
	MPIDI_STATE_DECL(MPID_STATE_DBG_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_DBG_INIT);
	
	MPIUI_dbg_state = 0;

	envstr = MPIU_GetStaticEnv(L"DeinoMPI_DBG_OUTPUT");
	if (envstr == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_DBG_INIT);
		return;
	}

	/*
	 * TODO:
	 *
	 * - parse environment variable to determine number of log lines, etc.
	 *
	 * - add support for writing to a (per-process or global?) file
	 *
	 * - add support for sending to a log server, perhaps with global time
	 *   sequencing information ???
	 */
	if (wcsstr(envstr, L"stdout"))
	{
		MPIUI_dbg_state |= MPIU_DBG_STATE_STDOUT;
	}
	if (wcsstr(envstr, L"memlog"))
	{
		MPIUI_dbg_state |= MPIU_DBG_STATE_MEMLOG;
	}
	if (wcsstr(envstr, L"file"))
	{
		MPIUI_dbg_state |= MPIU_DBG_STATE_FILE;
	}

	/* If memlog is enabled, the we need to allocate some memory for it */
	if (MPIUI_dbg_state & MPIU_DBG_STATE_MEMLOG)
	{
		dbg_memlog = MPIU_Malloc(dbg_memlog_num_lines * sizeof(wchar_t *) + dbg_memlog_num_lines * (dbg_memlog_line_size * sizeof(wchar_t)));
		if (dbg_memlog != NULL)
		{
			int i;
			
			for (i = 0; i < dbg_memlog_num_lines ; i++)
			{
				dbg_memlog[i] = ((wchar_t *) &dbg_memlog[dbg_memlog_num_lines]) + i * dbg_memlog_line_size;
			}
		}
		else
		{
			MPIUI_dbg_state &= ~MPIU_DBG_STATE_MEMLOG;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_DBG_INIT);
}

#undef FUNCNAME
#define FUNCNAME MPIU_dbglog_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_dbglog_printf(const wchar_t *str, ...)
{
	int n = 0;
	va_list list;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_DBGLOG_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DBGLOG_PRINTF);

	if (MPIUI_dbg_state == MPIU_DBG_STATE_UNINIT)
	{
		dbg_init();
	}

	if (MPIUI_dbg_state & MPIU_DBG_STATE_MEMLOG)
	{
		/* FIXME: put everything on one line until a \n is found */
		
		dbg_memlog[dbg_memlog_next][0] = L'\0';
		va_start(list, str);
		n = _vsnwprintf_s(dbg_memlog[dbg_memlog_next], dbg_memlog_line_size, _TRUNCATE, str, list);
		va_end(list);

		/* if the output was truncated, we null terminate the end of the
		   string, on the off chance that _vsnwprintf_s() didn't do that.  we also
		   check to see if any data has been written over the null we set at
		   the beginning of the string.  this is mostly paranoia, but the man
		   page does not clearly state what happens when truncation occurs.  if
		   data was written to the string, we would like to output it, but we
		   want to avoid reading past the end of the array or outputing garbage
		   data. */

		if (n < 0 || n >= dbg_memlog_line_size)
		{
			dbg_memlog[dbg_memlog_next][dbg_memlog_line_size - 1] = L'\0';
			n = (int)wcslen(dbg_memlog[dbg_memlog_next]);
		}

		if (dbg_memlog[dbg_memlog_next][0] != L'\0')
		{
			dbg_memlog_next = (dbg_memlog_next + 1) % dbg_memlog_num_lines;
			dbg_memlog_count++;
		}
	}

	if (MPIUI_dbg_state & MPIU_DBG_STATE_STDOUT)
	{
		va_start(list, str);
		n = vwprintf(str, list);
		va_end(list);
	}

	if ((MPIUI_dbg_state & MPIU_DBG_STATE_FILE) && MPIUI_dbg_fp != NULL)
	{
		va_start(list, str);
		n = vfwprintf(MPIUI_dbg_fp, str, list);
		va_end(list);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBGLOG_PRINTF);
	return n;
}

#undef FUNCNAME
#define FUNCNAME MPIU_dbglog_vprintf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_dbglog_vprintf(const wchar_t *str, va_list ap)
{
	int n = 0;
	va_list list;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_DBGLOG_VPRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DBGLOG_VPRINTF);

	if (MPIUI_dbg_state == MPIU_DBG_STATE_UNINIT)
	{
		dbg_init();
	}

	if (MPIUI_dbg_state & MPIU_DBG_STATE_MEMLOG)
	{
		va_copy(list,ap);
		dbg_memlog[dbg_memlog_next][0] = L'\0';
		n = _vsnwprintf_s(dbg_memlog[dbg_memlog_next], dbg_memlog_line_size, _TRUNCATE, str, list);
		va_copy_end(list);

		/* if the output was truncated, we null terminate the end of the
		   string, on the off chance that _vsnwprintf_s() didn't do that.  we also
		   check to see if any data has been written over the null we set at
		   the beginning of the string.  this is mostly paranoia, but the man
		   page does not clearly state what happens when truncation occurs.  if
		   data was written to the string, we would like to output it, but we
		   want to avoid reading past the end of the array or outputing garbage
		   data. */

		if (n < 0 || n >= dbg_memlog_line_size)
		{
			dbg_memlog[dbg_memlog_next][dbg_memlog_line_size - 1] = L'\0';
			n = (int)wcslen(dbg_memlog[dbg_memlog_next]);
		}

		if (dbg_memlog[dbg_memlog_next][0] != L'\0')
		{
			dbg_memlog_next = (dbg_memlog_next + 1) % dbg_memlog_num_lines;
			dbg_memlog_count++;
		}
	}

	if (MPIUI_dbg_state & MPIU_DBG_STATE_STDOUT)
	{
		va_copy(list, ap);
		n = vwprintf(str, list);
		va_copy_end(list);
	}

	if ((MPIUI_dbg_state & MPIU_DBG_STATE_FILE) && MPIUI_dbg_fp != NULL)
	{
		va_copy(list, ap);
		n = vfwprintf(MPIUI_dbg_fp, str, list);
		va_end(list);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBGLOG_VPRINTF);
	return n;
}

#undef FUNCNAME
#define FUNCNAME MPIU_dbg_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_dbg_printf(const wchar_t * str, ...)
{
	int n;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_DBG_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DBG_PRINTF);
	
	/* MPID_Common_thread_lock(); */
	{
		va_list list;

		MPIU_dbglog_printf(L"[%d]", dbg_rank);
		va_start(list, str);
		n = MPIU_dbglog_vprintf(str, list);
		va_end(list);
		MPIU_dbglog_flush();
	}
	/* MPID_Common_thread_unlock(); */

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBG_PRINTF);
	return n;
}

#undef FUNCNAME
#define FUNCNAME MPIU_dump_dbg_memlog_to_stdout
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIU_dump_dbg_memlog_to_stdout(void)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIU_DUMP_DBG_MEMLOG_TO_STDOUT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DUMP_DBG_MEMLOG_TO_STDOUT);
	MPIU_dump_dbg_memlog(stdout);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DUMP_DBG_MEMLOG_TO_STDOUT);
}

#undef FUNCNAME
#define FUNCNAME MPIU_dump_dbg_memlog_to_file
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIU_dump_dbg_memlog_to_file(const wchar_t *filename)
{
	FILE *fout;
	errno_t error;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_DUMP_DBG_MEMLOG_TO_FILE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DUMP_DBG_MEMLOG_TO_FILE);
	error = _wfopen_s(&fout, filename, L"wb");
	if (error == 0 && fout != NULL)
	{
		MPIU_dump_dbg_memlog(fout);
		fclose(fout);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DUMP_DBG_MEMLOG_TO_FILE);
}

#undef FUNCNAME
#define FUNCNAME MPIU_dump_dbg_memlog
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIU_dump_dbg_memlog(FILE * fp)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIU_DUMP_DBG_MEMLOG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DUMP_DBG_MEMLOG);
	if (dbg_memlog_count != 0)
	{
		int ent;
		int last_ent;

		/* there is a small issue with counter rollover which will need to be
		   fixed if more than 2^32 lines are going to be logged */
		ent = (dbg_memlog_next == dbg_memlog_count) ? 0 : dbg_memlog_next;
		last_ent = (ent + dbg_memlog_num_lines - 1) % dbg_memlog_num_lines;
		
		do
		{
			fputws(dbg_memlog[ent], fp);
			ent = (ent + 1) % dbg_memlog_num_lines;
		}
		while(ent != last_ent);
		fflush(fp);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DUMP_DBG_MEMLOG);
}

#ifdef USE_DBG_LOGGING
/* 
 * NEW ROUTINES FOR DEBUGGING
 */
int MPIU_DBG_ActiveClasses = 0;
int MPIU_DBG_MaxLevel	  = MPIU_DBG_TYPICAL;
static int mpiu_dbg_initialized = 0;
static FILE *MPIU_DBG_fp = 0;
static wchar_t *filePattern = L"-stdout-"; /* L"log%d.log"; */
/* 
   We keep information on the state of the file.  
   The issue is that sometimes we want to generate DBG output before we
   know the rank (and world) of the process.

   Thus, we have two values: whether the file has been opened, and if so,
   if it was opened before or after the rank was available; and how the
   file should be handled if I/O is performed before the rank is available.

   Another option that could be considered is to store values that would
   otherwise be written before the rank is available for opening the file.
   Currently, this is not used to avoid allocating the necessary space.
 */
typedef enum { MPIU_DBG_CLOSED, MPIU_DBG_PRERANK, MPIU_DBG_OPEN } FileState_t;
static FileState_t filestate = MPIU_DBG_CLOSED;
typedef enum { MPIU_DBG_REOPEN, MPIU_DBG_IGNORE_PRERANK, MPIU_DBG_OPENONCE }
	FileOpenMode_t;
static FileOpenMode_t filemode = MPIU_DBG_REOPEN;
static wchar_t *defaultFilePattern = L"dbg@W%w-@%d@T-%t@.log";
static int worldNum  = 0;
static int worldRank = -1;
static int threadID  = 0;
static int whichRank = -1;			 /* all ranks */
static double timeOrigin = 0.0;

static int MPIU_DBG_Usage( const wchar_t *, const wchar_t * );
static int MPIU_DBG_OpenFile( void );
static int setDBGClass( const wchar_t * );
static int SetDBGLevel( const wchar_t *s, const wchar_t *(names[]) );

#undef FUNCNAME
#define FUNCNAME MPIU_DBG_Outevent
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_DBG_Outevent( const wchar_t *file, int line, int class, int kind, 
					   const wchar_t *fmat, ... )
{
	va_list list;
	wchar_t *str, stmp[MPIU_DBG_MAXLINE];
	int  i;
	void *p;
	MPID_Time_t t;
	double  curtime;
	//MPIDI_STATE_DECL(MPID_STATE_MPIU_DBG_OUTEVENT);

	//MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DBG_OUTEVENT);

	if (!mpiu_dbg_initialized)
	{
		//MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBG_OUTEVENT);
		return 0;
	}

#if MPICH_IS_THREADED
	{
		MPE_Thread_id_t tid;
		MPE_Thread_self(&tid);
		threadID = (int)tid;
	}
#endif
	if (!MPIU_DBG_fp) {
		MPIU_DBG_OpenFile();
	}

	MPID_Wtime( &t );
	MPID_Wtime_todouble( &t, &curtime );
	curtime = curtime - timeOrigin;

	/* The kind values are used with the macros to simplify these cases */
	switch (kind) {
		case 0:
			fwprintf( MPIU_DBG_fp, L"%d\t%d\t%d\t%d\t%f\t%s\t%d\t%s\n",
					 worldNum, worldRank, threadID, class, curtime, 
					 file, line, fmat );
			break;
		case 1:
			va_start(list,fmat);
			str = va_arg(list,wchar_t *);
			MPIU_Snprintf( stmp, MPIU_DBG_MAXLINE, _TRUNCATE, fmat, str );
			va_end(list);
			fwprintf( MPIU_DBG_fp, L"%d\t%d\t%d\t%d\t%f\t%s\t%d\t%s\n",
					 worldNum, worldRank, threadID, class, curtime, 
					 file, line, stmp );
			break;
		case 2: 
			va_start(list,fmat);
			i = va_arg(list,int);
			MPIU_Snprintf( stmp, MPIU_DBG_MAXLINE, _TRUNCATE, fmat, i);
			va_end(list);
			fwprintf( MPIU_DBG_fp, L"%d\t%d\t%d\t%d\t%f\t%s\t%d\t%s\n",
					 worldNum, worldRank, threadID, class, curtime, 
					 file, line, stmp );
			break;
		case 3: 
			va_start(list,fmat);
			p = va_arg(list,void *);
			MPIU_Snprintf( stmp, MPIU_DBG_MAXLINE, _TRUNCATE, fmat, p);
			va_end(list);
			fwprintf( MPIU_DBG_fp, L"%d\t%d\t%d\t%d\t%f\t%s\t%d\t%s\n",
					 worldNum, worldRank, threadID, class, curtime, 
					 file, line, stmp );
			break;
		default:
			break;
	}
	fflush(MPIU_DBG_fp);
	//MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBG_OUTEVENT);
	return 0;
}

/* These are used to simplify the handling of options.  
   To add a new name, add an MPIU_DBG_ClassName element to the array
   MPIU_Classnames.  The "classbits" values are defined by MPIU_DBG_CLASS
   in src/include/mpidbg.h 
 */

typedef struct MPIU_DBG_ClassName {
	int		classbits;
	const wchar_t *UCName, *LCName; 
} MPIU_DBG_ClassName;

static const MPIU_DBG_ClassName MPIU_Classnames[] = {
	{ MPIU_DBG_PT2PT,		 L"PT2PT",		 L"pt2pt" },
	{ MPIU_DBG_RMA,		   L"RMA",		   L"rma"   },
	{ MPIU_DBG_THREAD,		L"THREAD",		L"thread" },
	{ MPIU_DBG_PM,			L"PM",			L"pm" },
	{ MPIU_DBG_ROUTINE_ENTER, L"ROUTINE_ENTER", L"routine_enter" },
	{ MPIU_DBG_ROUTINE_EXIT,  L"ROUTINE_EXIT",  L"routine_exit" },
	{ MPIU_DBG_ROUTINE_ENTER |
	  MPIU_DBG_ROUTINE_EXIT,  L"ROUTINE",	   L"routine" },
	{ MPIU_DBG_SYSCALL,	   L"SYSCALL",	   L"syscall" },
	{ MPIU_DBG_DATATYPE,	  L"DATATYPE",	  L"datatype" },
	{ MPIU_DBG_HANDLE,		L"HANDLE",		L"handle" },
	{ MPIU_DBG_COMM,		  L"COMM",		  L"comm" },
	{ MPIU_DBG_BSEND,		 L"BSEND",		 L"bsend" },
	{ MPIU_DBG_OTHER,		 L"OTHER",		 L"other" },
	{ MPIU_DBG_CH3_CONNECT,   L"CH3_CONNECT",   L"ch3_connect" },
	{ MPIU_DBG_CH3_DISCONNECT,L"CH3_DISCONNECT",L"ch3_disconnect" },
	{ MPIU_DBG_CH3_PROGRESS,  L"CH3_PROGRESS",  L"ch3_progress" },
	{ MPIU_DBG_CH3_CHANNEL,   L"CH3_CHANNEL",   L"ch3_channel" },
	{ MPIU_DBG_CH3,		   L"CH3_MSG",	   L"ch3_msg" },
	{ MPIU_DBG_CH3_OTHER,	 L"CH3_OTHER",	 L"ch3_other" },
	{ MPIU_DBG_CH3,		   L"CH3",		   L"ch3" },
	{ MPIU_DBG_NEM_SOCK_FUNC, L"NEM_SOCK_FUNC", L"nem_sock_func"},
	{ MPIU_DBG_NEM_SOCK_DET,  L"NEM_SOCK_DET",  L"nem_sock_det"},
	{ MPIU_DBG_VC,			L"VC",			L"vc"},
	{ MPIU_DBG_REFCOUNT,	  L"REFCOUNT",	  L"refcount"},
	{ MPIU_DBG_ALL,		   L"ALL",		   L"all" }, 
	{ 0,					  0,			   0 }
};

/* Because the level values are simpler and are rarely changed, these
   use a simple set of parallel arrays */
static const int  MPIU_Levelvalues[] = { MPIU_DBG_TERSE,
										 MPIU_DBG_TYPICAL,
										 MPIU_DBG_VERBOSE, 100 };
static const wchar_t *MPIU_Levelname[] = { L"TERSE", L"TYPICAL", L"VERBOSE", 0 };
static const wchar_t *MPIU_LCLevelname[] = { L"terse", L"typical", L"verbose", 0 };

/* 
 * Initialize the DBG_MSG system.  This is called during MPI_Init to process
 * command-line arguments as well as checking the MPICH_DBG environment
 * variables.  The initialization is split into two steps: a preinit and an 
 * init. This makes it possible to enable most of the features before calling 
 * MPID_Init, where a significant amount of the initialization takes place.
 */

#undef FUNCNAME
#define FUNCNAME MPIU_DBG_ProcessArgs
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIU_DBG_ProcessArgs( int *argc_p, wchar_t ***argv_p )
{
	int i, rc;

	/* Here's where we do the same thing with the command-line options */
	if (argc_p) {
		for (i=1; i<*argc_p; i++) {
			if (wcsncmp((*argv_p)[i], L"-mpich-dbg", 10) == 0) {
				wchar_t *s = (*argv_p)[i] + 10;
				/* Found a command */
				if (*s == 0) {
					/* Just -mpich-dbg */
					MPIU_DBG_MaxLevel	  = MPIU_DBG_TYPICAL;
					MPIU_DBG_ActiveClasses = MPIU_DBG_ALL;
				}
				else if (*s == L'=') {
					/* look for file */
					MPIU_DBG_MaxLevel	  = MPIU_DBG_TYPICAL;
					MPIU_DBG_ActiveClasses = MPIU_DBG_ALL;
					s++;
					if (wcsncmp( s, L"file", 4 ) == 0) {
						filePattern = defaultFilePattern;
					}
				}
				else if (wcsncmp(s, L"-level",6) == 0) {
					wchar_t *p = s + 6;
					if (*p == L'=') {
						p++;
						rc = SetDBGLevel( p, MPIU_LCLevelname );
						if (rc) 
							MPIU_DBG_Usage( L"-mpich-dbg-level", L"terse, typical, verbose" );
					}
				}
				else if (wcsncmp(s, L"-class",6) == 0) {
					wchar_t *p = s + 6;
					if (*p == L'=') {
						p++;
						rc = setDBGClass( p );
						if (rc)
							MPIU_DBG_Usage( L"-mpich-dbg-class", 0 );
					}
				}
				else if (wcsncmp( s, L"-filename", 9 ) == 0) {
					wchar_t *p = s + 9;
					if (*p == L'=') {
						p++;
						/* A special case for a filepattern of "-default",
						   use the predefined default pattern */
						if (wcscmp( p, L"-default" ) == 0) {
							filePattern = defaultFilePattern;
						}
						else {
							filePattern = MPIU_Strdup( p );
						}
					}
				}
				else if (wcsncmp( s, L"-rank", 5 ) == 0) {
					wchar_t *p = s + 5;
					if (*p == L'=' && p[1] != L'\0') {
						wchar_t *sOut;
						p++;
						whichRank = wcstol( p, &sOut, 10 );
						if (p == sOut) {
							MPIU_DBG_Usage( L"-mpich-dbg-rank", 0 );
							whichRank = -1;
						}
					}
				}
				else {
					MPIU_DBG_Usage( (*argv_p)[i], 0 );
				}

				/* Eventually, should null it out and reduce argc value */
			}
		}
	}
	return MPI_SUCCESS;
}

static int MPIU_DBG_ProcessEnv( void )
{
	int rc;
	size_t ret_val;
	wchar_t env_val[1024];

	_wgetenv_s(&ret_val, env_val, 1024, L"MPICH_DBG" );
	if (!ret_val) {
		/* Set the defaults */
		MPIU_DBG_MaxLevel = MPIU_DBG_TYPICAL;
		MPIU_DBG_ActiveClasses = MPIU_DBG_ALL;
		if (wcsncmp(env_val, L"FILE",4) == 0) {
			filePattern = defaultFilePattern;
		}
	}
	_wgetenv_s(&ret_val, env_val, 1024, L"MPICH_DBG_LEVEL" );
	if (!ret_val) {
		rc = SetDBGLevel( env_val, MPIU_Levelname );
		if (rc) 
			MPIU_DBG_Usage( L"MPICH_DBG_LEVEL", L"TERSE, TYPICAL, VERBOSE" );
	}

	_wgetenv_s(&ret_val, env_val, 1024, L"MPICH_DBG_CLASS" );
	if (!ret_val)
		rc = setDBGClass( env_val );
	else
		rc = setDBGClass( NULL );
	if (rc) 
		MPIU_DBG_Usage( L"MPICH_DBG_CLASS", 0 );

	_wgetenv_s(&ret_val, env_val, 1024, L"MPICH_DBG_FILENAME" );
	if (!ret_val) {
		filePattern = MPIU_Strdup( env_val );
	}

	_wgetenv_s(&ret_val, env_val, 1024, L"MPICH_DBG_RANK" );
	if (!ret_val) {
		wchar_t *sOut;
		whichRank = wcstol( env_val, &sOut, 10 );
		if (env_val == sOut) {
			MPIU_DBG_Usage( L"MPICH_DBG_RANK", 0 );
			whichRank = -1;
		}
	}
	return MPI_SUCCESS;
}

/*
 * Attempt to initialize the logging system.  This works only if MPID_Init
 * is not responsible for updating the environment and/or command-line
 * arguments. 
 */
int MPIU_DBG_PreInit( int *argc_p, wchar_t ***argv_p )
{
	MPID_Time_t t;

	/* if the DBG_MSG system was already initialized, say by the device, then
	   return immediately */
	if (mpiu_dbg_initialized != 0) return 0;

	/* Check to see if any debugging was selected.  The order of these
	   tests is important, as they allow general defaults to be set,
	   followed by more specific modifications */
	/* First, the environment variables */
	MPIU_DBG_ProcessEnv();

	MPIU_DBG_ProcessArgs( argc_p, argv_p );

	MPID_Wtime( &t );
	MPID_Wtime_todouble( &t, &timeOrigin );

	mpiu_dbg_initialized = 1;

	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIU_DBG_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_DBG_Init( int *argc_p, wchar_t ***argv_p, int has_args, int has_env, int wrank )
{
	//MPIDI_STATE_DECL(MPID_STATE_MPIU_DBG_INIT);

	//MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DBG_INIT);

	/* if the DBG_MSG system was already initialized, say by the device, then
	   return immediately.  Note that the device is then responsible
	   for handling the file mode (e.g., reopen when the rank become 
	   available) */
	if (mpiu_dbg_initialized == 2)
	{
		return 0;
	}

	/* Check to see if any debugging was selected.  The order of these
	   tests is important, as they allow general defaults to be set,
	   followed by more specific modifications */
	/* Both of these may have already been set in the PreInit call; 
	   if the command line and/or environment variables are set before
	   MPID_Init, then don't call the routines to check those values 
	   (as they were already handled in DBG_PreInit) */
	/* First, the environment variables */
	if (!has_env)
		MPIU_DBG_ProcessEnv();
	/* Now the command-line arguments */
	if (!has_args)
		MPIU_DBG_ProcessArgs( argc_p, argv_p );

	worldRank = wrank;

	if (whichRank >= 0 && whichRank != wrank) {
		/* Turn off logging on this process */
		MPIU_DBG_ActiveClasses = 0;
	}

	/* If the file has already been opened and we need to reopen it,
	   do that now.  We only need to close the file, since 
	   OutEvent will reopen it as necessary */
	if (filemode == MPIU_DBG_REOPEN && filestate == MPIU_DBG_PRERANK) {
		fclose( MPIU_DBG_fp );
		MPIU_DBG_fp = NULL;
	}

	mpiu_dbg_initialized = 2;
	//MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBG_INIT);
	return 0;
}

/* Print the usage statement to stderr */
#undef FUNCNAME
#define FUNCNAME MPIU_DBG_Usage
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIU_DBG_Usage( const wchar_t *cmd, const wchar_t *vals )
{
	//MPIDI_STATE_DECL(MPID_STATE_MPIU_DBG_USAGE);

	//MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DBG_USAGE);
	if (vals) {
		fwprintf( stderr, L"Incorrect value for %s, should be one of %s\n",
				 cmd, vals );
	}
	else {
		fwprintf( stderr, L"Incorrect value for %s\n", cmd );
	}
	fwprintf( stderr, 
L"Command line for debug switches\n\
	-mpich-dbg-class=name[,name,...]\n\
	-mpich-dbg-level=name   (one of terse, typical, verbose)\n\
	-mpich-dbg-filename=pattern (includes %%d for world rank, %%t for thread id\n\
	-mpich-dbg-rank=val	(only this rank in COMM_WORLD will be logged)\n\
	-mpich-dbg   (shorthand for -mpich-dbg-class=all -mpich-dbg-level=typical)\n\
	-mpich-dbg=file (shorthand for -mpich-dbg -mpich-dbg-filename=%s)\n\
Environment variables\n\
	MPICH_DBG_CLASS=NAME[,NAME...]\n\
	MPICH_DBG_LEVEL=NAME\n\
	MPICH_DBG_FILENAME=pattern\n\
	MPICH_DBG_RANK=val\n\
	MPICH_DBG=YES or FILE\n", defaultFilePattern );

	fflush(stderr);

	//MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBG_USAGE);
	return 0;
}
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

/* This routine can make no MPI calls, since it may be logging those
   calls. */
#undef FUNCNAME
#define FUNCNAME MPIU_DBG_OpenFile
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIU_DBG_OpenFile( void )
{
	int withinMworld = 0,		 /* True if within an @W...@ */
		withinMthread = 0;		/* True if within an @T...@ */
	/* FIXME: Need to know how many MPI_COMM_WORLDs are known */
	int nWorld = 1;
#if MPICH_IS_THREADED
	int nThread = 2;
#else
	int nThread = 1;
#endif
	static wchar_t worldNumAsChar[10] = L"0";
	errno_t error;
	//MPIDI_STATE_DECL(MPID_STATE_MPIU_DBG_OPENFILE);

	//MPIDI_FUNC_ENTER(MPID_STATE_MPIU_DBG_OPENFILE);

	/* FIXME: This is a hack to handle the common case of two worlds */
	if (MPIR_Process.comm_parent != NULL) {
		nWorld = 2;
		worldNumAsChar[0] = L'1';
		worldNumAsChar[1] = L'\0';
	}
		
	if (!filePattern || *filePattern == 0 ||
		wcscmp(filePattern, L"-stdout-" ) == 0) {
		filestate   = MPIU_DBG_OPEN;
		MPIU_DBG_fp = stdout;
	}
	else if (wcscmp( filePattern, L"-stderr-" ) == 0) {
		filestate   = MPIU_DBG_OPEN;
		MPIU_DBG_fp = stderr;
	}
	else {
		wchar_t filename[MAXPATHLEN], *pDest, *p;
		p	 = filePattern;
		pDest = filename;
		*filename = 0;
		while (*p && (pDest-filename) < MAXPATHLEN) {
			/* There are two special cases that allow text to
			   be optionally included.  Those patterns are
			   @T...@ (only if multi-threaded) and
			   @W...@ (only if more than one MPI_COMM_WORLD) 
			   UNIMPLEMENTED/UNTESTED */
			if (*p == L'@') {
				/* Escaped @? */
				if (p[1] == L'@') {
					*pDest++ = *++p;
					continue;
				}
				/* If within an @...@, terminate it */
				if (withinMworld) {
					withinMworld = 0;
					p++;
				}
				else if (withinMthread) {
					withinMthread = 0;
					p++;
				}
				else {
					/* Look for command */
					p++;
					if (*p == L'W') {
						p++;
						withinMworld = 1;
					}
					else if (*p == L'T') {
						p++;
						withinMthread = 1;
					}
					else {
						/* Unrecognized wchar_t */
						*pDest++ = *p++;
					}
				}
			}
			else if ( (withinMworld && nWorld == 1) ||
					  (withinMthread && nThread == 1) ) {
				/* Simply skip this character since we're not showing
				   this string */
				p++;
			}
			else if (*p == L'%') {
				p++;
				if (*p == L'd') {
					wchar_t rankAsChar[20];
					MPIU_Snprintf( rankAsChar, sizeof(rankAsChar)/sizeof(wchar_t), _TRUNCATE, L"%d", 
								   worldRank );
					*pDest = 0;
					MPIU_Wcsnapp( filename, rankAsChar, MAXPATHLEN );
					pDest += wcslen(rankAsChar);
				}
				else if (*p == L't') {
#if MPICH_IS_THREADED
					wchar_t threadIDAsChar[20];
					MPE_Thread_id_t tid;
					MPE_Thread_self(&tid);
					threadID = (int)tid;

					MPIU_Snprintf( threadIDAsChar, sizeof(threadIDAsChar)/sizeof(wchar_t), _TRUNCATE, 
								   L"%d", threadID );
					*pDest = 0;
					MPIU_Wcsnapp( filename, threadIDAsChar, MAXPATHLEN );
					pDest += wcslen(threadIDAsChar);
#else
					*pDest++ = L'0';
#endif
				}
				else if (*p == L'w') {
					/* FIXME: Get world number */
					/* *pDest++ = L'0'; */
					*pDest = 0;
					MPIU_Wcsnapp( filename, worldNumAsChar, MAXPATHLEN );
					pDest += wcslen(worldNumAsChar);
				}
				else {
					*pDest++ = L'%';
					*pDest++ = *p;
				}
				p++;
			}
			else {
				*pDest++ = *p++;
			}
		}
		*pDest = 0;
		if (worldRank == -1) filestate = MPIU_DBG_PRERANK;
		else				 filestate = MPIU_DBG_OPEN;
		error = _wfopen_s( &MPIU_DBG_fp, filename, L"w" );
		if (error != 0 || !MPIU_DBG_fp) {
			MPIU_Error_printf( L"Could not open log file %s\n", filename );
		}
	}
	//MPIDI_FUNC_EXIT(MPID_STATE_MPIU_DBG_OPENFILE);
	return 0;
}

/* Support routines for processing mpich-dbg values */
/* Update the GLOBAL variable MPIU_DBG_ActiveClasses with
   the bits corresponding to this name */
#undef FUNCNAME
#define FUNCNAME setDBGClass
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int setDBGClass( const wchar_t *s )
{
	int i;
	int slen = 0;
	int len = 0;
	//MPIDI_STATE_DECL(MPID_STATE_SETDBGCLASS);

	//MPIDI_FUNC_ENTER(MPID_STATE_SETDBGCLASS);
			
	if (s && *s) slen = (int)wcslen(s);

	while (s && *s) {
		for (i=0; MPIU_Classnames[i].LCName; i++) {
			/* The LCLen and UCLen *should* be the same, but
			   just in case, we separate them */
			int LClen = (int)wcslen(MPIU_Classnames[i].LCName);
			int UClen = (int)wcslen(MPIU_Classnames[i].UCName);
			int matchClass = 0;

			/* Allow the upper case and lower case in all cases */
			if (slen >= LClen && 
				wcsncmp(s,MPIU_Classnames[i].LCName, LClen) == 0 &&
				(s[LClen] == L',' || s[LClen] == 0) ) {
				matchClass = 1;
				len = LClen;
			}
			else if (slen >= UClen && 
				wcsncmp(s,MPIU_Classnames[i].UCName, UClen) == 0 &&
				(s[UClen] == L',' || s[UClen] == 0) ) {
				matchClass = 1;
				len = UClen;
			}
			if (matchClass) {
				MPIU_DBG_ActiveClasses |= MPIU_Classnames[i].classbits;
				s += len;
				slen -= len;
				if (*s == L',') { s++; slen--; }
				/* If we found a name, we need to restart the for loop */
				break;
			}
		}
		if (!MPIU_Classnames[i].LCName)
		{
			//MPIDI_FUNC_EXIT(MPID_STATE_SETDBGCLASS);
			return 1;
		}
	}
	//MPIDI_FUNC_EXIT(MPID_STATE_SETDBGCLASS);
	return 0;
}

/* Set the global MPIU_DBG_MaxLevel if there is a match with the known level
   names 
*/
#undef FUNCNAME
#define FUNCNAME SetDBGLevel
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int SetDBGLevel( const wchar_t *s, const wchar_t *(names[]) )
{
	int i;
	//MPIDI_STATE_DECL(MPID_STATE_SETDBGLEVEL);

	//MPIDI_FUNC_ENTER(MPID_STATE_SETDBGLEVEL);

	for (i=0; names[i]; i++) {
		if (wcscmp( names[i], s ) == 0) {
			MPIU_DBG_MaxLevel = MPIU_Levelvalues[i];
			//MPIDI_FUNC_EXIT(MPID_STATE_SETDBGLEVEL);
			return 0;
		}
	}
	//MPIDI_FUNC_EXIT(MPID_STATE_SETDBGLEVEL);
	return 1;
}
#endif /* USE_DBG_LOGGING */
