/*  (C) 2009 Deino Software. */

#include "mpiimpl.h"
#include "mpiinfo.h"
#include <string.h>

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_setA
#undef MPI_Info_setW
#define MPI_Info_setA PMPI_Info_setA
#define MPI_Info_setW PMPI_Info_setW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_set

/*@
	MPI_Info_set - Adds a (key,value) pair to info

Input Parameters:
+ info - info object (handle)
. key - key (string)
- value - value (string)

.N NotThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INFO_KEY
.N MPI_ERR_INFO_VALUE
.N MPI_ERR_EXHAUSTED
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_setA( MPI_Info info, char *key, char *value )
{
	static const wchar_t FCNAME[] = L"MPI_Info_setA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wkey = NULL, *wvalue = NULL;
	size_t key_len, value_len, len;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_SETA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_SETA);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			size_t keylen;
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
			keylen = strlen(key);
			MPIU_ERR_CHKANDJUMP((keylen > MPI_MAX_INFO_KEY), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeylong");
			MPIU_ERR_CHKANDJUMP((keylen == 0), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeyempty");

			/* Check value argument */
			MPIU_ERR_CHKANDJUMP((!value), mpi_errno, MPI_ERR_INFO_VALUE, L"**infovalnull");
			MPIU_ERR_CHKANDJUMP((strlen(value) > MPI_MAX_INFO_VAL), mpi_errno, MPI_ERR_INFO_VALUE, L"**infovallong");
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	key_len = strlen(key);
	value_len = strlen(value);
	wkey = MPIU_Malloc((key_len+1) * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, wkey, (key_len+1)*sizeof(wchar_t), L"wkey");
	wvalue = MPIU_Malloc((value_len+1) * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, wvalue, (value_len+1)*sizeof(wchar_t), L"wvalue");

	error = mbstowcs_s(&len, wkey, key_len+1, key, _TRUNCATE);
	if (error != 0)
	{
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}
	error = mbstowcs_s(&len, wvalue, value_len + 1, value, _TRUNCATE);
	if (error != 0)
	{
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}

	mpi_errno = MPI_Info_setW(info, wkey, wvalue);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wkey != NULL)
		MPIU_Free(wkey);
	if (wvalue != NULL)
		MPIU_Free(wvalue);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_SETA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Info_setW( MPI_Info info, wchar_t *key, wchar_t *value )
{
	static const wchar_t FCNAME[] = L"MPI_Info_set";
	int mpi_errno = MPI_SUCCESS;
	MPID_Info *info_ptr=0, *curr_ptr, *prev_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_SET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("info"); 
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_SET);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO(info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			size_t keylen;

			/* Validate info_ptr */
			MPID_Info_valid_ptr( info_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
			
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
			keylen = wcslen(key);
			MPIU_ERR_CHKANDJUMP((keylen > MPI_MAX_INFO_KEY), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeylong");
			MPIU_ERR_CHKANDJUMP((keylen == 0), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeyempty");

			/* Check value arguments */
			MPIU_ERR_CHKANDJUMP((!value), mpi_errno, MPI_ERR_INFO_VALUE, L"**infovalnull");
			MPIU_ERR_CHKANDJUMP((wcslen(value) > MPI_MAX_INFO_VAL), mpi_errno, MPI_ERR_INFO_VALUE, L"**infovallong");
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_set %I %s %s", info, key, value); }
	
	prev_ptr = info_ptr;
	curr_ptr = info_ptr->next;

	while (curr_ptr)
	{
		/*if (!wcsncmp(curr_ptr->key, key, MPI_MAX_INFO_KEY))*/
		if (!wcsncmp(curr_ptr->key, key, MPI_MAX_INFO_KEY))
		{
			/* Key already present; replace value */
			MPIU_Free(curr_ptr->value);  
			curr_ptr->value = MPIU_Strdup(value);
			MPIU_ERR_MEM_CHK_POP(mpi_errno, curr_ptr->value, (wcslen(value)+1) * sizeof(wchar_t), L"value");
			break;
		}
		prev_ptr = curr_ptr;
		curr_ptr = curr_ptr->next;
	}

	if (!curr_ptr)
	{
		/* Key not present, insert value */
		curr_ptr		 = (MPID_Info *)MPIU_Handle_obj_alloc( &MPID_Info_mem );

		MPIU_ERR_CHKANDJUMP1((!curr_ptr), mpi_errno, MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPI_Info" );

		/*wprintf( L"Inserting new elm %x at %x\n", curr_ptr->id, prev_ptr->id );*/
		prev_ptr->next   = curr_ptr;
		curr_ptr->key	= MPIU_Strdup(key);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, curr_ptr->key, (wcslen(key)+1)*sizeof(wchar_t), L"curr_ptr->key");
		curr_ptr->value  = MPIU_Strdup(value);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, curr_ptr->value, (wcslen(value)+1)*sizeof(wchar_t), L"curr_ptr->value");
		curr_ptr->next   = 0;
	}
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_SET);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_set",
			L"**mpi_info_set %I %s %s", info, key, value);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_get_valuelenA
#undef MPI_Info_get_valuelenW
#define MPI_Info_get_valuelenA PMPI_Info_get_valuelenA
#define MPI_Info_get_valuelenW PMPI_Info_get_valuelenW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_get_valuelen

/*@
	MPI_Info_get_valuelen - Retrieves the length of the value associated with 
	a key

	Input Parameters:
+ info - info object (handle)
- key - key (string)

	Output Parameters:
+ valuelen - length of value argument (integer)
- flag - true if key defined, false if not (boolean)

.N ThreadSafeInfoRead

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INFO_KEY
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_get_valuelenA( MPI_Info info, char *key, int *valuelen, int *flag )
{
	static const wchar_t FCNAME[] = L"MPI_Info_get_valuelenA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wkey = NULL;
	size_t key_len, len;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_GET_VALUELENA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_GET_VALUELENA);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			size_t keylen;
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
			keylen = strlen(key);
			MPIU_ERR_CHKANDJUMP((keylen > MPI_MAX_INFO_KEY), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeylong");
			MPIU_ERR_CHKANDJUMP((keylen == 0), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeyempty");
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	key_len = strlen(key);
	wkey = MPIU_Malloc((key_len+1) * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, wkey, (key_len+1)*sizeof(wchar_t), L"wkey");

	error = mbstowcs_s(&len, wkey, key_len+1, key, _TRUNCATE);
	if (error != 0)
	{
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}

	mpi_errno = MPI_Info_get_valuelenW(info, wkey, valuelen, flag);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wkey != NULL)
		MPIU_Free(wkey);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_GET_VALUELENA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Info_get_valuelenW( MPI_Info info, wchar_t *key, int *valuelen, int *flag )
{
	MPID_Info *curr_ptr, *info_ptr=0;
	static const wchar_t FCNAME[] = L"MPI_Info_get_valuelen";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_GET_VALUELEN);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("info");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_GET_VALUELEN);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO(info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			size_t keylen;

			/* Validate info_ptr */
			MPID_Info_valid_ptr( info_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
			
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
			keylen = wcslen(key);
			MPIU_ERR_CHKANDJUMP((keylen > MPI_MAX_INFO_KEY), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeylong");
			MPIU_ERR_CHKANDJUMP((keylen == 0), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeyempty");

			MPIR_ERRTEST_ARGNULL(valuelen, L"valuelen", mpi_errno);
			MPIR_ERRTEST_ARGNULL(flag, L"flag", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_get_valuelen %I %s %p %p", info, key, valuelen, flag); }

	curr_ptr = info_ptr->next;
	*flag = 0;

	while (curr_ptr)
	{
		if (!wcsncmp(curr_ptr->key, key, MPI_MAX_INFO_KEY))
		{
			*valuelen = (int)wcslen(curr_ptr->value);
			*flag = 1;
			break;
		}
		curr_ptr = curr_ptr->next;
	}

	/* ... end of body of routine ... */

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_GET_VALUELEN);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;

fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_get_valuelen",
			L"**mpi_info_get_valuelen %I %s %p %p", info, key, valuelen, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_get_nthkeyA
#undef MPI_Info_get_nthkeyW
#define MPI_Info_get_nthkeyA PMPI_Info_get_nthkeyA
#define MPI_Info_get_nthkeyW PMPI_Info_get_nthkeyW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_get_nthkey

/*@
	MPI_Info_get_nthkey - Returns the nth defined key in info

Input Parameters:
+ info - info object (handle)
- n - key number (integer)

Output Parameters:
. keys - key (string).  The maximum number of characters is 'MPI_MAX_INFO_KEY'.

.N ThreadSafeInfoRead

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_get_nthkeyA( MPI_Info info, int n, char *key )
{
	static const wchar_t FCNAME[] = L"MPI_Info_get_nthkeyA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t wkey[MPI_MAX_INFO_KEY+1];
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_GET_NTHKEYA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_GET_NTHKEYA);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	mpi_errno = MPI_Info_get_nthkeyW(info, n, wkey);
	MPIU_ERR_CHK_POP(mpi_errno);

	error = wcstombs_s(&num_converted, key, MPI_MAX_INFO_KEY, wkey, _TRUNCATE);
	if (error != 0)
	{
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}
	else if (num_converted == (size_t)-1)
	{
		/* unable to convert the wide characters to multi-byte characters */
		MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
	}
	if (num_converted >= MPI_MAX_INFO_KEY)
	{
		/* key name too long */
		MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
	}

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_GET_NTHKEYA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Info_get_nthkeyW( MPI_Info info, int n, wchar_t *key )
{
	MPID_Info *curr_ptr, *info_ptr=0;
	int	   nkeys;
	static const wchar_t FCNAME[] = L"MPI_Info_get_nthkey";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_GET_NTHKEY);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("info");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_GET_NTHKEY);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO(info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate info_ptr */
			MPID_Info_valid_ptr( info_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;

			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_get_nthkey %I %d %p", info, n, key); }
	
	curr_ptr = info_ptr->next;
	nkeys = 0;
	while (curr_ptr && nkeys != n)
	{
		curr_ptr = curr_ptr->next;
		nkeys++;
	}

	/* very that n is valid */
	MPIU_ERR_CHKANDJUMP2((!curr_ptr), mpi_errno, MPI_ERR_ARG, L"**infonkey", L"**infonkey %d %d", n, nkeys);

	MPIU_Wcsncpy(key, curr_ptr->key, MPI_MAX_INFO_KEY+1);
	/* Eventually, we could remember the location of this key in 
	   the head using the key/value locations (and a union datatype?) */

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_GET_NTHKEY);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_get_nthkey",
			L"**mpi_info_get_nthkey %I %d %p", info, n, key);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_get_nkeys
#define MPI_Info_get_nkeys PMPI_Info_get_nkeys
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_get_nkeys

/*@
	MPI_Info_get_nkeys - Returns the number of currently defined keys in info

Input Parameters:
. info - info object (handle)

Output Parameters:
. nkeys - number of defined keys (integer)

.N ThreadSafeInfoRead

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_get_nkeys( MPI_Info info, int *nkeys )
{
	MPID_Info *info_ptr=0;
	int	  n;
	static const wchar_t FCNAME[] = L"MPI_Info_get_nkeys";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_GET_NKEYS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("info");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_GET_NKEYS);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO(info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate info_ptr */
			MPID_Info_valid_ptr( info_ptr, mpi_errno );
			
			MPIR_ERRTEST_ARGNULL(nkeys,L"nkeys",mpi_errno);
			if (mpi_errno) goto fn_fail;

		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_get_nkeys %I %p", info, nkeys); }
	
	info_ptr = info_ptr->next;
	n = 0;

	while (info_ptr)
	{
		/*wprintf( L"Looking at %x\n", info_ptr->id );*/
		info_ptr = info_ptr->next;
		n ++;
	}
	*nkeys = n;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_GET_NKEYS);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_get_nkeys",
			L"**mpi_info_get_nkeys %I %p", info, nkeys);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_getA
#undef MPI_Info_getW
#define MPI_Info_getA PMPI_Info_getA
#define MPI_Info_getW PMPI_Info_getW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_get

/*@
	MPI_Info_get - Retrieves the value associated with a key

Input Parameters:
+ info - info object (handle)
. key - key (string)
- valuelen - length of value argument (integer)

Output Parameters:
+ value - value (string)
- flag - true if key defined, false if not (boolean)


.N ThreadSafeInfoRead

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
.N MPI_ERR_INFO_KEY
.N MPI_ERR_ARG
.N MPI_ERR_INFO_VALUE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_getA(MPI_Info info, char *key, int valuelen, char *value, int *flag)
{
	static const wchar_t FCNAME[] = L"MPI_Info_getA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wkey = NULL;
	wchar_t wvalue[MPI_MAX_INFO_VAL];
	size_t key_len, len;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_GETA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_GETA);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
			key_len = strlen(key);
			MPIU_ERR_CHKANDJUMP((key_len > MPI_MAX_INFO_KEY), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeylong");
			MPIU_ERR_CHKANDJUMP((key_len == 0), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeyempty");
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	key_len = strlen(key);
	wkey = MPIU_Malloc((key_len+1) * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, wkey, (key_len+1)*sizeof(wchar_t), L"wkey");

	error = mbstowcs_s(&len, wkey, key_len+1, key, _TRUNCATE);
	if (error != 0)
	{
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}

	mpi_errno = MPI_Info_getW(info, wkey, MPI_MAX_INFO_VAL, wvalue, flag);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (*flag)
	{
		error = wcstombs_s(&num_converted, value, valuelen, wvalue, _TRUNCATE);
		if (error != 0)
		{
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
		if (num_converted == (size_t)-1)
		{
			MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
		}
		if (num_converted >= MPI_MAX_INFO_VAL)
		{
			MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
		}
	}

fn_fail:
	if (wkey != NULL)
		MPIU_Free(wkey);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_GETA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Info_getW(MPI_Info info, wchar_t *key, int valuelen, wchar_t *value, int *flag)
{
	MPID_Info *curr_ptr, *info_ptr=0;
	static const wchar_t FCNAME[] = L"MPI_Info_get";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_GET);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("info");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_GET);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO(info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			size_t keylen;
			
			/* Validate info_ptr */
			MPID_Info_valid_ptr( info_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
			
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
			keylen = wcslen(key);
			MPIU_ERR_CHKANDJUMP((keylen > MPI_MAX_INFO_KEY), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeylong");
			MPIU_ERR_CHKANDJUMP((keylen == 0), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeyempty");

			/* Check value arguments */
			MPIR_ERRTEST_ARGNEG(valuelen, L"valuelen", mpi_errno);
			MPIU_ERR_CHKANDSTMT((!value), mpi_errno, MPI_ERR_INFO_VALUE,;, L"**infovalnull");
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_get %I %s %d %p %p", info, key, valuelen, value, flag); }

	curr_ptr = info_ptr->next;
	*flag = 0;

	while (curr_ptr)
	{
		/*if (!wcsncmp(curr_ptr->key, key, MPI_MAX_INFO_KEY))*/
		if (!wcsncmp(curr_ptr->key, key, MPI_MAX_INFO_KEY))
		{
			/*MPIU_Wcsncpy(value, curr_ptr->value, valuelen);*/
			MPIU_Wcsncpy(value, curr_ptr->value, valuelen);
			/* The following is problematic - if the user passes the
			   declared length, then this will access memory one
			   passed that point */
			/* value[valuelen] = L'\0'; */
			*flag = 1;
			break;
		}
		curr_ptr = curr_ptr->next;
	}
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_GET);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_get",
			L"**mpi_info_get %I %s %d %p %p", info, key, valuelen, value, flag);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_free
#define MPI_Info_free PMPI_Info_free
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_free

/*@
	MPI_Info_free - Frees an info object

Input Parameter:
. info - info object to be freed (handle)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INFO
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_free( MPI_Info *info )
{
	static const wchar_t FCNAME[] = L"MPI_Info_free";
	int mpi_errno = MPI_SUCCESS;
	MPID_Info *info_ptr=0;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_FREE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("info");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_FREE);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO(*info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( *info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate info_ptr */
			MPID_Info_valid_ptr( info_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_free %p", info); }
	
	MPIU_Info_free( info_ptr );
	*info = MPI_INFO_NULL;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_FREE);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_free",
			L"**mpi_info_free %p", info);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_dup
#define MPI_Info_dup PMPI_Info_dup
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_dup

/*@
	MPI_Info_dup - Returns a duplicate of the info object

Input Parameters:
. info - info object (handle)

Output Parameters:
. newinfo - duplicate of info object (handle)

.N ThreadSafeInfoRead

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_dup( MPI_Info info, MPI_Info *newinfo )
{
	MPID_Info *info_ptr=0, *curr_old, *curr_new;
	static const wchar_t FCNAME[] = L"MPI_Info_dup";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_DUP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("info");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_DUP);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO(info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate info_ptr */
			MPID_Info_valid_ptr( info_ptr, mpi_errno );
			MPIR_ERRTEST_ARGNULL(newinfo,L"newinfo",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_dup %I %p", info, newinfo); }
	
	/* Note that this routine allocates info elements one at a time.
	   In the multithreaded case, each allocation may need to acquire
	   and release the allocation lock.  If that is ever a problem, we
	   may want to add an "allocate n elements" routine and execute this
	   it two steps: count and then allocate */
	/* FIXME : multithreaded */
	curr_new		= (MPID_Info *)MPIU_Handle_obj_alloc( &MPID_Info_mem );
	curr_new->key   = 0;
	curr_new->value = 0;
	curr_new->next  = 0;
	*newinfo = curr_new->handle;

	curr_old		= info_ptr->next;
	while (curr_old)
	{
		curr_new->next = (MPID_Info *)MPIU_Handle_obj_alloc( &MPID_Info_mem );
		MPIU_ERR_CHKANDJUMP1((!curr_new->next), mpi_errno, MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPI_Info");

		curr_new		 = curr_new->next;
		curr_new->key		 = MPIU_Strdup(curr_old->key);
		curr_new->value		 = MPIU_Strdup(curr_old->value);
		curr_new->next		 = 0;
		
		curr_old		 = curr_old->next;
	}
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_DUP);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_dup",
			L"**mpi_info_dup %I %p", info, newinfo);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_deleteA
#undef MPI_Info_deleteW
#define MPI_Info_deleteA PMPI_Info_deleteA
#define MPI_Info_deleteW PMPI_Info_deleteW
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_delete

/*@
  MPI_Info_delete - Deletes a (key,value) pair from info

  Input Parameters:
+ info - info object (handle)
- key - key (string)

.N NotThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N 
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_deleteA( MPI_Info info, char *key )
{
	static const wchar_t FCNAME[] = L"MPI_Info_deleteA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *wkey = NULL;
	size_t key_len, len;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_DELETEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_DELETEA);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			size_t keylen;
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
			keylen = strlen(key);
			MPIU_ERR_CHKANDJUMP((keylen > MPI_MAX_INFO_KEY), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeylong");
			MPIU_ERR_CHKANDJUMP((keylen == 0), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeyempty");
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	key_len = strlen(key);
	wkey = MPIU_Malloc((key_len+1) * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, wkey, (key_len+1)*sizeof(wchar_t), L"wkey");

	error = mbstowcs_s(&len, wkey, key_len+1, key, _TRUNCATE);
	if (error != 0)
	{
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}

	mpi_errno = MPI_Info_deleteW(info, wkey);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	if (wkey != NULL)
		MPIU_Free(wkey);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_DELETEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Info_deleteW( MPI_Info info, wchar_t *key )
{
	static const wchar_t FCNAME[] = L"MPI_Info_delete";
	int mpi_errno = MPI_SUCCESS;
	MPID_Info *info_ptr=0, *prev_ptr, *curr_ptr;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_DELETE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("info");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_DELETE);
	

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_INFO(info, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Info_get_ptr( info, info_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			size_t keylen;
			
			/* Validate info_ptr */
			MPID_Info_valid_ptr( info_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
			
			/* Check key */
			MPIU_ERR_CHKANDJUMP((!key), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeynull");
			keylen = wcslen(key);
			MPIU_ERR_CHKANDJUMP((keylen > MPI_MAX_INFO_KEY), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeylong");
			MPIU_ERR_CHKANDJUMP((keylen == 0), mpi_errno, MPI_ERR_INFO_KEY, L"**infokeyempty");
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_delete %I %s", info, key); }
	
	prev_ptr = info_ptr;
	curr_ptr = info_ptr->next;

	while (curr_ptr)
	{
		if (!wcsncmp(curr_ptr->key, key, MPI_MAX_INFO_KEY))
		{
			MPIU_Free(curr_ptr->key);   
			MPIU_Free(curr_ptr->value);
			prev_ptr->next = curr_ptr->next;
			MPIU_Handle_obj_free( &MPID_Info_mem, curr_ptr );
			break;
		}
		prev_ptr = curr_ptr;
		curr_ptr = curr_ptr->next;
	}

	/* If curr_ptr is not defined, we never found the key */
	MPIU_ERR_CHKANDJUMP1((!curr_ptr), mpi_errno, MPI_ERR_INFO_NOKEY, L"**infonokey", L"**infonokey %s", key);
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_DELETE);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_delete",
			L"**mpi_info_delete %I %s", info, key);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Info_create
#define MPI_Info_create PMPI_Info_create
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Info_create

/*@
	MPI_Info_create - Creates a new info object

   Output Parameter:
. info - info object created (handle)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Info_create( MPI_Info *info )
{
	MPID_Info *info_ptr;
	static const wchar_t FCNAME[] = L"MPI_Info_create";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_INFO_CREATE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("info");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_INFO_CREATE);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(info, L"info", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_info_create %p", info); }
	
	info_ptr = (MPID_Info *)MPIU_Handle_obj_alloc( &MPID_Info_mem );
	MPIU_ERR_CHKANDJUMP1((!info_ptr), mpi_errno, MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPI_Info");

	*info			 = info_ptr->handle;
	/* (info_ptr)->cookie = MPIR_INFO_COOKIE; */
	info_ptr->key	= 0;
	info_ptr->value  = 0;
	info_ptr->next   = 0;
	/* this is the first structure in this linked list. it is 
	   always kept empty. new (key,value) pairs are added after it. */
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_INFO_CREATE);
	MPIU_THREAD_SINGLE_CS_EXIT("info");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_info_create",
			L"**mpi_info_create %p", info);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
