/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

/* 
 * This file contains "safe" versions of the various string and wprintf
 * operations.
 */

/*
 * MPIU_Wcsncpy - Copy at most n character.  Stop once a null is reached.
 *
 * This is different from wcsncpy, which null pads so that exactly
 * n characters are copied.  The wcsncpy behavior is correct for many 
 * applications because it guarantees that the string has no uninitialized
 * data.
 *
 * If n characters are copied without reaching a null, return an error.
 * Otherwise, return 0.
 *
 * Question: should we provide a way to request the length of the string,
 * since we know it?
 */
/*@ MPIU_Wcsncpy - Copy a string with a maximum length
  
	Input Parameters:
+   instr - String to copy
-   maxlen - Maximum total length of 'outstr'

	Output Parameter:
.   outstr - String to copy into

	Notes:
	This routine is the routine that you wish 'wcsncpy' was.  In copying 
	'instr' to 'outstr', it stops when either the end of 'outstr' (the 
	null character) is seen or the maximum length 'maxlen' is reached.
	Unlike 'wcsncpy', it does not add enough nulls to 'outstr' after 
	copying 'instr' in order to move precisely 'maxlen' characters.  
	Thus, this routine may be used anywhere 'wcscpy' is used, without any
	performance cost related to large values of 'maxlen'.

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Strncpy
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
int MPIU_Strncpy( char *dest, const char *src, size_t n )
{
	char * restrict d_ptr = dest;
	const char * restrict s_ptr = src;
	register int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STRNCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STRNCPY);

	i = (int)n;
	while (*s_ptr && i-- > 0) {
		*d_ptr++ = *s_ptr++;
	}
	
	if (i > 0) { 
		*d_ptr = 0;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STRNCPY);
		return 0;
	}

	/* We may want to force an error message here, at least in the debugging version */
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STRNCPY);
	return 1;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Wcsncpy
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
int MPIU_Wcsncpy( wchar_t *dest, const wchar_t *src, size_t n )
{
	wchar_t * restrict d_ptr = dest;
	const wchar_t * restrict s_ptr = src;
	register int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_WCSNCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_WCSNCPY);

	i = (int)n;
	while (*s_ptr && i-- > 0)
	{
		*d_ptr++ = *s_ptr++;
	}

	if (i > 0)
	{
		*d_ptr = 0;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_WCSNCPY);
		return 0;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_WCSNCPY);
	return 1;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Mbstowcs
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
int MPIU_Mbstowcs(const char *str, wchar_t **wstr_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	size_t len;
	wchar_t *wstr;
	size_t num_converted;
	errno_t error;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_MBSTOWCS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_MBSTOWCS);

	len = strlen(str) + 1;
	wstr = MPIU_Malloc(len * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, wstr, len * sizeof(wchar_t), "wstr");
	wstr[0] = L'\0';
	error = mbstowcs_s(&num_converted, wstr, len, str, _TRUNCATE);
	if (error != 0)
	{
		if (num_converted <= len)
		{
			MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", wstr, num_converted);
			goto fn_fail;
		}
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}
	/* This case cannot happen because any failure or truncation will cause error != 0
	if (num_converted != len)
	{
		MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", wstr, num_converted);
		goto fn_fail;
	}
	*/

	*wstr_ptr = wstr;
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_MBSTOWCS);
	return mpi_errno;
}

/* Append src to dest, but only allow dest to contain n characters (including
   any null, which is always added to the end of the line */
/*@ MPIU_Wcsnapp - Append to a string with a maximum length

	Input Parameters:
+   instr - String to copy
-   maxlen - Maximum total length of 'outstr'

	Output Parameter:
.   outstr - String to copy into

	Notes:
	This routine is similar to 'wcsncat_s' except that the 'maxlen' argument
	is the maximum total length of 'outstr', rather than the maximum 
	number of characters to move from 'instr'.  Thus, this routine is
	easier to use when the declared size of 'instr' is known.

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Strnapp
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
int MPIU_Strnapp( char *dest, const char *src, size_t n )
{
	char * restrict d_ptr = dest;
	const char * restrict s_ptr = src;
	register int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STRNAPP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STRNAPP);

	/* Get to the end of dest */
	i = (int)n;
	while (i-- > 0 && *d_ptr) d_ptr++;
	if (i <= 0)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STRNAPP);
		return 1;
	}

	/* Append.  d_ptr points at first null and i is remaining space. */
	while (*s_ptr && i-- > 0) {
		*d_ptr++ = *s_ptr++;
	}

	/* We allow i >= (not just >) here because the first while decrements
	   i by one more than there are characters, leaving room for the null */
	if (i >= 0) { 
		*d_ptr = 0;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STRNAPP);
		return 0;
	}

	/* Force the null at the end */
	*--d_ptr = 0;

	/* We may want to force an error message here, at least in the debugging version */
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STRNAPP);
	return 1;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Wcsnapp
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
int MPIU_Wcsnapp( wchar_t *dest, const wchar_t *src, size_t n )
{
	wchar_t * restrict d_ptr = dest;
	const wchar_t * restrict s_ptr = src;
	register int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_WCSNAPP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_WCSNAPP);

	/* Get to the end of dest */
	i = (int)n;
	while (i-- > 0 && *d_ptr) d_ptr++;
	if (i <= 0)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_WCSNAPP);
		return 1;
	}

	/* Append.  d_ptr points at first null and i is remaining space. */
	while (*s_ptr && i-- > 0)
	{
		*d_ptr++ = *s_ptr++;
	}

	/* We allow i >= (not just >) here because the first while decrements
	   i by one more than there are characters, leaving room for the null */
	if (i >= 0)
	{
		*d_ptr = 0;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_WCSNAPP);
		return 0;
	}

	/* Force the null at the end */
	*--d_ptr = 0;

	/* We may want to force an error message here, at least in the
	   debugging version */
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_WCSNAPP);
	return 1;
}

#ifndef HAVE_STRDUP
#ifdef MPIU_Strdup
#undef MPIU_Strdup
#endif
/*@ 
  MPIU_Strdup - Duplicate a string

  Synopsis:
.vb
	char *MPIU_Strdup( const char *str )
.ve

  Input Parameter:
. str - null-terminated string to duplicate

  Return value:
  A pointer to a copy of the string, including the terminating null.  A
  null pointer is returned on error, such as out-of-memory.

  Notes:
  Like 'MPIU_Malloc' and 'MPIU_Free', this will often be implemented as a 
  macro but may use 'MPIU_trstrdup' to provide a tracing version.

  Module:
  Utility
  @*/
/*
#undef FUNCNAME
#define FUNCNAME MPIU_Strdup
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
char *MPIU_Strdup( const char *str )
{
	char * restrict p = (char *)MPIU_Malloc( strlen(str) + 1 );
	const char * restrict in_p = str;
	char *save_p;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STRDUP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STRDUP);

	save_p = p;
	if (p) {
		while (*in_p) {
			*p++ = *in_p++;
		}
		*p = 0;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STRDUP);
	return save_p;
}
*/

#undef FUNCNAME
#define FUNCNAME MPIU_Strdup
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
wchar_t *MPIU_Strdup( const wchar_t *str )
{
	wchar_t * restrict p = (wchar_t *)MPIU_Malloc( (wcslen(str) + 1) * sizeof(wchar_t) );
	const wchar_t * restrict in_p = str;
	wchar_t *save_p;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STRDUP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STRDUP);

	save_p = p;
	if (p)
	{
		while (*in_p)
		{
			*p++ = *in_p++;
		}
		*p = 0;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STRDUP);
	return save_p;
}
#endif

/* 
 * We need an _snwprintf replacement for systems without one
 */
#ifndef HAVE_SNPRINTF
/* 
 * This is an approximate form which is suitable for most uses within
 * the MPICH code
 */
#undef FUNCNAME
#define FUNCNAME MPIU_Snprintf
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
int MPIU_Snprintf( char *str, size_t size, const char *format, ... )
{
	int n;
	const char *p;
	char *out_str = str;
	va_list list;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_SNPRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_SNPRINTF);

	va_start(list, format);

	p = format;
	while (*p && size > 0) {
		char *nf;

		nf = wcschr(p, L'%');
		if (!nf) {
			/* No more format characters */
			while (size-- > 0 && *p) {
				*out_str++ = *p++;
			}
		}
		else {
			int nc;
			int width = -1;

			/* Copy until nf */
			while (p < nf && size-- > 0) {
				*out_str++ = *p++;
			}
			/* p now points at nf */
			/* Handle the format character */
			nc = nf[1];
			if (isdigit(nc)) {
				/* Get the field width */
				/* FIXME : Assumes ASCII */
				width = nc - L'0';
				p = nf + 2;
				while (*p && isdigit(*p)) {
					width = 10 * width + (*p++ - L'0');
				}
				/* When there is no longer a digit, get the format 
				   character */
				nc = *p++;
			}
			else {
				/* Skip over the format string */
				p += 2;
			}

			switch (nc) {
			case L'%':
				*out_str++ = L'%';
				size--;
				break;

			case L'd':
			{
				int val;
				char tmp[20];
				char *t = tmp;
				/* Get the argument, of integer type */
				val = va_arg( list, int );
				swprintf( tmp, L"%d", val );
				if (width > 0) {
					int tmplen = wcslen(tmp);
					/* If a width was specified, pad with spaces on the
					   left (on the right if %-3d given; not implemented yet */
					while (size-- > 0 && width-- > tmplen) 
						*out_str++ = L' ';
				}
				while (size-- > 0 && *t) {
					*out_str++ = *t++;
				}
			}
			break;

			case L'x':
			{
				int val;
				char tmp[20];
				char *t = tmp;
				/* Get the argument, of integer type */
				val = va_arg( list, int );
				swprintf( tmp, L"%x", val );
				if (width > 0) {
					int tmplen = wcslen(tmp);
					/* If a width was specified, pad with spaces on the
					   left (on the right if %-3d given; not implemented yet */
					while (size-- > 0 && width-- > tmplen) 
						*out_str++ = L' ';
				}
				while (size-- > 0 && *t) {
					*out_str++ = *t++;
				}
			}
			break;

			case L'p':
			{
				int val;
				char tmp[20];
				char *t = tmp;
				/* Get the argument, of integer type */
				val = va_arg( list, int );
				swprintf( tmp, L"%p", val );
				if (width > 0) {
					int tmplen = wcslen(tmp);
					/* If a width was specified, pad with spaces on the
					   left (on the right if %-3d given; not implemented yet */
					while (size-- > 0 && width-- > tmplen) 
						*out_str++ = L' ';
				}
				while (size-- > 0 && *t) {
					*out_str++ = *t++;
				}
			}
			break;

			case L's':
			{
				char *s_arg;
				/* Get the argument, of pointer to char type */
				s_arg = va_arg( list, char * );
				while (size-- > 0 && s_arg && *s_arg) {
					*out_str++ = *s_arg++;
				}
			}
			break;

			default:
				/* Error, unknown case */
				MPIDI_FUNC_EXIT(MPID_STATE_MPIU_SNPRINTF);
				return -1;
				break;
			}
		}
	}

	va_end(list);

	if (size-- > 0) *out_str++ = L'\0';

	n = (int)(out_str - str);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_SNPRINTF);
	return n;
}
#endif
