/*  (C) 2009 by Deino Software. */

#include "mpiimpl.h"
#include "datatype.h"

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_commit
#define MPI_Type_commit PMPI_Type_commit

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_commit

/*@
	MPI_Type_commit - Commits the datatype

Input Parameter:
. datatype - datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_commit(MPI_Datatype *datatype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_commit";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_COMMIT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_COMMIT);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			MPIR_ERRTEST_DATATYPE(*datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( *datatype, datatype_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_commit %p", datatype); }
	
	if (HANDLE_GET_KIND(*datatype) == HANDLE_KIND_BUILTIN) goto fn_exit;

	/* pair types stored as real types are a special case */
	if (*datatype == MPI_FLOAT_INT ||
		*datatype == MPI_DOUBLE_INT ||
		*datatype == MPI_LONG_INT ||
		*datatype == MPI_SHORT_INT ||
		*datatype == MPI_LONG_DOUBLE_INT) goto fn_exit;

	mpi_errno = MPID_Type_commit(datatype);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_COMMIT);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_commit",
			L"**mpi_type_commit %p", datatype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}


/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Status_set_elements
#define MPI_Status_set_elements PMPI_Status_set_elements

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Status_set_elements

/*@
   MPI_Status_set_elements - Set the number of elements in a status

   Input Parameters:
+ status - status to associate count with (Status) 
. datatype - datatype associated with count (handle) 
- count - number of elements to associate with status (integer) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_TYPE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Status_set_elements(MPI_Status *status, MPI_Datatype datatype, int count)
{
	static const wchar_t FCNAME[] = L"MPI_Status_set_elements";
	int mpi_errno = MPI_SUCCESS;
	int size;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_STATUS_SET_ELEMENTS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_STATUS_SET_ELEMENTS);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_COUNT(count,mpi_errno);
			MPIR_ERRTEST_ARGNULL(status,L"status",mpi_errno);
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;

			/* Validate datatype_ptr */
			MPID_Datatype_get_ptr( datatype, datatype_ptr );
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			/* If datatype_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_status_set_elements %p %D %d", status, datatype, count); }
	
	MPID_Datatype_get_size_macro(datatype, size);
	/* FIXME: The device may want something else here */
	status->count = count * size;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_STATUS_SET_ELEMENTS);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_status_set_elements",
			L"**mpi_status_set_elements %p %D %d", status, datatype, count);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Register_datarepA
#undef MPI_Register_datarepW
#define MPI_Register_datarepA PMPI_Register_datarepA
#define MPI_Register_datarepW PMPI_Register_datarepW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Register_datarep

/*@
   MPI_Register_datarep - Register a set of user-provided data conversion 
   functions

   Input Parameters:
+ datarep - data representation identifier (string) 
. read_conversion_fn - function invoked to convert from file representation to native representation (function) 
. write_conversion_fn - function invoked to convert from native representation to file representation (function) 
. dtype_file_extent_fn - function invoked to get the extent of a datatype as represented in the file (function) 
- extra_state - extra state that is passed to the conversion functions

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Register_datarepW(wchar_t *datarep, 
						 MPI_Datarep_conversion_function *read_conversion_fn, 
						 MPI_Datarep_conversion_function *write_conversion_fn, 
						 MPI_Datarep_extent_function *dtype_file_extent_fn, 
						 void *extra_state)
{
	static const wchar_t FCNAME[] = L"MPI_Register_datarepW";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_REGISTER_DATAREP);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_REGISTER_DATAREP);
	MPIU_UNREFERENCED_ARG(datarep);
	MPIU_UNREFERENCED_ARG(read_conversion_fn);
	MPIU_UNREFERENCED_ARG(write_conversion_fn);
	MPIU_UNREFERENCED_ARG(dtype_file_extent_fn);
	MPIU_UNREFERENCED_ARG(extra_state);
	mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**notimpl", 0);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_REGISTER_DATAREP);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Register_datarepA(char *datarep, 
						 MPI_Datarep_conversion_function *read_conversion_fn, 
						 MPI_Datarep_conversion_function *write_conversion_fn, 
						 MPI_Datarep_extent_function *dtype_file_extent_fn, 
						 void *extra_state)
{
	static const wchar_t FCNAME[] = L"MPI_Register_datarep";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_REGISTER_DATAREPA);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_REGISTER_DATAREPA);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_register_datarep %s %p %p %p %p", datarep, read_conversion_fn, write_conversion_fn,
		dtype_file_extent_fn, extra_state); }
	
	/* FIXME UNIMPLEMENTED */
	MPIU_UNREFERENCED_ARG(datarep);
	MPIU_UNREFERENCED_ARG(read_conversion_fn);
	MPIU_UNREFERENCED_ARG(write_conversion_fn);
	MPIU_UNREFERENCED_ARG(dtype_file_extent_fn);
	MPIU_UNREFERENCED_ARG(extra_state);
	mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**notimpl", 0);

	/* ... end of body of routine ... */
	
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_REGISTER_DATAREPA);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_register_datarep",
			L"**mpi_register_datarep %s %p %p %p %p", datarep, read_conversion_fn, write_conversion_fn,
			dtype_file_extent_fn, extra_state);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Pack_size
#define MPI_Pack_size PMPI_Pack_size

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Pack_size

/*@
   MPI_Pack_size - Returns the upper bound on the amount of space needed to
					pack a message

Input Parameters:
+ incount - count argument to packing call (integer) 
. datatype - datatype argument to packing call (handle) 
- comm - communicator argument to packing call (handle) 

Output Parameter:
. size - upper bound on size of packed message, in bytes (integer) 

Notes:
The MPI standard document describes this in terms of 'MPI_Pack', but it 
applies to both 'MPI_Pack' and 'MPI_Unpack'.  That is, the value 'size' is 
the maximum that is needed by either 'MPI_Pack' or 'MPI_Unpack'.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_TYPE
.N MPI_ERR_ARG

@*/
MPI_DLL_SPEC int MPI_CALL MPI_Pack_size(int incount, MPI_Datatype datatype, MPI_Comm comm, int *size)
{
	static const wchar_t FCNAME[] = L"MPI_Pack_size";
	MPID_Comm *comm_ptr = NULL;
	int mpi_errno = MPI_SUCCESS;
	int typesize;
#ifdef HAVE_ERROR_CHECKING
	MPID_Datatype *datatype_ptr = NULL;
#endif
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PACK_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_PACK_SIZE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr( comm, comm_ptr );
	
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COUNT(incount, mpi_errno);
			MPIR_ERRTEST_ARGNULL(size, L"size", mpi_errno);
			if (mpi_errno) goto fn_fail;
			
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
			
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;
			
			if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_pack_size %d %D %C %p", incount, datatype, comm, size); }
	
	MPID_Datatype_get_size_macro(datatype, typesize);
	*size = incount * typesize;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_PACK_SIZE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_pack_size",
			L"**mpi_pack_size %d %D %C %p", incount, datatype, comm, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Pack_external_sizeA
#undef MPI_Pack_external_sizeW
#define MPI_Pack_external_sizeA PMPI_Pack_external_sizeA
#define MPI_Pack_external_sizeW PMPI_Pack_external_sizeW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Pack_external_size

/*@
  MPI_Pack_external_size - Returns the upper bound on the amount of
  space needed to pack a message using MPI_Pack_external.

   Input Parameters:
+ datarep - data representation (string)  
. incount - number of input data items (integer)  
- datatype - datatype of each input data item (handle)  

   Output Parameters:
. size - output buffer size, in bytes (integer)  

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Pack_external_sizeA(char *datarep, int incount, MPI_Datatype datatype, MPI_Aint *size)
{
	static const wchar_t FCNAME[] = L"MPI_Pack_external_sizeA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *tmp_datarep = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PACK_EXTERNAL_SIZEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_PACK_EXTERNAL_SIZEA);

	if (datarep != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(datarep, &tmp_datarep);
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = MPI_Pack_external_sizeW(tmp_datarep, incount, datatype, size);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		/* Invalid datarep string */
		MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
	}

fn_fail:
	if (tmp_datarep != NULL)
		MPIU_Free(tmp_datarep);
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_PACK_EXTERNAL_SIZEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Pack_external_sizeW(wchar_t *datarep, int incount, MPI_Datatype datatype, MPI_Aint *size)
{
	static const wchar_t FCNAME[] = L"MPI_Pack_external_size";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PACK_EXTERNAL_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_PACK_EXTERNAL_SIZE);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COUNT(incount,mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr(datatype, datatype_ptr);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			/* If datatype_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_pack_external_size %s %d %D %p", datarep, incount, datatype, size); }
	
	*size = incount * MPID_Datatype_size_external32(datatype);

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_PACK_EXTERNAL_SIZE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_pack_external_size",
			L"**mpi_pack_external_size %s %d %D %p", datarep, incount, datatype, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Pack_externalA
#undef MPI_Pack_externalW
#define MPI_Pack_externalA PMPI_Pack_externalA
#define MPI_Pack_externalW PMPI_Pack_externalW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Pack_external

/*@
   MPI_Pack_external - Packs a datatype into contiguous memory, using the
	 external32 format

   Input Parameters:
+ datarep - data representation (string)  
. inbuf - input buffer start (choice)  
. incount - number of input data items (integer)  
. datatype - datatype of each input data item (handle)  
- outsize - output buffer size, in bytes (integer)  

   Output Parameter:
. outbuf - output buffer start (choice)  

   Input/Output Parameter:
. position - current position in buffer, in bytes (integer)  

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
.N MPI_ERR_COUNT
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Pack_externalA(char *datarep, void *inbuf, int incount, MPI_Datatype datatype, void *outbuf, MPI_Aint outcount, MPI_Aint *position)
{
	static const wchar_t FCNAME[] = L"MPI_Pack_externalA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *tmp_datarep = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PACK_EXTERNALA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_PACK_EXTERNALA);

	if (datarep != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(datarep, &tmp_datarep);
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = MPI_Pack_externalW(tmp_datarep, inbuf, incount, datatype, outbuf, outcount, position);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		/* Invalid datarep string */
		MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
	}

fn_fail:
	if (tmp_datarep != NULL)
	{
		MPIU_Free(tmp_datarep);
	}
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_PACK_EXTERNALA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Pack_externalW(wchar_t *datarep, void *inbuf, int incount, MPI_Datatype datatype, void *outbuf, MPI_Aint outcount, MPI_Aint *position)
{
	static const wchar_t FCNAME[] = L"MPI_Pack_external";
	int mpi_errno = MPI_SUCCESS;
	MPI_Aint first, last;
	MPID_Segment *segp;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PACK_EXTERNAL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_PACK_EXTERNAL);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COUNT(incount, mpi_errno);
			MPIR_ERRTEST_COUNT(outcount, mpi_errno);
			/* NOTE: inbuf could be null (MPI_BOTTOM) */
			MPIR_ERRTEST_ARGNULL(outbuf, L"output buffer", mpi_errno);
			MPIR_ERRTEST_ARGNULL(position, L"position", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;

			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno == MPI_SUCCESS)
			{
				if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
				{
					MPID_Datatype *datatype_ptr = NULL;

					MPID_Datatype_get_ptr(datatype, datatype_ptr);
					MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
					MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
				}
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
	
	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_pack_external %s %p %d %D %p %d %p", datarep, inbuf, incount, datatype, outbuf, outcount, position); }
	
	segp = MPID_Segment_alloc();
	/* --BEGIN ERROR HANDLING-- */
	if (segp == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPID_Segment");
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */
	mpi_errno = MPID_Segment_init(inbuf, incount, datatype, segp, 1);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* NOTE: the use of buffer values and positions in MPI_Pack_external and
	 * in MPID_Segment_pack_external are quite different.  See code or docs
	 * or something.
	 */
	first = 0;
	last  = SEGMENT_IGNORE_LAST;

	MPID_Segment_pack_external32(segp, first, &last, (void *)((char *) outbuf + *position));

	*position += (int) last;

	MPID_Segment_free(segp);

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_PACK_EXTERNAL);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_pack_external",
			L"**mpi_pack_external %s %p %d %D %p %d %p", datarep, inbuf, incount, datatype, outbuf, outcount, position);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Pack
#define MPI_Pack PMPI_Pack

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Pack

/*@
	MPI_Pack - Packs a datatype into contiguous memory

   Arguments:
+  void *inbuf - input buffer
.  int incount - input count
.  MPI_Datatype datatype - datatype
.  void *outbuf - output buffer
.  int outcount - output count
.  int *position - position
-  MPI_Comm comm - communicator

   Notes (from the specifications):

   The input value of position is the first location in the output buffer to be
   used for packing.  position is incremented by the size of the packed message,
   and the output value of position is the first location in the output buffer
   following the locations occupied by the packed message.  The comm argument is
   the communicator that will be subsequently used for sending the packed
   message.


.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Pack(void *inbuf,
			 int incount,
			 MPI_Datatype datatype,
			 void *outbuf, 
			 int outcount,
			 int *position,
			 MPI_Comm comm)
{
	static const wchar_t FCNAME[] = L"MPI_Pack";
	int mpi_errno = MPI_SUCCESS;
	MPI_Aint first, last;
	MPID_Comm *comm_ptr = NULL;
	MPID_Segment *segp;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_PACK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_PACK);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr(comm, comm_ptr);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COUNT(incount,mpi_errno);
			MPIR_ERRTEST_COUNT(outcount,mpi_errno);
			/* NOTE: inbuf could be null (MPI_BOTTOM) */
			MPIR_ERRTEST_ARGNULL(outbuf, L"output buffer", mpi_errno);
			MPIR_ERRTEST_ARGNULL(position, L"position", mpi_errno);
			/* Validate comm_ptr */
			/* If comm_ptr is not valid, it will be reset to null */
			MPID_Comm_valid_ptr(comm_ptr, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;

			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno == MPI_SUCCESS)
			{
				if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
				{
					MPID_Datatype *datatype_ptr = NULL;

					MPID_Datatype_get_ptr(datatype, datatype_ptr);
					MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
					MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
				}
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

#ifdef HAVE_ERROR_CHECKING /* IMPLEMENTATION-SPECIFIC ERROR CHECKS */
	{
		int tmp_sz;

		MPID_BEGIN_ERROR_CHECKS;
		/* Verify that there is space in the buffer to pack the type */
		MPID_Datatype_get_size_macro(datatype, tmp_sz);

		if (tmp_sz * incount > outcount - *position)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**arg", 0);
			goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#endif /* HAVE_ERROR_CHECKING */
	
	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_pack %p %d %D %p %d %p %C", inbuf, incount, datatype, outbuf, outcount, position, comm); }
	
	/* TODO: CHECK RETURN VALUES?? */
	/* TODO: SHOULD THIS ALL BE IN A MPID_PACK??? */
	segp = MPID_Segment_alloc();
	/* --BEGIN ERROR HANDLING-- */
	if (segp == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPID_Segment");
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */
	mpi_errno = MPID_Segment_init(inbuf, incount, datatype, segp, 0);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	/* NOTE: the use of buffer values and positions in MPI_Pack and in
	 * MPID_Segment_pack are quite different.  See code or docs or something.
	 */
	first = 0;
	last  = SEGMENT_IGNORE_LAST;

	MPID_Segment_pack(segp, first, &last, (void *) ((char *) outbuf + *position));

	*position += (int) last;

	MPID_Segment_free(segp);

	/* ... end of body of routine ... */
	
  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_PACK);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_pack",
			L"**mpi_pack %p %d %D %p %d %p %C", inbuf, incount, datatype, outbuf, outcount, position, comm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#ifndef MIN
#define MIN(__a, __b) (((__a) < (__b)) ? (__a) : (__b))
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Get_elements
#define MPI_Get_elements PMPI_Get_elements

PMPI_LOCAL int MPIR_Type_get_basic_type_elements(int *bytes_p, int count, MPI_Datatype datatype);
PMPI_LOCAL int MPIR_Type_get_elements(int *bytes_p, int count, MPI_Datatype datatype);

/* NOTE: I think that this is in here so that we don't get two copies of it
 * in the case where we don't have weak symbols.
 */

/* MPIR_Type_get_basic_type_elements()
 *
 * Arguments:
 * - bytes_p - input/output byte count
 * - count - maximum number of this type to subtract from the bytes; a count
 *		   of -1 indicates use as many as we like
 * - datatype - input datatype
 *
 * Returns number of elements available given the two constraints of number of
 * bytes and count of types.  Also reduces the byte count by the amount taken
 * up by the types.
 *
 * Assumptions:
 * - the type passed to this function must be a basic *or* a pairtype
 *   (which aren't 
 * - the count is not zero (otherwise we can't tell between a "no more
 *   complete types" case and a "zero count" case)
 *
 * As per section 4.9.3 of the MPI 1.1 specification, the two-part reduction
 * types are to be treated as structs of the constituent types.  So we have to
 * do something special to handle them correctly in here.
 *
 * As per section 3.12.5 get_count and get_elements report the same value for
 * basic datatypes; I'm currently interpreting this to *not* include these
 * reduction types, as they are considered structs.
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Type_get_basic_type_elements
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
PMPI_LOCAL int MPIR_Type_get_basic_type_elements(int *bytes_p, int count, MPI_Datatype datatype)
{
	int elements, usable_bytes, used_bytes, type1_sz, type2_sz;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_TYPE_GET_BASIC_TYPE_ELEMENTS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_TYPE_GET_BASIC_TYPE_ELEMENTS);

	if (count == 0)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_BASIC_TYPE_ELEMENTS);
		return 0;
	}

	/* determine the maximum number of bytes we should take from the byte count. */
	if (count < 0)
	{
		usable_bytes = *bytes_p;
	}
	else
	{
		usable_bytes = MIN(*bytes_p, count * MPID_Datatype_get_basic_size(datatype));
	}

	switch (datatype)
	{
		/* we don't get valid fortran datatype handles in all cases... */
#ifdef HAVE_FORTRAN_BINDING
		case MPI_2REAL:
 			type1_sz = type2_sz = MPID_Datatype_get_basic_size(MPI_REAL);
			break;
		case MPI_2DOUBLE_PRECISION:
 			type1_sz = type2_sz = MPID_Datatype_get_basic_size(MPI_DOUBLE_PRECISION);
			break;
		case MPI_2INTEGER:
 			type1_sz = type2_sz = MPID_Datatype_get_basic_size(MPI_INTEGER);
			break;
#endif
		case MPI_2INT:
 			type1_sz = type2_sz = MPID_Datatype_get_basic_size(MPI_INT);
			break;
		case MPI_FLOAT_INT:
			type1_sz = MPID_Datatype_get_basic_size(MPI_FLOAT);
			type2_sz = MPID_Datatype_get_basic_size(MPI_INT);
			break;
		case MPI_DOUBLE_INT:
			type1_sz = MPID_Datatype_get_basic_size(MPI_DOUBLE);
			type2_sz = MPID_Datatype_get_basic_size(MPI_INT);
			break;
		case MPI_LONG_INT:
			type1_sz = MPID_Datatype_get_basic_size(MPI_LONG);
			type2_sz = MPID_Datatype_get_basic_size(MPI_INT);
			break;
		case MPI_SHORT_INT:
			type1_sz = MPID_Datatype_get_basic_size(MPI_SHORT);
			type2_sz = MPID_Datatype_get_basic_size(MPI_INT);
			break;
		case MPI_LONG_DOUBLE_INT:
			type1_sz = MPID_Datatype_get_basic_size(MPI_LONG_DOUBLE);
			type2_sz = MPID_Datatype_get_basic_size(MPI_INT);
			break;
		default:
			/* all other types.  this is more complicated than
			 * necessary for handling these types, but it puts us in the
			 * same code path for all the basics, so we stick with it.
			 */
			type1_sz = type2_sz = MPID_Datatype_get_basic_size(datatype);
			break;
	}

	/* determine the number of elements in the region */
	elements = 2 * (usable_bytes / (type1_sz + type2_sz));
	if (usable_bytes % (type1_sz + type2_sz) >= type1_sz) elements++;

	/* determine how many bytes we used up with those elements */
	used_bytes = ((elements / 2) * (type1_sz + type2_sz));
	if (elements % 2 == 1) used_bytes += type1_sz;

	*bytes_p -= used_bytes;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_BASIC_TYPE_ELEMENTS);
	return elements;
}


/* MPIR_Type_get_elements
 *
 * Arguments:
 * - bytes_p - input/output byte count
 * - count - maximum number of this type to subtract from the bytes; a count
 *		   of -1 indicates use as many as we like
 * - datatype - input datatype
 *
 * Returns number of elements available given the two constraints of number of
 * bytes and count of types.  Also reduces the byte count by the amount taken
 * up by the types.
 *
 * This is called from MPI_Get_elements() when it sees a type with multiple
 * element types (datatype_ptr->element_sz = -1).  This function calls itself too.
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Type_get_elements
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
PMPI_LOCAL int MPIR_Type_get_elements(int *bytes_p, int count, MPI_Datatype datatype)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
	
	MPID_Datatype_get_ptr(datatype, datatype_ptr); /* invalid if builtin */

	/* if we have gotten down to a type with only one element type,
	 * call MPIR_Type_get_basic_type_elements() and return.
	 */
	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN ||
		datatype == MPI_FLOAT_INT ||
		datatype == MPI_DOUBLE_INT ||
		datatype == MPI_LONG_INT ||
		datatype == MPI_SHORT_INT ||
		datatype == MPI_LONG_DOUBLE_INT)
	{
		mpi_errno = MPIR_Type_get_basic_type_elements(bytes_p, count, datatype);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
		return mpi_errno;
	}
	else if (datatype_ptr->element_size >= 0)
	{
		mpi_errno = MPIR_Type_get_basic_type_elements(bytes_p, count * datatype_ptr->n_elements, datatype_ptr->eltype);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
		return mpi_errno;
	}
	else
	{
		/* we have bytes left and still don't have a single element size; must recurse. */
		int i, j, typecount = 0, *ints, nr_elements = 0, last_nr_elements;
		MPI_Aint *aints;
		MPI_Datatype *types;

		/* Establish locations of arrays; perhaps this should be a fn. call or
		 * this fn. should be an MPID one?
		 */
		types = (MPI_Datatype *) (((char *) datatype_ptr->contents) + sizeof(MPID_Datatype_contents));
		ints  = (int *) (((char *) types) + datatype_ptr->contents->nr_types * sizeof(MPI_Datatype));
		aints = (MPI_Aint *) (((char *) ints) + datatype_ptr->contents->nr_ints * sizeof(int));

		switch (datatype_ptr->contents->combiner)
		{
			case MPI_COMBINER_NAMED:
			case MPI_COMBINER_DUP:
			case MPI_COMBINER_RESIZED:
				mpi_errno = MPIR_Type_get_elements(bytes_p, count, *types);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
				return mpi_errno;
				break;
			case MPI_COMBINER_CONTIGUOUS:
			case MPI_COMBINER_VECTOR:
			case MPI_COMBINER_HVECTOR_INTEGER:
			case MPI_COMBINER_HVECTOR:
				/* count is first in ints array */
				mpi_errno = MPIR_Type_get_elements(bytes_p, count * (*ints), *types);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
				return mpi_errno;
				break;
			case MPI_COMBINER_INDEXED_BLOCK:
				/* count is first in ints array, blocklength is second */
				mpi_errno = MPIR_Type_get_elements(bytes_p, count * ints[0] * ints[1], *types);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
				return mpi_errno;
				break;
			case MPI_COMBINER_INDEXED:
			case MPI_COMBINER_HINDEXED_INTEGER:
			case MPI_COMBINER_HINDEXED:
				for (i=0; i < (*ints); i++)
				{
					/* add up the blocklengths to get a max. # of the next type */
					typecount += ints[i+1];
				}
				mpi_errno = MPIR_Type_get_elements(bytes_p, count * typecount, *types);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
				return mpi_errno;
				break;
			case MPI_COMBINER_STRUCT_INTEGER:
			case MPI_COMBINER_STRUCT:
				/* In this case we can't simply multiply the count of the next
				 * type by the count of the current type, because we need to
				 * cycle through the types just as the struct would.  thus the
				 * nested loops.
				 *
				 * We need to keep going until we see a "0" elements returned
				 * or we run out of bytes.
				 */


				last_nr_elements = 1; /* seed value */
				for (j=0;
					 (count == -1 || j < count) &&
						 *bytes_p > 0 && last_nr_elements > 0;
					 j++)
				{
					/* recurse on each type; bytes are reduced in calls */
					for (i=0; i < (*ints); i++)
					{
						/* skip zero-count elements of the struct */
						if (ints[i+1] == 0) continue;

						last_nr_elements = MPIR_Type_get_elements(bytes_p, ints[i+1], types[i]);
						nr_elements += last_nr_elements;

						MPIU_Assert(last_nr_elements >= 0);

						if (last_nr_elements == 0) break;
					}
				}
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
				return nr_elements;
				break;
			case MPI_COMBINER_SUBARRAY:
			case MPI_COMBINER_DARRAY:
			case MPI_COMBINER_F90_REAL:
			case MPI_COMBINER_F90_COMPLEX:
			case MPI_COMBINER_F90_INTEGER:
			default:
				/* --BEGIN ERROR HANDLING-- */
				MPIU_Assert(0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
				return -1;
				break;
				/* --END ERROR HANDLING-- */
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_ELEMENTS);
}
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Get_elements

/*@
   MPI_Get_elements - get_elements

   Arguments:
+  MPI_Status *status - status
.  MPI_Datatype datatype - datatype
-  int *elements - elements

   Notes:

 If the size of the datatype is zero and the amount of data returned as
 determined by 'status' is also zero, this routine will return a count of
 zero.  This is consistent with a clarification made by the MPI Forum.

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Get_elements(MPI_Status *status, MPI_Datatype datatype, int *elements)
{
#undef FCNAME
	static const wchar_t FCNAME[] = L"MPI_Get_elements";
	int mpi_errno = MPI_SUCCESS, byte_count;
	MPID_Datatype *datatype_ptr = NULL;

	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GET_ELEMENTS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GET_ELEMENTS);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr(datatype, datatype_ptr);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(status, L"status", mpi_errno);
			MPIR_ERRTEST_ARGNULL(elements, L"elements", mpi_errno);
			/* Validate datatype_ptr */
			if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				if (mpi_errno == MPI_SUCCESS)
				{
					MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
				}
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_get_elements %p %D %p", status, datatype, elements); }

	/* three cases:
	 * - nice, simple, single element type
	 * - derived type with a zero size
	 * - type with multiple element types (nastiest)
	 */
	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN ||
		(datatype_ptr->element_size != -1 && datatype_ptr->size > 0))
	{
		byte_count = status->count;

		/* QUESTION: WHAT IF SOMEONE GAVE US AN MPI_UB OR MPI_LB???
		 */

		/* in both cases we do not limit the number of types that might
		 * be in bytes
		 */
		if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
		{
			*elements = MPIR_Type_get_basic_type_elements(&byte_count, -1, datatype_ptr->eltype);
		}
		else
		{
			*elements = MPIR_Type_get_basic_type_elements(&byte_count, -1, datatype);
		}
		MPIU_Assert(byte_count >= 0);
	}
	else if (datatype_ptr->size == 0)
	{
		if (status->count > 0)
		{
			/* --BEGIN ERROR HANDLING-- */

			/* datatype size of zero and count > 0 should never happen. */

			(*elements) = MPI_UNDEFINED;
			/* --END ERROR HANDLING-- */
		}
		else
		{
			/* This is ambiguous.  However, discussions on MPI Forum
			 * reached a consensus that this is the correct return 
			 * value
			 */
			(*elements) = 0;
		}
	}
	else /* derived type with weird element type or weird size */
	{
		MPIU_Assert(datatype_ptr->element_size == -1);

		byte_count = status->count;
		*elements = MPIR_Type_get_elements(&byte_count, -1, datatype);
	}

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GET_ELEMENTS);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_get_elements",
			L"**mpi_get_elements %p %D %p", status, datatype, elements);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Get_count
#define MPI_Get_count PMPI_Get_count
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Get_count

/*@
  MPI_Get_count - Gets the number of "top level" elements

Input Parameters:
+ status - return status of receive operation (Status) 
- datatype - datatype of each receive buffer element (handle) 

Output Parameter:
. count - number of received elements (integer) 
Notes:
If the size of the datatype is zero, this routine will return a count of
zero.  If the amount of data in 'status' is not an exact multiple of the 
size of 'datatype' (so that 'count' would not be integral), a 'count' of
'MPI_UNDEFINED' is returned instead.

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Get_count( MPI_Status *status, MPI_Datatype datatype, int *count )
{
	static const wchar_t FCNAME[] = L"MPI_Get_count";
	int mpi_errno = MPI_SUCCESS;
	int size;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GET_COUNT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GET_COUNT);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_ARGNULL(status, L"status", mpi_errno);
			MPIR_ERRTEST_ARGNULL(count, L"count", mpi_errno);
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;

			/* Validate datatype_ptr */
			if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				/* Q: Must the type be committed to be used with this function? */
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_get_count %p %D %p", status, datatype, count); }
	
	/* Check for correct number of bytes */
	MPID_Datatype_get_size_macro(datatype, size);
	if (size != 0)
	{
		if ((status->count % size) != 0)
			(*count) = MPI_UNDEFINED;
		else
			(*count) = status->count / size;
	}
	else
	{
		if (status->count > 0)
		{
			/* --BEGIN ERROR HANDLING-- */

			/* case where datatype size is 0 and count is > 0 should
			 * never occur.
			 */

			(*count) = MPI_UNDEFINED;
			/* --END ERROR HANDLING-- */
		}
		else
		{
			/* This is ambiguous.  However, discussions on MPI Forum
			   reached a consensus that this is the correct return 
			   value
			*/
			(*count) = 0;
		}
	}
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GET_COUNT);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_get_count",
			L"**mpi_get_count %p %D %p", status, datatype, count);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Get_address
#define MPI_Get_address PMPI_Get_address

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Get_address

/*@
   MPI_Get_address - Get the address of a location in memory 

Input Parameter:
. location - location in caller memory (choice) 

Output Parameter:
. address - address of location (address) 

   Notes:
	This routine is provided for both the Fortran and C programmers.
	On many systems, the address returned by this routine will be the same
	as produced by the C '&' operator, but this is not required in C and
	may not be true of systems with word- rather than byte-oriented 
	instructions or systems with segmented address spaces.  

	This routine should be used instead of 'MPI_Address'. 

.N SignalSafe

.N Fortran

 In Fortran, the integer type is always signed.  This can cause problems
 on systems where the address fits into a four byte unsigned integer but
 the value is larger than the largest signed integer.  For example, a system
 with more than 2 GBytes of memory may have addresses that do not fit within
 a four byte signed integer.  Unfortunately, there is no easy solution to 
 this problem, as there is no Fortran datatype that can be used here (using
 a longer integer type will cause other problems, as well as surprising
 users when the size of the integer type is larger that the size of a pointer
 in C).  In this case, it is recommended that you use C to manipulate 
 addresses.

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Get_address(void *location, MPI_Aint *address)
{
	static const wchar_t FCNAME[] = L"MPI_Get_address";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_GET_ADDRESS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_GET_ADDRESS);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(address,L"address",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_get_address %p %p", location, address); }
	
	/* SX_4 needs to set CHAR_PTR_IS_ADDRESS 
	   The reason is that it computes the different in two pointers in
	   an "int", and addresses typically have the high (bit 31) bit set;
	   thus the difference, when cast as MPI_Aint (long), is sign-extended, 
	   making the absolute address negative.  Without a copy of the C 
	   standard, I can't tell if this is a compiler bug or a language bug.
	*/
#ifdef CHAR_PTR_IS_ADDRESS
	*address = (MPI_Aint) ((char *)location);
#else
	/* Note that this is the "portable" way to generate an address.
	   The difference of two pointers is the number of elements
	   between them, so this gives the number of chars between location
	   and ptr.  As long as sizeof(char) represents one byte, 
	   of bytes from 0 to location */
	*address = (MPI_Aint) ((char *)location - (char *)MPI_BOTTOM);
#endif
	/* The same code is used in MPI_Address */
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_GET_ADDRESS);
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_get_address",
			L"**mpi_get_address %p %p", location, address);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Address
#define MPI_Address PMPI_Address

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Address

/*@
	MPI_Address - Gets the address of a location in memory  

Input Parameter:
. location - location in caller memory (choice) 

Output Parameter:
. address - address of location (integer) 

	Note:
	This routine is provided for both the Fortran and C programmers.
	On many systems, the address returned by this routine will be the same
	as produced by the C '&' operator, but this is not required in C and
	may not be true of systems with word- rather than byte-oriented 
	instructions or systems with segmented address spaces.  

.N SignalSafe
 
.N Deprecated
The replacement for this routine is 'MPI_Get_address'.

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Address( void *location, MPI_Aint *address )
{
	static const wchar_t FCNAME[] = L"MPI_Address";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_ADDRESS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_ADDRESS);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(address,L"address",mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_address %p %p", location, address); }
	
	/* SX_4 needs to set CHAR_PTR_IS_ADDRESS 
	   The reason is that it computes the different in two pointers in
	   an "int", and addresses typically have the high (bit 31) bit set;
	   thus the difference, when cast as MPI_Aint (long), is sign-extended, 
	   making the absolute address negative.  Without a copy of the C 
	   standard, I can't tell if this is a compiler bug or a language bug.
	*/
#ifdef CHAR_PTR_IS_ADDRESS
	*address = (MPI_Aint) ((char *)location);
#else
	/* Note that this is the "portable" way to generate an address.
	   The difference of two pointers is the number of elements
	   between them, so this gives the number of chars between location
	   and ptr.  As long as sizeof(char) represents one byte, 
	   of bytes from 0 to location */
	*address = (MPI_Aint) ((char *)location - (char *)MPI_BOTTOM);
#endif
	/* The same code is used in MPI_Get_address */
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_ADDRESS);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_address",
			L"**mpi_address %p %p", location, address);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Unpack_externalA
#undef MPI_Unpack_externalW
#define MPI_Unpack_externalA PMPI_Unpack_externalA
#define MPI_Unpack_externalW PMPI_Unpack_externalW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Unpack_external

/*@
   MPI_Unpack_external - Unpack a buffer (packed with MPI_Pack_external) 
   according to a datatype into contiguous memory

   Input Parameters:
+ datarep - data representation (string)  
. inbuf - input buffer start (choice)  
. insize - input buffer size, in bytes (integer)  
. outcount - number of output data items (integer)  
. datatype - datatype of output data item (handle)  

   Input/Output Parameter:
. position - current position in buffer, in bytes (integer)  

   Output Parameter:
. outbuf - output buffer start (choice)  

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Unpack_externalA(char *datarep, void *inbuf, MPI_Aint insize, MPI_Aint *position, void *outbuf, int outcount, MPI_Datatype datatype)
{
	static const wchar_t FCNAME[] = L"MPI_Unpack_externalA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t *tmp_datarep = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_UNPACK_EXTERNALA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_UNPACK_EXTERNALA);

	if (datarep != NULL)
	{
		mpi_errno = MPIU_Mbstowcs(datarep, &tmp_datarep);
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = MPI_Unpack_externalW(tmp_datarep, inbuf, insize, position, outbuf, outcount, datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		/* Invalid datarep string */
		MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
	}

fn_fail:
	if (tmp_datarep != NULL)
	{
		MPIU_Free(tmp_datarep);
	}
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_UNPACK_EXTERNALA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Unpack_externalW(wchar_t *datarep, void *inbuf, MPI_Aint insize, MPI_Aint *position, void *outbuf, int outcount, MPI_Datatype datatype)
{
	static const wchar_t FCNAME[] = L"MPI_Unpack_external";
	int mpi_errno = MPI_SUCCESS;
	MPI_Aint first, last;
	MPID_Segment *segp;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_UNPACK_EXTERNAL);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_UNPACK_EXTERNAL);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(inbuf, L"input buffer", mpi_errno);
			/* NOTE: outbuf could be MPI_BOTTOM; don't test for NULL */
			MPIR_ERRTEST_COUNT(insize, mpi_errno);
			MPIR_ERRTEST_COUNT(outcount, mpi_errno);

			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			
			if (datatype != MPI_DATATYPE_NULL && HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype *datatype_ptr = NULL;

				MPID_Datatype_get_ptr(datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}
				
			/* If datatye_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_unpack_external %s %p %d %p %p %d %D", datarep, inbuf, insize, position, outbuf, outcount, datatype); }
	
	segp = MPID_Segment_alloc();
	MPIU_ERR_CHKANDJUMP1((segp == NULL), mpi_errno, MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPID_Segment_alloc");
	mpi_errno = MPID_Segment_init(outbuf, outcount, datatype, segp, 1);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* NOTE: buffer values and positions in MPI_Unpack_external are used very
	 * differently from use in MPID_Segment_unpack_external...
	 */
	first = 0;
	last  = SEGMENT_IGNORE_LAST;

	MPID_Segment_unpack_external32(segp, first, &last, (void *) ((char *) inbuf + *position));

	*position += (int) last;

	MPID_Segment_free(segp);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_UNPACK_EXTERNAL);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_unpack_external",
			L"**mpi_unpack_external %s %p %d %p %p %d %D", datarep, inbuf, insize, position, outbuf, outcount, datatype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Unpack
#define MPI_Unpack PMPI_Unpack

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Unpack

/*@
	MPI_Unpack - Unpack a buffer according to a datatype into contiguous memory

Input Parameters:
+ inbuf - input buffer start (choice) 
. insize - size of input buffer, in bytes (integer) 
. position - current position in bytes (integer) 
. outcount - number of items to be unpacked (integer) 
. datatype - datatype of each output data item (handle) 
- comm - communicator for packed message (handle) 

Output Parameter:
. outbuf - output buffer start (choice) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_ARG

.seealso: MPI_Pack, MPI_Pack_size
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Unpack(void *inbuf, int insize, int *position, void *outbuf, int outcount, MPI_Datatype datatype, MPI_Comm comm)
{
	static const wchar_t FCNAME[] = L"MPI_Unpack";
	int mpi_errno = MPI_SUCCESS;
	MPI_Aint first, last;
	MPID_Comm *comm_ptr = NULL;
	MPID_Segment *segp;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_UNPACK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_UNPACK);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_COMM(comm, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Comm_get_ptr(comm, comm_ptr);
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(inbuf, L"input buffer", mpi_errno);
			/* Note: outbuf could be MPI_BOTTOM; don't test for NULL */
			MPIR_ERRTEST_COUNT(insize, mpi_errno);
			MPIR_ERRTEST_COUNT(outcount, mpi_errno);

			/* Validate comm_ptr */
			MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */

			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			
			if (datatype != MPI_DATATYPE_NULL &&HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype *datatype_ptr = NULL;

				MPID_Datatype_get_ptr(datatype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				MPID_Datatype_committed_ptr(datatype_ptr, mpi_errno);
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_unpack %p %d %p %p %d %D %C", inbuf, insize, position, outbuf, outcount, datatype, comm); }
	
	segp = MPID_Segment_alloc();
	MPIU_ERR_CHKANDJUMP1((segp == NULL), mpi_errno, MPI_ERR_OTHER, L"**nomem", L"**nomem %s", L"MPID_Segment_alloc");
	MPID_Segment_init(outbuf, outcount, datatype, segp, 0);

	/* NOTE: the use of buffer values and positions in MPI_Unpack and in
	 * MPID_Segment_unpack are quite different.  See code or docs or something.
	 */
	first = 0;
	last  = SEGMENT_IGNORE_LAST;

	MPID_Segment_unpack(segp, first, &last, (void *) ((char *) inbuf + *position));

	*position += (int) last;

	MPID_Segment_free(segp);

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_UNPACK);
	return mpi_errno;


  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_unpack", 
			L"**mpi_unpack %p %d %p %p %d %D %C", inbuf, insize, position, outbuf, outcount, datatype, comm);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_vector
#define MPI_Type_vector PMPI_Type_vector

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_vector

/*@
	MPI_Type_vector - Creates a vector (strided) datatype

Input Parameters:
+ count - number of blocks (nonnegative integer) 
. blocklength - number of elements in each block 
  (nonnegative integer)
. stride - number of elements between start of each block (integer) 
- oldtype - old datatype (handle) 

Output Parameter:
. newtype_p - new datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_vector(int count, int blocklength, int stride, MPI_Datatype old_type, MPI_Datatype *newtype_p)
{
	static const wchar_t FCNAME[] = L"MPI_Type_vector";
	int mpi_errno = MPI_SUCCESS;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_VECTOR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_VECTOR);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Datatype *old_ptr = NULL;

			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_ARGNEG(blocklength, L"blocklen", mpi_errno);
			MPIR_ERRTEST_DATATYPE(old_type, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			
			if (old_type != MPI_DATATYPE_NULL && HANDLE_GET_KIND(old_type) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(old_type, old_ptr);
				MPID_Datatype_valid_ptr(old_ptr, mpi_errno);
			}
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_vector %d %d %d %D %p", count, blocklength, stride, old_type, newtype_p); }
	
	mpi_errno = MPID_Type_vector(count, blocklength, (MPI_Aint) stride, 0 /* stride not in bytes, in extents */, old_type, newtype_p);

	if (mpi_errno == MPI_SUCCESS)
	{
		MPID_Datatype *new_dtp;
		int ints[3];

		ints[0] = count;
		ints[1] = blocklength;
		ints[2] = stride;
		MPID_Datatype_get_ptr(*newtype_p, new_dtp);
		mpi_errno = MPID_Datatype_set_contents(new_dtp,
											   MPI_COMBINER_VECTOR,
											   3, /* ints (cnt, blklen, str) */
											   0, /* aints */
											   1, /* types */
											   ints,
											   NULL,
											   &old_type);
	}

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_VECTOR);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_vector",
			L"**mpi_type_vector %d %d %d %D %p", count, blocklength, stride, old_type, newtype_p);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_ub
#define MPI_Type_ub PMPI_Type_ub

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_ub

/*@
	MPI_Type_ub - Returns the upper bound of a datatype

Input Parameters:
. datatype - datatype (handle) 

Output Parameter:
. displacement - displacement of upper bound from origin, 
							 in bytes (integer) 

.N Deprecated
The replacement for this routine is 'MPI_Type_get_extent'

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_ub(MPI_Datatype datatype, MPI_Aint *displacement)
{
	static const wchar_t FCNAME[] = L"MPI_Type_ub";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_UB);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_UB);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_ub %D %p", datatype, displacement); }
	
	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN)
		*displacement = MPID_Datatype_get_basic_size(datatype);
	else
		*displacement = datatype_ptr->ub;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_UB);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_ub",
			L"**mpi_type_ub %D %p", datatype, displacement);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_struct
#define MPI_Type_struct PMPI_Type_struct

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_struct

/*@
	MPI_Type_struct - Creates a struct datatype

Input Parameters:
+ count - number of blocks (integer) -- also number of 
entries in arrays array_of_types ,
array_of_displacements  and array_of_blocklengths  
. blocklens - number of elements in each block (array)
. indices - byte displacement of each block (array)
- old_types - type of elements in each block (array 
of handles to datatype objects) 

Output Parameter:
. newtype - new datatype (handle) 

.N Deprecated
The replacement for this routine is 'MPI_Type_create_struct'

Notes:
If an upperbound is set explicitly by using the MPI datatype 'MPI_UB', the
corresponding index must be positive.

The MPI standard originally made vague statements about padding and alignment;
this was intended to allow the simple definition of structures that could
be sent with a count greater than one.  For example,
.vb
	struct { int a; char b; } foo;
.ve
may have 'sizeof(foo) > sizeof(int) + sizeof(char)'; for example, 
'sizeof(foo) == 2*sizeof(int)'.  The initial version of the MPI standard
defined the extent of a datatype as including an `epsilon` that would have 
allowed an implementation to make the extent an MPI datatype
for this structure equal to '2*sizeof(int)'.  
However, since different systems might define different paddings, there was 
much discussion by the MPI Forum about what was the correct value of
epsilon, and one suggestion was to define epsilon as zero.
This would have been the best thing to do in MPI 1.0, particularly since 
the 'MPI_UB' type allows the user to easily set the end of the structure.
Unfortunately, this change did not make it into the final document.  
Currently, this routine does not add any padding, since the amount of 
padding needed is determined by the compiler that the user is using to
build their code, not the compiler used to construct the MPI library.
A later version of MPICH may provide for some natural choices of padding
(e.g., multiple of the size of the largest basic member), but users are
advised to never depend on this, even with vendor MPI implementations.
Instead, if you define a structure datatype and wish to send or receive
multiple items, you should explicitly include an 'MPI_UB' entry as the
last member of the structure.  For example, the following code can be used
for the structure foo
.vb
	blen[0] = 1; indices[0] = 0; oldtypes[0] = MPI_INT;
	blen[1] = 1; indices[1] = &foo.b - &foo; oldtypes[1] = MPI_CHAR;
	blen[2] = 1; indices[2] = sizeof(foo); oldtypes[2] = MPI_UB;
	MPI_Type_struct( 3, blen, indices, oldtypes, &newtype );
.ve

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_COUNT
.N MPI_ERR_EXHAUSTED
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_struct(int count,
					int blocklens[],
					MPI_Aint indices[],
					MPI_Datatype old_types[],
					MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_struct";
	int mpi_errno = MPI_SUCCESS;
	MPIU_CHKLMEM_DECL(1);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_STRUCT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_STRUCT);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			int i;
			MPID_Datatype *datatype_ptr;

			MPIR_ERRTEST_COUNT(count,mpi_errno);
			if (count > 0)
			{
				MPIR_ERRTEST_ARGNULL(blocklens, L"blocklens", mpi_errno);
				MPIR_ERRTEST_ARGNULL(indices, L"indices", mpi_errno);
				MPIR_ERRTEST_ARGNULL(old_types, L"types", mpi_errno);
			}
			if (mpi_errno == MPI_SUCCESS)
			{
				for (i=0; i < count; i++)
				{
					MPIR_ERRTEST_ARGNEG(blocklens[i], L"blocklen", mpi_errno);
					MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"blocklens", i, blocklens[i]);
					MPIR_ERRTEST_DATATYPE(old_types[i], L"datatype[i]", mpi_errno);
					/* stop before we dereference the null type */
					MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"old_types", i, old_types[i]);

					if (old_types[i] != MPI_DATATYPE_NULL && HANDLE_GET_KIND(old_types[i]) != HANDLE_KIND_BUILTIN)
					{
						MPID_Datatype_get_ptr(old_types[i], datatype_ptr);
						MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
					}
				}
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_struct %d %p %p %p %p", count, blocklens, indices, old_types, newtype); }
	
	mpi_errno = MPID_Type_struct(count, blocklens, indices, old_types, newtype);

	if (mpi_errno == MPI_SUCCESS)
	{
		int i, *ints;
		MPID_Datatype *new_dtp;

		MPIU_CHKLMEM_MALLOC(ints, int *, (count + 1) * sizeof(int), mpi_errno, L"contents integer array");

		ints[0] = count;
		for (i=0; i < count; i++)
		{
			ints[i+1] = blocklens[i];
		}

		MPID_Datatype_get_ptr(*newtype, new_dtp);
		mpi_errno = MPID_Datatype_set_contents(new_dtp,
											   MPI_COMBINER_STRUCT,
											   count+1, /* ints (count, blocklengths) */
											   count, /* aints (displacements) */
											   count, /* types */
											   ints,
											   indices,
											   old_types);

	}
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_STRUCT);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_struct",
			L"**mpi_type_struct %d %p %p %p %p", count, blocklens, indices, old_types, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_size
#define MPI_Type_size PMPI_Type_size

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_size

/*@
	MPI_Type_size - Return the number of bytes occupied by entries
					in the datatype

Input Parameters:
. datatype - datatype (handle) 

Output Parameter:
. size - datatype size (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_size(MPI_Datatype datatype, int *size)
{
	static const wchar_t FCNAME[] = L"MPI_Type_size";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_SIZE);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */
			
	/* If this is a built-in datatype, then get the size out of the handle */
	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN)
	{
		MPID_Datatype_get_size_macro(datatype, *size);
		goto fn_exit;
	}

	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_size %D %p", datatype, size); }

	MPID_Datatype_get_size_macro(datatype, *size);

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_SIZE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_size", 
			L"**mpi_type_size %D %p", datatype, size);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_set_nameA
#undef MPI_Type_set_nameW
#define MPI_Type_set_nameA PMPI_Type_set_nameA
#define MPI_Type_set_nameW PMPI_Type_set_nameW

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_set_name

/*@
   MPI_Type_set_name - set datatype name

   Input Parameters:
+ type - datatype whose identifier is to be set (handle) 
- type_name - the character string which is remembered as the name (string) 

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_TYPE
.N MPI_ERR_OTHER
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_set_nameA(MPI_Datatype type, char *type_name)
{
	static const wchar_t FCNAME[] = L"MPI_Type_set_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t name[MPI_MAX_OBJECT_NAME];
	wchar_t *name_ptr = L"";
	size_t len = 0;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_SET_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_SET_NAMEA);

	/* convert the ASCII string to a wide character string and call the wide character version */

	if (type_name != NULL)
	{
		error = mbstowcs_s(&len, name, MPI_MAX_OBJECT_NAME, type_name, _TRUNCATE);
		if (error != 0)
		{
			if (len >= MPI_MAX_OBJECT_NAME)
			{
				/* truncation error */
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"type_name", (int)len);
				goto fn_fail;
			}
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
			goto fn_fail;
		}
		if (len >= MPI_MAX_OBJECT_NAME)
		{
			/* truncation error */
			type_name[MPI_MAX_OBJECT_NAME-1] = '\0';
			MPIU_ERR_SET2(mpi_errno, MPI_ERR_OTHER, L"**conversion_truncation", L"**conversion_truncation %s %d", L"type_name", (int)len);
			goto fn_fail;
		}
		if (len > 0 && len < MPI_MAX_OBJECT_NAME)
			name_ptr = name;
	}
	else
	{
		name_ptr = NULL;
	}

	mpi_errno = MPI_Type_set_nameW(type, name_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_SET_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Type_set_nameW(MPI_Datatype type, wchar_t *type_name)
{
	static const wchar_t FCNAME[] = L"MPI_Type_set_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	static int setup = 0;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_SET_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_SET_NAME);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(type, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( type, datatype_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			int slen;
			
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			/* If datatype_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(type_name,L"type_name", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;

			slen = (int)wcslen( type_name );
			MPIU_ERR_CHKANDSTMT1((slen >= MPI_MAX_OBJECT_NAME), mpi_errno, MPI_ERR_ARG,;, L"**typenamelen", L"**typenamelen %d", slen );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_set_name %D %s", type, type_name); }
	
	/* If this is the first call, initialize all of the predefined names.
	   Note that type_get_name must also make the same call */
	if (!setup)
	{ 
		MPIR_Datatype_init_names();
		setup = 1;
	}

	/* Include the null in MPI_MAX_OBJECT_NAME */
	/*MPIU_Wcsncpy( datatype_ptr->name, type_name, MPI_MAX_OBJECT_NAME );*/
	MPIU_Wcsncpy( datatype_ptr->name, type_name, MPI_MAX_OBJECT_NAME );
	
	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_SET_NAME);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_set_name",
			L"**mpi_type_set_name %D %s", type, type_name);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_match_size
#define MPI_Type_match_size PMPI_Type_match_size

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_match_size

/*@
   MPI_Type_match_size - Find an MPI datatype matching a specified size

   Input Parameters:
+ typeclass - generic type specifier (integer) 
- size - size, in bytes, of representation (integer) 

   Output Parameter:
. type - datatype with correct type, size (handle) 

Notes:
'typeclass' is one of 'MPI_TYPECLASS_REAL', 'MPI_TYPECLASS_INTEGER' and 
'MPI_TYPECLASS_COMPLEX', corresponding to the desired typeclass. 
The function returns an MPI datatype matching a local variable of type 
'( typeclass, size )'. 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_match_size(int typeclass, int size, MPI_Datatype *datatype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_match_size";
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t *tname = 0;
	static MPI_Datatype real_types[] = { MPI_FLOAT, MPI_DOUBLE
#ifdef HAVE_LONG_DOUBLE
										 ,MPI_LONG_DOUBLE
#endif
	};
	static MPI_Datatype int_types[] = { MPI_CHAR, MPI_SHORT, MPI_INT, 
										MPI_LONG
#ifdef HAVE_LONG_LONG_INT
										, MPI_LONG_LONG
#endif
	};
	static MPI_Datatype complex_types[] = { MPI_COMPLEX, MPI_DOUBLE_COMPLEX };
	MPI_Datatype matched_datatype = MPI_DATATYPE_NULL;
	int i, tsize;
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_MATCH_SIZE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_MATCH_SIZE);

	MPIU_THREADPRIV_GET;
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL( datatype, L"datatype", mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_match_size %d %d %p",typeclass, size, datatype); }
	
	/* FIXME: Should make use of Fortran optional types (e.g., MPI_INTEGER2) */

	/* The following implementation follows the suggestion in the
	   MPI-2 standard.  
	   The version in the MPI-2 spec makes use of the Fortran optional types;
	   currently, we don't support these from C (see mpi.h.in).  
	   Thus, we look at the candidate types and make use of the first fit.
	*/
	switch (typeclass)
	{
	case MPI_TYPECLASS_REAL:
		tname = L"MPI_TYPECLASS_REAL";
		for (i=0; i<sizeof(real_types)/sizeof(MPI_Datatype); i++)
		{
			if (real_types[i] == MPI_DATATYPE_NULL) { continue; }
			NMPI_Type_size(mpi_errno, ( real_types[i], &tsize ));
			if (tsize == size)
			{
				matched_datatype = real_types[i];
				break;
			}
		}
		break;
	case MPI_TYPECLASS_INTEGER:
		tname = L"MPI_TYPECLASS_INTEGER";
		for (i=0; i<sizeof(int_types)/sizeof(MPI_Datatype); i++)
		{
			if (int_types[i] == MPI_DATATYPE_NULL) { continue; }
			NMPI_Type_size(mpi_errno, ( int_types[i], &tsize ));
			if (tsize == size)
			{
				matched_datatype = int_types[i];
				break;
			}
		}
		break;
	case MPI_TYPECLASS_COMPLEX:
		tname = L"MPI_TYPECLASS_COMPLEX";
		for (i=0; i<sizeof(complex_types)/sizeof(MPI_Datatype); i++)
		{
			if (complex_types[i] == MPI_DATATYPE_NULL) { continue; }
			NMPI_Type_size(mpi_errno, ( complex_types[i], &tsize ));
			if (tsize == size)
			{
				matched_datatype = complex_types[i];
				break;
			}
		}
		break;
	default:
		/* --BEGIN ERROR HANDLING-- */
		MPIU_ERR_SETANDSTMT(mpi_errno, MPI_ERR_ARG, break, L"**typematchnoclass");
		/* --END ERROR HANDLING-- */
	}

	if (mpi_errno == MPI_SUCCESS)
	{
		if (matched_datatype == MPI_DATATYPE_NULL)
		{
			/* --BEGIN ERROR HANDLING-- */
			MPIU_ERR_SETANDSTMT2(mpi_errno, MPI_ERR_ARG,;, L"**typematchsize", L"**typematchsize %s %d", tname, size);
			/* --END ERROR HANDLING-- */
		}
		else
		{
			*datatype = matched_datatype;
		}
	}
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_MATCH_SIZE);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_match_size",
			L"**mpi_type_match_size %d %d %p",typeclass, size, datatype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_lb
#define MPI_Type_lb PMPI_Type_lb

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_lb

/*@
	MPI_Type_lb - Returns the lower-bound of a datatype

Input Parameters:
. datatype - datatype (handle) 

Output Parameter:
. displacement - displacement of lower bound from origin, 
							 in bytes (integer) 

.N Deprecated
The replacement for this routine is 'MPI_Type_Get_extent'.

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_lb(MPI_Datatype datatype, MPI_Aint *displacement)
{
	static const wchar_t FCNAME[] = L"MPI_Type_lb";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_LB);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_LB);

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_lb %D %p", datatype, displacement); }
	
	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN)
	{
		*displacement = 0;
	}
	else
	{
		*displacement = datatype_ptr->lb;
	}

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_LB);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_lb",
			L"**mpi_type_lb %D %p", datatype, displacement);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_indexed
#define MPI_Type_indexed PMPI_Type_indexed

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_indexed

/*@
	MPI_Type_indexed - Creates an indexed datatype

Input Parameters:
+ count - number of blocks -- also number of entries in indices and blocklens
. blocklens - number of elements in each block (array of nonnegative integers) 
. indices - displacement of each block in multiples of old_type (array of 
  integers)
- old_type - old datatype (handle) 

Output Parameter:
. newtype - new datatype (handle) 

.N ThreadSafe

.N Fortran

The indices are displacements, and are based on a zero origin.  A common error
is to do something like to following
.vb
	integer a(100)
	integer blens(10), indices(10)
	do i=1,10
		 blens(i)   = 1
10	   indices(i) = 1 + (i-1)*10
	call MPI_TYPE_INDEXED(10,blens,indices,MPI_INTEGER,newtype,ierr)
	call MPI_TYPE_COMMIT(newtype,ierr)
	call MPI_SEND(a,1,newtype,...)
.ve
expecting this to send 'a(1),a(11),...' because the indices have values 
'1,11,...'.   Because these are `displacements` from the beginning of 'a',
it actually sends 'a(1+1),a(1+11),...'.

If you wish to consider the displacements as indices into a Fortran array,
consider declaring the Fortran array with a zero origin
.vb
	integer a(0:99)
.ve

.N Errors
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
.N MPI_ERR_EXHAUSTED
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_indexed(int count, int blocklens[], int indices[], MPI_Datatype old_type, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_indexed";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *new_dtp;
	int i, *ints;
	MPIU_CHKLMEM_DECL(1);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_INDEXED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_INDEXED);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			int i;
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_COUNT(count,mpi_errno);
			if (count > 0)
			{
				MPIR_ERRTEST_ARGNULL(blocklens, L"blocklens", mpi_errno);
				MPIR_ERRTEST_ARGNULL(indices, L"indices", mpi_errno);
			}
			MPIR_ERRTEST_DATATYPE(old_type, L"datatype", mpi_errno);
			if (mpi_errno == MPI_SUCCESS)
			{
 				if (HANDLE_GET_KIND(old_type) != HANDLE_KIND_BUILTIN)
				{
					MPID_Datatype_get_ptr( old_type, datatype_ptr );
					MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
				}
				/* verify that all blocklengths are >= 0 */
				for (i=0; i < count; i++)
				{
					MPIR_ERRTEST_ARGNEG(blocklens[i], L"blocklen", mpi_errno);
					MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"blocklens", i, blocklens[i]);
				}
			}
			MPIR_ERRTEST_ARGNULL(newtype, L"newtype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_indexed %d %p %p %D %p", count,blocklens, indices, old_type, newtype); }
	
	mpi_errno = MPID_Type_indexed(count, blocklens, indices, 0, old_type, newtype);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* copy all integer values into a temporary buffer; this
	 * includes the count, the blocklengths, and the displacements.
	 */
	MPIU_CHKLMEM_MALLOC(ints, int *, (2 * count + 1) * sizeof(int), mpi_errno, L"contents integer array");

	ints[0] = count;

	for (i=0; i < count; i++)
	{
		ints[i+1] = blocklens[i];
	}
	for (i=0; i < count; i++)
	{
		ints[i + count + 1] = indices[i];
	}
	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_INDEXED,
										   2*count + 1, /* ints */
										   0, /* aints  */
										   1, /* types */
										   ints,
										   NULL,
										   &old_type);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_INDEXED);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_indexed",
			L"**mpi_type_indexed %d %p %p %D %p", count,blocklens, indices, old_type, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_hvector
#define MPI_Type_hvector PMPI_Type_hvector

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_hvector

/*@
   MPI_Type_hvector - type_hvector

   Arguments:
+  int count - count
.  int blocklen - blocklen
.  MPI_Aint stride - stride
.  MPI_Datatype old_type - old datatype
-  MPI_Datatype *newtype - new datatype

   Notes:

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_hvector(int count, int blocklen, MPI_Aint stride, MPI_Datatype old_type, MPI_Datatype *newtype_p)
{
	static const wchar_t FCNAME[] = L"MPI_Type_hvector";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *new_dtp;
	int ints[2];
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_HVECTOR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_HVECTOR);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_ARGNEG(blocklen,L"blocklen",mpi_errno);
			MPIR_ERRTEST_DATATYPE(old_type, L"datatype", mpi_errno);
			if (mpi_errno == MPI_SUCCESS)
			{
				if (HANDLE_GET_KIND(old_type) != HANDLE_KIND_BUILTIN)
				{
					MPID_Datatype_get_ptr(old_type, datatype_ptr);
					MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				}
			}
			MPIR_ERRTEST_ARGNULL(newtype_p, L"newtype", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_hvector %d %d %d %D %p", count, blocklen, stride, old_type, newtype_p); }
	
	mpi_errno = MPID_Type_vector(count, blocklen, (MPI_Aint) stride, 1 /* stride in bytes */, old_type, newtype_p);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	ints[0] = count;
	ints[1] = blocklen;
	MPID_Datatype_get_ptr(*newtype_p, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_HVECTOR,
											   2, /* ints (count, blocklen) */
										   1, /* aints */
										   1, /* types */
										   ints,
										   &stride,
										   &old_type);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_HVECTOR);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_hvector",
			L"**mpi_type_hvector %d %d %d %D %p", count, blocklen, stride, old_type, newtype_p);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_hindexed
#define MPI_Type_hindexed PMPI_Type_hindexed

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_hindexed

/*@
	MPI_Type_hindexed - Creates an indexed datatype with offsets in bytes

Input Parameters:
+ count - number of blocks -- also number of entries in indices and blocklens
. blocklens - number of elements in each block (array of nonnegative integers) 
. indices - byte displacement of each block (array of MPI_Aint) 
- old_type - old datatype (handle) 

Output Parameter:
. newtype - new datatype (handle) 

.N Deprecated
This routine is replaced by 'MPI_Type_create_hindexed'.

.N ThreadSafe

.N Fortran

The indices are displacements, and are based on a zero origin.  A common error
is to do something like to following
.vb
	integer a(100)
	integer blens(10), indices(10)
	do i=1,10
		 blens(i)   = 1
10	   indices(i) = (1 + (i-1)*10) * sizeofint
	call MPI_TYPE_HINDEXED(10,blens,indices,MPI_INTEGER,newtype,ierr)
	call MPI_TYPE_COMMIT(newtype,ierr)
	call MPI_SEND(a,1,newtype,...)
.ve
expecting this to send 'a(1),a(11),...' because the indices have values 
'1,11,...'.   Because these are `displacements` from the beginning of 'a',
it actually sends 'a(1+1),a(1+11),...'.

If you wish to consider the displacements as indices into a Fortran array,
consider declaring the Fortran array with a zero origin
.vb
	integer a(0:99)
.ve

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_COUNT
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_hindexed(int count, int blocklens[], MPI_Aint indices[], MPI_Datatype old_type, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_hindexed";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *new_dtp;
	int i, *ints;
	MPIU_CHKLMEM_DECL(1);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_HINDEXED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_HINDEXED);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			int i;
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_DATATYPE(old_type, L"datatype", mpi_errno);
			if (count > 0)
			{
				MPIR_ERRTEST_ARGNULL(blocklens, L"blocklens", mpi_errno);
				MPIR_ERRTEST_ARGNULL(indices, L"indices", mpi_errno);
			}
			if (mpi_errno == MPI_SUCCESS)
			{
				if (HANDLE_GET_KIND(old_type) != HANDLE_KIND_BUILTIN)
				{
					MPID_Datatype_get_ptr( old_type, datatype_ptr );
					MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
				}
				/* verify that all blocklengths are >= 0 */
				for (i=0; i < count; i++)
				{
					MPIR_ERRTEST_ARGNEG(blocklens[i], L"blocklen", mpi_errno);
					MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"blocklens", i, blocklens[i]);
				}
			}
			MPIR_ERRTEST_ARGNULL(newtype, L"newtype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_hindexed %d %p %p %D %p", count, blocklens, indices, old_type, newtype); }
	
	mpi_errno = MPID_Type_indexed(count, blocklens, indices, 1 /* displacements in bytes */, old_type, newtype);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	MPIU_CHKLMEM_MALLOC(ints, int *, (count + 1) * sizeof(int), mpi_errno, L"contents integer array");

	/* copy ints into temporary buffer (count and blocklengths) */
	ints[0] = count;
	for (i=0; i < count; i++)
	{
		ints[i+1] = blocklens[i];
	}

	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_HINDEXED,
										   count+1, /* ints */
										   count, /* aints (displs) */
										   1, /* types */
										   ints,
										   indices,
										   &old_type);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_HINDEXED);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_hindexed",
			L"**mpi_type_hindexed %d %p %p %D %p", count, blocklens, indices, old_type, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_get_true_extent
#define MPI_Type_get_true_extent PMPI_Type_get_true_extent

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_get_true_extent

/*@
   MPI_Type_get_true_extent - Get the true lower bound and extent for a 
	 datatype

   Input Parameter:
. datatype - datatype to get information on (handle) 

   Output Parameters:
+ true_lb - true lower bound of datatype (integer) 
- true_extent - true size of datatype (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_true_extent(MPI_Datatype datatype, MPI_Aint *true_lb, MPI_Aint *true_extent)
{
	static const wchar_t FCNAME[] = L"MPI_Type_get_true_extent";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_GET_TRUE_EXTENT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_GET_TRUE_EXTENT);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_get_true_extent %D %p %p", datatype, true_lb, true_extent); }
	
	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN)
	{
		*true_lb	 = 0;
		*true_extent = MPID_Datatype_get_basic_size(datatype);
	}
	else
	{
		*true_lb	 = datatype_ptr->true_lb;
		*true_extent = datatype_ptr->true_ub - datatype_ptr->true_lb;
	}

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_GET_TRUE_EXTENT);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_get_true_extent", 
			L"**mpi_type_get_true_extent %D %p %p", datatype, true_lb, true_extent);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_get_nameA
#undef MPI_Type_get_nameW
#define MPI_Type_get_nameA PMPI_Type_get_nameA
#define MPI_Type_get_nameW PMPI_Type_get_nameW

#endif

/* This routine initializes all of the name fields in the predefined 
   datatypes */
#ifndef MPICH_MPI_FROM_PMPI
/* Include these definitions only with the PMPI version */
typedef struct mpi_names_t { MPI_Datatype dtype; const wchar_t *name; } mpi_names_t;
/* The MPI standard specifies that the names must be the MPI names,
   not the related language names (e.g., MPI_CHAR, not char) */
static mpi_names_t mpi_names[] =
{
	{ MPI_CHAR, L"MPI_CHAR" },
	{ MPI_SIGNED_CHAR, L"MPI_SIGNED_CHAR" },
	{ MPI_UNSIGNED_CHAR, L"MPI_UNSIGNED_CHAR" },
	{ MPI_BYTE, L"MPI_BYTE" },
	{ MPI_WCHAR, L"MPI_WCHAR" },
	{ MPI_SHORT, L"MPI_SHORT" },
	{ MPI_UNSIGNED_SHORT, L"MPI_UNSIGNED_SHORT" },
	{ MPI_INT, L"MPI_INT" },
	{ MPI_UNSIGNED, L"MPI_UNSIGNED" },
	{ MPI_LONG, L"MPI_LONG" },
	{ MPI_UNSIGNED_LONG, L"MPI_UNSIGNED_LONG" },
	{ MPI_FLOAT, L"MPI_FLOAT" },
	{ MPI_DOUBLE, L"MPI_DOUBLE" },
	{ MPI_LONG_DOUBLE, L"MPI_LONG_DOUBLE" },
	/* LONG_LONG_INT is allowed as an alias; we don't make it a separate
	   type */
/*	{ MPI_LONG_LONG_INT, L"MPI_LONG_LONG_INT" }, */
	{ MPI_LONG_LONG, L"MPI_LONG_LONG" },
	{ MPI_UNSIGNED_LONG_LONG, L"MPI_UNSIGNED_LONG_LONG" }, 
	{ MPI_PACKED, L"MPI_PACKED" },
	{ MPI_LB, L"MPI_LB" },
	{ MPI_UB, L"MPI_UB" },

/* The maxloc/minloc types are not builtin because they're size and
   extent may not be the same. */
/*
	{ MPI_FLOAT_INT, L"MPI_FLOAT_INT" },
	{ MPI_DOUBLE_INT, L"MPI_DOUBLE_INT" },
	{ MPI_LONG_INT, L"MPI_LONG_INT" },
	{ MPI_SHORT_INT, L"MPI_SHORT_INT" },
*/
	/* Fortran */
	{ MPI_COMPLEX, L"MPI_COMPLEX" },
	{ MPI_DOUBLE_COMPLEX, L"MPI_DOUBLE_COMPLEX" },
	{ MPI_LOGICAL, L"MPI_LOGICAL" },
	{ MPI_REAL, L"MPI_REAL" },
	{ MPI_DOUBLE_PRECISION, L"MPI_DOUBLE_PRECISION" },
	{ MPI_INTEGER, L"MPI_INTEGER" },
	{ MPI_2INTEGER, L"MPI_2INTEGER" },
	{ MPI_2COMPLEX, L"MPI_2COMPLEX" },
	{ MPI_2DOUBLE_COMPLEX, L"MPI_2DOUBLE_COMPLEX" },
	{ MPI_2REAL, L"MPI_2REAL" },
	{ MPI_2DOUBLE_PRECISION, L"MPI_2DOUBLE_PRECISION" },
	{ MPI_CHARACTER, L"MPI_CHARACTER" },
	/* Size-specific types (C, Fortran, and C++) */
	{ MPI_REAL4, L"MPI_REAL4" },
	{ MPI_REAL8, L"MPI_REAL8" },
	{ MPI_REAL16, L"MPI_REAL16" },
	{ MPI_COMPLEX8, L"MPI_COMPLEX8" },
	{ MPI_COMPLEX16, L"MPI_COMPLEX16" },
	{ MPI_COMPLEX32, L"MPI_COMPLEX32" },
	{ MPI_INTEGER1, L"MPI_INTEGER1" },
	{ MPI_INTEGER2, L"MPI_INTEGER2" },
	{ MPI_INTEGER4, L"MPI_INTEGER4" },
	{ MPI_INTEGER8, L"MPI_INTEGER8" },
	{ MPI_INTEGER16, L"MPI_INTEGER16" },
	{ 0, (wchar_t *)0 },  /* Sentinal used to indicate the last element */
};

static mpi_names_t mpi_maxloc_names[] =
{
	{ MPI_FLOAT_INT, L"MPI_FLOAT_INT" },
	{ MPI_DOUBLE_INT, L"MPI_DOUBLE_INT" },
	{ MPI_LONG_INT, L"MPI_LONG_INT" },
	{ MPI_SHORT_INT, L"MPI_SHORT_INT" },
	{ MPI_2INT, L"MPI_2INT" },
	{ MPI_LONG_DOUBLE_INT, L"MPI_LONG_DOUBLE_INT" },
	{ 0, (wchar_t *)0 },  /* Sentinal used to indicate the last element */
};
/* This routine is also needed by type_set_name */

int MPIR_Datatype_init_names( void ) 
{
	static const wchar_t FCNAME[] = L"MPIR_Datatype_init_names";
	int mpi_errno = MPI_SUCCESS;
	int i;
	MPID_Datatype *datatype_ptr = NULL;
	MPIU_THREADSAFE_INIT_DECL(needsInit);

	if (needsInit)
	{
		MPIU_THREADSAFE_INIT_BLOCK_BEGIN(needsInit);
		/* Make sure that the basics have datatype structures allocated
		 * and filled in for them.  They are just integers prior to this
		 * call.
		 */
		mpi_errno = MPIR_Datatype_builtin_fillin();
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_POPFATAL(mpi_errno);
		}
		
		/* For each predefined type, ensure that there is a corresponding
		   object and that the object's name is set */
		for (i=0; mpi_names[i].name != 0; i++)
		{
			/* The size-specific types may be DATATYPE_NULL, as might be those
			   based on 'long long' and 'long double' if those types were
			   disabled at configure time. */
			if (mpi_names[i].dtype == MPI_DATATYPE_NULL) continue;
			
			MPID_Datatype_get_ptr( mpi_names[i].dtype, datatype_ptr );
			
			if (datatype_ptr < MPID_Datatype_builtin || 
				datatype_ptr > MPID_Datatype_builtin + MPID_DATATYPE_N_BUILTIN)
				{
					MPIU_ERR_SETFATALANDJUMP1(mpi_errno,MPI_ERR_INTERN, L"**typeinitbadmem", L"**typeinitbadmem %d", i );
				}
			if (!datatype_ptr)
			{
				MPIU_ERR_SETFATALANDJUMP1(mpi_errno,MPI_ERR_INTERN, L"**typeinitfail", L"**typeinitfail %d", i - 1 )
			}

			/* MPIU_dbg_printf("mpi_names[%d].name = %x\n", i, (int) mpi_names[i].name ); */
			MPIU_Wcsncpy( datatype_ptr->name, mpi_names[i].name, MPI_MAX_OBJECT_NAME );
		}
		/* Handle the minloc/maxloc types */
		for (i=0; mpi_maxloc_names[i].name != 0; i++)
		{
			/* types based on 'long long' and 'long double' may be disabled at
			   configure time, and their values set to MPI_DATATYPE_NULL.  skip
			   those types. */
			if (mpi_maxloc_names[i].dtype == MPI_DATATYPE_NULL) continue;
			
			MPID_Datatype_get_ptr( mpi_maxloc_names[i].dtype, datatype_ptr );
			if (!datatype_ptr)
			{
				MPIU_ERR_SETFATALANDJUMP(mpi_errno,MPI_ERR_INTERN, L"**typeinitminmaxloc");
			}
			MPIU_Wcsncpy( datatype_ptr->name, mpi_maxloc_names[i].name, MPI_MAX_OBJECT_NAME );
		}
		MPIU_THREADSAFE_INIT_CLEAR(needsInit);
	fn_fail:;
	MPIU_THREADSAFE_INIT_BLOCK_END(needsInit);
	}

	return mpi_errno;
}
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_get_name

/*@
   MPI_Type_get_name - Get the print name for a datatype

   Input Parameter:
. type - datatype whose name is to be returned (handle) 

   Output Parameters:
+ type_name - the name previously stored on the datatype, or a empty string 
  if no such name exists (string) 
- resultlen - length of returned name (integer) 

.N ThreadSafeNoUpdate

.N NULL

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_nameA(MPI_Datatype datatype, char *type_name, int *resultlen)
{
	static const wchar_t FCNAME[] = L"MPI_Type_get_nameA";
	int mpi_errno = MPI_SUCCESS;
	wchar_t name[MPI_MAX_OBJECT_NAME] = L"";
	int len;
	size_t num_converted;
	errno_t error;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_GET_NAMEA);

	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_GET_NAMEA);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(type_name, L"type_name", mpi_errno);
			MPIR_ERRTEST_ARGNULL(resultlen, L"resultlen", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	mpi_errno = MPI_Type_get_nameW(datatype, name, &len);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (len == 0)
	{
		*type_name = '\0';
		*resultlen = 0;
		MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_GET_NAMEA);
		return mpi_errno;
	}

	error = wcstombs_s(&num_converted, type_name, MPI_MAX_OBJECT_NAME, name, _TRUNCATE);
	if (error != 0)
	{
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_OTHER, L"**fail", L"**fail %d", error);
		goto fn_fail;
	}
	if (num_converted > 0)
	{
		*resultlen = (int)num_converted - 1;
	}
	else
	{
		/* conversion error */
		MPIU_ERR_SET(mpi_errno, MPI_ERR_INTERN, L"**fail");
	}

fn_fail:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_GET_NAMEA);
	return mpi_errno;
}

MPI_DLL_SPEC int MPI_CALL MPI_Type_get_nameW(MPI_Datatype datatype, wchar_t *type_name, int *resultlen)
{
	static const wchar_t FCNAME[] = L"MPI_Type_get_name";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	static int setup = 0;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_GET_NAME);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_GET_NAME);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif

	/* Note that MPI_DATATYPE_NULL is invalid input to this routine;
	   it must not return a string for MPI_DATATYPE_NULL.  Instead, 
	   it must return an error indicating an invalid datatype argument */

	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			/* If datatype_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(type_name,L"type_name", mpi_errno);
			MPIR_ERRTEST_ARGNULL(resultlen,L"resultlen", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_get_name %D %p %p", datatype, type_name, resultlen); }
	
	/* If this is the first call, initialize all of the predefined names */
	if (!setup)
	{ 
		mpi_errno = MPIR_Datatype_init_names();
		if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		setup = 1;
	}

	/* Include the null in MPI_MAX_OBJECT_NAME */
	/*
	MPIU_Wcsncpy( type_name, datatype_ptr->name, MPI_MAX_OBJECT_NAME );
	*resultlen = (int)wcslen( type_name );
	*/
	MPIU_Wcsncpy( type_name, datatype_ptr->name, MPI_MAX_OBJECT_NAME );
	*resultlen = (int)wcslen( type_name );

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_GET_NAME);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_get_name", 
			L"**mpi_type_get_name %D %p %p", datatype, type_name, resultlen);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_get_extent
#define MPI_Type_get_extent PMPI_Type_get_extent

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_get_extent

/*@
   MPI_Type_get_extent - Get the lower bound and extent for a Datatype

   Input Parameter:
. datatype - datatype to get information on (handle) 

   Output Parameters:
+ lb - lower bound of datatype (integer) 
- extent - extent of datatype (integer) 

.N SignalSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_TYPE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_extent(MPI_Datatype datatype, MPI_Aint *lb, MPI_Aint *extent)
{
	static const wchar_t FCNAME[] = L"MPI_Type_get_extent";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_GET_EXTENT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_GET_EXTENT);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_get_extent %D %p %p", datatype, lb, extent); }
	
	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN)
	{
		*lb	 = 0;
		*extent = MPID_Datatype_get_basic_size(datatype);
	}
	else
	{
		*lb	 = datatype_ptr->lb;
		*extent = datatype_ptr->extent; /* derived, should be same as ub - lb */
	}

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_GET_EXTENT);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_get_extent",
			L"**mpi_type_get_extent %D %p %p", datatype, lb, extent);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_get_envelope
#define MPI_Type_get_envelope PMPI_Type_get_envelope

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_get_envelope

/*@
   MPI_Type_get_envelope - get type envelope

   Arguments:
+  MPI_Datatype datatype - datatype
.  int *num_integers - num integers
.  int *num_addresses - num addresses
.  int *num_datatypes - num datatypes
-  int *combiner - combiner

   Notes:

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_envelope(MPI_Datatype datatype, int *num_integers, int *num_addresses, int *num_datatypes, int *combiner)
{
	static const wchar_t FCNAME[] = L"MPI_Type_get_envelope";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_GET_ENVELOPE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_GET_ENVELOPE);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );
	
	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
			/* If comm_ptr is not value, it will be reset to null */
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_get_envelope %D %p %p %p %p", datatype, num_integers, num_addresses, num_datatypes, combiner); }
	
	/* handle all the predefined types here */
	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN ||
		datatype == MPI_FLOAT_INT ||
		datatype == MPI_DOUBLE_INT ||
		datatype == MPI_LONG_INT ||
		datatype == MPI_SHORT_INT ||
		datatype == MPI_LONG_DOUBLE_INT)
	{
		*num_integers  = 0;
		*num_addresses = 0;
		*num_datatypes = 0;
		*combiner = MPI_COMBINER_NAMED;
		goto fn_exit;
	}

	mpi_errno = MPID_Type_get_envelope(datatype, num_integers, num_addresses, num_datatypes, combiner);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_GET_ENVELOPE);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_get_envelope",
			L"**mpi_type_get_envelope %D %p %p %p %p", datatype, num_integers, num_addresses, num_datatypes, combiner);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_get_contents
#define MPI_Type_get_contents PMPI_Type_get_contents

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_get_contents

/*@
   MPI_Type_get_contents - get type contents

   Arguments:
+  MPI_Datatype datatype - datatype
.  int max_integers - max integers
.  int max_addresses - max addresses
.  int max_datatypes - max datatypes
.  int array_of_integers[] - integers
.  MPI_Aint array_of_addresses[] - addresses
-  MPI_Datatype array_of_datatypes[] - datatypes

   Notes:

.N Fortran

.N Errors
.N MPI_SUCCESS
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_contents(MPI_Datatype datatype, int max_integers, int max_addresses, int max_datatypes, int array_of_integers[], MPI_Aint array_of_addresses[], MPI_Datatype array_of_datatypes[])
{
	static const wchar_t FCNAME[] = L"MPI_Type_get_contents";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_GET_CONTENTS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_GET_CONTENTS);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr(datatype, datatype_ptr);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Check for built-in type */
			if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypeperm", 0);
				goto fn_fail;
			}

			/* all but MPI_2INT of the pair types are not stored as builtins
			 * but should be treated similarly.
			 */
			if (datatype == MPI_FLOAT_INT ||
				datatype == MPI_DOUBLE_INT ||
				datatype == MPI_LONG_INT ||
				datatype == MPI_SHORT_INT ||
				datatype == MPI_LONG_DOUBLE_INT)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypeperm", 0);
				goto fn_fail;
			}

			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
			/* If comm_ptr is not value, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_get_contents %D %d %d %d %p %p %p", datatype, max_integers, max_addresses, max_datatypes,
		array_of_integers, array_of_addresses, array_of_datatypes); }
	
	mpi_errno = MPID_Type_get_contents(datatype, max_integers, max_addresses, max_datatypes, array_of_integers, array_of_addresses, array_of_datatypes);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_GET_CONTENTS);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_get_contents",
			L"**mpi_type_get_contents %D %d %d %d %p %p %p", datatype, max_integers, max_addresses, max_datatypes,
			array_of_integers, array_of_addresses, array_of_datatypes);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_free
#define MPI_Type_free PMPI_Type_free

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_free

/*@
	MPI_Type_free - Frees the datatype

Input Parameter:
. datatype - datatype that is freed (handle) 

Predefined types:

The MPI standard states that (in Opaque Objects)
.Bqs
MPI provides certain predefined opaque objects and predefined, static handles
to these objects. Such objects may not be destroyed.
.Bqe

Thus, it is an error to free a predefined datatype.  The same section makes
it clear that it is an error to free a null datatype.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_free(MPI_Datatype *datatype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_free";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_FREE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_FREE);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			MPIR_ERRTEST_DATATYPE(*datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Validate parameters, especially handles needing to be converted */
	MPID_Datatype_get_ptr( *datatype, datatype_ptr );
	
	/* Convert MPI object handles to object pointers */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Check for built-in type */
			if (HANDLE_GET_KIND(*datatype) == HANDLE_KIND_BUILTIN)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypeperm", 0);
				goto fn_fail;
			}

			/* all but MPI_2INT of the pair types are not stored as builtins
			 * but should be treated similarly.
			 */
			if (*datatype == MPI_FLOAT_INT ||
				*datatype == MPI_DOUBLE_INT ||
				*datatype == MPI_LONG_INT ||
				*datatype == MPI_SHORT_INT ||
				*datatype == MPI_LONG_DOUBLE_INT)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypeperm", 0);
				goto fn_fail;
			}
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_free %p", datatype); }
	
	MPID_Datatype_release(datatype_ptr);
	*datatype = MPI_DATATYPE_NULL;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_FREE);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_free",
			L"**mpi_type_free %p", datatype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_extent
#define MPI_Type_extent PMPI_Type_extent

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_extent

/*@
	MPI_Type_extent - Returns the extent of a datatype

Input Parameters:
. datatype - datatype (handle) 

Output Parameter:
. extent - datatype extent (integer) 

.N SignalSafe

.N Deprecated
The replacement for this routine is 'MPI_Type_get_extent'.

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_extent(MPI_Datatype datatype, MPI_Aint *extent)
{
	static const wchar_t FCNAME[] = L"MPI_Type_extent";
	int mpi_errno = MPI_SUCCESS;
	MPI_Aint type_extent;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_EXTENT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_EXTENT);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_extent %D %p", datatype, extent); }
	
	MPID_Datatype_get_extent_macro(datatype, type_extent);
	*extent = type_extent;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_EXTENT);
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_extent",
			L"**mpi_type_extent %D %p", datatype, extent);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_dup
#define MPI_Type_dup PMPI_Type_dup

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_dup

/*@
   MPI_Type_dup - Duplicate a datatype

   Input Parameter:
. type - datatype (handle) 

   Output Parameter:
. newtype - copy of type (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_dup(MPI_Datatype datatype, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_dup";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr = NULL;
	MPID_Datatype *new_dtp;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_DUP);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_DUP);
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(datatype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( datatype, datatype_ptr );
	
	/* Convert MPI object handles to object pointers */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			/* If comm_ptr is not valid, it will be reset to null */
			MPIR_ERRTEST_ARGNULL(newtype, L"newtype", mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ...  */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_dup %D %p", datatype, newtype); }
	
	mpi_errno = MPID_Type_dup(datatype, newtype);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_DUP,
										   0, /* ints */
										   0, /* aints */
										   1, /* types */
										   NULL,
										   NULL,
										   &datatype);

	/* Copy attributes, executing the attribute copy functions */
	/* This accesses the attribute dup function through the perprocess
	   structure to prevent type_dup from forcing the linking of the
	   attribute functions.  The actual function is (by default)
	   MPIR_Attr_dup_list 
	*/
	if (mpi_errno == MPI_SUCCESS && MPIR_Process.attr_dup)
	{
		new_dtp->attributes = 0;
		mpi_errno = MPIR_Process.attr_dup( datatype, datatype_ptr->attributes, &new_dtp->attributes );
		if (mpi_errno)
		{
			MPID_Datatype_release(new_dtp);
			*newtype = MPI_DATATYPE_NULL;
			goto fn_fail;
		}
	}

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */

  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_DUP);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_dup",
			L"**mpi_type_dup %D %p", datatype, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_create_subarray
#define MPI_Type_create_subarray PMPI_Type_create_subarray

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_create_subarray

/*@
   MPI_Type_create_subarray - Create a datatype for a subarray of a regular, 
	multidimensional array

   Input Parameters:
+ ndims - number of array dimensions (positive integer) 
. array_of_sizes - number of elements of type oldtype in each dimension of the
  full array (array of positive integers) 
. array_of_subsizes - number of elements of type oldtype in each dimension of
  the subarray (array of positive integers) 
. array_of_starts - starting coordinates of the subarray in each dimension
  (array of nonnegative integers) 
. order - array storage order flag (state) 
- oldtype - array element datatype (handle) 

   Output Parameter:
. newtype - new datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_subarray(int ndims, int array_of_sizes[], int array_of_subsizes[], int array_of_starts[], int order, MPI_Datatype oldtype, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_create_subarray";
	int mpi_errno = MPI_SUCCESS, i;

	/* these variables are from the original version in ROMIO */
	MPI_Aint size, extent, disps[3];
	int blklens[3];
	MPI_Datatype tmp1, tmp2, types[3];

	/* for saving contents */
	int *ints;
	MPID_Datatype *new_dtp;

	MPID_Datatype *datatype_ptr = NULL;
	MPIU_CHKLMEM_DECL(1);
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_SUBARRAY);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CREATE_SUBARRAY);

	MPIU_THREADPRIV_GET;

	/* Get handles to MPI objects. */
	MPID_Datatype_get_ptr( oldtype, datatype_ptr );
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Check parameters */
			MPIR_ERRTEST_ARGNONPOS(ndims, L"ndims", mpi_errno);
			MPIR_ERRTEST_ARGNULL(array_of_sizes, L"array_of_sizes", mpi_errno);
			MPIR_ERRTEST_ARGNULL(array_of_subsizes, L"array_of_subsizes", mpi_errno);
			MPIR_ERRTEST_ARGNULL(array_of_starts, L"array_of_starts", mpi_errno);
			for (i=0; mpi_errno == MPI_SUCCESS && i < ndims; i++)
			{
				MPIR_ERRTEST_ARGNONPOS(array_of_sizes[i], L"size", mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_sizes", i, array_of_sizes[i]);
				MPIR_ERRTEST_ARGNONPOS(array_of_subsizes[i], L"subsize", mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_subsizes", i, array_of_subsizes[i]);
				MPIR_ERRTEST_ARGNEG(array_of_starts[i], L"start", mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_starts", i, array_of_starts[i]);
				if (array_of_subsizes[i] > array_of_sizes[i])
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**argrange", L"**argrange %s %d %d", L"array_of_subsizes", array_of_subsizes[i], array_of_sizes[i]);
				}
				if (array_of_starts[i] > (array_of_sizes[i] - array_of_subsizes[i]))
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**argrange", L"**argrange %s %d %d", L"array_of_starts", array_of_starts[i], array_of_sizes[i] - array_of_subsizes[i]);
				}
			}
			if (order != MPI_ORDER_FORTRAN && order != MPI_ORDER_C)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**arg", L"**arg %s", L"order");
			}

#if 0
			/* TODO: INTEGRATE THIS CHECK AS WELL */
			NMPI_Type_extent(mpi_errno, (oldtype, &extent));

			/* check if MPI_Aint is large enough for size of global array. 
			   if not, complain. */

			size_with_aint = extent;
			for (i=0; i<ndims; i++) size_with_aint *= array_of_sizes[i];
			size_with_offset = extent;
			for (i=0; i<ndims; i++) size_with_offset *= array_of_sizes[i];
			if (size_with_aint != size_with_offset)
			{
				FPRINTF(stderr, L"MPI_Type_create_subarray: Can't use an array of this size unless the MPI implementation defines a 64-bit MPI_Aint\n");
				MPI_Abort(MPI_COMM_WORLD, 1);
	}
#endif

			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
			/* If datatype_ptr is not valid, it will be reset to null */
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_create_subarray %d %p %p %p %d %D %p", ndims, array_of_sizes, array_of_subsizes,
		array_of_starts, order, oldtype, newtype); }
	
	/* TODO: CHECK THE ERROR RETURNS FROM ALL THESE!!! */

	/* TODO: GRAB EXTENT WITH A MACRO OR SOMETHING FASTER */
	NMPI_Type_extent(mpi_errno, (oldtype, &extent));

	if (order == MPI_ORDER_FORTRAN)
	{
		if (ndims == 1)
			mpi_errno = MPID_Type_contiguous(array_of_subsizes[0], oldtype, &tmp1);
		else
		{
			mpi_errno = MPID_Type_vector(array_of_subsizes[1], array_of_subsizes[0], array_of_sizes[0], 0 /* stride in types */, oldtype, &tmp1);
			
			size = array_of_sizes[0]*extent;
			for (i=2; i<ndims; i++)
			{
				size *= array_of_sizes[i-1];
				mpi_errno = MPID_Type_vector(array_of_subsizes[i], 1, size, 1 /* stride in bytes */, tmp1, &tmp2);
				NMPI_Type_free(mpi_errno, (&tmp1));
				tmp1 = tmp2;
			}
		}
		
		/* add displacement and UB */
		
		disps[1] = array_of_starts[0];
		size = 1;
		for (i=1; i<ndims; i++)
		{
			size *= array_of_sizes[i-1];
			disps[1] += size*array_of_starts[i];
		}  
		/* rest done below for both Fortran and C order */
	}
	else /* MPI_ORDER_C */
	{
		/* dimension ndims-1 changes fastest */
		if (ndims == 1)
		{
			mpi_errno = MPID_Type_contiguous(array_of_subsizes[0], oldtype, &tmp1);

		}
		else
		{
			mpi_errno = MPID_Type_vector(array_of_subsizes[ndims-2],
										 array_of_subsizes[ndims-1],
										 array_of_sizes[ndims-1],
										 0, /* stride in types */
										 oldtype,
										 &tmp1);
			

			size = array_of_sizes[ndims-1]*extent;
			for (i=ndims-3; i>=0; i--)
			{
				size *= array_of_sizes[i+1];
				mpi_errno = MPID_Type_vector(array_of_subsizes[i],
											 1,	/* blocklen */
											 size, /* stride */
											 1,	/* stride in bytes */
											 tmp1, /* old type */
											 &tmp2);

				NMPI_Type_free(mpi_errno, (&tmp1));
				tmp1 = tmp2;
			}
		}
		
		/* add displacement and UB */
		
		disps[1] = array_of_starts[ndims-1];
		size = 1;
		for (i=ndims-2; i>=0; i--)
		{
			size *= array_of_sizes[i+1];
			disps[1] += size*array_of_starts[i];
		}
	}

	disps[1] *= extent;
	
	disps[2] = extent;
	for (i=0; i<ndims; i++) disps[2] *= array_of_sizes[i];
	
	disps[0] = 0;
	blklens[0] = blklens[1] = blklens[2] = 1;
	types[0] = MPI_LB;
	types[1] = tmp1;
	types[2] = MPI_UB;
	
	/* TODO:
	 * if we were to do all this as an mpid function, we could just
	 * directly adjust the LB and UB in the MPID_Datatype structure
	 * instead of jumping through this hoop.
	 *
	 * i suppose we could do the same thing here...
	 *
	 * another alternative would be to use MPID_Type_create_resized() 
	 * instead of building the struct.  that would also be cleaner.
	 */
	mpi_errno = MPID_Type_struct(3, blklens, disps, types, newtype);

	NMPI_Type_free(mpi_errno, (&tmp1));

	/* at this point we have the new type, and we've cleaned up any
	 * intermediate types created in the process.  we just need to save
	 * all our contents/envelope information.
	 */

	/* Save contents */
	MPIU_CHKLMEM_MALLOC_ORJUMP(ints, int *, (3 * ndims + 2) * sizeof(int), mpi_errno, L"content description");

	ints[0] = ndims;
	for (i=0; i < ndims; i++)
	{
		ints[i + 1] = array_of_sizes[i];
	}
	for(i=0; i < ndims; i++)
	{
		ints[i + ndims + 1] = array_of_subsizes[i];
	}
	for(i=0; i < ndims; i++)
	{
		ints[i + 2*ndims + 1] = array_of_starts[i];
	}
	ints[3*ndims + 1] = order;

	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_SUBARRAY,
										   3 * ndims + 2, /* ints */
										   0, /* aints */
										   1, /* types */
										   ints,
										   NULL,
										   &oldtype);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */
	MPIR_Nest_decr();
	
  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CREATE_SUBARRAY);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	MPIR_Nest_decr();
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_create_subarray", 
			L"**mpi_type_create_subarray %d %p %p %p %d %D %p", ndims, array_of_sizes, array_of_subsizes,
			array_of_starts, order, oldtype, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_create_struct
#define MPI_Type_create_struct PMPI_Type_create_struct

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_create_struct

/*@
   MPI_Type_create_struct - Create an MPI datatype from a general set of 
   datatypes, displacements, and block sizes

   Input Parameters:
+ count - number of blocks (integer) --- also number of entries in arrays 
  array_of_types, array_of_displacements and array_of_blocklengths 
. array_of_blocklength - number of elements in each block (array of integer) 
. array_of_displacements - byte displacement of each block (array of integer) 
- array_of_types - type of elements in each block (array of handles to 
  datatype objects) 

   Output Parameter:
. newtype - new datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_TYPE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_struct(int count,
						   int array_of_blocklengths[],
						   MPI_Aint array_of_displacements[],
						   MPI_Datatype array_of_types[],
						   MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_create_struct";
	int mpi_errno = MPI_SUCCESS;
	int i, *ints;
	MPID_Datatype *new_dtp;
	MPIU_CHKLMEM_DECL(1);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_STRUCT);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CREATE_STRUCT);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			int i;
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_COUNT(count,mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;

			if (count > 0)
			{ 
				MPIR_ERRTEST_ARGNULL(array_of_blocklengths, L"blocklens", mpi_errno);
				MPIR_ERRTEST_ARGNULL(array_of_displacements, L"indices", mpi_errno);
				MPIR_ERRTEST_ARGNULL(array_of_types, L"types", mpi_errno);
				if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			}

			for (i=0; i < count; i++)
			{
				MPIR_ERRTEST_ARGNEG(array_of_blocklengths[i], L"blocklen", mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_blocklengths", i, array_of_blocklengths[i]);
				MPIR_ERRTEST_DATATYPE(array_of_types[i], L"datatype[i]", mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_types", i, array_of_types[i]);
						
				if (array_of_types[i] != MPI_DATATYPE_NULL && HANDLE_GET_KIND(array_of_types[i]) != HANDLE_KIND_BUILTIN)
				{
					MPID_Datatype_get_ptr(array_of_types[i], datatype_ptr);
					MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
					MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_types", i, array_of_types[i]);
				}
			}
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_create_struct %d %p %p %p %p", count, array_of_blocklengths, array_of_displacements,
		array_of_types, newtype); }

	mpi_errno = MPID_Type_struct(count, array_of_blocklengths, array_of_displacements, array_of_types, newtype);

	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
		goto fn_fail;
	/* --END ERROR HANDLING-- */

	MPIU_CHKLMEM_MALLOC_ORJUMP(ints, int *, (count + 1) * sizeof(int), mpi_errno, L"content description");

	ints[0] = count;
	for (i=0; i < count; i++)
	{
		ints[i+1] = array_of_blocklengths[i];
	}

	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_STRUCT,
										   count+1, /* ints (cnt,blklen) */
										   count, /* aints (disps) */
										   count, /* types */
										   ints,
										   array_of_displacements,
										   array_of_types);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CREATE_STRUCT);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_create_struct",
			L"**mpi_type_create_struct %d %p %p %p %p", count, array_of_blocklengths, array_of_displacements,
			array_of_types, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_create_resized
#define MPI_Type_create_resized PMPI_Type_create_resized

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_create_resized

/*@
   MPI_Type_create_resized - Create a datatype with a new lower bound and 
	 extent from an existing datatype

   Input Parameters:
+ oldtype - input datatype (handle) 
. lb - new lower bound of datatype (integer) 
- extent - new extent of datatype (integer) 

   Output Parameter:
. newtype - output datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_resized(MPI_Datatype oldtype, MPI_Aint lb, MPI_Aint extent, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_create_resized";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *new_dtp;
	MPI_Aint aints[2];
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_RESIZED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CREATE_RESIZED);

	/* Get handles to MPI objects. */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Datatype *datatype_ptr = NULL;
			
			MPIR_ERRTEST_DATATYPE(oldtype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			
			/* Validate datatype_ptr */
			MPID_Datatype_get_ptr( oldtype, datatype_ptr );
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			/* If datatype_ptr is not valid, it will be reset to null */
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_create_resized %D %d %d %p", oldtype, lb, extent, newtype); }
	
	mpi_errno = MPID_Type_create_resized(oldtype, lb, extent, newtype);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
		goto fn_fail;
	/* --END ERROR HANDLING-- */

	aints[0] = lb;
	aints[1] = extent;

	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_RESIZED,
										   0,
										   2, /* Aints */
										   1,
										   NULL,
										   aints,
										   &oldtype);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CREATE_RESIZED);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_create_resized",
			L"**mpi_type_create_resized %D %d %d %p", oldtype, lb, extent, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_create_indexed_block
#define MPI_Type_create_indexed_block PMPI_Type_create_indexed_block

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_create_indexed_block

/*@
   MPI_Type_create_indexed_block - Create an indexed
	 datatype with constant-sized blocks

   Input Parameters:
+ count - length of array of displacements (integer) 
. blocklength - size of block (integer) 
. array_of_displacements - array of displacements (array of integer) 
- oldtype - old datatype (handle) 

	Output Parameter:
. newtype - new datatype (handle) 

Notes:
The indices are displacements, and are based on a zero origin.  A common error
is to do something like to following
.vb
	integer a(100)
	integer blens(10), indices(10)
	do i=1,10
10	   indices(i) = 1 + (i-1)*10
	call MPI_TYPE_CREATE_INDEXED_BLOCK(10,1,indices,MPI_INTEGER,newtype,ierr)
	call MPI_TYPE_COMMIT(newtype,ierr)
	call MPI_SEND(a,1,newtype,...)
.ve
expecting this to send 'a(1),a(11),...' because the indices have values 
'1,11,...'.   Because these are `displacements` from the beginning of 'a',
it actually sends 'a(1+1),a(1+11),...'.

If you wish to consider the displacements as indices into a Fortran array,
consider declaring the Fortran array with a zero origin
.vb
	integer a(0:99)
.ve

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_indexed_block(int count, int blocklength, int array_of_displacements[], MPI_Datatype oldtype, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_create_indexed_block";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *new_dtp;
	int i, *ints;
	MPIU_CHKLMEM_DECL(1);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_INDEXED_BLOCK);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CREATE_INDEXED_BLOCK);
	
	/* Validate parameters and objects */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_ARGNEG(blocklength, L"blocklen", mpi_errno);
			if (count > 0)
			{
				MPIR_ERRTEST_ARGNULL(array_of_displacements, L"indices", mpi_errno);
			}
			MPIR_ERRTEST_DATATYPE(oldtype, L"datatype", mpi_errno);
			if (mpi_errno) goto fn_fail;
			
			if (HANDLE_GET_KIND(oldtype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(oldtype, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
			}

			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_create_indexed_block %d %d %p %D %p", count, blocklength, array_of_displacements, oldtype, newtype); }
	
	mpi_errno = MPID_Type_blockindexed(count, blocklength, array_of_displacements, 0 /* dispinbytes */, oldtype, newtype);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	MPIU_CHKLMEM_MALLOC_ORJUMP(ints, int *, (count + 2) * sizeof(int), mpi_errno, L"content description");

	ints[0] = count;
	ints[1] = blocklength;

	for (i=0; i < count; i++)
	{
		ints[i+2] = array_of_displacements[i];
	}

	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_INDEXED_BLOCK,
										   count + 2, /* ints */
										   0, /* aints */
										   1, /* types */
										   ints,
										   NULL,
										   &oldtype);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CREATE_INDEXED_BLOCK);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_create_indexed_block",
			L"**mpi_type_create_indexed_block %d %d %p %D %p", count, blocklength, array_of_displacements, oldtype, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_create_hvector
#define MPI_Type_create_hvector PMPI_Type_create_hvector

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_create_hvector

/*@
   MPI_Type_create_hvector - Create a datatype with a constant stride given
	 in bytes

   Input Parameters:
+ count - number of blocks (nonnegative integer) 
. blocklength - number of elements in each block (nonnegative integer) 
. stride - number of bytes between start of each block (integer) 
- oldtype - old datatype (handle) 

   Output Parameter:
. newtype - new datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_hvector(int count, int blocklength, MPI_Aint stride, MPI_Datatype oldtype, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_create_hvector";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *new_dtp;
	int ints[2];
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_HVECTOR);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CREATE_HVECTOR);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_COUNT(count, mpi_errno);
			MPIR_ERRTEST_ARGNEG(blocklength, L"blocklen", mpi_errno);
			MPIR_ERRTEST_DATATYPE(oldtype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;

			MPID_Datatype_get_ptr(oldtype, datatype_ptr);
			MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_create_hvector %d %d %d %D %p", count,
		blocklength, stride, oldtype, newtype); }
	
	mpi_errno = MPID_Type_vector(count, blocklength, stride, 1 /* stride in bytes */, oldtype, newtype);
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	ints[0] = count;
	ints[1] = blocklength;
	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_HVECTOR,
										   3, /* ints (count, blocklength) */
										   1, /* aints */
										   1, /* types */
										   ints,
										   &stride,
										   &oldtype);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CREATE_HVECTOR);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__,
			MPI_ERR_OTHER, L"**mpi_type_create_hvector",
			L"**mpi_type_create_hvector %d %d %d %D %p", count,
			blocklength, stride, oldtype, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_create_hindexed
#define MPI_Type_create_hindexed PMPI_Type_create_hindexed

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_create_hindexed

/*@
   MPI_Type_create_hindexed - Create a datatype for an indexed datatype with 
   displacements in bytes

   Input Parameters:
+ count - number of blocks --- also number of entries in 
  displacements and blocklengths (integer) 
. blocklengths - number of elements in each block (array of nonnegative integers) 
. displacements - byte displacement of each block (array of integer) 
- oldtype - old datatype (handle) 

   Output Parameter:
. newtype - new datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_hindexed(int count, int blocklengths[], MPI_Aint displacements[], MPI_Datatype oldtype, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_create_hindexed";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *new_dtp;
	int i, *ints;
	MPIU_CHKLMEM_DECL(1);
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_HINDEXED);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CREATE_HINDEXED);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			int i;
			MPID_Datatype *datatype_ptr = NULL;

			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (count > 0)
			{
				MPIR_ERRTEST_ARGNULL(blocklengths, L"blocklens", mpi_errno);
				MPIR_ERRTEST_ARGNULL(displacements, L"indices", mpi_errno);
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			
			MPIR_ERRTEST_DATATYPE(oldtype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			
			if (HANDLE_GET_KIND(oldtype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr( oldtype, datatype_ptr );
				MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			}
			for (i=0; i < count; i++)
			{
				MPIR_ERRTEST_ARGNEG(blocklengths[i], L"blocklen", mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"blocklengths", i, blocklengths[i]);
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_create_hindexed %d %p %p %D %p", count, blocklengths, displacements, oldtype, newtype); }
	
	mpi_errno = MPID_Type_indexed(count, blocklengths, displacements, 1 /* displacements in bytes */, oldtype, newtype);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	MPIU_CHKLMEM_MALLOC_ORJUMP(ints, int *, (count + 1) * sizeof(int), mpi_errno, L"content description");

	ints[0] = count;

	for (i=0; i < count; i++)
	{
		ints[i+1] = blocklengths[i];
	}
	MPID_Datatype_get_ptr(*newtype, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp,
										   MPI_COMBINER_HINDEXED,
										   count+1, /* ints (count, blocklengths) */
										   count, /* aints (displacements) */
										   1, /* types */
										   ints,
										   displacements,
										   &oldtype);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	
	/* ... end of body of routine ... */

  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CREATE_HINDEXED);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_create_hindexed",
			L"**mpi_type_create_hindexed %d %p %p %D %p", count, blocklengths, displacements, oldtype, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#ifndef MIN
#define MIN(__a, __b) (((__a) < (__b)) ? (__a) : (__b))
#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_create_darray
#define MPI_Type_create_darray PMPI_Type_create_darray

PMPI_LOCAL int MPIR_Type_block(int *array_of_gsizes, int dim, int ndims, int nprocs, int rank, int darg, int order, MPI_Aint orig_extent, MPI_Datatype type_old, MPI_Datatype *type_new, MPI_Aint *st_offset);
PMPI_LOCAL int MPIR_Type_cyclic(int *array_of_gsizes, int dim, int ndims, int nprocs, int rank, int darg, int order, MPI_Aint orig_extent, MPI_Datatype type_old, MPI_Datatype *type_new, MPI_Aint *st_offset);


#undef FUNCNAME
#define FUNCNAME MPIR_Type_block
PMPI_LOCAL int MPIR_Type_block(int *array_of_gsizes, int dim, int ndims, int nprocs, int rank, int darg, int order, MPI_Aint orig_extent, MPI_Datatype type_old, MPI_Datatype *type_new, MPI_Aint *st_offset)
{
/* nprocs = no. of processes in dimension dim of grid
   rank = coordinate of this process in dimension dim */
	static const wchar_t FCNAME[] = L"MPIR_Type_block";
	int mpi_errno, blksize, global_size, mysize, i, j;
	MPI_Aint stride;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_TYPE_BLOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_TYPE_BLOCK);
	
	global_size = array_of_gsizes[dim];

	if (darg == MPI_DISTRIBUTE_DFLT_DARG)
		blksize = (global_size + nprocs - 1)/nprocs;
	else
	{
		blksize = darg;

#ifdef HAVE_ERROR_CHECKING
		if (blksize <= 0)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**darrayblock", L"**darrayblock %d", blksize);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_BLOCK);
			return mpi_errno;
		}
		if (blksize * nprocs < global_size)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**darrayblock2", L"**darrayblock2 %d %d", blksize*nprocs, global_size);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_BLOCK);
			return mpi_errno;
		}
#endif
	}

	j = global_size - blksize*rank;
	mysize = MIN(blksize, j);
	if (mysize < 0) mysize = 0;

	stride = orig_extent;
	if (order == MPI_ORDER_FORTRAN)
	{
		if (dim == 0)
		{
			mpi_errno = MPID_Type_contiguous(mysize, type_old, type_new);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_BLOCK);
				return mpi_errno;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			for (i=0; i<dim; i++) stride *= array_of_gsizes[i];
			mpi_errno = MPID_Type_vector(mysize, 1, stride, 1 /* stride in bytes */, type_old, type_new);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_BLOCK);
				return mpi_errno;
			}
			/* --END ERROR HANDLING-- */
		}
	}
	else
	{
		if (dim == ndims-1)
		{
			mpi_errno = MPID_Type_contiguous(mysize, type_old, type_new);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_BLOCK);
				return mpi_errno;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			for (i=ndims-1; i>dim; i--) stride *= array_of_gsizes[i];
			mpi_errno = MPID_Type_vector(mysize, 1, stride, 1 /* stride in bytes */, type_old, type_new);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_BLOCK);
				return mpi_errno;
			}
			/* --END ERROR HANDLING-- */
		}
	}

	*st_offset = blksize * rank;
	 /* in terms of no. of elements of type oldtype in this dimension */
	if (mysize == 0) *st_offset = 0;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_BLOCK);
	return MPI_SUCCESS;
}


#undef FUNCNAME
#define FUNCNAME MPIR_Type_cyclic
PMPI_LOCAL int MPIR_Type_cyclic(int *array_of_gsizes, int dim, int ndims, int nprocs, int rank, int darg, int order, MPI_Aint orig_extent, MPI_Datatype type_old, MPI_Datatype *type_new, MPI_Aint *st_offset)
{
/* nprocs = no. of processes in dimension dim of grid
   rank = coordinate of this process in dimension dim */
	static const wchar_t FCNAME[] = L"MPIR_Type_cyclic";
	int mpi_errno,blksize, i, blklens[3], st_index, end_index,
		local_size, rem, count;
	MPI_Aint stride, disps[3];
	MPI_Datatype type_tmp, types[3];
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_TYPE_CYCLIC);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_TYPE_CYCLIC);

	MPIU_THREADPRIV_GET;

	if (darg == MPI_DISTRIBUTE_DFLT_DARG) blksize = 1;
	else blksize = darg;

#ifdef HAVE_ERROR_CHECKING
	if (blksize <= 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**darraycyclic", L"**darraycyclic %d", blksize);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_CYCLIC);
		return mpi_errno;
	}
#endif
	
	st_index = rank*blksize;
	end_index = array_of_gsizes[dim] - 1;

	if (end_index < st_index) local_size = 0;
	else
	{
		local_size = ((end_index - st_index + 1)/(nprocs*blksize))*blksize;
		rem = (end_index - st_index + 1) % (nprocs*blksize);
		local_size += MIN(rem, blksize);
	}

	count = local_size/blksize;
	rem = local_size % blksize;
	
	stride = nprocs*blksize*orig_extent;
	if (order == MPI_ORDER_FORTRAN)
		for (i=0; i<dim; i++) stride *= array_of_gsizes[i];
	else for (i=ndims-1; i>dim; i--) stride *= array_of_gsizes[i];

	mpi_errno = MPID_Type_vector(count, blksize, stride, 1 /* stride in bytes */, type_old, type_new);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_CYCLIC);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */

	if (rem)
	{
		/* if the last block is of size less than blksize, include
		   it separately using MPI_Type_struct */

		types[0] = *type_new;
		types[1] = type_old;
		disps[0] = 0;
		disps[1] = count*stride;
		blklens[0] = 1;
		blklens[1] = rem;

		mpi_errno = MPID_Type_struct(2, blklens, disps, types, &type_tmp);
		NMPI_Type_free(mpi_errno, (type_new));
		*type_new = type_tmp;

		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_CYCLIC);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */
	}

	/* In the first iteration, we need to set the displacement in that
	   dimension correctly. */ 
	if ( ((order == MPI_ORDER_FORTRAN) && (dim == 0)) ||
		 ((order == MPI_ORDER_C) && (dim == ndims-1)) )
	{
		types[0] = MPI_LB;
		disps[0] = 0;
		types[1] = *type_new;
		disps[1] = rank * blksize * orig_extent;
		types[2] = MPI_UB;
		disps[2] = orig_extent * array_of_gsizes[dim];
		blklens[0] = blklens[1] = blklens[2] = 1;
		mpi_errno = MPID_Type_struct(3, blklens, disps, types, &type_tmp);
		NMPI_Type_free(mpi_errno, (type_new));
		*type_new = type_tmp;

		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_CYCLIC);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */

		*st_offset = 0;  /* set it to 0 because it is taken care of in
							the struct above */
	}
	else
	{
		*st_offset = rank * blksize; 
		/* st_offset is in terms of no. of elements of type oldtype in
		 * this dimension */ 
	}

	if (local_size == 0) *st_offset = 0;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_CYCLIC);
	return MPI_SUCCESS;
}
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_create_darray


/*@
   MPI_Type_create_darray - Create a datatype representing a distributed array

   Input Parameters:
+ size - size of process group (positive integer) 
. rank - rank in process group (nonnegative integer) 
. ndims - number of array dimensions as well as process grid dimensions (positive integer) 
. array_of_gsizes - number of elements of type oldtype in each dimension of global array (array of positive integers) 
. array_of_distribs - distribution of array in each dimension (array of state) 
. array_of_dargs - distribution argument in each dimension (array of positive integers) 
. array_of_psizes - size of process grid in each dimension (array of positive integers) 
. order - array storage order flag (state) 
- oldtype - old datatype (handle) 

	Output Parameter:
. newtype - new datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_darray(int size, int rank, int ndims, int array_of_gsizes[], int array_of_distribs[], int array_of_dargs[], int array_of_psizes[], int order, MPI_Datatype oldtype, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"MPI_Type_create_darray";
	int mpi_errno = MPI_SUCCESS, i;

	int procs, tmp_rank, tmp_size, blklens[3], *coords;
	MPI_Aint *st_offsets, orig_extent, disps[3];
	MPI_Datatype type_old, type_new = MPI_DATATYPE_NULL, types[3];

	int *ints;
	MPID_Datatype *datatype_ptr = NULL;
	MPIU_CHKLMEM_DECL(3);
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_DARRAY);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CREATE_DARRAY);

	MPIU_THREADPRIV_GET;
	
	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_DATATYPE(oldtype, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif
	
	/* Convert MPI object handles to object pointers */
	MPID_Datatype_get_ptr( oldtype, datatype_ptr );
	MPID_Datatype_get_extent_macro(oldtype, orig_extent);

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Check parameters */
			MPIR_ERRTEST_ARGNEG(rank, L"rank", mpi_errno);
			MPIR_ERRTEST_ARGNONPOS(size, L"size", mpi_errno);
			MPIR_ERRTEST_ARGNONPOS(ndims, L"ndims", mpi_errno);

			MPIR_ERRTEST_ARGNULL(array_of_gsizes, L"array_of_gsizes", mpi_errno);
			MPIR_ERRTEST_ARGNULL(array_of_distribs, L"array_of_distribs", mpi_errno);
			MPIR_ERRTEST_ARGNULL(array_of_dargs, L"array_of_dargs", mpi_errno);
			MPIR_ERRTEST_ARGNULL(array_of_psizes, L"array_of_psizes", mpi_errno);
			if (order != MPI_ORDER_C && order != MPI_ORDER_FORTRAN)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**arg", L"**arg %s", L"order");
			}

			for (i=0; mpi_errno == MPI_SUCCESS && i < ndims; i++)
			{
				MPIR_ERRTEST_ARGNONPOS(array_of_gsizes[i], L"gsize", mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_gsizes", i, array_of_gsizes[i]);
				MPIR_ERRTEST_ARGNONPOS(array_of_psizes[i], L"psize", mpi_errno);
				MPIU_ERR_CHK_POP_ARRAY(mpi_errno, L"array_of_psizes", i, array_of_psizes[i]);

				if ((array_of_distribs[i] != MPI_DISTRIBUTE_NONE) &&
					(array_of_distribs[i] != MPI_DISTRIBUTE_BLOCK) &&
					(array_of_distribs[i] != MPI_DISTRIBUTE_CYCLIC))
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**darrayunknown", 0);
				}

				if ((array_of_dargs[i] != MPI_DISTRIBUTE_DFLT_DARG) && 
					(array_of_dargs[i] <= 0))
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**arg", L"**arg %s", L"array_of_dargs");
				}

				if ((array_of_distribs[i] == MPI_DISTRIBUTE_NONE) &&
					(array_of_psizes[i] != 1))
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_ARG, L"**darraydist", L"**darraydist %d %d", i, array_of_psizes[i]);
				}
			}

			/* TODO: GET THIS CHECK IN ALSO */
#if 0
			/* check if MPI_Aint is large enough for size of global array. 
			   if not, complain. */
			
			size_with_aint = orig_extent;
			for (i=0; i<ndims; i++) size_with_aint *= array_of_gsizes[i];
			size_with_offset = orig_extent;
			for (i=0; i<ndims; i++) size_with_offset *= array_of_gsizes[i];
			if (size_with_aint != size_with_offset)
			{
				FPRINTF(stderr, L"MPI_Type_create_darray: Can't use an array of this size unless the MPI implementation defines a 64-bit MPI_Aint\n");
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
#endif			

			/* Validate datatype_ptr */
			MPID_Datatype_valid_ptr( datatype_ptr, mpi_errno );
			/* If datatype_ptr is not valid, it will be reset to null */
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno) goto fn_fail;
			/* --END ERROR HANDLING-- */
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_create_darray %d %d %d %p %p %p %p %d %D %p", size, rank, ndims, array_of_gsizes,
		array_of_distribs, array_of_dargs, array_of_psizes, order, oldtype, newtype); }
	
/* calculate position in Cartesian grid as MPI would (row-major
   ordering) */
	MPIU_CHKLMEM_MALLOC_ORJUMP(coords, int *, ndims * sizeof(int), mpi_errno, L"position is Cartesian grid");

	procs = size;
	tmp_rank = rank;
	for (i=0; i<ndims; i++)
	{
		procs = procs/array_of_psizes[i];
		coords[i] = tmp_rank/procs;
		tmp_rank = tmp_rank % procs;
	}

	MPIU_CHKLMEM_MALLOC_ORJUMP(st_offsets, MPI_Aint *, ndims * sizeof(MPI_Aint), mpi_errno, L"st_offsets");

	type_old = oldtype;

	if (order == MPI_ORDER_FORTRAN)
	{
	  /* dimension 0 changes fastest */
		for (i=0; i<ndims; i++)
		{
			switch(array_of_distribs[i])
			{
			case MPI_DISTRIBUTE_BLOCK:
				mpi_errno = MPIR_Type_block(array_of_gsizes, i, ndims, array_of_psizes[i], coords[i], array_of_dargs[i], order, orig_extent, type_old, &type_new, st_offsets+i); 
				break;
			case MPI_DISTRIBUTE_CYCLIC:
				mpi_errno = MPIR_Type_cyclic(array_of_gsizes, i, ndims, array_of_psizes[i], coords[i], array_of_dargs[i], order, orig_extent, type_old, &type_new, st_offsets+i);
				break;
			case MPI_DISTRIBUTE_NONE:
				/* treat it as a block distribution on 1 process */
				mpi_errno = MPIR_Type_block(array_of_gsizes, i, ndims, 1, 0, MPI_DISTRIBUTE_DFLT_DARG, order, orig_extent, type_old, &type_new, st_offsets+i); 
				break;
			}
			if (i)
			{
				NMPI_Type_free(mpi_errno, (&type_old));
			}
			type_old = type_new;

			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			/* --END ERROR HANDLING-- */
		}

		/* add displacement and UB */
		disps[1] = st_offsets[0];
		tmp_size = 1;
		for (i=1; i<ndims; i++)
		{
			tmp_size *= array_of_gsizes[i-1];
			disps[1] += tmp_size*st_offsets[i];
		}
		/* rest done below for both Fortran and C order */
	}

	else /* order == MPI_ORDER_C */
	{
		/* dimension ndims-1 changes fastest */
		for (i=ndims-1; i>=0; i--)
		{
			switch(array_of_distribs[i])
			{
			case MPI_DISTRIBUTE_BLOCK:
				mpi_errno = MPIR_Type_block(array_of_gsizes, i, ndims, array_of_psizes[i], coords[i], array_of_dargs[i], order, orig_extent, type_old, &type_new, st_offsets+i); 
				break;
			case MPI_DISTRIBUTE_CYCLIC:
				mpi_errno = MPIR_Type_cyclic(array_of_gsizes, i, ndims, array_of_psizes[i], coords[i], array_of_dargs[i], order, orig_extent, type_old, &type_new, st_offsets+i);
				break;
			case MPI_DISTRIBUTE_NONE:
				/* treat it as a block distribution on 1 process */
				mpi_errno = MPIR_Type_block(array_of_gsizes, i, ndims, array_of_psizes[i], coords[i], MPI_DISTRIBUTE_DFLT_DARG, order, orig_extent, type_old, &type_new, st_offsets+i);
				break;
			}
			if (i != ndims-1)
			{
				NMPI_Type_free(mpi_errno, (&type_old));
			}
			type_old = type_new;

			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			/* --END ERROR HANDLING-- */
		}

		/* add displacement and UB */
		disps[1] = st_offsets[ndims-1];
		tmp_size = 1;
		for (i=ndims-2; i>=0; i--)
		{
			tmp_size *= array_of_gsizes[i+1];
			disps[1] += tmp_size*st_offsets[i];
		}
	}

	disps[1] *= orig_extent;

	disps[2] = orig_extent;
	for (i=0; i<ndims; i++) disps[2] *= array_of_gsizes[i];
		
	disps[0] = 0;
	blklens[0] = blklens[1] = blklens[2] = 1;
	types[0] = MPI_LB;
	types[1] = type_new;
	types[2] = MPI_UB;
	
	mpi_errno = MPID_Type_struct(3, blklens, disps, types, newtype);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	/* --END ERROR HANDLING-- */

	NMPI_Type_free(mpi_errno, (&type_new));

	/* at this point we have the new type, and we've cleaned up any
	 * intermediate types created in the process.  we just need to save
	 * all our contents/envelope information.
	 */

	/* Save contents */
	MPIU_CHKLMEM_MALLOC_ORJUMP(ints, int *, (4 * ndims + 4) * sizeof(int), mpi_errno, L"content description");

	ints[0] = size;
	ints[1] = rank;
	ints[2] = ndims;

	for (i=0; i < ndims; i++)
	{
		ints[i + 3] = array_of_gsizes[i];
	}
	for (i=0; i < ndims; i++)
	{
		ints[i + ndims + 3] = array_of_distribs[i];
	}
	for (i=0; i < ndims; i++)
	{
		ints[i + 2*ndims + 3] = array_of_dargs[i];
	}
	for (i=0; i < ndims; i++)
	{
		ints[i + 3*ndims + 3] = array_of_psizes[i];
	}
	ints[4*ndims + 3] = order;
	MPID_Datatype_get_ptr(*newtype, datatype_ptr);
	mpi_errno = MPID_Datatype_set_contents(datatype_ptr, MPI_COMBINER_DARRAY, 4*ndims + 4, 0, 1, ints, NULL, &oldtype);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	/* --END ERROR HANDLING-- */

	/* ... end of body of routine ... */
	
  fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CREATE_DARRAY);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_create_darray",
			L"**mpi_type_create_darray %d %d %d %p %p %p %p %d %D %p", size, rank, ndims, array_of_gsizes,
			array_of_distribs, array_of_dargs, array_of_psizes, order, oldtype, newtype);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build the MPI routines. */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Type_contiguous
#define MPI_Type_contiguous PMPI_Type_contiguous
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Type_contiguous

/*@
	MPI_Type_contiguous - Creates a contiguous datatype

Input Parameters:
+ count - replication count (nonnegative integer) 
- oldtype - old datatype (handle) 

Output Parameter:
. newtype - new datatype (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_COUNT
.N MPI_ERR_EXHAUSTED
@*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_contiguous(int count, MPI_Datatype old_type, MPI_Datatype *new_type_p)
{
	static const wchar_t FCNAME[] = L"MPI_Type_contiguous";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *new_dtp;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_TYPE_CONTIGUOUS);

	MPIR_ERRTEST_INITIALIZED_ORDIE();
	
	MPIU_THREAD_SINGLE_CS_ENTER("datatype");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_TYPE_CONTIGUOUS);

#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPID_Datatype *datatype_ptr = NULL;

			/* MPIR_ERRTEST_XXX macros defined in mpiimpl.h */
			MPIR_ERRTEST_COUNT(count, mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			
			MPIR_ERRTEST_DATATYPE(old_type, L"datatype", mpi_errno);
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
			
			if (HANDLE_GET_KIND(old_type) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(old_type, datatype_ptr);
				MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
			}
			if (mpi_errno != MPI_SUCCESS) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* ... body of routine ... */
	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_type_contiguous %d %D %p", count, old_type, new_type_p); }
	
	mpi_errno = MPID_Type_contiguous(count, old_type, new_type_p);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	MPID_Datatype_get_ptr(*new_type_p, new_dtp);
	mpi_errno = MPID_Datatype_set_contents(new_dtp, MPI_COMBINER_CONTIGUOUS, 1 /* ints (count) */, 0, 1, &count, NULL, &old_type);

	if (mpi_errno != MPI_SUCCESS) goto fn_fail;

	/* ... end of body of routine ... */
	
  fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_TYPE_CONTIGUOUS);
	MPIU_THREAD_SINGLE_CS_EXIT("datatype");
	return mpi_errno;

  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
	mpi_errno = MPIR_Err_create_code(
		mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_type_contiguous",
		L"**mpi_type_contiguous %d %D %p", count, old_type, new_type_p);
	}
#   endif
	mpi_errno = MPIR_Err_return_comm( NULL, FCNAME, mpi_errno );
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
