/*  (C) 2009 by Deino Software. */

#include "mpidimpl.h"
#include "mpidi_ch3_impl.h"
#include "pmi.h"
#include <stdio.h>
#include <stdarg.h>
#include <limits.h>
/* Needed for isdigit */
#include <ctype.h>

static int MPIDI_CH3I_PMI_Abort(int exit_code, const wchar_t *error_msg);

static wchar_t processorName[MPI_MAX_PROCESSOR_NAME];
static int  setProcessorName = 0;
static int  processorNameLen = -1;

static inline void setupProcessorName( void );

static int setupPortFunctions = 1;
static int MPIDI_Open_port(MPID_Info *, wchar_t *);

/* Define the functions that are used to implement the port operations */
static MPIDI_PortFns portFns = { MPIDI_Open_port, 0, 
								 MPIDI_Comm_accept, 
								 MPIDI_Comm_connect };

static int setupRMAFunctions = 1;
/* Define the functions that are used to implement the RMA operations */
static MPIDI_RMAFns RMAFns = { MPIDI_Win_create, MPIDI_Win_free,
							   MPIDI_Put, MPIDI_Get, MPIDI_Accumulate,
							   MPIDI_Win_fence, MPIDI_Win_post, 
							   MPIDI_Win_start, MPIDI_Win_complete,
							   MPIDI_Win_wait, MPIDI_Win_lock,
							   MPIDI_Win_unlock, MPIDI_Alloc_mem, MPIDI_Free_mem };

MPIDI_msg_sz_t MPIDI_CH3I_Eager_max_msg_size;
MPIDI_CH3I_Process_t MPIDI_CH3I_Process = {NULL};
volatile unsigned int MPIDI_CH3I_progress_completion_count = 0;
MPIDU_Sock_set_t MPIDI_CH3I_sock_set = NULL;

#define MPIDI_CH3I_PORT_NAME_TAG_KEY L"tag"

/*S
 * MPIDI_VCRT - virtual connection reference table
 *
 * handle - this element is not used, but exists so that we may use the 
 * MPIU_Object routines for reference counting
 *
 * ref_count - number of references to this table
 *
 * vcr_table - array of virtual connection references
 S*/
typedef struct MPIDI_VCRT
{
	int handle;
	volatile int ref_count;
	int size;
	MPIDI_VC_t * vcr_table[1];
}
MPIDI_VCRT_t;

/* What is the arrangement of VCRT and VCR and VC? 
   
   Each VC (the virtual connection itself) is refered to by a reference 
   (pointer) or VCR.  
   Each communicator has a VCRT, which is nothing more than a 
   structure containing a count (size) and an array of pointers to 
   virtual connections (as an abstraction, this could be a sparse
   array, allowing a more scalable representation on massively 
   parallel systems).

 */

static int MPIDI_CH3U_VC_FinishPending( MPIDI_VCRT_t *vcrt );

/*
 * The lpid counter counts new processes that this process knows about.
 */
static int lpid_counter = 0;

static MPIDI_PG_t * MPIDI_PG_list = NULL;
static MPIDI_PG_t * MPIDI_PG_iterator_next = NULL;
static MPIDI_PG_Compare_ids_fn_t MPIDI_PG_Compare_ids_fn;
static MPIDI_PG_Destroy_fn_t MPIDI_PG_Destroy_fn;

/* Key track of the process group corresponding to the MPI_COMM_WORLD 
   of this process */
static MPIDI_PG_t *pg_world = NULL;

#define MPIDI_MAX_KVS_KEY_LEN	  256

/* For process groups that are not our MPI_COMM_WORLD, store the connection
   information in an array of strings.  These routines and structure
   implement the access to this information. */
typedef struct
{
	int	 toStringLen;			/* Length needed to encode this connection info */
	wchar_t ** connStrings;		/* pointer to an array, indexed by rank, containing connection information */
} MPIDI_ConnInfo;

#ifdef USE_MPIU_DBG_PRINT_VC
wchar_t *MPIU_DBG_parent_str = L"?";
#endif

static int InitPG( int *argc_p, wchar_t ***argv_p, int *has_args, int *has_env, int *has_parent, int *pg_rank_p, MPIDI_PG_t **pg_p );
static int MPIDI_CH3I_PG_Compare_ids(const void * id1, const void * id2);
static int MPIDI_CH3I_PG_Destroy(MPIDI_PG_t * pg );

MPIDI_Process_t MPIDI_Process = { NULL };
MPIDI_CH3_Functions_t MPIDI_CH3 = { NULL };
MPIDI_CH3U_SRBuf_element_t * MPIDI_CH3U_SRBuf_pool = NULL;

/* FIXME: This routine *or* MPI_Abort should provide abort callbacks, similar to the support in MPI_Finalize */
#undef FUNCNAME
#define FUNCNAME MPID_Abort
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Abort(MPID_Comm * comm, int mpi_errno, int exit_code, const wchar_t *error_msg)
{
	int rank;
	wchar_t msg[MPI_MAX_ERROR_STRING] = L"";
	wchar_t error_str[MPI_MAX_ERROR_STRING + 100];
	/* This function cannot be profiled because it does not exit
	MPIDI_STATE_DECL(MPID_STATE_MPID_ABORT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_ABORT);
	*/

	if (error_msg == NULL)
	{
		/* Create a default error message */
		error_msg = error_str;
		/* FIXME: Do we want the rank of the input communicator here 
		   or the rank of comm world?  The message gives the rank but not the 
		   communicator, so using other than the rank in comm world does not 
		   identify the process, as the message suggests */
		if (comm)
		{
			rank = comm->rank;
		}
		else
		{
			if (MPIR_Process.comm_world != NULL)
			{
				rank = MPIR_Process.comm_world->rank;
			}
			else
			{
				rank = -1;
			}
		}

		if (mpi_errno != MPI_SUCCESS)
		{
			MPIR_Err_get_string(mpi_errno, msg, MPI_MAX_ERROR_STRING, NULL);
			/* FIXME: Not internationalized */
			MPIU_Snprintf(error_str, sizeof(error_str)/sizeof(wchar_t), _TRUNCATE, L"internal ABORT - process %d: %s", rank, msg);
		}
		else
		{
			/* FIXME: Not internationalized */
			MPIU_Snprintf(error_str, sizeof(error_str)/sizeof(wchar_t), _TRUNCATE, L"internal ABORT - process %d", rank);
		}
	}

	MPIDI_CH3I_PMI_Abort(exit_code, error_msg);

	/*MPIU_UNREACHABLE_CODE();*/ /* Can't use this macro since it calls MPID_Abort leading to infinite recursion */

	/* ch3_abort should not return but if it does, exit here */
	ExitProcess(exit_code);

	/*MPIDI_FUNC_EXIT(MPID_STATE_MPID_ABORT);*/
	return MPI_ERR_INTERN;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_PMI_Abort
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_PMI_Abort(int exit_code, const wchar_t *error_msg)
{
	/*
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PMI_ABORT);
	
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PMI_ABORT);
	*/

	/* FIXME: What is the scope for PMI_Abort?  Shouldn't it be one or more
	   process groups?  Shouldn't abort of a communicator abort either the
	   process groups of the communicator or only the current process?
	   Should PMI_Abort have a parameter for which of these two cases to
	   perform? */
	PMI_AbortW(exit_code, error_msg);

	/* if abort returns for some reason, exit here */

	/* FIXME: Why is the message only printed if PMI_Abort fails to return?
	   The simple_pmi.c implementation of PMI_Abort performs an exit; even a 
	   more sophisticated version is likely to cause this process to exit 
	   before it can do much more */
	   
	MPIU_Error_printf(L"%s", error_msg);
	fflush(stderr);
	ExitProcess(exit_code);

	/*
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PMI_ABORT);
	return MPI_ERR_INTERN;
	*/
}

#undef FUNCNAME
#define FUNCNAME MPID_Cancel_recv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Cancel_recv(MPID_Request * rreq)
{
	MPIDI_STATE_DECL(MPID_STATE_MPID_CANCEL_RECV);
	
	MPIDI_FUNC_ENTER(MPID_STATE_MPID_CANCEL_RECV);
	
	MPIU_Assert(rreq->kind == MPID_REQUEST_RECV);
	
	if (MPIDI_CH3U_Recvq_DP(rreq))
	{
		MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"request 0x%08x cancelled", rreq->handle);
		rreq->status.cancelled = TRUE;
		rreq->status.count = 0;
		MPID_Request_set_completed(rreq);
		MPID_Request_release(rreq);
	}
	else
	{
		MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"request 0x%08x already matched, unable to cancel", rreq->handle);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_CANCEL_RECV);
	return MPI_SUCCESS;
}


/* FIXME: This should call a channel-provided routine to deliver the 
   cancel message, once the code decides that the request can still
   be cancelled */

#undef FUNCNAME
#define FUNCNAME MPID_Cancel_send
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Cancel_send(MPID_Request * sreq)
{
	MPIDI_VC_t * vc;
	int proto;
	int flag;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_CANCEL_SEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_CANCEL_SEND);
	
	MPIU_Assert(sreq->kind == MPID_REQUEST_SEND);

	MPIDI_Request_cancel_pending(sreq, &flag);
	if (flag)
	{
		goto fn_exit;
	}

	/*
	 * FIXME: user requests returned by MPI_Ibsend() have a NULL comm pointer and no pointer to the underlying communication
	 * request.  For now, we simply fail to cancel the request.  In the future, we should add a new request kind to indicate that
	 * the request is a BSEND.  Then we can properly cancel the request, much in the way we do persistent requests.
	 */
	if (sreq->comm == NULL)
	{
		goto fn_exit;
	}

	MPIDI_Comm_get_vc(sreq->comm, sreq->dev.match.rank, &vc);

	proto = MPIDI_Request_get_msg_type(sreq);

	if (proto == MPIDI_REQUEST_SELF_MSG)
	{
		MPID_Request * rreq;
		
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"attempting to cancel message sent to self");
		
		rreq = MPIDI_CH3U_Recvq_FDU(sreq->handle, &sreq->dev.match);
		if (rreq)
		{
			MPIU_Assert(rreq->partner_request == sreq);
			
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"send-to-self cancellation successful, sreq=0x%08x, rreq=0x%08x", sreq->handle, rreq->handle));
			
			MPIU_Object_set_ref(rreq, 0);
			MPIDI_CH3_Request_destroy(rreq);
			
			sreq->status.cancelled = TRUE;
			/* no other thread should be waiting on sreq, so it is safe to reset ref_count and cc */
			sreq->cc = 0;
			MPIU_Object_set_ref(sreq, 1);
		}
		else
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"send-to-self cancellation failed, sreq=0x%08x, rreq=0x%08x", sreq->handle, rreq->handle));
		}
		
		goto fn_exit;
	}

	/* Check to see if the send is still in the send queue.  If so, remove it, mark the request and cancelled and complete, and
	   release the device's reference to the request object.  QUESTION: what is the right interface for MPIDI_CH3_Send_cancel()?
	   It needs to be able to cancel requests to send a RTS packet for this request.  Perhaps we can use the partner request
	   field to track RTS requests. */
	{
		int cancelled;
		
		if (proto == MPIDI_REQUEST_RNDV_MSG)
		{
			MPID_Request * rts_sreq;
			/* The cancellation of the RTS request needs to be atomic through the destruction of the RTS request to avoid
			   conflict with release of the RTS request if the CTS is received (see handling of a rendezvous CTS packet in
			   MPIDI_CH3U_Handle_recv_pkt()).  MPID_Request_fetch_and_clear_rts_sreq() is used to gurantee that atomicity. */
			MPIDI_Request_fetch_and_clear_rts_sreq(sreq, &rts_sreq);
			if (rts_sreq != NULL) 
			{
				cancelled = FALSE;
				/*mpi_errno = MPIDI_CH3_Cancel_send(vc, rts_sreq, &cancelled);*/
				
				/* since we attempted to cancel a RTS request, then we are responsible for releasing that request */
				MPID_Request_release(rts_sreq);

				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|cancelrndv", 0);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
				
				if (cancelled)
				{
					sreq->status.cancelled = TRUE;
					/* no other thread should be waiting on sreq, so it is safe to reset ref_count and cc */
					sreq->cc = 0;
					MPIU_Object_set_ref(sreq, 1);
					goto fn_exit;
				}
			}
		}
		else
		{
			cancelled = FALSE;
			/*mpi_errno = MPIDI_CH3_Cancel_send(vc, sreq, &cancelled);*/
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|canceleager", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			if (cancelled)
			{
				sreq->status.cancelled = TRUE;
				/* no other thread should be waiting on sreq, so it is safe to reset ref_count and cc */
				sreq->cc = 0;
				MPIU_Object_set_ref(sreq, 1);
				goto fn_exit;
			}
		}
	}

	/* Part or all of the message has already been sent, so we need to send a 
	   cancellation request to the receiver in an attempt
	   to catch the message before it is matched. */
	{
		int was_incomplete;
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_cancel_send_req_t * const csr_pkt = &upkt.cancel_send_req;
		MPID_Request * csr_sreq;
		
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending cancel request to %d for 0x%08x", sreq->dev.match.rank, sreq->handle));

		/* The completion counter and reference count are incremented to keep 
		   the request around long enough to receive a
		   response regardless of what the user does (free the request before 
		   waiting, etc.). */
		MPIDI_CH3U_Request_increment_cc(sreq, &was_incomplete);
		if (!was_incomplete)
		{
			/* The reference count is incremented only if the request was 
			   complete before the increment. */
			MPIR_Request_add_ref(sreq);
		}

		MPIDI_Pkt_init(csr_pkt, MPIDI_CH3_PKT_CANCEL_SEND_REQ);
		csr_pkt->match.rank = (int16_t)sreq->comm->rank;
		csr_pkt->match.tag = sreq->dev.match.tag;
		csr_pkt->match.context_id = sreq->dev.match.context_id;
		csr_pkt->sender_req_id = sreq->handle;
		
		mpi_errno = MPIDI_CH3.iStartMsg(vc, csr_pkt, sizeof(*csr_pkt), &csr_sreq);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|cancelreq", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		if (csr_sreq != NULL)
		{
			MPID_Request_release(csr_sreq);
		}
	}
	
	/* FIXME: if send cancellation packets are allowed to arrive out-of-order 
	   with respect to send packets, then we need to
	   timestamp send and cancel packets to insure that a cancellation request 
	   does not bypass the send packet to be cancelled
	   and erroneously cancel a previously sent message with the same request 
	   handle. */
	/* FIXME: A timestamp is more than is necessary; a message sequence number
	   should be adequate. */

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_CANCEL_SEND);
	return mpi_errno;
}

/*
 * Handler routines called when cancel send packets arrive
 */

int MPIDI_CH3_PktHandler_CancelSendReq( MPIDI_VC_t *vc, MPIDI_CH3_Pkt_t *pkt, MPID_Request **rreqp )
{
	MPIDI_CH3_Pkt_cancel_send_req_t * req_pkt = &pkt->cancel_send_req;
	MPID_Request * rreq;
	int ack;
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_cancel_send_resp_t * resp_pkt = &upkt.cancel_send_resp;
	MPID_Request * resp_sreq;
	int mpi_errno = MPI_SUCCESS;
	
	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"received cancel send req pkt, sreq=0x%08x, rank=%d, tag=%d, context=%d", req_pkt->sender_req_id, req_pkt->match.rank, req_pkt->match.tag, req_pkt->match.context_id));
			
	rreq = MPIDI_CH3U_Recvq_FDU(req_pkt->sender_req_id, &req_pkt->match);
	if (rreq != NULL)
	{
		MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"message cancelled");
		if (MPIDI_Request_get_msg_type(rreq) == MPIDI_REQUEST_EAGER_MSG && rreq->dev.recv_data_sz > 0)
		{
			MPIU_Free(rreq->dev.tmpbuf);
		}
		MPID_Request_release(rreq);
		ack = TRUE;
	}
	else
	{
		MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"unable to cancel message");
		ack = FALSE;
	}
	
	MPIDI_Pkt_init(resp_pkt, MPIDI_CH3_PKT_CANCEL_SEND_RESP);
	resp_pkt->sender_req_id = req_pkt->sender_req_id;
	resp_pkt->ack = ack;
	mpi_errno = MPIDI_CH3.iStartMsg(vc, resp_pkt, sizeof(*resp_pkt), &resp_sreq);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|cancelresp");
	}
	if (resp_sreq != NULL)
	{
		MPID_Request_release(resp_sreq);
	}
	
	*rreqp = NULL;

 fn_fail:
	return mpi_errno;
}

int MPIDI_CH3_PktHandler_CancelSendResp( MPIDI_VC_t *vc, MPIDI_CH3_Pkt_t *pkt, MPID_Request **rreqp )
{
	MPIDI_CH3_Pkt_cancel_send_resp_t * resp_pkt = &pkt->cancel_send_resp;
	MPID_Request * sreq;
	
	MPIU_UNREFERENCED_ARG(vc);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"received cancel send resp pkt, sreq=0x%08x, ack=%d", resp_pkt->sender_req_id, resp_pkt->ack));
			
	MPID_Request_get_ptr(resp_pkt->sender_req_id, sreq);
	
	if (resp_pkt->ack)
	{
		sreq->status.cancelled = TRUE;
		
		if (MPIDI_Request_get_msg_type(sreq) == MPIDI_REQUEST_RNDV_MSG ||
			MPIDI_Request_get_type(sreq) == MPIDI_REQUEST_TYPE_SSEND)
		{
			int cc;
			
			/* decrement the CC one additional time for the CTS/sync ack that 
			   is never going to arrive */
			MPIDI_CH3U_Request_decrement_cc(sreq, &cc);
		}
				
		MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"message cancelled");
	}
	else
	{
		sreq->status.cancelled = FALSE; 
		MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"unable to cancel message");
	}
	
	MPIDI_CH3U_Request_complete(sreq);
	
	*rreqp = NULL;

	return MPI_SUCCESS;
}

/*
 * Define the routines that can print out the cancel packets if 
 * debugging is enabled.
 */
#ifdef MPICH_DBG_OUTPUT
int MPIDI_CH3_PktPrint_CancelSendReq( FILE *fp, MPIDI_CH3_Pkt_t *pkt )
{
	MPIU_DBG_PRINTF((L" type ......... CANCEL_SEND\n"));
	MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->cancel_send_req.sender_req_id));
	MPIU_DBG_PRINTF((L" context_id ... %d\n", pkt->cancel_send_req.match.context_id));
	MPIU_DBG_PRINTF((L" tag .......... %d\n", pkt->cancel_send_req.match.tag));
	MPIU_DBG_PRINTF((L" rank ......... %d\n", pkt->cancel_send_req.match.rank));

	return MPI_SUCCESS;
}

int MPIDI_CH3_PktPrint_CancelSendResp( FILE *fp, MPIDI_CH3_Pkt_t *pkt )
{
	MPIU_DBG_PRINTF((L" type ......... CANCEL_SEND_RESP\n"));
	MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->cancel_send_resp.sender_req_id));
	MPIU_DBG_PRINTF((L" ack .......... %d\n", pkt->cancel_send_resp.ack));
	
	return MPI_SUCCESS;
}
#endif

/*@
   MPID_Comm_disconnect - Disconnect a communicator 

   Arguments:
.  comm_ptr - communicator

   Notes:

.N Errors
.N MPI_SUCCESS
@*/
#undef FUNCNAME
#define FUNCNAME MPID_Comm_disconnect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Comm_disconnect(MPID_Comm *comm_ptr)
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPID_COMM_DISCONNECT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_COMM_DISCONNECT);

	/* Before releasing the communicator, we need to ensure that all VCs are
	   in a stable state.  In particular, if a VC is still in the process of
	   connecting, complete the connection before tearing it down */
	/* FIXME: How can we get to a state where we are still connecting a VC but
	   the MPIR_Comm_release will find that the ref count decrements to zero 
	   (it may be that some operation fails to increase/decrease the reference 
	   count.  A patch could be to increment the reference count while 
	   connecting, then decrement it.  But the increment in the reference 
	   count should come 
	   from the step that caused the connection steps to be initiated.  
	   Possibility: if the send queue is not empty, the ref count should
	   be higher.  */
	/* FIXME: This doesn't work yet */
	/*
	mpi_errno = MPIDI_CH3U_Comm_FinishPending( comm_ptr );
	*/

	/* it's more than a comm_release, but ok for now */
	/* FIXME: Describe what more might be required */
	/* MPIU_PG_Printall( stdout ); */
	mpi_errno = MPIR_Comm_release(comm_ptr, 1);
	/* If any of the VCs were released by this Comm_release, wait
	 for those close operations to complete */
	MPIDI_CH3U_VC_WaitForClose();
	/* MPIU_PG_Printall( stdout ); */

	MPIU_PrintCurrentMemoryUsage();

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_COMM_DISCONNECT);
	return mpi_errno;
}


/* FIXME: Correct description of function */
/*@
   MPID_Comm_spawn_multiple - 

   Input Arguments:
+  int count - count
.  wchar_t *array_of_commands[] - commands
.  wchar_t* *array_of_argv[] - arguments
.  int array_of_maxprocs[] - maxprocs
.  MPI_Info array_of_info[] - infos
.  int root - root
-  MPI_Comm comm - communicator

   Output Arguments:
+  MPI_Comm *intercomm - intercommunicator
-  int array_of_errcodes[] - error codes

   Notes:

.N Errors
.N MPI_SUCCESS
@*/
#undef FUNCNAME
#define FUNCNAME MPID_Comm_spawn_multiple
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Comm_spawn_multiple(int count, wchar_t *array_of_commands[], 
							 wchar_t ** array_of_argv[], int array_of_maxprocs[],
							 MPID_Info * array_of_info_ptrs[], int root, 
							 MPID_Comm * comm_ptr, MPID_Comm ** intercomm,
							 int array_of_errcodes[]) 
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_COMM_SPAWN_MULTIPLE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_COMM_SPAWN_MULTIPLE);

	mpi_errno = MPIDI_Comm_spawn_multiple(count, array_of_commands, 
										  array_of_argv, array_of_maxprocs,
										  array_of_info_ptrs,
										  root, comm_ptr, intercomm, 
										  array_of_errcodes);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**fail");
	}
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_COMM_SPAWN_MULTIPLE);
	return mpi_errno;
}


/* FIXME: This routine needs to be factored into finalize actions per module,
   In addition, we should consider registering callbacks for those actions
   rather than direct routine calls.
 */

#undef FUNCNAME
#define FUNCNAME MPID_Finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Finalize(void)
{
	int mpi_errno = MPI_SUCCESS;
	MPIU_THREADPRIV_DECL;						/* Required for Nest_incr */
	MPIDI_STATE_DECL(MPID_STATE_MPID_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_FINALIZE);

	/*
	 * Wait for all posted receives to complete.  For now we are not doing 
	 * this since it will cause invalid programs to hang.
	 * The side effect of not waiting is that posted any source receives 
	 * may erroneous blow up.
	 *
	 * For now, we are placing a warning at the end of MPID_Finalize() to 
	 * inform the user if any outstanding posted receives exist.
	 */
	 /* FIXME: The correct action here is to begin a shutdown protocol
	  * that lets other processes know that this process is now
	  * in finalize.  
	  *
	  * Note that only requests that have been freed with MPI_Request_free
	  * are valid at this point; other pending receives can be ignored 
	  * since a valid program should wait or test for them before entering
	  * finalize.  
	  * 
	  * The easist fix is to allow an MPI_Barrier over comm_world (and 
	  * any connected processes in the MPI-2 case).  Once the barrier
	  * completes, all processes are in finalize and any remaining 
	  * unmatched receives will never be matched (by a correct program; 
	  * a program with a send in a separate thread that continues after
	  * some thread calls MPI_Finalize is erroneous).
	  * 
	  * Avoiding the barrier is hard.  Consider this sequence of steps:
	  * Send in-finalize message to all connected processes.  Include
	  * information on whether there are pending receives.
	  *   (Note that a posted receive with any source is a problem)
	  *   (If there are many connections, then this may take longer than
	  *   the barrier)
	  * Allow connection requests from anyone who has not previously
	  * connected only if there is an possible outstanding receive; 
	  * reject others with a failure (causing the source process to 
	  * fail).
	  * Respond to an in-finalize message with the number of posted receives
	  * remaining.  If both processes have no remaining receives, they 
	  * can both close the connection.
	  * 
	  * Processes with no pending receives and no connections can exit, 
	  * calling PMI_Finalize to let the process manager know that they
	  * are in a controlled exit.  
	  *
	  * Processes that still have open connections must then try to contact
	  * the remaining processes.
	  * 
	  */
	
/* commenting out the close protocol and simply using MPI_Barrier until 
   MPI_Comm_disconnect correctly disconnects all VCs */

	MPIU_THREADPRIV_GET;
	NMPI_Barrier(mpi_errno, (MPI_COMM_WORLD));
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = MPID_VCRT_Release(MPIR_Process.comm_self->vcrt, 0);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPID_Dev_comm_destroy_hook(MPIR_Process.comm_world);

	mpi_errno = MPID_VCRT_Release(MPIR_Process.comm_world->vcrt, 0);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* FIXME: The close actions should use the same code as the other connection close code */
	MPIDI_PG_Close_VCs();
	/*
	 * Wait for all VCs to finish the close protocol
	 */
	mpi_errno = MPIDI_CH3U_VC_WaitForClose();
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	mpi_errno = MPIDI_CH3.Finalize();
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	/* Tell the process group code that we're done with the process groups.
	   This will notify PMI (with PMI_Finalize) if necessary.  It
	   also frees al PG structures, including the PG for COMM_WORLD, whose 
	   pointer is also saved in MPIDI_Process.my_pg */
	mpi_errno = MPIDI_PG_Finalize();
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	MPIDI_CH3_FreeParentPort();

	/* Release any SRbuf pool storage */
	if (MPIDI_CH3U_SRBuf_pool)
	{
		MPIDI_CH3U_SRBuf_element_t *p, *pNext;
		p = MPIDI_CH3U_SRBuf_pool;
		while (p)
		{
			pNext = p->next;
			MPIU_Free(p);
			p = pNext;
		}
		MPIDI_CH3U_SRBuf_pool = NULL;
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_FINALIZE);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}


/*
 * MPID_Get_universe_size - Get the universe size from the process manager
 *
 * Notes: The ch3 device requires that the PMI routines are used to 
 * communicate with the process manager.  If a channel wishes to 
 * bypass the standard PMI implementations, it is the responsibility of the
 * channel to provide an implementation of the PMI routines.
 */
#undef FUNCNAME
#define FUNCNAME MPID_Get_universe_size
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Get_universe_size(int  * universe_size)
{
	int mpi_errno = MPI_SUCCESS;
	int pmi_errno = PMI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_GET_UNIVERSE_SIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_GET_UNIVERSE_SIZE);

	pmi_errno = PMI_Get_universe_size(universe_size);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno, MPI_ERR_OTHER, L"**pmi_get_universe_size", L"**pmi_get_universe_size %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}
	if (*universe_size < 0)
	{
		*universe_size = MPIR_UNIVERSE_SIZE_NOT_AVAILABLE;
	}
	
fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_GET_UNIVERSE_SIZE);
	return mpi_errno;

	/* --BEGIN ERROR HANDLING-- */
fn_fail:
	*universe_size = MPIR_UNIVERSE_SIZE_NOT_AVAILABLE;
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/*
 * MPID_Get_processor_name()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Get_processor_name
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Get_processor_name(wchar_t * name, int namelen, int * resultlen)
{
	MPIDI_STATE_DECL(MPID_STATE_MPID_GET_PROCESSOR_NAME);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_GET_PROCESSOR_NAME);
	/* FIXME: Make thread safe */
	if (!setProcessorName)
	{
		setupProcessorName();
		setProcessorName = 1;
	}
	if (processorNameLen > 0)
	{
		/* MPIU_Wcsncpy only copies until (and including) the null, 
		   unlink wcsncpy, it does not blank pad.  This is a good thing
		   here, because users don't always allocated MPI_MAX_PROCESSOR_NAME
		   characters */
		MPIU_Wcsncpy(name, processorName, namelen );
		if (resultlen) 
			*resultlen = processorNameLen;
	}
	/* --BEGIN ERROR HANDLING-- */
	else
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPID_GET_PROCESSOR_NAME);
		return MPI_ERR_UNKNOWN;
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_GET_PROCESSOR_NAME);
	return MPI_SUCCESS;
}

/* Here we define an internal routine to get the processor name, based on 
   which system or facilities are available to us */

#undef FUNCNAME
#define FUNCNAME setupProcessorName
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void setupProcessorName( void )
{
	DWORD size = MPI_MAX_PROCESSOR_NAME;
	MPIDI_STATE_DECL(MPID_STATE_SETUPPROCESSORNAME);

	MPIDI_FUNC_ENTER(MPID_STATE_SETUPPROCESSORNAME);

	/* Use the fully qualified name instead of the short name because the SSPI security functions require the full name */
	if (GetComputerNameExW(ComputerNameDnsFullyQualified, processorName, &size))
	{
		processorNameLen = (int)wcslen(processorName);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_SETUPPROCESSORNAME);
}

#undef FUNCNAME
#define FUNCNAME MPID_Iprobe
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Iprobe(int source, int tag, MPID_Comm * comm, int context_offset, int * flag, MPI_Status * status)
{
	const int context = comm->recvcontext_id + context_offset;
	int found = 0;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_IPROBE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_IPROBE);

	if (source == MPI_PROC_NULL)
	{
		MPIR_Status_set_procnull(status);
		/* We set the flag to true because an MPI_Recv with this rank will return immediately */
		*flag = TRUE;
		goto fn_exit;
	}

	/* FIXME: The routine CH3U_Recvq_FU is used only by the probe functions;
	   it should atomically return the flag and status rather than create 
	   a request.  Note that in some cases it will be possible to 
	   atomically query the unexpected receive list (which is what the
	   probe routines are for). */
	found = MPIDI_CH3U_Recvq_FU( source, tag, context, status );
	if (!found)
	{
		/* Always try to advance progress before returning failure
		   from the iprobe test.  */
		/* FIXME: It would be helpful to know if the Progress_poke
		   operation causes any change in state; we could then avoid
		   a second test of the receive queue if we knew that nothing
		   had changed */
		mpi_errno = MPID_Progress_poke();
		found = MPIDI_CH3U_Recvq_FU( source, tag, context, status );
	}
		
	*flag = found;

 fn_exit:	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_IPROBE);
	return mpi_errno;
}


/*
 * MPID_Isend()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Irecv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Irecv(void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPID_Request * rreq;
	int found;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_IRECV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_IRECV);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"rank=%d, tag=%d, context=%d", rank, tag, comm->recvcontext_id + context_offset));
	
	if (rank == MPI_PROC_NULL)
	{
		rreq = MPID_Request_create();
		if (rreq != NULL)
		{
			MPIU_Object_set_ref(rreq, 1);
			rreq->cc = 0;
			rreq->kind = MPID_REQUEST_RECV;
			MPIR_Status_set_procnull(&rreq->status);
			/* FIXME: What is this code for?  Why should the behavior of
			   this routine depend on whether the "assert" routine is
			   enabled? As a reminder, if NDEBUG is defined, assert 
			   generates no code; thus, this block of code (which even 
			   adds a reference to a comm!) is used only if asserts are
			   enabled.  It is a bad idea to piggy back this kind of 
			   code on NDEBUG.  In addition, a justification for this 
			   code is needed, since NDEBUG should normally not change the 
			   actions performed with correct code. */
#			if !defined(NDEBUG)
			{
				rreq->comm = comm;
				MPIR_Comm_add_ref(comm);
				rreq->dev.match.rank = rank;
				rreq->dev.match.tag = tag;
				rreq->dev.match.context_id = comm->context_id + context_offset;
				rreq->dev.user_buf = buf;
				rreq->dev.user_count = count;
				rreq->dev.datatype = datatype;
			}
#			endif
		}
		else
		{
			/* --BEGIN ERROR HANDLING-- */
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			/* --END ERROR HANDLING-- */
		}
		goto fn_exit;
	}

	rreq = MPIDI_CH3U_Recvq_FDU_or_AEP(rank, tag, comm->recvcontext_id + context_offset, &found);
	/* --BEGIN ERROR HANDLING-- */
	if (rreq == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

	rreq->comm = comm;
	MPIR_Comm_add_ref(comm);
	rreq->dev.user_buf = buf;
	rreq->dev.user_count = count;
	rreq->dev.datatype = datatype;

	if (found)
	{
		MPIDI_VC_t * vc;
		
		/* Message was found in the unexepected queue */
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"request found in unexpected queue");

		MPIDI_Comm_get_vc(comm, rreq->dev.match.rank, &vc);
		
		if (MPIDI_Request_get_msg_type(rreq) == MPIDI_REQUEST_EAGER_MSG)
		{
			int recv_pending;
			
			/* This is an eager message */
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"eager message in the request");
			
			/* If this is a eager synchronous message, then we need to send an 
			   acknowledgement back to the sender. */
			if (MPIDI_Request_get_sync_send_flag(rreq))
			{
				MPIDI_CH3_Pkt_t upkt;
				MPIDI_CH3_Pkt_eager_sync_ack_t * const esa_pkt = &upkt.eager_sync_ack;
				MPID_Request * esa_req;
					
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending eager sync ack");
				MPIDI_Pkt_init(esa_pkt, MPIDI_CH3_PKT_EAGER_SYNC_ACK);
				esa_pkt->sender_req_id = rreq->dev.sender_req_id;
				mpi_errno = MPIDI_CH3.iStartMsg(vc, esa_pkt, sizeof(*esa_pkt), &esa_req);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
				if (esa_req != NULL)
				{
					MPID_Request_release(esa_req);
				}
			}

			MPIDI_Request_recv_pending(rreq, &recv_pending);
			if (!recv_pending)
			{
				/* All of the data has arrived, we need to copy the data and then free the buffer and the request. */
				if (rreq->dev.recv_data_sz > 0)
				{
					MPIDI_CH3U_Request_unpack_uebuf(rreq);
					MPIU_Free(rreq->dev.tmpbuf);
				}

				mpi_errno = rreq->status.MPI_ERROR;
				goto fn_exit;
			}
			else
			{
				/* The data is still being transfered across the net.  We'll leave it to the progress engine to handle once the
				   entire message has arrived. */
				if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
				{
					MPID_Datatype_get_ptr(datatype, rreq->dev.datatype_ptr);
					MPID_Datatype_add_ref(rreq->dev.datatype_ptr);
				}
			
			}
		}
		else if (MPIDI_Request_get_msg_type(rreq) == MPIDI_REQUEST_RNDV_MSG)
		{
			/* A rendezvous request-to-send (RTS) message has arrived.  We need to send a CTS message to the remote process. */

			mpi_errno = MPIDI_CH3U_Post_data_receive(found, &rreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code (mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_RNDV_REQ_TO_SEND");
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			mpi_errno = MPIDI_CH3.iStartRndvTransfer (vc, rreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code (mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(datatype, rreq->dev.datatype_ptr);
				MPID_Datatype_add_ref(rreq->dev.datatype_ptr);
			}
		}
		else if (MPIDI_Request_get_msg_type(rreq) == MPIDI_REQUEST_SELF_MSG)
		{
			MPID_Request * const sreq = rreq->partner_request;

			if (sreq != NULL)
			{
				MPIDI_msg_sz_t data_sz;
				
				MPIDI_CH3U_Buffer_copy(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype, &sreq->status.MPI_ERROR,
									   buf, count, datatype, &data_sz, &rreq->status.MPI_ERROR);
				rreq->status.count = (int)data_sz;
				MPID_Request_set_completed(sreq);
				MPID_Request_release(sreq);
			}
			else
			{
				/* The sreq is missing which means an error occurred.  rreq->status.MPI_ERROR should have been set when the
				   error was detected. */
			}
			
			/* no other thread can possibly be waiting on rreq, so it is safe to reset ref_count and cc */
			rreq->cc = 0;
			MPIU_Object_set_ref(rreq, 1);
		}
		else
		{
			/* --BEGIN ERROR HANDLING-- */
			MPID_Request_release(rreq);
			rreq = NULL;
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|badmsgtype", L"**ch3|badmsgtype %d", MPIDI_Request_get_msg_type(rreq));
			goto fn_exit;
			/* --END ERROR HANDLING-- */
		}
	}
	else
	{
		/* Message has yet to arrived.  The request has been placed on the list of posted receive requests and populated with
		   information supplied in the arguments. */
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"request allocated in posted queue");
		
		if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
		{
			MPID_Datatype_get_ptr(datatype, rreq->dev.datatype_ptr);
			MPID_Datatype_add_ref(rreq->dev.datatype_ptr);
		}

		rreq->dev.recv_pending_count = 1;
	}

  fn_exit:
	*request = rreq;
	MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE,L"request allocated, handle=0x%08x", rreq->handle);

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_IRECV);
	return mpi_errno;
}


/* FIXME: HOMOGENEOUS SYSTEMS ONLY -- no data conversion is performed */

/*
 * MPID_Irsend()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Irsend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Irsend(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_ready_send_t * const ready_pkt = &upkt.ready_send;
	MPIDI_msg_sz_t data_sz;
	int dt_contig;
	MPI_Aint dt_true_lb;
	MPID_Datatype * dt_ptr;
	MPID_Request * sreq;
	MPID_IOV iov[MPID_IOV_LIMIT];
	MPIDI_VC_t * vc;
	int mpi_errno = MPI_SUCCESS;	
	MPIDI_STATE_DECL(MPID_STATE_MPID_IRSEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_IRSEND);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"rank=%d, tag=%d, context=%d", rank, tag, comm->context_id + context_offset));
	
	if (rank == comm->rank && comm->comm_kind != MPID_INTERCOMM)
	{
		mpi_errno = MPIDI_Isend_self(buf, count, datatype, rank, tag, comm, context_offset, MPIDI_REQUEST_TYPE_RSEND, &sreq);
		goto fn_exit;
	}
	
	MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_RSEND);
	MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_EAGER_MSG);
	
	if (rank == MPI_PROC_NULL)
	{
		MPIU_Object_set_ref(sreq, 1);
		sreq->cc = 0;
		goto fn_exit;
	}
	
	MPIDI_Datatype_get_info(count, datatype, dt_contig, data_sz, dt_ptr, dt_true_lb);

	MPIDI_Comm_get_vc(comm, rank, &vc);
	
	MPIDI_Pkt_init(ready_pkt, MPIDI_CH3_PKT_READY_SEND);
	ready_pkt->match.rank = (int16_t)comm->rank;
	ready_pkt->match.tag = tag;
	ready_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
	ready_pkt->sender_req_id = MPI_REQUEST_NULL;
	ready_pkt->data_sz = data_sz;

	if (data_sz == 0)
	{
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending zero length message");

		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;

		mpi_errno = MPIDI_CH3.iSend(vc, sreq, ready_pkt, sizeof(*ready_pkt));
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		goto fn_exit;
	}
	
	iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)ready_pkt;
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*ready_pkt);

	if (dt_contig)
	{
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending contiguous ready-mode message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;

		/* FIXME: handle case where data_sz is greater than what can be stored in iov.MPID_IOV_LEN.  hand off to segment code? */
		iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char*) buf + dt_true_lb);
		iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;

		mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, 2);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
	}
	else
	{
		int iov_n;
			
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending non-contiguous ready-mode message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

		MPID_Segment_init(buf, count, datatype, &sreq->dev.segment, 0);
		sreq->dev.segment_first = 0;
		sreq->dev.segment_size = data_sz;
			
		iov_n = MPID_IOV_LIMIT - 1;
		mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &iov[1], &iov_n);
		if (mpi_errno == MPI_SUCCESS)
		{
			iov_n += 1;

			if (sreq->dev.ca != MPIDI_CH3_CA_COMPLETE)
			{
				sreq->dev.datatype_ptr = dt_ptr;
				MPID_Datatype_add_ref(dt_ptr);
			}
			
			mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, iov_n);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			/* --BEGIN ERROR HANDLING-- */
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
			goto fn_exit;
			/* --END ERROR HANDLING-- */
		}
	}
 
  fn_exit:
	*request = sreq;

	MPIU_DBG_STMT(CH3_OTHER,VERBOSE, { if (sreq != NULL) { MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE,L"request allocated, handle=0x%08x", sreq->handle); } } );
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_IRSEND);
	return mpi_errno;
}


/* FIXME: HOMOGENEOUS SYSTEMS ONLY -- no data conversion is performed */

/* FIXME: The routines MPID_Isend, MPID_Issend, MPID_Irsend are nearly 
   identical. It would be better if these did roughly:

   MPID_Irsend -> always eager send (with ready mode for error detection)
   MPID_Issend -> always rendezvous send
   MPID_Isend  -> chose eager/rendezvous based on a threshold (and consider
   making the threshold configurable at either compile time (for best 
   low-latency performance) or run-time (for application tuning).

   Then the 3 routines share little code, particularly if the eager/rendezvous 
   implementations are in their own routine
   */
/*
 * MPID_Isend()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Isend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Isend(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPIDI_msg_sz_t data_sz;
	int dt_contig;
	MPI_Aint dt_true_lb;
	MPID_Datatype * dt_ptr;
	MPID_Request * sreq;
	MPIDI_VC_t * vc;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_ISEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_ISEND);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"rank=%d, tag=%d, context=%d", rank, tag, comm->context_id + context_offset));

	if (rank == comm->rank && comm->comm_kind != MPID_INTERCOMM)
	{
		mpi_errno = MPIDI_Isend_self(buf, count, datatype, rank, tag, comm, context_offset, MPIDI_REQUEST_TYPE_SEND, &sreq);
		goto fn_exit;
	}
	
	MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SEND);
	
	if (rank == MPI_PROC_NULL)
	{
		MPIU_Object_set_ref(sreq, 1);
		sreq->cc = 0;
		goto fn_exit;
	}

	MPIDI_Datatype_get_info(count, datatype, dt_contig, data_sz, dt_ptr, dt_true_lb);
	
	MPIDI_Comm_get_vc(comm, rank, &vc);
	
	if (data_sz == 0)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_eager_send_t * const eager_pkt = &upkt.eager_send;

		MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_EAGER_MSG);
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
			
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending zero length message");
		MPIDI_Pkt_init(eager_pkt, MPIDI_CH3_PKT_EAGER_SEND);
		eager_pkt->match.rank = (int16_t)comm->rank;
		eager_pkt->match.tag = tag;
		eager_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		eager_pkt->sender_req_id = sreq->handle;
		eager_pkt->data_sz = 0;

		mpi_errno = MPIDI_CH3.iSend(vc, sreq, eager_pkt, sizeof(*eager_pkt));
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */

		goto fn_exit;
	}
	
	/* FIXME: flow control: limit number of outstanding eager messsages containing data and need to be buffered by the receiver */

	if ((MPIDI_msg_sz_t)(data_sz + sizeof(MPIDI_CH3_Pkt_eager_send_t)) <=		MPIDI_CH3I_Eager_max_msg_size)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_eager_send_t * const eager_pkt = &upkt.eager_send;
		MPID_IOV iov[MPID_IOV_LIMIT];
		int iov_n;
		
		MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_EAGER_MSG);
		
		MPIDI_Pkt_init(eager_pkt, MPIDI_CH3_PKT_EAGER_SEND);
		eager_pkt->match.rank = (int16_t)comm->rank;
		eager_pkt->match.tag = tag;
		eager_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		eager_pkt->sender_req_id = sreq->handle;
		eager_pkt->data_sz = data_sz;
		
		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)eager_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*eager_pkt);
		
		if (dt_contig)
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending contiguous eager message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

			sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;

			iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) ((char *)buf + dt_true_lb);
			iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;

			mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, 2);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			/* FIXME: We do *not* want to use iovs for sending non-contiguous
			   messages if (a) the total message is short (faster to pack)
			   (b) the size of the individual blocks is small (faster to pack 
			   than to create the iov).  This fixme applies to all send
			   modes */
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending non-contiguous eager message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

			MPID_Segment_init(buf, count, datatype, &sreq->dev.segment, 0);
			sreq->dev.segment_first = 0;
			sreq->dev.segment_size = data_sz;
			
			iov_n = MPID_IOV_LIMIT - 1;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &iov[1], &iov_n);
			if (mpi_errno == MPI_SUCCESS)
			{
				iov_n += 1;

				mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, iov_n);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_Object_set_ref(sreq, 0);
					MPIDI_CH3_Request_destroy(sreq);
					sreq = NULL;
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
				
				if (sreq->dev.ca != MPIDI_CH3_CA_COMPLETE)
				{
					sreq->dev.datatype_ptr = dt_ptr;
					MPID_Datatype_add_ref(dt_ptr);
				}
			}
			else
			{
				/* --BEGIN ERROR HANDLING-- */
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
				/* --END ERROR HANDLING-- */
			}
		}
	}
	else
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_rndv_req_to_send_t * const rts_pkt = &upkt.rndv_req_to_send;
			
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending rndv RTS, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));
			
		MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_RNDV_MSG);
		/* FIXME: The partner request should be set to the rts_sreq so that local cancellation can occur; however, this requires
		   allocating the RTS request early to avoid a race condition. */
		sreq->partner_request = NULL;
		
		MPIDI_Pkt_init(rts_pkt, MPIDI_CH3_PKT_RNDV_REQ_TO_SEND);
		rts_pkt->match.rank = (int16_t)comm->rank;
		rts_pkt->match.tag = tag;
		rts_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		rts_pkt->sender_req_id = sreq->handle;
		rts_pkt->data_sz = data_sz;

		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"Rendezvous send using iStartRndvMsg");
	
		if (dt_contig) 
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"  contiguous rndv data, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

			sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
			
			sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) ((char*)sreq->dev.user_buf + dt_true_lb);
			sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;
			sreq->dev.iov_count = 1;
		}
		else
		{
			MPID_Segment_init(sreq->dev.user_buf, sreq->dev.user_count,
							  sreq->dev.datatype, &sreq->dev.segment, 0);
			sreq->dev.iov_count = MPID_IOV_LIMIT;
			sreq->dev.segment_first = 0;
			sreq->dev.segment_size = data_sz;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &sreq->dev.iov[0],
														 &sreq->dev.iov_count);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		mpi_errno = MPIDI_CH3.iStartRndvMsg (vc, sreq, &upkt);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		
		/* FIXME: fill temporary IOV or pack temporary buffer after send to hide some latency.  This requires synchronization
		   because the CTS packet could arrive and be processed before the above iStartmsg completes (depending on the progress
		   engine, threads, etc.). */
		
		if (dt_ptr != NULL)
		{
			sreq->dev.datatype_ptr = dt_ptr;
			MPID_Datatype_add_ref(dt_ptr);
		}
	}

  fn_exit:
	*request = sreq;

	MPIU_DBG_STMT(CH3_OTHER,VERBOSE,
	{
		if (sreq != NULL)
		{
			MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE,L"request allocated, handle=0x%08x", sreq->handle);
		}
	}
				  );
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_ISEND);
	return mpi_errno;
}


/* FIXME: HOMOGENEOUS SYSTEMS ONLY -- no data conversion is performed */

/*
 * MPID_Issend()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Issend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Issend(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPIDI_msg_sz_t data_sz;
	int dt_contig;
	MPI_Aint dt_true_lb;
	MPID_Datatype * dt_ptr;
	MPID_Request * sreq;
	MPIDI_VC_t * vc;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_ISSEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_ISSEND);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"rank=%d, tag=%d, context=%d", rank, tag, comm->context_id + context_offset));
	
	if (rank == comm->rank && comm->comm_kind != MPID_INTERCOMM)
	{
		mpi_errno = MPIDI_Isend_self(buf, count, datatype, rank, tag, comm, context_offset, MPIDI_REQUEST_TYPE_SSEND, &sreq);
		goto fn_exit;
	}
	
	MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SSEND);
	
	if (rank == MPI_PROC_NULL)
	{
		MPIU_Object_set_ref(sreq, 1);
		sreq->cc = 0;
		goto fn_exit;
	}

	MPIDI_Datatype_get_info(count, datatype, dt_contig, data_sz, dt_ptr, dt_true_lb);
	
	MPIDI_Comm_get_vc(comm, rank, &vc);
	
	if (data_sz == 0)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_eager_sync_send_t * const es_pkt = &upkt.eager_sync_send;

		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending zero length message");

		sreq->cc = 2;
		MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_EAGER_MSG);
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		
		MPIDI_Pkt_init(es_pkt, MPIDI_CH3_PKT_EAGER_SYNC_SEND);
		es_pkt->match.rank = (int16_t)comm->rank;
		es_pkt->match.tag = tag;
		es_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		es_pkt->sender_req_id = sreq->handle;
		es_pkt->data_sz = 0;

		mpi_errno = MPIDI_CH3.iSend(vc, sreq, es_pkt, sizeof(*es_pkt));
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		
		goto fn_exit;
	}
	
	/* FIXME: flow control: limit number of outstanding eager messsages containing data and need to be buffered by the receiver */

	if ((MPIDI_msg_sz_t)(data_sz + sizeof(MPIDI_CH3_Pkt_eager_sync_send_t)) <= MPIDI_CH3I_Eager_max_msg_size)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_eager_sync_send_t * const es_pkt = &upkt.eager_sync_send;
		MPID_IOV iov[MPID_IOV_LIMIT];
			
		sreq->cc = 2;
		MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_EAGER_MSG);
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		
		MPIDI_Pkt_init(es_pkt, MPIDI_CH3_PKT_EAGER_SYNC_SEND);
		es_pkt->match.rank = (int16_t)comm->rank;
		es_pkt->match.tag = tag;
		es_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		es_pkt->sender_req_id = sreq->handle;
		es_pkt->data_sz = data_sz;

		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)es_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*es_pkt);

		if (dt_contig)
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending contiguous sync eager message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));
			
			iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) ((char *)buf + dt_true_lb);
			iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;

			mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, 2);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			int iov_n;
			
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending non-contiguous sync eager message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));
			
			MPID_Segment_init(buf, count, datatype, &sreq->dev.segment, 0);
			sreq->dev.segment_first = 0;
			sreq->dev.segment_size = data_sz;
			
			iov_n = MPID_IOV_LIMIT - 1;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &iov[1], &iov_n);
			if (mpi_errno == MPI_SUCCESS)
			{
				iov_n += 1;

				mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, iov_n);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_Object_set_ref(sreq, 0);
					MPIDI_CH3_Request_destroy(sreq);
					sreq = NULL;
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */

				if (sreq->dev.ca != MPIDI_CH3_CA_COMPLETE)
				{
					sreq->dev.datatype_ptr = dt_ptr;
					MPID_Datatype_add_ref(dt_ptr);
				}
			}
			else
			{
				/* --BEGIN ERROR HANDLING-- */
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
				/* --END ERROR HANDLING-- */
			}
		}
	}
	else
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_rndv_req_to_send_t * const rts_pkt = &upkt.rndv_req_to_send;
		
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending rndv RTS, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

		MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_RNDV_MSG);
		/* FIXME: The partner request should be set to the rts_sreq so that local cancellation can occur; however, this requires
		   allocating the RTS request early to avoid a race condition. */
		sreq->partner_request = NULL;
		
		MPIDI_Pkt_init(rts_pkt, MPIDI_CH3_PKT_RNDV_REQ_TO_SEND);
		rts_pkt->match.rank = (int16_t)comm->rank;
		rts_pkt->match.tag = tag;
		rts_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		rts_pkt->sender_req_id = sreq->handle;
		rts_pkt->data_sz = data_sz;

		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"Rendezvous send using iStartRndvMsg");
	
		if (dt_contig) 
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"  contiguous rndv data, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

			sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;

			sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) ((char*)sreq->dev.user_buf + dt_true_lb);
			sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;
			sreq->dev.iov_count = 1;
		}
		else
		{
			MPID_Segment_init(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype, &sreq->dev.segment, 0);
			sreq->dev.iov_count = MPID_IOV_LIMIT;
			sreq->dev.segment_first = 0;
			sreq->dev.segment_size = data_sz;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &sreq->dev.iov[0], &sreq->dev.iov_count);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		mpi_errno = MPIDI_CH3.iStartRndvMsg (vc, sreq, &upkt);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		
		/* FIXME: fill temporary IOV or pack temporary buffer after send to hide some latency.  This requires synchronization
		   because the CTS packet could arrive and be processed before the above iStartmsg completes (depending on the progress
		   engine, threads, etc.). */
		
		if (dt_ptr != NULL)
		{
			sreq->dev.datatype_ptr = dt_ptr;
			MPID_Datatype_add_ref(dt_ptr);
		}
	}

  fn_exit:
	*request = sreq;
	
	MPIU_DBG_STMT(CH3_OTHER,VERBOSE, { if (sreq != NULL) { MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"request allocated, handle=0x%08x", sreq->handle); } } )
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_ISSEND);
	return mpi_errno;
}


/*@
   MPID_Open_port - Open an MPI Port

   Input Arguments:
.  MPI_Info info - info

   Output Arguments:
.  wchar_t *port_name - port name

   Notes:

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER
@*/
#undef FUNCNAME
#define FUNCNAME MPID_Open_port
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Open_port(MPID_Info *info_ptr, wchar_t *port_name)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_OPEN_PORT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_OPEN_PORT);

	/* Check to see if we need to setup channel-specific functions
	   for handling the port operations */
	/* FIXME: if this routine is not called within a critical section,
	   this initialization should be made thread-safe */
	if (setupPortFunctions)
	{
		MPIDI_CH3.PortFnsInit( &portFns );
		setupPortFunctions = 0;
	}

	/* The default for this function is MPIDI_Open_port.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Open_port.
	   In addition, not all channels can implement this operation, so
	   those channels will set the function pointer to NULL */
	if (portFns.OpenPort)
	{
		mpi_errno = portFns.OpenPort( info_ptr, port_name );
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl" );
	}

	/* FIXME: The only purpose of this call is to add this function
	   to the stack of routines involved in this error */
	MPIU_ERR_CHK_POP(mpi_errno);

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_OPEN_PORT);
	return mpi_errno;
}

/*@
   MPID_Close_port - Close port

   Input Parameter:
.  port_name - Name of MPI port to close

   Notes:

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_OTHER

@*/
#undef FUNCNAME
#define FUNCNAME MPID_Close_port
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Close_port(const wchar_t *port_name)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_CLOSE_PORT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_CLOSE_PORT);

	/* Check to see if we need to setup channel-specific functions
	   for handling the port operations */
	if (setupPortFunctions)
	{
		MPIDI_CH3.PortFnsInit( &portFns );
		setupPortFunctions = 0;
	}

	/* The default for this function is 0 (no function).
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Close_port */
	if (portFns.ClosePort)
	{
		mpi_errno = portFns.ClosePort( port_name );
		MPIU_ERR_CHK_POP(mpi_errno);
	}

 fn_fail:		
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_CLOSE_PORT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPID_Comm_accept
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Comm_accept(wchar_t * port_name, MPID_Info * info, int root, MPID_Comm * comm, MPID_Comm ** newcomm_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_COMM_ACCEPT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_COMM_ACCEPT);

	/* Check to see if we need to setup channel-specific functions
	   for handling the port operations */
	if (setupPortFunctions)
	{
		MPIDI_CH3.PortFnsInit( &portFns );
		setupPortFunctions = 0;
	}

	/* A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Comm_accept.
	   If the function is null, we signal a not-implemented error */
	if (portFns.CommAccept)
	{
		mpi_errno = portFns.CommAccept( port_name, info, root, comm, newcomm_ptr );
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl" );
	}

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_COMM_ACCEPT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPID_Comm_connect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Comm_connect(const wchar_t * port_name, MPID_Info * info, int root, MPID_Comm * comm, MPID_Comm ** newcomm_ptr)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_COMM_CONNECT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_COMM_CONNECT);

	/* Check to see if we need to setup channel-specific functions for handling the port operations */
	if (setupPortFunctions)
	{
		MPIDI_CH3.PortFnsInit( &portFns );
		setupPortFunctions = 0;
	}

	/* A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Comm_connect.
	   If the function is null, we signal a not-implemented error */
	if (portFns.CommConnect)
	{
		mpi_errno = portFns.CommConnect( port_name, info, root, comm, newcomm_ptr );
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl" );
	}

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_COMM_CONNECT);
	return mpi_errno;
}

/*
 * Here are the routines that provide some of the default implementations
 * for the Port routines.
 *
 * MPIDI_Open_port - creates a port "name" that includes a tag value that
 * is used to separate different MPI Port values.  That tag value is
 * extracted with MPIDI_GetTagFromPort
 * MPIDI_GetTagFromPort - Routine to return the tag associated with a port.
 */

/*
 * MPIDI_Open_port()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_Open_port
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_Open_port(MPID_Info *info_ptr, wchar_t *port_name)
{
	int mpi_errno = MPI_SUCCESS;
	int len;
	static int port_name_tag=0;   /* this tag is incremented and added to the 
									 business card, which is then returned 
									 as the port name */
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_OPEN_PORT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_OPEN_PORT);

	MPIU_UNREFERENCED_ARG(info_ptr);

	len = MPI_MAX_PORT_NAME;
	mpi_errno = MPIU_Str_add_int_arg(&port_name, &len, MPIDI_CH3I_PORT_NAME_TAG_KEY, port_name_tag++);
	/* FIXME: MPIU_xxx routines should return regular mpi error codes */
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		/* FIXME: We can't use mpi_errno because it isn't a valid mpich2 error code */
		mpi_errno = MPI_ERR_OTHER;
		MPIU_ERR_POP(mpi_errno);
	}

	/* This works because Get_business_card uses the same MPIU_Str_xxx 
	   functions as above to add the business card to the input string */
	/* FIXME: We should instead ask the mpid_pg routines to give us
	   a connection string. There may need to be a separate step to 
	   restrict us to a connection information that is only valid for
	   connections between processes that are started separately (e.g.,
	   may not use shared memory).  We may need a channel-specific 
	   function to create an exportable connection string.  */
	mpi_errno = MPIDI_CH3.Get_business_card(port_name, len);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	/*wprintf(L"open port: <%s>\n", port_name);fflush(stdout);*/
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_OPEN_PORT);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/*
 * The connect and accept routines use this routine to get the port tag
 * from the port name.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_GetTagFromPort
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_GetTagFromPort( const wchar_t *port_name, int *port_name_tag )
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_GETTAGFROMPORT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_GETTAGFROMPORT);

	mpi_errno = MPIU_Str_get_int_arg(port_name, MPIDI_CH3I_PORT_NAME_TAG_KEY, port_name_tag);
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**argstr_port_name_tag");
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_GETTAGFROMPORT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPID_Probe
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Probe(int source, int tag, MPID_Comm * comm, int context_offset, MPI_Status * status)
{
	MPID_Progress_state progress_state;
	const int context = comm->recvcontext_id + context_offset;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_PROBE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_PROBE);

	if (source == MPI_PROC_NULL)
	{
		MPIR_Status_set_procnull(status);
		goto fn_exit;
	}

	MPIDI_CH3_Progress_start(&progress_state);
	do
	{
		if (MPIDI_CH3U_Recvq_FU( source, tag, context, status ))
		{
			break;
		}

		mpi_errno = MPIDI_CH3.Progress_wait(&progress_state);
	}
	while(mpi_errno == MPI_SUCCESS);
	MPIDI_CH3_Progress_end(&progress_state);

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_PROBE);
	return mpi_errno;
}

int MPID_Progress_wait(MPID_Progress_state *state)
{
	return MPIDI_CH3.Progress_wait(state);
}

int MPID_Progress_test(void)
{
	return MPIDI_CH3.Progress_test();
}

int MPID_Progress_poke(void)
{
	return MPIDI_CH3.Progress_test();
}

#undef FUNCNAME
#define FUNCNAME MPID_Recv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Recv(void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPI_Status * status, MPID_Request ** request)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * rreq;
	int found;
	MPIDI_STATE_DECL(MPID_STATE_MPID_RECV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_RECV);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"rank=%d, tag=%d, context=%d", rank, tag, comm->recvcontext_id + context_offset));
	
	if (rank == MPI_PROC_NULL)
	{
		MPIR_Status_set_procnull(status);
		rreq = NULL;
		goto fn_exit;
	}

	rreq = MPIDI_CH3U_Recvq_FDU_or_AEP(rank, tag, comm->recvcontext_id + context_offset, &found);
	if (rreq == NULL)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_NO_MEM, L"**nomem");
	}

	/* FIXME: in the common case, we want to simply complete the message
	   and make as few updates as possible.
	   Note in addition that this routine is used only by MPI_Recv (a
	   blocking routine; the intent of the interface (which returns 
	   a request) was to simplify the handling of the case where the
	   message was not found in the unexpected queue. */
	/* FIXME: why do we add the ref count to comm?  The routine that
	   calls this is required to complete before returning, so
	   no valid user program can free the communicator while we
	   are within this routine, and no change to the ref count should 
	   be needed.  Ditto for remembering the datatype and user buffer
	   statistics (no request should need to be returned by
	   this routine if the message is already available) */
	rreq->comm = comm;
	MPIR_Comm_add_ref(comm);
	rreq->dev.user_buf = buf;
	rreq->dev.user_count = count;
	rreq->dev.datatype = datatype;

	if (found)
	{
		MPIDI_VC_t * vc;

		/* FIXME: We do not need the vc unless we are performing
		   a rendezvous (or synchronous eager(!!!)) send.  This code should 
		   instead be optimized for the low-latency case of an eager 
		   message, already received first, and then handle the more
		   expensive case of rendezvous messages separately. */
		MPIDI_Comm_get_vc(comm, rreq->dev.match.rank, &vc);

		/* Message was found in the unexepected queue */
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"request found in unexpected queue");

		if (MPIDI_Request_get_msg_type(rreq) == MPIDI_REQUEST_EAGER_MSG)
		{
			int recv_pending;
			
			/* This is an eager message. */
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"eager message in the request");

			/* FIXME: There should be no synchronous eager messages.  It
			   is not worth the extra complexity to support these separate
			   from the rendezvous case (we did this early in MPICH-1 and
			   took it out, we should *not* commit the same mistake in MPICH-2)
			*/
			/* FIXME: This appears to be identical code to that in MPID_Isend;
			   that is a maintenance problem (duplicate code should rarely be
			   used; if it is necessary to duplicate code, then it is vital 
			   that all locations of the duplicate code contain a complete
			   list of locations where the same code appears, so that any
			   changes may be uniformly applied */
			/* If this is a eager synchronous message, then we need to send an acknowledgement back to the sender. */
			if (MPIDI_Request_get_sync_send_flag(rreq))
			{
				MPIDI_CH3_Pkt_t upkt;
				MPIDI_CH3_Pkt_eager_sync_ack_t * const esa_pkt = &upkt.eager_sync_ack;
				MPID_Request * esa_req;
					
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending eager sync ack");
				MPIDI_Pkt_init(esa_pkt, MPIDI_CH3_PKT_EAGER_SYNC_ACK);
				esa_pkt->sender_req_id = rreq->dev.sender_req_id;
				mpi_errno = MPIDI_CH3.iStartMsg(vc, esa_pkt, sizeof(*esa_pkt), &esa_req);
				MPIU_ERR_CHK_POP(mpi_errno);
				if (esa_req != NULL)
				{
					MPID_Request_release(esa_req);
				}
			}
			
			MPIDI_Request_recv_pending(rreq, &recv_pending);
			if (!recv_pending)
			{
				/* All of the data has arrived, we need to unpack the data and then free the buffer and the request. */
				if (rreq->dev.recv_data_sz > 0)
				{
					MPIDI_CH3U_Request_unpack_uebuf(rreq);
					MPIU_Free(rreq->dev.tmpbuf);
				}
				
				mpi_errno = rreq->status.MPI_ERROR;
				if (status != MPI_STATUS_IGNORE)
				{
					*status = rreq->status;
				}
				
				MPID_Request_release(rreq);
				rreq = NULL;
				
				goto fn_exit;
			}
			else
			{
				/* The data is still being transfered across the net.  We'll leave it to the progress engine to handle once the
				   entire message has arrived. */
				if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
				{
					MPID_Datatype_get_ptr(datatype, rreq->dev.datatype_ptr);
					MPID_Datatype_add_ref(rreq->dev.datatype_ptr);
				}
			}
		}
		else if (MPIDI_Request_get_msg_type(rreq) == MPIDI_REQUEST_RNDV_MSG)
		{
			/* A rendezvous request-to-send (RTS) message has arrived.  We need to send a CTS message to the remote process. */
			mpi_errno = MPIDI_CH3U_Post_data_receive(found, &rreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_RNDV_REQ_TO_SEND");
			}
			/* --END ERROR HANDLING-- */
			mpi_errno = MPIDI_CH3.iStartRndvTransfer (vc, rreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|ctspkt");
			}
			/* --END ERROR HANDLING-- */

			if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(datatype, rreq->dev.datatype_ptr);
				MPID_Datatype_add_ref(rreq->dev.datatype_ptr);
			}
		}
		else if (MPIDI_Request_get_msg_type(rreq) == MPIDI_REQUEST_SELF_MSG)
		{
			MPID_Request * const sreq = rreq->partner_request;

			if (sreq != NULL)
			{
				MPIDI_msg_sz_t data_sz;
				
				MPIDI_CH3U_Buffer_copy(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype, &sreq->status.MPI_ERROR,
									   buf, count, datatype, &data_sz, &rreq->status.MPI_ERROR);
				rreq->status.count = (int)data_sz;
				MPID_Request_set_completed(sreq);
				MPID_Request_release(sreq);
			}
			else
			{
				/* The sreq is missing which means an error occurred.  rreq->status.MPI_ERROR should have been set when the
				   error was detected. */
			}

			if (status != MPI_STATUS_IGNORE)
			{
				*status = rreq->status;
			}

			/* no other thread can possibly be waiting on rreq, so it is safe to reset ref_count and cc */
			rreq->cc = 0;
			MPIU_Object_set_ref(rreq, 1);
		}
		else
		{
			/* --BEGIN ERROR HANDLING-- */
			MPID_Request_release(rreq);
			rreq = NULL;
			MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_INTERN, L"**ch3|badmsgtype", L"**ch3|badmsgtype %d", MPIDI_Request_get_msg_type(rreq));
			/* --END ERROR HANDLING-- */
		}
	}
	else
	{
		/* Message has yet to arrived.  The request has been placed on the list of posted receive requests and populated with
		   information supplied in the arguments. */
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"request allocated in posted queue");
		
		if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
		{
			MPID_Datatype_get_ptr(datatype, rreq->dev.datatype_ptr);
			MPID_Datatype_add_ref(rreq->dev.datatype_ptr);
		}

		rreq->dev.recv_pending_count = 1;
	}

  fn_exit:
	*request = rreq;
	MPIU_DBG_STMT(CH3_OTHER,VERBOSE,
	if (rreq)
	{
		MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"request allocated, handle=0x%08x", rreq->handle);
	}
	else
	{
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"operation complete, no requests allocated");
	});

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_RECV);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_create
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_create(void *base, MPI_Aint size, int disp_unit, MPID_Info *info, 
					MPID_Comm *comm_ptr, MPID_Win **win_ptr)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_CREATE);
	
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_CREATE);

	MPIU_UNREFERENCED_ARG(info);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_create.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Win_create.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	/* We pass the RMAFns function table to this function because a channel may 
	   want to reset it to the default if it finds that it cannot optimize for this 
	   set of windows. The sshm channel does this if windows are not allocated in 
	   shared memory. */
	   
	if (RMAFns.Win_create)
	{
		mpi_errno = RMAFns.Win_create(base, size, disp_unit, info, comm_ptr, win_ptr, &RMAFns);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_WIN_CREATE);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_free
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_free(MPID_Win **win_ptr)
{
   int mpi_errno=MPI_SUCCESS;
   MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_FREE);
		
   MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_FREE);
		
	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_free.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Win_free.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Win_free)
	{
		mpi_errno = RMAFns.Win_free(win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_WIN_FREE);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Put
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Put(void *origin_addr, int origin_count, MPI_Datatype
			origin_datatype, int target_rank, MPI_Aint target_disp,
			int target_count, MPI_Datatype target_datatype, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_PUT);
		
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_PUT);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Put.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Put.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Put)
	{
		mpi_errno = RMAFns.Put(origin_addr, origin_count, origin_datatype, 
							   target_rank, target_disp, target_count, target_datatype,
							   win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_PUT);	
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Get
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Get(void *origin_addr, int origin_count, MPI_Datatype
			origin_datatype, int target_rank, MPI_Aint target_disp,
			int target_count, MPI_Datatype target_datatype, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_GET);
		
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_GET);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Get.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Get.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Get)
	{
		mpi_errno = RMAFns.Get(origin_addr, origin_count, origin_datatype, 
							   target_rank, target_disp, target_count, target_datatype,
							   win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_GET);
	return mpi_errno;
}



#undef FUNCNAME
#define FUNCNAME MPID_Accumulate
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Accumulate(void *origin_addr, int origin_count, MPI_Datatype
					origin_datatype, int target_rank, MPI_Aint target_disp,
					int target_count, MPI_Datatype target_datatype, MPI_Op op,
					MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_ACCUMULATE);
		
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_ACCUMULATE);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Accumulate.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_CH3_Accumulate.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Accumulate)
	{
		mpi_errno = RMAFns.Accumulate(origin_addr, origin_count, origin_datatype, 
							   target_rank, target_disp, target_count, target_datatype,
							   op, win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_ACCUMULATE);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_fence
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_fence(int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_FENCE);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_FENCE);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_fence.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Win_fence.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Win_fence)
	{
		mpi_errno = RMAFns.Win_fence(assert, win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_FENCE);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_post
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_post(MPID_Group *group_ptr, int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_POST);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_POST);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_post.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Win_post.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Win_post)
	{
		mpi_errno = RMAFns.Win_post(group_ptr, assert, win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_POST);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_start
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_start(MPID_Group *group_ptr, int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_START);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_START);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_start.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Win_start.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Win_start)
	{
		mpi_errno = RMAFns.Win_start(group_ptr, assert, win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_START);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_complete
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_complete(MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_COMPLETE);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_COMPLETE);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_complete.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Win_complete.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Win_complete)
	{
		mpi_errno = RMAFns.Win_complete(win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_COMPLETE);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_wait(MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_WAIT);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_WAIT);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_wait.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Win_wait.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Win_wait)
	{
		mpi_errno = RMAFns.Win_wait(win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_WAIT);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_lock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_lock(int lock_type, int dest, int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_LOCK);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_LOCK);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_lock.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Win_lock.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Win_lock)
	{
		mpi_errno = RMAFns.Win_lock(lock_type, dest, assert, win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_LOCK);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_unlock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_unlock(int dest, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_UNLOCK);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_UNLOCK);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Win_unlock.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Win_unlock.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Win_unlock)
	{
		mpi_errno = RMAFns.Win_unlock(dest, win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}

 fn_fail:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_UNLOCK);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Alloc_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void *MPID_Alloc_mem( size_t size, MPID_Info *info_ptr )
{
	void *ap=NULL;
	MPIDI_STATE_DECL(MPID_STATE_MPID_ALLOC_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_ALLOC_MEM);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Alloc_mem.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Alloc_mem.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Alloc_mem)
	{
		ap = RMAFns.Alloc_mem(size, info_ptr);
	}
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_ALLOC_MEM);
	return ap;
}


#undef FUNCNAME
#define FUNCNAME MPID_Free_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Free_mem( void *ptr )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_FREE_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_FREE_MEM);

	/* Check to see if we need to setup channel-specific functions
	   for handling the RMA operations */
	if (setupRMAFunctions)
	{
		MPIDI_CH3.RMAFnsInit( &RMAFns );
		setupRMAFunctions = 0;
	}

	/* The default for this function is MPIDI_Free_mem.
	   A channel may define its own function and set it in the 
	   init check above; such a function may be named MPIDI_Free_mem.
	   If a channel does not implement this operation, it will set
	   the function pointer to NULL */

	if (RMAFns.Free_mem)
	{
		mpi_errno = RMAFns.Free_mem(ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**notimpl");
	}
		
 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_FREE_MEM);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPID_Win_test
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Win_test(MPID_Win *win_ptr, int *flag)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_WIN_TEST);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPID_WIN_TEST);

	mpi_errno = MPID_Progress_test();
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_TEST);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */

	*flag = (win_ptr->my_counter) ? 0 : 1;

	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPID_WIN_TEST);
	return mpi_errno;
}


/* FIXME: It would be better to install this as an abort and a finalize 
   handler, ranter than expecting all routines to call this.  It
   can be installed on first use of alloc_mem */
#undef FUNCNAME
#define FUNCNAME MPID_Cleanup_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPID_Cleanup_mem()
{
	MPIDI_STATE_DECL(MPID_STATE_MPID_CLEANUP_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_CLEANUP_MEM);
	MPIDI_CH3_Cleanup_mem();
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_CLEANUP_MEM);
}


/* FIXME: HOMOGENEOUS SYSTEMS ONLY -- no data conversion is performed */

/* FIXME: How does this differ from eager send?  It should differ in 
   only a few bits (e.g., indicate that the send is ready and should
   fail if there is no matching receive) */

/*
 * MPID_Rsend()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Rsend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Rsend(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_ready_send_t * const ready_pkt = &upkt.ready_send;
	MPIDI_msg_sz_t data_sz;
	int dt_contig;
	MPI_Aint dt_true_lb;
	MPID_Datatype * dt_ptr;
	MPID_Request * sreq = NULL;
	MPID_IOV iov[MPID_IOV_LIMIT];
	MPIDI_VC_t * vc;
	int mpi_errno = MPI_SUCCESS;	
	MPIDI_STATE_DECL(MPID_STATE_MPID_RSEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_RSEND);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"rank=%d, tag=%d, context=%d", rank, tag, comm->context_id + context_offset));

	if (rank == comm->rank && comm->comm_kind != MPID_INTERCOMM)
	{
		mpi_errno = MPIDI_Isend_self(buf, count, datatype, rank, tag, comm, context_offset, MPIDI_REQUEST_TYPE_RSEND, request);
		goto fn_exit;
	}

	if (rank == MPI_PROC_NULL)
	{
		goto fn_exit;
	}

	MPIDI_Datatype_get_info(count, datatype, dt_contig, data_sz, dt_ptr, dt_true_lb);

	MPIDI_Comm_get_vc(comm, rank, &vc);
	
	MPIDI_Pkt_init(ready_pkt, MPIDI_CH3_PKT_READY_SEND);
	ready_pkt->match.rank = (int16_t)comm->rank;
	ready_pkt->match.tag = tag;
	ready_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
	ready_pkt->sender_req_id = MPI_REQUEST_NULL;
	ready_pkt->data_sz = data_sz;

	if (data_sz == 0)
	{
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending zero length message");

		mpi_errno = MPIDI_CH3.iStartMsg(vc, ready_pkt, sizeof(*ready_pkt), &sreq);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		if (sreq != NULL)
		{
			MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_RSEND);
			/* sreq->comm = comm;
			   MPIR_Comm_add_ref(comm); -- not needed for blocking operations */
		}

		goto fn_exit;
	}
	
	iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)ready_pkt;
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*ready_pkt);

	if (dt_contig)
	{
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending contiguous ready-mode message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

		/* FIXME: handle case where data_sz is greater than what can be stored in iov.MPID_IOV_LEN.  hand off to segment code? */
		iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) buf + dt_true_lb);
		iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;

		mpi_errno = MPIDI_CH3.iStartMsgv(vc, iov, 2, &sreq);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		if (sreq != NULL)
		{
			MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_RSEND);
			/* sreq->comm = comm;
			   MPIR_Comm_add_ref(comm); -- not needed for blocking operations */
		}
	}
	else
	{
		int iov_n;
			
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending non-contiguous ready-mode message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

		MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit);
		MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_RSEND);
			
		MPID_Segment_init(buf, count, datatype, &sreq->dev.segment, 0);
		sreq->dev.segment_first = 0;
		sreq->dev.segment_size = data_sz;
			
		iov_n = MPID_IOV_LIMIT - 1;
		mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &iov[1], &iov_n);
		if (mpi_errno == MPI_SUCCESS)
		{
			iov_n += 1;

			if (sreq->dev.ca != MPIDI_CH3_CA_COMPLETE)
			{
				/* sreq->dev.datatype_ptr = dt_ptr;
				   MPID_Datatype_add_ref(dt_ptr); -- not needed for blocking operations */
			}
			
			mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, iov_n);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			/* --BEGIN ERROR HANDLING-- */
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
			goto fn_exit;
			/* --END ERROR HANDLING-- */
		}
	}

  fn_exit:
	*request = sreq;

	MPIU_DBG_STMT(CH3_OTHER,VERBOSE,
	{
		if (mpi_errno == MPI_SUCCESS)
		{
			if (sreq)
			{
				MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE,L"request allocated, handle=0x%08x", sreq->handle);
			}
			else
			{
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"operation complete, no requests allocated");
			}
		}
	}
				  );
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_RSEND);
	return mpi_errno;
}


/* FIXME - HOMOGENEOUS SYSTEMS ONLY -- no data conversion is performed */

/* FIXME: This should share the rendezvous and eager code with 
   the related operations in MPID_Rsend (always eager) and MPID_Ssend
   (always rendezvous, though an eager sync mode is possible, it isn't
   worth the effort). */

/*
 * MPID_Send()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Send
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Send(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPIDI_msg_sz_t data_sz;
	int dt_contig;
	MPI_Aint dt_true_lb;
	MPID_Datatype * dt_ptr;
	MPID_Request * sreq = NULL;
	MPIDI_VC_t * vc;
	int mpi_errno = MPI_SUCCESS;	
	MPIDI_STATE_DECL(MPID_STATE_MPID_SEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_SEND);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"rank=%d, tag=%d, context=%d", rank, tag, comm->context_id + context_offset));

	if (rank == comm->rank && comm->comm_kind != MPID_INTERCOMM)
	{
		mpi_errno = MPIDI_Isend_self(buf, count, datatype, rank, tag, comm, context_offset, MPIDI_REQUEST_TYPE_SEND, &sreq);

		/* In the single threaded case, sending to yourself will cause 
		   deadlock.  Note that in the runtime-thread case, this check
		   will not be made (long-term FIXME) */
#	   ifndef MPICH_IS_THREADED
		{
			/* --BEGIN ERROR HANDLING-- */
			if (sreq != NULL && sreq->cc != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**dev|selfsenddeadlock", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
#		endif
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		}
		/* --END ERROR HANDLING-- */
		goto fn_exit;
	}

	if (rank == MPI_PROC_NULL)
	{
		goto fn_exit;
	}

	MPIDI_Datatype_get_info(count, datatype, dt_contig, data_sz, dt_ptr, dt_true_lb);

	MPIDI_Comm_get_vc(comm, rank, &vc);
	
	if (data_sz == 0)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_eager_send_t * const eager_pkt = &upkt.eager_send;

		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending zero length message");
		MPIDI_Pkt_init(eager_pkt, MPIDI_CH3_PKT_EAGER_SEND);
		eager_pkt->match.rank = (int16_t)comm->rank;
		eager_pkt->match.tag = tag;
		eager_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		eager_pkt->sender_req_id = MPI_REQUEST_NULL;
		eager_pkt->data_sz = 0;

		mpi_errno = MPIDI_CH3.iStartMsg(vc, eager_pkt, sizeof(*eager_pkt), &sreq);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			/* FIXME: this is a fatal error because a sequence number has already been allocated.  If sequence numbers are not
			   being used then this could be a recoverable error.  A check needs to be added that sets the error to fatal or
			   recoverable depending on the use of sequence numbers. */
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		if (sreq != NULL)
		{
			MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SEND);
			/* sreq->comm = comm;
			  MPIR_Comm_add_ref(comm); -- not necessary for blocking functions */
		}
		
		goto fn_exit;
	}
	
	/* FIXME: flow control: limit number of outstanding eager messsages containing data and need to be buffered by the receiver */

	if ((MPIDI_msg_sz_t)(data_sz + sizeof(MPIDI_CH3_Pkt_eager_send_t)) <=		MPIDI_CH3I_Eager_max_msg_size)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_eager_send_t * const eager_pkt = &upkt.eager_send;
		MPID_IOV iov[MPID_IOV_LIMIT];
			
		MPIDI_Pkt_init(eager_pkt, MPIDI_CH3_PKT_EAGER_SEND);
		eager_pkt->match.rank = (int16_t)comm->rank;
		eager_pkt->match.tag = tag;
		eager_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		eager_pkt->sender_req_id = MPI_REQUEST_NULL;
		eager_pkt->data_sz = data_sz;

		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)eager_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*eager_pkt);

		if (dt_contig)
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending contiguous eager message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));
			
			iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) ((char *)buf + dt_true_lb);
			iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;

			mpi_errno = MPIDI_CH3.iStartMsgv(vc, iov, 2, &sreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			if (sreq != NULL)
			{
				MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SEND);
				/* sreq->comm = comm;
				   MPIR_Comm_add_ref(comm); -- not necessary for blocking functions */

			}
		}
		else
		{
			int iov_n;
			
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending non-contiguous eager message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));
			
			MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit);
			MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SEND);
			
			MPID_Segment_init(buf, count, datatype, &sreq->dev.segment, 0);
			sreq->dev.segment_first = 0;
			sreq->dev.segment_size = data_sz;
			
			iov_n = MPID_IOV_LIMIT - 1;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &iov[1], &iov_n);
			if (mpi_errno == MPI_SUCCESS)
			{
				iov_n += 1;

				if (sreq->dev.ca != MPIDI_CH3_CA_COMPLETE)
				{
					/* sreq->dev.datatype_ptr = dt_ptr;
					   MPID_Datatype_add_ref(dt_ptr); -- not necessary for blocking functions */
				}
				
				mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, iov_n);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_Object_set_ref(sreq, 0);
					MPIDI_CH3_Request_destroy(sreq);
					sreq = NULL;
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}
			else
			{
				/* --BEGIN ERROR HANDLING-- */
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
				/* --END ERROR HANDLING-- */
			}
		}
	}
	else
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_rndv_req_to_send_t * const rts_pkt = &upkt.rndv_req_to_send;

		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending rndv RTS, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

		MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit);
		MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SEND);
		sreq->partner_request = NULL;
		
		/* FIXME - Since the request is never returned to the user and they can't do things like cancel it or wait on it, we may
		   not need to fill in all of the fields.  For example, it may be completely unnecessary to supply the matching
		   information.  Also, some of the fields can be set after the message has been sent.  These issues should be looked at
		   more closely when we are trying to squeeze those last few nanoseconds out of the code.  */
		
		MPIDI_Pkt_init(rts_pkt, MPIDI_CH3_PKT_RNDV_REQ_TO_SEND);
		rts_pkt->match.rank = (int16_t)comm->rank;
		rts_pkt->match.tag = tag;
		rts_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		rts_pkt->sender_req_id = sreq->handle;
		rts_pkt->data_sz = data_sz;

		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"Rendezvous send using iStartRndvMsg");
	
		if (dt_contig) 
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"  contiguous rndv data, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));

			sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;

			sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) ((char*)sreq->dev.user_buf + dt_true_lb);
			sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;
			sreq->dev.iov_count = 1;
		}
		else
		{
			MPID_Segment_init(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype, &sreq->dev.segment, 0);
			sreq->dev.iov_count = MPID_IOV_LIMIT;
			sreq->dev.segment_first = 0;
			sreq->dev.segment_size = data_sz;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &sreq->dev.iov[0], &sreq->dev.iov_count);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}

		mpi_errno = MPIDI_CH3.iStartRndvMsg (vc, sreq, &upkt);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		
		/* FIXME: fill temporary IOV or pack temporary buffer after send to hide some latency.  This requires synchronization
		   because the CTS packet could arrive and be processed before the above iStartmsg completes (depending on the progress
		   engine, threads, etc.). */
		
		if (dt_ptr != NULL)
		{
			/* sreq->dev.datatype_ptr = dt_ptr;
			   MPID_Datatype_add_ref(dt_ptr);  -- no necessary for blocking send */
		}

	}

  fn_exit:
	*request = sreq;

	MPIU_DBG_STMT(CH3_OTHER,VERBOSE,
	{
		if (mpi_errno == MPI_SUCCESS)
		{
			if (sreq)
			{
				MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"request allocated, handle=0x%08x", sreq->handle);
			}
			else
			{
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"operation complete, no requests allocated");
			}
		}
	}
				  );
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_SEND);
	return mpi_errno;
}


/* FIXME: HOMOGENEOUS SYSTEMS ONLY -- no data conversion is performed */

/*
 * MPID_Ssend()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Ssend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Ssend(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset,
			   MPID_Request ** request)
{
	MPIDI_msg_sz_t data_sz;
	int dt_contig;
	MPI_Aint dt_true_lb;
	MPID_Datatype * dt_ptr;
	MPID_Request * sreq = NULL;
	MPIDI_VC_t * vc;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_SSEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_SSEND);

	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"rank=%d, tag=%d, context=%d", rank, tag, comm->context_id + context_offset));

	if (rank == comm->rank && comm->comm_kind != MPID_INTERCOMM)
	{
		mpi_errno = MPIDI_Isend_self(buf, count, datatype, rank, tag, comm, context_offset, MPIDI_REQUEST_TYPE_SSEND, &sreq);
		/* In the single threaded case, sending to yourself will cause 
		   deadlock.  Note that in the runtime-thread case, this check
		   will not be made (long-term FIXME) */
#	   ifndef MPICH_IS_THREADED
		{
			/* --BEGIN ERROR HANDLING-- */
			if (sreq != NULL && sreq->cc != 0)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**dev|selfsenddeadlock", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
#		endif
		goto fn_exit;
	}
	
	if (rank == MPI_PROC_NULL)
	{
		goto fn_exit;
	}

	MPIDI_Datatype_get_info(count, datatype, dt_contig, data_sz, dt_ptr, dt_true_lb);

	MPIDI_Comm_get_vc(comm, rank, &vc);
		
	MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SSEND);
	
	/* FIXME: Since the request is never returned to the user and they can't do things like cancel it or wait on it, we may not
	   need to fill in all of the fields.  For example, it may be completely unnecessary to supply the matching information.  Also,
	   some of the fields can be set after the message has been sent.  These issues should be looked at more closely when we are
	   trying to squeeze those last few nanoseconds out of the code.  */
	
	if (data_sz == 0)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_eager_sync_send_t * const es_pkt = &upkt.eager_sync_send;

		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending zero length message");

		sreq->cc = 2;
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		
		MPIDI_Pkt_init(es_pkt, MPIDI_CH3_PKT_EAGER_SYNC_SEND);
		es_pkt->match.rank = (int16_t)comm->rank;
		es_pkt->match.tag = tag;
		es_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		es_pkt->sender_req_id = sreq->handle;
		es_pkt->data_sz = 0;

		mpi_errno = MPIDI_CH3.iSend(vc, sreq, es_pkt, sizeof(*es_pkt));
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		
		goto fn_exit;
	}
	
	/* FIXME: flow control: limit number of outstanding eager messsages containing data and need to be buffered by the receiver */

	if ((MPIDI_msg_sz_t)(data_sz + sizeof(MPIDI_CH3_Pkt_eager_sync_send_t)) <= MPIDI_CH3I_Eager_max_msg_size)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_eager_sync_send_t * const es_pkt = &upkt.eager_sync_send;
		MPID_IOV iov[MPID_IOV_LIMIT];
			
		sreq->cc = 2;
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		
		MPIDI_Pkt_init(es_pkt, MPIDI_CH3_PKT_EAGER_SYNC_SEND);
		es_pkt->match.rank = (int16_t)comm->rank;
		es_pkt->match.tag = tag;
		es_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		es_pkt->sender_req_id = sreq->handle;
		es_pkt->data_sz = data_sz;

		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)es_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*es_pkt);

		if (dt_contig)
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending contiguous sync eager message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));
			
			iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) ((char *)buf + dt_true_lb);
			iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;

			mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, 2);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			int iov_n;
			
			MPIU_DBG_MSG_D(CH3_OTHER,VERBOSE, L"sending non-contiguous sync eager message, data_sz=" MPIDI_MSG_SZ_FMT, data_sz);

			MPID_Segment_init(buf, count, datatype, &sreq->dev.segment, 0);
			sreq->dev.segment_first = 0;
			sreq->dev.segment_size = data_sz;
			
			iov_n = MPID_IOV_LIMIT - 1;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &iov[1], &iov_n);
			if (mpi_errno == MPI_SUCCESS)
			{
				iov_n += 1;

				if (sreq->dev.ca != MPIDI_CH3_CA_COMPLETE)
				{
					/* sreq->dev.datatype_ptr = dt_ptr;
					   MPID_Datatype_add_ref(dt_ptr); -- not needed for blocking operations */
				}

				mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, iov_n);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_Object_set_ref(sreq, 0);
					MPIDI_CH3_Request_destroy(sreq);
					sreq = NULL;
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|eagermsg", 0);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}
			else
			{
				/* --BEGIN ERROR HANDLING-- */
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				sreq = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
				/* --END ERROR HANDLING-- */
			}
		}
	}
	else
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_rndv_req_to_send_t * const rts_pkt = &upkt.rndv_req_to_send;
		
		MPIU_DBG_MSG_D(CH3_OTHER,VERBOSE, L"sending rndv RTS, data_sz=" MPIDI_MSG_SZ_FMT, data_sz);
			
		sreq->partner_request = NULL;
		
		MPIDI_Pkt_init(rts_pkt, MPIDI_CH3_PKT_RNDV_REQ_TO_SEND);
		rts_pkt->match.rank = (int16_t)comm->rank;
		rts_pkt->match.tag = tag;
		rts_pkt->match.context_id = (int16_t)(comm->context_id + context_offset);
		rts_pkt->sender_req_id = sreq->handle;
		rts_pkt->data_sz = data_sz;

		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"Rendezvous send using iStartRndvMsg");
	
		if (dt_contig) 
		{
			MPIU_DBG_MSG_D(CH3_OTHER,VERBOSE,L"  contiguous rndv data, data_sz=" MPIDI_MSG_SZ_FMT, data_sz);
				
			sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
			
			sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) ((char*)sreq->dev.user_buf + dt_true_lb);
			sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;
			sreq->dev.iov_count = 1;
		}
		else
		{
			MPID_Segment_init(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype, &sreq->dev.segment, 0);
			sreq->dev.iov_count = MPID_IOV_LIMIT;
			sreq->dev.segment_first = 0;
			sreq->dev.segment_size = data_sz;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &sreq->dev.iov[0], &sreq->dev.iov_count);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		mpi_errno = MPIDI_CH3.iStartRndvMsg (vc, sreq, &upkt);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */

		/* FIXME: fill temporary IOV or pack temporary buffer after send to hide some latency.  This requires synchronization
		   because the CTS packet could arrive and be processed before the above iStartmsg completes (depending on the progress
		   engine, threads, etc.). */
		
		if (dt_ptr != NULL)
		{
			/* sreq->dev.datatype_ptr = dt_ptr;
			   MPID_Datatype_add_ref(dt_ptr); -- not needed for blocking operations */
		}
	}

  fn_exit:
	*request = sreq;
	
	MPIU_DBG_STMT(CH3_OTHER,VERBOSE,{if (sreq!=NULL) { MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"request allocated, handle=0x%08x", sreq->handle);}});
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_SSEND);
	return mpi_errno;
}


/* FIXME: This bsend header shouldn't be needed (the function prototype
   should be in mpiimpl.h), to allow a devices MPID_Startall to use the
   MPIR_Bsend_isend function */
#include "../../../mpi/pt2pt/bsendutil.h"

/* FIXME: Consider using function pointers for invoking persistent requests;
   if we made those part of the public request structure, the top-level routine
   could implement startall unless the device wanted to study the requests
   and reorder them */

/* FIXME: Where is the memory registration call in the init routines, 
   in case the channel wishes to take special action (such as pinning for DMA)
   the memory? This was part of the design. */

/* This macro initializes all of the fields in a persistent request */
#define MPIDI_Request_create_psreq(sreq_, mpi_errno_, FAIL_)				\
{																		\
	(sreq_) = MPID_Request_create();								\
	if ((sreq_) == NULL)												\
	{																		\
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"send request allocation failed");\
		(mpi_errno_) = MPIR_ERR_MEMALLOCFAILED;								\
		FAIL_;																\
	}																		\
																		\
	MPIU_Object_set_ref((sreq_), 1);										\
	(sreq_)->cc   = 0;												  \
	(sreq_)->kind = MPID_PREQUEST_SEND;										\
	(sreq_)->comm = comm;												\
	MPIR_Comm_add_ref(comm);												\
	(sreq_)->dev.match.rank = (int16_t)rank;										\
	(sreq_)->dev.match.tag = tag;										\
	(sreq_)->dev.match.context_id = (int16_t)(comm->context_id + context_offset);		\
	(sreq_)->dev.user_buf = (void *) buf;								\
	(sreq_)->dev.user_count = count;										\
	(sreq_)->dev.datatype = datatype;										\
	(sreq_)->partner_request = NULL;										\
}

		
/*
 * MPID_Startall()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Startall
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Startall(int count, MPID_Request * requests[])
{
	int i;
	int rc;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_STARTALL);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_STARTALL);

	for (i = 0; i < count; i++)
	{
		MPID_Request * const preq = requests[i];

		/* FIXME: The odd 7th arg (match.context_id - comm->context_id) 
		   is probably to get the context offset.  Do we really need the
		   context offset? Is there any case where the offset isn't zero? */
		switch (MPIDI_Request_get_type(preq))
		{
			case MPIDI_REQUEST_TYPE_RECV:
			{
				rc = MPID_Irecv(preq->dev.user_buf, preq->dev.user_count, preq->dev.datatype, preq->dev.match.rank,
					preq->dev.match.tag, preq->comm, preq->dev.match.context_id - preq->comm->recvcontext_id,
					&preq->partner_request);
				break;
			}
			
			case MPIDI_REQUEST_TYPE_SEND:
			{
				rc = MPID_Isend(preq->dev.user_buf, preq->dev.user_count, preq->dev.datatype, preq->dev.match.rank,
					preq->dev.match.tag, preq->comm, preq->dev.match.context_id - preq->comm->context_id,
					&preq->partner_request);
				break;
			}
				
			case MPIDI_REQUEST_TYPE_RSEND:
			{
				rc = MPID_Irsend(preq->dev.user_buf, preq->dev.user_count, preq->dev.datatype, preq->dev.match.rank,
					preq->dev.match.tag, preq->comm, preq->dev.match.context_id - preq->comm->context_id,
					&preq->partner_request);
				break;
			}
				
			case MPIDI_REQUEST_TYPE_SSEND:
			{
				rc = MPID_Issend(preq->dev.user_buf, preq->dev.user_count, preq->dev.datatype, preq->dev.match.rank,
					preq->dev.match.tag, preq->comm, preq->dev.match.context_id - preq->comm->context_id,
					&preq->partner_request);
				break;
			}

			case MPIDI_REQUEST_TYPE_BSEND:
			{
#if 0
				MPID_Request * sreq;
				
				sreq = MPID_Request_create();
				if (sreq != NULL)
				{
					rc = MPIR_Bsend_isend(preq->dev.user_buf, preq->dev.user_count, preq->dev.datatype, preq->dev.match.rank,
										  preq->dev.match.tag, preq->comm, BSEND_INIT, &preq->partner_request);

					MPIU_Object_set_ref(sreq, 1);
					sreq->kind = MPID_REQUEST_SEND;
					sreq->cc   = 0;
					sreq->comm = NULL;
/* FIXME: why the #if 0? */					
#if 0					
					sreq->comm = preq->comm;
					MPIR_Comm_add_ref(sreq->comm);
#endif					
					sreq->status.MPI_ERROR = rc;
					preq->partner_request = sreq;
					rc = MPI_SUCCESS;
				}
				else
				{
					/* --BEGIN ERROR HANDLING-- */
					rc = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					/* --END ERROR HANDLING-- */
				}
#else
				MPI_Request sreq_handle;

				NMPI_Ibsend(rc, (preq->dev.user_buf, preq->dev.user_count,
					preq->dev.datatype, preq->dev.match.rank,
					preq->dev.match.tag, preq->comm->handle, 
					&sreq_handle));
				if (rc == MPI_SUCCESS)
				{
					MPID_Request_get_ptr(sreq_handle, preq->partner_request);
				}
#endif				
				break;
			}

			default:
			{
				/* --BEGIN ERROR HANDLING-- */
				rc = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|badreqtype", L"**ch3|badreqtype %d", MPIDI_Request_get_type(preq));
				/* --END ERROR HANDLING-- */
			}
		}
		
		if (rc == MPI_SUCCESS)
		{
			preq->status.MPI_ERROR = MPI_SUCCESS;
			preq->cc_ptr = &preq->partner_request->cc;
		}
		/* --BEGIN ERROR HANDLING-- */
		else
		{
			/* If a failure occurs attempting to start the request, then we 
			   assume that partner request was not created, and stuff
			   the error code in the persistent request.  The wait and test
			   routines will look at the error code in the persistent
			   request if a partner request is not present. */
			preq->partner_request = NULL;
			preq->status.MPI_ERROR = rc;
			preq->cc_ptr = &preq->cc;
			preq->cc = 0;
		}
		/* --END ERROR HANDLING-- */
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_STARTALL);
	return mpi_errno;
}

/* FIXME:
   Move the routines that initialize the persistent requests into this file,
   since startall must be used with all of them */

/*
 * MPID_Send_init()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Send_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Send_init(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPID_Request * sreq;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_SEND_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_SEND_INIT);

	MPIDI_Request_create_psreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SEND);
	if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
	{
		MPID_Datatype_get_ptr(datatype, sreq->dev.datatype_ptr);
		MPID_Datatype_add_ref(sreq->dev.datatype_ptr);
	}
	*request = sreq;

  fn_exit:	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_SEND_INIT);
	return mpi_errno;
}

/*
 * MPID_Ssend_init()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Ssend_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Ssend_init(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset,
					MPID_Request ** request)
{
	MPID_Request * sreq;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_SSEND_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_SSEND_INIT);

	MPIDI_Request_create_psreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SSEND);
	if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
	{
		MPID_Datatype_get_ptr(datatype, sreq->dev.datatype_ptr);
		MPID_Datatype_add_ref(sreq->dev.datatype_ptr);
	}
	*request = sreq;

  fn_exit:	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_SSEND_INIT);
	return mpi_errno;
}

/*
 * MPID_Rsend_init()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Rsend_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Rsend_init(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPID_Request * sreq;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_RSEND_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_RSEND_INIT);

	MPIDI_Request_create_psreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_RSEND);
	if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
	{
		MPID_Datatype_get_ptr(datatype, sreq->dev.datatype_ptr);
		MPID_Datatype_add_ref(sreq->dev.datatype_ptr);
	}
	*request = sreq;

  fn_exit:	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_RSEND_INIT);
	return mpi_errno;
}

/*
 * MPID_Bsend_init()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Bsend_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Bsend_init(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPID_Request * sreq;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_BSEND_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_BSEND_INIT);

	MPIDI_Request_create_psreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_BSEND);
	if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
	{
		MPID_Datatype_get_ptr(datatype, sreq->dev.datatype_ptr);
		MPID_Datatype_add_ref(sreq->dev.datatype_ptr);
	}
	*request = sreq;

  fn_exit:	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_BSEND_INIT);
	return mpi_errno;
}

/* 
 * FIXME: The ch3 implmentation of the persistent routines should
 * be very simple and use common code as much as possible.  All
 * persistent routine should be in the same file, along with 
 * startall.  Consider using function pointers to specify the 
 * start functions, as if these were generalized requests, 
 * rather than having MPID_Startall look at the request type.
 */
/*
 * MPID_Recv_init()
 */
#undef FUNCNAME
#define FUNCNAME MPID_Recv_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Recv_init(void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, MPID_Request ** request)
{
	MPID_Request * rreq;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_RECV_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_RECV_INIT);
	
	rreq = MPID_Request_create();
	if (rreq == NULL)
	{
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		/* --END ERROR HANDLING-- */
		goto fn_exit;
	}
	
	MPIU_Object_set_ref(rreq, 1);
	rreq->kind = MPID_PREQUEST_RECV;
	rreq->comm = comm;
	rreq->cc   = 0;
	MPIR_Comm_add_ref(comm);
	rreq->dev.match.rank = (int16_t)rank;
	rreq->dev.match.tag = tag;
	rreq->dev.match.context_id = (int16_t)(comm->recvcontext_id + context_offset);
	rreq->dev.user_buf = (void *) buf;
	rreq->dev.user_count = count;
	rreq->dev.datatype = datatype;
	rreq->partner_request = NULL;
	MPIDI_Request_set_type(rreq, MPIDI_REQUEST_TYPE_RECV);
	if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
	{
		MPID_Datatype_get_ptr(datatype, rreq->dev.datatype_ptr);
		MPID_Datatype_add_ref(rreq->dev.datatype_ptr);
	}
	*request = rreq;

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_RECV_INIT);
	return mpi_errno;
}


/*@
  MPID_VCRT_Create - Create a table of VC references

  Notes:
  This routine only provides space for the VC references.  Those should
  be added by assigning to elements of the vc array within the 
  'MPID_VCRT' object.
  @*/
#undef FUNCNAME
#define FUNCNAME MPID_VCRT_Create
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_VCRT_Create(int size, MPID_VCRT *vcrt_ptr)
{
	MPIDI_VCRT_t * vcrt;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_VCRT_CREATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_VCRT_CREATE);
	
	vcrt = MPIU_Malloc(sizeof(MPIDI_VCRT_t) + (size - 1) * sizeof(MPIDI_VC_t *));
	if (vcrt != NULL)
	{
		MPIU_Object_set_ref(vcrt, 1);
		vcrt->size = size;
		*vcrt_ptr = vcrt;
	}
	else
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**nomem");
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCRT_CREATE);
	return mpi_errno;
}

/*@
  MPID_VCRT_Add_ref - Add a reference to a VC reference table

  Notes:
  This is called when a communicator duplicates its group of processes.
  It is used in 'commutil.c' and in routines to create communicators from
  dynamic process operations.  It does not change the state of any of the
  virtural connections (VCs).
  @*/
#undef FUNCNAME
#define FUNCNAME MPID_VCRT_Add_ref
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_VCRT_Add_ref(MPID_VCRT vcrt)
{
	MPIDI_STATE_DECL(MPID_STATE_MPID_VCRT_ADD_REF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_VCRT_ADD_REF);
	MPIU_Object_add_ref(vcrt);
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCRT_ADD_REF);
	return MPI_SUCCESS;
}

/*@
  MPID_VCRT_Release - Release a reference to a VC reference table

  Notes:
  
  @*/
#undef FUNCNAME
#define FUNCNAME MPID_VCRT_Release
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_VCRT_Release(MPID_VCRT vcrt, int isDisconnect)
{
	int in_use;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPID_VCRT_RELEASE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_VCRT_RELEASE);

	/* If this VC reference table is no longer in use, we can
	   decrement the reference count of each of the VCs.  If the
	   count on the VCs goes to zero, then we can decrement the
	   ref count on the process group and so on. 
	*/
	MPIU_Object_release_ref(vcrt, &in_use);
	if (!in_use)
	{
		int i, inuse;

		/* FIXME: Need a better way to define how vc's are closed that 
		 takes into account pending operations on vcs, including 
		 close events received from other processes. */
		/*
		mpi_errno = MPIDI_CH3U_VC_FinishPending( vcrt );
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		*/

		for (i = 0; i < vcrt->size; i++)
		{
			MPIDI_VC_t * const vc = vcrt->vcr_table[i];
			
			MPIDI_VC_Release_ref(vc, &in_use);
			/* The rule for disconnect that we use is that if
			   MPI_Comm_disconnect removes the last reference to this
			   VC, we fully remove the VC.  This is not quite what the
			   MPI standard says, but this is sufficient to give the 
			   expected behavior for most user programs that
			   use MPI_Comm_disconnect */
			if (isDisconnect && vc->ref_count == 1)
			{
				MPIDI_VC_Release_ref(vc, &in_use);
			}

			if (!in_use)
			{
				/* If the VC is myself then skip the close message */
				if (vc->pg == MPIDI_Process.my_pg && 
					vc->pg_rank == MPIDI_Process.my_pg_rank)
				{
					MPIDI_PG_Release_ref(vc->pg, &inuse);
					if (inuse == 0)
					{
						MPIDI_PG_Destroy(vc->pg);
					}
					continue;
				}
				
				/* FIXME: the correct test is ACTIVE or REMOTE_CLOSE */
				if (vc->state != MPIDI_VC_STATE_INACTIVE)
				{
					/*printf("%s:%d SendClose pg=%p\n", __FILE__, __LINE__, vc->pg);fflush(stdout);*/
					MPIDI_CH3.VC_SendClose( vc, i );
				}
				else
				{
					if (vc->pg != NULL) /* The pg used in accept/connect is NULL */
					{
						MPIDI_PG_Release_ref(vc->pg, &inuse);
						if (inuse == 0)
						{
							MPIDI_PG_Destroy(vc->pg);
						}
					}

					MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"vc=%p: not sending a close to %d, vc in state %s", vc, i, MPIDI_VC_GetStateString(vc->state)));
				}
			}
		}

		MPIU_Free(vcrt);
	}

/*fn_exit:*/
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCRT_RELEASE);
	return mpi_errno;
	/*
 fn_fail:
	goto fn_exit;
	*/
}

#undef FUNCNAME
#define FUNCNAME MPID_VCRT_Get_ptr
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_VCRT_Get_ptr(MPID_VCRT vcrt, MPID_VCR **vc_pptr)
{
	MPIDI_STATE_DECL(MPID_STATE_MPID_VCRT_GET_PTR);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_VCRT_GET_PTR);
	*vc_pptr = vcrt->vcr_table;
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCRT_GET_PTR);
	return MPI_SUCCESS;
}

/*@
  MPID_VCR_Dup - Duplicate a virtual connection reference 

  Notes:
  If the VC is being used for the first time in a VC reference
  table, the reference count is set to two, not one, in order to
  distinquish between freeing a communicator with 'MPI_Comm_free' and
  'MPI_Comm_disconnect', and the reference count on the process group
  is incremented (to indicate that the process group is in use).
  While this has no effect on the process group of 'MPI_COMM_WORLD',
  it is important for process groups accessed through 'MPI_Comm_spawn'
  or 'MPI_Comm_connect/MPI_Comm_accept'.
  
  @*/
#undef FUNCNAME
#define FUNCNAME MPID_VCR_Dup
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_VCR_Dup(MPID_VCR orig_vcr, MPID_VCR * new_vcr)
{
	MPIDI_STATE_DECL(MPID_STATE_MPID_VCR_DUP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_VCR_DUP);
	/* We are allowed to create a vc that belongs to no process group 
	 as part of the initial connect/accept action, so in that case,
	 ignore the pg ref count update */
	if (orig_vcr->ref_count == 0 && orig_vcr->pg)
	{
		MPIU_Object_set_ref( orig_vcr, 2 );
		MPIDI_PG_Add_ref( orig_vcr->pg );
	}
	else
	{
		MPIU_Object_add_ref(orig_vcr);
	}
	MPIU_DBG_MSG_FMT(REFCOUNT,TYPICAL,(MPIU_DBG_FDEST, "Incr VCR %p ref count to %d",orig_vcr,orig_vcr->ref_count));
	*new_vcr = orig_vcr;
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCR_DUP);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPID_VCR_Get_lpid
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_VCR_Get_lpid(MPID_VCR vcr, int * lpid_ptr)
{
	MPIDI_STATE_DECL(MPID_STATE_MPID_VCR_GET_LPID);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_VCR_GET_LPID);
	*lpid_ptr = vcr->lpid;
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCR_GET_LPID);
	return MPI_SUCCESS;
}

/* 
 * The following routines convert to/from the global pids, which are 
 * represented as pairs of ints (process group id, rank in that process group)
 */

/* FIXME: These routines belong in a different place */
#undef FUNCNAME
#define FUNCNAME MPID_GPID_GetAllInComm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_GPID_GetAllInComm( MPID_Comm *comm_ptr, int local_size, int local_gpids[], int *singlePG )
{
	int mpi_errno = MPI_SUCCESS;
	int i;
	int *gpid = local_gpids;
	int lastPGID = -1, pgid;
	MPID_VCR vc;
	MPIDI_STATE_DECL(MPID_STATE_MPID_GPID_GETALLINCOMM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_GPID_GETALLINCOMM);

	MPIU_UNREFERENCED_ARG(local_size);

	*singlePG = 1;
	for (i=0; i<comm_ptr->local_size; i++)
	{
		vc = comm_ptr->vcr[i];

		/* Get the process group id as an int */
		MPIDI_PG_IdToNum( vc->pg, &pgid );

		*gpid++ = pgid;
		if (lastPGID != pgid)
		{
			if (lastPGID != -1)
				*singlePG = 0;
			lastPGID = pgid;
		}
		*gpid++ = vc->pg_rank;
		if (vc->pg_rank != vc->lpid)
		{
			MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**fail");
			goto fn_fail;
/*			wprintf( L"Unexpected results %d != %d\n",
			vc->pg_rank, vc->lpid ); */
		}
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_GPID_GETALLINCOMM);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPID_GPID_Get
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_GPID_Get( MPID_Comm *comm_ptr, int rank, int gpid[] )
{
	int	  pgid;
	MPID_VCR vc;
	MPIDI_STATE_DECL(MPID_STATE_MPID_GPID_GET);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_GPID_GET);
	
	vc = comm_ptr->vcr[rank];

	/* Get the process group id as an int */
	MPIDI_PG_IdToNum( vc->pg, &pgid );

	gpid[0] = pgid;
	gpid[1] = vc->pg_rank;
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_GPID_GET);
	return MPI_SUCCESS;
}

/* 
 * The following is a very simple code for looping through 
 * the GPIDs.  Note that this code requires that all processes
 * have information on the process groups.
 */
#undef FUNCNAME
#define FUNCNAME MPID_GPID_ToLpidArray
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_GPID_ToLpidArray( int size, int gpid[], int lpid[] )
{
	int i, mpi_errno = MPI_SUCCESS;
	int pgid;
	MPIDI_PG_t *pg = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPID_GPID_TOLPIDARRAY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_GPID_TOLPIDARRAY);

	for (i=0; i<size; i++)
	{
		MPIDI_PG_Iterate_reset();
		for (;;)
		{
			MPIDI_PG_Get_next( &pg );
			if (!pg)
			{
				/* Internal error.  This gpid is unknown on this process */
				lpid[i] = -1;
				MPIU_ERR_SET2(mpi_errno,MPI_ERR_INTERN, L"**unknowngpid", L"**unknowngpid %d %d", gpid[0], gpid[1] );
				MPIDI_FUNC_EXIT(MPID_STATE_MPID_GPID_TOLPIDARRAY);
				return mpi_errno;
			}
			MPIDI_PG_IdToNum( pg, &pgid );
/*			wprintf( L"Pg id = %d in %s\n", pgid, (char *)pg->id );fflush(stdout); */
			if (pgid == gpid[0])
			{
				/* found the process group.  gpid[1] is the rank in this process group */
				/* Sanity check on size */
				if (pg->size > gpid[1])
				{
					lpid[i] = pg->vct[gpid[1]].lpid;
				}
				else
				{
					lpid[i] = -1;
					MPIU_ERR_SET2(mpi_errno,MPI_ERR_INTERN, L"**unknowngpid", L"**unknowngpid %d %d", gpid[0], gpid[1] );
					MPIDI_FUNC_EXIT(MPID_STATE_MPID_GPID_TOLPIDARRAY);
					return mpi_errno;
				}
				/* wprintf( L"lpid[%d] = %d for gpid = (%d)%d\n", i, lpid[i], gpid[0], gpid[1] ); */
				break;
			}
		}
		gpid += 2;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_GPID_TOLPIDARRAY);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPID_VCR_CommFromLpids
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_VCR_CommFromLpids( MPID_Comm *newcomm_ptr, int size, const int lpids[] )
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *commworld_ptr;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPID_VCR_COMMFROMLPIDS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_VCR_COMMFROMLPIDS);

	commworld_ptr = MPIR_Process.comm_world;
	/* Setup the communicator's vc table: remote group */
	MPID_VCRT_Create( size, &newcomm_ptr->vcrt );
	MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
	for (i=0; i<size; i++)
	{
		MPIDI_VC_t *vc = 0;

		/* For rank i in the new communicator, find the corresponding
		   virtual connection.  For lpids less than the size of comm_world,
		   we can just take the corresponding entry from comm_world.
		   Otherwise, we need to search through the process groups.
		*/
		/* wprintf( L"[%d] Remote rank %d has lpid %d\n", 
		   MPIR_Process.comm_world->rank, i, lpids[i] ); */
		if (lpids[i] < commworld_ptr->remote_size)
		{
			vc = commworld_ptr->vcr[lpids[i]];
		}
		else
		{
			/* We must find the corresponding vcr for a given lpid */		
			/* For now, this means iterating through the process groups */
			MPIDI_PG_t *pg = 0;
			int j;

			MPIDI_PG_Iterate_reset();
			/* Skip comm_world */
			MPIDI_PG_Get_next( &pg );
			do
			{
				MPIDI_PG_Get_next( &pg );
				if (!pg)
				{
					mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, L"MPID_VCR_CommFromLpids", __LINE__, MPI_ERR_INTERN, L"**intern", 0 );
					MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCR_COMMFROMLPIDS);
					return mpi_errno;
				}
				/* FIXME: a quick check on the min/max values of the lpid
				   for this process group could help speed this search */
				for (j=0; j<pg->size; j++)
				{
					/*wprintf( L"Checking lpid %d against %d in pg %s\n",
							lpids[i], pg->vct[j].lpid, (char *)pg->id );
							fflush(stdout); */
					if (pg->vct[j].lpid == lpids[i])
					{
						vc = &pg->vct[j];
						/*wprintf( L"found vc %x for lpid = %d in another pg\n", 
						  (int)vc, lpids[i] );*/
						break;
					}
				}
			} while (!vc);
		}

		/* wprintf( L"about to dup vc %x for lpid = %d in another pg\n", (int)vc, lpids[i] ); */
		MPID_VCR_Dup( vc, &newcomm_ptr->vcr[i] );
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_VCR_COMMFROMLPIDS);
	return mpi_errno;
}

/* The following is a temporary hook to ensure that all processes in 
   a communicator have a set of process groups.
 
   All arguments are input (all processes in comm must have gpids)

   First: all processes check to see if they have information on all
   of the process groups mentioned by id in the array of gpids.

   The local result is LANDed with Allreduce.
   If any process is missing process group information, then the
   root process broadcasts the process group information as a string; 
   each process then uses this information to update to local process group
   information (in the KVS cache that contains information about 
   contacting any process in the process groups).
*/
#undef FUNCNAME
#define FUNCNAME MPID_PG_ForwardPGInfo
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_PG_ForwardPGInfo( MPID_Comm *peer_ptr, MPID_Comm *comm_ptr, int nPGids, const int gpids[], int root )
{
	int mpi_errno = MPI_SUCCESS;
	int i, allfound = 1, pgid, pgidWorld;
	MPIDI_PG_t *pg = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPID_PG_FORWARDPGINFO);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_PG_FORWARDPGINFO);

	/* Get the pgid for CommWorld (always attached to the first process group) */
	mpi_errno = MPIDI_PG_Iterate_reset();
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = MPIDI_PG_Get_next( &pg );
	MPIU_ERR_CHK_POP(mpi_errno);
	MPIDI_PG_IdToNum( pg, &pgidWorld );

	/* Extract the unique process groups */
	for (i=0; i<nPGids && allfound; i++)
	{
		if (gpids[0] != pgidWorld)
		{
			/* Add this gpid to the list of values to check */
			/* FIXME: For testing, we just test in place */
			mpi_errno = MPIDI_PG_Iterate_reset();
			MPIU_ERR_CHK_POP(mpi_errno);
			do
			{
				mpi_errno = MPIDI_PG_Get_next( &pg );
				MPIU_ERR_CHK_POP(mpi_errno);
				if (!pg)
				{
					/* We don't know this pgid */
					allfound = 0;
					break;
				}
				MPIDI_PG_IdToNum( pg, &pgid );
			} while (pgid != gpids[0]);
		}
		gpids += 2;
	}

	/* See if everyone is happy */
	NMPI_Allreduce(mpi_errno, ( MPI_IN_PLACE, &allfound, 1, MPI_INT, MPI_LAND, comm_ptr->handle ));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (allfound)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPID_PG_FORWARDPGINFO);
		return MPI_SUCCESS;
	}

	/* We need to share the process groups.  We use routines from ch3u_port.c */
	mpi_errno = MPID_PG_BCast( peer_ptr, comm_ptr, root );
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_PG_FORWARDPGINFO);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_VC_FinishPending
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3U_VC_FinishPending( MPIDI_VCRT_t *vcrt )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_VC_t **vc;
	int i, size, nPending;
	MPID_Progress_state progress_state; 
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_VC_FINISHPENDING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_VC_FINISHPENDING);

	do
	{
		/* Compute the number of pending ops.
		   A virtual connection has pending operations if the state
		   is not INACTIVE or if the sendq is not null */
		nPending = 0;
		vc	   = vcrt->vcr_table;
		size	 = vcrt->size;
		/* wprintf( L"Size = %d\n", size ); fflush(stdout); */
		for (i=0; i<size; i++)
		{
			if (vc[i]->state != MPIDI_VC_STATE_INACTIVE)
			{
				wprintf (L"state for vc[%d] is %d\n", i, vc[i]->state ); fflush(stdout);
				nPending++;
			}
			if (vc[i]->ch.sendq_head)
			{
				wprintf( L"Nonempty sendQ for vc[%d]\n", i ); fflush(stdout);
				nPending++;
			}
		}
		if (nPending > 0)
		{
			wprintf( L"Panic! %d pending operations!\n", nPending );
			fflush(stdout);
			MPIU_Assert( nPending == 0 );
		}
		else
		{
			break;
		}

		MPID_Progress_start(&progress_state);
		MPIU_DBG_MSG_D(CH3_CONNECT,VERBOSE,L"Waiting for %d close operations", nPending);
		mpi_errno = MPID_Progress_wait(&progress_state);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPID_Progress_end(&progress_state);
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|close_progress");
		}
		/* --END ERROR HANDLING-- */
		MPID_Progress_end(&progress_state);
	} while(nPending > 0);

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_VC_FINISHPENDING);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/*
 * MPIDI_CH3U_Comm_FinishPending - Complete any pending operations on the 
 * communicator.  
 *
 * Notes: 
 * This should be used before freeing or disconnecting a communicator.
 *
 * For better scalability, we might want to form a list of VC's with 
 * pending operations.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Comm_FinishPending
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Comm_FinishPending( MPID_Comm *comm_ptr )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_COMM_FINISHPENDING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_COMM_FINISHPENDING);

	mpi_errno = MPIDI_CH3U_VC_FinishPending( comm_ptr->vcrt );
	MPIU_ERR_CHK_POP(mpi_errno);
	if (!mpi_errno && comm_ptr->local_vcrt)
	{
		mpi_errno = MPIDI_CH3U_VC_FinishPending( comm_ptr->local_vcrt );
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_COMM_FINISHPENDING);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* ----------------------------------------------------------------------- */
/* Routines to initialize a VC */

int MPIDI_VC_Init( MPIDI_VC_t *vc, MPIDI_PG_t *pg, int rank )
{
	vc->state = MPIDI_VC_STATE_INACTIVE;
	MPIU_Object_set_ref(vc, 0);
	vc->handle  = MPID_VCONN;
	vc->pg	  = pg;
	vc->pg_rank = rank;
	vc->lpid = lpid_counter++;
	/*
    vc->rndvSend_fn      = MPIDI_CH3_RndvSend;
    vc->rndvRecv_fn      = MPIDI_CH3_RecvRndv;
    vc->eager_max_msg_sz = MPIDI_CH3_EAGER_MAX_MSG_SIZE;
    vc->sendNoncontig_fn = MPIDI_CH3_SendNoncontig_iov;
	*/
	MPIDI_CH3.VC_Init(vc);
	MPIU_DBG_PrintVCState(vc);

	return MPI_SUCCESS;
}


/* FIXME: Explain this function */

/* FIXME: should there be a simpler version of this for short messages, 
   particularly contiguous ones?  See also the FIXME about buffering
   short messages */

#undef FUNCNAME
#define FUNCNAME MPIDI_Isend_self
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Isend_self(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset, int type, MPID_Request ** request)
{
	MPIDI_Message_match match;
	MPID_Request * sreq = NULL;
	MPID_Request * rreq;
	MPIDI_VC_t * vc;
	int found;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_ISEND_SELF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_ISEND_SELF);
		
	MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending message to self");
		
	MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit);
	MPIDI_Request_set_type(sreq, type);
	MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_SELF_MSG);
	
	match.rank = (int16_t)rank;
	match.tag = tag;
	match.context_id = (int16_t)(comm->context_id + context_offset);
	rreq = MPIDI_CH3U_Recvq_FDP_or_AEU(&match, &found);
	/* --BEGIN ERROR HANDLING-- */
	if (rreq == NULL)
	{
		MPIU_Object_set_ref(sreq, 0);
		MPIDI_CH3_Request_destroy(sreq);
		sreq = NULL;
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_NO_MEM, L"**nomem", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

	MPIDI_Comm_get_vc(comm, rank, &vc);
	
	rreq->status.MPI_SOURCE = rank;
	rreq->status.MPI_TAG = tag;
	
	if (found)
	{
		MPIDI_msg_sz_t data_sz;
		
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"found posted receive request; copying data");
			
		MPIDI_CH3U_Buffer_copy(buf, count, datatype, &sreq->status.MPI_ERROR,
							   rreq->dev.user_buf, rreq->dev.user_count, rreq->dev.datatype, &data_sz, &rreq->status.MPI_ERROR);
		rreq->status.count = (int)data_sz;
		MPID_Request_set_completed(rreq);
		MPID_Request_release(rreq);
		/* sreq has never been seen by the user or outside this thread, so it is safe to reset ref_count and cc */
		MPIU_Object_set_ref(sreq, 1);
		sreq->cc = 0;
	}
	else
	{
		if (type != MPIDI_REQUEST_TYPE_RSEND)
		{
			int dt_sz;
		
			/* FIXME: Insert code here to buffer small sends in a temporary buffer? */

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"added receive request to unexpected queue; attaching send request");
			if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
			{
				MPID_Datatype_get_ptr(datatype, sreq->dev.datatype_ptr);
				MPID_Datatype_add_ref(sreq->dev.datatype_ptr);
			}
			rreq->partner_request = sreq;
			rreq->dev.sender_req_id = sreq->handle;
			MPID_Datatype_get_size_macro(datatype, dt_sz);
			rreq->status.count = count * dt_sz;
		}
		else
		{
			/* --BEGIN ERROR HANDLING-- */
			MPIU_DBG_MSG(CH3_OTHER,TYPICAL, L"ready send unable to find matching recv req");
			sreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**rsendnomatch", L"**rsendnomatch %d %d", rank, tag);
			rreq->status.MPI_ERROR = sreq->status.MPI_ERROR;
			
			rreq->partner_request = NULL;
			rreq->dev.sender_req_id = MPI_REQUEST_NULL;
			rreq->status.count = 0;
			
			/* sreq has never been seen by the user or outside this thread, so it is safe to reset ref_count and cc */
			MPIU_Object_set_ref(sreq, 1);
			sreq->cc = 0;
			/* --END ERROR HANDLING-- */
		}
			
		MPIDI_Request_set_msg_type(rreq, MPIDI_REQUEST_SELF_MSG);
		MPIDI_CH3_Progress_signal_completion();
	}

  fn_exit:
	*request = sreq;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_ISEND_SELF);
	return mpi_errno;
}


/* FIXME: These routines need a description.  What is their purpose?  Who
   calls them and why?  What does each one do?
*/
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_Init(int *argc_p, wchar_t ***argv_p, MPIDI_PG_Compare_ids_fn_t compare_ids_fn, MPIDI_PG_Destroy_fn_t destroy_fn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_INIT);
	
	MPIU_UNREFERENCED_ARG(argc_p);
	MPIU_UNREFERENCED_ARG(argv_p);

	MPIDI_PG_Compare_ids_fn = compare_ids_fn;
	MPIDI_PG_Destroy_fn	 = destroy_fn;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_INIT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@ 
   MPIDI_PG_Finalize - Finalize the process groups, including freeing all
   process group structures
  @*/
int MPIDI_PG_Finalize(void)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_PG_t *pg, *pgNext;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_FINALIZE);

	/* FIXME - straighten out the use of PG_Finalize - no use after PG_Finalize */
	if (pg_world->connData)
	{
		int rc;
		rc = PMI_Finalize();
		if (rc)
		{
			MPIU_ERR_SET2(mpi_errno,MPI_ERR_OTHER, L"**ch3|pmi_finalize", L"**ch3|pmi_finalize %d %s", rc, PMI_Get_last_error_msgW());
		}
	}

	/* Free the storage associated with the process groups */
	pg = MPIDI_PG_list;
	while (pg)
	{
		pgNext = pg->next;

		/* In finalize, we free all process group information, even if
		   the ref count is not zero.  This can happen if the user
		   fails to use MPI_Comm_disconnect on communicators that
		   were created with the dynamic process routines.*/
		/*if (pg->ref_count == 0 || 1)*/
		{
			if (pg == MPIDI_Process.my_pg)
				MPIDI_Process.my_pg = NULL;
			MPIDI_PG_Destroy(pg);
		}
		pg	 = pgNext;
	}

	/* If COMM_WORLD is still around (it normally should be), 
	   try to free it here.  The reason that we need to free it at this 
	   point is that comm_world (and comm_self) still exist, and 
	   hence the usual process to free the related VC structures will
	   not be invoked. */
	if (MPIDI_Process.my_pg)
	{
		MPIDI_PG_Destroy(MPIDI_Process.my_pg);
	} 
	MPIDI_Process.my_pg = NULL;

	/* ifdefing out this check because the list will not be NULL in 
	   Ch3_finalize because
	   one additional reference is retained in MPIDI_Process.my_pg. 
	   That reference is released
	   only after ch3_finalize returns. If I release it before ch3_finalize, 
	   the ssm channel crashes. */
#if 0

	if (MPIDI_PG_list != NULL)
	{ 
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN,
		L"**dev|pg_finalize|list_not_empty", NULL); 
		/* --END ERROR HANDLING-- */
	}
#endif

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_FINALIZE);
	return mpi_errno;
}

/* FIXME: This routine needs to make it clear that the pg_id, for example
   is saved; thus, if the pg_id is a string, then that string is not 
   copied and must be freed by a PG_Destroy routine */

/* This routine creates a new process group description and appends it to 
   the list of the known process groups.  The pg_id is saved, not copied.
   The PG_Destroy routine that was set with MPIDI_PG_Init is responsible for
   freeing any storage associated with the pg_id. 

   The new process group is returned in pg_ptr 
*/
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Create
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_Create(int vct_sz, const void * pg_id, MPIDI_PG_t ** pg_ptr)
{
	MPIDI_PG_t * pg = NULL, *pgnext;
	int p;
	int mpi_errno = MPI_SUCCESS;
	MPIU_CHKPMEM_DECL(2);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_CREATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_CREATE);

	MPIU_CHKPMEM_MALLOC(pg,MPIDI_PG_t*,sizeof(MPIDI_PG_t),mpi_errno,L"pg");
	MPIU_CHKPMEM_MALLOC(pg->vct,MPIDI_VC_t *,sizeof(MPIDI_VC_t)*vct_sz, mpi_errno,L"pg->vct");

	pg->handle = 0;
	/* The reference count indicates the number of vc's that are or 
	   have been in use and not disconnected. It starts at zero,
	   except for MPI_COMM_WORLD. */
	MPIU_Object_set_ref(pg, 0);
	pg->size = vct_sz;
	pg->id = pg_id;
	/* Initialize the connection information to null.  Use
	   the appropriate MPIDI_PG_InitConnXXX routine to set up these 
	   fields */
	pg->connData		   = 0;
	pg->getConnInfo		= 0;
	pg->connInfoToString   = 0;
	pg->connInfoFromString = 0;
	pg->freeConnInfo	   = 0;

	for (p = 0; p < vct_sz; p++)
	{
		/* Initialize device fields in the VC object */
		MPIDI_VC_Init(&pg->vct[p], pg, p);
	}

	/* The first process group is always the world group */
	if (!pg_world) { pg_world = pg; }
	/* Add pg's at the tail so that comm world is always the first pg */
	pg->next = 0;
	if (!MPIDI_PG_list)
	{
		MPIDI_PG_list = pg;
	}
	else
	{
		pgnext = MPIDI_PG_list; 
		while (pgnext->next)
		{
			pgnext = pgnext->next;
		}
		pgnext->next = pg;
	}
	*pg_ptr = pg;
	
  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_CREATE);
	return mpi_errno;
	
  fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Destroy
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_Destroy(MPIDI_PG_t * pg)
{
	MPIDI_PG_t * pg_prev;
	MPIDI_PG_t * pg_cur;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_DESTROY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_DESTROY);

	pg_prev = NULL;
	pg_cur = MPIDI_PG_list;
	while(pg_cur != NULL)
	{
		if (pg_cur == pg)
		{
			if (MPIDI_PG_iterator_next == pg)
			{ 
				MPIDI_PG_iterator_next = MPIDI_PG_iterator_next->next;
			}

			if (pg_prev == NULL)
				MPIDI_PG_list = pg->next; 
			else
				pg_prev->next = pg->next;

			MPIDI_PG_Destroy_fn(pg);
			MPIU_Free(pg->vct);
			if (pg->connData != NULL)
			{
				if (pg->freeConnInfo)
					(*pg->freeConnInfo)( pg );
				else
					MPIU_Free(pg->connData);
				pg->connData = NULL;
			}
			MPIU_Free(pg);

			goto fn_exit;
		}

		pg_prev = pg_cur;
		pg_cur = pg_cur->next;
	}

	/* PG not found if we got here */
	MPIU_ERR_SET1(mpi_errno,MPI_ERR_OTHER, L"**dev|pg_not_found", L"**dev|pg_not_found %p", pg);

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_DESTROY);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Find
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_Find(const void * id, MPIDI_PG_t ** pg_ptr)
{
	MPIDI_PG_t * pg;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_FIND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_FIND);
	
	pg = MPIDI_PG_list;
	while (pg != NULL)
	{
		if (MPIDI_PG_Compare_ids_fn(id, pg->id) != FALSE)
		{
			*pg_ptr = pg;
			goto fn_exit;
		}

		pg = pg->next;
	}

	*pg_ptr = NULL;

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_FIND);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Id_compare
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_Id_compare(const void * id1, const void *id2)
{
	int result;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_ID_COMPARE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_ID_COMPARE);
	result = MPIDI_PG_Compare_ids_fn(id1, id2);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_ID_COMPARE);
	return result;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Get_next
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_Get_next(MPIDI_PG_t ** pg_ptr)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_GET_NEXT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_GET_NEXT);
	*pg_ptr = MPIDI_PG_iterator_next;
	if (MPIDI_PG_iterator_next != NULL)
	{ 
		MPIDI_PG_iterator_next = MPIDI_PG_iterator_next->next;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_GET_NEXT);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Iterate_reset
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_Iterate_reset()
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_ITERATE_RESET);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_ITERATE_RESET);
	MPIDI_PG_iterator_next = MPIDI_PG_list;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_ITERATE_RESET);
	return MPI_SUCCESS;
}

/* Note: Allocated memory that is returned in str_ptr.  The user of this routine must free that data */
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_To_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_To_string(MPIDI_PG_t *pg_ptr, wchar_t **str_ptr, int *lenStr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_TO_STRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_TO_STRING);

	/* Replace this with the new string */
	if (pg_ptr->connInfoToString)
	{
		(*pg_ptr->connInfoToString)( str_ptr, lenStr, pg_ptr );
	}
	else
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_INTERN,L"**noConnInfoToString");
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_TO_STRING);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* This routine takes a string description of a process group (created with 
   MPIDI_PG_To_string, usually on a different process) and returns a pointer to
   the matching process group.  If the group already exists, flag is set to 
   false.  If the group does not exist, it is created with MPIDI_PG_Create (and
   hence is added to the list of active process groups) and flag is set to 
   true.  In addition, the connection information is set up using the 
   information in the input string.
*/
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Create_from_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_Create_from_string(const wchar_t * str, MPIDI_PG_t ** pg_pptr, int *flag)
{
	int mpi_errno = MPI_SUCCESS;
	const wchar_t *p;
	int vct_sz;
	MPIDI_PG_t *existing_pg, *pg_ptr=0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_CREATE_FROM_STRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_CREATE_FROM_STRING);

	/* The pg_id is at the beginning of the string, so we can just pass
	   it to the find routine */
	/* wprintf( L"Looking for pg with id %s\n", str );fflush(stdout); */
	mpi_errno = MPIDI_PG_Find((void*)str, &existing_pg);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (existing_pg != NULL)
	{
		/* return the existing PG */
		*pg_pptr = existing_pg;
		*flag = 0;
		/* Note that the memory for the pg_id is freed in the exit */
		goto fn_exit;
	}
	*flag = 1;

	/* Get the size from the string */
	p = str;
	while (*p) p++; p++;
	vct_sz = _wtoi(p);

	mpi_errno = MPIDI_PG_Create(vct_sz, (void *)str, pg_pptr);
	MPIU_ERR_CHK_POP(mpi_errno);
	
	pg_ptr = *pg_pptr;
	pg_ptr->id = MPIU_Strdup( str );
	
	/* Set up the functions to use strings to manage connection information */
	MPIDI_PG_InitConnString( pg_ptr );
	/*wprintf(L"Creating PG from string: <%s>\n", str);fflush(stdout);*/
	(*pg_ptr->connInfoFromString)( str, pg_ptr );

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_CREATE_FROM_STRING);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_PG_IdToNum
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_PG_IdToNum( MPIDI_PG_t *pg, int *id )
{
	const wchar_t *p = (const wchar_t *)pg->id;
	int pgid = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_IDTONUM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_IDTONUM);

	while (*p && !isdigit(*p)) p++;
	if (!*p)
	{
		p = (const wchar_t *)pg->id;
		while (*p)
		{
			pgid += *p - L' ';
		}
		pgid = pgid ^ 0x100;
	}
	else
	{
		/* mpd uses (pid_num) as part of the kvs name, so we skip over - and _ */
		while (*p)
		{
			if (isdigit(*p))
			{
				pgid = pgid * 10 + (*p - L'0');
			}
			else if (*p != L'-' && *p != L'_')
			{
				break;
			}
			p++;
		}
	}
	*id = pgid;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_IDTONUM);
}

/*
 * Managing connection information for process groups
 * 
 *
 */

/* Setting a process's connection information 
   
   This is a collective call (for scalability) over all of the processes in 
   the same MPI_COMM_WORLD.
*/
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_SetConnInfo
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_SetConnInfo( int rank, const wchar_t *connString )
{
	int mpi_errno = MPI_SUCCESS;
	int pmi_errno;
	int len;
#define BC_KEY_LENGTH 128
	wchar_t key[BC_KEY_LENGTH];
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_SETCONNINFO);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_SETCONNINFO);

	MPIU_Assert(pg_world->connData);
	
	len = MPIU_Snprintf(key, BC_KEY_LENGTH, _TRUNCATE, L"P%d-businesscard", rank);
	if (len < 0 || len >= BC_KEY_LENGTH)
	{
		MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**_snwprintf", L"**_snwprintf %d", len);
	}
	/*wprintf(L"Putting my connString: <%s>\n", connString);fflush(stdout);*/
	pmi_errno = PMI_KVS_PutW(pg_world->connData, key, connString );
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_kvs_put", L"**pmi_kvs_put %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}
	pmi_errno = PMI_KVS_CommitW(pg_world->connData);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_kvs_commit", L"**pmi_kvs_commit %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}
	
	pmi_errno = PMI_Barrier();
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_barrier", L"**pmi_barrier %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}
 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_SETCONNINFO);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* For all of these routines, the format of the process group description
   that is created and used by the connTo/FromString routines is this:
   (All items are strings, terminated by null)

   process group id string
   sizeof process group (as string)
   conninfo for rank 0
   conninfo for rank 1
   ...

   The "conninfo for rank 0" etc. for the original (MPI_COMM_WORLD)
   process group are stored in the PMI_KVS space with the keys 
   p<rank>-businesscard .  

   Fixme: Add a routine to publish the connection info to this file so that
   the key for the businesscard is defined in just this one file.
*/


/* The "KVS" versions are for the process group to which the calling 
   process belongs.  These use the PMI_KVS routines to access the
   process information */
#undef FUNCNAME
#define FUNCNAME getConnInfoKVS
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int getConnInfoKVS( int rank, wchar_t *buf, int bufsize, MPIDI_PG_t *pg )
{
	wchar_t key[MPIDI_MAX_KVS_KEY_LEN];
	int  mpi_errno = MPI_SUCCESS, rc, pmi_errno;
	MPIDI_STATE_DECL(MPID_STATE_GETCONNINFOKVS);

	MPIDI_FUNC_ENTER(MPID_STATE_GETCONNINFOKVS);

	rc = MPIU_Snprintf(key, MPIDI_MAX_KVS_KEY_LEN, _TRUNCATE, L"P%d-businesscard", rank );
	if (rc < 0 || rc > MPIDI_MAX_KVS_KEY_LEN)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem");
	}
	/*wprintf(L"Getting %s:%s\n", pg->connData, key);*/
	pmi_errno = PMI_KVS_GetW(pg->connData, key, buf, bufsize );
	if (pmi_errno)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER,L"**pmi_kvs_get", L"**pmi_kvs_get %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}
	/*wprintf(L"got connData for rank %d: <%s>\n", rank, buf);fflush(stdout);*/

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_GETCONNINFOKVS);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME connToStringKVS
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connToStringKVS( wchar_t **buf_p, int *slen, MPIDI_PG_t *pg )
{
	wchar_t *string = NULL;
	wchar_t *pg_idStr = (wchar_t *)pg->id;	  /* In the PMI/KVS space,
											 the pg id is a string */
	wchar_t buf[MPIDI_MAX_KVS_VALUE_LEN];
	int   i, j, vallen, rc, mpi_errno = MPI_SUCCESS, len;
	int   curSlen;
	MPIDI_STATE_DECL(MPID_STATE_CONNTOSTRINGKVS);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNTOSTRINGKVS);

	/* Make an initial allocation of a string with an estimate of the
	   needed space */
	len = 0;
	curSlen = 10 + pg->size * 128;
	string = (wchar_t *)MPIU_Malloc( curSlen * sizeof(wchar_t) );

	/* Start with the id of the pg */
	while (*pg_idStr && len < curSlen) 
		string[len++] = *pg_idStr++;
	string[len++] = '\0';
	
	/* Add the size of the pg */
	MPIU_Snprintf( &string[len], curSlen - len, _TRUNCATE, L"%d", pg->size );
	while (string[len]) len++;
	len++;

	for (i=0; i<pg->size; i++)
	{
		rc = getConnInfoKVS( i, buf, MPIDI_MAX_KVS_VALUE_LEN, pg );
		MPIU_ERR_CHK_POP(rc);
		/* Add the information to the output buffer */
		vallen = (int)wcslen(buf);
		/* Check that this will fix in the remaining space */
		if (len + vallen + 1 >= curSlen)
		{
			wchar_t *nstring = 0;
			nstring = MPIU_Realloc( string, (curSlen + (pg->size - i) * (vallen + 1 )) * sizeof(wchar_t));
			if (!nstring)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem");
			}
			/* This curSlen update line was removed in Argonne's code. Why? */
			curSlen = curSlen + (pg->size - i) * (vallen + 1 );
			string = nstring;
		}
		/* Append to string */
		for (j=0; j<vallen+1; j++)
		{
			string[len++] = buf[j];
		}
	}

	*buf_p = string;
	*slen  = len;
 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_CONNTOSTRINGKVS);
	return mpi_errno;
 fn_fail:
	if (string) MPIU_Free(string);
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME connFromStringKVS
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connFromStringKVS( const wchar_t *buf, MPIDI_PG_t *pg )
{
	MPIDI_STATE_DECL(MPID_STATE_CONNFROMSTRINGKVS);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNFROMSTRINGKVS);
	MPIU_UNREFERENCED_ARG(buf);
	MPIU_UNREFERENCED_ARG(pg);
	/* Fixme: this should be a failure to call this routine */
	MPIDI_FUNC_EXIT(MPID_STATE_CONNFROMSTRINGKVS);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME connFreeKVS
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connFreeKVS( MPIDI_PG_t *pg )
{
	MPIDI_STATE_DECL(MPID_STATE_CONNFREEKVS);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNFREEKVS);
	if (pg->connData)
	{
		MPIU_Free( pg->connData );
		pg->connData = NULL;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_CONNFREEKVS);
	return MPI_SUCCESS;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_PG_InitConnKVS
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_InitConnKVS( MPIDI_PG_t *pg )
{
	int pmi_errno, kvs_name_sz;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_INITCONNKVS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_INITCONNKVS);

	pmi_errno = PMI_KVS_Get_name_length_max( &kvs_name_sz );
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_kvs_get_name_length_max", L"**pmi_kvs_get_name_length_max %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}
	
	pg->connData = (wchar_t *)MPIU_Malloc((kvs_name_sz + 1)*sizeof(wchar_t));
	if (pg->connData == NULL)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomem");
	}
	
	pmi_errno = PMI_KVS_Get_my_nameW(pg->connData, kvs_name_sz);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_kvs_get_my_name", L"**pmi_kvs_get_my_name %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	pg->getConnInfo			= getConnInfoKVS;
	pg->connInfoToString	= connToStringKVS;
	pg->connInfoFromString	= connFromStringKVS;
	pg->freeConnInfo		= connFreeKVS;

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_INITCONNKVS);
	return mpi_errno;
 fn_fail:
	if (pg->connData)
	{
		MPIU_Free(pg->connData);
		pg->connData = NULL;
	}
	goto fn_exit;
}

/* Return the kvsname associated with the MPI_COMM_WORLD of this process. */
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_GetConnKVSname
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_GetConnKVSname( wchar_t ** kvsname )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_GETCONNKVSNAME);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_GETCONNKVSNAME);
	*kvsname = pg_world->connData;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_GETCONNKVSNAME);
	return MPI_SUCCESS;
}

/* For process groups that are not our MPI_COMM_WORLD, store the connection
   information in an array of strings.  These routines and structure
   implement the access to this information. */

#undef FUNCNAME
#define FUNCNAME getConnInfo
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int getConnInfo( int rank, wchar_t *buf, int bufsize, MPIDI_PG_t *pg )
{
	MPIDI_ConnInfo *connInfo = (MPIDI_ConnInfo *)pg->connData;
	MPIDI_STATE_DECL(MPID_STATE_GETCONNINFO);

	MPIDI_FUNC_ENTER(MPID_STATE_GETCONNINFO);

	/* wprintf( L"Entering getConnInfo\n" ); fflush(stdout); */
	if (!connInfo || !connInfo->connStrings || !connInfo->connStrings[rank])
	{
		/* FIXME: Turn this into a valid error code create/return */
		wprintf( L"Fatal error in getConnInfo (rank = %d)\n", rank );
		wprintf( L"connInfo = %p\n", connInfo );fflush(stdout);
		if (connInfo)
		{
			wprintf( L"connInfo->connStrings = %p\n", connInfo->connStrings );
		}
		/* Fatal error.  Connection information missing */
		fflush(stdout);   
	}

	/*wprintf(L"Copying <%s> to buf\n", connInfo->connStrings[rank] ); fflush(stdout);*/
	MPIU_Wcsncpy( buf, connInfo->connStrings[rank], bufsize );
	/*wprintf(L"Result of copy: <%s>\n", buf);fflush(stdout);*/
	MPIDI_FUNC_EXIT(MPID_STATE_GETCONNINFO);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME connToString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connToString( wchar_t **buf_p, int *slen, MPIDI_PG_t *pg )
{
	int mpi_errno = MPI_SUCCESS;
	wchar_t *string = NULL, *str;
	const wchar_t *pg_id;
	int  i, len=0;
	MPIDI_ConnInfo *connInfo = (MPIDI_ConnInfo *)pg->connData;
	MPIDI_STATE_DECL(MPID_STATE_CONNTOSTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNTOSTRING);

	/* Create this from the string array */
	string = (wchar_t *)MPIU_Malloc( connInfo->toStringLen * sizeof(wchar_t) );
	MPIU_ERR_MEM_CHK_POP(mpi_errno, string, connInfo->toStringLen * sizeof(wchar_t), L"string");
	str = string;

	pg_id = pg->id;
	while (*pg_id)
		str[len++] = *pg_id++;
	str[len++] = 0;
	
	MPIU_Snprintf( &str[len], connInfo->toStringLen - len, _TRUNCATE, L"%d", pg->size);
	/* Skip over the length */
	while (str[len++]);

	/* Copy each connection string */
	for (i=0; i<pg->size; i++)
	{
		wchar_t *p = connInfo->connStrings[i];
		while (*p)
		{
			str[len++] = *p++;
		}
		str[len++] = 0;
	}

	if (len > connInfo->toStringLen)
	{
		*buf_p = 0;
		*slen  = 0;
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,MPIR_ERR_FATAL,L"connToString", __LINE__, MPI_ERR_INTERN, L"**intern", NULL);
		goto fn_fail;
	}

	*buf_p = string;
	*slen = len;
	/*wprintf(L"connToString: <%s>\n", string);fflush(stdout);*/

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_CONNTOSTRING);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME connFromString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connFromString( const wchar_t *buf, MPIDI_PG_t *pg )
{
	MPIDI_ConnInfo *conninfo = 0;
	int i, mpi_errno = MPI_SUCCESS;
	const wchar_t *buf0 = buf;   /* save the start of buf */
	MPIDI_STATE_DECL(MPID_STATE_CONNFROMSTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNFROMSTRING);

	/* wprintf( L"Starting with buf = %s\n", buf );fflush(stdout); */

	/* Skip the pg id */
	while (*buf) buf++; buf++;

	/* Determine the size of the pg */
	pg->size = _wtoi( buf );
	while (*buf) buf++; buf++;

	conninfo = (MPIDI_ConnInfo *)MPIU_Malloc( sizeof(MPIDI_ConnInfo) );
	MPIU_ERR_MEM_CHK_POP(mpi_errno, conninfo, sizeof(MPIDI_ConnInfo), L"conninfo");
	conninfo->connStrings = (wchar_t **)MPIU_Malloc( pg->size * sizeof(wchar_t *));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, conninfo->connStrings, pg->size * sizeof(wchar_t*), L"connStrings");

	/* For now, make a copy of each item */
	for (i=0; i<pg->size; i++)
	{
		/*wprintf( L"Adding conn[%d] = <%s>\n", i, buf );fflush(stdout);*/
		conninfo->connStrings[i] = MPIU_Strdup( buf );
		MPIU_ERR_MEM_CHK_POP(mpi_errno, conninfo->connStrings, wcslen(buf)+1, L"connStrings");
		while (*buf) buf++;
		buf++;
	}
	pg->connData = conninfo;
		
	/* Save the length of the string needed to encode the connection
	   information */
	conninfo->toStringLen = (int)(buf - buf0) + 1;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_CONNFROMSTRING);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME connFree
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connFree( MPIDI_PG_t *pg )
{
	MPIDI_ConnInfo *conninfo = (MPIDI_ConnInfo *)pg->connData;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_CONNFREE);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNFREE);

	for (i=0; i<pg->size; i++) {
		MPIU_Free( conninfo->connStrings[i] );
	}
	MPIU_Free( conninfo->connStrings );
	MPIU_Free( conninfo );

	MPIDI_FUNC_EXIT(MPID_STATE_CONNFREE);
	return MPI_SUCCESS;
}

#ifdef USE_DBG_LOGGING
/* This is a temporary routine that is used to print out the pg string.
   A better approach may be to convert it into a single (long) string
   with no nulls. */
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_InitConnString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PrintConnStr( const wchar_t *file, int line, 
						const wchar_t *label, const wchar_t *str )
{
	int pg_size, i;

	MPIU_DBG_Outevent( file, line, MPIU_DBG_CH3_CONNECT, 0, label );
	MPIU_DBG_Outevent( file, line, MPIU_DBG_CH3_CONNECT, 0, str );
	
	/* Skip the pg id */
	while (*str) str++; str++;

	/* Determine the size of the pg */
	pg_size = atoi( str );
	while (*str) str++; str++;

	for (i=0; i<pg_size; i++) {
		MPIU_DBG_Outevent( file, line, MPIU_DBG_CH3_CONNECT, 0, str );
		while (*str) str++;
		str++;
	}
	return 0;
}
#endif

int MPIDI_PG_InitConnString( MPIDI_PG_t *pg )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_INITCONNSTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_INITCONNSTRING);

	pg->connData		   = 0;
	pg->getConnInfo		= getConnInfo;
	pg->connInfoToString   = connToString;
	pg->connInfoFromString = connFromString;
	pg->freeConnInfo	   = connFree;

/* fn_exit:*/
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_INITCONNSTRING);
	return mpi_errno;
	/*
 fn_fail:
	goto fn_exit;
	*/
}

/* Temp to get connection value for rank r */
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_GetConnString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_PG_GetConnString( MPIDI_PG_t *pg, int rank, wchar_t *val, int vallen )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_GETCONNSTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_GETCONNSTRING);

	if (pg->getConnInfo)
	{
		mpi_errno = (*pg->getConnInfo)( rank, val, vallen, pg );
	}
	else
	{
		MPIU_Internal_error_printf( L"Panic: no getConnInfo defined!\n" );
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_GETCONNSTRING);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Dup_vcr
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
  MPIDI_PG_Dup_vcr - Duplicate a virtual connection from a process group

  Notes:
  This routine provides a dup of a virtual connection given a process group
  and a rank in that group.  This routine is used only in initializing
  the MPI-1 communicators 'MPI_COMM_WORLD' and 'MPI_COMM_SELF', and in creating
  the initial intercommunicator after an 'MPI_Comm_spawn', 
  'MPI_Comm_spawn_multiple', or 'MPI_Comm_connect/MPI_Comm_accept'.  

  In addition to returning a dup of the virtual connection, it manages the
  reference count of the process group, which is always the number of inuse
  virtual connections.
  @*/
int MPIDI_PG_Dup_vcr( MPIDI_PG_t *pg, int rank, MPIDI_VC_t **vc_p )
{
	MPIDI_VC_t *vc;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_DUP_VCR);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_DUP_VCR);

	vc = &pg->vct[rank];
	/* Increase the reference count of the vc.  If the reference count 
	   increases from 0 to 1, increase the reference count of the 
	   process group *and* the reference count of the vc (this
	   allows us to distinquish between Comm_free and Comm_disconnect) */
	/* FIXME: This should be a fetch and increment for thread-safety */
	if (vc->ref_count == 0)
	{
		MPIDI_PG_Add_ref(pg);
		MPIDI_VC_Add_ref(vc);
	}
	MPIDI_VC_Add_ref(vc);
	*vc_p = vc;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_DUP_VCR);
	return MPI_SUCCESS;
}

/* FIXME: This routine should invoke a close method on the connection,
   rather than have all of the code here */
#undef FUNCNAME
#define FUNCNAME MPIDI_PG_Close_VCs
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
  MPIDI_PG_Close_VCs - Close all virtual connections on all process groups.
  
  Note:
  This routine is used in MPID_Finalize.  It is here to keep the process-group
  related functions together.
  @*/
int MPIDI_PG_Close_VCs( void )
{
	MPIDI_PG_t * pg = MPIDI_PG_list;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PG_CLOSE_VCS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_PG_CLOSE_VCS);

	while (pg)
	{
		int i, inuse;

		MPIU_DBG_MSG_S(CH3_CONNECT,VERBOSE,L"Closing vcs for pg %s",(wchar_t *)pg->id );


		for (i = 0; i < pg->size; i++)
		{
			MPIDI_VC_t * vc = &pg->vct[i];
			/* If the VC is myself then skip the close message */
			if (pg == MPIDI_Process.my_pg && i == MPIDI_Process.my_pg_rank)
			{
				if (vc->ref_count != 0)
				{
					MPIDI_PG_Release_ref(pg, &inuse);
				}
				continue;
			}

			if (MPIDI_CH3.VC_NeedsSendClose(vc))
			{
				/*printf("%s:%d SendClose pg=%p\n", __FILE__, __LINE__, vc->pg);fflush(stdout);*/
				MPIDI_CH3.VC_SendClose( vc, i );
			}
			else
			{
				if (vc->state == MPIDI_VC_STATE_INACTIVE && vc->ref_count != 0) {
					/* FIXME: If the reference count for the vc is not 0, something is wrong */
					MPIDI_PG_Release_ref(pg, &inuse);
				}

				MPIU_DBG_MSG_FMT(CH3_CONNECT,VERBOSE,(MPIU_DBG_FDEST, L"vc=%p: not sending a close to %d, vc in state %s", vc, i, MPIDI_VC_GetStateString(vc->state)));
			}
		}
		pg = pg->next;
	}
	/* Note that we do not free the process groups within this routine, even
	   if the reference counts have gone to zero. That is done once the 
	   connections are in fact closed (by the final progress loop that
	   handles any close requests that this code generates) */

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_PG_CLOSE_VCS);
	return mpi_errno;
}


/* FIXME: What are these routines for?  Who uses them?  Why are they different 
   from the src/util/dbg routines? */

/*
 * Note on thread safety.  These routines originally used 
 * MPID_Common_thread_lock/unlock, but that lock was not defined or used
 * consistently with the global mutex approach (now defined as
 * SINGLE_CS_ENTER/EXIT).  As these debugging routines should also
 * be withdrawn in favor of the general messaging utility, the
 * Common_thread_lock/unlock has been removed.
 */

#undef MPIDI_dbg_printf
#undef FUNCNAME
#define FUNCNAME MPIDI_dbg_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_dbg_printf(int level, wchar_t * func, wchar_t * fmt, ...)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_DBG_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_DBG_PRINTF);
	/* FIXME: This "unreferenced_arg" is an example of a problem with the 
	   API (unneeded level argument) or the code (failure to check the 
	   level argument).  Inserting these "unreference_arg" macros erroneously
	   suggests that the code is correct with this ununsed argument, and thus
	   commits the grave harm of obscuring a real problem */
	MPIU_UNREFERENCED_ARG(level);
	{
		va_list list;

		if (MPIR_Process.comm_world)
		{
			MPIU_dbglog_printf(L"[%d] %s(): ", MPIR_Process.comm_world->rank, func);
		}
		else
		{
			MPIU_dbglog_printf(L"[-1] %s(): ", func);
		}
		va_start(list, fmt);
		MPIU_dbglog_vprintf(fmt, list);
		va_end(list);
		MPIU_dbglog_printf(L"\n");
		fflush(stdout);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DBG_PRINTF);
}

#undef MPIDI_err_printf
#undef FUNCNAME
#define FUNCNAME MPIDI_err_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_err_printf(wchar_t * func, wchar_t * fmt, ...)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_ERR_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_ERR_PRINTF);
	{
		va_list list;

		if (MPIR_Process.comm_world)
		{
			wprintf(L"[%d] ERROR - %s(): ", MPIR_Process.comm_world->rank, func);
		}
		else
		{
			wprintf(L"[-1] ERROR - %s(): ", func);
		}
		va_start(list, fmt);
		vwprintf(fmt, list);
		va_end(list);
		wprintf(L"\n");
		fflush(stdout);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_ERR_PRINTF);
}

/* FIXME: It would be much better if the routine to print packets used
   routines defined by packet type, making it easier to modify the handling
   of packet types (and allowing channels to customize the printing
   of packets). For example, an array of function pointers, indexed by
   packet type, could be used.
   Also, these routines should not use MPIU_DBG_PRINTF, instead they should
   us a simple fwprintf with a style allowance (so that the style checker
   won't flag the use as a possible problem).  */

#ifdef MPICH_DBG_OUTPUT
#undef FUNCNAME
#define FUNCNAME MPIDI_DBG_Print_packet
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_DBG_Print_packet(MPIDI_CH3_Pkt_t *pkt)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_DBG_PRINT_PACKET);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_DBG_PRINT_PACKET);
	{
		MPIU_DBG_PRINTF((L"MPIDI_CH3_Pkt_t:\n"));
		switch(pkt->type)
		{
			case MPIDI_CH3_PKT_EAGER_SEND:
				MPIU_DBG_PRINTF((L" type ......... EAGER_SEND\n"));
				MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->eager_send.sender_req_id));
				MPIU_DBG_PRINTF((L" context_id ... %d\n", pkt->eager_send.match.context_id));
				MPIU_DBG_PRINTF((L" tag .......... %d\n", pkt->eager_send.match.tag));
				MPIU_DBG_PRINTF((L" rank ......... %d\n", pkt->eager_send.match.rank));
				MPIU_DBG_PRINTF((L" data_sz ...... %d\n", pkt->eager_send.data_sz));
				break;
			case MPIDI_CH3_PKT_EAGER_SYNC_SEND:
				MPIU_DBG_PRINTF((L" type ......... EAGER_SYNC_SEND\n"));
				MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->eager_sync_send.sender_req_id));
				MPIU_DBG_PRINTF((L" context_id ... %d\n", pkt->eager_sync_send.match.context_id));
				MPIU_DBG_PRINTF((L" tag .......... %d\n", pkt->eager_sync_send.match.tag));
				MPIU_DBG_PRINTF((L" rank ......... %d\n", pkt->eager_sync_send.match.rank));
				MPIU_DBG_PRINTF((L" data_sz ...... %d\n", pkt->eager_sync_send.data_sz));
				break;
			case MPIDI_CH3_PKT_EAGER_SYNC_ACK:
				MPIU_DBG_PRINTF((L" type ......... EAGER_SYNC_ACK\n"));
				MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->eager_sync_ack.sender_req_id));
				break;
			case MPIDI_CH3_PKT_READY_SEND:
				MPIU_DBG_PRINTF((L" type ......... READY_SEND\n"));
				MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->ready_send.sender_req_id));
				MPIU_DBG_PRINTF((L" context_id ... %d\n", pkt->ready_send.match.context_id));
				MPIU_DBG_PRINTF((L" tag .......... %d\n", pkt->ready_send.match.tag));
				MPIU_DBG_PRINTF((L" rank ......... %d\n", pkt->ready_send.match.rank));
				MPIU_DBG_PRINTF((L" data_sz ...... %d\n", pkt->ready_send.data_sz));
				break;
			case MPIDI_CH3_PKT_RNDV_REQ_TO_SEND:
				MPIU_DBG_PRINTF((L" type ......... REQ_TO_SEND\n"));
				MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->rndv_req_to_send.sender_req_id));
				MPIU_DBG_PRINTF((L" context_id ... %d\n", pkt->rndv_req_to_send.match.context_id));
				MPIU_DBG_PRINTF((L" tag .......... %d\n", pkt->rndv_req_to_send.match.tag));
				MPIU_DBG_PRINTF((L" rank ......... %d\n", pkt->rndv_req_to_send.match.rank));
				MPIU_DBG_PRINTF((L" data_sz ...... %d\n", pkt->rndv_req_to_send.data_sz));
				break;
			case MPIDI_CH3_PKT_RNDV_CLR_TO_SEND:
				MPIU_DBG_PRINTF((L" type ......... CLR_TO_SEND\n"));
				MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->rndv_clr_to_send.sender_req_id));
				MPIU_DBG_PRINTF((L" recvr_reqid .. 0x%08X\n", pkt->rndv_clr_to_send.receiver_req_id));
				break;
			case MPIDI_CH3_PKT_RNDV_SEND:
				MPIU_DBG_PRINTF((L" type ......... RNDV_SEND\n"));
				MPIU_DBG_PRINTF((L" recvr_reqid .. 0x%08X\n", pkt->rndv_send.receiver_req_id));
				break;
			case MPIDI_CH3_PKT_CANCEL_SEND_REQ:
				MPIU_DBG_PRINTF((L" type ......... CANCEL_SEND\n"));
				MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->cancel_send_req.sender_req_id));
				MPIU_DBG_PRINTF((L" context_id ... %d\n", pkt->cancel_send_req.match.context_id));
				MPIU_DBG_PRINTF((L" tag .......... %d\n", pkt->cancel_send_req.match.tag));
				MPIU_DBG_PRINTF((L" rank ......... %d\n", pkt->cancel_send_req.match.rank));
				break;
			case MPIDI_CH3_PKT_CANCEL_SEND_RESP:
				MPIU_DBG_PRINTF((L" type ......... CANCEL_SEND_RESP\n"));
				MPIU_DBG_PRINTF((L" sender_reqid . 0x%08X\n", pkt->cancel_send_resp.sender_req_id));
				MPIU_DBG_PRINTF((L" ack .......... %d\n", pkt->cancel_send_resp.ack));
				break;
			case MPIDI_CH3_PKT_PUT:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_PUT\n"));
				MPIU_DBG_PRINTF((L" addr ......... %p\n", pkt->put.addr));
				MPIU_DBG_PRINTF((L" count ........ %d\n", pkt->put.count));
				MPIU_DBG_PRINTF((L" datatype ..... 0x%08X\n", pkt->put.datatype));
				MPIU_DBG_PRINTF((L" dataloop_size. 0x%08X\n", pkt->put.dataloop_size));
				MPIU_DBG_PRINTF((L" target ....... 0x%08X\n", pkt->put.target_win_handle));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->put.source_win_handle));
				/*MPIU_DBG_PRINTF((L" win_ptr ...... 0x%08X\n", pkt->put.win_ptr));*/
				break;
			case MPIDI_CH3_PKT_GET:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_GET\n"));
				MPIU_DBG_PRINTF((L" addr ......... %p\n", pkt->get.addr));
				MPIU_DBG_PRINTF((L" count ........ %d\n", pkt->get.count));
				MPIU_DBG_PRINTF((L" datatype ..... 0x%08X\n", pkt->get.datatype));
				MPIU_DBG_PRINTF((L" dataloop_size. %d\n", pkt->get.dataloop_size));
				MPIU_DBG_PRINTF((L" request ...... 0x%08X\n", pkt->get.request_handle));
				MPIU_DBG_PRINTF((L" target ....... 0x%08X\n", pkt->get.target_win_handle));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->get.source_win_handle));
				/*
				MPIU_DBG_PRINTF((L" request ...... 0x%08X\n", pkt->get.request));
				MPIU_DBG_PRINTF((L" win_ptr ...... 0x%08X\n", pkt->get.win_ptr));
				*/
				break;
			case MPIDI_CH3_PKT_GET_RESP:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_GET_RESP\n"));
				MPIU_DBG_PRINTF((L" request ...... 0x%08X\n", pkt->get_resp.request_handle));
				/*MPIU_DBG_PRINTF((L" request ...... 0x%08X\n", pkt->get_resp.request));*/
				break;
			case MPIDI_CH3_PKT_ACCUMULATE:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_ACCUMULATE\n"));
				MPIU_DBG_PRINTF((L" addr ......... %p\n", pkt->accum.addr));
				MPIU_DBG_PRINTF((L" count ........ %d\n", pkt->accum.count));
				MPIU_DBG_PRINTF((L" datatype ..... 0x%08X\n", pkt->accum.datatype));
				MPIU_DBG_PRINTF((L" dataloop_size. %d\n", pkt->accum.dataloop_size));
				MPIU_DBG_PRINTF((L" op ........... 0x%08X\n", pkt->accum.op));
				MPIU_DBG_PRINTF((L" target ....... 0x%08X\n", pkt->accum.target_win_handle));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->accum.source_win_handle));
				/*MPIU_DBG_PRINTF((L" win_ptr ...... 0x%08X\n", pkt->accum.win_ptr));*/
				break;
			case MPIDI_CH3_PKT_LOCK:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_LOCK\n"));
				MPIU_DBG_PRINTF((L" lock_type .... %d\n", pkt->lock.lock_type));
				MPIU_DBG_PRINTF((L" target ....... 0x%08X\n", pkt->lock.target_win_handle));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->lock.source_win_handle));
				break;
			case MPIDI_CH3_PKT_LOCK_PUT_UNLOCK:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_LOCK_PUT_UNLOCK\n"));
				MPIU_DBG_PRINTF((L" addr ......... %p\n", pkt->lock_put_unlock.addr));
				MPIU_DBG_PRINTF((L" count ........ %d\n", pkt->lock_put_unlock.count));
				MPIU_DBG_PRINTF((L" datatype ..... 0x%08X\n", pkt->lock_put_unlock.datatype));
				MPIU_DBG_PRINTF((L" lock_type .... %d\n", pkt->lock_put_unlock.lock_type));
				MPIU_DBG_PRINTF((L" target ....... 0x%08X\n", pkt->lock_put_unlock.target_win_handle));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->lock_put_unlock.source_win_handle));
				break;
			case MPIDI_CH3_PKT_LOCK_ACCUM_UNLOCK:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_LOCK_ACCUM_UNLOCK\n"));
				MPIU_DBG_PRINTF((L" addr ......... %p\n", pkt->lock_accum_unlock.addr));
				MPIU_DBG_PRINTF((L" count ........ %d\n", pkt->lock_accum_unlock.count));
				MPIU_DBG_PRINTF((L" datatype ..... 0x%08X\n", pkt->lock_accum_unlock.datatype));
				MPIU_DBG_PRINTF((L" lock_type .... %d\n", pkt->lock_accum_unlock.lock_type));
				MPIU_DBG_PRINTF((L" target ....... 0x%08X\n", pkt->lock_accum_unlock.target_win_handle));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->lock_accum_unlock.source_win_handle));
				break;
			case MPIDI_CH3_PKT_LOCK_GET_UNLOCK:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_LOCK_GET_UNLOCK\n"));
				MPIU_DBG_PRINTF((L" addr ......... %p\n", pkt->lock_get_unlock.addr));
				MPIU_DBG_PRINTF((L" count ........ %d\n", pkt->lock_get_unlock.count));
				MPIU_DBG_PRINTF((L" datatype ..... 0x%08X\n", pkt->lock_get_unlock.datatype));
				MPIU_DBG_PRINTF((L" lock_type .... %d\n", pkt->lock_get_unlock.lock_type));
				MPIU_DBG_PRINTF((L" target ....... 0x%08X\n", pkt->lock_get_unlock.target_win_handle));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->lock_get_unlock.source_win_handle));
				MPIU_DBG_PRINTF((L" request ...... 0x%08X\n", pkt->lock_get_unlock.request_handle));
				break;
			case MPIDI_CH3_PKT_PT_RMA_DONE:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_PT_RMA_DONE\n"));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->lock_accum_unlock.source_win_handle));
				break;
			case MPIDI_CH3_PKT_LOCK_GRANTED:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_LOCK_GRANTED\n"));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->lock_granted.source_win_handle));
				break;
				/*
			case MPIDI_CH3_PKT_SHARED_LOCK_OPS_DONE:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_SHARED_LOCK_OPS_DONE\n"));
				MPIU_DBG_PRINTF((L" source ....... 0x%08X\n", pkt->shared_lock_ops_done.source_win_handle));
				break;
				*/
			case MPIDI_CH3_PKT_FLOW_CNTL_UPDATE:
				MPIU_DBG_PRINTF((L" FLOW_CNTRL_UPDATE\n"));
				break;
			case MPIDI_CH3_PKT_RTS_IOV:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_RTS_IOV\n"));
				MPIU_DBG_PRINTF((L" sreq ......... 0x%08X\n", pkt->rts_iov.sreq));
				MPIU_DBG_PRINTF((L" iov_len ...... %d\n", pkt->rts_iov.iov_len));
				break;
			case MPIDI_CH3_PKT_CTS_IOV:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_CTS_IOV\n"));
				MPIU_DBG_PRINTF((L" sreq ......... 0x%08X\n", pkt->cts_iov.sreq));
				MPIU_DBG_PRINTF((L" rreq ......... 0x%08X\n", pkt->cts_iov.rreq));
				MPIU_DBG_PRINTF((L" iov_len ...... %d\n", pkt->cts_iov.iov_len));
				break;
			case MPIDI_CH3_PKT_RELOAD:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_RELOAD\n"));
				MPIU_DBG_PRINTF((L" send_recv .... %d\n", pkt->reload.send_recv));
				MPIU_DBG_PRINTF((L" sreq ......... 0x%08X\n", pkt->reload.sreq));
				MPIU_DBG_PRINTF((L" rreq ......... 0x%08X\n", pkt->reload.rreq));
				break;
			case MPIDI_CH3_PKT_IOV:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_IOV\n"));
				MPIU_DBG_PRINTF((L" req .......... 0x%08X\n", pkt->iov.req));
				MPIU_DBG_PRINTF((L" send_recv .... %d\n", pkt->iov.send_recv));
				MPIU_DBG_PRINTF((L" iov_len ...... %d\n", pkt->iov.iov_len));
				break;
			case MPIDI_CH3_PKT_CLOSE:
				MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_CLOSE\n"));
				MPIU_DBG_PRINTF((L" ack ......... %s\n", pkt->close.ack ? L"TRUE" : L"FALSE"));
				break;
			
			default:
				MPIU_DBG_PRINTF((L" INVALID PACKET\n"));
				MPIU_DBG_PRINTF((L" unknown type ... %d\n", pkt->type));
				MPIU_DBG_PRINTF((L"  type .......... EAGER_SEND\n"));
				MPIU_DBG_PRINTF((L"   sender_reqid . 0x%08X\n", pkt->eager_send.sender_req_id));
				MPIU_DBG_PRINTF((L"   context_id ... %d\n", pkt->eager_send.match.context_id));
				MPIU_DBG_PRINTF((L"   data_sz ...... %d\n", pkt->eager_send.data_sz));
				MPIU_DBG_PRINTF((L"   tag .......... %d\n", pkt->eager_send.match.tag));
				MPIU_DBG_PRINTF((L"   rank ......... %d\n", pkt->eager_send.match.rank));
				MPIU_DBG_PRINTF((L"  type .......... REQ_TO_SEND\n"));
				MPIU_DBG_PRINTF((L"   sender_reqid . 0x%08X\n", pkt->rndv_req_to_send.sender_req_id));
				MPIU_DBG_PRINTF((L"   context_id ... %d\n", pkt->rndv_req_to_send.match.context_id));
				MPIU_DBG_PRINTF((L"   data_sz ...... %d\n", pkt->rndv_req_to_send.data_sz));
				MPIU_DBG_PRINTF((L"   tag .......... %d\n", pkt->rndv_req_to_send.match.tag));
				MPIU_DBG_PRINTF((L"   rank ......... %d\n", pkt->rndv_req_to_send.match.rank));
				MPIU_DBG_PRINTF((L"  type .......... CLR_TO_SEND\n"));
				MPIU_DBG_PRINTF((L"   sender_reqid . 0x%08X\n", pkt->rndv_clr_to_send.sender_req_id));
				MPIU_DBG_PRINTF((L"   recvr_reqid .. 0x%08X\n", pkt->rndv_clr_to_send.receiver_req_id));
				MPIU_DBG_PRINTF((L"  type .......... RNDV_SEND\n"));
				MPIU_DBG_PRINTF((L"   recvr_reqid .. 0x%08X\n", pkt->rndv_send.receiver_req_id));
				MPIU_DBG_PRINTF((L"  type .......... CANCEL_SEND\n"));
				MPIU_DBG_PRINTF((L"   context_id ... %d\n", pkt->cancel_send_req.match.context_id));
				MPIU_DBG_PRINTF((L"   tag .......... %d\n", pkt->cancel_send_req.match.tag));
				MPIU_DBG_PRINTF((L"   rank ......... %d\n", pkt->cancel_send_req.match.rank));
				MPIU_DBG_PRINTF((L"   sender_reqid . 0x%08X\n", pkt->cancel_send_req.sender_req_id));
				MPIU_DBG_PRINTF((L"  type .......... CANCEL_SEND_RESP\n"));
				MPIU_DBG_PRINTF((L"   sender_reqid . 0x%08X\n", pkt->cancel_send_resp.sender_req_id));
				MPIU_DBG_PRINTF((L"   ack .......... %d\n", pkt->cancel_send_resp.ack));
				break;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_DBG_PRINT_PACKET);
}
#endif


#undef FUNCNAME
#define FUNCNAME MPIDI_VC_GetStateString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
const wchar_t * MPIDI_VC_GetStateString(int state)
{
	const wchar_t * str;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_VC_GETSTATESTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_VC_GETSTATESTRING);
	switch (state)
	{
		case MPIDI_VC_STATE_INACTIVE:
			str = L"MPIDI_VC_STATE_INACTIVE";
		case MPIDI_VC_STATE_ACTIVE:
			str = L"MPIDI_VC_STATE_ACTIVE";
		case MPIDI_VC_STATE_LOCAL_CLOSE:
			str = L"MPIDI_VC_STATE_LOCAL_CLOSE";
		case MPIDI_VC_STATE_REMOTE_CLOSE:
			str = L"MPIDI_VC_STATE_REMOTE_CLOSE";
		case MPIDI_VC_STATE_CLOSE_ACKED:
			str = L"MPIDI_VC_STATE_CLOSE_ACKED";
		default:
			str = L"unknown";
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_VC_GETSTATESTRING);
	return str;
}

/* This routine is not thread safe and should only be used while
   debugging.  It is used to encode a brief description of a message
   packet into a string to make it easy to include in the message log
   output (with no newlines to simplify extracting info from the log file) 
*/
const wchar_t *MPIDI_Pkt_GetDescString( MPIDI_CH3_Pkt_t *pkt ) 
{
	static wchar_t pktmsg[256];

	/* For data messages, the string (...) is (context,tag,rank,size) */
	switch(pkt->type) {
	case MPIDI_CH3_PKT_EAGER_SEND:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"EAGER_SEND - (%d,%d,%d,%d)", 
					   pkt->eager_send.match.context_id,
					   (int)pkt->eager_send.match.tag, 
					   pkt->eager_send.match.rank, 
					   pkt->eager_send.data_sz );
		break;
	case MPIDI_CH3_PKT_EAGER_SYNC_SEND:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"EAGER_SYNC_SEND - (%d,%d,%d,%d) req=%d", 
					   pkt->eager_sync_send.match.context_id,
					   (int)pkt->eager_sync_send.match.tag, 
					   pkt->eager_sync_send.match.rank, 
					   pkt->eager_sync_send.data_sz,
					   pkt->eager_sync_send.sender_req_id );
				break;
	case MPIDI_CH3_PKT_EAGER_SYNC_ACK:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"EAGER_SYNC_ACK - req=%d", 
					   pkt->eager_sync_ack.sender_req_id );
		break;
	case MPIDI_CH3_PKT_READY_SEND:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"READY_SEND - (%d,%d,%d,%d)", 
					   pkt->ready_send.match.context_id,
					   (int)pkt->ready_send.match.tag, 
					   pkt->ready_send.match.rank, 
					   pkt->ready_send.data_sz );
		break;
	case MPIDI_CH3_PKT_RNDV_REQ_TO_SEND:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"RNDV_REQ_TO_SEND - (%d,%d,%d,%d) req=%d", 
					   pkt->rndv_req_to_send.match.context_id,
					   (int)pkt->rndv_req_to_send.match.tag, 
					   pkt->rndv_req_to_send.match.rank, 
					   pkt->rndv_req_to_send.data_sz,
					   pkt->rndv_req_to_send.sender_req_id );
		break;
	case MPIDI_CH3_PKT_RNDV_CLR_TO_SEND:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"RNDV_CLRTO_SEND - req=%d, recv req=%d", 
					   pkt->rndv_clr_to_send.sender_req_id,
					   pkt->rndv_clr_to_send.receiver_req_id );
				break;
	case MPIDI_CH3_PKT_RNDV_SEND:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"RNDV_SEND - recv req=%d", 
					   pkt->rndv_send.receiver_req_id );
		break;
	case MPIDI_CH3_PKT_CANCEL_SEND_REQ:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"CANCEL_SEND_REQ - req=%d", 
					   pkt->cancel_send_req.sender_req_id );
		break;
	case MPIDI_CH3_PKT_CANCEL_SEND_RESP:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"CANCEL_SEND_RESP - req=%d ack=%d", 
					   pkt->cancel_send_resp.sender_req_id, 
					   pkt->cancel_send_resp.ack );
		break;
	case MPIDI_CH3_PKT_PUT:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"PUT - (%p,%d,0x%08X)", 
					   pkt->put.addr, 
					   pkt->put.count,
					   pkt->put.target_win_handle );
				break;
	case MPIDI_CH3_PKT_GET:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"GET - (%p,%d,0x%08X) req=%d", 
					   pkt->get.addr, 
					   pkt->get.count,
					   pkt->get.target_win_handle,
					   pkt->get.request_handle );
		break;
	case MPIDI_CH3_PKT_GET_RESP:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"GET_RESP - req=%d", 
					   pkt->get_resp.request_handle );
		break;
	case MPIDI_CH3_PKT_ACCUMULATE:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"ACCUMULATE - (%p,%d,0x%08X)", 
					   pkt->accum.addr,
					   pkt->accum.count, 
					   pkt->accum.target_win_handle );
		break;
	case MPIDI_CH3_PKT_LOCK:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"LOCK - %d", 
					   pkt->lock.target_win_handle );
		break;
	case MPIDI_CH3_PKT_LOCK_PUT_UNLOCK:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"PUT_UNLOCK - (%p,%d,0x%08X)", 
					   pkt->lock_put_unlock.addr,
					   pkt->lock_put_unlock.count,
					   pkt->lock_put_unlock.target_win_handle );
		break;
	case MPIDI_CH3_PKT_LOCK_ACCUM_UNLOCK:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"LOCK_ACCUM_UNLOCK - (%p,%d,0x%08X)", 
					   pkt->lock_accum_unlock.addr,
					   pkt->lock_accum_unlock.count,
					   pkt->lock_accum_unlock.target_win_handle );
		break;
	case MPIDI_CH3_PKT_LOCK_GET_UNLOCK:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"LOCK_GET_UNLOCK - (%p,%d,0x%08X) req=%d", 
					   pkt->lock_get_unlock.addr,
					   pkt->lock_get_unlock.count,
					   pkt->lock_get_unlock.target_win_handle, 
					   pkt->lock_get_unlock.request_handle );
		break;
	case MPIDI_CH3_PKT_PT_RMA_DONE:
		/* There is no rma_done packet type */
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"RMA_DONE - 0x%08X", 
					   pkt->lock_accum_unlock.source_win_handle );
		break;
	case MPIDI_CH3_PKT_LOCK_GRANTED:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"LOCK_GRANTED - 0x%08X", 
					   pkt->lock_granted.source_win_handle );
				break;
	case MPIDI_CH3_PKT_FLOW_CNTL_UPDATE:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"FLOW_CNTL_UPDATE" );
		break;
	case MPIDI_CH3_PKT_RTS_IOV:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"RTS_IOV - sreq=0x%08X, len=%d", 
					   pkt->rts_iov.sreq, 
					   pkt->rts_iov.iov_len );
		break;
	case MPIDI_CH3_PKT_CTS_IOV:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"CTS_IOV - sreq=0x%08X, rreq=0x%08X, len=%d", 
					   pkt->cts_iov.sreq,
					   pkt->cts_iov.rreq,
					   pkt->cts_iov.iov_len );
		break;
	case MPIDI_CH3_PKT_RELOAD:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"RELOAD  - sreq=0x%08X, rreq=0x%08X, sendrecv=%d", 
					   pkt->reload.sreq,
					   pkt->reload.rreq,
					   pkt->reload.send_recv );
		break;
	case MPIDI_CH3_PKT_IOV:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"IOV - req=%d, sendrecv=%d, len=%d",
					   pkt->iov.req,
					   pkt->iov.send_recv,
					   pkt->iov.iov_len );
		break;
	case MPIDI_CH3_PKT_CLOSE:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"CLOSE ack=%d", 
					   pkt->close.ack );
		break;
			
	default:
		MPIU_Snprintf( pktmsg, sizeof(pktmsg)/sizeof(wchar_t), _TRUNCATE, 
					   L"INVALID PACKET type=%d", pkt->type );
		break;
	}

	return pktmsg;
}


#undef FUNCNAME
#define FUNCNAME MPID_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Init(int *argc, wchar_t ***argv, int requested, int *provided, int *has_args, int *has_env)
{
	int mpi_errno = MPI_SUCCESS;
	int has_parent;
	MPIDI_PG_t * pg = NULL;
	int pg_rank = -1;
	int pg_size;
	MPID_Comm * comm;
	int p;
	const wchar_t *channel;
	MPIDI_STATE_DECL(MPID_STATE_MPID_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_INIT);

	/* FIXME: This is a good place to check for environment variables
	   and command line options that may control the device */

	/*
	 * Set global process attributes.  These can be overridden by the channel if necessary.
	 */
	MPIR_Process.attrs.tag_ub = MPIDI_TAG_UB;

	/*
	 * Load the CH3 interface
	 */
	channel = MPIU_GetStaticEnv(L"DeinoMPI_CHANNEL");
	if (channel == NULL)
	{
		//wprintf(L"Using SOCK channel.\r\n");fflush(stdout);
		mpi_errno = sockMPIDI_CH3_Init_fns(&MPIDI_CH3);
		//wprintf(L"Using SHM channel.\r\n");fflush(stdout);
		//mpi_errno = shmMPIDI_CH3_Init_fns(&MPIDI_CH3);
		//wprintf(L"Using SSHM channel.\r\n");fflush(stdout);
		//mpi_errno = sshmMPIDI_CH3_Init_fns(&MPIDI_CH3);
		//wprintf(L"Using SSM channel.\r\n");fflush(stdout);
		//mpi_errno = ssmMPIDI_CH3_Init_fns(&MPIDI_CH3);
	}
	else
	{
		if (_wcsicmp(channel, L"shm") == 0)
		{
			//wprintf(L"Using SHM channel.\r\n");fflush(stdout);
			mpi_errno = shmMPIDI_CH3_Init_fns(&MPIDI_CH3);
		}
		else if (_wcsicmp(channel, L"sshm") == 0)
		{
			//wprintf(L"Using SSHM channel.\r\n");fflush(stdout);
			mpi_errno = sshmMPIDI_CH3_Init_fns(&MPIDI_CH3);
		}
		else if (_wcsicmp(channel, L"ssm") == 0)
		{
			//wprintf(L"Using SSM channel.\r\n");fflush(stdout);
			mpi_errno = ssmMPIDI_CH3_Init_fns(&MPIDI_CH3);
		}
		else
		{
			//wprintf(L"Using SOCK channel.\r\n");fflush(stdout);
			mpi_errno = sockMPIDI_CH3_Init_fns(&MPIDI_CH3);
		}
	}

	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno, MPI_ERR_OTHER, L"**ch3|ch3_init");
	}

	/*
	 * Perform channel-independent PMI initialization
	 */
	mpi_errno = InitPG( argc, argv, has_args, has_env, &has_parent, &pg_rank, &pg );
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|ch3_init");
	}

	/* FIXME: Why are pg_size and pg_rank handled differently? */
	pg_size = MPIDI_PG_Get_size(pg);
	MPIDI_Process.my_pg = pg;  /* this is rework for shared memories because they need this set earlier for getting the business card */
	MPIDI_Process.my_pg_rank = pg_rank;
	/* FIXME: Why do we add a ref to pg here? */
	MPIDI_PG_Add_ref(pg);

	/*
	 * Let the channel perform any necessary initialization
	 * The channel init should assume that PMI_Init has been called and that
	 * the basic information about the job has been extracted from PMI (e.g.,
	 * the size and rank of this process, and the process group id)
	 */
	mpi_errno = MPIDI_CH3.Init(has_parent, pg, pg_rank);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|ch3_init");
	}

	/*
	 * Initialize the MPI_COMM_WORLD object
	 */
	comm = MPIR_Process.comm_world;

	comm->rank			= pg_rank;
	comm->remote_size	= pg_size;
	comm->local_size	= pg_size;

	mpi_errno = MPID_VCRT_Create(comm->remote_size, &comm->vcrt);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,L"**dev|vcrt_create", L"**dev|vcrt_create %s", L"MPI_COMM_WORLD");
	}
	
	mpi_errno = MPID_VCRT_Get_ptr(comm->vcrt, &comm->vcr);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,L"**dev|vcrt_get_ptr", L"dev|vcrt_get_ptr %s", L"MPI_COMM_WORLD");
	}
	
	/* Initialize the connection table on COMM_WORLD from the process group's
	   connection table */
	for (p = 0; p < pg_size; p++)
	{
		MPID_VCR_Dup(&pg->vct[p], &comm->vcr[p]);
	}

	MPID_Dev_comm_create_hook (comm);
	mpi_errno = MPIR_Comm_commit(comm);
	MPIU_ERR_CHK_POP(mpi_errno);

	/*
	 * Initialize the MPI_COMM_SELF object
	 */
	comm = MPIR_Process.comm_self;
	comm->rank			= 0;
	comm->remote_size	= 1;
	comm->local_size	= 1;
	
	mpi_errno = MPID_VCRT_Create(comm->remote_size, &comm->vcrt);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**dev|vcrt_create", L"**dev|vcrt_create %s", L"MPI_COMM_SELF");
	}
	
	mpi_errno = MPID_VCRT_Get_ptr(comm->vcrt, &comm->vcr);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**dev|vcrt_get_ptr", L"dev|vcrt_get_ptr %s", L"MPI_COMM_WORLD");
	}
	
	MPID_VCR_Dup(&pg->vct[pg_rank], &comm->vcr[0]);

#ifdef MPID_NEEDS_ICOMM_WORLD
    /*
     * Initialize the MPIR_ICOMM_WORLD object (an internal, private version of MPI_COMM_WORLD) 
     */
    comm = MPIR_Process.icomm_world;

    comm->rank        = pg_rank;
    comm->remote_size = pg_size;
    comm->local_size  = pg_size;
#if 0    
    mpi_errno = MPID_VCRT_Create(comm->remote_size, &comm->vcrt);
    if (mpi_errno != MPI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,"**dev|vcrt_create", 
			     "**dev|vcrt_create %s", "MPI_COMM_WORLD");
    }
    
    mpi_errno = MPID_VCRT_Get_ptr(comm->vcrt, &comm->vcr);
    if (mpi_errno != MPI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,"**dev|vcrt_get_ptr", 
			     "dev|vcrt_get_ptr %s", "MPI_COMM_WORLD");
    }
    
    /* Initialize the connection table on COMM_WORLD from the process group's
       connection table */
    for (p = 0; p < pg_size; p++)
    {
	MPID_VCR_Dup(&pg->vct[p], &comm->vcr[p]);
    }
#endif
    MPID_VCRT_Add_ref( MPIR_Process.comm_world->vcrt );
    comm->vcrt = MPIR_Process.comm_world->vcrt;
    comm->vcr  = MPIR_Process.comm_world->vcr;
    
    MPID_Dev_comm_create_hook (comm);
    mpi_errno = MPIR_Comm_commit(comm);
    MPIU_ERR_CHK_POP(mpi_errno);
#endif
    
	
	/*
	 * If this process group was spawned by a MPI application, then
	 * form the MPI_COMM_PARENT inter-communicator.
	 */

	if (has_parent)
	{
		wchar_t * parent_port;

		/* FIXME: To allow just the "root" process to 
		   request the port and then use MPIR_Bcast to 
		   distribute it to the rest of the processes,
		   we need to perform the Bcast after MPI is
		   otherwise initialized.  We could do this
		   by adding another MPID call that the MPI_Init(_thread)
		   routine would make after the rest of MPI is 
		   initialized, but before MPI_Init returns.
		   In fact, such a routine could be used to 
		   perform various checks, including parameter
		   consistency value (e.g., all processes have the
		   same environment variable values). Alternately,
		   we could allow a few routines to operate with 
		   predefined parameter choices (e.g., bcast, allreduce)
		   for the purposes of initialization. */
		mpi_errno = MPIDI_CH3_Get_parent_port(&parent_port);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|get_parent_port");
		}
		MPIU_DBG_MSG_S(CH3_CONNECT,VERBOSE,L"Parent port is %s\n", parent_port);
			
		mpi_errno = MPID_Comm_connect(parent_port, NULL, 0, MPIR_Process.comm_world, &comm);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**ch3|conn_parent", L"**ch3|conn_parent %s", parent_port);
		}

		MPIR_Process.comm_parent = comm;
		MPIU_Assert(MPIR_Process.comm_parent != NULL);
		MPIU_Wcsncpy(comm->name, L"MPI_COMM_PARENT", MPI_MAX_OBJECT_NAME);
		
		/* FIXME: Check that this intercommunicator gets freed in MPI_Finalize if not already freed.  */
	}
	
	/*
	 * Set provided thread level
	 */
	if (provided != NULL)
	{
		/* This must be min(requested,MPICH_THREAD_LEVEL) if runtime
		   control of thread level is available */
		*provided = (MPICH_THREAD_LEVEL < requested) ? MPICH_THREAD_LEVEL : requested;
	}

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_INIT);
	return mpi_errno;

	/* --BEGIN ERROR HANDLING-- */
  fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* This function is registered with PMI and is called when the job is aborted and this process is about to be killed. */
/* Code in this function cannot be relied upon since the process manager may not be able to create a thread to execute this function. */
static int print_call_history_on_abort = 0;
int MPID_Abort_callback(int abort_exit_code)
{
	MPIU_UNREFERENCED_ARG(abort_exit_code);
	/* Here we could do some cleanup in preparation for being killed.
	 * For example the logging buffers could be written out to disk and the log files closed.
	 */
	if (print_call_history_on_abort)
	{
		MPIU_Print_call_history();
	}
	MPIU_Timer_finalize();
	//printf("MPID_Abort_callback(%d) called for rank %d\n", abort_exit_code, MPIR_Process.comm_world->rank);fflush(stdout);
	return 0;
}

/* This function is registered with PMI and is called by the process management system when it wants a print out of the process queue status */
int MPID_Print_queues(wchar_t *str, int max_length)
{
	int mpi_errno = MPI_SUCCESS;

	MPIU_dump_dbg_memlog_to_stdout();
	//MPIU_Print_call_history();

	if (max_length > 0)
		str[0] = L'\0';

	mpi_errno = MPIU_Queues_to_string(str, max_length);
	if (wcslen(str) < 1)
		MPIU_Snprintf(str, max_length, _TRUNCATE, L"Rank %d queues empty", MPIR_Process.comm_world->rank);

	return mpi_errno;
}

/* This function is registered with PMI and is called by the process management system when it wants to print out the process MPI call history */
int MPID_Print_call_history(wchar_t *str, int max_length)
{
	int mpi_errno = MPI_SUCCESS;

	if (max_length > 0)
		str[0] = L'\0';

	mpi_errno = MPIU_Call_history_to_string(str, max_length);
	return mpi_errno;
}

/*
 * Initialize the process group structure by using PMI calls.
 * This routine initializes PMI and uses PMI calls to setup the 
 * process group structures.
 * 
 */
#undef FUNCNAME
#define FUNCNAME InitPG
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int InitPG( int *argc, wchar_t ***argv, int *has_args, int *has_env, int *has_parent, int *pg_rank_p, MPIDI_PG_t **pg_p )
{
	int pmi_errno;
	int mpi_errno = MPI_SUCCESS;
	int pg_rank=0, pg_size=1, appnum, pg_id_sz;
	int usePMI=1;
	wchar_t *pg_id;
	MPIDI_PG_t *pg = 0;
	MPIDI_STATE_DECL(MPID_STATE_INITPG);

	MPIDI_FUNC_ENTER(MPID_STATE_INITPG);

	/* If we use PMI here, make the PMI calls to get the
	   basic values.  Note that systems that return setvals == true
	   do not make use of PMI for the KVS routines either (it is
	   assumed that the discover connection information through some
	   other mechanism */
	/* FIXME: We may want to allow the channel to ifdef out the use
	   of PMI calls, or ask the channel to provide stubs that 
	   return errors if the routines are in fact used */
	if (usePMI)
	{
		/*
		 * Initialize the process manangement interface (PMI), 
		 * and get rank and size information about our process group
		 */
		pmi_errno = PMI_Init(has_parent);
		if (pmi_errno != PMI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_init", L"**pmi_init %d %s", pmi_errno, PMI_Get_last_error_msgW());
		}

		/* Ignore the return value of this function since the functionality may not be implemented */
		PMI_Register_abort_hook(MPID_Abort_callback);
		PMI_Register_print_queues_fn(MPID_Print_queues);
		PMI_Register_print_call_history_fn(MPID_Print_call_history);
		MPIU_GetEnvBool(L"DeinoMPI_PRINT_HISTORY_ON_ABORT", &print_call_history_on_abort);

		pmi_errno = PMI_Get_rank(&pg_rank);
		if (pmi_errno != PMI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_rank", L"**pmi_get_rank %d %s", pmi_errno, PMI_Get_last_error_msgW());
		}

		pmi_errno = PMI_Get_size(&pg_size);
		if (pmi_errno != 0)
		{
			MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_size", L"**pmi_get_size %d %s", pmi_errno, PMI_Get_last_error_msgW());
		}
		
		pmi_errno = PMI_Get_appnum(&appnum);
		if (pmi_errno != PMI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_appnum", L"**pmi_get_appnum %d %s", pmi_errno, PMI_Get_last_error_msgW());
		}

		/* Note that if pmi is not availble, the value of MPI_APPNUM is 
		   not set */
		if (appnum != -1)
		{
			MPIR_Process.attrs.appnum = appnum;
		}
		
		/* Now, initialize the process group information with PMI calls */
		/*
		 * Get the process group id
		 */
		pmi_errno = PMI_Get_id_length_max(&pg_id_sz);
		if (pmi_errno != PMI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_id_length_max", L"**pmi_get_id_length_max %d %s", pmi_errno, PMI_Get_last_error_msgW());
		}

		/* This memory will be freed by the PG_Destroy if there is an error */
		pg_id = MPIU_Malloc((pg_id_sz + 1) * sizeof(wchar_t));
		if (pg_id == NULL)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomem");
		}
	
		pmi_errno = PMI_Get_idW(pg_id, pg_id_sz);
		if (pmi_errno != PMI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_id", L"**pmi_get_id %d %s", pmi_errno, PMI_Get_last_error_msgW());
		}
	}
	else
	{
		/* Create a default pg id */
		pg_id = MPIU_Malloc(2*sizeof(wchar_t));
		if (pg_id == NULL)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomem");
		}
		MPIU_Wcsncpy( pg_id, L"0", 2 );
	}

	/*
	 * Initialize the process group tracking subsystem
	 */
	mpi_errno = MPIDI_PG_Init(argc, argv, MPIDI_CH3I_PG_Compare_ids, MPIDI_CH3I_PG_Destroy);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**dev|pg_init");
	}

	/*
	 * Create a new structure to track the process group for our MPI_COMM_WORLD
	 */
	mpi_errno = MPIDI_PG_Create(pg_size, pg_id, &pg);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**dev|pg_create");
	}

	/* FIXME: We can allow the channels to tell the PG how to get
	   connection information by passing the pg to the channel init routine */
	if (usePMI)
	{
		/* Tell the process group how to get connection information */
		mpi_errno = MPIDI_PG_InitConnKVS( pg );
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* FIXME: Who is this for and where does it belong? */
#ifdef USE_MPIU_DBG_PRINT_VC
	MPIU_DBG_parent_str = (*has_parent) ? L"+" : L"";
#endif

	/* FIXME: has_args and has_env need to come from PMI eventually... */
	*has_args = TRUE;
	*has_env  = TRUE;

	*pg_p	  = pg;
	*pg_rank_p = pg_rank;
	
 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_INITPG);
	return mpi_errno;
 fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	if (pg)
	{
		MPIDI_PG_Destroy( pg );
	}
	/* --END ERROR HANDLING-- */
	goto fn_exit;
}

/*
 * Create the storage for the business card. 
 *
 * The routine MPIDI_CH3I_BCFree should be called with the original 
 * value *bc_val_p .  Note that the routines that set the value 
 * of the businesscard return a pointer to the first free location,
 * so you need to remember the original location in order to free 
 * it later.
 *
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BCInit
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BCInit( wchar_t **bc_val_p, int *val_max_sz_p )
{
	int pmi_errno;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BCINIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BCINIT);

	pmi_errno = PMI_KVS_Get_value_length_max(val_max_sz_p);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_kvs_get_value_length_max", L"**pmi_kvs_get_value_length_max %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	/* This memroy is returned by this routine */
	*bc_val_p = MPIU_Malloc((*val_max_sz_p) * sizeof(wchar_t));
	if (*bc_val_p == NULL)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomem");
	}
	
	/* Add a null to simplify looking at the bc */
	**bc_val_p = 0;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BCINIT);
	return mpi_errno;

  fn_fail:
	goto fn_exit;
}

/* Free the business card.  This routine should be called once the business
   card is published. */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BCFree
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BCFree( wchar_t *bc_val )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BCFREE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BCFREE);

	if (bc_val)
	{
		MPIU_Free( bc_val );
	}
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BCFREE);
	return 0;
}

/* FIXME: The PG code should supply these, since it knows how the 
   pg_ids and other data are represented */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_PG_Compare_ids
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_PG_Compare_ids(const void * id1, const void * id2)
{
	int result;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PG_COMPARE_IDS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PG_COMPARE_IDS);
	result = (wcscmp((wchar_t *) id1, (wchar_t *) id2) == 0) ? TRUE : FALSE;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PG_COMPARE_IDS);
	return result;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_PG_Destroy
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_PG_Destroy(MPIDI_PG_t * pg)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PG_DESTROY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PG_DESTROY);
	if (pg->id != NULL)
	{ 
		MPIU_Free(pg->id);
		pg->id = NULL;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PG_DESTROY);
	return MPI_SUCCESS;
}
